Package: wizaRdry
Title: A Magical Framework for Collaborative & Reproducible Data
        Analysis
Version: 0.6.16
Authors@R: c(
    person("Joshua G.", "Kenney", email = "joshua.kenney@yale.edu", role = c("aut", "cre")),
    person("Trevor F.", "Williams", email = "trevormsu@gmail.com", role = "aut"),
    person("Minerva K.", "Pappu", email = "minerva.pappu@yale.edu", role = "aut"),
    person("Michael J.", "Spilka", email = "michael.j.spilka@gmail.com", role = "aut"),
    person("Danielle N. Pratt", email = "danielle.pratt33@gmail.com", role = "ctb"),
    person("Victor J.", "Pokorny", email = "victor.pokorny@northwestern.edu", role = "ctb"),
    person("Santiago", "Castiello de Obeso", email = "santiago.castiellodeobeso@yale.edu", role = "ctb"),
    person("Praveen", "Suthaharan", email = "praveen.suthaharan@yale.edu", role = "ctb"),
    person("Christian R.", "Horgan", email = "christian.horgan@yale.edu", role = "ctb")
  )
Description: A comprehensive data analysis framework for NIH-funded research that streamlines workflows for both data cleaning and preparing NIH Data Archive ('NDA') submission templates. Provides unified access to multiple data sources ('REDCap', 'MongoDB', 'Qualtrics', 'SQL', 'ORACLE') through interfaces to their APIs, with specialized functions for data cleaning, filtering, merging, and parsing. Features automatic validation, field harmonization, and memory-aware processing to enhance reproducibility in multi-site collaborative research as described in Mittal et al. (2021) <doi:10.20900/jpbs.20210011>.
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.2
Imports: beepr, cli, config, dplyr, future, future.apply, haven, httr,
        jsonlite, knitr, mongolite, parallel, qualtRics, REDCapR,
        rlang, stringdist, testthat, rstudioapi, lubridate, DBI,
        RMariaDB, odbc, R6, openxlsx, openxlsx2
Depends: R (>= 4.1.0)
URL: https://github.com/belieflab/wizaRdry
BugReports: https://github.com/belieflab/wizaRdry/issues
Suggests: rmarkdown, yaml
NeedsCompilation: no
Packaged: 2026-06-10 15:22:09 UTC; Joshua
Author: Joshua G. Kenney [aut, cre],
  Trevor F. Williams [aut],
  Minerva K. Pappu [aut],
  Michael J. Spilka [aut],
  Danielle N. Pratt [ctb],
  Victor J. Pokorny [ctb],
  Santiago Castiello de Obeso [ctb],
  Praveen Suthaharan [ctb],
  Christian R. Horgan [ctb]
Maintainer: Joshua G. Kenney <joshua.kenney@yale.edu>
Repository: CRAN
Date/Publication: 2026-06-10 16:40:02 UTC
Built: R 4.5.2; ; 2026-06-10 20:45:32 UTC; unix

Package: wizaRdry
Title: A Magical Framework for Collaborative & Reproducible Data
        Analysis
Version: 0.6.7
Authors@R: c(
    person("Joshua G.", "Kenney", email = "joshua.kenney@yale.edu", role = c("aut", "cre")),
    person("Trevor F.", "Williams", email = "trevormsu@gmail.com", role = "aut"),
    person("Minerva K.", "Pappu", email = "minerva.pappu@yale.edu", role = "aut"),
    person("Michael J.", "Spilka", email = "michael.j.spilka@gmail.com", role = "aut"),
    person("Danielle N. Pratt", email = "danielle.pratt33@gmail.com", role = "ctb"),
    person("Victor J.", "Pokorny", email = "victor.pokorny@northwestern.edu", role = "ctb"),
    person("Santiago", "Castiello de Obeso", email = "santiago.castiellodeobeso@yale.edu", role = "ctb"),
    person("Praveen", "Suthaharan", email = "praveen.suthaharan@yale.edu", role = "ctb"),
    person("Christian R.", "Horgan", email = "christian.horgan@yale.edu", role = "ctb")
  )
Description: A comprehensive data analysis framework for NIH-funded research that streamlines workflows for both data cleaning and preparing NIH Data Archive ('NDA') submission templates. Provides unified access to multiple data sources ('REDCap', 'MongoDB', 'Qualtrics', 'SQL', 'ORACLE') through interfaces to their APIs, with specialized functions for data cleaning, filtering, merging, and parsing. Features automatic validation, field harmonization, and memory-aware processing to enhance reproducibility in multi-site collaborative research as described in Mittal et al. (2021) <doi:10.20900/jpbs.20210011>.
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.2
Imports: beepr, cli, config, dplyr, future, future.apply, haven, httr,
        jsonlite, knitr, mongolite, parallel, qualtRics, REDCapR,
        rlang, stringdist, testthat, rstudioapi, lubridate, DBI,
        RMariaDB, odbc, R6, openxlsx, openxlsx2
Depends: R (>= 4.1.0)
URL: https://github.com/belieflab/wizaRdry
BugReports: https://github.com/belieflab/wizaRdry/issues
Suggests: rmarkdown, yaml
NeedsCompilation: no
Packaged: 2026-04-13 23:30:28 UTC; Joshua
Author: Joshua G. Kenney [aut, cre],
  Trevor F. Williams [aut],
  Minerva K. Pappu [aut],
  Michael J. Spilka [aut],
  Danielle N. Pratt [ctb],
  Victor J. Pokorny [ctb],
  Santiago Castiello de Obeso [ctb],
  Praveen Suthaharan [ctb],
  Christian R. Horgan [ctb]
Maintainer: Joshua G. Kenney <joshua.kenney@yale.edu>
Repository: CRAN
Date/Publication: 2026-04-14 01:00:02 UTC

# Generated by roxygen2: do not edit by hand

export(clean)
export(createCsv)
export(createRds)
export(createSpss)
export(dataFilter)
export(dataMerge)
export(dataRequest)
export(getRedcap)
export(getSurvey)
export(getTask)
export(meld)
export(mongo)
export(mongo.index)
export(mongo.rune)
export(nda)
export(ndaRequest)
export(oracle)
export(oracle.desc)
export(oracle.index)
export(oracle.query)
export(oracle.test)
export(qualtrics)
export(qualtrics.dict)
export(qualtrics.index)
export(qualtrics.rune)
export(redcap)
export(redcap.dict)
export(redcap.index)
export(redcap.rune)
export(rune)
export(scry)
export(sift)
export(sql)
export(sql.desc)
export(sql.index)
export(sql.query)
export(to.csv)
export(to.nda)
export(to.rds)
export(to.sav)
import(dplyr)
import(haven)
import(lubridate)
import(qualtRics)
importFrom(R6,R6Class)
importFrom(REDCapR,redcap_instruments)
importFrom(REDCapR,redcap_metadata_read)
importFrom(config,get)
importFrom(dplyr,"%>%")
importFrom(dplyr,bind_rows)
importFrom(dplyr,contains)
importFrom(dplyr,filter)
importFrom(dplyr,first)
importFrom(dplyr,group_by)
importFrom(dplyr,mutate)
importFrom(dplyr,select)
importFrom(dplyr,setdiff)
importFrom(future,future)
importFrom(future,multisession)
importFrom(future,plan)
importFrom(future.apply,future_lapply)
importFrom(jsonlite,toJSON)
importFrom(knitr,kable)
importFrom(lubridate,parse_date_time)
importFrom(mongolite,mongo)
importFrom(mongolite,ssl_options)
importFrom(odbc,dbConnect)
importFrom(odbc,dbDisconnect)
importFrom(odbc,dbGetInfo)
importFrom(odbc,dbGetQuery)
importFrom(odbc,dbListFields)
importFrom(odbc,dbListTables)
importFrom(parallel,detectCores)
importFrom(rlang,.data)
importFrom(stats,setNames)
importFrom(stringdist,stringdist)
importFrom(testthat,expect_true)
importFrom(testthat,test_that)
importFrom(utils,capture.output)
importFrom(utils,flush.console)
importFrom(utils,getFromNamespace)
importFrom(utils,head)
importFrom(utils,install.packages)
importFrom(utils,setTxtProgressBar)
importFrom(utils,txtProgressBar)

#' Fetch data from SQL database to be stored in a data frame
#'
#' Retrieves data from a SQL table and optionally joins it with a primary keys table
#' as specified in the configuration.
#'
#' @param table_name Name of the SQL table or view to query
#' @param ... Optional column names to filter for. Only rows with non-missing values
#'        in ALL specified columns will be returned.
#' @param fields Optional vector of specific fields to select
#' @param where_clause Optional WHERE clause to filter results (without the "WHERE" keyword)
#' @param join_primary_keys Boolean, whether to join with the primary keys table (default: TRUE)
#' @param custom_query Optional custom SQL query to execute instead of building one
#' @param max_rows Optional limit on the number of rows to return
#' @param date_format Optional format for date fields (default uses ISO format)
#' @param batch_size Number of records to retrieve per batch for large datasets
#' @param pii Logical; if FALSE (default), remove fields marked as PII. TRUE keeps PII.
#' @param interview_date Optional; can be either:
#'        - A date string in various formats (ISO, US, etc.) to filter data up to that date
#'        - A boolean TRUE to return only rows with non-NA interview_date values
#' @param all Logical; if TRUE, use LEFT OUTER JOIN instead of INNER JOIN (default: FALSE),
#'        similar to the 'all' parameter in base R's merge() function
#'
#' @return A data frame containing the requested SQL data
#' @export
#' @examples
#' \dontrun{
#' # Get data from a specific table
#' data <- sql("participants")
#'
#' # Get data with a where clause
#' survey_data <- sql("vw_surveyquestionresults",
#'                   where_clause = "resultidentifier = 'NRS'")
#'
#' # Get all records, including those without matching primary key
#' all_data <- sql("candidate", all = TRUE)
#' }
sql <- function(table_name = NULL, ..., fields = NULL, where_clause = NULL,
                join_primary_keys = TRUE, custom_query = NULL, max_rows = NULL,
                date_format = NULL, batch_size = 1000, pii = FALSE,
                interview_date = NULL, all = FALSE) {
  # Check if required packages are available
  if (!requireNamespace("RMariaDB", quietly = TRUE)) {
    stop("Package 'RMariaDB' is needed for this function to work. Please install it.",
         call. = FALSE)
  }
  if (!requireNamespace("DBI", quietly = TRUE)) {
    stop("Package 'DBI' is needed for this function to work. Please install it.",
         call. = FALSE)
  }

  start_time <- Sys.time()

  # Validate secrets and config - MUST COME FIRST
  validate_secrets("sql")
  config <- validate_config("sql")

  # Initialize loading animation
  pb <- initializeLoadingAnimation(20)

  # Validate parameters
  if (is.null(table_name) && is.null(custom_query)) {
    tables_info <- sql.index()
    if (is.null(tables_info)) {
      stop("No table name or custom query provided, and could not retrieve table list")
    }
    stop("No table name or custom query provided. Use sql.index() to see available tables.")
  }

  # Get connection parameters using get_secret
  host_value <- get_secret("host")
  user_id <- get_secret("uid")
  password <- get_secret("pwd")

  # Parse the host value to extract database name and port if present
  # Default values
  db_name <- if (!is.null(config) && !is.null(config$sql$database)) {
    config$sql$database  # Use configured database name
  } else {
    "mysql"  # Fallback default database
  }
  port_number <- 3306  # Default MySQL/MariaDB port

  # Extract host, port, and database from conn if it's a connection string
  if (grepl("Database=|dbname=", host_value)) {
    # It's a connection string, extract components
    if (grepl("Database=", host_value)) {
      db_name <- gsub(".*Database=([^;]+).*", "\\1", host_value)
    } else if (grepl("dbname=", host_value)) {
      db_name <- gsub(".*dbname=([^;]+).*", "\\1", host_value)
    }
    if (grepl("Port=", host_value)) {
      port_number <- as.integer(gsub(".*Port=([^;]+).*", "\\1", host_value))
    }
    if (grepl("Server=", host_value)) {
      host_value <- gsub(".*Server=([^;]+).*", "\\1", host_value)
    } else if (grepl("host=", host_value)) {
      host_value <- gsub(".*host=([^;]+).*", "\\1", host_value)
    }
  }

  # Display loading message
  if (!is.null(custom_query)) {
    message(sprintf("\nExecuting custom SQL query..."))
  } else {
    message(sprintf("\nRetrieving data from SQL table: %s%s",
                    table_name,
                    ifelse(!is.null(where_clause), sprintf(" (with filters)"), "")))
  }

  # Update progress bar
  for (i in 1:10) {
    updateLoadingAnimation(pb, i)
    Sys.sleep(0.05)
  }

  # Define the primary key information
  primary_key_column <- NULL
  superkey_table <- NULL
  if (join_primary_keys) {
    if (!is.null(config$sql$primary_key)) {
      primary_key_column <- config$sql$primary_key
    } else {
      message("Warning: Primary key not specified in config. Using default 'PARTICIPANTIDENTIFIER'")
      primary_key_column <- "PARTICIPANTIDENTIFIER"
    }
    if (!is.null(config$sql$superkey)) {
      superkey_table <- config$sql$superkey
    } else {
      message("Warning: Superkey table not specified in config. Join with primary keys disabled.")
      join_primary_keys <- FALSE
    }
  }

  # Determine if the requested table is the configured superkey table
  is_superkey_request <- FALSE
  if (!is.null(superkey_table) && !is.null(table_name)) {
    # Extract table name without schema for comparison
    table_name_only <- if (grepl("\\.", table_name)) {
      strsplit(table_name, "\\.")[[1]][2]
    } else {
      table_name
    }
    
    # Extract superkey table name without schema for comparison
    superkey_name_only <- if (grepl("\\.", superkey_table)) {
      strsplit(superkey_table, "\\.")[[1]][2]
    } else {
      superkey_table
    }
    
    # Check if the requested table matches the superkey table
    is_superkey_request <- identical(trimws(toupper(table_name_only)), trimws(toupper(superkey_name_only)))
    
    if (is_superkey_request) {
      message("Requested table matches configured superkey; returning without joins.")
      join_primary_keys <- FALSE
    }
  }

  # Determine fields to exclude if PII exclusion is enabled
  pii_fields <- character(0)
  if (!pii && !is.null(config$sql$pii_fields)) {
    pii_fields <- config$sql$pii_fields
    if (length(pii_fields) > 0) {
      message(sprintf("Will exclude %d PII fields: %s",
                      length(pii_fields),
                      paste(pii_fields, collapse = ", ")))
    }
  }

  # Establish database connection using RMariaDB
  tryCatch({
    # Connect to MariaDB
    db_conn <- DBI::dbConnect(RMariaDB::MariaDB(),
                              host = host_value,
                              user = user_id,
                              password = password,
                              port = port_number,
                              dbname = db_name)
    message("Database connection established successfully.")
  }, error = function(e) {
    stop(paste("Error connecting to database:", e$message))
  })

  # Update progress bar
  for (i in 11:15) {
    updateLoadingAnimation(pb, i)
    Sys.sleep(0.05)
  }

  # Check if we need to add a schema to the table name
  if (!is.null(table_name) && !grepl("\\.", table_name) && !is.null(config$sql$database)) {
    # Use the configured database as schema context
    schema <- config$sql$database
    
    # Build a query to check if the table exists in this schema
    check_query <- sprintf(
      "SELECT 1 FROM information_schema.tables WHERE table_schema = '%s' AND table_name = '%s' LIMIT 1",
      schema, table_name
    )

    # Execute the query
    check_result <- DBI::dbGetQuery(db_conn, check_query)

    # If table exists in this schema, use it
    if (nrow(check_result) > 0) {
      message("Found table '", table_name, "' in schema '", schema, "'")
      table_name <- paste0(schema, ".", table_name)
    } else {
      # If table not found in the configured schema, show an informative error
      stop(sprintf(
        "Table '%s' not found in configured database '%s'. Please check your config.yml or specify explicitly with '%s.%s'.",
        table_name, schema, schema, table_name
      ))
    }
  }

  # Build or execute SQL query
  result_data <- NULL
  tryCatch({
    if (!is.null(custom_query)) {
      # Execute custom query directly
      result_data <- DBI::dbGetQuery(db_conn, custom_query)

      # If PII exclusion is enabled, filter out PII columns from the result
      if (!pii && !is.null(pii_fields) && length(pii_fields) > 0) {
        pii_cols_in_result <- intersect(names(result_data), pii_fields)
        pii_cols_missing <- setdiff(pii_fields, names(result_data))
        if (length(pii_cols_in_result) > 0) {
          message("Removing PII columns from result: ", paste(pii_cols_in_result, collapse = ", "))
          result_data <- result_data[, !(names(result_data) %in% pii_fields), drop = FALSE]
        }
        if (length(pii_cols_missing) > 0) {
          message("Configured PII columns not present in result (skipped): ", paste(pii_cols_missing, collapse = ", "))
        }
      }
    } else {
      # For PII exclusion with SELECT *, we need to get all column names first
      # with this updated version:

      # For PII exclusion with SELECT *, we need to get all column names first
      if (!pii && !is.null(pii_fields) && length(pii_fields) > 0 && is.null(fields)) {
        # Check if we're joining tables
        if (!is.null(superkey_table) && !is.null(primary_key_column)) {
          # Get column names from both tables
          table_parts <- strsplit(table_name, "\\.")[[1]]
          main_schema <- NULL
          main_table_only <- table_name
          if (length(table_parts) > 1) {
            main_schema <- table_parts[1]
            main_table_only <- table_parts[2]
          }

          # Extract schema and table name for superkey table
          superkey_parts <- strsplit(superkey_table, "\\.")[[1]]
          superkey_schema <- NULL
          superkey_table_only <- superkey_table
          if (length(superkey_parts) > 1) {
            superkey_schema <- superkey_parts[1]
            superkey_table_only <- superkey_parts[2]
          }

          # Get columns from main table
          main_col_query <- sprintf("SELECT column_name FROM information_schema.columns WHERE table_name = '%s'", main_table_only)
          if (!is.null(main_schema)) {
            main_col_query <- sprintf("%s AND table_schema = '%s'", main_col_query, main_schema)
          }

          # Get columns from superkey table
          superkey_col_query <- sprintf("SELECT column_name FROM information_schema.columns WHERE table_name = '%s'", superkey_table_only)
          if (!is.null(superkey_schema)) {
            superkey_col_query <- sprintf("%s AND table_schema = '%s'", superkey_col_query, superkey_schema)
          }

          message("Fetching column lists from both tables to exclude PII fields...")
          main_columns <- DBI::dbGetQuery(db_conn, main_col_query)
          superkey_columns <- DBI::dbGetQuery(db_conn, superkey_col_query)

          if (nrow(main_columns) > 0 && nrow(superkey_columns) > 0) {
            # Filter out PII fields from both tables
            main_valid_columns <- setdiff(main_columns$column_name, pii_fields)
            superkey_valid_columns <- setdiff(superkey_columns$column_name, pii_fields)

            # Build column list with table aliases
            main_column_list <- paste0("t.", main_valid_columns, collapse = ", ")
            superkey_column_list <- paste0("pk.", superkey_valid_columns, collapse = ", ")
            column_list <- paste(main_column_list, superkey_column_list, sep = ", ")

            # Parse table name parts for schema handling
            table_parts <- strsplit(table_name, "\\.")[[1]]
            if (length(table_parts) > 1) {
              schema <- table_parts[1]
              # Apply schema to superkey_table if needed
              if (!grepl("\\.", superkey_table)) {
                superkey_table <- paste0(schema, ".", superkey_table)
              }
            }

            # Determine join type based on 'all' parameter
            join_type <- ifelse(all, "LEFT OUTER JOIN", "INNER JOIN")
            query <- sprintf(
              "SELECT DISTINCT %s FROM %s t %s %s pk ON t.%s = pk.%s %s %s",
              column_list,
              table_name,
              join_type,
              superkey_table,
              primary_key_column, primary_key_column,
              ifelse(!is.null(where_clause) && nchar(where_clause) > 0, paste("WHERE", where_clause), ""),
              ifelse(!is.null(max_rows) && is.numeric(max_rows) && max_rows > 0, paste("LIMIT", as.integer(max_rows)), "")
            )
            message("Generated SQL query with PII exclusion for joined tables: ", query)
          } else {
            # Fall back to standard query building if column fetch fails
            query <- build_sql_query(
              table_name = table_name,
              fields = fields,
              where_clause = where_clause,
              superkey_table = if (join_primary_keys) superkey_table else NULL,
              primary_key_column = if (join_primary_keys) primary_key_column else NULL,
              max_rows = max_rows,
              pii_fields = NULL,  # Don't filter in query, will filter result after
              all = all
            )
          }
        } else {
          # Single table case (existing code)
          table_parts <- strsplit(table_name, "\\.")[[1]]
          schema_name <- NULL
          table_only <- table_name
          if (length(table_parts) > 1) {
            schema_name <- table_parts[1]
            table_only <- table_parts[2]
          }

          # Get all column names from the table
          col_query <- sprintf("SELECT column_name FROM information_schema.columns WHERE table_name = '%s'", table_only)
          # If table_name includes schema, add schema filter
          if (!is.null(schema_name)) {
            col_query <- sprintf("%s AND table_schema = '%s'", col_query, schema_name)
          }
          message("Fetching column list to exclude PII fields...")
          all_columns <- DBI::dbGetQuery(db_conn, col_query)
          if (nrow(all_columns) > 0) {
            # Filter out PII fields
            valid_columns <- setdiff(all_columns$column_name, pii_fields)
            column_list <- paste(valid_columns, collapse = ", ")
            query <- sprintf(
              "SELECT DISTINCT %s FROM %s %s %s",
              column_list,
              table_name,
              ifelse(!is.null(where_clause) && nchar(where_clause) > 0, paste("WHERE", where_clause), ""),
              ifelse(!is.null(max_rows) && is.numeric(max_rows) && max_rows > 0, paste("LIMIT", as.integer(max_rows)), "")
            )
            message("Generated SQL query with PII exclusion: ", query)
          } else {
            # Fall back to standard query building if column fetch fails
            query <- build_sql_query(
              table_name = table_name,
              fields = fields,
              where_clause = where_clause,
              superkey_table = if (join_primary_keys) superkey_table else NULL,
              primary_key_column = if (join_primary_keys) primary_key_column else NULL,
              max_rows = max_rows,
              pii_fields = NULL,  # Don't filter in query, will filter result after
              all = all
            )
          }
        }
      } else {
        # Standard query building
        query <- build_sql_query(
          table_name = table_name,
          fields = fields,
          where_clause = where_clause,
          superkey_table = if (join_primary_keys) superkey_table else NULL,
          primary_key_column = if (join_primary_keys) primary_key_column else NULL,
          max_rows = max_rows,
          pii_fields = if (!pii) pii_fields else NULL,
          all = all
        )
      }

      # Execute the query
      result_data <- DBI::dbGetQuery(db_conn, query)

      # Double-check PII exclusion on the result
      if (!pii && !is.null(pii_fields) && length(pii_fields) > 0) {
        pii_cols_in_result <- intersect(names(result_data), pii_fields)
        pii_cols_missing <- setdiff(pii_fields, names(result_data))
        if (length(pii_cols_in_result) > 0) {
          message("Removing PII columns from result: ", paste(pii_cols_in_result, collapse = ", "))
          result_data <- result_data[, !(names(result_data) %in% pii_fields), drop = FALSE]
        }
        if (length(pii_cols_missing) > 0) {
          message("Configured PII columns not present in result (skipped): ", paste(pii_cols_missing, collapse = ", "))
        }
      }
    }
  }, error = function(e) {
    DBI::dbDisconnect(db_conn)
    stop(paste("Error executing SQL query:", e$message))
  })

  # Update progress bar
  for (i in 16:20) {
    updateLoadingAnimation(pb, i)
    Sys.sleep(0.05)
  }

  completeLoadingAnimation(pb)

  # Close the database connection
  DBI::dbDisconnect(db_conn)

  # Process and clean the data
  if (nrow(result_data) > 0) {
    # Process date fields
    if (!is.null(date_format)) {
      result_data <- process_date_fields(result_data, date_format)
    }

    # Handle interview_date filtering if specified
    if (!is.null(interview_date) && "interview_date" %in% names(result_data)) {
      result_data <- filter_by_interview_date(result_data, interview_date)
    }

    # Handle missing values using the configuration
    result_data <- handle_missing_values(result_data, config)

    # Standardize column names for key fields (similar to redcap function)
    age_cols <- grep("_interview_age$", base::names(result_data))
    if (length(age_cols) > 0) {
      base::names(result_data)[age_cols] <- "interview_age"
    }

    date_patterns <- c("_interview_date$", "interview_date")
    date_cols <- NULL
    for (pattern in date_patterns) {
      found_cols <- grep(pattern, base::names(result_data), ignore.case = TRUE)
      if (length(found_cols) > 0) {
        date_cols <- found_cols
        break  # Stop at first pattern that finds matches
      }
    }
    if (!is.null(date_cols) && length(date_cols) > 0) {
      base::names(result_data)[date_cols] <- "interview_date"
    }

    # Process column filtering based on ... parameters
    dots_args <- list(...)
    if (length(dots_args) > 0) {
      result_data <- filter_by_column_values(result_data, dots_args)
    }
  }

  # Add metadata to the result
  attr(result_data, "sql_table") <- table_name
  if (!is.null(where_clause)) {
    attr(result_data, "where_clause") <- where_clause
  }

  # Show duration
  end_time <- Sys.time()
  duration <- difftime(end_time, start_time, units = "secs")
  message(sprintf("\nData frame retrieved from SQL in %s with %d rows and %d columns.",
                  formatDuration(duration),
                  nrow(result_data),
                  ncol(result_data)))

  return(result_data)
}

#' Get a list of tables from the SQL database
#'
#' @param schema Optional schema name to filter tables
#' @return A data frame with table information
#' @export
sql.index <- function(schema = NULL) {
  # Check if required packages are available
  if (!requireNamespace("RMariaDB", quietly = TRUE)) {
    stop("Package 'RMariaDB' is needed for this function to work. Please install it.",
         call. = FALSE)
  }
  if (!requireNamespace("DBI", quietly = TRUE)) {
    stop("Package 'DBI' is needed for this function to work. Please install it.",
         call. = FALSE)
  }
  if (!requireNamespace("knitr", quietly = TRUE)) {
    stop("Package 'knitr' is needed for this function to work. Please install it.",
         call. = FALSE)
  }

  # Validate secrets and config
  validate_secrets("sql")
  config <- validate_config("sql")

  # Get connection parameters using get_secret
  host_value <- get_secret("host")
  user_id <- get_secret("uid")
  password <- get_secret("pwd")

  # Parse the host value to extract database name and port if present
  # Default values
  db_name <- if (!is.null(config) && !is.null(config$sql$schema)) {
    config$sql$schema  # Use first configured schema
  } else {
    "mysql"  # Fallback default database
  }
  port_number <- 3306  # Default MySQL/MariaDB port

  # Extract host, port, and database from conn if it's a connection string
  if (grepl("Database=|dbname=", host_value)) {
    # It's a connection string, extract components
    if (grepl("Database=", host_value)) {
      db_name <- gsub(".*Database=([^;]+).*", "\\1", host_value)
    } else if (grepl("dbname=", host_value)) {
      db_name <- gsub(".*dbname=([^;]+).*", "\\1", host_value)
    }

    if (grepl("Port=", host_value)) {
      port_number <- as.integer(gsub(".*Port=([^;]+).*", "\\1", host_value))
    }

    if (grepl("Server=", host_value)) {
      host_value <- gsub(".*Server=([^;]+).*", "\\1", host_value)
    } else if (grepl("host=", host_value)) {
      host_value <- gsub(".*host=([^;]+).*", "\\1", host_value)
    }
  }

  # Connect to database
  db_conn <- NULL
  tables <- NULL

  tryCatch({
    # Connect to MariaDB
    db_conn <- DBI::dbConnect(RMariaDB::MariaDB(),
                              host = host_value,
                              user = user_id,
                              password = password,
                              port = port_number,
                              dbname = db_name)

    # Get tables
    if (is.null(schema)) {
      # Query for all tables
      tables <- DBI::dbGetQuery(db_conn,
                                "SELECT table_schema AS 'Schema',
                               table_name AS 'Table',
                               table_type AS 'Type'
                               FROM information_schema.tables
                               ORDER BY table_schema, table_name")
    } else {
      # Query for tables in the specified schema
      tables <- DBI::dbGetQuery(db_conn,
                                sprintf("SELECT table_schema AS 'Schema',
                                      table_name AS 'Table',
                                      table_type AS 'Type'
                                      FROM information_schema.tables
                                      WHERE table_schema = '%s'
                                      ORDER BY table_schema, table_name",
                                        schema))
    }
  }, error = function(e) {
    message(paste("Error retrieving tables:", e$message))
    return(NULL)
  }, finally = {
    # Close connection if it was opened
    if (!is.null(db_conn)) {
      DBI::dbDisconnect(db_conn)
    }
  })

  # Format the result
  if (!is.null(tables) && nrow(tables) > 0) {
    return(knitr::kable(tables, format = "simple"))
  } else {
    message("No tables found")
    return(NULL)
  }
}

#' Get SQL table columns/metadata
#'
#' @param table_name Name of the table to get metadata for
#' @return A data frame with column information
#' @export
sql.desc <- function(table_name) {
  # Check if required packages are available
  if (!requireNamespace("RMariaDB", quietly = TRUE)) {
    stop("Package 'RMariaDB' is needed for this function to work. Please install it.",
         call. = FALSE)
  }
  if (!requireNamespace("DBI", quietly = TRUE)) {
    stop("Package 'DBI' is needed for this function to work. Please install it.",
         call. = FALSE)
  }
  if (!requireNamespace("knitr", quietly = TRUE)) {
    stop("Package 'knitr' is needed for this function to work. Please install it.",
         call. = FALSE)
  }

  if (is.null(table_name)) {
    stop("Table name is required")
  }

  # Validate secrets
  validate_secrets("sql")
  config <- validate_config("sql")

  # Get connection parameters using get_secret
  host_value <- get_secret("host")
  user_id <- get_secret("uid")
  password <- get_secret("pwd")

  # Parse table_name to extract schema and table
  parts <- strsplit(table_name, "\\.")[[1]]
  schema_name <- NULL
  table_only <- table_name

  if (length(parts) > 1) {
    # Explicit schema.table notation
    schema_name <- parts[1]
    table_only <- parts[2]
  } else {
    # No schema specified, use the first configured schema
    if (!is.null(config) && !is.null(config$sql$schema)) {
      schema_name <- config$sql$schema
      message(sprintf("No schema specified. Using default schema: %s", schema_name))
    }
  }

  # Parse the host value to extract database name and port if present
  # Default values
  db_name <- if (!is.null(config) && !is.null(config$sql$schema)) {
    config$sql$schema  # Use first configured schema
  } else {
    "mysql"  # Fallback default database
  }
  port_number <- 3306  # Default MySQL/MariaDB port

  # Extract host, port, and database from conn if it's a connection string
  if (grepl("Database=|dbname=", host_value)) {
    # It's a connection string, extract components
    if (grepl("Database=", host_value)) {
      db_name <- gsub(".*Database=([^;]+).*", "\\1", host_value)
    } else if (grepl("dbname=", host_value)) {
      db_name <- gsub(".*dbname=([^;]+).*", "\\1", host_value)
    }
    if (grepl("Port=", host_value)) {
      port_number <- as.integer(gsub(".*Port=([^;]+).*", "\\1", host_value))
    }
    if (grepl("Server=", host_value)) {
      host_value <- gsub(".*Server=([^;]+).*", "\\1", host_value)
    } else if (grepl("host=", host_value)) {
      host_value <- gsub(".*host=([^;]+).*", "\\1", host_value)
    }
  }

  # Connect to database
  db_conn <- NULL
  columns <- NULL

  tryCatch({
    # Connect to MariaDB
    db_conn <- DBI::dbConnect(RMariaDB::MariaDB(),
                              host = host_value,
                              user = user_id,
                              password = password,
                              port = port_number,
                              dbname = db_name)

    # Query for column information
    if (!is.null(schema_name)) {
      # Use the specified schema
      columns <- DBI::dbGetQuery(db_conn,
                                 sprintf("SELECT column_name AS 'Column',
                                       data_type AS 'Type',
                                       character_maximum_length AS 'Size',
                                       is_nullable AS 'Nullable',
                                       column_comment AS 'Description'
                                       FROM information_schema.columns
                                       WHERE table_schema = '%s' AND table_name = '%s'
                                       ORDER BY ordinal_position",
                                         schema_name, table_only))

      # If no columns found with the specified schema, try to find the table in any schema
      if (nrow(columns) == 0) {
        message(sprintf("No columns found for table '%s' in schema '%s'. Checking all schemas...",
                        table_only, schema_name))

        columns <- DBI::dbGetQuery(db_conn,
                                   sprintf("SELECT table_schema AS 'Schema',
                                          column_name AS 'Column',
                                          data_type AS 'Type',
                                          character_maximum_length AS 'Size',
                                          is_nullable AS 'Nullable',
                                          column_comment AS 'Description'
                                          FROM information_schema.columns
                                          WHERE table_name = '%s'
                                          ORDER BY table_schema, ordinal_position",
                                           table_only))
      }
    } else {
      # Try to find the table in any schema
      columns <- DBI::dbGetQuery(db_conn,
                                 sprintf("SELECT table_schema AS 'Schema',
                                      column_name AS 'Column',
                                      data_type AS 'Type',
                                      character_maximum_length AS 'Size',
                                      is_nullable AS 'Nullable',
                                      column_comment AS 'Description'
                                      FROM information_schema.columns
                                      WHERE table_name = '%s'
                                      ORDER BY table_schema, ordinal_position",
                                         table_only))
    }
  }, error = function(e) {
    if (!is.null(db_conn)) {
      DBI::dbDisconnect(db_conn)
    }
    stop(paste("Error retrieving columns for table", table_name, ":", e$message))
  }, finally = {
    # Close connection if it was opened
    if (!is.null(db_conn)) {
      DBI::dbDisconnect(db_conn)
    }
  })

  # Format the result
  if (!is.null(columns) && nrow(columns) > 0) {
    return(knitr::kable(columns, format = "simple"))
  } else {
    message(paste("No columns found for table", table_name))
    return(NULL)
  }
}

#' Perform a direct SQL query with minimal processing
#'
#' @param query The SQL query to execute
#' @param pii Logical; if FALSE (default), remove fields marked as PII. TRUE keeps PII.
#' @return A data frame with the query results
#' @export
sql.query <- function(query, pii = FALSE) {
  # Check if required packages are available
  if (!requireNamespace("RMariaDB", quietly = TRUE)) {
    stop("Package 'RMariaDB' is needed for this function to work. Please install it.",
         call. = FALSE)
  }
  if (!requireNamespace("DBI", quietly = TRUE)) {
    stop("Package 'DBI' is needed for this function to work. Please install it.",
         call. = FALSE)
  }

  if (is.null(query) || !is.character(query) || length(query) != 1) {
    stop("A valid SQL query string is required")
  }

  # Validate secrets and config
  validate_secrets("sql")
  config <- validate_config("sql")

  # Get connection parameters using get_secret
  host_value <- get_secret("host")
  user_id <- get_secret("uid")
  password <- get_secret("pwd")

  # Parse the host value to extract database name and port if present
  # Default values
  db_name <- if (!is.null(config) && !is.null(config$sql$schema)) {
    config$sql$schema  # Use first configured schema
  } else {
    "mysql"  # Fallback default database
  }
  port_number <- 3306  # Default MySQL/MariaDB port

  # Extract host, port, and database from conn if it's a connection string
  if (grepl("Database=|dbname=", host_value)) {
    # It's a connection string, extract components
    if (grepl("Database=", host_value)) {
      db_name <- gsub(".*Database=([^;]+).*", "\\1", host_value)
    } else if (grepl("dbname=", host_value)) {
      db_name <- gsub(".*dbname=([^;]+).*", "\\1", host_value)
    }

    if (grepl("Port=", host_value)) {
      port_number <- as.integer(gsub(".*Port=([^;]+).*", "\\1", host_value))
    }

    if (grepl("Server=", host_value)) {
      host_value <- gsub(".*Server=([^;]+).*", "\\1", host_value)
    } else if (grepl("host=", host_value)) {
      host_value <- gsub(".*host=([^;]+).*", "\\1", host_value)
    }
  }

  # Get config for PII exclusion if needed
  config <- NULL
  pii_fields <- NULL
  if (!pii) {
    tryCatch({
      config <- validate_config("sql")
      if (!is.null(config$sql$pii_fields)) {
        pii_fields <- config$sql$pii_fields
        if (length(pii_fields) > 0) {
          message(sprintf("Will exclude %d PII fields if present in results: %s",
                          length(pii_fields),
                          paste(pii_fields, collapse=", ")))
        }
      }
    }, error = function(e) {
      warning("Could not load PII field configuration: ", e$message)
    })
  }

  # Connect to database
  db_conn <- NULL
  result <- NULL

  tryCatch({
    # Connect to MariaDB
    db_conn <- DBI::dbConnect(RMariaDB::MariaDB(),
                              host = host_value,
                              user = user_id,
                              password = password,
                              port = port_number,
                              dbname = db_name)

    # Execute query
    result <- DBI::dbGetQuery(db_conn, query)

    # Apply PII exclusion if enabled
    if (!pii && !is.null(pii_fields) && length(pii_fields) > 0) {
      pii_cols_present <- intersect(names(result), pii_fields)
      if (length(pii_cols_present) > 0) {
        message(sprintf("Removing %d PII fields from results: %s",
                        length(pii_cols_present),
                        paste(pii_cols_present, collapse=", ")))
        result <- result[, !names(result) %in% pii_fields, drop = FALSE]
      }
    }

    # Apply missing value handling if configured
    if (!is.null(config) && !is.null(config$missing_data_codes) && nrow(result) > 0) {
      result <- handle_missing_values(result, config)
    }

  }, error = function(e) {
    if (!is.null(db_conn)) {
      DBI::dbDisconnect(db_conn)
    }
    stop(paste("Error executing query:", e$message))
  }, finally = {
    # Close connection if it was opened
    if (!is.null(db_conn)) {
      DBI::dbDisconnect(db_conn)
    }
  })

  return(result)
}

build_sql_query <- function(table_name, fields = NULL, where_clause = NULL,
                            superkey_table = NULL, primary_key_column = NULL,
                            max_rows = NULL, pii_fields = NULL, all = FALSE) {

  # Define table aliases
  main_table_alias <- "t"
  pk_table_alias <- "pk"

  # Determine which fields to select
  if (is.null(fields)) {
    # No specific fields requested
    if (!is.null(superkey_table) && !is.null(primary_key_column)) {
      # When joining, always include fields from both tables
      if (!is.null(pii_fields) && length(pii_fields) > 0) {
        # PII exclusion needed but no specific fields requested
        # This requires getting table metadata first, which we can't do in this function
        message("Note: PII exclusion with * requires listing all non-PII columns explicitly")
        select_clause <- "SELECT DISTINCT t.*, pk.*" # Will be replaced in main function
      } else {
        # No PII exclusion needed - include all fields from both tables
        select_clause <- "SELECT DISTINCT t.*, pk.*"
      }
    } else {
      # No join - just select from main table
      if (!is.null(pii_fields) && length(pii_fields) > 0) {
        message("Note: PII exclusion with * requires listing all non-PII columns explicitly")
        select_clause <- "SELECT DISTINCT *" # Will be replaced in main function
      } else {
        select_clause <- "SELECT DISTINCT *"
      }
    }
  } else {
    # Specific fields requested, exclude PII fields
    if (!is.null(pii_fields)) {
      fields <- setdiff(fields, pii_fields)
    }
    # If fields array is not empty after filtering
    if (length(fields) > 0) {
      # For joins, qualify field names with table alias
      if (!is.null(superkey_table) && !is.null(primary_key_column)) {
        # Check if fields already have table qualifiers
        qualified_fields <- character(0)
        for (field in fields) {
          if (grepl("\\.", field)) {
            # Field already has table qualifier (e.g., "t.sub_id" or "pk.phi_field")
            qualified_fields <- c(qualified_fields, field)
          } else {
            # Add main table alias to field names
            qualified_fields <- c(qualified_fields, paste0(main_table_alias, ".", field))
          }
        }
        select_clause <- paste0("SELECT DISTINCT ", paste(qualified_fields, collapse = ", "))
      } else {
        select_clause <- paste0("SELECT DISTINCT ", paste(fields, collapse = ", "))
      }
    } else {
      # All requested fields were PII, so select a dummy field
      select_clause <- "SELECT DISTINCT 1 AS dummy_column" # Fallback
    }
  }

  # Build the FROM clause
  # Handle schema in table name
  table_parts <- strsplit(table_name, "\\.")[[1]]
  if (!is.null(superkey_table) && !is.null(primary_key_column)) {
    # If table_name includes a schema, apply it to superkey_table if needed
    if (length(table_parts) > 1) {
      schema <- table_parts[1]
      # Apply schema to superkey_table if it doesn't already have one
      if (!grepl("\\.", superkey_table)) {
        superkey_table <- paste0(schema, ".", superkey_table)
      }
    }
    # Determine join type based on 'all' parameter
    join_type <- ifelse(all, "LEFT OUTER JOIN", "INNER JOIN")
    # Create JOIN clause
    from_clause <- sprintf(
      "FROM %s %s %s %s %s ON %s.%s = %s.%s",
      table_name, main_table_alias,
      join_type,
      superkey_table, pk_table_alias,
      main_table_alias, primary_key_column,
      pk_table_alias, primary_key_column
    )
  } else {
    # No join needed
    from_clause <- sprintf("FROM %s %s", table_name, main_table_alias)
  }

  # Add WHERE clause if provided
  where_part <- ""
  if (!is.null(where_clause) && nchar(where_clause) > 0) {
    where_part <- paste("WHERE", where_clause)
  }

  # Add LIMIT clause if max_rows is specified
  limit_part <- ""
  if (!is.null(max_rows) && is.numeric(max_rows) && max_rows > 0) {
    limit_part <- paste("LIMIT", as.integer(max_rows))
  }

  # Combine all parts into the final query
  query <- paste(select_clause, from_clause, where_part, limit_part)

  # For debugging
  message("Generated SQL query: ", query)

  # Add an attribute to indicate if column list expansion is needed for PII exclusion
  if (!is.null(pii_fields) && length(pii_fields) > 0 && is.null(fields)) {
    attr(query, "needs_column_list") <- TRUE
  }

  return(query)
}

#' Process date fields in the dataset
#'
#' @param df The data frame to process
#' @param date_format The desired output date format
#'
#' @return The processed data frame
#' @noRd
process_date_fields <- function(df, date_format = NULL) {
  # If no format specified, leave dates in database format
  if (is.null(date_format)) {
    return(df)
  }

  # Identify potential date columns by name pattern
  date_patterns <- c("date", "dt", "timestamp")
  potential_date_cols <- character(0)

  for (pattern in date_patterns) {
    found_cols <- grep(pattern, names(df), value = TRUE, ignore.case = TRUE)
    potential_date_cols <- c(potential_date_cols, found_cols)
  }

  # Remove duplicates
  potential_date_cols <- unique(potential_date_cols)

  # Process each potential date column
  for (col in potential_date_cols) {
    # Only process if the column exists and has data
    if (col %in% names(df) && !all(is.na(df[[col]]))) {
      # Check if it's already a Date or POSIXct object
      if (inherits(df[[col]], "Date") || inherits(df[[col]], "POSIXct")) {
        # Format according to preference
        if (date_format == "mdy") {
          df[[col]] <- format(df[[col]], "%m/%d/%Y")
        } else if (date_format == "dmy") {
          df[[col]] <- format(df[[col]], "%d/%m/%Y")
        } else if (date_format == "ymd") {
          df[[col]] <- format(df[[col]], "%Y-%m-%d")
        }
      }
      # Check if it's character data that looks like dates
      else if (is.character(df[[col]])) {
        # Try to convert string dates to consistent format
        if (any(grepl("-|/|\\.", df[[col]][!is.na(df[[col]])]))) {
          tryCatch({
            parsed_dates <- parseAnyDate(df[[col]])
            if (!all(is.na(parsed_dates))) {
              # Format according to preference
              if (date_format == "mdy") {
                df[[col]] <- format(parsed_dates, "%m/%d/%Y")
              } else if (date_format == "dmy") {
                df[[col]] <- format(parsed_dates, "%d/%m/%Y")
              } else if (date_format == "ymd") {
                df[[col]] <- format(parsed_dates, "%Y-%m-%d")
              }
            }
          }, error = function(e) {
            message(sprintf("Note: Could not parse dates in column %s", col))
          })
        }
      }
    }
  }

  return(df)
}

#' Parse dates in various formats
#'
#' @param date_strings Vector of date strings to parse
#'
#' @return Vector of Date objects
#' @noRd
parseAnyDate <- function(date_strings) {
  # Create a result vector of the same length as input
  result <- rep(as.Date(NA), length(date_strings))

  # Skip NA values
  mask_na <- !is.na(date_strings)

  if (!any(mask_na)) {
    return(result)
  }

  valid_strings <- date_strings[mask_na]

  # Try different date formats
  parsed <- tryCatch({
    # Try lubridate parsing functions in sequence
    if (requireNamespace("lubridate", quietly = TRUE)) {
      # Try ISO format (YYYY-MM-DD)
      parsed <- lubridate::ymd(valid_strings, quiet = TRUE)

      # If that fails, try US format (MM/DD/YYYY)
      if (all(is.na(parsed))) {
        parsed <- lubridate::mdy(valid_strings, quiet = TRUE)
      }

      # If that fails, try European format (DD/MM/YYYY)
      if (all(is.na(parsed))) {
        parsed <- lubridate::dmy(valid_strings, quiet = TRUE)
      }

      parsed
    } else {
      # Fallback if lubridate is not available
      as.Date(valid_strings)
    }
  }, error = function(e) {
    rep(as.Date(NA), length(valid_strings))
  })

  # Update the result with parsed dates
  result[mask_na] <- parsed

  return(result)
}

#' Filter data frame by interview_date
#'
#' @param df The data frame to filter
#' @param interview_date The interview date filter value
#'
#' @return The filtered data frame
#' @noRd
filter_by_interview_date <- function(df, interview_date) {
  if (is.null(interview_date) || !("interview_date" %in% names(df))) {
    return(df)
  }

  # Create a temporary date column for filtering
  df$temp_date <- parseAnyDate(df$interview_date)

  if (is.logical(interview_date) && interview_date == TRUE) {
    # Keep only rows with non-NA interview_date values
    df <- df[!is.na(df$temp_date), ]
  } else if (is.character(interview_date) || inherits(interview_date, "Date")) {
    # Filter by specific date
    target_date <- if (inherits(interview_date, "Date")) {
      interview_date
    } else {
      tryCatch({
        parseAnyDate(interview_date)
      }, error = function(e) {
        stop("Failed to parse interview_date parameter: ", interview_date)
      })
    }

    if (is.na(target_date)) {
      stop("Failed to parse interview_date parameter: ", interview_date)
    }

    # Keep rows with dates before or equal to target_date
    df <- df[!is.na(df$temp_date) & df$temp_date <= target_date, ]
  }

  # Remove temporary column
  df$temp_date <- NULL

  return(df)
}

#' Filter data frame by column values
#'
#' @param df The data frame to filter
#' @param cols_to_check List of column names to check for non-NA values
#'
#' @return The filtered data frame
#' @noRd
filter_by_column_values <- function(df, cols_to_check) {
  if (length(cols_to_check) == 0) {
    return(df)
  }

  # Convert to character vector
  requested_cols <- as.character(unlist(cols_to_check))

  # Find which columns exist in the data
  existing_cols <- intersect(requested_cols, names(df))

  if (length(existing_cols) > 0) {
    # Display columns found
    message(sprintf("Found %d of %d requested columns: %s",
                    length(existing_cols),
                    length(requested_cols),
                    paste(existing_cols, collapse = ", ")))

    # Create a filter to keep rows with data in all requested columns
    rows_to_keep <- rep(TRUE, nrow(df))

    for (col in existing_cols) {
      # Check for non-NA values
      not_na <- !is.na(df[[col]])

      # For character columns, also check for non-empty strings
      not_empty <- rep(TRUE, nrow(df))
      if (is.character(df[[col]])) {
        not_empty <- df[[col]] != ""
      }

      # Update filter
      rows_to_keep <- rows_to_keep & not_na & not_empty
    }

    # Apply the filter
    original_rows <- nrow(df)
    df <- df[rows_to_keep, ]
    kept_rows <- nrow(df)

    message(sprintf("Kept %d of %d rows where all requested columns have values.",
                    kept_rows, original_rows))
  } else {
    warning("None of the requested columns were found in the dataset.")
  }

  return(df)
}

#' Format time duration in a human-readable way
#'
#' @param duration The duration to format
#' @return A formatted string
#' @noRd
formatDuration <- function(duration) {
  secs <- as.numeric(duration, units = "secs")

  if (secs < 60) {
    return(sprintf("%.1f seconds", secs))
  } else {
    mins <- floor(secs / 60)
    remaining_secs <- round(secs %% 60, 1)

    if (remaining_secs > 0) {
      return(sprintf("%d minutes and %.1f seconds", mins, remaining_secs))
    } else {
      return(sprintf("%d minutes", mins))
    }
  }
}

#' Initialize a loading animation
#'
#' @param steps Total number of steps
#' @return A list with animation state
#' @noRd
initializeLoadingAnimation <- function(steps) {
  # Get console width
  width <- tryCatch({
    if (requireNamespace("cli", quietly = TRUE)) {
      cli::console_width() - 10  # Leave some margin
    } else {
      getOption("width", 80) - 10  # Fallback to R's width setting
    }
  }, error = function(e) 80)  # Default if all else fails

  list(
    steps = steps,
    current = 0,
    width = width,
    start_time = Sys.time()
  )
}

#' Update the loading animation
#'
#' @param pb The progress bar object
#' @param current Current step
#' @return Nothing
#' @noRd
updateLoadingAnimation <- function(pb, current) {
  pb$current <- current
  percentage <- round(current / pb$steps * 100)
  filled <- round(pb$width * current / pb$steps)
  bar <- paste0(
    strrep("=", filled),
    strrep(" ", pb$width - filled)
  )
  cat(sprintf("\r|%s| %3d%%", bar, percentage))
  utils::flush.console()
}

#' Complete the loading animation
#'
#' @param pb The progress bar object
#' @return Nothing
#' @noRd
completeLoadingAnimation <- function(pb) {
  updateLoadingAnimation(pb, pb$steps)
  cat("\n")
}

#' Handle missing values in SQL data based on configuration
#'
#' This function processes a data frame and replaces values that match the missing data codes
#' defined in the configuration with NA. It handles different types of missing data
#' (skipped, refused, unknown, missing) according to the configuration.
#'
#' @param df Data frame containing SQL query results
#' @param config Configuration object loaded from config.yml
#' @return Data frame with missing values handled according to configuration
#' @noRd
handle_missing_values <- function(df, config) {
  # Return the dataframe unchanged if no missing_data_codes configuration
  if (is.null(config) || is.null(config$missing_data_codes)) {
    return(df)
  }

  # Get missing data codes from configuration
  missing_codes <- config$missing_data_codes
  transformed_count <- 0
  transformed_cols <- character(0)

  # Process each category of missing codes
  for (category in names(missing_codes)) {
    codes <- missing_codes[[category]]

    # Skip if no codes defined for this category
    if (length(codes) == 0) next

    # Apply to all columns in the dataframe
    for (col in names(df)) {
      # Only process numeric or character columns
      if (is.numeric(df[[col]]) || is.character(df[[col]])) {
        # For each code in this category
        for (code in codes) {
          # Convert to the appropriate type
          typed_code <- if (is.numeric(df[[col]])) {
            suppressWarnings(as.numeric(code))
          } else {
            as.character(code)
          }

          # Skip if conversion failed (e.g., non-numeric code for numeric column)
          if (is.na(typed_code) && !is.character(df[[col]])) next

          # Count matches before replacement
          matches <- sum(df[[col]] == typed_code, na.rm = TRUE)

          # Replace the code with NA if matches found
          if (matches > 0) {
            df[[col]][df[[col]] == typed_code] <- NA
            transformed_count <- transformed_count + matches

            # Track which columns were affected
            if (!(col %in% transformed_cols)) {
              transformed_cols <- c(transformed_cols, col)
            }
          }
        }
      }
    }
  }

  # Report changes if any were made
  if (transformed_count > 0) {
    message(sprintf("Replaced %d missing value codes with NA in %d column(s): %s",
                    transformed_count,
                    length(transformed_cols),
                    paste(transformed_cols, collapse=", ")))
  }

  return(df)
}

#' NDA Field Mapping and Detection Functions
#'
#' @description
#' Functions for detecting, mapping, and renaming fields to match NDA structure requirements.
#' Includes similarity calculation, field matching, and interactive renaming.
#'

#' Calculate Jaro-Winkler similarity between two strings
#'
#' @description
#' Calculates string similarity using Jaro-Winkler algorithm.
#' Enhanced for NDA structure name matching with scientific abbreviations.
#'
#' @param str1 First string
#' @param str2 Second string
#' @return Numeric similarity score (0-1)
#' @noRd
calculate_jaro_winkler <- function(str1, str2) {
  # Convert to lowercase for comparison
  str1 <- tolower(str1)
  str2 <- tolower(str2)
  
  # Handle exact matches first
  if (str1 == str2) return(1.0)
  if (nchar(str1) == 0 || nchar(str2) == 0) return(0.0)
  
  # Jaro-Winkler similarity algorithm implementation
  jaro_winkler <- function(s1, s2, prefix_weight = 0.1) {
    if (s1 == s2) return(1.0)
    if (nchar(s1) == 0 || nchar(s2) == 0) return(0.0)
    
    len1 <- nchar(s1)
    len2 <- nchar(s2)
    
    # Calculate match window
    match_window <- floor(max(len1, len2) / 2) - 1
    if (match_window < 0) match_window <- 0
    
    # Track matches
    s1_matches <- rep(FALSE, len1)
    s2_matches <- rep(FALSE, len2)
    matches <- 0
    transpositions <- 0
    
    # Find matches within the window
    for (i in 1:len1) {
      start <- max(1, i - match_window)
      end <- min(len2, i + match_window)
      
      for (j in start:end) {
        if (s2_matches[j] || substr(s1, i, i) != substr(s2, j, j)) next
        s1_matches[i] <- TRUE
        s2_matches[j] <- TRUE
        matches <- matches + 1
        break
      }
    }
    
    if (matches == 0) return(0.0)
    
    # Count transpositions
    k <- 1
    for (i in 1:len1) {
      if (!s1_matches[i]) next
      while (!s2_matches[k]) k <- k + 1
      if (substr(s1, i, i) != substr(s2, k, k)) transpositions <- transpositions + 1
      k <- k + 1
    }
    
    # Calculate Jaro similarity
    jaro <- (matches/len1 + matches/len2 + (matches - transpositions/2)/matches) / 3
    
    # Add Winkler prefix bonus
    prefix_length <- 0
    max_prefix <- min(len1, len2, 4)
    for (i in 1:max_prefix) {
      if (substr(s1, i, i) == substr(s2, i, i)) {
        prefix_length <- prefix_length + 1
      } else {
        break
      }
    }
    
    return(jaro + (prefix_weight * prefix_length * (1 - jaro)))
  }
  
  # Start with Jaro-Winkler base similarity
  base_score <- jaro_winkler(str1, str2)
  
  return(base_score)
}

#' Calculate Levenshtein distance similarity between two strings
#'
#' @description
#' Alternative similarity metric using edit distance.
#'
#' @param str1 First string
#' @param str2 Second string
#' @return Numeric similarity score (0-1)
#' @noRd
calculate_levenshtein <- function(str1, str2) {
  # Convert to lowercase
  str1 <- tolower(str1)
  str2 <- tolower(str2)
  
  # Create matrix
  m <- nchar(str1)
  n <- nchar(str2)
  d <- matrix(0, nrow = m + 1, ncol = n + 1)
  d[1,] <- 0:n
  d[,1] <- 0:m
  
  # Fill matrix
  for (i in 2:(m + 1)) {
    for (j in 2:(n + 1)) {
      cost <- if (substr(str1, i-1, i-1) == substr(str2, j-1, j-1)) 0 else 1
      d[i,j] <- min(
        d[i-1,j] + 1,      # deletion
        d[i,j-1] + 1,      # insertion
        d[i-1,j-1] + cost  # substitution
      )
    }
  }
  
  # Return similarity score (1 - normalized distance)
  return(1 - d[m+1,n+1] / max(m,n))
}

#' Extract mapping rules from NDA field notes
#'
#' @description
#' Parses the "notes" field from NDA structure to extract value mappings
#' (e.g., "1=Never; 2=Sometimes; 999=Missing")
#'
#' @param notes Notes string from NDA dataElement
#' @return Named list of code -> meaning mappings, or NULL
#' @noRd
get_mapping_rules <- function(notes) {
  if (is.null(notes) || is.na(notes) || notes == "") return(NULL)
  
  rules <- list()
  
  tryCatch({
    # Handle more elaborate mapping formats like "-999=Missing" or "999=Not Applicable"
    pattern1 <- "(-?\\d+)\\s*=\\s*([^;,]+)"
    matches1 <- gregexpr(pattern1, notes, perl = TRUE)
    if (matches1[[1]][1] != -1) {
      all_matches <- regmatches(notes, matches1)[[1]]
      for (match in all_matches) {
        parts <- regexec(pattern1, match, perl = TRUE)
        extracted <- regmatches(match, parts)[[1]]
        if (length(extracted) >= 3) {
          code <- trimws(extracted[2])
          meaning <- trimws(extracted[3])
          rules[[code]] <- meaning
        }
      }
    }
    
    # Handle simple mappings like "0=Never" or "0 = Never"
    if (grepl("[^=]+=[^;]+", notes)) {
      patterns <- strsplit(notes, ";\\s*")[[1]]
      for (pattern in patterns) {
        if (grepl("=", pattern)) {
          parts <- strsplit(pattern, "=")[[1]]
          if (length(parts) >= 2) {
            code <- trimws(parts[1])
            value <- trimws(parts[2])
            if (!(code %in% names(rules))) {
              rules[[code]] <- value
            }
          }
        }
      }
    }
    
    # Handle array notation like "1=(0.9, 0.5, 0.1)"
    if (grepl("=\\(.*\\)", notes)) {
      pattern_matches <- gregexpr("(\\d+)=\\(([^)]+)\\)", notes)
      if (pattern_matches[[1]][1] != -1) {
        matches <- regmatches(notes, pattern_matches)[[1]]
        for (match in matches) {
          code_match <- regexec("(\\d+)=\\(([^)]+)\\)", match)
          parts <- regmatches(match, code_match)[[1]]
          if (length(parts) >= 3) {
            code <- parts[2]
            values <- sprintf("[%s]", parts[3])
            rules[[code]] <- values
          }
        }
      }
    }
  }, error = function(e) {
    warning(sprintf("Error parsing mapping rules: %s\nNotes: %s", e$message, notes))
    return(list())
  })
  
  return(rules)
}

#' Handle missing required fields
#'
#' @description
#' Auto-adds missing required fields with appropriate missing data codes
#' extracted from the NDA structure notes.
#'
#' @param df Data frame
#' @param elements NDA structure dataElements
#' @param missing_required Character vector of missing required field names
#' @param verbose Logical - print details
#' @return Data frame with missing fields added
#' @noRd
handle_missing_fields <- function(df, elements, missing_required, verbose = FALSE) {
  if(verbose) {
    message("\nAuto-adding missing required fields with missing data codes:")
  }
  
  for (field in missing_required) {
    element <- elements[elements$name == field, ]
    
    if(verbose) {
      cat("\nNotes for", field, ":", element$notes)
    }
    
    # Extract missing value code from notes (e.g. -999)
    missing_code <- NULL
    if (!is.na(element$notes)) {
      # Look for pattern like "999 = Missing" or "999 = Missing/NA"
      if (grepl("\\d+\\s*=\\s*Missing(?:/NA)?", element$notes, perl = TRUE)) {
        value <- gsub(".*?(\\d+)\\s*=\\s*Missing(?:/NA)?.*", "\\1", element$notes, perl = TRUE)
        missing_code <- paste0("-", value)  # Make it negative
      }
    }
    
    if (!is.null(missing_code)) {
      # Convert to proper type and fill entire column
      if (element$type == "Float") {
        df[[field]] <- as.numeric(missing_code)
      } else if (element$type == "Integer") {
        df[[field]] <- as.integer(missing_code)
      } else {
        df[[field]] <- missing_code
      }
      
      if(verbose) {
        cat(sprintf("\n- Added %s with missing code %s as type %s",
                   field, missing_code, element$type))
      }
    }
  }
  
  if(verbose) cat("\n")
  return(df)
}

#' Get field value range from NDA elements
#'
#' @param field_name Name of the field
#' @param elements NDA data elements
#' @return Value range string or "No range specified"
#' @noRd
get_field_value_range <- function(field_name, elements) {
  if (is.null(elements) || nrow(elements) == 0) {
    return("No range specified")
  }
  
  # Find the field in elements
  field_idx <- which(elements$name == field_name)
  if (length(field_idx) == 0) {
    return("No range specified")
  }
  
  # Get the value range
  value_range <- elements$valueRange[field_idx[1]]
  if (is.null(value_range) || is.na(value_range) || value_range == "") {
    return("No range specified")
  }
  
  return(as.character(value_range))
}

#' Parse a value range string to extract numeric ranges and missing codes
#'
#' @param value_range Value range string (e.g., "0::6;999=missing")
#' @return List with parsed components (numeric_range, missing_codes)
#' @noRd
parse_value_range <- function(value_range) {
  if (is.null(value_range) || is.na(value_range) || value_range == "") {
    return(list(numeric_range = NULL, missing_codes = NULL))
  }
  
  # Split by semicolon
  parts <- trimws(strsplit(value_range, ";")[[1]])
  
  numeric_range <- NULL
  missing_codes <- NULL
  
  for (part in parts) {
    if (grepl("::", part)) {
      # This is a numeric range (e.g., "0::6")
      range_parts <- strsplit(part, "::")[[1]]
      if (length(range_parts) == 2) {
        min_val <- as.numeric(range_parts[1])
        max_val <- as.numeric(range_parts[2])
        if (!is.na(min_val) && !is.na(max_val)) {
          numeric_range <- list(min = min_val, max = max_val)
        }
      }
    } else if (grepl("=", part)) {
      # This is a missing code with meaning (e.g., "999=missing")
      code_parts <- strsplit(part, "=")[[1]]
      if (length(code_parts) == 2) {
        code <- trimws(code_parts[1])
        meaning <- trimws(code_parts[2])
        if (grepl("missing|na|not.*applicable", meaning, ignore.case = TRUE)) {
          if (is.null(missing_codes)) missing_codes <- character(0)
          missing_codes <- c(missing_codes, code)
        }
      }
    } else {
      # This is a standalone special code (e.g., "888", "-777", "999")
      if (grepl("^-?\\d+$", part)) {
        if (is.null(missing_codes)) missing_codes <- character(0)
        missing_codes <- c(missing_codes, part)
      }
    }
  }
  
  return(list(numeric_range = numeric_range, missing_codes = missing_codes))
}

#' Check if two value ranges are compatible for renaming
#'
#' @param source_range Value range of source field
#' @param target_range Value range of target field
#' @param source_field Name of source field
#' @param target_field Name of target field
#' @param verbose Whether to print verbose output
#' @return List with compatibility check results
#' @noRd
check_value_range_compatibility <- function(source_range, target_range, source_field, target_field, verbose = FALSE) {
  
  # If either range is missing, they're compatible
  if (source_range == "No range specified" || target_range == "No range specified") {
    return(list(
      compatible = TRUE,
      message = "One or both fields have no value range specified",
      recommendation = "Proceed with caution - ranges cannot be validated"
    ))
  }
  
  # If ranges are identical, they're compatible
  if (source_range == target_range) {
    return(list(
      compatible = TRUE,
      message = "Value ranges are identical",
      recommendation = "Safe to rename"
    ))
  }
  
  # Parse the ranges to extract numeric ranges and special codes
  source_parts <- parse_value_range(source_range)
  target_parts <- parse_value_range(target_range)
  
  # Check if both have numeric ranges
  if (!is.null(source_parts$numeric_range) && !is.null(target_parts$numeric_range)) {
    source_min <- source_parts$numeric_range$min
    source_max <- source_parts$numeric_range$max
    target_min <- target_parts$numeric_range$min
    target_max <- target_parts$numeric_range$max
    
    # Check if ranges overlap significantly
    overlap_min <- max(source_min, target_min)
    overlap_max <- min(source_max, target_max)
    overlap_size <- max(0, overlap_max - overlap_min + 1)
    
    source_size <- source_max - source_min + 1
    target_size <- target_max - target_min + 1
    
    # Calculate overlap percentage
    overlap_percentage <- overlap_size / min(source_size, target_size) * 100
    
    if (overlap_percentage >= 80) {
      return(list(
        compatible = TRUE,
        message = sprintf("Numeric ranges overlap by %.1f%% - likely compatible", overlap_percentage),
        recommendation = "Proceed with rename"
      ))
    } else if (overlap_percentage >= 50) {
      return(list(
        compatible = FALSE,
        message = sprintf("Numeric ranges overlap by only %.1f%% - potential conflict", overlap_percentage),
        recommendation = "Consider creating a new field instead of renaming"
      ))
    } else {
      return(list(
        compatible = FALSE,
        message = sprintf("Numeric ranges have minimal overlap (%.1f%%) - likely incompatible", overlap_percentage),
        recommendation = "Create a new field instead of renaming"
      ))
    }
  }
  
  # Default: ranges are compatible if we can't determine otherwise
  return(list(
    compatible = TRUE,
    message = "Cannot determine compatibility - special codes or no numeric ranges",
    recommendation = "Proceed with caution"
  ))
}

#' Get Matrix Group Name for REDCap Field
#' @param field_name Character - field to check
#' @param api Character - API type
#' @param measure_name Character - for REDCap dictionary lookup
#' @return Character - matrix group name or empty string
#' @noRd
get_matrix_group_name <- function(field_name, api, measure_name) {
  if (tolower(api) != "redcap") return("")
  
  redcap_dict <- tryCatch({
    redcap.dict(measure_name)
  }, error = function(e) NULL)
  
  if (is.null(redcap_dict) || !field_name %in% redcap_dict$field_name) {
    return("")
  }
  
  matrix_group <- redcap_dict$matrix_group_name[redcap_dict$field_name == field_name]
  if (is.na(matrix_group)) return("")
  
  return(as.character(matrix_group))
}

#' Fetch NDA structure from API
#'
#' @param structure_name Structure short name
#' @param nda_base_url NDA API base URL
#' @return List with structure definition
#' @noRd
fetch_nda_structure <- function(structure_name, nda_base_url) {
  url <- sprintf("%s/datastructure/%s", nda_base_url, structure_name)
  response <- httr::GET(url, httr::timeout(10))
  
  if (httr::status_code(response) != 200) {
    stop(sprintf("Failed to fetch structure '%s' - HTTP %d", 
                structure_name, httr::status_code(response)))
  }
  
  content <- jsonlite::fromJSON(rawToChar(response$content))
  
  if (!"dataElements" %in% names(content)) {
    stop("Unexpected API response format - no dataElements found")
  }
  
  return(content)
}

#' DataEnvironment R6 Class
#'
#' @description
#' Manages dataframe storage in package environment (.pkg_env$.wizaRdry_env) with optional
#' convenience assignment to calling environment. CRAN-compliant environment management.
#'
#' @details
#' This class provides a clean interface for getting and setting dataframes using the package
#' environment (.pkg_env) as the authoritative source, with optional assignment to the calling
#' environment for user convenience. This eliminates global environment pollution and follows
#' R package best practices.
#'
#' @keywords internal
DataEnvironment <- R6::R6Class("DataEnvironment",
  public = list(
    #' @field measure_name Character string - name of the measure/dataframe
    measure_name = NULL,
    
    #' @description
    #' Create a new DataEnvironment instance
    #' @param measure_name Name of the measure/dataframe
    #' @param df Initial dataframe to store
    #' @return A new DataEnvironment object
    initialize = function(measure_name, df) {
      if (missing(measure_name) || is.null(measure_name) || measure_name == "") {
        stop("measure_name is required and cannot be empty")
      }
      if (missing(df) || is.null(df) || !is.data.frame(df)) {
        stop("df must be a valid data.frame")
      }
      
      self$measure_name <- measure_name
      self$set_df(df)
    },
    
    #' @description
    #' Get dataframe from package environment or calling environment
    #' @return The dataframe stored in package environment
    get_df = function() {
      # Priority order: .pkg_env$.wizaRdry_env (authoritative), then calling environment
      
      # Check package environment first (authoritative source)
      if (exists(".wizaRdry_env", envir = .pkg_env, inherits = FALSE)) {
        wizaRdry_env <- .pkg_env$.wizaRdry_env
        if (exists(self$measure_name, envir = wizaRdry_env, inherits = FALSE)) {
          return(base::get(self$measure_name, envir = wizaRdry_env))
        }
      }
      
      # Fallback: Try calling environment (for user convenience)
      calling_env <- parent.frame(2)
      if (exists(self$measure_name, envir = calling_env, inherits = FALSE)) {
        return(base::get(self$measure_name, envir = calling_env))
      }
      
      stop(sprintf("Dataframe '%s' not found in package environment or calling environment", self$measure_name))
    },
    
    #' @description
    #' Set dataframe in package environment with optional calling environment assignment
    #' @param df Data frame to set
    #' @return Self (invisibly) for method chaining
    set_df = function(df) {
      if (is.null(df) || !is.data.frame(df)) {
        stop("df must be a valid data.frame")
      }
      
      # Ensure .wizaRdry_env exists in package environment (CRAN compliant)
      if (!exists(".wizaRdry_env", envir = .pkg_env, inherits = FALSE)) {
        assign(".wizaRdry_env", new.env(parent = emptyenv()), envir = .pkg_env)
      }
      
      # Get wizaRdry environment from package environment
      wizaRdry_env <- .pkg_env$.wizaRdry_env
      
      # Set in package environment (authoritative source)
      base::assign(self$measure_name, df, envir = wizaRdry_env)
      
      # Also set in calling environment for user convenience (HYBRID approach)
      tryCatch({
        calling_env <- parent.frame(2)
        base::assign(self$measure_name, df, envir = calling_env)
      }, error = function(e) {
        # Calling env not accessible - this is OK, we have package env
      })
      
      invisible(self)
    },
    
    #' @description
    #' Get column names from the dataframe
    #' @return Character vector of column names
    get_colnames = function() {
      names(self$get_df())
    },
    
    #' @description
    #' Get number of rows in the dataframe
    #' @return Integer number of rows
    nrow = function() {
      nrow(self$get_df())
    },
    
    #' @description
    #' Get number of columns in the dataframe
    #' @return Integer number of columns
    ncol = function() {
      ncol(self$get_df())
    },
    
    #' @description
    #' Print method for DataEnvironment
    #' @return Self (invisibly)
    print = function() {
      cat("DataEnvironment:\n")
      cat(sprintf("  Measure: %s\n", self$measure_name))
      cat(sprintf("  Dimensions: %d rows x %d columns\n", self$nrow(), self$ncol()))
      cat(sprintf("  Columns: %s\n", paste(head(self$get_colnames(), 5), collapse = ", ")))
      if (self$ncol() > 5) {
        cat(sprintf("           ... and %d more\n", self$ncol() - 5))
      }
      invisible(self)
    }
  )
)

#' Check for Duplicates in Qualtrics Data
#'
#' This function checks for duplicate records in a specified Qualtrics dataset
#' based on certain identifiers and time points. If duplicates are found,
#' they are exported to a CSV file for review and additionally displayed in the viewer.
#'
#' @param measure_alias A string representing the name of the dataset to check for duplicates.
#' @param measure_type A string specifying the type of measure, currently supports 'qualtrics' only.
#' @return The function does not return a value but will generate a CSV file if duplicates are found
#'         and display those duplicates in the RStudio viewer.
#' @examples
#' \donttest{
#' checkDuplicates("your_dataset_alias", "qualtrics")
#' }
#' @importFrom dplyr filter %>%
#' @importFrom testthat test_that expect_true
#' @note This function requires the dplyr and testthat packages. It is specifically designed for
#'       Qualtrics data and expects the data frame to be named with a '_clean' suffix.
#'       It checks for duplicates based on 'src_subject_id' combined with 'visit' or 'week' columns.
#'       The function will stop and throw an error if the necessary columns are not present.
#' @noRd

checkQualtricsDuplicates <- function(measure_alias, measure_type, verbose = TRUE) {

  # Ensure required packages are loaded

  # Generate the name of the dataframe and get it
  output_df_name <- paste0(measure_alias, "_clean")
  df <- base::get(output_df_name)
  
  # possible identifiers
  possible_identifiers <- c("src_subject_id", "workerId", "PROLIFIC_PID", "participantId")

  if (measure_type == "qualtrics") {
    # Find which identifiers exist in the dataframe
    existing_identifiers <- possible_identifiers[possible_identifiers %in% colnames(df)]
    
    if (length(existing_identifiers) == 0) {
      stop("Please provide valid identifiers: src_subject_id, workerId, PROLIFIC_PID, participantId")
    }
    
    # Find the first identifier with non-NA values
    identifier <- NA
    for (key in existing_identifiers) {
      non_na_count <- sum(!is.na(df[[key]]))
      if (non_na_count > 0) {
        identifier <- key
        break
      }
    }
    
    if (is.na(identifier)) {
      stop(paste("No identifier found with non-NA values. Available identifiers:", paste(existing_identifiers, collapse = ", ")))
    }

    for (col in c("visit", "week")) {
      if (col %in% base::colnames(df)) {
        df$duplicates <- base::duplicated(df[c(identifier, col)], first = TRUE)
        df_dup_ids <- base::subset(df, duplicates == TRUE)[, c(identifier, col)]

        if (base::nrow(df_dup_ids) > 0) {
          # Using inner_join to filter duplicates correctly
          df_duplicates <- df %>%
            dplyr::inner_join(df_dup_ids, by = c(identifier, col))

          if (base::nrow(df_duplicates) > 0) {
            # Export and create a CSV file if duplicates found
            duplicate_extract <- paste0("duplicates_", measure_alias)
            createCsv(df_duplicates, paste0("duplicates_", measure_alias))

            duplicates_summary <- toString(base::unique(df_duplicates[[identifier]]))
            error_msg <- paste("DATA ERROR: Duplicates detected in '", measure_alias, "': ",
                              "Offending IDs: ", duplicates_summary,
                              "\nDetails exported to ", paste0("./tmp/",duplicate_extract,".csv"))
            
            tryCatch({
              testthat::test_that("Check for Qualtrics duplicates", {
                testthat::expect_true(FALSE, info = error_msg)
              })
            }, error = function(e) {
              message(error_msg)
            })

            # Optionally view the offending records in RStudio's data viewer
            if (verbose) {
              View(df_duplicates, title = paste0("Duplicates: ", measure_alias))
            }
          }
        }
      }
    }
  }
}

#' Check for Duplicates in Qualtrics Data
#'
#' This function checks for duplicate records in a specified Qualtrics dataset
#' based on certain identifiers and time points. If duplicates are found,
#' they are exported to a CSV file for review and additionally displayed in the viewer.
#'
#' @param measure_alias A string representing the name of the dataset to check for duplicates.
#' @param measure_type A string specifying the type of measure, currently supports 'qualtrics' only.
#' @return The function does not return a value but will generate a CSV file if duplicates are found
#'         and display those duplicates in the RStudio viewer.
#' @examples
#' \donttest{
#' checkDuplicates("your_dataset_alias", "qualtrics")
#' }
#' @importFrom dplyr filter %>% 
#' @importFrom testthat test_that expect_true
#' @note This function requires the dplyr and testthat packages. It is specifically designed for
#'       Qualtrics data and expects the data frame to be named with a '_clean' suffix.
#'       It checks for duplicates based on 'src_subject_id' combined with 'visit' or 'week' columns.
#'       The function will stop and throw an error if the necessary columns are not present.
#' @noRd
ndaCheckQualtricsDuplicates <- function(measure_alias, measure_type) {
  
  # Ensure required packages are loaded
  
  identifier <- "src_subject_id"
  
  df <- base::get0(measure_alias)
  
  if (measure_type == "qualtrics") {
    if (!(identifier %in% colnames(df))) {
      stop("Please provide a valid identifier: src_subject_id, workerId, PROLIFIC_PID")
    }
    
    for (col in c("visit", "week")) {
      if (col %in% base::colnames(df)) {
        df$duplicates <- base::duplicated(df[c(identifier, col)], first = TRUE)
        df_dup_ids <- base::subset(df, duplicates == TRUE)[, c(identifier, col)]
        
        if (base::nrow(df_dup_ids) > 0) {
          # Using inner_join to filter duplicates correctly
          df_duplicates <- df %>% 
            dplyr::inner_join(df_dup_ids, by = c(identifier, col))
          
          if (base::nrow(df_duplicates) > 0) {
            # Export and create a CSV file if duplicates found
            duplicate_extract <- paste0("duplicates_", measure_alias)
            createCsv(df_duplicates, paste0("duplicates_", measure_alias))
            
            tryCatch({
              testthat::test_that("Check for Qualtrics duplicates", {
                testthat::expect_true(base::nrow(df_duplicates) == 0, 
                                      info = paste("DATA ERROR: Duplicates detected in '", measure_alias, "': ", 
                                                   "Offending IDs: ", toString(base::unique(df_duplicates[[identifier]]))))
              })
            }, error = function(e) {
              duplicates_summary <- toString(base::unique(df_duplicates[[identifier]]))
              message(paste("Error in testing for duplicates in '", measure_alias, "': ", e$message, 
                            "\nOffending IDs: ", duplicates_summary,
                            "\nDetails exported to ", paste0("./tmp/",duplicate_extract,".csv")))
            })
            
            # Optionally view the offending records in RStudio's data viewer
            View(df_duplicates)
          }
        }
      }
    }
  }
}

#' Filter data frame by superkey parameters, rows, and columns
#'
#' @param df Dataframe to be filtered and trimmed based on the provided parameters.
#' @param rows Optional; either a single row name or a vector of row names to be retained in the final output. 
#'        If NULL or empty, all rows in the dataframe are retained.
#' @param cols Optional; either a single column name or a vector of column names to be retained in the final output. 
#'        If NULL or empty, all columns in the dataframe are retained.#' Data Filter
#' @param record_id Optional; either a single record_id or a vector of record_ids to filter the dataframe by
#' @param src_subject_id Optional; either a single subject ID or a vector of subject IDs to filter the dataframe by
#' @param subjectkey Optional; either a single subjectkey or a vector of subjectkeys to filter the dataframe by
#' @param site Optional; either a single site value or a vector of site values to filter the dataframe by (e.g., Yale, NU)
#' @param subsiteid Optional; either a single subsiteid or a vector of subsiteids to filter the dataframe by
#' @param sex Optional; either a single sex value or a vector of sex values at birth to filter the dataframe by (e.g., 'M', 'F')
#' @param race Optional; either a single race value or a vector of race values to filter the dataframe by
#' @param ethnic_group Optional; either a single ethnic_group value or a vector of ethnic_group values to filter the dataframe by
#' @param phenotype Optional; either a single phenotype value or a vector of phenotype values to filter the dataframe by
#' @param phenotype_description Optional; either a single phenotype_description or a vector of phenotype_descriptions to filter the dataframe by
#' @param status Optional; either a single status string or a vector of status conditions to filter the dataframe by. Used 
#'        if either 'state' or 'status' column exists in the dataframe. Can include values like 'complete', 
#'        'completed baseline', 'completed 12m', 'completed 24m', etc.
#' @param lost_to_followup Optional; either a single value or a vector of values to filter the dataframe by (checks both 'lost_to_followup' and 'lost_to_follow-up' columns)
#' @param twins_study Optional; either a single twins_study value or a vector of twins_study values to filter the dataframe by
#' @param sibling_study Optional; either a single sibling_study value or a vector of sibling_study values to filter the dataframe by
#' @param family_study Optional; either a single family_study value or a vector of family_study values to filter the dataframe by
#' @param sample_taken Optional; either a single sample_taken value or a vector of sample_taken values to filter the dataframe by
#' @param visit Optional; either a single visit value or a vector of visit values to filter the dataframe by. 
#'        Only used if 'visit' column exists in the dataframe.
#' @param week Optional; either a single week value or a vector of week values to filter the dataframe by. 
#'        Only used if 'week' column exists in the dataframe.
#' @param arm Optional; either a single arm value or a vector of arm values to filter the dataframe by (e.g., drug, placebo)
#' @param interview_date Optional; can be either:
#'        - A date string in various formats (ISO, US, etc.) to filter data up to that date
#'        - A boolean TRUE to return only rows with non-NA interview_date values
#'
#' @return A filtered dataframe based on the provided parameters, 
#'         and containing only the columns specified in 'cols'. If no columns 
#'         are specified, returns the entire dataframe with applied row filters.
#'
#' @examples
#' # Create a sample dataframe
#' sample_df <- data.frame(
#'   record_id = c("REC001", "REC002", "REC003", "REC004"),
#'   src_subject_id = c("SUB001", "SUB002", "SUB003", "SUB004"),
#'   subjectkey = c("KEY001", "KEY002", "KEY003", "KEY004"),
#'   site = c("Yale", "NU", "Yale", "NU"),
#'   phenotype = c("A", "B", "A", "C"),
#'   visit = c(1, 2, 2, 1),
#'   state = c("complete", "completed baseline", "in progress", NA),
#'   status = c(NA, NA, NA, "complete"),
#'   lost_to_followup = c(FALSE, FALSE, TRUE, NA),
#'   interview_date = c("2023-01-15", "2023/02/20", NA, "2023-03-10")
#' )
#' 
#' # Set row names for demonstration
#' rownames(sample_df) <- c("foo", "bar", "baz", "qux")
#' 
#' # Filter by specific date
#' filtered1 <- sift(sample_df, 
#'                  cols = c("src_subject_id", "phenotype"), 
#'                  visit = 2, 
#'                  interview_date = "01/31/2023")
#'                  
#' # Filter to include only rows with non-NA interview dates
#' filtered2 <- sift(sample_df, 
#'                  interview_date = TRUE)
#'                  
#' # Filter by status (works with either state or status column)
#' filtered3 <- sift(sample_df,
#'                  status = c("complete", "completed baseline"))
#'                  
#' # Filter with specific row names
#' filtered4 <- sift(sample_df,
#'                  rows = c("foo", "qux"))
#'                  
#' # Filter with vector of visit values
#' filtered6 <- sift(sample_df,
#'                  visit = c(1, 2))
#'                  
#' # Filter by lost_to_followup
#' filtered10 <- sift(sample_df,
#'                  lost_to_followup = FALSE)
#'                  
#' # Filter by src_subject_id
#' filtered11 <- sift(sample_df,
#'                  src_subject_id = c("SUB001", "SUB004"))
#'                  
#' # Multiple filters combined
#' filtered12 <- sift(sample_df,
#'                  site = "Yale",
#'                  visit = 1,
#'                  cols = c("record_id", "src_subject_id", "site"))
#' @import dplyr
#' @import lubridate
#' @export
sift <- function(df, rows = NULL, cols = NULL,
                 record_id = NULL, src_subject_id = NULL, subjectkey = NULL, 
                 site = NULL, subsiteid = NULL, sex = NULL, race = NULL, 
                 ethnic_group = NULL, phenotype = NULL, phenotype_description = NULL, 
                 status = NULL, lost_to_followup = NULL, twins_study = NULL, 
                 sibling_study = NULL, family_study = NULL, sample_taken = NULL,
                 visit = NULL, week = NULL, arm = NULL, interview_date = NULL) {
  
  
  # Create a copy of the original dataframe to preserve original values
  original_df <- df
  
  # Advanced date parsing function that handles multiple formats
  parseAnyDate <- function(date_string) {
    if (is.na(date_string) || is.null(date_string)) {
      return(NA)
    }
    
    # Try multiple date formats sequentially
    date <- NULL
    
    # Try ISO format (YYYY-MM-DD)
    if (grepl("^\\d{4}-\\d{1,2}-\\d{1,2}$", date_string)) {
      date <- tryCatch(ymd(date_string), error = function(e) NULL)
    } 
    # Try US format (MM/DD/YYYY)
    else if (grepl("^\\d{1,2}/\\d{1,2}/\\d{4}$", date_string)) {
      date <- tryCatch(mdy(date_string), error = function(e) NULL)
    } 
    # Try European format (DD.MM.YYYY)
    else if (grepl("^\\d{1,2}\\.\\d{1,2}\\.\\d{4}$", date_string)) {
      date <- tryCatch(dmy(date_string), error = function(e) NULL)
    }
    # Try Canadian format (YYYY/MM/DD)
    else if (grepl("^\\d{4}/\\d{1,2}/\\d{1,2}$", date_string)) {
      date <- tryCatch(ymd(date_string), error = function(e) NULL)
    }
    # Try other format (DD-MM-YYYY)
    else if (grepl("^\\d{1,2}-\\d{1,2}-\\d{4}$", date_string)) {
      date <- tryCatch(dmy(date_string), error = function(e) NULL)
    }
    # Try abbreviated month name (15-Jan-2023 or Jan 15, 2023)
    else if (grepl("[A-Za-z]", date_string)) {
      date <- tryCatch(parse_date_time(date_string, c("dmy", "mdy")), error = function(e) NULL)
    }
    
    # If all attempts fail, return NA
    if (is.null(date) || all(is.na(date))) {
      warning("Failed to parse date: ", date_string, ". Treating as NA.")
      return(NA)
    }
    
    return(as.Date(date))
  }
  
  # Handle interview_date filtering
  if ("interview_date" %in% names(df)) {
    # Create a temporary date column for filtering but don't modify the original
    df$temp_date <- sapply(df$interview_date, parseAnyDate)
    
    # Handle the interview_date parameter
    if (!is.null(interview_date)) {
      if (is.logical(interview_date) && interview_date == TRUE) {
        # Keep only rows with non-NA interview_date values
        rows_to_keep <- !is.na(df$temp_date)
        df <- df[rows_to_keep, ]
        original_df <- original_df[rows_to_keep, ]
      } else if (is.character(interview_date) || inherits(interview_date, "Date")) {
        # Filter by specific date
        input_date <- tryCatch({
          if (inherits(interview_date, "Date")) {
            interview_date
          } else {
            parseAnyDate(interview_date)
          }
        }, error = function(e) {
          stop("Failed to parse interview_date parameter: ", interview_date)
        })
        
        if (is.na(input_date)) {
          stop("Failed to parse interview_date parameter: ", interview_date)
        }
        
        rows_to_keep <- df$temp_date <= input_date
        df <- df[rows_to_keep, ]
        original_df <- original_df[rows_to_keep, ]
      } else {
        stop("interview_date must be either a date string or TRUE")
      }
    }
    
    # Remove the temporary date column
    df$temp_date <- NULL
  }
  
  # Use the filtered dataframe for applying other filters, but keep the original values
  
  # Check for either state or status column and apply filtering
  if (!is.null(status)) {
    # Convert single string to vector if needed
    status_values <- if(is.character(status) && length(status) == 1) {
      c(status)
    } else {
      status
    }
    
    if ("state" %in% names(df)) {
      rows_to_keep <- df$state %in% status_values
      df <- df[rows_to_keep, ]
      original_df <- original_df[rows_to_keep, ]
    } else if ("status" %in% names(df)) {
      rows_to_keep <- df$status %in% status_values
      df <- df[rows_to_keep, ]
      original_df <- original_df[rows_to_keep, ]
    } else {
      warning("Neither 'state' nor 'status' column found in the dataframe. Status filtering skipped.")
    }
  }
  
  if ("visit" %in% names(df) && !is.null(visit)) {
    # Convert single value to vector if needed
    visit_values <- if(length(visit) == 1) {
      c(visit)
    } else {
      visit
    }
    rows_to_keep <- df$visit %in% visit_values
    df <- df[rows_to_keep, ]
    original_df <- original_df[rows_to_keep, ]
  }
  
  if ("week" %in% names(df) && !is.null(week)) {
    # Convert single value to vector if needed
    week_values <- if(length(week) == 1) {
      c(week)
    } else {
      week
    }
    rows_to_keep <- df$week %in% week_values
    df <- df[rows_to_keep, ]
    original_df <- original_df[rows_to_keep, ]
  }
  
  if ("arm" %in% names(df) && !is.null(arm)) {
    rows_to_keep <- df$arm %in% arm
    df <- df[rows_to_keep, ]
    original_df <- original_df[rows_to_keep, ]
  }
  
  if ("site" %in% names(df) && !is.null(site)) {
    rows_to_keep <- df$site %in% site
    df <- df[rows_to_keep, ]
    original_df <- original_df[rows_to_keep, ]
  }
  
  if ("sex" %in% names(df) && !is.null(sex)) {
    rows_to_keep <- df$sex %in% sex
    df <- df[rows_to_keep, ]
    original_df <- original_df[rows_to_keep, ]
  }
  
  if ("phenotype" %in% names(df) && !is.null(phenotype)) {
    rows_to_keep <- df$phenotype %in% phenotype
    df <- df[rows_to_keep, ]
    original_df <- original_df[rows_to_keep, ]
  }
  
  if ("record_id" %in% names(df) && !is.null(record_id)) {
    rows_to_keep <- df$record_id %in% record_id
    df <- df[rows_to_keep, ]
    original_df <- original_df[rows_to_keep, ]
  }
  
  if ("src_subject_id" %in% names(df) && !is.null(src_subject_id)) {
    rows_to_keep <- df$src_subject_id %in% src_subject_id
    df <- df[rows_to_keep, ]
    original_df <- original_df[rows_to_keep, ]
  }
  
  if ("subjectkey" %in% names(df) && !is.null(subjectkey)) {
    rows_to_keep <- df$subjectkey %in% subjectkey
    df <- df[rows_to_keep, ]
    original_df <- original_df[rows_to_keep, ]
  }
  
  if (!is.null(lost_to_followup)) {
    if ("lost_to_followup" %in% names(df)) {
      rows_to_keep <- df$lost_to_followup == lost_to_followup
      df <- df[rows_to_keep, ]
      original_df <- original_df[rows_to_keep, ]
    } else if ("lost_to_follow-up" %in% names(df)) {
      rows_to_keep <- df$`lost_to_follow-up` == lost_to_followup
      df <- df[rows_to_keep, ]
      original_df <- original_df[rows_to_keep, ]
    }
  }
  
  # Apply row and column filters to the original_df which has original values
  result_df <- original_df
  
  if (!is.null(cols)) {
    # Convert single column name to vector if needed
    col_values <- if(is.character(cols) && length(cols) == 1) {
      c(cols)
    } else {
      cols
    }
    result_df <- result_df[, col_values, drop = FALSE]
  } else {
    message("No columns provided; all columns will be included.")
  }
  
  if (!is.null(rows)) {
    # Convert single row name to vector if needed
    row_values <- if(is.character(rows) && length(rows) == 1) {
      c(rows)
    } else {
      rows
    }
    result_df <- result_df[row_values, , drop = FALSE]
  }
  
  return(result_df)
}

#' Alias for 'sift' (DEPRECATED)
#'
#' This function is deprecated. Please use 'sift' instead.
#' This is a legacy alias for the 'sift' function to maintain compatibility with older code.
#'
#' @param ... Additional arguments passed through to \code{sift()}.
#' @inherit sift return
#' @export
#' @examples
#' \dontrun{
#' # DEPRECATED - use sift() instead
#' filtered <- dataFilter(df, sex="F")
#' }
dataFilter <- function(...) {
  .Deprecated("sift", package = "wizaRdry")
  sift(...)
}

#' Add prefix to columns and handle text normalization
#'
#' @param df A data frame
#' @param collection_name The prefix to add
#' @param excluded_cols Columns to exclude from prefixing
#' @param similarity_threshold Threshold for considering a prefix similar (0-1)
#' @param normalize_case Whether to normalize column case (lowercase, etc)
#' @param normalize_type Type of case normalization ("lower", "upper", "title", etc)
#' @return A data frame with prefixed and normalized column names
#' @importFrom stringdist stringdist
#' @noRd
addPrefixToColumnss <- function(df, collection_name, 
                                  excluded_cols = c("src_subject_id", "visit", "phenotype", 
                                                    "interview_age", "interview_date", "site", 
                                                    "sex", "subjectkey", "arm", "handedness", 
                                                    "state", "status"),
                                  similarity_threshold = 0.8,
                                  normalize_case = FALSE,
                                  normalize_type = "lower") {
  
  # Ensure stringdist package is available
  if (!requireNamespace("stringdist", quietly = TRUE)) {
    stop("The stringdist package is required. Please install it with install.packages('stringdist')")
  }
  
  # Helper function to check if a column has a similar prefix
  has_similar_prefix <- function(col_name, prefix) {
    # If the column is too short, it can't have a meaningful prefix
    if (nchar(col_name) < nchar(prefix)) {
      return(FALSE)
    }
    
    # Extract potential prefix (same length as collection_name)
    potential_prefix <- substring(col_name, 1, nchar(prefix))
    
    # Calculate similarity
    similarity <- 1 - stringdist::stringdist(tolower(potential_prefix), 
                                             tolower(prefix), 
                                             method = "jw") / nchar(prefix)
    
    return(similarity >= similarity_threshold)
  }
  
  # Get the current column names
  col_names <- names(df)
  
  # Replace dots with underscores
  col_names <- gsub("\\.", "_", col_names)
  
  # Create a new set of column names
  new_col_names <- sapply(col_names, function(col) {
    # Check if column should be excluded from prefixing
    if (col %in% excluded_cols) {
      return(col)
    }
    
    # Check if the column already has the exact prefix
    if (startsWith(col, paste0(collection_name, "_"))) {
      return(col)
    }
    
    # Check if the column has a similar prefix using string distance
    if (has_similar_prefix(col, collection_name)) {
      return(col)  # Keep as is if it has a similar prefix
    }
    
    # Add the prefix if none of the above conditions are met
    return(paste0(collection_name, "_", col))
  })
  
  # Normalize case if requested
  if (normalize_case) {
    if (normalize_type == "lower") {
      new_col_names <- tolower(new_col_names)
    } else if (normalize_type == "upper") {
      new_col_names <- toupper(new_col_names)
    } else if (normalize_type == "title") {
      # Title case (capitalize first letter of each word)
      new_col_names <- gsub("(^|_)([a-z])", "\\1\\U\\2", tolower(new_col_names), perl = TRUE)
    }
  }
  
  # Rename the columns
  names(df) <- new_col_names
  
  return(df)
}

#' CategoricalVariables R6 Class
#'
#' @description
#' Manages categorical variables that should be preserved during data processing.
#' Categorizes variables by statistical type (nominal vs ordinal).
#'
#' These are STUDY-SPECIFIC categorical fields that contain meaningful data,
#' as opposed to StandardOutput fields which are API-generated metadata.
#'
#' Defaults:
#' - Nominal (no inherent order): phenotype
#' - Ordinal (ordered): visit, week
#'
#' @keywords internal
#' @noRd
CategoricalVariables <- R6::R6Class(
  "CategoricalVariables",
  
  public = list(
    #' @field nominal Character vector - Nominal categorical variables (no inherent order)
    #' Examples: phenotype, site, arm
    nominal = character(0),
    
    #' @field ordinal Character vector - Ordinal categorical variables (ordered)
    #' Examples: visit, week, session, timepoint
    ordinal = character(0),
    
    #' @description
    #' Initialize with default categorical variables
    initialize = function() {
      self$nominal <- c("phenotype")
      self$ordinal <- c("visit", "week")
    },
    
    #' @description
    #' Get all categorical variables (all subtypes combined)
    #'
    #' @return Character vector of all categorical variable names
    get_all = function() {
      unique(c(self$nominal, self$ordinal))
    },
    
    #' @description
    #' Get categorical variables by subtype
    #'
    #' @param subtype Character - Either "nominal" or "ordinal"
    #'
    #' @return Character vector of variable names for that subtype
    get_by_type = function(subtype = c("nominal", "ordinal")) {
      subtype <- match.arg(subtype)
      self[[subtype]]
    },
    
    #' @description
    #' Check if a field is categorical
    #'
    #' @param field Character - Field name to check
    #'
    #' @return Logical - TRUE if field is categorical (in any subtype)
    is_categorical = function(field) {
      if (!is.character(field) || length(field) != 1) {
        stop("field must be a single character string")
      }
      field %in% self$get_all()
    },
    
    #' @description
    #' Get categorical variables that exist in a dataframe
    #'
    #' @param df Data frame to check
    #'
    #' @return Character vector of categorical variables found in df
    get_existing = function(df) {
      if (!is.data.frame(df)) {
        return(character(0))
      }
      all_categorical <- self$get_all()
      intersect(all_categorical, names(df))
    }
  )
)

#' Get Oracle connection parameters from secrets
#'
#' Retrieves connection parameters (DSN, DBQ, or host) from secrets configuration.
#' Returns a list indicating which type of connection to use.
#'
#' @return A list with `use_dbq` (logical) and `value` (character) elements
#' @noRd
get_oracle_connection_params <- function() {
  # Get connection parameters, supporting both DSN and DBQ (TNS alias)
  # DBQ is used for TNS aliases from tnsnames.ora (Oracle convention: uppercase)
  # DSN is used for ODBC Data Source Names (Oracle convention: uppercase)
  
  dsn <- get_secret_optional("DSN")
  dbq <- get_secret_optional("DBQ")
  host <- get_secret_optional("host")  # Fallback for backward compatibility (lowercase for legacy)
  
  # If DBQ is specified, use it (TNS alias)
  if (!is.null(dbq) && nchar(dbq) > 0) {
    return(list(use_dbq = TRUE, value = trimws(dbq)))
  }
  
  # If DSN is specified, use it
  if (!is.null(dsn) && nchar(dsn) > 0) {
    return(list(use_dbq = FALSE, value = trimws(dsn)))
  }
  
  # Fallback to host (treat as dsn for backward compatibility)
  if (!is.null(host) && nchar(host) > 0) {
    return(list(use_dbq = FALSE, value = trimws(host)))
  }
  
  stop("No connection target specified. Please set 'DSN', 'DBQ', or 'host' in secrets.R")
}

#'
#' @noRd
get_oracle_driver <- function(validate = FALSE) {
  # Try to get driver from secrets first
  driver <- get_secret_optional("driver")
  
  if (!is.null(driver) && nchar(driver) > 0) {
    # Trim whitespace from driver name
    driver <- trimws(driver)
    
    # If validation requested, check if driver exists (with case-insensitive matching)
    if (validate && requireNamespace("odbc", quietly = TRUE)) {
      tryCatch({
        available_drivers <- odbc::odbcListDrivers()
        # Check exact match first
        if (!driver %in% available_drivers$name) {
          # Try case-insensitive match
          driver_lower <- tolower(driver)
          available_lower <- tolower(available_drivers$name)
          matched_idx <- which(available_lower == driver_lower)
          
          if (length(matched_idx) > 0) {
            # Found case-insensitive match, use the actual name from the list
            driver <- available_drivers$name[matched_idx[1]]
            message(sprintf("Note: Using driver name '%s' (case-adjusted from secrets.R)", driver))
          } else {
            # No match found, but don't fail - let the connection attempt handle it
            warning(sprintf("Driver '%s' from secrets.R not found in available drivers list.", driver))
            message("Available drivers:")
            print(available_drivers)
            message("Will attempt connection anyway - driver name may still work.")
          }
        }
      }, error = function(e) {
        # If we can't check, proceed anyway
        message("Could not validate driver name: ", e$message)
      })
    }
    return(driver)
  }
  
  # If not in secrets, try to find an Oracle driver automatically
  if (requireNamespace("odbc", quietly = TRUE)) {
    tryCatch({
      available_drivers <- odbc::odbcListDrivers()
      oracle_drivers <- available_drivers[grepl("Oracle", available_drivers$name, ignore.case = TRUE), ]
      if (nrow(oracle_drivers) > 0) {
        # Return the first Oracle driver found
        return(oracle_drivers$name[1])
      }
    }, error = function(e) {
      # Silently fail and return NULL
    })
  }
  
  # If no driver found, return NULL (will try connection without explicit driver)
  return(NULL)
}

#' Fetch data from Oracle database to be stored in a data frame
#'
#' Retrieves data from an Oracle table or view and optionally joins it with a primary keys table
#' as specified in the configuration.
#'
#' @param table_name Name of the SQL table or view to query
#' @param ... Optional column names to filter for. Only rows with non-missing values
#'        in ALL specified columns will be returned.
#' @param fields Optional vector of specific fields to select
#' @param where_clause Optional WHERE clause to filter results (without the "WHERE" keyword)
#' @param join_primary_keys Boolean, whether to join with the primary keys table (default: TRUE)
#' @param custom_query Optional custom SQL query to execute instead of building one
#' @param max_rows Optional limit on the number of rows to return
#' @param date_format Optional format for date fields (default uses ISO format)
#' @param batch_size Number of records to retrieve per batch for large datasets
#' @param pii Logical; if FALSE (default), remove fields marked as PII. TRUE keeps PII.
#' @param interview_date Optional; can be either:
#'        - A date string in various formats (ISO, US, etc.) to filter data up to that date
#'        - A boolean TRUE to return only rows with non-NA interview_date values
#' @param all Logical; if TRUE, use LEFT OUTER JOIN instead of INNER JOIN (default: FALSE),
#'        similar to the 'all' parameter in base R's merge() function
#' @param schema Optional schema name to use for table qualification
#'
#' @return A data frame containing the requested SQL data
#' @importFrom odbc dbConnect dbGetInfo dbListTables dbGetQuery dbDisconnect
#' @export
#' @examples
#' \dontrun{
#' # Get data from a specific table
#' data <- oracle("participants")
#'
#' # Get data with a where clause
#' survey_data <- oracle("vw_surveyquestionresults",
#'                   where_clause = "resultidentifier = 'NRS'")
#'
#' # Get all records, including those without matching primary key
#' all_data <- oracle("candidate", all = TRUE)
#'
#' # Specify schema explicitly
#' schema_data <- oracle("survey_results", schema = "STUDY_DATA")
#' }
oracle <- function(table_name = NULL, ..., fields = NULL, where_clause = NULL,
                   join_primary_keys = TRUE, custom_query = NULL, max_rows = NULL,
                   date_format = NULL, batch_size = 1000, pii = FALSE,
                   interview_date = NULL, all = FALSE, schema = NULL) {

  # Check if required packages are available
  if (!requireNamespace("odbc", quietly = TRUE)) {
    stop("Package 'odbc' is needed for this function to work. Please install it.",
         call. = FALSE)
  }

  start_time <- Sys.time()

  # Validate secrets and config - MUST COME FIRST
  validate_secrets("sql")
  config <- validate_config("sql")

  # If no schema provided and config specifies a default schema, use it
  if (is.null(schema)) {
    cfg_schema <- tryCatch({
      if (!is.null(config) && !is.null(config$sql) && !is.null(config$sql$database)) {
        as.character(config$sql$database)
      } else NULL
    }, error = function(e) NULL)
    if (!is.null(cfg_schema) && nzchar(cfg_schema)) {
      schema <- cfg_schema
    }
  }

  # If table_name includes schema (e.g., SCHEMA.TABLE) and no schema arg provided,
  # extract schema for consistent handling and checks
  if (is.null(schema) && !is.null(table_name) && grepl("\\.", table_name)) {
    parts <- strsplit(table_name, "\\.")[[1]]
    if (length(parts) >= 2) {
      schema <- parts[1]
      table_name <- parts[2]
      message(sprintf("Using schema '%s' from table name", schema))
    }
  }

  # Initialize loading animation
  pb <- initializeLoadingAnimation(20)

  # Validate parameters
  if (is.null(table_name) && is.null(custom_query)) {
    tables_info <- oracle.index(schema)
    if (is.null(tables_info)) {
      stop("No table name or custom query provided, and could not retrieve table list")
    }
    stop("No table name or custom query provided. Use oracle.index() to see available tables.")
  }

  # Set Oracle environment variables from secrets if they exist
  tns_admin <- get_secret_optional("TNS_ADMIN")
  oracle_home <- get_secret_optional("ORACLE_HOME")
  if (!is.null(tns_admin) && nchar(tns_admin) > 0) {
    Sys.setenv(TNS_ADMIN = tns_admin)
  }
  if (!is.null(oracle_home) && nchar(oracle_home) > 0) {
    Sys.setenv(ORACLE_HOME = oracle_home)
  }

  # Get connection parameters (supports both DSN and DBQ for TNS aliases)
  conn_params <- get_oracle_connection_params()
  user_id <- get_secret("uid")
  password <- get_secret("pwd")
  driver <- get_oracle_driver()

  # Display loading message
  if (!is.null(custom_query)) {
    message(sprintf("\nExecuting custom SQL query..."))
  } else {
    message(sprintf("\nRetrieving data from Oracle table: %s%s%s",
                    ifelse(!is.null(schema), paste0(schema, "."), ""),
                    table_name,
                    ifelse(!is.null(where_clause), sprintf(" (with filters)"), "")))
  }

  # Update progress bar
  for (i in 1:10) {
    updateLoadingAnimation(pb, i)
    Sys.sleep(0.05)
  }

  # Establish database connection
  channel <- NULL
  result_data <- NULL

  tryCatch({
    # Build connection parameters with Driver as first parameter if available
    # Use DBQ for TNS aliases, dsn for ODBC DSNs
    conn_args <- list(
      UID = user_id,
      PWD = password
    )
    
    if (!is.null(driver)) {
      conn_args$Driver <- driver
    }
    
    if (conn_params$use_dbq) {
      conn_args$DBQ <- conn_params$value
    } else {
      conn_args$dsn <- conn_params$value
    }
    
    channel <- do.call(odbc::dbConnect, c(list(odbc::odbc()), conn_args))
    if (is.null(channel)) {
      stop("Failed to connect to database. Please check your connection settings.")
    }

    # Get database type information for diagnostic purposes
    db_info <- tryCatch({
      odbc::dbGetInfo(channel)
    }, error = function(e) {
      list(dbms.name = "Oracle")  # Default to Oracle since we're using the oracle function
    })

    message(sprintf("Connected to %s database", db_info$dbms.name))

    # Update progress bar
    for (i in 11:15) {
      updateLoadingAnimation(pb, i)
      Sys.sleep(0.05)
    }

    # Define the primary key information with proper validation
    primary_key_column <- NULL
    superkey_table <- NULL

    if (join_primary_keys) {
      # Check config structure like your SQL functions do
      if (!is.null(config) && !is.null(config$sql) && !is.null(config$sql$primary_key)) {
        primary_key_column <- config$sql$primary_key
      } else {
        message("Warning: Primary key not specified in config. Using default 'PARTICIPANTIDENTIFIER'")
        primary_key_column <- "PARTICIPANTIDENTIFIER"
      }

      if (!is.null(config) && !is.null(config$sql) && !is.null(config$sql$superkey)) {
        superkey_table <- config$sql$superkey
      } else {
        message("Warning: Superkey table not specified in config. Join with primary keys disabled.")
        join_primary_keys <- FALSE
      }
    }

    # Determine if the requested table is the configured superkey table
    is_superkey_request <- FALSE
    if (!is.null(superkey_table) && !is.null(table_name)) {
      # Extract table name without schema for comparison
      table_name_only <- table_name
      
      # Extract superkey table name without schema for comparison
      superkey_name_only <- if (grepl("\\.", superkey_table)) {
        strsplit(superkey_table, "\\.")[[1]][2]
      } else {
        superkey_table
      }
      
      # Check if the requested table matches the superkey table
      is_superkey_request <- identical(trimws(toupper(table_name_only)), trimws(toupper(superkey_name_only)))
      
      if (is_superkey_request) {
        message("Requested table matches configured superkey; returning without joins.")
        join_primary_keys <- FALSE
      }
    }

    # Determine fields to exclude based on PII settings with proper validation
    pii_fields <- character(0)
    if (!pii && !is.null(config) && !is.null(config$sql) && !is.null(config$sql$pii_fields)) {
      pii_fields <- config$sql$pii_fields
      if (length(pii_fields) > 0) {
        message(sprintf("Will exclude %d PII fields: %s",
                        length(pii_fields),
                        paste(pii_fields, collapse = ", ")))
      }
    }

    # Check if table exists and get correct schema if needed
    table_exists <- FALSE
    if (!is.null(table_name)) {
      tables_list <- odbc::dbListTables(channel, schema_name = schema)
      # Convert table name to uppercase for case-insensitive comparison
      # (Oracle typically stores object names in uppercase)
      upper_table_name <- toupper(table_name)

      if (!is.null(schema)) {
        # Check if table exists in specified schema
        upper_schema <- toupper(schema)
        table_exists <- any(toupper(tables_list) == upper_table_name)
        if (!table_exists) {
          # Try with exact case
          table_exists <- any(tables_list == table_name)
        }
      } else {
        # Check if table exists in any schema
        table_exists <- any(toupper(tables_list) == upper_table_name)
        # If table exists but no schema specified, get the schema
        if (table_exists) {
          message(sprintf("Found table: %s", table_name))
        }
      }

      if (!table_exists && is.null(custom_query)) {
        warning(sprintf("Table '%s' not found in the database. Attempting query anyway...", table_name))
      }
    }

    # Build or execute SQL query
    query <- NULL
    if (!is.null(custom_query)) {
      # Execute custom query directly
      query <- custom_query
    } else {
      # For Oracle syntax
      is_oracle <- TRUE  # Always TRUE for oracle() function

      # Build query based on parameters
      query <- build_oracle_query(
        table_name = table_name,
        fields = fields,
        where_clause = where_clause,
        superkey_table = if (join_primary_keys) superkey_table else NULL,
        primary_key_column = if (join_primary_keys) primary_key_column else NULL,
        max_rows = max_rows,
        pii_fields = if (!pii) pii_fields else NULL,
        schema = schema,
        all = all,
        is_oracle = is_oracle
      )
    }

    message(sprintf("Executing query: %s", query))

    # Execute the query
    result_data <- odbc::dbGetQuery(channel, query)

    # Check if result is an error message
    if (is.character(result_data) && length(result_data) == 1) {
      stop(paste("SQL error:", result_data))
    }

    # Ensure we have a data frame, even if empty
    if (!is.data.frame(result_data)) {
      result_data <- data.frame()
    }

    # Update progress bar
    for (i in 16:20) {
      updateLoadingAnimation(pb, i)
      Sys.sleep(0.05)
    }

    completeLoadingAnimation(pb)

    # Process and clean the data
    if (is.data.frame(result_data) && nrow(result_data) > 0) {
      # Apply PII exclusion if needed
      if (!pii && !is.null(pii_fields) && length(pii_fields) > 0) {
        pii_cols_in_result <- intersect(names(result_data), pii_fields)
        pii_cols_missing <- setdiff(pii_fields, names(result_data))
        if (length(pii_cols_in_result) > 0) {
          message("Removing PII columns from result: ", paste(pii_cols_in_result, collapse = ", "))
          result_data <- result_data[, !(names(result_data) %in% pii_fields), drop = FALSE]
        }
        if (length(pii_cols_missing) > 0) {
          message("Configured PII columns not present in result (skipped): ", paste(pii_cols_missing, collapse = ", "))
        }
      }

      # Process date fields
      result_data <- process_date_fields(result_data, date_format)

      # Handle missing values using the configuration
      result_data <- handle_missing_values(result_data, config)

      # Handle interview_date filtering if specified
      if (!is.null(interview_date) && "interview_date" %in% names(result_data)) {
        result_data <- filter_by_interview_date(result_data, interview_date)
      }

      # Standardize column names for key fields
      age_cols <- grep("_interview_age$", base::names(result_data), ignore.case = TRUE)
      if (length(age_cols) > 0) {
        base::names(result_data)[age_cols] <- "interview_age"
      }

      date_patterns <- c("_interview_date$", "interview_date")
      date_cols <- NULL
      for (pattern in date_patterns) {
        found_cols <- grep(pattern, base::names(result_data), ignore.case = TRUE)
        if (length(found_cols) > 0) {
          date_cols <- found_cols
          break  # Stop at first pattern that finds matches
        }
      }

      if (!is.null(date_cols) && length(date_cols) > 0) {
        base::names(result_data)[date_cols] <- "interview_date"
      }

      # Process column filtering based on ... parameters
      dots_args <- list(...)
      if (length(dots_args) > 0) {
        result_data <- filter_by_column_values(result_data, dots_args)
      }
    } else {
      # Create an empty data frame if no results
      if (!is.data.frame(result_data)) {
        result_data <- data.frame()
      }
      message("No data returned from query")
    }

  }, error = function(e) {
    stop(paste("Error:", e$message))
  }, finally = {
    # Close the database connection - fixed the issue here
    if (!is.null(channel)) {
      tryCatch({
        odbc::dbDisconnect(channel)
      }, error = function(e) {
        # Silently ignore close errors
      })
    }
  })

  # Add metadata to the result
  attr(result_data, "sql_table") <- table_name
  if (!is.null(schema)) {
    attr(result_data, "schema") <- schema
  }
  if (!is.null(where_clause)) {
    attr(result_data, "where_clause") <- where_clause
  }

  # Show duration
  end_time <- Sys.time()
  duration <- difftime(end_time, start_time, units = "secs")
  message(sprintf("\nData frame retrieved from Oracle in %s with %d rows and %d columns.",
                  formatDuration(duration),
                  nrow(result_data),
                  ncol(result_data)))

  return(result_data)
}

#' Get a list of tables from the Oracle database
#'
#' @param schema Optional schema name to filter tables
#' @return A data frame with table information
#' @importFrom odbc dbConnect dbListTables dbDisconnect
#' @export
oracle.index <- function(schema = NULL) {

  # Validate secrets
  validate_secrets("sql")
  config <- validate_config("sql")

  # Set Oracle environment variables from secrets if they exist
  tns_admin <- get_secret_optional("TNS_ADMIN")
  oracle_home <- get_secret_optional("ORACLE_HOME")
  if (!is.null(tns_admin) && nchar(tns_admin) > 0) {
    Sys.setenv(TNS_ADMIN = tns_admin)
  }
  if (!is.null(oracle_home) && nchar(oracle_home) > 0) {
    Sys.setenv(ORACLE_HOME = oracle_home)
  }

  # Get connection parameters (supports both DSN and DBQ for TNS aliases)
  conn_params <- get_oracle_connection_params()
  user_id <- get_secret("uid")
  password <- get_secret("pwd")
  driver <- get_oracle_driver()

  # Connect to database
  channel <- NULL
  tables <- NULL

  tryCatch({
    # Build connection parameters with Driver as first parameter if available
    # Use DBQ for TNS aliases, dsn for ODBC DSNs
    conn_args <- list(
      UID = user_id,
      PWD = password
    )
    
    if (!is.null(driver)) {
      conn_args$Driver <- driver
    }
    
    if (conn_params$use_dbq) {
      conn_args$DBQ <- conn_params$value
    } else {
      conn_args$dsn <- conn_params$value
    }
    
    channel <- do.call(odbc::dbConnect, c(list(odbc::odbc()), conn_args))
    if (is.null(channel)) {
      stop("Failed to connect to database")
    }

    # Get database type for diagnostics
    db_info <- tryCatch({
      odbc::dbGetInfo(channel)
    }, error = function(e) {
      list(dbms.name = "Oracle")  # Default to Oracle since we're using oracle.index
    })

    message(sprintf("Connected to %s database", db_info$dbms.name))

    # Oracle-specific handling
    is_oracle <- TRUE  # Always TRUE for oracle.index

    # Get table list
    if (is.null(schema)) {
      # Get all tables
      tables <- odbc::dbListTables(channel)
      message("Retrieving tables from all schemas")
    } else {
      # Get tables from specific schema
      # Note: Oracle treats schema name as case-sensitive
      tables <- odbc::dbListTables(channel, schema_name = schema)
      message(sprintf("Retrieving tables from schema: %s", schema))

      # If no tables found, try uppercase (Oracle often stores schemas in uppercase)
      if (is_oracle && (is.null(tables) || length(tables) == 0)) {
        upper_schema <- toupper(schema)
        tables <- odbc::dbListTables(channel, schema_name = upper_schema)
        message(sprintf("Trying uppercase schema: %s", upper_schema))
      }
    }

    # If still no tables, try direct query for Oracle
    if (is_oracle && (is.null(tables) || length(tables) == 0)) {
      message("Trying direct Oracle query to list tables...")
      if (is.null(schema)) {
        # Query all accessible tables in Oracle
        query <- "SELECT owner AS TABLE_SCHEM, table_name AS TABLE_NAME,
                 'TABLE' AS TABLE_TYPE FROM all_tables
                 UNION
                 SELECT owner AS TABLE_SCHEM, view_name AS TABLE_NAME,
                 'VIEW' AS TABLE_TYPE FROM all_views
                 ORDER BY 1, 2"
      } else {
        # Query tables from specific schema in Oracle
        query <- sprintf(
          "SELECT owner AS TABLE_SCHEM, table_name AS TABLE_NAME,
           'TABLE' AS TABLE_TYPE FROM all_tables WHERE owner = '%s'
           UNION
           SELECT owner AS TABLE_SCHEM, view_name AS TABLE_NAME,
           'VIEW' AS TABLE_TYPE FROM all_views WHERE owner = '%s'
           ORDER BY 1, 2",
          schema, schema
        )
      }
      tables <- odbc::dbGetQuery(channel, query)
    }

  }, error = function(e) {
    message(paste("Error retrieving tables:", e$message))
  }, finally = {
    # Fixed connection close handling
    if (!is.null(channel)) {
      tryCatch({
        odbc::dbDisconnect(channel)
      }, error = function(e) {
        # Silently ignore close errors
      })
    }
  })

  # Format the result
  if (!is.null(tables) && length(tables) > 0) {
    # If tables is a character vector (from dbListTables), convert to data frame
    if (is.character(tables)) {
      tables_df <- data.frame(
        Schema = rep(ifelse(is.null(schema), "Unknown", schema), length(tables)),
        Table = tables,
        Type = rep("TABLE", length(tables)),
        stringsAsFactors = FALSE
      )
    } else if (is.data.frame(tables)) {
      # Keep only relevant columns
      if (all(c("TABLE_SCHEM", "TABLE_NAME", "TABLE_TYPE") %in% names(tables))) {
        tables_df <- tables[, c("TABLE_SCHEM", "TABLE_NAME", "TABLE_TYPE")]
        names(tables_df) <- c("Schema", "Table", "Type")
      } else if (ncol(tables) >= 3) {
        # Handle case where column names might be different
        tables_df <- tables[, 1:3]
        names(tables_df) <- c("Schema", "Table", "Type")
      } else {
        tables_df <- tables
      }
    } else {
      tables_df <- data.frame()
    }

    # Sort by schema and table name
    if (nrow(tables_df) > 0) {
      tables_df <- tables_df[order(tables_df$Schema, tables_df$Table), ]
      return(knitr::kable(tables_df, format = "simple"))
    }
  }
  
  message("No tables found")
  return(NULL)
}

#' Get Oracle table columns/metadata
#'
#' @param table_name Name of the table to get metadata for
#' @param schema Optional schema name
#' @return A data frame with column information
#' @importFrom odbc dbConnect dbListFields dbGetQuery dbDisconnect
#' @export
oracle.desc <- function(table_name, schema = NULL) {

  if (is.null(table_name)) {
    stop("Table name is required")
  }

  # Validate secrets
  validate_secrets("sql")
  config <- validate_config("sql")

  # Set Oracle environment variables from secrets if they exist
  tns_admin <- get_secret_optional("TNS_ADMIN")
  oracle_home <- get_secret_optional("ORACLE_HOME")
  if (!is.null(tns_admin) && nchar(tns_admin) > 0) {
    Sys.setenv(TNS_ADMIN = tns_admin)
  }
  if (!is.null(oracle_home) && nchar(oracle_home) > 0) {
    Sys.setenv(ORACLE_HOME = oracle_home)
  }

  # Parse table_name to extract schema if not provided separately
  if (is.null(schema) && grepl("\\.", table_name)) {
    parts <- strsplit(table_name, "\\.")[[1]]
    if (length(parts) > 1) {
      schema <- parts[1]
      table_name <- parts[2]
      message(sprintf("Using schema '%s' from table name", schema))
    }
  }

  # Get connection parameters (supports both DSN and DBQ for TNS aliases)
  conn_params <- get_oracle_connection_params()
  user_id <- get_secret("uid")
  password <- get_secret("pwd")
  driver <- get_oracle_driver()

  # Connect to database
  channel <- NULL
  columns <- NULL

  tryCatch({
    # Build connection parameters with Driver as first parameter if available
    # Use DBQ for TNS aliases, dsn for ODBC DSNs
    conn_args <- list(
      UID = user_id,
      PWD = password
    )
    
    if (!is.null(driver)) {
      conn_args$Driver <- driver
    }
    
    if (conn_params$use_dbq) {
      conn_args$DBQ <- conn_params$value
    } else {
      conn_args$dsn <- conn_params$value
    }
    
    channel <- do.call(odbc::dbConnect, c(list(odbc::odbc()), conn_args))
    if (is.null(channel)) {
      stop("Failed to connect to database")
    }

    # Get database type for diagnostics
    db_info <- tryCatch({
      odbc::dbGetInfo(channel)
    }, error = function(e) {
      list(dbms.name = "Oracle")  # Default to Oracle for oracle.desc
    })

    message(sprintf("Connected to %s database", db_info$dbms.name))

    is_oracle <- TRUE  # Always TRUE for oracle.desc

    # Try to get column information
    # Oracle stores table names in uppercase, so try both original and uppercase
    columns <- NULL
    upper_table <- toupper(table_name)
    
    # First, try direct Oracle query (more reliable than dbListFields for Oracle)
    if (is_oracle) {
      tryCatch({
        # Try user_tab_columns first (current user's schema) if no schema specified
        if (is.null(schema)) {
          # Try current user's schema first
          oracle_query <- sprintf(
            "SELECT column_name, data_type, data_length,
             DECODE(nullable, 'Y', 'Yes', 'No') as nullable
             FROM user_tab_columns
             WHERE table_name = '%s'
             ORDER BY column_id",
            upper_table
          )
          columns <- odbc::dbGetQuery(channel, oracle_query)
          
          # If no results, try all_tab_columns (all accessible schemas)
          if (is.null(columns) || !is.data.frame(columns) || nrow(columns) == 0) {
            oracle_query <- sprintf(
              "SELECT owner as schema_name, column_name, data_type, data_length,
               DECODE(nullable, 'Y', 'Yes', 'No') as nullable
               FROM all_tab_columns
               WHERE table_name = '%s'
               ORDER BY owner, column_id",
              upper_table
            )
            columns <- odbc::dbGetQuery(channel, oracle_query)
          }
        } else {
          # With schema specified
          upper_schema <- toupper(schema)
          oracle_query <- sprintf(
            "SELECT column_name, data_type, data_length,
             DECODE(nullable, 'Y', 'Yes', 'No') as nullable
             FROM all_tab_columns
             WHERE owner = '%s' AND table_name = '%s'
             ORDER BY column_id",
            upper_schema, upper_table
          )
          columns <- odbc::dbGetQuery(channel, oracle_query)
        }
        
        # Process the results
        if (!is.null(columns) && is.data.frame(columns) && nrow(columns) > 0) {
          # Rename columns to match expected format
          if ("OWNER" %in% names(columns) || "schema_name" %in% names(columns)) {
            # Has schema column
            schema_col <- if ("OWNER" %in% names(columns)) "OWNER" else "schema_name"
            col_order <- c(schema_col, "COLUMN_NAME", "DATA_TYPE", "DATA_LENGTH", "nullable")
            col_order <- col_order[col_order %in% names(columns)]
            columns <- columns[, col_order, drop = FALSE]
            names(columns) <- c("Schema", "Column", "Type", "Size", "Nullable")
          } else {
            # No schema column
            col_order <- c("COLUMN_NAME", "DATA_TYPE", "DATA_LENGTH", "nullable")
            col_order <- col_order[col_order %in% names(columns)]
            columns <- columns[, col_order, drop = FALSE]
            names(columns) <- c("Column", "Type", "Size", "Nullable")
          }
        } else {
          columns <- NULL
        }
      }, error = function(e) {
        message(paste("Error querying Oracle data dictionary:", e$message))
        columns <- NULL
      })
    }
    
    # Fallback to dbListFields if Oracle query didn't work
    if (is.null(columns) || !is.data.frame(columns) || nrow(columns) == 0) {
      tryCatch({
        # Try with uppercase table name for Oracle
        table_to_use <- if (is_oracle) upper_table else table_name
        if (!is.null(schema)) {
          columns <- odbc::dbListFields(channel, name = table_to_use, schema_name = schema)
        } else {
          columns <- odbc::dbListFields(channel, name = table_to_use)
        }
        
        # Convert to data frame format
        if (!is.null(columns) && length(columns) > 0) {
          columns_df <- data.frame(
            Column = columns,
            Type = rep("Unknown", length(columns)),
            Size = rep(NA, length(columns)),
            Nullable = rep("Unknown", length(columns)),
            stringsAsFactors = FALSE
          )
          columns <- columns_df
        }
      }, error = function(e) {
        # dbListFields failed, but we already tried Oracle query, so just note it
        if (is.null(columns)) {
          message(paste("Error using dbListFields:", e$message))
        }
      })
    }

  }, error = function(e) {
    message(paste("Error retrieving columns for table", table_name, ":", e$message))
  }, finally = {
    # Fixed connection close handling
    if (!is.null(channel)) {
      tryCatch({
        odbc::dbDisconnect(channel)
      }, error = function(e) {
        # Silently ignore close errors
      })
    }
  })

  # Format the result
  if (!is.null(columns) && is.data.frame(columns) && nrow(columns) > 0 && !is.na(nrow(columns))) {
    # Keep only relevant columns
    if (all(c("COLUMN_NAME", "TYPE_NAME", "COLUMN_SIZE", "NULLABLE") %in% names(columns))) {
      columns <- columns[, c("COLUMN_NAME", "TYPE_NAME", "COLUMN_SIZE", "NULLABLE")]
      # Rename NULLABLE column for clarity
      columns$NULLABLE <- ifelse(columns$NULLABLE == 1, "Yes", "No")
      # Rename columns
      names(columns) <- c("Column", "Type", "Size", "Nullable")
    } else if (!all(c("Column", "Type", "Size", "Nullable") %in% names(columns))) {
      # Try to use whatever columns we have
      message("Warning: Column metadata format is non-standard")
      return(columns)
    }
    return(knitr::kable(columns, format = "simple"))
  } else {
    message(paste("No columns found for table", table_name))
    return(NULL)
  }
}

#' Perform a direct Oracle query with minimal processing
#'
#' @param query The SQL query to execute
#' @param pii Logical; if FALSE (default), remove fields marked as PII. TRUE keeps PII.
#' @param schema Optional schema name to qualify table names in the query
#' @return A data frame with the query results
#' @importFrom odbc dbConnect dbGetQuery dbDisconnect
#' @export
oracle.query <- function(query, pii = FALSE, schema = NULL) {

  if (is.null(query) || !is.character(query) || length(query) != 1) {
    stop("A valid SQL query string is required")
  }

  # Validate secrets
  validate_secrets("sql")
  config <- validate_config("sql")

  # Set Oracle environment variables from secrets if they exist
  tns_admin <- get_secret_optional("TNS_ADMIN")
  oracle_home <- get_secret_optional("ORACLE_HOME")
  if (!is.null(tns_admin) && nchar(tns_admin) > 0) {
    Sys.setenv(TNS_ADMIN = tns_admin)
  }
  if (!is.null(oracle_home) && nchar(oracle_home) > 0) {
    Sys.setenv(ORACLE_HOME = oracle_home)
  }

  # Get connection parameters (supports both DSN and DBQ for TNS aliases)
  conn_params <- get_oracle_connection_params()
  user_id <- get_secret("uid")
  password <- get_secret("pwd")
  driver <- get_oracle_driver()

  # Get PII fields configuration if needed with proper validation
  pii_fields <- NULL
  if (!pii && !is.null(config) && !is.null(config$sql) && !is.null(config$sql$pii_fields)) {
    pii_fields <- config$sql$pii_fields
    if (length(pii_fields) > 0) {
      message(sprintf("Will exclude %d PII fields if present in results: %s",
                      length(pii_fields),
                      paste(pii_fields, collapse=", ")))
    }
  }

  # Modify query to include schema if provided
  if (!is.null(schema) && !grepl("\\.", query)) {
    # Simple schema qualification for table references
    # This is a basic approach - a proper SQL parser would be better
    query <- gsub("\\b(FROM|JOIN)\\s+(\\w+)",
                  paste0("\\1 ", schema, ".\\2"),
                  query,
                  ignore.case = TRUE)
    message(sprintf("Using schema: %s", schema))
    message(sprintf("Modified query: %s", query))
  }

  # Connect to database
  channel <- NULL
  result <- NULL

  tryCatch({
    # Build connection parameters with Driver as first parameter if available
    # Use DBQ for TNS aliases, dsn for ODBC DSNs
    conn_args <- list(
      UID = user_id,
      PWD = password
    )
    
    if (!is.null(driver)) {
      conn_args$Driver <- driver
    }
    
    if (conn_params$use_dbq) {
      conn_args$DBQ <- conn_params$value
    } else {
      conn_args$dsn <- conn_params$value
    }
    
    # Attempt connection with better error handling for IM006
    tryCatch({
      channel <- do.call(odbc::dbConnect, c(list(odbc::odbc()), conn_args))
    }, error = function(e) {
      error_msg <- e$message
      if (grepl("IM006", error_msg, ignore.case = TRUE)) {
        message("\nTroubleshooting IM006 error (Driver's SQLSetConnectAttr failed):")
        if (conn_params$use_dbq) {
          message("1. Verify TNS_ADMIN is set correctly and points to directory containing tnsnames.ora")
          message(sprintf("2. Verify DBQ '%s' exists in tnsnames.ora file", conn_params$value))
          message("3. Check that ORACLE_HOME is set correctly")
          message("4. Verify the TNS alias name matches exactly (case-sensitive)")
        } else {
          message("1. Verify DSN is correctly configured in ODBC Data Source Administrator")
          message("2. Check for 32-bit vs 64-bit driver mismatch")
          message("3. Try recreating the DSN")
        }
        message("5. Ensure Oracle client is properly installed and configured")
        stop("Connection failed with IM006 error. See troubleshooting steps above.")
      } else {
        stop(error_msg)
      }
    })
    
    if (is.null(channel)) {
      stop("Failed to connect to database")
    }

    # Get database info for diagnostics
    db_info <- tryCatch({
      odbc::dbGetInfo(channel)
    }, error = function(e) {
      list(dbms.name = "Oracle")  # Default to Oracle for oracle.query
    })

    message(sprintf("Connected to %s database", db_info$dbms.name))

    # Execute query
    result <- odbc::dbGetQuery(channel, query)

    # Check if result is an error message
    if (is.character(result) && length(result) == 1) {
      stop(paste("SQL error:", result))
    }

    # Make sure we have a data frame
    if (!is.data.frame(result)) {
      result <- data.frame()
    }

    # Apply PII exclusion if enabled
    if (!pii && !is.null(pii_fields) && length(pii_fields) > 0 && nrow(result) > 0) {
      pii_cols_present <- intersect(names(result), pii_fields)
      if (length(pii_cols_present) > 0) {
        message(sprintf("Removing %d PII fields from results: %s",
                        length(pii_cols_present),
                        paste(pii_cols_present, collapse=", ")))
        result <- result[, !names(result) %in% pii_fields, drop = FALSE]
      }
    }

    # Apply missing value handling if configured with proper validation
    if (!is.null(config) && !is.null(config$missing_data_codes) && nrow(result) > 0) {
      result <- handle_missing_values(result, config)
    }

  }, error = function(e) {
    message(paste("Error executing query:", e$message))
    stop(paste("Oracle Error:", e$message))
  }, finally = {
    # Fixed connection close handling
    if (!is.null(channel)) {
      tryCatch({
        odbc::dbDisconnect(channel)
      }, error = function(e) {
        # Silently ignore close errors
      })
    }
  })

  return(result)
}

build_oracle_query <- function(table_name, fields = NULL, where_clause = NULL,
                               superkey_table = NULL, primary_key_column = NULL,
                               max_rows = NULL, pii_fields = NULL, all = FALSE,
                               schema = NULL, is_oracle = TRUE) {

  # Add schema prefix to table names if specified
  table_with_schema <- table_name
  superkey_with_schema <- superkey_table

  if (!is.null(schema)) {
    if (!grepl("\\.", table_name)) {
      table_with_schema <- paste0(schema, ".", table_name)
    }
    if (!is.null(superkey_table) && !grepl("\\.", superkey_table)) {
      superkey_with_schema <- paste0(schema, ".", superkey_table)
    }
  }

  # Define table aliases
  main_table_alias <- "t"
  pk_table_alias <- "pk"

  # Determine which fields to select
  if (is.null(fields)) {
    # No specific fields requested
    if (!is.null(superkey_table) && !is.null(primary_key_column)) {
      # When joining, include fields from both tables
      if (!is.null(pii_fields) && length(pii_fields) > 0) {
        # PII exclusion needed - this will be handled in the main function
        message("Note: PII exclusion with * requires listing all non-PII columns explicitly")
        select_clause <- paste0("SELECT ", main_table_alias, ".*, ", pk_table_alias, ".*")
      } else {
        # No PII exclusion needed - include all fields from both tables
        select_clause <- paste0("SELECT ", main_table_alias, ".*, ", pk_table_alias, ".*")
      }
    } else {
      # No join - just select from main table
      if (!is.null(pii_fields) && length(pii_fields) > 0) {
        message("Note: PII exclusion with * requires listing all non-PII columns explicitly")
        select_clause <- "SELECT *"
      } else {
        select_clause <- "SELECT *"
      }
    }
  } else {
    # Specific fields requested
    # Filter out PII fields if specified
    if (!is.null(pii_fields)) {
      fields <- setdiff(fields, pii_fields)
    }

    # If fields array is not empty after filtering
    if (length(fields) > 0) {
      # For joins, qualify field names with table alias if they don't already have one
      if (!is.null(superkey_table) && !is.null(primary_key_column)) {
        # Check if fields specify table aliases or need them
        qualified_fields <- character(0)
        for (field in fields) {
          if (grepl("\\.", field)) {
            # Field already has table qualifier (e.g., "t.sub_id" or "pk.phi_field")
            qualified_fields <- c(qualified_fields, field)
          } else {
            # Add main table alias to field names
            qualified_fields <- c(qualified_fields, paste0(main_table_alias, ".", field))
          }
        }
        select_clause <- paste0("SELECT ", paste(qualified_fields, collapse = ", "))
      } else {
        # No join, use fields as-is
        select_clause <- paste0("SELECT ", paste(fields, collapse = ", "))
      }
    } else {
      # All requested fields were PII, so select a dummy field
      select_clause <- "SELECT 1 AS dummy_column"
    }
  }

  # Build the FROM clause
  if (!is.null(superkey_with_schema) && !is.null(primary_key_column)) {
    # Determine join type based on 'all' parameter
    join_type <- ifelse(all, "LEFT OUTER JOIN", "INNER JOIN")

    # Join with superkey table - using Oracle syntax format
    # Ensure primary_key_column doesn't have schema prefix
    clean_primary_key <- if (grepl("\\.", primary_key_column)) {
      strsplit(primary_key_column, "\\.")[[1]][2]
    } else {
      primary_key_column
    }
    
    from_clause <- sprintf(
      "FROM %s %s\n%s %s %s\n    ON %s.%s = %s.%s",
      table_with_schema, main_table_alias,
      join_type,
      superkey_with_schema, pk_table_alias,
      main_table_alias, clean_primary_key,
      pk_table_alias, clean_primary_key
    )
  } else {
    from_clause <- sprintf("FROM %s %s", table_with_schema, main_table_alias)
  }

  # Add WHERE clause if provided
  where_part <- ""
  if (!is.null(where_clause) && nchar(where_clause) > 0) {
    where_part <- paste("WHERE", where_clause)
  }

  # Add LIMIT/TOP clause if max_rows is specified
  limit_part <- ""
  if (!is.null(max_rows) && is.numeric(max_rows) && max_rows > 0) {
    if (is_oracle) {
      # Oracle uses ROWNUM
      if (nchar(where_part) > 0) {
        where_part <- paste0(where_part, " AND ROWNUM <= ", as.integer(max_rows))
      } else {
        where_part <- paste0("WHERE ROWNUM <= ", as.integer(max_rows))
      }
    } else {
      # Many databases use TOP in the SELECT clause
      select_clause <- gsub(
        "SELECT DISTINCT",
        paste0("SELECT DISTINCT TOP ", as.integer(max_rows)),
        select_clause
      )
    }
  }

  # Combine all parts into the final query
  query <- paste(select_clause, from_clause, where_part)

  # For debugging
  message("Generated SQL query: ", query)

  return(query)
}

#' Process date fields in the dataset
#'
#' @param df The data frame to process
#' @param date_format The desired output date format
#'
#' @return The processed data frame
#' @noRd
process_date_fields <- function(df, date_format = NULL) {
  # If no format specified, leave dates in database format
  if (is.null(date_format)) {
    return(df)
  }

  # Identify potential date columns by name pattern
  date_patterns <- c("date", "dt", "timestamp")
  potential_date_cols <- character(0)

  for (pattern in date_patterns) {
    found_cols <- grep(pattern, names(df), value = TRUE, ignore.case = TRUE)
    potential_date_cols <- c(potential_date_cols, found_cols)
  }

  # Remove duplicates
  potential_date_cols <- unique(potential_date_cols)

  # Process each potential date column
  for (col in potential_date_cols) {
    # Only process if the column exists and has data
    if (col %in% names(df) && !all(is.na(df[[col]]))) {
      # Check if it's already a Date or POSIXct object
      if (inherits(df[[col]], "Date") || inherits(df[[col]], "POSIXct")) {
        # Format according to preference
        if (date_format == "mdy") {
          df[[col]] <- format(df[[col]], "%m/%d/%Y")
        } else if (date_format == "dmy") {
          df[[col]] <- format(df[[col]], "%d/%m/%Y")
        } else if (date_format == "ymd") {
          df[[col]] <- format(df[[col]], "%Y-%m-%d")
        }
      }
      # Check if it's character data that looks like dates
      else if (is.character(df[[col]])) {
        # Try to convert string dates to consistent format
        if (any(grepl("-|/|\\.", df[[col]][!is.na(df[[col]])]))) {
          tryCatch({
            parsed_dates <- parseAnyDate(df[[col]])
            if (!all(is.na(parsed_dates))) {
              # Format according to preference
              if (date_format == "mdy") {
                df[[col]] <- format(parsed_dates, "%m/%d/%Y")
              } else if (date_format == "dmy") {
                df[[col]] <- format(parsed_dates, "%d/%m/%Y")
              } else if (date_format == "ymd") {
                df[[col]] <- format(parsed_dates, "%Y-%m-%d")
              }
            }
          }, error = function(e) {
            message(sprintf("Note: Could not parse dates in column %s", col))
          })
        }
      }
    }
  }

  return(df)
}

#' Parse dates in various formats
#'
#' @param date_strings Vector of date strings to parse
#' @return Vector of Date objects
#' @noRd
parseAnyDate <- function(date_strings) {
  # Create a result vector of the same length as input
  result <- rep(as.Date(NA), length(date_strings))

  # Skip NA values
  mask_na <- !is.na(date_strings)
  if (!any(mask_na)) {
    return(result)
  }

  valid_strings <- date_strings[mask_na]

  # Try different date formats
  parsed <- tryCatch({
    # Try lubridate parsing functions in sequence
    if (requireNamespace("lubridate", quietly = TRUE)) {
      # Try ISO format (YYYY-MM-DD)
      parsed <- lubridate::ymd(valid_strings, quiet = TRUE)
      # If that fails, try US format (MM/DD/YYYY)
      if (all(is.na(parsed))) {
        parsed <- lubridate::mdy(valid_strings, quiet = TRUE)
      }
      # If that fails, try European format (DD/MM/YYYY)
      if (all(is.na(parsed))) {
        parsed <- lubridate::dmy(valid_strings, quiet = TRUE)
      }
      parsed
    } else {
      # Fallback if lubridate is not available
      as.Date(valid_strings)
    }
  }, error = function(e) {
    rep(as.Date(NA), length(valid_strings))
  })

  # Update the result with parsed dates
  result[mask_na] <- parsed

  return(result)
}

#' Filter data frame by interview_date
#'
#' @param df The data frame to filter
#' @param interview_date The interview date filter value
#' @return The filtered data frame
#' @noRd
filter_by_interview_date <- function(df, interview_date) {
  if (is.null(interview_date) || !("interview_date" %in% names(df))) {
    return(df)
  }

  # Create a temporary date column for filtering
  df$temp_date <- parseAnyDate(df$interview_date)

  if (is.logical(interview_date) && interview_date == TRUE) {
    # Keep only rows with non-NA interview_date values
    df <- df[!is.na(df$temp_date), ]
  } else if (is.character(interview_date) || inherits(interview_date, "Date")) {
    # Filter by specific date
    target_date <- if (inherits(interview_date, "Date")) {
      interview_date
    } else {
      tryCatch({
        parseAnyDate(interview_date)
      }, error = function(e) {
        stop("Failed to parse interview_date parameter: ", interview_date)
      })
    }

    if (is.na(target_date)) {
      stop("Failed to parse interview_date parameter: ", interview_date)
    }

    # Keep rows with dates before or equal to target_date
    df <- df[!is.na(df$temp_date) & df$temp_date <= target_date, ]
  }

  # Remove temporary column
  df$temp_date <- NULL

  return(df)
}

#' Filter data frame by column values
#'
#' @param df The data frame to filter
#' @param cols_to_check List of column names to check for non-NA values
#' @return The filtered data frame
#' @noRd
filter_by_column_values <- function(df, cols_to_check) {
  if (length(cols_to_check) == 0) {
    return(df)
  }

  # Convert to character vector
  requested_cols <- as.character(unlist(cols_to_check))

  # Find which columns exist in the data
  existing_cols <- intersect(requested_cols, names(df))

  if (length(existing_cols) > 0) {
    # Display columns found
    message(sprintf("Found %d of %d requested columns: %s",
                    length(existing_cols),
                    length(requested_cols),
                    paste(existing_cols, collapse = ", ")))

    # Create a filter to keep rows with data in all requested columns
    rows_to_keep <- rep(TRUE, nrow(df))

    for (col in existing_cols) {
      # Check for non-NA values
      not_na <- !is.na(df[[col]])

      # For character columns, also check for non-empty strings
      not_empty <- rep(TRUE, nrow(df))
      if (is.character(df[[col]])) {
        not_empty <- df[[col]] != ""
      }

      # Update filter
      rows_to_keep <- rows_to_keep & not_na & not_empty
    }

    # Apply the filter
    original_rows <- nrow(df)
    df <- df[rows_to_keep, ]
    kept_rows <- nrow(df)

    message(sprintf("Kept %d of %d rows where all requested columns have values.",
                    kept_rows, original_rows))
  } else {
    warning("None of the requested columns were found in the dataset.")
  }

  return(df)
}

#' Format time duration in a human-readable way
#'
#' @param duration The duration to format
#' @return A formatted string
#' @noRd
formatDuration <- function(duration) {
  secs <- as.numeric(duration, units = "secs")

  if (secs < 60) {
    return(sprintf("%.1f seconds", secs))
  } else {
    mins <- floor(secs / 60)
    remaining_secs <- round(secs %% 60, 1)
    if (remaining_secs > 0) {
      return(sprintf("%d minutes and %.1f seconds", mins, remaining_secs))
    } else {
      return(sprintf("%d minutes", mins))
    }
  }
}

#' Initialize a loading animation
#'
#' @param steps Total number of steps
#' @return A list with animation state
#' @noRd
initializeLoadingAnimation <- function(steps) {
  # Get console width
  width <- tryCatch({
    if (requireNamespace("cli", quietly = TRUE)) {
      cli::console_width() - 10  # Leave some margin
    } else {
      getOption("width", 80) - 10  # Fallback to R's width setting
    }
  }, error = function(e) 80)  # Default if all else fails

  list(
    steps = steps,
    current = 0,
    width = width,
    start_time = Sys.time()
  )
}

#' Update the loading animation
#'
#' @param pb The progress bar object
#' @param current Current step
#' @return Nothing
#' @noRd
updateLoadingAnimation <- function(pb, current) {
  pb$current <- current
  percentage <- round(current / pb$steps * 100)
  filled <- round(pb$width * current / pb$steps)

  bar <- paste0(
    strrep("=", filled),
    strrep(" ", pb$width - filled)
  )

  cat(sprintf("\r|%s| %3d%%", bar, percentage))
  utils::flush.console()
}

#' Complete the loading animation
#'
#' @param pb The progress bar object
#' @return Nothing
#' @noRd
completeLoadingAnimation <- function(pb) {
  updateLoadingAnimation(pb, pb$steps)
  cat("\n")
}

#' Handle missing values based on configuration
#'
#' @param df The data frame to process
#' @param config The configuration containing missing value definitions
#' @return The processed data frame
#' @noRd
handle_missing_values <- function(df, config) {
  # If missing_data_codes is not configured, return the data frame as is
  if (is.null(config) || !is.list(config) || !("missing_data_codes" %in% names(config))) {
    return(df)
  }

  # Get missing value codes
  missing_codes <- config$missing_data_codes

  # Additional check that missing_codes is a list
  if (!is.list(missing_codes)) {
    return(df)
  }

  # Process each category of missing values
  transformed_count <- 0
  transformed_cols <- character(0)

  for (category in names(missing_codes)) {
    values <- missing_codes[[category]]

    # Skip if no values defined
    if (length(values) == 0) {
      next
    }

    # Replace values with NA in the data frame
    for (col in names(df)) {
      if (is.character(df[[col]])) {
        # For character columns, convert missing value codes to NA
        for (value in values) {
          # Case insensitive comparison for string values
          if (is.character(value)) {
            mask <- tolower(df[[col]]) == tolower(value)
            if (any(mask, na.rm = TRUE)) {
              matches <- sum(mask, na.rm = TRUE)
              df[[col]][mask] <- NA
              transformed_count <- transformed_count + matches
              # Track which columns were affected
              if (!(col %in% transformed_cols)) {
                transformed_cols <- c(transformed_cols, col)
              }
            }
          }
        }
      } else if (is.numeric(df[[col]])) {
        # For numeric columns, only match numeric missing codes
        for (value in values) {
          if (is.numeric(value)) {
            mask <- df[[col]] == value
            if (any(mask, na.rm = TRUE)) {
              matches <- sum(mask, na.rm = TRUE)
              df[[col]][mask] <- NA
              transformed_count <- transformed_count + matches
              # Track which columns were affected
              if (!(col %in% transformed_cols)) {
                transformed_cols <- c(transformed_cols, col)
              }
            }
          }
        }
      }
    }
  }

  # Report changes if any were made
  if (transformed_count > 0) {
    message(sprintf("Replaced %d missing value codes with NA in %d column(s): %s",
                    transformed_count,
                    length(transformed_cols),
                    paste(transformed_cols, collapse=", ")))
  }

  return(df)
}

#' Get missing data codes from configuration
#'
#' @return A list of missing data codes by category
#' @noRd
get_missing_data_codes <- function() {
  # Get configuration
  config <- validate_config("sql")

  # Return missing data codes if configured with proper validation
  if (!is.null(config) && is.list(config) && "missing_data_codes" %in% names(config)) {
    return(config$missing_data_codes)
  }

  return(NULL)
}

#' Test Oracle database connection
#'
#' Tests the connection to the Oracle database using the configured DSN and credentials.
#' This is a simple connectivity test that doesn't perform any data operations.
#'
#' @return A logical value indicating whether the connection was successful
#' @importFrom odbc dbConnect dbDisconnect
#' @export
#' @examples
#' \dontrun{
#' # Test the Oracle connection
#' if (oracle.test()) {
#'   message("Oracle connection successful!")
#' } else {
#'   message("Oracle connection failed!")
#' }
#' }
oracle.test <- function() {
  
  # Check if required packages are available
  if (!requireNamespace("odbc", quietly = TRUE)) {
    message("Package 'odbc' is needed for this function to work. Please install it.")
    return(FALSE)
  }
  
  # Validate secrets
  tryCatch({
    validate_secrets("sql")
  }, error = function(e) {
    message("Error validating secrets: ", e$message)
    return(FALSE)
  })
  
  # Set Oracle environment variables from secrets if they exist
  tns_admin <- get_secret_optional("TNS_ADMIN")
  oracle_home <- get_secret_optional("ORACLE_HOME")
  if (!is.null(tns_admin) && nchar(tns_admin) > 0) {
    Sys.setenv(TNS_ADMIN = tns_admin)
    message(sprintf("Set TNS_ADMIN = %s", tns_admin))
  }
  if (!is.null(oracle_home) && nchar(oracle_home) > 0) {
    Sys.setenv(ORACLE_HOME = oracle_home)
    message(sprintf("Set ORACLE_HOME = %s", oracle_home))
  }
  
  # Get connection parameters (supports both DSN and DBQ for TNS aliases)
  tryCatch({
    conn_params <- get_oracle_connection_params()
  }, error = function(e) {
    message("Error getting connection parameters: ", e$message)
    return(FALSE)
  })
  
  user_id <- get_secret("uid")
  password <- get_secret("pwd")
  driver <- get_oracle_driver(validate = TRUE)  # Validate driver name if specified
  
  # Validate that we have the required connection parameters
  if (is.null(conn_params$value) || is.null(user_id) || is.null(password)) {
    message("Missing connection parameters. Please check your secrets configuration.")
    return(FALSE)
  }
  
  # Test the connection
  channel <- NULL
  connection_successful <- FALSE
  
  tryCatch({
    if (conn_params$use_dbq) {
      message(sprintf("Testing connection to TNS alias (DBQ): %s", conn_params$value))
    } else {
      message(sprintf("Testing connection to DSN: %s", conn_params$value))
    }
    if (!is.null(driver)) {
      message(sprintf("Using driver: %s", driver))
    } else {
      message("No driver specified - attempting connection without explicit driver")
    }
    
    # Check connection target length (IM010 can be caused by name being too long)
    if (!conn_params$use_dbq && nchar(conn_params$value) > 32) {
      message(sprintf("Warning: DSN name is %d characters long. Some ODBC drivers have a 32-character limit.", nchar(conn_params$value)))
    }
    
    # List available DSNs for diagnostics
    tryCatch({
      available_dsns <- odbc::odbcListDataSources()
        if (nrow(available_dsns) > 0) {
          message("Available DSNs:")
          print(available_dsns)
          if (!conn_params$use_dbq && !conn_params$value %in% available_dsns$name) {
            message(sprintf("Warning: DSN '%s' not found in list of available DSNs.", conn_params$value))
            message("This may indicate a configuration issue or 32-bit/64-bit driver mismatch.")
            message("If you're using a TNS alias, make sure to set 'DBQ' in secrets.R instead of 'dsn' or 'host'.")
          }
        }
    }, error = function(e) {
      message("Could not list available DSNs: ", e$message)
    })
    
    # List available drivers for diagnostics
    tryCatch({
      available_drivers <- odbc::odbcListDrivers()
      if (nrow(available_drivers) > 0) {
        message("All available ODBC drivers:")
        print(available_drivers)
        oracle_drivers <- available_drivers[grepl("Oracle", available_drivers$name, ignore.case = TRUE), ]
        if (nrow(oracle_drivers) > 0) {
          message("\nAvailable Oracle drivers:")
          print(oracle_drivers)
          # Don't fail here - let the actual connection attempt determine if driver works
          # The driver validation already happened in get_oracle_driver()
          if (!is.null(driver)) {
            # Check if driver matches (case-insensitive)
            driver_lower <- tolower(trimws(driver))
            available_lower <- tolower(available_drivers$name)
            if (!driver_lower %in% available_lower) {
              message(sprintf("\nWarning: Driver '%s' from secrets.R not found in available drivers list.", driver))
              message("Will attempt connection anyway - the driver name may still work.")
            } else {
              message(sprintf("\nDriver '%s' found in available drivers.", driver))
            }
          }
        } else {
          message("Warning: No Oracle drivers found. You may need to install an Oracle ODBC driver.")
        }
      }
    }, error = function(e) {
      message("Could not list available drivers: ", e$message)
    })
    
    # Attempt to connect with Driver as first parameter if available
    # Use DBQ for TNS aliases, dsn for ODBC DSNs
    conn_args <- list(
      UID = user_id,
      PWD = password
    )
    
    if (!is.null(driver)) {
      conn_args$Driver <- driver
    }
    
    if (conn_params$use_dbq) {
      conn_args$DBQ <- conn_params$value
      if (!is.null(driver)) {
        message(sprintf("\nAttempting connection with Driver='%s', DBQ='%s' (TNS alias)", driver, conn_params$value))
      } else {
        message(sprintf("\nAttempting connection with DBQ='%s' (TNS alias, no explicit driver)", conn_params$value))
      }
    } else {
      conn_args$dsn <- conn_params$value
      if (!is.null(driver)) {
        message(sprintf("\nAttempting connection with Driver='%s', DSN='%s'", driver, conn_params$value))
      } else {
        message(sprintf("\nAttempting connection with DSN='%s' (no explicit driver)", conn_params$value))
      }
    }
    
    channel <- do.call(odbc::dbConnect, c(list(odbc::odbc()), conn_args))
    
    if (is.null(channel)) {
      message("Connection failed: channel is null")
      return(FALSE)
    }
    
    # Get database info to confirm connection is working
    db_info <- tryCatch({
      odbc::dbGetInfo(channel)
    }, error = function(e) {
      message("Warning: Could not retrieve database info: ", e$message)
      list(dbms.name = "Oracle")
    })
    
    message(sprintf("Successfully connected to %s database", db_info$dbms.name))
    connection_successful <- TRUE
    
  }, error = function(e) {
    error_msg <- e$message
    message("Connection failed: ", error_msg)
    
    # Provide specific guidance for IM010 error
    if (grepl("IM010", error_msg, ignore.case = TRUE)) {
      message("\nTroubleshooting IM010 error:")
      message("1. Verify the DSN name is correct and exists in ODBC Data Source Administrator")
      message("2. Check for 32-bit vs 64-bit driver mismatch (R may be 64-bit but DSN uses 32-bit driver)")
      message("3. Ensure Oracle ODBC driver is properly installed")
      message("4. Try recreating the DSN in ODBC Data Source Administrator")
      message("5. If DSN name is longer than 32 characters, try using a shorter name")
      message("6. Consider using a driver name instead of DSN (may require config changes)")
    }
    
    # Provide specific guidance for IM002 error (driver not found)
    if (grepl("IM002", error_msg, ignore.case = TRUE)) {
      message("\nTroubleshooting IM002 error (Driver not found):")
      message("1. The driver name specified does not match any installed ODBC driver")
      if (!is.null(driver)) {
        message(sprintf("2. Driver specified: '%s'", driver))
      }
      message("3. Check the list of available drivers printed above")
      message("4. Update your secrets.R file with the exact driver name from the list")
      message("5. Driver names are case-sensitive and must match exactly")
      message("6. Common Oracle driver names include:")
      message("   - 'Oracle in instantclient_19_3'")
      message("   - 'Oracle in instantclient_21_1'")
      message("   - 'Oracle 19 ODBC driver'")
      message("   - 'Oracle in instantclient_23_1'")
    }
    
    connection_successful <- FALSE
  }, finally = {
    # Always try to close the connection
    if (!is.null(channel)) {
      tryCatch({
        odbc::dbDisconnect(channel)
        if (connection_successful) {
          message("Connection closed successfully")
        }
      }, error = function(e) {
        message("Warning: Error closing connection: ", e$message)
      })
    }
  })
  
  return(connection_successful)
}

#' Run a Suite of Tests on Dataset
#'
#' This function performs a series of validation tests on a dataset to ensure it meets
#' quality standards and contains required variables for NDA submission.
#'
#' @param measure_alias Character string, the name of the measure to be tested
#' @param measure_type Character string, the type of measure (e.g., "redcap", "qualtrics", "mongo")
#' @param script_path Character string, the path to the script that processes the measure
#' @param super_key Character vector, one or more identifier variables to check for in the dataset
#'
#' @return No return value, called for side effects
#'
#' @details
#' The function performs the following checks:
#' \itemize{
#'   \item Checks for duplicate records in Qualtrics data
#'   \item Verifies that a cleaned data frame exists
#'   \item Checks that all NDA required variables are present
#'   \item Ensures column naming is consistent with measure prefix standards
#'   \item Validates that interview age is within acceptable range
#'   \item Verifies that super key variables are present in both raw and cleaned data
#' }
#'
#' @examples
#' \donttest{
#'   testSuite("rgpts", "qualtrics", "./clean/qualtrics/rgpts.R", "src_subject_id")
#' }
#'
#' @importFrom utils head
#' @noRd
testSuite <- function(measure_alias, measure_type, script_path, super_key) {

  # Get identifier from config
  config <- validate_config()
  identifier <- config$identifier

  # List of NDA required variables
  nda_required_variables <- c("src_subject_id", "phenotype", "site", "arm", "visit", "week",
                              "subjectkey", "sex", "interview_date", "interview_age", "state")

  # Perform generic tests
  checkQualtricsDuplicates(measure_alias, measure_type) # and give allow to View them in a table
  cleanDataFrameExists(measure_alias, measure_type) #checkin_clean x
  checkColumnPrefix(measure_alias, measure_type, nda_required_variables, identifier) # checkin_distress

  # perform nda-specific tests
  if (identifier == "src_subject_id") {
    ndaRequiredVariablesExist(measure_alias, measure_type, nda_required_variables) # do Nda req variables exist
    checkInterviewAge(measure_alias) # <240 >860
  }

  if (exists(measure_alias)) {
    message("Raw data found. Looking for super keys...")
    # Check for presence of super_key variables in the raw data
    candidate_keys <- checkKeys(measure_alias, super_key, "raw")
    # Assuming you still want to verify these keys are present in the cleaned data
    if (exists(paste0(measure_alias, "_clean"))) {
      message("Clean data found. Looking for candidate keys... ")
      candidate_keys <- checkKeys(paste0(measure_alias, "_clean"), candidate_keys, "clean")
    }
  }


  # ...add additional functions here, making sure they pass in measure_alias and measure_type
}

# Define the loading animation function
show_loading_animation <- function() {
  cat("Loading ")
  pb <- txtProgressBar(min = 0, max = 20, style = 3)
  
  for (i in 1:20) {
    Sys.sleep(0.1) # Simulate some computation time
    setTxtProgressBar(pb, i)
  }
  
  close(pb)
}

# Define the progress callback function
progress_callback <- function(count, total) {
  setTxtProgressBar(pb, count) # Update the loading animation
}

#' @noRd
initializeLoadingAnimation <- function(max_progress) {
  # cat("Loading...\n")
  pb <- txtProgressBar(min = 0, max = max_progress, style = 3)
  return(pb)
}

#' @noRd
updateLoadingAnimation <- function(pb, current_progress) {
  setTxtProgressBar(pb, current_progress)
}
#' @noRd
completeLoadingAnimation <- function(pb) {
  close(pb)
}


#' Initialize or Update Loading Animation
#'
#' Initializes a new loading animation or updates an existing one based on the progress.
#'
#' @param step Current step (chunk being processed).
#' @param total_steps Total number of chunks.
#' @param pb An existing txtProgressBar object; if NULL, a new progress bar is created.
#' @return A txtProgressBar object representing the current state of the progress bar.
#' @examples
#' \donttest{
#' pb <- show_loading_animation_with_chunks(step = 1, total_steps = 20)  # Start new animation
#' pb <- show_loading_animation_with_chunks(step = 2, total_steps = 20, pb = pb)  # Update animation
#' }
#' @noRd
show_loading_animation_with_chunks <- function(step = NULL, total_steps = NULL, pb = NULL) {
  if (is.null(pb)) {
    # cat("Loading:\n")  # Comment this out if you don't want the static "Loading..." message
    pb <- txtProgressBar(min = 0, max = total_steps, style = 3)
  }
  if (!is.null(step) && !is.null(total_steps)) {
    setTxtProgressBar(pb, step)
  }
  return(pb)
}

#' @importFrom utils capture.output head install.packages setTxtProgressBar txtProgressBar getFromNamespace
NULL

# Create a package environment to store state between functions
.pkg_env <- new.env(parent = emptyenv())

#' NDA Super Required Fields
#'
#' @description
#' The 5 mandatory fields required for ALL NDA submissions.
#' These are always sourced from ndar_subject01 and added to all structures.
#' 
#' Order matches NDA documentation convention:
#' 1. subjectkey - Unique GUID identifier
#' 2. src_subject_id - Study-specific subject ID
#' 3. interview_date - Date of data collection
#' 4. interview_age - Age at interview (months)
#' 5. sex - Biological sex (M/F/O)
#'
#' @keywords internal
#' @noRd
SUPER_REQUIRED_FIELDS <- c(
  "subjectkey",
  "src_subject_id",
  "interview_date",
  "interview_age",
  "sex"
)

#' DCC Required Fields
#'
#' @description
#' The 7 required fields from ndar_subject01 that are mandatory for
#' Data Coordinating Center (DCC) submissions. These fields are only
#' included when dcc = TRUE.
#'
#' @keywords internal
#' @noRd
DCC_REQUIRED_FIELDS <- c(
  "race",
  "phenotype",
  "phenotype_description",
  "twins_study",
  "sibling_study",
  "family_study",
  "sample_taken"
)

#' DCC Recommended Fields
#'
#' @description
#' The 4 recommended fields from ndar_subject01 for Data Coordinating
#' Center (DCC) submissions. These fields are only included when dcc = TRUE.
#'
#' @keywords internal
#' @noRd
DCC_RECOMMENDED_FIELDS <- c(
  "ethnic_group",
  "site",
  "study",
  "subsiteid"
)

#' All DCC Fields
#'
#' @description
#' Combined list of all DCC required and recommended fields (11 total).
#'
#' @keywords internal
#' @noRd
DCC_FIELDS <- c(DCC_REQUIRED_FIELDS, DCC_RECOMMENDED_FIELDS)

#' Fields to automatically drop during unexpected field processing
#'
#' @description
#' Combined list of DCC fields (from ndar_subject01) and internal wizaRdry
#' tracking fields that should never be included in NDA data structures.
#' These are silently dropped during interactive field mapping without
#' prompting the user.
#'
#' @keywords internal
#' @noRd
NDAR_SKIP_FIELDS <- c(
  DCC_FIELDS,                      # race, phenotype, subsiteid, etc.
  "state",                         # Internal wizaRdry tracking field
  "lost_to_followup",              # Internal wizaRdry tracking field
  "lost_to_follow-up",             # Hyphenated variant
  "study_status"                   # Internal wizaRdry tracking field
)

.onLoad <- function(libname, pkgname) {
  options(mongolite.quiet = TRUE)
  options(wizaRdry.nda_base_url = "https://nda.nih.gov/api/datadictionary/v2")

  # Try to load secrets silently without any messages
  if (file.exists("secrets.R")) {
    tryCatch({
      load_secrets("secrets.R")
    }, error = function(e) {
      # Store the error for potential use in .onAttach
      assign("secrets_load_error", TRUE, envir = .pkg_env)
      assign("secrets_error_message", conditionMessage(e), envir = .pkg_env)
    })
  }
}

# run scry() on launch if no project structure created
.onAttach <- function(libname, pkgname) {
  # Check if we need to report a secrets loading error
  if (exists("secrets_load_error", envir = .pkg_env)) {
    error_msg <- if (exists("secrets_error_message", envir = .pkg_env)) {
      base::get("secrets_error_message", envir = .pkg_env)
    } else {
      "unknown error"
    }
    packageStartupMessage("Note: Could not load secrets.R at package load time: ", error_msg)
  }

  if (interactive()) {
    # Check if the current directory has a wizaRdry structure
    current_dir <- getwd()
    has_config <- file.exists(file.path(current_dir, "config.yml"))
    has_clean <- dir.exists(file.path(current_dir, "clean"))
    has_nda <- dir.exists(file.path(current_dir, "nda"))
    structure_exists <- has_config || has_clean || has_nda

    if (!structure_exists) {
      packageStartupMessage("Welcome to wizaRdry! Use scry() to initialize a new project structure.")
      #packageStartupMessage("  Tip: Use scry(examples = FALSE) to skip creating example scripts if you're already familiar with wizaRdry.")
    } else {
      # Check if structure is complete
      dirs_to_check <- c(
        file.path(current_dir, "clean"),
        file.path(current_dir, "clean", "mongo"),
        file.path(current_dir, "clean", "qualtrics"),
        file.path(current_dir, "clean", "redcap"),
        file.path(current_dir, "nda"),
        file.path(current_dir, "nda", "mongo"),
        file.path(current_dir, "nda", "qualtrics"),
        file.path(current_dir, "nda", "redcap"),
        file.path(current_dir, "tmp")
      )
      files_to_check <- c(
        file.path(current_dir, "config.yml"),
        file.path(current_dir, "secrets.R"),
        file.path(current_dir, "main.R")
      )
      structure_complete <- all(sapply(dirs_to_check, dir.exists)) &&
        all(sapply(files_to_check, file.exists))

      if (!structure_complete) {
        packageStartupMessage("wizaRdry structure detected but incomplete. Use scry(repair = TRUE) to repair it.")
      } else {
        ver <- tryCatch(as.character(utils::packageVersion(pkgname)), error = function(e) "unknown")
        packageStartupMessage(paste0("wizaRdry v", ver, " project structure detected and complete."))

        # Check for duplicate script names - use get to access the internal function
        tryCatch({
          check_duplicate_scripts <- getFromNamespace("check_duplicate_scripts", "wizaRdry")
          dups <- check_duplicate_scripts(current_dir, quiet = TRUE)
          if (!is.null(dups)) {
            packageStartupMessage("\nWARNING: Duplicate script names found across folders.")
            packageStartupMessage("This may cause conflicts when processing data.")
          }
        }, error = function(e) {
          # Silently ignore any errors in startup checks
        })
      }
    }
  }
}

#' Merge two or more data frames magically according to their candidate key
#'
#' This function simplifies the process of merging multiple cleaned data frames by automatically determining 
#' common merge keys or utilizing user-specified keys. Supports both inner and outer join methods, 
#' and offers options for exporting the merged data.
#'
#' @param ... Clean data frames to be merged.
#' @param by A vector of strings specifying the column names to be used as merge keys. If NULL, 
#'           the function automatically determines common keys from the provided data frames.
#' @param all Logical; if TRUE, performs an OUTER JOIN. If FALSE, performs an INNER JOIN.
#' @param no.dups Logical; if TRUE, duplicates are removed post-merge.
#' @param csv Logical; if TRUE, the merged data frame is exported as a CSV file.
#' @param rdata Logical; if TRUE, the merged data frame is saved as an Rda file.
#' @param spss Logical; if TRUE, the merged data frame is exported as an SPSS file.
#'
#' @examples
#' \dontrun{
#' # Create sample dataframes for demonstration
#' df1 <- data.frame(
#'   src_subject_id = c("S001", "S002", "S003"),
#'   visit = c(1, 2, 1),
#'   measure1 = c(10, 15, 12),
#'   stringsAsFactors = FALSE
#' )
#' 
#' df2 <- data.frame(
#'   src_subject_id = c("S001", "S002", "S004"),
#'   visit = c(1, 2, 2),
#'   measure2 = c(85, 92, 78),
#'   stringsAsFactors = FALSE
#' )
#' 
#' # Perform an OUTER JOIN using default keys:
#' merged1 <- meld(df1, df2, all = TRUE)
#' 
#' # Perform an INNER JOIN using specified keys:
#' merged2 <- meld(df1, df2, by = "src_subject_id", all = FALSE)
#' }
#' @return A merged data frame based on the specified or common candidate keys.
#' @author Joshua Kenney <joshua.kenney@yale.edu>
#' @export
meld <- function(..., by = NULL, all = TRUE, no.dups = FALSE, csv = FALSE, rdata = FALSE, spss = FALSE) {
  
  # Inform about the type of join being performed
  message(ifelse(all, "Performing an OUTER JOIN.", "Performing an INNER JOIN."))
  
  config <- validate_config()
  
  if (config$study_alias == "capr") {
    # NDA variables suitable for merging fromr capr
    super_key <- c("src_subject_id", "subjectkey", "phenotype", "visit", "week", "sex", "site", "arm")
  } else {
    super_key <- c("src_subject_id", "subjectkey", "phenotype", "visit", "week", "sex", "site", "arm", "state", "PROLIFIC_PID", "participantId", "workerId", "rat_id")
  }
  
  # Load custom scripts if any
  
  data_list <- list(...)
  
  #if one inputs a list of dataframes then we need to unpack the list using the
  #following code. For example you can now do this:
  #create a list of dfs to merge
  #  dfs_to_merge = NULL
  #  for (name in variables_to_merge){
  #      dfs_to_merge[[name]]<-base::get(name)
  #      }
  #then input that list into dataMerge
  #  merged_df<-dataMerge(dfs_to_merge)
  
  if (length(data_list) == 1){
    data_list = data_list[[1]]
  }
  
  # Preprocess data frames: Remove specified columns and ungroup
 # data_list <- lapply(data_list, function(df) {
#    # Remove 'interview_date' and 'interview_age' columns
#    df <- df[setdiff(names(df), c("interview_date", "interview_age"))]
#    return(df)
#  })
  
  # Determine the keys to use for merging
  if (is.null(by)) {
    by <- Reduce(intersect, lapply(data_list, function(df) intersect(super_key, names(df))))
    message("Detected common candidate keys for merge: ", toString(by))
  } else {
    by <- by
    message("Using user-specified keys for merge: ", toString(by))
  }
  # Perform the merging process
  dfs <- Reduce(function(x, y) base::merge(x, y, by = by,
                                            all = all, no.dups = no.dups), data_list)
  
  # Export merged data if requested
  if (csv) { createCsv(dfs, "merged_dfs.csv") }
  if (rdata) { createRds(dfs, "merged_dfs") }
  if (spss) { createSpss(dfs, "merged_dfs.sav") }
  
  
  return(dfs)
}


#' Alias for 'meld' (DEPRECATED)
#'
#' This function is deprecated. Please use 'meld' instead.
#' This is a legacy alias for the 'meld' function to maintain compatibility with older code.
#'
#' @inheritParams meld
#' @inherit meld return
#' @export
#' @examples
#' \dontrun{
#' # DEPRECATED - use meld() instead
#' merged <- dataMerge(df1_clean, df2_clean)
#' }
dataMerge <- function(...) {
  .Deprecated("meld", package = "wizaRdry")
  meld(...)
}

#' NDA Data Structure Validator
#'
#' @description
#' Main validation function for NDA data structure compliance.
#' This is the refactored version using modular components.
#' 
#' This is an internal function called by nda(). Users should not call this directly.
#'
#' @param measure_name Name of measure/structure
#' @param api API type (redcap, qualtrics, mongo, csv, oracle, sql)
#' @param limited_dataset Logical - if TRUE, skip de-identification (date-shifting, age-capping)
#' @param nda_base_url NDA API base URL
#' @param verbose Logical - print detailed output
#' @param debug Logical - print debug information
#' @param auto_drop_unknown Logical - automatically drop unknown fields
#' @param interactive_mode Logical - allow user prompts
#' @param modified_structure Pre-enhanced NDA structure (from ndaRequest.R)
#' @param strict Logical - if TRUE (default), enforce strict validation for missing data
#' @param dcc Logical - if TRUE, validate DCC fields from ndar_subject01
#' @return ValidationState object with validation results
#' @keywords internal
#' @noRd
ndaValidator <- function(measure_name,
                         api,
                         limited_dataset = FALSE,
                             nda_base_url = "https://nda.nih.gov/api/datadictionary/v2",
                             verbose = FALSE,
                             debug = FALSE,
                             auto_drop_unknown = FALSE,
                             interactive_mode = TRUE,
                             modified_structure = NULL,
                             strict = TRUE,
                             dcc = FALSE,
                             ndar_additions = character()) {
  
  tryCatch({
    # Initialize environment in package namespace (CRAN compliant)
    # NOTE: Uses .pkg_env for cross-function data sharing instead of globalenv
    if (!exists(".wizaRdry_env", envir = .pkg_env, inherits = FALSE)) {
      assign(".wizaRdry_env", new.env(parent = emptyenv()), envir = .pkg_env)
    }
    
    # Get wizaRdry environment from package environment
    wizaRdry_env <- .pkg_env$.wizaRdry_env
    
    # Get dataframe from package environment
    df <- base::get(measure_name, envir = wizaRdry_env)
    if (is.null(df) || !is.data.frame(df)) {
      stop(sprintf("Dataframe '%s' not found or is not a data.frame", measure_name))
    }
    
    if (verbose) {
      message(sprintf("\n\n=== Starting NDA Validation for '%s' ===", measure_name))
      message(sprintf("API: %s | Rows: %d | Columns: %d", api, nrow(df), ncol(df)))
    }
    
    # Get or fetch NDA structure
    if (!is.null(modified_structure)) {
      if (verbose) message("\nUsing enhanced NDA structure with ndar_subject01 definitions...")
      elements <- modified_structure$dataElements
      nda_structure <- modified_structure
    } else {
      if (verbose) message(sprintf("\nFetching '%s' from NDA API...", measure_name))
      nda_structure <- fetch_nda_structure(measure_name, nda_base_url)
      elements <- nda_structure$dataElements
    }
    
    if (is.null(elements) || nrow(elements) == 0) {
      stop("No dataElements found in structure definition")
    }
    
    # Create ValidationState object
    state <- ValidationState$new(measure_name, api, df, nda_structure, dcc = dcc)
    
    # Store ndar_subject01 additions in ValidationState
    if (length(ndar_additions) > 0) {
      state$ndar_subject_additions <- ndar_additions
      state$is_modified_structure <- TRUE  # DCC additions modify the structure
    }
    
    if (verbose) {
      message(sprintf("Structure type: %s", 
                     if(state$is_new_structure) "NEW (not in NDA)" else "EXISTING"))
      message(sprintf("Structure has %d field definitions", nrow(elements)))
    }
    
    # ============================================================================
    # PHASE 1: Data Cleaning
    # ============================================================================
    if (verbose) message("\n--- PHASE 1: Data Cleaning ---")
    
    # Convert problematic column types
    df <- convert_problematic_column_types(df, measure_name, verbose)
    state$set_df(df)
    
    # Convert logical columns to character
    df <- convert_logical_to_character(df, verbose)
    state$set_df(df)
    
    # ============================================================================
    # PHASE 2: Field Standardization
    # ============================================================================
    if (verbose) message("\n--- PHASE 2: Field Standardization ---")
    
    # Standardize column names
    df <- standardize_column_names(df, measure_name, verbose)
    df <- standardize_field_names(df, measure_name, verbose)
    state$set_df(df)
    
    # Handle missing required fields
    required_fields <- elements$name[elements$required == "Required"]
    missing_required <- required_fields[!required_fields %in% names(df)]
    if (length(missing_required) > 0) {
      df <- handle_missing_fields(df, elements, missing_required, verbose)
      state$set_df(df)
    }
    
    # ============================================================================
    # PHASE 2.5: Interactive Field Mapping (MODIFIED Structures Only)
    # ============================================================================
    # Handle unexpected fields (fields in data but not in NDA structure)
    # This only applies to EXISTING structures, not NEW structures
    if (!state$is_new_structure) {
      unexpected_count <- length(setdiff(names(df), elements$name))
      
      if (unexpected_count > 0) {
        if (verbose) {
          message(sprintf("\n--- PHASE 2.5: Interactive Field Mapping ---"))
          message(sprintf("Found %d unexpected field(s) not in NDA structure", unexpected_count))
        }
        
        # Process unexpected fields interactively
        mapping_result <- process_unexpected_fields(
          df = df,
          elements = elements,
          structure_name = measure_name,
          measure_name = measure_name,
          api = api,
          verbose = verbose,
          auto_drop_unknown = auto_drop_unknown,
          interactive_mode = interactive_mode
        )
        
        # Update dataframe
        df <- mapping_result$df
        state$set_df(df)
        
        # Drop fields marked for removal
        if (length(mapping_result$columns_to_drop) > 0) {
          df <- df[, !names(df) %in% mapping_result$columns_to_drop, drop = FALSE]
          state$set_df(df)
          if (verbose) {
            message(sprintf("Dropped %d field(s) from dataframe", 
                           length(mapping_result$columns_to_drop)))
          }
        }
        
        # Mark as modified structure if any changes were made
        if (length(mapping_result$renamed_fields) > 0 || length(mapping_result$columns_to_drop) > 0) {
          state$is_modified_structure <- TRUE
        }
        
        # Update cleaning script with auto-generated field operations
        if (length(mapping_result$renames) > 0 || length(mapping_result$columns_to_drop) > 0) {
          # Construct script path based on API type
          script_dir <- file.path(".", "nda", tolower(api))
          script_path <- file.path(script_dir, paste0(measure_name, ".R"))
          
          if (file.exists(script_path)) {
            update_cleaning_script(
              script_path = script_path,
              measure_name = measure_name,
              renames = mapping_result$renames,
              drops = mapping_result$columns_to_drop,
              verbose = verbose
            )
          } else if (verbose) {
            message(sprintf("[SCRIPT UPDATE] Cleaning script not found: %s", script_path))
            message("[SCRIPT UPDATE] Field operations will not persist to next run")
          }
        }
      }
    }
    
    # ============================================================================
    # PHASE 3: Required Field Data Validation
    # ============================================================================
    
    # Add STEP 2 header (only in non-verbose mode)
    if (!verbose) {
      if (state$is_new_structure) {
        message("\n=== STEP 2: Preparing New Structure ===")
      } else {
        message("\n=== STEP 2: Validating Data Structure ===")
        message(sprintf("Validating against NDA structure '%s'...", measure_name))
      }
    }
    
    if (verbose) message("\n--- PHASE 3: Required Field Data Validation ---")
    
    # Check field data completeness (only validates 5 super required fields)
    data_violations <- check_field_data_completeness(
      state, 
      elements, 
      super_required_fields = SUPER_REQUIRED_FIELDS,
      strict = strict, 
      verbose = verbose
    )
    required_violations <- data_violations$required
    recommended_violations <- data_violations$recommended
    
    # Handle violations based on strict mode
    has_violations <- length(required_violations) > 0 || length(recommended_violations) > 0
    
    if (has_violations) {
      # BOTH strict and lenient: Mark as invalid and store violations
      state$is_valid <- FALSE
      
      # Store violations in state (same for both modes)
      for (field_name in names(required_violations)) {
        state$errors <- c(state$errors, 
                         sprintf("Required field '%s': %s", 
                                field_name, 
                                required_violations[[field_name]]$issue))
      }
      
      for (field_name in names(recommended_violations)) {
        state$errors <- c(state$errors,
                         sprintf("Recommended field '%s': %s",
                                field_name,
                                recommended_violations[[field_name]]$issue))
      }
      
      if (strict) {
        # STRICT MODE: Stop processing, skip remaining phases
        
        # Non-verbose output
        if (!verbose) {
          message("")  # Blank line
          message("[ERROR] Field validation failed:\n")
          
          if (length(required_violations) > 0) {
            message("  Required Fields:")
            for (field_name in names(required_violations)) {
              violation <- required_violations[[field_name]]
              message(sprintf("    - %s: %s", field_name, violation$issue))
            }
            message("")
          }
          
          if (length(recommended_violations) > 0) {
            message("  Recommended Fields (strict mode):")
            for (field_name in names(recommended_violations)) {
              violation <- recommended_violations[[field_name]]
              message(sprintf("    - %s: %s", field_name, violation$issue))
            }
            message("")
          }
          
          total_violations <- length(required_violations) + length(recommended_violations)
          message(sprintf("[ERROR] Validation failed with %d field violation%s",
                         total_violations,
                         if (total_violations > 1) "s" else ""))
          message("")  # Blank line
          
          # Skip remaining phases
          message("\n=== STEP 3: De-identifying Data ===")
          message("[SKIPPED - Validation failed]")
          message("")
          message("\n=== STEP 4: Generating NDA Files ===")
          message("[SKIPPED - Validation failed]")
          message("")
        }
        
        # Return early with failed state
        return(state)
        
      } else {
        # LENIENT MODE: Show warnings but continue processing
        state$warnings <- c(state$warnings, 
                           sprintf("Found %d field(s) with missing data", 
                                  length(required_violations) + length(recommended_violations)))
        
        if (!verbose) {
          message("")  # Blank line
          message("[WARN] Field validation failed (lenient mode):\n")
          
          if (length(required_violations) > 0) {
            message("  Required Fields:")
            for (field_name in names(required_violations)) {
              violation <- required_violations[[field_name]]
              message(sprintf("    - %s: %s", field_name, violation$issue))
            }
            message("")
          }
          
          if (length(recommended_violations) > 0) {
            message("  Recommended Fields:")
            for (field_name in names(recommended_violations)) {
              violation <- recommended_violations[[field_name]]
              message(sprintf("    - %s: %s", field_name, violation$issue))
            }
            message("")
          }
          
          total_violations <- length(required_violations) + length(recommended_violations)
          message(sprintf("[WARN] Validation failed with %d field violation%s (continuing anyway)",
                         total_violations,
                         if (total_violations > 1) "s" else ""))
          message("[WARN] Files will be created with caveats")
          message("")  # Blank line
        }
      }
    }
    
    # ============================================================================
    # PHASE 3.5: DCC Fields Validation (if dcc = TRUE)
    # ============================================================================
    if (state$dcc) {
      if (verbose || strict) {
        message("\n--- Checking DCC Fields ---")
      }
      
      dcc_result <- check_dcc_fields(
        state = state,
        elements = elements,
        strict = strict,
        verbose = verbose
      )
      
      dcc_required_violations <- dcc_result$required_violations
      dcc_recommended_violations <- dcc_result$recommended_violations
      
      # Add to state
      state$add_violations("dcc_required", dcc_required_violations)
      state$add_violations("dcc_recommended", dcc_recommended_violations)
      
      # Update validation status if violations found
      if (length(dcc_required_violations) > 0 || (strict && length(dcc_recommended_violations) > 0)) {
        state$set_valid(FALSE)
        
        if (strict) {
          # STRICT MODE: Show errors and return early
          if (!verbose) {
            message("")  # Blank line
            message("[ERROR] DCC field validation failed:\n")
            
            if (length(dcc_required_violations) > 0) {
              message("  DCC Required Fields:")
              for (field_name in names(dcc_required_violations)) {
                violation <- dcc_required_violations[[field_name]]
                message(sprintf("    - %s: %s", field_name, violation$issue))
              }
              message("")
            }
            
            if (length(dcc_recommended_violations) > 0) {
              message("  DCC Recommended Fields:")
              for (field_name in names(dcc_recommended_violations)) {
                violation <- dcc_recommended_violations[[field_name]]
                message(sprintf("    - %s: %s", field_name, violation$issue))
              }
              message("")
            }
            
            total_violations <- length(dcc_required_violations) + length(dcc_recommended_violations)
            message(sprintf("[ERROR] Validation failed with %d DCC field violation%s",
                           total_violations,
                           if (total_violations > 1) "s" else ""))
            message("")
          }
          
          # Skip remaining phases
          message("\n=== STEP 3: De-identifying Data ===")
          message("[SKIPPED - DCC validation failed]")
          message("")
          message("\n=== STEP 4: Generating NDA Files ===")
          message("[SKIPPED - DCC validation failed]")
          message("")
          
          # Return early with failed state
          return(state)
          
        } else {
          # LENIENT MODE: Show warnings but continue
          if (!verbose) {
            total_violations <- length(dcc_required_violations) + length(dcc_recommended_violations)
            message(sprintf("[WARN] %d DCC field violation%s found (continuing anyway)",
                           total_violations,
                           if (total_violations > 1) "s" else ""))
            message("")
          }
        }
      }
    }
    
    # ============================================================================
    # PHASE 4: Value Range Validation
    # ============================================================================
    tryCatch({
      if (verbose) message("\n--- PHASE 4: Value Range Validation ---")
      
      # This is the critical function that properly tracks violations
      violations <- check_value_range_violations(state, elements, verbose)
      
      # Show value range summary (violations only, unless verbose)
      if (!verbose) {
        total_fields <- length(intersect(names(state$get_df()), elements$name))
        
        if (length(violations) == 0) {
          message(sprintf("Checking %d fields for NDA compliance...", total_fields))
          message("")  # Blank line before [OK]
          message(sprintf("[OK] All %d fields validated successfully (0 violations)", total_fields))
          message("")  # Blank line after [OK]
        } else {
          message(sprintf("Checking %d fields for NDA compliance...", total_fields))
          message("[WARN] Value range violations found:\n")
          
          for (field_name in names(violations)) {
            violation <- violations[[field_name]]
            expected_str <- if (is.null(violation$expected_range)) {
              "no range defined"
            } else {
              violation$expected_range
            }
            
            violating_vals <- violation$violating_values
            message(sprintf("  Field: %s", field_name))
            message(sprintf("  Expected range: %s", expected_str))
            message(sprintf("  Violating values found: %s (%d occurrence%s)", 
                           paste(violating_vals, collapse = ", "),
                           length(violating_vals),
                           if (length(violating_vals) > 1) "s" else ""))
            message("")
          }
          
          message("")  # Blank line before final summary
          message(sprintf("[WARN] Validation completed with %d violation%s",
                         length(violations),
                         if (length(violations) > 1) "s" else ""))
          message("")  # Blank line after summary
        }
      }
      
      if (length(violations) > 0 && verbose) {
        message(sprintf("\nDetected %d field(s) with value range issues", length(violations)))
        message("These will trigger data definition file creation")
      }
    }, error = function(e) {
      stop(sprintf("Error in PHASE 4 (Value Range Validation): %s", e$message))
    })
    
    # ============================================================================
    # PHASE 5: New Field Detection
    # ============================================================================
    tryCatch({
      if (verbose) message("\n--- PHASE 5: New Field Detection ---")
      
      state$new_fields <- detect_new_fields(state$get_df(), elements, dcc = state$dcc)
      
      if (length(state$new_fields) > 0) {
        state$is_modified_structure <- TRUE
        if (verbose) {
          message(sprintf("\nDetected %d new field(s) not in NDA structure:", 
                         length(state$new_fields)))
          message(sprintf("  %s", paste(state$new_fields, collapse = ", ")))
        }
      }
    }, error = function(e) {
      stop(sprintf("Error in PHASE 5 (New Field Detection): %s", e$message))
    })
    
    # ============================================================================
    # PHASE 6: De-identification
    # ============================================================================
    if (!verbose) {
      message("\n=== STEP 3: De-identifying Data ===")
    } else {
      message("\n--- PHASE 6: De-identification ---")
    }
    
    if (limited_dataset) {
      # Limited dataset mode - skip de-identification
      if (!verbose) {
        message("[OK] Using limited dataset (de-identification already applied per your configuration)")
        message("")  # Blank line
      } else {
        message("[OK] Limited dataset mode - de-identification already applied")
      }
    } else {
      # Normal mode - perform de-identification
      
      # Single intro message
      if (!verbose) {
        message("Applying HIPAA Safe Harbor de-identification to limited dataset")
        message("")
      }
      
      # Process date-shifting silently
      if ("interview_date" %in% names(df)) {
        df <- standardize_dates(df, verbose = verbose, limited_dataset = limited_dataset)
      }
      
      # Process age-capping silently
      if ("interview_age" %in% names(df)) {
        age_result <- standardize_age(df, verbose = verbose, limited_dataset = limited_dataset)
        
        # Defensive check: ensure age_result is a list with df and stats
        if (is.list(age_result) && "df" %in% names(age_result) && "stats" %in% names(age_result)) {
          df <- age_result$df
          age_stats <- age_result$stats
        } else if (is.data.frame(age_result)) {
          # Fallback: age_result might be just a dataframe (old format)
          df <- age_result
          age_stats <- NULL
        } else {
          age_stats <- NULL
        }
      }
      
      # Show completion messages together
      if (!verbose) {
        message("[OK] Date-shifting complete")
        message("[OK] Age-capping complete")
        message("")  # Blank line after completion
      }
      
      if (verbose) {
        message("\nDataset has been de-identified using date-shifting and age-capping.")
      }
    }
    
    state$set_df(df)
    
    # ============================================================================
    # PHASE 7: Final Validation Summary
    # ============================================================================
    if (verbose) {
      message("\n--- PHASE 6: Validation Summary ---")
      print_validation_summary(state)
    }
    
    # Return ValidationState object (not a plain list)
    return(state)
    
  }, error = function(e) {
    error_msg <- sprintf("Error in ndaValidator for '%s': %s", measure_name, e$message)
    message(error_msg)
    
    if (debug) {
      message("\nTraceback:")
      message(paste(capture.output(traceback()), collapse="\n"))
    }
    
    # Try to get the dataframe for error state from package environment
    error_df <- if (exists(".wizaRdry_env", envir = .pkg_env, inherits = FALSE)) {
      wizaRdry_env <- .pkg_env$.wizaRdry_env
      tryCatch(base::get(measure_name, envir = wizaRdry_env), error = function(e2) data.frame())
    } else {
      data.frame()
    }
    
    # Use the modified_structure if provided (existing structures), otherwise NULL (new structures)
    error_nda_structure <- if (!is.null(modified_structure)) modified_structure else NULL
    
    # Return error state for graceful failure
    error_state <- ValidationState$new(measure_name, api, 
                                       error_df, error_nda_structure, dcc = dcc)
    error_state$is_valid <- FALSE
    error_state$errors <- c(error_msg)
    return(error_state)
  })
}

#' Cross-platform memory check function
#' @return List containing total and available memory in GB
#' @noRd
getAvailableMemory <- function() {
  tryCatch({
    if (.Platform$OS.type == "windows") {
      # Windows-specific memory detection with better error handling
      total_mem <- tryCatch({
        mem_info <- system('wmic ComputerSystem get TotalPhysicalMemory /Value', intern = TRUE)
        mem_line <- grep("TotalPhysicalMemory=", mem_info, value = TRUE)
        if (length(mem_line) == 0) return(NULL)
        total <- as.numeric(sub("TotalPhysicalMemory=", "", mem_line))
        if (is.na(total)) return(NULL)
        total / (1024^3)  # Convert to GB
      }, error = function(e) NULL)

      avail_mem <- tryCatch({
        # Get multiple memory metrics for better available memory calculation
        mem_info <- system('wmic OS get FreePhysicalMemory,TotalVisibleMemorySize /Value', intern = TRUE)
        # Extract free physical memory
        free_line <- grep("FreePhysicalMemory=", mem_info, value = TRUE)
        if (length(free_line) == 0) return(NULL)
        free_mem <- as.numeric(sub("FreePhysicalMemory=", "", free_line))
        if (is.na(free_mem)) return(NULL)
        # Get total visible memory for percentage calculation
        total_line <- grep("TotalVisibleMemorySize=", mem_info, value = TRUE)
        if (length(total_line) == 0) return(NULL)
        total_visible <- as.numeric(sub("TotalVisibleMemorySize=", "", total_line))
        if (is.na(total_visible)) return(NULL)
        # Convert KB to GB and add 20% buffer for cached memory
        available <- (free_mem / (1024^2)) * 1.2
        # Sanity check - don't return more than 90% of total memory
        max_available <- (total_visible / (1024^2)) * 0.9
        min(available, max_available)
      }, error = function(e) NULL)

      # Return both metrics, with NULL handling
      return(list(
        total = if (is.null(total_mem)) NULL else round(total_mem, 1),
        available = if (is.null(avail_mem)) NULL else round(avail_mem, 1)
      ))

    } else if (Sys.info()["sysname"] == "Darwin") {
      # MacOS
      total_mem <- tryCatch({
        mem_info <- system("sysctl hw.memsize", intern = TRUE)
        as.numeric(strsplit(mem_info, " ")[[1]][2]) / (1024^3)
      }, error = function(e) NULL)

      # More accurate available memory detection for Mac
      avail_mem <- tryCatch({
        vm_stat <- system("vm_stat", intern = TRUE)
        page_size <- 4096  # Default page size for Mac
        # Extract different memory stats
        get_pages <- function(pattern) {
          line <- grep(pattern, vm_stat, value = TRUE)
          as.numeric(sub(".*: *(\\d+).*", "\\1", line))
        }
        free_pages <- get_pages("Pages free:")
        inactive_pages <- get_pages("Pages inactive:")
        purgeable_pages <- get_pages("Pages purgeable:")
        cached_pages <- get_pages("File-backed pages:")
        # Calculate available memory including cache and purgeable
        total_available_pages <- free_pages + inactive_pages + purgeable_pages + cached_pages
        (total_available_pages * page_size) / (1024^3)  # Convert to GB
      }, error = function(e) NULL)

      return(list(
        total = total_mem,
        available = avail_mem
      ))

    } else {
      # Linux
      if (file.exists("/proc/meminfo")) {
        mem_info <- readLines("/proc/meminfo")
        # Helper function to extract memory values
        get_mem_value <- function(pattern) {
          line <- grep(pattern, mem_info, value = TRUE)
          value <- as.numeric(strsplit(line, "\\s+")[[1]][2])  # Get the number
          value / (1024^2)  # Convert KB to GB
        }
        # Get all relevant memory metrics
        total_mem <- get_mem_value("MemTotal:")
        free_mem <- get_mem_value("MemFree:")
        available_mem <- get_mem_value("MemAvailable:")  # Modern Linux kernels provide this
        cached_mem <- get_mem_value("Cached:")
        buffers_mem <- get_mem_value("Buffers:")
        slab_mem <- get_mem_value("SReclaimable:")  # Reclaimable kernel memory

        # Calculate true available memory
        # MemAvailable is already calculated by kernel using a sophisticated algorithm
        # But we can fall back to our own calculation if needed
        if (!is.na(available_mem)) {
          avail_mem <- available_mem
        } else {
          # Similar to how the kernel calculates it:
          # free + ((cached + buffers + slab) * 0.8)
          avail_mem <- free_mem + ((cached_mem + buffers_mem + slab_mem) * 0.8)
        }

        return(list(
          total = total_mem,
          available = avail_mem
        ))
      }
    }
  }, error = function(e) {
    return(list(total = NULL, available = NULL))
  })

  return(list(total = NULL, available = NULL))
}

#' Calculate optimal resource parameters
#' @param total_records Total number of records to process
#' @param mem_info Memory information structure
#' @param num_cores Number of CPU cores
#' @return List containing optimal chunk size and number of workers
#' @noRd
calculateResourceParams <- function(total_records, mem_info, num_cores) {
  # Default values
  params <- list(
    chunk_size = 1000,
    workers = min(4, num_cores)  # OPTIMIZATION: Fewer workers for consistency
  )

  # Adjust chunk size based on available memory and total records
  if (!is.null(mem_info$available)) {
    if (mem_info$available < 4) {
      params$chunk_size <- 500
    } else if (mem_info$available < 8) {
      params$chunk_size <- 1000
    } else if (mem_info$available < 16) {
      params$chunk_size <- 2000
    } else {
      params$chunk_size <- 5000
    }
  }

  # OPTIMIZATION: Larger chunks for bigger datasets
  if (total_records > 50000) {
    params$chunk_size <- max(params$chunk_size, 10000)
  }

  # Adjust for very small datasets
  if (total_records < params$chunk_size * 2) {
    params$chunk_size <- max(500, floor(total_records / 2))
  }

  # Calculate resulting chunks
  params$num_chunks <- ceiling(total_records / params$chunk_size)

  return(params)
}

#' Initialize a clean loading animation
#' @param steps Number of steps in the process
#' @return Loading animation object
#' @noRd
initializeLoadingAnimation <- function(steps) {
  list(
    steps = steps,
    current = 0,
    width = 50,
    start_time = Sys.time()
  )
}

#' Update the loading animation
#' @param pb Loading animation object
#' @param current Current step
#' @noRd
updateLoadingAnimation <- function(pb, current) {
  pb$current <- current
  percentage <- round(current / pb$steps * 100)
  filled <- round(pb$width * current / pb$steps)
  bar <- paste0(
    strrep("=", filled),
    strrep(" ", pb$width - filled)
  )
  cat(sprintf("\r  |%s| %3d%%", bar, percentage))
  utils::flush.console()
}

#' Complete the loading animation
#' @param pb Loading animation object
#' @noRd
completeLoadingAnimation <- function(pb) {
  updateLoadingAnimation(pb, pb$steps)
  cat("\n")
}

#' Format a time duration in a human-readable way
#'
#' @name formatDuration
#' @param duration The duration to format in seconds or minutes
#' @return A formatted string representing the duration
#' @noRd
formatDuration <- function(duration) {
  secs <- as.numeric(duration, units = "secs")
  if (secs < 60) {
    return(sprintf("%.1f seconds", secs))
  } else {
    mins <- floor(secs / 60)
    remaining_secs <- round(secs %% 60, 1)
    if (remaining_secs > 0) {
      return(sprintf("%d minutes and %.1f seconds", mins, remaining_secs))
    } else {
      return(sprintf("%d minutes", mins))
    }
  }
}

#' Create consistent chunks using unique IDs instead of skip/limit
#' @param Mongo MongoDB connection object
#' @param identifier Field to use as identifier
#' @param chunk_size Number of records per chunk
#' @return List containing chunks with specific IDs and total record count
#' @noRd
createConsistentChunks <- function(Mongo, identifier, chunk_size) {
  message("Getting unique identifiers for consistent chunking...")

  # Get ALL unique identifiers first (eliminates skip/limit issues)
  all_ids <- Mongo$distinct(identifier,
                            query = sprintf('{"%s": {"$exists": true, "$ne": ""}}', identifier))

  total_records <- length(all_ids)
  num_chunks <- ceiling(total_records / chunk_size)

  chunks <- vector("list", num_chunks)
  for (i in seq_len(num_chunks)) {
    start_idx <- (i - 1) * chunk_size + 1
    end_idx <- min(i * chunk_size, total_records)

    chunks[[i]] <- list(
      ids = all_ids[start_idx:end_idx],  # Specific IDs, not skip/limit
      size = length(all_ids[start_idx:end_idx])
    )
  }

  return(list(chunks = chunks, total_records = total_records))
}

#' Retrieve MongoDB data using $in queries for consistency
#' @param Mongo MongoDB connection object
#' @param identifier Field to use as identifier
#' @param chunk_info Chunk information with specific IDs
#' @param verbose Logical for verbose output
#' @return Data frame with consistent results
#' @noRd
getMongoDataConsistent <- function(Mongo, identifier, chunk_info, verbose = FALSE) {
  if (length(chunk_info$ids) == 0) {
    return(data.frame())
  }

  # Use $in query with specific IDs (eliminates skip/limit race conditions)
  query_json <- sprintf('{"%s": {"$in": %s}}',
                        identifier,
                        jsonlite::toJSON(chunk_info$ids, auto_unbox = FALSE))

  if(verbose) message(paste("Using $in query for", length(chunk_info$ids), "specific IDs"))

  df <- Mongo$find(query = query_json)

  if(verbose) message(paste("Retrieved", nrow(df), "rows"))

  return(df)
}

#' Validate results for duplicates and consistency
#' @param df Data frame to validate
#' @param identifier Field to check for duplicates
#' @return Logical indicating if validation passed
#' @noRd
validateResults <- function(df, identifier) {
  if (nrow(df) == 0) return(TRUE)

  duplicates <- sum(duplicated(df[[identifier]]))
  if (duplicates > 0) {
    warning(sprintf("Found %d duplicate records for identifier %s", duplicates, identifier))
    return(FALSE)
  }

  message(sprintf("Validation passed: %d unique records", nrow(df)))
  return(TRUE)
}

#' Fetch data from MongoDB to be stored in a data frame - UPDATED VERSION
#'
#' @param collection The name of the MongoDB collection
#' @param ... Optional column names to filter for. Only rows with non-missing values
#'        in ALL specified columns will be returned. This is useful for filtering
#'        data to only include complete cases for specific variables of interest.
#' @param database The database name (optional)
#' @param identifier Field to use as identifier (optional)
#' @param chunk_size Number of records per chunk (optional)
#' @param verbose Logical; if TRUE, displays detailed progress messages. Default is FALSE.
#' @param interview_date Optional; can be either:
#'        - A date string in various formats (ISO, US, etc.) to filter data up to that date
#'        - A boolean TRUE to return only rows with non-NA interview_date values
#'
#' @importFrom mongolite mongo ssl_options
#' @importFrom parallel detectCores
#' @importFrom future plan multisession
#' @importFrom future future
#' @importFrom future.apply future_lapply
#' @importFrom dplyr bind_rows
#' @importFrom utils flush.console
#' @importFrom stats setNames
#' @importFrom jsonlite toJSON
#'
#' @return A data frame containing the MongoDB data with superkeys first
#' @export
#' @examples
#' \dontrun{
#' # Get data from MongoDB collection
#' data <- mongo("collection")
#' }
mongo <- function(collection, ..., database = NULL, identifier = NULL, chunk_size = NULL, verbose = FALSE, interview_date = NULL) {
  start_time <- Sys.time()
  Mongo <- NULL  # Initialize to NULL for cleanup in on.exit

  # Setup cleanup on exit
  on.exit({
    disconnectMongo(Mongo)
  })

  # Suppress MongoDB messages globally
  options(mongolite.quiet = TRUE)

  # Get configuration
  cfg <- validate_config("mongo")
  if (is.null(database)) {
    database <- cfg$mongo$collection
  }

  # Validate identifier
  if (is.null(identifier)) {
    identifier <- cfg$identifier
  }
  if (is.null(identifier) || any(identifier == "")) {
    stop("No identifier specified in the config file.")
  }

  # Try connecting - will now throw explicit error if collection doesn't exist
  Mongo <- ConnectMongo(collection, database)

  # Find valid identifier
  if (is.null(identifier)) {
    for (key in trimws(strsplit(identifier, ",")[[1]])) {
      count <- Mongo$count(sprintf('{"$s": {"$exists": true, "$ne": ""}}', key))
      if (count > 0) {
        identifier <- key
        break
      }
    }
  }

  if (is.null(identifier)) {
    stop("No valid identifier found in the collection.")
  }

  # Get and display system resources
  mem_info <- getAvailableMemory()
  num_cores <- parallel::detectCores(logical = TRUE)

  # Display system info
  if (!is.null(mem_info$total)) {
    message(sprintf("System resources: %.0fGB RAM, %d-core CPU",
                    mem_info$total, num_cores))
  } else {
    message(sprintf("System resources: %d-core CPU.", num_cores))
  }

  if (!is.null(mem_info$available)) {
    message(sprintf("Memory available: %.0fGB RAM", mem_info$available))
  }

  # NEW: Use consistent chunking instead of skip/limit
  chunk_result <- createConsistentChunks(Mongo, identifier,
                                         if(is.null(chunk_size)) 1000 else chunk_size)
  chunks <- chunk_result$chunks
  total_records <- chunk_result$total_records  # Use actual count from distinct()

  # Calculate optimal parameters with the new optimizations
  params <- calculateResourceParams(total_records, mem_info, num_cores)

  # Use optimized chunk size if not manually specified
  if (is.null(chunk_size)) {
    # Recreate chunks with optimized size
    chunk_result <- createConsistentChunks(Mongo, identifier, params$chunk_size)
    chunks <- chunk_result$chunks
  }

  message(sprintf("Processing: %d chunks x ~%d records in parallel (%d workers)",
                  length(chunks), params$chunk_size, params$workers))

  # Setup parallel processing with optimized worker count
  plan(future::multisession, workers = params$workers)

  # Progress message
  message(sprintf("\nImporting %s records from %s/%s into dataframe...",
                  formatC(total_records, format = "d", big.mark = ","),
                  database, collection))

  # Initialize custom progress bar
  pb <- initializeLoadingAnimation(length(chunks))

  # Process chunks with consistent retrieval
  future_results <- vector("list", length(chunks))
  for (i in seq_along(chunks)) {
    future_results[[i]] <- future({
      temp <- tempfile()
      sink(temp)
      chunk_mongo <- NULL
      on.exit({
        sink()
        unlink(temp)
        disconnectMongo(chunk_mongo)
      })

      tryCatch({
        chunk_mongo <- ConnectMongo(collection, database)
        # Use NEW consistent retrieval method
        data_chunk <- getMongoDataConsistent(chunk_mongo, identifier, chunks[[i]], verbose)
        data_chunk
      }, error = function(e) {
        warning(sprintf("Error processing chunk %d: %s", i, e$message))
        NULL
      })
    })
    updateLoadingAnimation(pb, i)
  }

  # Collect results
  # results <- lapply(future_results, future::value)
  # Recommended new syntax by Futureverse maintainer
  results <- value(future_results)

  # Combine results
  df <- dplyr::bind_rows(results)
  completeLoadingAnimation(pb)

  # NEW: Validate results for consistency
  validation_passed <- validateResults(df, identifier)
  if (!validation_passed) {
    warning("Data consistency validation failed. Results may contain duplicates.")
  }

  # Handle interview_date filtering if needed
  if (!is.null(interview_date) && "interview_date" %in% names(df)) {
    message("Filtering by interview date...", appendLF = FALSE)
    # Convert dates only once for the whole dataframe - much more efficient
    if (!inherits(df$interview_date, "Date")) {
      df$interview_date <- parse_dates_to_iso(df$interview_date, "interview_date")
      # The parse_dates_to_iso function already outputs a message with success percentage
    } else {
      message(" using existing Date format.")
    }

    # Apply the filter based on the parameter type
    if (is.logical(interview_date) && interview_date) {
      # Keep only non-NA interview dates
      df <- df[!is.na(df$interview_date), ]
    } else if (is.character(interview_date) || inherits(interview_date, "Date")) {
      # Parse the cutoff date once
      if (is.character(interview_date)) {
        cutoff_date <- parse_dates_to_iso(interview_date, "filter_date")
        if (all(is.na(cutoff_date))) {
          stop("Failed to parse interview_date parameter: ", interview_date)
        }
      } else {
        cutoff_date <- interview_date
      }
      # Keep only rows with dates up to the cutoff
      # Make sure we only keep non-NA dates before or equal to the cutoff
      df <- df[!is.na(df$interview_date) & df$interview_date <= cutoff_date, ]
    }
  }

  # Harmonize data
  message(sprintf("Harmonizing data on %s...", identifier), appendLF = FALSE)
  clean_df <- taskHarmonization(df, identifier, collection)
  message(sprintf("\rHarmonizing data on %s...done.", identifier))

  # List of allowed superkey columns to prioritize
  allowed_superkey_cols <- c(
    "record_id",
    "src_subject_id",
    "subjectkey",
    "site",
    "subsiteid",
    "sex",
    "race",
    "ethnic_group",
    "phenotype",
    "phenotype_description",
    "state",
    "status",
    "lost_to_followup",
    "lost_to_follow-up",
    "twins_study",
    "sibling_study",
    "family_study",
    "sample_taken",
    "interview_date",
    "interview_age",
    "visit",
    "week"
  )

  # Reorder columns to have superkeys first
  if (is.data.frame(clean_df) && ncol(clean_df) > 0) {
    # Identify which superkey columns are actually in the data
    present_superkeys <- intersect(allowed_superkey_cols, names(clean_df))
    # Get all other columns (non-superkeys)
    other_cols <- setdiff(names(clean_df), present_superkeys)
    # If there are matching superkeys, reorder the columns
    if (length(present_superkeys) > 0) {
      # Create new column order with superkeys first, then other columns
      new_order <- c(present_superkeys, other_cols)
      # Reorder the dataframe
      clean_df <- clean_df[, new_order, drop = FALSE]
    }
  }

  # Check if any column requests were passed via ...
  dots_args <- list(...)
  if (length(dots_args) > 0) {
    # Convert the dots arguments to a character vector
    requested_cols <- as.character(unlist(dots_args))
    # Find which of the requested columns actually exist in the data
    existing_cols <- intersect(requested_cols, names(clean_df))
    if (length(existing_cols) > 0) {
      # Display the names of the columns that were found
      message(sprintf("Found %d of %d requested columns: %s",
                      length(existing_cols),
                      length(requested_cols),
                      paste(existing_cols, collapse = ", ")))
      # Create a filter to keep only rows where ALL requested columns have data
      rows_to_keep <- rep(TRUE, nrow(clean_df))
      for (col in existing_cols) {
        # Check if column values are not NA
        not_na <- !is.na(df[[col]])
        # For non-NA values, check if they're not empty strings (if character type)
        not_empty <- rep(TRUE, nrow(clean_df))
        if (is.character(clean_df[[col]])) {
          not_empty <- clean_df[[col]] != ""
        }
        # Combine the conditions - both not NA and not empty (if applicable)
        has_data <- not_na & not_empty
        # Update the rows_to_keep vector
        rows_to_keep <- rows_to_keep & has_data
      }
      # Apply the filter to keep only rows with data in all requested columns
      original_rows <- nrow(clean_df)
      clean_df <- clean_df[rows_to_keep, ]
      kept_rows <- nrow(clean_df)
      message(sprintf("Kept %d of %d rows where all requested columns have values.",
                      kept_rows, original_rows))
    } else {
      if (length(requested_cols) > 0) {
        warning("None of the requested columns were found in the dataset.")
      }
    }
  }

  # Report execution time
  end_time <- Sys.time()
  duration <- difftime(end_time, start_time, units = "secs")
  message(sprintf("\nData frame '%s' retrieved in %s.", collection, formatDuration(duration)))

  return(clean_df)
}

# ################
## Helper Functions
## ################

#' Setup MongoDB connection with suppressed messages
#' @param collection The name of the collection you want to connect to.
#' @param database The name of the database you cant to connect to.
#' @return A mongolite::mongo object representing the connection to the MongoDB collection.
#' @noRd
ConnectMongo <- function(collection, database) {
  # Validate secrets
  validate_secrets("mongo")
  config <- validate_config("mongo")

  # Get secrets using get_secret() to keep it secret, keep it safe
  connectionString <- get_secret("connectionString")

  if (is.null(database)) {
    database = config$mongo$database
  }

  options <- ssl_options(weak_cert_validation = TRUE, key = "rds-combined-ca-bundle.pem")

  # The key is to use sink() to capture and discard the messages
  temp <- tempfile()
  sink(temp)

  # Create connection without specifying collection first
  base_connection <- mongolite::mongo(
    collection = collection, # This is a system collection that always exists
    db = database,
    url = connectionString,
    verbose = FALSE,
    options = options
  )

  # Check if the collection exists
  collections_list <- getCollectionsFromConnection(base_connection)

  # Close the base connection
  base_connection$disconnect()
  sink()
  unlink(temp)

  # Validate that collection exists
  if (!collection %in% collections_list) {
    stop(sprintf("Collection '%s' does not exist in database '%s'. Available collections: %s",
                 collection, database, paste(collections_list, collapse=", ")))
  }

  # If we get here, the collection exists - create normal connection
  sink(temp)
  on.exit({
    sink()
    unlink(temp)
  })

  Mongo <- mongolite::mongo(
    collection = collection,
    db = database,
    url = connectionString,
    verbose = FALSE,
    options = options
  )

  return(Mongo)
}

#' Safely close MongoDB connection
#' @param mongo A mongolite::mongo connection object
#' @noRd
disconnectMongo <- function(mongo_conn) {
  if (!is.null(mongo_conn)) {
    # Create a temporary sink to capture all output during disconnect
    temp <- tempfile()
    sink(file = temp, type = "output")
    sink(file = temp, type = "message")
    # Try disconnect with warning suppression
    tryCatch({
      suppressWarnings({
        mongo_conn$disconnect()
      })
    }, error = function(e) {
      # Do nothing, we'll handle errors after restoring output
    }, finally = {
      # Restore output streams
      sink(type = "message")
      sink(type = "output")
      unlink(temp)
    })
  }
}

getCollectionsFromConnection <- function(mongo_connection) {
  collections <- mongo_connection$run('{"listCollections":1,"nameOnly":true}')
  return(collections$cursor$firstBatch$name)
}

#' Task Data Harmonization Function
#'
#' This function performs data cleaning and preparation tasks, including handling missing values,
#' converting date formats, and adding necessary columns. It is tailored for a specific dataset
#' structure used in psychological or medical research.
#'
#' @param df A data frame containing the data to be harmonized.
#' @param identifier A string that specifies the unique identifier for the dataset;
#' it influences how date conversions and subsetting are handled.
#' @param collection A string representing the specific collection that needs harmonization.
#'
#' @return A data frame with the harmonized data, including standardized 'visit' column entries,
#' converted interview dates, and added 'measure' column based on the task.
#'
#' @examples
#' \donttest{
#' # Create a sample dataset
#' df <- data.frame(
#'   src_subject_id = 1:3,
#'   visit = c("bl", "6m", "12m"),
#'   score = c(10, 20, 30)
#' )
#' harmonized_data <- taskHarmonization(df, 'src_subject_id', 'task1')
#' }
#'
#' @importFrom stats setNames
#' @noRd
taskHarmonization <- function(df, identifier, collection) {
  # Ensure 'visit' column exists and update it as necessary
  if (!("visit" %in% colnames(df))) {
    df$visit <- "bl"  # Add 'visit' column with all values as "bl" if it doesn't exist
  } else {
    df$visit <- ifelse(is.na(df$visit) | df$visit == "", "bl", df$visit)  # Replace empty or NA 'visit' values with "bl"
  }

  # convert dates (from string ("m/d/Y") to iso date format)
  if ("interview_date" %in% colnames(df)) {
    df$interview_date <- parse_dates_to_iso(df$interview_date, "interview_date")
  }

  return(df)
}

#' Display table of available MongoDB collections
#'
#'
#' Retrieves a list of all available collections in the configured MongoDB database.
#' @param database Optional; the name of the database to connect to. If NULL, uses the database
#'   specified in the configuration file.
#' @return A character vector containing the names of all available collections
#'   in the configured MongoDB database.
#' @export
mongo.index <- function(database = NULL) {
  # Temporarily suppress warnings
  old_warn <- options("warn")

  # Function to suppress specific warnings by pattern
  suppressSpecificWarning <- function(expr, pattern) {
    withCallingHandlers(
      expr,
      warning = function(w) {
        # Check for multiple patterns that might indicate endSessions warning
        if (grepl(pattern, w$message, fixed = TRUE) ||
            grepl("endSessions", w$message, fixed = TRUE) ||
            grepl("Feature not supported", w$message, fixed = TRUE)) {
          invokeRestart("muffleWarning")
        }
      }
    )
  }

  # Function to detect if we're connecting to DocumentDB vs MongoDB
  detectDatabaseType <- function(connectionString) {
    # DocumentDB typically uses different host patterns or connection strings
    # Check for common DocumentDB indicators
    if (grepl("docdb|documentdb", connectionString, ignore.case = TRUE)) {
      return("documentdb")
    }
    # Check for AWS DocumentDB cluster endpoints (they often contain 'docdb')
    if (grepl("amazonaws\\.com.*docdb", connectionString, ignore.case = TRUE)) {
      return("documentdb")
    }
    return("mongodb")
  }

  validate_secrets("mongo")
  config <- validate_config("mongo")

  # Get secrets using get_secret() to keep it secret, keep it safe
  connectionString <- get_secret("connectionString")

  if (is.null(database)) {
    database = config$mongo$database
  }

  options <- ssl_options(weak_cert_validation = TRUE, key = "rds-combined-ca-bundle.pem")

  # Create a temporary sink to capture MongoDB connection messages
  temp <- tempfile()
  sink(temp)
  result <- NULL

  # Detect database type to handle warnings appropriately
  db_type <- detectDatabaseType(connectionString)
  message(sprintf("Detected database type: %s", db_type))

  # Create a direct connection to the database without specifying a collection
  tryCatch({
    # Use suppressSpecificWarning to handle the endSessions warning
    suppressSpecificWarning({
      # Connect directly to the database, not a specific collection
      base_connection <- mongolite::mongo(
        collection = "system.namespaces", # This is a system collection that always exists
        db = database,
        url = connectionString,
        verbose = FALSE,
        options = options
      )

      # Get the list of collections
      collections <- base_connection$run('{"listCollections":1,"nameOnly":true}')
      result <- collections$cursor$firstBatch$name

      # Try to disconnect with warning suppression
      suppressSpecificWarning({
        base_connection$disconnect()
      }, "endSessions")

      # Force garbage collection to clean up any lingering connections
      rm(base_connection)
      invisible(gc(verbose = FALSE))
    }, "endSessions")

    # Additional suppression for any remaining warnings
    suppressWarnings({
      # This should catch any warnings that slip through
    })

    sink()
    unlink(temp)

    # Display collections in a nice format
    if (length(result) > 0) {
      message(sprintf("Available %s Collections:", toupper(database)))
      message("===================================")

      # Calculate adaptive padding based on longest collection name
      max_name_length <- max(nchar(result))
      padding_width <- max(max_name_length + 2, 15)  # At least 15 chars, or longest name + 2

      # Display collections in columns (4 per row)
      for (i in seq(1, length(result), by = 4)) {
        row_collections <- result[i:min(i + 3, length(result))]
        # Pad shorter names to align columns with adaptive width
        padded_names <- sprintf(paste0("%-", padding_width, "s"), row_collections)
        message(paste(padded_names, collapse = " "))
      }

      message("")
      message(sprintf("Total: %d collections", length(result)))
    } else {
      message("No collections found in the database.")
    }

    # Restore previous warning setting
    options(old_warn)
    invisible(result)  # Use invisible() to prevent console printing

  }, error = function(e) {
    sink()
    unlink(temp)
    # Restore previous warning setting before stopping
    options(old_warn)
    stop(paste("Error connecting to MongoDB:", e$message))
  })
}

#' Convert dates to ISO format robustly
#'
#' This function attempts to intelligently parse dates in various formats
#' and convert them to ISO format (YYYY-MM-DD).
#'
#' @param date_vector A vector of date strings to be parsed
#' @param column_name The name of the column being parsed (for error messages)
#' @return A Date vector in ISO format (YYYY-MM-DD)
#' @importFrom lubridate parse_date_time
#' @noRd
parse_dates_to_iso <- function(date_vector, column_name = "date") {
  if (is.null(date_vector) || length(date_vector) == 0) {
    return(date_vector)
  }

  # Skip if already in Date format
  if (inherits(date_vector, "Date")) {
    return(date_vector)
  }

  # If already a POSIXct or POSIXlt, convert to Date
  if (inherits(date_vector, "POSIXt")) {
    return(as.Date(date_vector))
  }

  # Convert to character if not already
  date_vector <- as.character(date_vector)

  # Remove any NA values for analysis
  non_na_dates <- date_vector[!is.na(date_vector) & date_vector != ""]
  if (length(non_na_dates) == 0) {
    # All NA or empty, just return a vector of NAs
    return(as.Date(date_vector))
  }

  # Define a set of possible date formats to try
  possible_formats <- c(
    # American formats
    "mdy", "mdY", "m/d/y", "m/d/Y", "m-d-y", "m-d-Y",
    # European/ISO formats
    "ymd", "Ymd", "y/m/d", "Y/m/d", "y-m-d", "Y-m-d",
    # Other common formats
    "dmy", "dmY", "d/m/y", "d/m/Y", "d-m-y", "d-m-Y",
    # Month name formats
    "mdy_b", "mdY_b", "b_d_y", "b_d_Y",
    "dmy_b", "dmY_b", "d_b_y", "d_b_Y",
    "ymd_b", "Ymd_b", "y_b_d", "Y_b_d"
  )

  # Try to detect the date format
  tryCatch({
    # Sample the first few non-NA dates to guess format
    sample_size <- min(100, length(non_na_dates))
    sample_dates <- non_na_dates[1:sample_size]

    # Try parsing with each format and keep track of success rate
    format_success <- numeric(length(possible_formats))
    for (i in seq_along(possible_formats)) {
      parsed_dates <- suppressWarnings(
        lubridate::parse_date_time(sample_dates, possible_formats[i], quiet = TRUE)
      )
      format_success[i] <- sum(!is.na(parsed_dates)) / length(sample_dates)
    }

    # Find the format with the highest success rate
    best_format_idx <- which.max(format_success)
    best_format <- possible_formats[best_format_idx]

    # If the best format doesn't parse at least 50% of dates, try combo of top formats
    if (format_success[best_format_idx] < 0.5) {
      # Get top 3 formats
      top_formats <- possible_formats[order(format_success, decreasing = TRUE)[1:3]]
      # Try parsing with these formats
      parsed_dates <- suppressWarnings(
        lubridate::parse_date_time(date_vector, top_formats, quiet = TRUE)
      )
    } else {
      # Parse all dates with the best format
      parsed_dates <- suppressWarnings(
        lubridate::parse_date_time(date_vector, best_format, quiet = TRUE)
      )
    }

    # Convert to Date class
    result <- as.Date(parsed_dates)

    # Basic validation: check for impossibly old dates (before 1900) or future dates
    result[result < as.Date("1900-01-01") | result > Sys.Date() + 30] <- NA

    # Log stats about parsing
    success_rate <- sum(!is.na(result)) / length(date_vector) * 100
    message(sprintf("Parsed %s: %.1f%% successful using %s format",
                    column_name, success_rate,
                    ifelse(format_success[best_format_idx] < 0.5,
                           paste(top_formats, collapse=", "), best_format)))

    return(result)

  }, error = function(e) {
    # Fallback: try base R's as.Date with common formats
    warning(sprintf("Advanced date parsing failed for %s: %s. Falling back to basic parsing.",
                    column_name, e$message))

    fallback_formats <- c("%Y-%m-%d", "%m/%d/%Y", "%d/%m/%Y", "%Y/%m/%d")
    for (fmt in fallback_formats) {
      parsed <- suppressWarnings(as.Date(date_vector, format = fmt))
      if (sum(!is.na(parsed)) / length(parsed) > 0.5) {
        message(sprintf("Basic parsing of %s succeeded with format: %s", column_name, fmt))
        return(parsed)
      }
    }

    # If all else fails, return NA
    warning(sprintf("All date parsing methods failed for %s", column_name))
    return(as.Date(rep(NA, length(date_vector))))
  })
}

#' Alias for 'mongo' (DEPRECATED)
#'
#' This function is deprecated. Please use 'mongo' instead.
#' This is a legacy alias for the 'mongo' function to maintain compatibility with older code.
#'
#' @inheritParams mongo
#' @inherit mongo return
#' @export
#' @examples
#' \dontrun{
#' # DEPRECATED - use mongo() instead
#' survey_data <- getTask("task_alias")
#' }
getTask <- function(...) {
  .Deprecated("mongo", package = "wizaRdry")
  mongo(...)
}

#' NDA Field Selection Helper Functions
#'
#' @description
#' Centralized field selection logic to eliminate redundant prompts when creating
#' both submission templates and data definition files.
#'

#' Select NDA Fields Interactively
#'
#' @description
#' Prompts user ONCE for field selection and returns choices.
#' Eliminates redundant prompts between submission template and data definition.
#' This function consolidates all field selection logic that was previously
#' duplicated across createNdaSubmissionTemplate() and createNdaDataDefinition().
#'
#' @param validation_state ValidationState object containing the measure data
#' @param nda_structure NDA structure definition (list with dataElements)
#' @param verbose Logical - print messages (default: TRUE)
#' @param interactive_mode Logical - allow user prompts (default: interactive())
#' @return List with components:
#'   \itemize{
#'     \item selected_fields - Character vector of field names to include
#'     \item user_choices - List of user decisions for reproducibility
#'     \item missing_required - Required fields not in data

#'     \item timepoint_fields - Timepoint fields that were added
#'   }
#' @noRd
select_nda_fields <- function(validation_state, 
                               nda_structure, 
                               verbose = TRUE,
                               interactive_mode = interactive()) {
  
  df <- validation_state$get_df()
  elements <- nda_structure$dataElements
  
  # Start with fields present in the data
  selected_fields <- names(df)
  
  # STEP 1: Remove DCC fields if dcc=FALSE
  # DCC fields should only be included when dcc=TRUE
  if (!validation_state$dcc && length(validation_state$ndar_subject_additions) > 0) {
    dcc_fields_in_selection <- intersect(selected_fields, validation_state$ndar_subject_additions)
    if (length(dcc_fields_in_selection) > 0) {
      selected_fields <- setdiff(selected_fields, validation_state$ndar_subject_additions)
      if (verbose) {
        message(sprintf("Excluding %d DCC field(s) from submission file (dcc=FALSE): %s", 
                       length(dcc_fields_in_selection),
                       paste(dcc_fields_in_selection, collapse=", ")))
      }
    }
  }
  
  # Define super-required ndar_subject01 fields (always needed)
  ndar_required <- c("subjectkey", "src_subject_id", "interview_date", 
                     "interview_age", "sex")
  
  # Get all fields defined in the NDA structure
  structure_fields <- elements$name
  
  # Find required fields that exist in structure definition
  required_in_structure <- elements$name[elements$required == "Required"]
  
  # Find required fields that are NOT in the data and NOT ndar_subject01 fields
  missing_required <- setdiff(
    required_in_structure, 
    c(ndar_required, selected_fields)
  )
  
  # Initialize user choices tracking
  user_choices <- list(
    include_required = FALSE
  )
  
  # Auto-add required fields — required means required, no prompt needed
  if (length(missing_required) > 0) {
    selected_fields <- unique(c(selected_fields, missing_required))
    user_choices$include_required <- TRUE

    # Safe readline with error handling (defined locally for this scope)
    safe_readline_local <- function(prompt, default = "") {
      result <- tryCatch({
        readline(prompt = prompt)
      }, error = function(e) {
        return(default)
      }, interrupt = function(e) {
        message("\nInput cancelled")
        return(default)
      })
      if (is.null(result) || trimws(result) == "") default else result
    }

    # Collect value assignments for remediation script
    assignments <- c()

    for (field in missing_required) {
      # Look up metadata from nda_structure$dataElements
      field_row   <- elements[elements$name == field, ]
      value_range <- if (nrow(field_row) > 0 && !is.na(field_row$valueRange[1])) field_row$valueRange[1] else ""
      notes       <- if (nrow(field_row) > 0 && !is.na(field_row$notes[1]))      field_row$notes[1]      else ""
      desc        <- if (nrow(field_row) > 0 && !is.na(field_row$description[1])) field_row$description[1] else ""

      message(sprintf("\nRequired field auto-added: %s", field))
      if (nchar(desc) > 0)        message(sprintf("  Description: %s", desc))
      if (nchar(value_range) > 0) message(sprintf("  Value range: %s", value_range))
      if (nchar(notes) > 0)       message(sprintf("  Notes:       %s", notes))

      if (interactive_mode) {
        val <- safe_readline_local(
          prompt = sprintf("  Enter value for %s$%s (Enter to leave as NA): ",
                           validation_state$measure_name, field),
          default = ""
        )
        if (nchar(trimws(val)) > 0) {
          assignments[field] <- trimws(val)
        }
      }
    }

    # Write assignments to remediation script
    if (length(assignments) > 0) {
      script_path <- sprintf("./nda/%s/%s.R",
                             validation_state$api,
                             validation_state$measure_name)
      if (file.exists(script_path)) {
        update_cleaning_script(script_path, validation_state$measure_name,
                               renames = c(), drops = c(), assignments = assignments,
                               verbose = verbose)
        message(sprintf("  Written to remediation script: %s", script_path))
      } else {
        if (verbose) message(sprintf("  Script not found, skipping write: %s", script_path))
      }
    }
  }
  
  # STEP 2: Ensure all super required fields are included
  # These are mandatory for NDA submission and must ALWAYS be present
  missing_super_required <- setdiff(ndar_required, selected_fields)
  if (length(missing_super_required) > 0) {
    selected_fields <- c(selected_fields, missing_super_required)
    if (verbose) {
      message(sprintf("Adding %d missing super required field(s): %s", 
                     length(missing_super_required),
                     paste(missing_super_required, collapse=", ")))
    }
  }
  
  # STEP 3: Reorder fields - super required first (in order), then other fields
  # This ensures consistent field ordering across all NDA submissions
  super_required_present <- ndar_required[ndar_required %in% selected_fields]
  other_fields <- setdiff(selected_fields, ndar_required)
  selected_fields <- c(super_required_present, other_fields)
  
  if (verbose) {
    message(sprintf("Field ordering: %d super required first, then %d other fields",
                   length(super_required_present), length(other_fields)))
  }
  
  return(list(
    selected_fields = selected_fields,
    user_choices = user_choices,
    missing_required = missing_required
  ))
}

#' Generate validated NDA submission templates created in the ./nda directory
#'
#' This function processes requests for clean data sequentially for specified measures.
#' It makes a request to the NIH NDA API for the named data structures
#' and runs the associated data remediation routines. It then runs a series of
#' unit tests to verify that the data quality standards are met.
#'
#' @param ... Strings, specifying the measures to process, which can be a Mongo collection, REDCap instrument, or Qualtrics survey.
#' @param csv Optional; Boolean, if TRUE creates a .csv extract in ./tmp.
#' @param rdata Optional; Boolean, if TRUE creates an .rdata extract in ./tmp.
#' @param spss Optional; Boolean, if TRUE creates a .sav extract in ./tmp.
#' @param limited_dataset Optional; Boolean, if TRUE does not perform date-shifting of interview_date or age-capping of interview_age
#' @param skip_prompt Logical. If TRUE (default), skips confirmation prompts unless preferences aren't set yet. If FALSE,
#'   prompts for confirmation unless the user has previously chosen to remember their preference.
#' @param verbose Logical. If TRUE, shows detailed processing information. If FALSE (default), shows only essential user-facing messages.
#' @param strict Logical. If TRUE (default), enforce strict NDA validation: required fields with ANY missing data or
#'   recommended fields with ALL missing data will cause validation failure. If FALSE (lenient mode), missing data
#'   triggers warnings but allows processing to continue.
#' @param dcc Logical. If TRUE, include 11 DCC (Data Coordinating Center) fields from ndar_subject01 
#'   (7 required + 4 recommended). Default FALSE.
#' @return Prints the time taken for the data request process.
#' @importFrom stats na.omit
#' @export
#' @examples
#' \dontrun{
#'   nda("prl", csv=TRUE)
#'   nda("rgpts", "kamin", rdata=TRUE)
#'
#'   # Skip confirmation prompts
#'   nda("prl", csv=TRUE, skip_prompt=TRUE)
#'   
#'   # Show detailed processing information
#'   nda("prl", verbose=TRUE)
#'   
#'   # Use lenient validation mode (allow missing data with warnings)
#'   nda("prl", strict=FALSE)
#'   
#'   # Include DCC fields from ndar_subject01
#'   nda("prl", dcc=TRUE)
#' }
#'
#' @author Joshua Kenney <joshua.kenney@yale.edu>
nda <- function(..., csv = FALSE, rdata = FALSE, spss = FALSE, limited_dataset = FALSE, skip_prompt = TRUE, verbose = FALSE, strict = TRUE, dcc = FALSE) {
  start_time <- Sys.time()

  # Define base path
  path <- "."


  # Required Libraries Setup

  # Prepare lists for REDCap, Qualtrics, and tasks
  csv_list <- tools::file_path_sans_ext(list.files("./nda/csv"))
  redcap_list <- tools::file_path_sans_ext(list.files("./nda/redcap"))
  qualtrics_list <- tools::file_path_sans_ext(list.files("./nda/qualtrics"))
  mongo_list <- tools::file_path_sans_ext(list.files("./nda/mongo"))
  oracle_list <- tools::file_path_sans_ext(list.files("./nda/oracle"))
  sql_list <- tools::file_path_sans_ext(list.files("./nda/sql"))

  # Get identifier from config
  config <- validate_config()
  identifier <- config$identifier
  if (is.null(identifier) || identifier == "") {
    stop("No identifier specified in the config file.")
  }

  # Split identifier if it's a comma-separated string
  if (is.character(identifier)) {
    identifier <- strsplit(identifier, ",")[[1]]
  }

  # Source necessary R scripts from the 'api' directory

  # Create a mapping to store the API type for newly created scripts
  new_script_apis <- list()

  # Validate Measures Function
  validateMeasures <- function(data_list) {
    # Check if input is a dataframe
    if (is.data.frame(data_list)) {
      # Get the name of the dataframe as a string
      data_list <- deparse(substitute(data_list))
    }

    # Ensure data_list is a character vector (in case it's a single string)
    if (!is.character(data_list)) {
      data_list <- as.character(data_list)
    }

    user_prefs_file <- file.path(path, ".wizaRdry_prefs")
    user_prefs <- list(shown_tree = FALSE, auto_create = FALSE, auto_clean = FALSE, auto_nda = FALSE)

    if (file.exists(user_prefs_file)) {
      tryCatch({
        user_prefs <- readRDS(user_prefs_file)
        # Add the auto_nda field if it doesn't exist
        if (is.null(user_prefs$auto_nda)) {
          user_prefs$auto_nda <- FALSE
        }
      }, error = function(e) {
        # If file exists but can't be read, create a new one
        user_prefs <- list(shown_tree = FALSE, auto_create = FALSE, auto_clean = FALSE, auto_nda = FALSE)
      })
    }

    # Create a list name_changes to store updated script_names
    name_changes <- list()

    # Create a character vector for failed validation tracking
    failed_validations <- character(0)

    # Your existing code to determine structures to create
    # For example:
    invalid_structures <- Filter(function(measure) !measure %in% c(csv_list, redcap_list, qualtrics_list, mongo_list, oracle_list, sql_list), data_list)

    # If we have structures to create and need to prompt
    if (length(invalid_structures) > 0) {
      # If skip_prompt is TRUE or user has previously set auto_nda to TRUE, bypass the prompt
      if (!skip_prompt | !user_prefs$auto_nda) {

        template_word <- ifelse(length(invalid_structures) > 1, "templates", "template")
        response <- readline(prompt = sprintf("Would you like to create NDA submission %s for %s now? y/n ",
                                              template_word,
                                              paste(invalid_structures, collapse = ", ")))


        while (!tolower(response) %in% c("y", "n")) {
          response <- readline(prompt = "Please enter either y or n: ")
        }

        # If response is 'y', update auto_nda to TRUE in user preferences
        if (tolower(response) == "y") {
          user_prefs$auto_nda <- TRUE
          saveRDS(user_prefs, user_prefs_file)
        }

        if (tolower(response) == "n") {
          message("NDA submission template script creation cancelled.")
          invokeRestart("abort")  # This exits without the "Error:" prefix
        }
      }

      # Process each invalid script
      for (script_name in invalid_structures) {
        message(sprintf("Processing script: %s\n", script_name))

        original_name <- script_name  # Store the original name for tracking changes

        # Improved validation function for NDA data structure names
        validate_script_name <- function(script_name, nda_base_url = "https://nda.nih.gov/api/datadictionary/v2") {

          # First, check if it's a valid structure name directly
          url <- sprintf("%s/datastructure/%s", nda_base_url, script_name)

          # Add proper error handling for the API request
          response <- tryCatch({
            httr::GET(url, httr::timeout(10))
          }, error = function(e) {
            message("Network error when connecting to NDA API: ", e$message)
            message("Check your internet connection or try again later.")
            return(NULL)
          })

          # If we got a response, check if it's valid
          if (!is.null(response)) {
            status_code <- httr::status_code(response)

            if (status_code == 200) {
              # Try to parse the JSON with handling for errors and empty string responses
              content <- tryCatch({
                raw_content <- rawToChar(response$content) # store raw_content
                if (nchar(raw_content) > 0) { # if there is content
                  jsonlite::fromJSON(raw_content)
                } else { # response is valid but content is empty
                  NULL # set content to NULL and go to similarity search
                }
              }, error = function(e) { # throw an error if something is wrong
                message("Error parsing API response: ", e$message)
                return(NULL)
              })

              if (!is.null(content) && "dataElements" %in% names(content)) {
                message(sprintf("Successfully validated '%s' - found in NDA data dictionary", script_name))
                return(TRUE)
              } else if (!is.null(content)) {
                message(sprintf("API returned content but no dataElements for '%s'", script_name))
              }
            } else if (status_code == 404) {
              message(sprintf("Data structure '%s' not found in NDA data dictionary", script_name))
            } else {
              message(sprintf("API returned status code %d for '%s'", status_code, script_name))
            }
          } else {
            message("Couldn't connect to NDA API")
            return(FALSE)
          }

          # If we get here, the direct lookup failed, so search for similar structures
          search_url <- sprintf("%s/datastructure", nda_base_url) # API endpoint to search and get all NDA data structures

          search_response <- tryCatch({
            httr::GET(search_url, httr::timeout(10)) # Get and store all NDA data structures
          }, error = function(e) {
            message("Network error when searching NDA API: ", e$message)
            return(NULL)
          })

          if (!is.null(search_response) && httr::status_code(search_response) == 200) {
            # Try to parse the JSON with error handling
            all_structures <- tryCatch({
              raw_content <- rawToChar(search_response$content)
              if (nchar(raw_content) > 0) {
                content <- jsonlite::fromJSON(raw_content)
                if (is.data.frame(content)) {
                  content
                } else {
                  message("Unexpected API response format: not a data frame")
                  return(data.frame())
                }
              } else {
                data.frame() # Pass in an empty dataframe if api response is empty
              }
            }, error = function(e) { # Throw an error if something else is wrong.
              message("Error parsing API response: ", e$message)
              return(data.frame())
            })

            if (nrow(all_structures) > 0) {
              # Look for exact or similar matches
              exact_match <- all_structures[which(all_structures$shortName == script_name), ]

              if (nrow(exact_match) > 0) {
                message(sprintf("Found exact match for '%s' in NDA data dictionary", script_name))
                return(TRUE)
              } else {
                # Show closest matches
                # Safely calculate similarities
                similarities <- tryCatch({
                  sapply(all_structures$shortName, function(name) {
                    calculate_jaro_winkler(script_name, name)
                  })
                }, error = function(e) {
                  message("Error calculating similarities: ", e$message)
                  return(numeric(0))
                })

                if (length(similarities) > 0) {
                  # Get all matches sorted by similarity (let pagination handle display)
                  top_matches <- tryCatch({
                    sort(similarities, decreasing = TRUE)
                  }, error = function(e) {
                    message("Error sorting similarities: ", e$message)
                    return(numeric(0))
                  })

                  if (length(top_matches) > 0 && top_matches[1] > 0.7) {
                    message("Data structure name not found in NDA dictionary. Did you mean one of these?")
                    # Paginated selector: show 5 at a time, support n(ext)/p(rev)/m(anual)/number
                    page_size <- 5
                    total <- length(top_matches)
                    page <- 1
                    repeat {
                      start_idx <- (page - 1) * page_size + 1
                      end_idx <- min(page * page_size, total)
                      cat(sprintf("\nShowing matches %d-%d of %d:\n", start_idx, end_idx, total))
                      for (i in seq(start_idx, end_idx)) {
                        local_num <- i - start_idx + 1
                        match_name <- names(top_matches)[i]
                        match_score <- top_matches[i]
                        message(sprintf("%d. %s (%.1f%% match)", local_num, match_name, match_score * 100))
                      }
                      prompt_msg <- "Select 1-5, 'n' next, 'p' prev, 'm' manual, Enter keep original, Esc quit: "
                      use_suggested <- tryCatch({
                        readline(prompt = prompt_msg)
                      }, interrupt = function(e) {
                        message("Interactive selection cancelled (Esc pressed). Keeping original.")
                        ""  # Treat as keep original
                      })
                      # Keep original if empty
                      if (use_suggested == "") {
                        break
                      }
                      # Manual entry
                      if (tolower(use_suggested) == "m") {
                        manual <- readline(prompt = "Enter NDA structure shortName manually (or Enter to cancel): ")
                        if (nzchar(manual)) {
                          script_name <- manual
                          message(sprintf("Using '%s' (manual entry)", script_name))
                          return(script_name)
                        } else {
                          next
                        }
                      }
                      # Next/Prev navigation
                      if (tolower(use_suggested) == "n") {
                        if (page * page_size < total) {
                          page <- page + 1
                        } else {
                          message("Already at last page")
                        }
                        next
                      }
                      if (tolower(use_suggested) == "p") {
                        if (page > 1) {
                          page <- page - 1
                        } else {
                          message("Already at first page")
                        }
                        next
                      }
                      # Numeric selection within current page
                      if (grepl("^[0-9]+$", use_suggested)) {
                        selection_local <- as.numeric(use_suggested)
                        if (!is.na(selection_local) && selection_local >= 1 && selection_local <= (end_idx - start_idx + 1)) {
                          global_index <- start_idx + selection_local - 1
                          script_name <- names(top_matches)[global_index]
                          message(sprintf("Using '%s' instead", script_name))
                          return(script_name)
                        } else {
                          message(sprintf("Please enter a number between 1 and %d", (end_idx - start_idx + 1)))
                          next
                        }
                      }
                      message("Invalid input. Please choose 1-5, 'n', 'p', 'm', or Enter.")
                    }
                  } else {
                    message(sprintf("Warning: '%s' not found in NDA data dictionary and no close matches found", script_name))
                    proceed <- readline(prompt = "Proceed anyway? (y/n): ")
                    # Check that proceed was given a valid y or n response
                    while (!tolower(proceed) %in% c("y", "n")){
                      proceed <- readline(prompt = "Please enter either y or n: ")
                    }
                    if (tolower(proceed) == "y") {
                      return(script_name)  # Return the original name
                    } else {
                      return(FALSE)  # Indicate validation failed
                    }
                  }
                } else {
                  message(sprintf("No similarity data available for '%s'", script_name))
                }
              }
            } else {
              message("No data structures returned from NDA API")
            }
          } else {
            message("Failed to search NDA data dictionary")
          }

          message(sprintf("Warning: Unable to validate '%s' against NDA data dictionary", script_name))
          proceed <- readline(prompt = "Proceed anyway? (y/n): ")
          while (!tolower(proceed) %in% c("y", "n")){
            proceed <- readline(prompt = "Please enter either y or n: ")
          }
          if (tolower(proceed) == "y") {
            return(script_name)  # Return the original name
          } else {
            return(FALSE)  # Indicate validation failed
          }
        }

        # Validate the script name
        validated_name <- validate_script_name(script_name)

        if (is.logical(validated_name) && !validated_name) { # Only runs if validation fails
          message("Script creation cancelled due to validation failure.")
          # Store current original (script) name in failed_validations
          failed_validations <- c(failed_validations, original_name)
          next  # Skip this script and continue with the next one
        } else if (is.character(validated_name)) {
          # A different name was selected
          script_name <- validated_name
        }

        # Track name changes after validated name is stored in script_name
        name_changes[[original_name]] <- script_name

        # If script passes validation, allow user to select api:
        api_selection <- function() {
          options <- c("csv", "mongo", "qualtrics", "redcap", "oracle", "sql")

          cat("\nSelect script type (choose only one):\n")

          for (i in 1:length(options)) {
            cat(i, ":", options[i], "\n")
          }

          # Get user choice
          choice <- as.numeric(readline("Enter number to select option: "))

          while(is.na(choice) || choice < 1 || choice > length(options)) {
            cat("Please enter a valid number between 1 and", length(options), "\n")
            choice <- as.numeric(readline("Enter number to select option: "))
          }

          # Return the selected API
          cat("\n") # Improve spacing and readability
          return(options[choice])
        }

        # Use the function
        selected_api <- api_selection()

        # Store the selected API for this script in our mapping
        new_script_apis[[script_name]] <<- selected_api

        clean_templates <- list(
          mongo = list(
            path = sprintf(file.path(path, "nda", "mongo", "%s.R"), script_name),  # Added .R extension
            content = paste(
              "#",
              sprintf("# nda/mongo/%s.R", script_name),
              "#",
              '# config:  database name is defined in config.yml',
              '# secrets: connectionString is defined in secrets.R',
              '# encrypt: the *.pem file must be placed in the root of this repository',
              "#",
              "# return a list of the collection(s) from MongoDB",
              "# mongo.index()",
              "#",
              "# get collection from MongoDB",
              "# IMPORTANT: both variable name and script filename must match",
              sprintf("%s <- mongo(\"%s\")", script_name, original_name),
              "",
              "# nda remediation code...",
              "",
              "# IMPORTANT: final df name must still match the NDA data structure alias",
              "",
              sep = "\n"
            )
          ),
          qualtrics = list(
            path = sprintf(file.path(path, "nda", "qualtrics", "%s.R"), script_name),  # Added .R extension
            content = paste(
              "#",
              sprintf("# nda/qualtrics/%s.R", script_name),
              "#",
              "# get survey from Qualtrics database",
              "# config:  surveys are defined in config.yml as key-value pairs",
              "# secrets: baseUrls and apiKeys are defined in secrets.R",
              "#",
              "# return a list of the survey(s) from Qualtrics",
              "# qualtrics.index()",
              "#",
              "# get survey from Qualtrics",
              "# IMPORTANT: both variable name and script filename must match",
              sprintf("%s <- qualtrics(\"%s\")", script_name, original_name),
              "",
              "# nda remediation code...",
              "",
              "# IMPORTANT: final df name must still match the NDA data structure alias",
              "",
              sep = "\n"
            )
          ),
          redcap = list(
            path = sprintf(file.path(path, "nda", "redcap", "%s.R"), script_name),  # Added .R extension
            content = paste(
              "#",
              sprintf("# nda/redcap/%s.R", script_name),
              "#",
              "# config:  superkey instrument is defined in config.yml",
              "# secrets: uri and token are defined in secrets.R",
              "#",
              "# return a list of the instrument_name(s) from REDCap",
              "# redcap.index()",
              "#",
              "# get the instrument_name from REDCap",
              "# IMPORTANT: both variable name and script filename must match the NDA data structure alias",
              sprintf("%s <- redcap(\"%s\")", script_name, original_name),
              "",
              "# nda remediation code...",
              "",
              "# IMPORTANT: final df name must still match the NDA data structure alias",
              "",
              sep = "\n"
            )
          ),
          oracle = list(
            path = sprintf(file.path(path, "nda", "oracle", "%s.R"), script_name),  # Added .R extension
            content = paste(
              "#",
              sprintf("# nda/oracle/%s.R", script_name),
              "#",
              "# config:  coming soon...",
              "# secrets: coming soon...",
              "#",
              "# return a list of the table(s) from ORACLE",
              "# oracle.index()",
              "#",
              "# get the table data from ORACLE",
              "# IMPORTANT: both variable name and script filename must match the NDA data structure alias",
              sprintf("%s <- oracle(\"%s\")", script_name, original_name),
              "",
              "# nda remediation code...",
              "",
              "# IMPORTANT: final df name must still match the NDA data structure alias",
              "",
              sep = "\n"
            )
          ),
          sql = list(
            path = sprintf(file.path(path, "nda", "sql", "%s.R"), script_name),  # Added .R extension
            content = paste(
              "#",
              sprintf("# nda/sql/%s.R", script_name),
              "#",
              "# config:  coming soon...",
              "# secrets: coming soon...",
              "#",
              "# return a list of the table(s) from SQL",
              "# sql.index()",
              "#",
              "# get the table data from SQL",
              "# IMPORTANT: both variable name and script filename must match the NDA data structure alias",
              sprintf("%s <- sql(\"%s\")", script_name, original_name),
              "",
              "# nda remediation code...",
              "",
              "# IMPORTANT: final df name must still match the NDA data structure alias",
              "",
              sep = "\n"
            )
          ),
          csv = list(
            path = sprintf(file.path(path, "nda", "csv", "%s.R"), script_name),  # Added .R extension
            content = paste(
              "#",
              sprintf("# nda/csv/%s.R", script_name),
              "#",
              "# get the data from CSV file",
              "# IMPORTANT: both variable name and script filename must match the NDA data structure alias",
              sprintf("%s <- read.csv(\"%s.csv\")", script_name, original_name),
              "",
              "# nda remediation code...",
              "",
              "# IMPORTANT: final df name must still match the NDA data structure alias",
              "",
              sep = "\n"
            )
          )
        )

        # Create directory if it doesn't exist
        template <- clean_templates[[selected_api]]
        dir_path <- dirname(template$path)
        if (!dir.exists(dir_path)) {
          dir.create(dir_path, recursive = TRUE)
        }

        # Write the file if it doesn't exist
        if (!file.exists(template$path)) {
          writeLines(template$content, template$path)
          message(sprintf("Created file: %s", template$path))
          # open script for editing
          rstudioapi::navigateToFile(sprintf("%s",template$path))
        } else {
          message(sprintf("File already exists: %s (skipped)", template$path))
        }
      }
    }

    # After creating new scripts in validateMeasures, update the lists
    csv_list <<- tools::file_path_sans_ext(list.files("./nda/csv"))
    redcap_list <<- tools::file_path_sans_ext(list.files("./nda/redcap"))
    qualtrics_list <<- tools::file_path_sans_ext(list.files("./nda/qualtrics"))
    mongo_list <<- tools::file_path_sans_ext(list.files("./nda/mongo"))
    oracle_list <<- tools::file_path_sans_ext(list.files("./nda/oracle"))
    sql_list <<- tools::file_path_sans_ext(list.files("./nda/sql"))

    # Update data_list with validated names if necessary (only if name_changes is not empty)
    if (length(name_changes) > 0) {
      for (original_name in names(name_changes)) {
        new_name <- name_changes[[original_name]]
        if (original_name != new_name) {
          # Replace the original name with the new name in data_list
          data_list[data_list == original_name] <- new_name
        }
      }
    }

    # Remove failed validations (scripts) from data_list if they exist
    # This will make sure unvalidated scripts are not processed later on
    if (length(failed_validations) > 0) {
      data_list <- data_list[!data_list %in% failed_validations]
    }

    # Return the updated data_list and name_changes invisibly instead of stopping execution
    return(invisible(data_list))
  }

  # Compile data list and validate measures
  data_list <- list(...)

  # This is so the function doesn't break if user enters a variable storing a character vector
  # or a list of strings
  # in other words it let's you do this:
  # vars_i_want <- c('demo','sps','sips_p')
  # dataRequest(vars_i_want)
  if (length(data_list) == 1) {
    data_list = data_list[[1]]
  }

  # Validate measures and potentially create new scripts
  # Validation results contains data_list (updated if necessary)
  validation_results <- validateMeasures(data_list)

  # Update data_list with validated names in main function call (for invalid structures if neccessary)
  data_list <- validation_results # This where data_list is updated globally if necessary

  # Process each measure using processNda function
  for (measure in data_list) {
    # Check if this is a newly created script with a known API
    if (measure %in% names(new_script_apis)) {
      api <- new_script_apis[[measure]]
    } else {
      # Otherwise determine the API from updated lists
      api <- ifelse(measure %in% redcap_list, "redcap",
                    ifelse(measure %in% qualtrics_list, "qualtrics",
                           ifelse(measure %in% mongo_list, "mongo",
                                  ifelse(measure %in% oracle_list, "oracle",
                                         ifelse(measure %in% sql_list, "sql", "csv")))))
    }

    processNda(measure, api, csv, rdata, spss, identifier, start_time, limited_dataset, verbose, strict, dcc)
  }

  # Clean up and record processing time
  # performCleanup()
  # message(Sys.time() - start_time)  # Print time taken for processing
}

processNda <- function(measure, api, csv, rdata, spss, identifier, start_time, limited_dataset = FALSE, verbose = FALSE, strict = TRUE, dcc = FALSE) {
  # Store the origin environment (the one that called this function)
  origin_env <- parent.frame()

  # Add message to user to signify which data structure is being processed
  # This will help when iterating through multiple data structures

  # Create dynamic border length based on the length of the data structure name (measure)
  # 24 is the length of the preceeding message "Processing NDA Structure:"
  border_length <- 27 + nchar(measure)
  border <- paste(rep("=", border_length), collapse = "")

  # Format api name for processing message
  api_formatted <- switch(api,
                          redcap = "REDCap",
                          qualtrics = "Qualtrics",
                          mongo = "MongoDB",
                          oracle = "Oracle",
                          sql = "SQL",
                          csv = "CSV",
                          api)

  # Use dynamic border and formatted api name in message
  message(sprintf("\n%s\n Processing NDA Structure: %s \n Data source: %s\n%s\n", border, measure, api_formatted, border))

  # Add debugging flag
  DEBUG <- FALSE

  if (DEBUG) message("\n[DEBUG] Starting processNda with measure: ", measure, ", api: ", api)

  # Check if input is a dataframe
  if (is.data.frame(measure)) {
    # Get the name of the dataframe as a string
    measure_name <- deparse(substitute(measure))
    if (DEBUG) message("[DEBUG] Input is a dataframe, assigned name: ", measure_name)
  } else {
    measure_name <- measure
    if (DEBUG) message("[DEBUG] Input is not a dataframe, using as name: ", measure_name)
  }

  # Ensure data_list is a character vector (in case it's a single string)
  if (!is.character(measure)) {
    measure <- as.character(measure)
    if (DEBUG) message("[DEBUG] Converted measure to character: ", measure)
  }

  # Construct the path to the measure's cleaning script
  file_path <- sprintf("./nda/%s/%s.R", api, measure)
  message("Fetching ", measure, " with ./nda/", api, "/", measure, ".R\n")

  # Setup cleanup on exit
  on.exit({
    if (exists("mongo_conn") && !is.null(mongo_conn)) {
      tryCatch({
        mongo_conn$disconnect()
      }, error = function(e) {
        warning(sprintf("Error disconnecting from MongoDB: %s", e$message))
      })
    }
    # Clear the mongo connection from memory
    if (exists("mongo_conn")) {
      rm(mongo_conn)
    }
    gc()  # Force garbage collection
  })

  result <- tryCatch({
    if (DEBUG) message("[DEBUG] Sourcing script file: ", file_path)
    base::source(file_path)

    # Initialize package environment for NDA workflow (CRAN compliant)
    if (!exists(".wizaRdry_env", envir = .pkg_env, inherits = FALSE)) {
      if (DEBUG) message("[DEBUG] Creating .wizaRdry_env in package environment")
      assign(".wizaRdry_env", new.env(parent = emptyenv()), envir = .pkg_env)
    }
    wizaRdry_env <- .pkg_env$.wizaRdry_env

    # Get the data frame
    if (DEBUG) message("[DEBUG] Attempting to get dataframe: ", measure)
    df <- base::get0(measure)

    # Store required metadata for later use
    required_field_metadata <- NULL

    if (!is.null(df) && is.data.frame(df)) {
      if (DEBUG) {
        message("[DEBUG] Found dataframe, dims: ", nrow(df), "x", ncol(df))
      }

      # ADD REQUIRED AND RECOMMENDED ELEMENTS
      enhancement_result <- addNdarSubjectElements(df, measure, verbose, dcc)
      df <- enhancement_result$df
      required_field_metadata <- enhancement_result$required_metadata
      recommended_field_metadata <- enhancement_result$recommended_metadata  # NEW

      # Save redcap_event_name values before removal for possible auto-mapping later
      saved_event_names <- NULL
      if (api == "redcap" && "redcap_event_name" %in% names(df)) {
        event_vals <- unique(na.omit(df$redcap_event_name))
        if (length(event_vals) > 1) {
          saved_event_names <- df$redcap_event_name
          if (verbose) message(sprintf("[AUTO] Saved %d event name values for potential visit/week mapping",
                                       length(saved_event_names)))
        }
      }

      # REMOVE STANDARD OUTPUT FIELDS while preserving categorical variables
      # This must happen BEFORE any NDA processing to ensure these fields
      # never make it into submission templates or data definitions
      std_output <- StandardOutput$new()
      cat_vars <- CategoricalVariables$new()
      
      df <- std_output$remove_from_df(
        df = df,
        api = api,
        measure_name = measure,
        categorical_vars = cat_vars,
        verbose = verbose
      )

      # Store cleaned dataframe in all relevant environments
      # This ensures the cleaned version is used by all subsequent operations
      
      # 1. Package environment (authoritative source)
      base::assign(measure, df, envir = wizaRdry_env)
      
      # 2. Global environment (where user data typically lives)
      base::assign(measure, df, envir = .GlobalEnv)
      
      # 3. Origin environment (if accessible)
      tryCatch({
        base::assign(measure, df, envir = origin_env)
      }, error = function(e) {
        # Origin environment not accessible
      })
      
      # 4. Calling environment for user convenience
      tryCatch({
        calling_env <- parent.frame()
        base::assign(measure, df, envir = calling_env)
      }, error = function(e) {
        # Calling environment not accessible
      })

      if (DEBUG) message("[DEBUG] Enhanced dataframe with required and recommended elements")
    }

    # Qualtrics-specific cleanup now handled by StandardOutput class above
    # (Old hardcoded Qualtrics column removal removed - now uses StandardOutput uniformly for all APIs)
    
    if (api == "qualtrics") {
      if (DEBUG) message("[DEBUG] Processing as Qualtrics data")
      if (DEBUG) message("[DEBUG] Calling ndaCheckQualtricsDuplicates")
      ndaCheckQualtricsDuplicates(measure, "qualtrics")
    }

    if (api == "redcap") {
      if (DEBUG) message("[DEBUG] Processing as REDCap data")

      # Re-get the data to ensure we have the latest version
      if (DEBUG) message("[DEBUG] Re-getting dataframe from environment")
      if (exists(measure, envir = origin_env)) {
        df <- base::get(measure, envir = origin_env)
        if (DEBUG) message("[DEBUG] Got from package environment")
      } else if (exists(measure, envir = wizaRdry_env)) {
        df <- base::get(measure, envir = wizaRdry_env)
        if (DEBUG) message("[DEBUG] Got from package environment")
      } else if (exists(measure, envir = origin_env)) {
        df <- base::get(measure, envir = origin_env)
        if (DEBUG) message("[DEBUG] Got from origin_env")
      } else {
        if (DEBUG) message("[DEBUG] ERROR: Can't find dataframe in any environment")
        stop(paste("Object", measure, "not found in any environment"))
      }


      # Define config so you can access primary key
      config <- validate_config()

      # Remove specified REDCap columns, including configured primary key
      cols_to_remove <- c(config$redcap$primary_key, "redcap_event_name")

      if (DEBUG) {
        message("[DEBUG] Current columns: ", paste(names(df), collapse=", "))
        intersection <- intersect(names(df), cols_to_remove)
        message("[DEBUG] Columns to remove (", length(intersection), "): ", paste(intersection, collapse=", "))
      }

      # Remove the columns - different approach
      cols_to_keep <- setdiff(names(df), cols_to_remove)
      if (DEBUG) message("[DEBUG] Columns to keep (", length(cols_to_keep), "): ", paste(head(cols_to_keep, 10), collapse=", "), "...")

      # Create a new dataframe with only the columns to keep
      df_new <- df[, cols_to_keep, drop = FALSE]

      if (DEBUG) {
        message("[DEBUG] After removal: ", ncol(df_new), " columns remain")
        message("[DEBUG] New columns: ", paste(names(df_new), collapse=", "))
      }

      # Reassign the filtered dataframe to BOTH environments
      if (DEBUG) message("[DEBUG] Assigning filtered dataframe back to environments")

      # Update in globalenv
      if (DEBUG) message("[DEBUG] Assigned to package environment")

      # Update in origin_env
      base::assign(measure, df_new, envir = origin_env)
      if (DEBUG) message("[DEBUG] Assigned to origin_env")

        base::assign(measure, df_new, envir = wizaRdry_env)
        if (DEBUG) message("[DEBUG] Assigned to package environment")

      # Update our local df variable for continuing the function
      df <- df_new

      # Verify the changes took effect
      if (DEBUG) {
        if (exists(measure, envir = origin_env)) {
          df_check <- base::get(measure, envir = origin_env)
          message("[DEBUG] Verification - globalenv columns: ", paste(head(names(df_check), 5), collapse=", "), "...")
        }
      }
    }

    if (api == "mongo") {
      if (DEBUG) message("[DEBUG] Processing as Mongo data")

      # Re-get the data to ensure we have the latest version
      if (DEBUG) message("[DEBUG] Re-getting dataframe from environment")
      if (exists(measure, envir = origin_env)) {
        df <- base::get(measure, envir = origin_env)
        if (DEBUG) message("[DEBUG] Got from package environment")
      } else if (exists(measure, envir = wizaRdry_env)) {
        df <- base::get(measure, envir = wizaRdry_env)
        if (DEBUG) message("[DEBUG] Got from package environment")
      } else if (exists(measure, envir = origin_env)) {
        df <- base::get(measure, envir = origin_env)
        if (DEBUG) message("[DEBUG] Got from origin_env")
      } else {
        if (DEBUG) message("[DEBUG] ERROR: Can't find dataframe in any environment")
        stop(paste("Object", measure, "not found in any environment"))
      }

      # Remove specified qualtrics columns
      cols_to_remove <- c("internal_node_id", "trial_type", "trial_index", "stimulus", "time_elapsed")

      if (DEBUG) {
        message("[DEBUG] Current columns: ", paste(names(df), collapse=", "))
        intersection <- intersect(names(df), cols_to_remove)
        message("[DEBUG] Columns to remove (", length(intersection), "): ", paste(intersection, collapse=", "))
      }

      # Remove the columns - different approach
      cols_to_keep <- setdiff(names(df), cols_to_remove)
      if (DEBUG) message("[DEBUG] Columns to keep (", length(cols_to_keep), "): ", paste(head(cols_to_keep, 10), collapse=", "), "...")

      # Create a new dataframe with only the columns to keep
      df_new <- df[, cols_to_keep, drop = FALSE]

      if (DEBUG) {
        message("[DEBUG] After removal: ", ncol(df_new), " columns remain")
        message("[DEBUG] New columns: ", paste(names(df_new), collapse=", "))
      }

      # Reassign the filtered dataframe to BOTH environments
      if (DEBUG) message("[DEBUG] Assigning filtered dataframe back to environments")

      # Update in globalenv
      if (DEBUG) message("[DEBUG] Assigned to package environment")

      # Update in origin_env
      base::assign(measure, df_new, envir = origin_env)
      if (DEBUG) message("[DEBUG] Assigned to origin_env")

        base::assign(measure, df_new, envir = wizaRdry_env)
        if (DEBUG) message("[DEBUG] Assigned to package environment")

      # Update our local df variable for continuing the function
      df <- df_new

      # Verify the changes took effect
      if (DEBUG) {
        if (exists(measure, envir = origin_env)) {
          df_check <- base::get(measure, envir = origin_env)
          message("[DEBUG] Verification - globalenv columns: ", paste(head(names(df_check), 5), collapse=", "), "...")
        }
      }

    }

    # Normalize specific values prior to validation (e.g., race mapping)
    # Difference between NIH reporting and ndar_subject01
    if (!is.null(df) && is.data.frame(df) && "race" %in% names(df)) {
      # Ensure character for safe replacement
      if (is.factor(df$race)) df$race <- as.character(df$race)
      df$race <- trimws(df$race)
      # Map NDA-disallowed label to allowed value
      idx <- which(!is.na(df$race) & df$race == "Native Hawaiian or Pacific Islander")
      if (length(idx) > 0) {
        df$race[idx] <- "Hawaiian or Pacific Islander"
        # Propagate updates to environments used downstream
        base::assign(measure, df, envir = origin_env)
          base::assign(measure, df, envir = wizaRdry_env)
        message(sprintf("Normalized %d 'race' value(s) to 'Hawaiian or Pacific Islander'", length(idx)))
      }
    }

    # show missing data that needs filled
    if (DEBUG) message("[DEBUG] Checking for missing data in required fields")
    missing_data <- df[is.na(df$src_subject_id) | is.na(df$subjectkey) | is.na(df$interview_age) | is.na(df$interview_date) | is.na(df$sex), ]
    if (nrow(missing_data) > 0) {
      if (DEBUG) message("[DEBUG] Found ", nrow(missing_data), " rows with missing required data")
      # Only show View in interactive mode
      if (interactive()) {
        tryCatch({
          View(missing_data)
        }, error = function(e) {
          message("Could not open View (X11 not available)")
        })
      }
    }

    # Re-integrate ndaValidator with proper environment management
    if (DEBUG) message("[DEBUG] Running validation")

    # First check if NDA structure exists
    nda_structure_exists <- TRUE
    nda_structure <- NULL

    tryCatch({
      nda_base_url <- "https://nda.nih.gov/api/datadictionary/v2"
      url <- sprintf("%s/datastructure/%s", nda_base_url, measure)

      response <- httr::GET(url, httr::timeout(10))

      if (httr::status_code(response) == 200) {
        raw_content <- rawToChar(response$content)
        if (nchar(raw_content) > 0) {
          nda_structure <- jsonlite::fromJSON(raw_content)
          if (!"dataElements" %in% names(nda_structure)) {
            nda_structure_exists <- FALSE
          }
        } else {
          nda_structure_exists <- FALSE
        }
      } else if (httr::status_code(response) == 404) {
        nda_structure_exists <- FALSE
        message(sprintf("\nNDA structure '%s' not found in data dictionary - treating as new structure", measure))
      } else {
        nda_structure_exists <- FALSE
        message(sprintf("Could not validate NDA structure '%s' (status: %d) - treating as new structure",
                        measure, httr::status_code(response)))
      }
    }, error = function(e) {
      nda_structure_exists <- FALSE
      message(sprintf("Network error checking NDA structure '%s' - treating as new structure: %s",
                      measure, e$message))
    })

    if (nda_structure_exists && !is.null(nda_structure)) {

      # Enhance existing structure with ndar_subject01 metadata BEFORE validation
      if (!is.null(required_field_metadata) || !is.null(recommended_field_metadata)) {
        if (verbose) message("Enhancing existing NDA structure with ndar_subject01 metadata...")
        merge_result <- mergeNdarSubjectIntoExisting(nda_structure, required_field_metadata, recommended_field_metadata, verbose)
        nda_structure <- merge_result$structure
        ndar_additions <- merge_result$ndar_additions
      } else {
        ndar_additions <- character(0)
      }

      # Auto-map redcap_event_name -> visit or week if NDA structure requires it
      if (!is.null(saved_event_names) && "dataElements" %in% names(nda_structure)) {
        nda_field_names <- nda_structure$dataElements$name
        for (visit_col in c("visit", "week")) {
          if (visit_col %in% nda_field_names) {
            current_vals <- if (visit_col %in% names(df)) unique(na.omit(df[[visit_col]])) else character(0)
            if (length(current_vals) <= 1) {
              df[[visit_col]] <- saved_event_names
              base::assign(measure, df, envir = wizaRdry_env)
              base::assign(measure, df, envir = .GlobalEnv)
              tryCatch(base::assign(measure, df, envir = origin_env), error = function(e) NULL)
              message(sprintf("[AUTO] Mapped redcap_event_name -> %s (%d unique event values)",
                              visit_col, length(unique(na.omit(saved_event_names)))))
              break
            }
          }
        }
      }

      # Recode study-specific missing data codes to NDA-defined special codes
      if (!is.null(config$missing_data_codes)) {
        df <- recode_missing_data_codes(df, config, nda_structure, verbose = verbose)
        base::assign(measure, df, envir = origin_env)
        base::assign(measure, df, envir = wizaRdry_env)
      }

      # EXISTING NDA STRUCTURE - Run full validation
      validation_state <- ndaValidator(measure, api, limited_dataset,
                                       modified_structure = nda_structure,
                                       verbose = verbose,
                                       strict = strict,
                                       dcc = dcc,
                                       ndar_additions = ndar_additions)

      # Handle validation errors gracefully
      if (is.null(validation_state)) {
        message("Warning: Validation returned NULL - this should not happen with new validator")
        # Create error state
        validation_state <- ValidationState$new(measure, api, base::get0(measure), nda_structure, dcc = dcc)
        validation_state$is_valid <- FALSE
        validation_state$errors <- c("Validation returned NULL")
      }
      
      # Store metadata in ValidationState
      if (!is.null(required_field_metadata)) {
        validation_state$required_metadata <- required_field_metadata
      }
      if (!is.null(recommended_field_metadata)) {
        validation_state$recommended_metadata <- recommended_field_metadata
      }
      
      # Store complete ndar_subject01 field list for consistent Excel formatting
      # This ensures fields from ndar_subject01 are formatted identically regardless of dcc parameter
      if (!is.null(enhancement_result$ndar_subject01_all_fields)) {
        validation_state$ndar_subject01_all_fields <- enhancement_result$ndar_subject01_all_fields
        if (verbose) {
          message(sprintf("[CACHE] Stored %d ndar_subject01 fields in validation_state for Excel formatting",
                         length(enhancement_result$ndar_subject01_all_fields)))
        }
      }

      # Update local df variable (DataEnvironment already handled the environments)
      df <- validation_state$get_df()
      if (DEBUG) {
        message(sprintf("[DEBUG] Retrieved dataframe from ValidationState (nrow=%d, ncol=%d)", 
                       nrow(df), ncol(df)))
      }
      
      # Remove DCC fields from dataframe if dcc = FALSE
      # Only remove DCC fields that are NOT in the base NDA structure
      # EXCEPT preserve categorical variables (phenotype, visit, week)
      if (!dcc && !is.null(nda_structure) && "dataElements" %in% names(nda_structure)) {
        base_structure_fields <- nda_structure$dataElements$name
        
        # Get categorical variables
        cat_vars <- CategoricalVariables$new()
        categorical_fields <- cat_vars$get_all()
        
        dcc_fields_in_data <- intersect(names(df), DCC_FIELDS)
        # Only remove DCC fields that are NOT part of the base structure AND NOT categorical
        dcc_fields_to_remove <- setdiff(dcc_fields_in_data, base_structure_fields)
        dcc_fields_to_remove <- setdiff(dcc_fields_to_remove, categorical_fields)
        
        if (length(dcc_fields_to_remove) > 0) {
          df <- df[, !names(df) %in% dcc_fields_to_remove, drop = FALSE]
          
          # Update both environments
          base::assign(measure, df, envir = wizaRdry_env)
          base::assign(measure, df, envir = .GlobalEnv)
          
          # Update ValidationState dataframe
          validation_state$set_df(df)
          
          if (verbose) {
            message(sprintf("\n[DCC EXCLUDED] Removed %d DCC fields from dataframe (dcc=FALSE): %s",
                           length(dcc_fields_to_remove),
                           paste(dcc_fields_to_remove, collapse = ", ")))
            
            # Show preserved categorical DCC fields
            preserved_categorical <- intersect(DCC_FIELDS, categorical_fields)
            preserved_in_data <- intersect(names(df), preserved_categorical)
            if (length(preserved_in_data) > 0) {
              message(sprintf("[DCC PRESERVED] Kept %d categorical DCC field(s): %s",
                             length(preserved_in_data),
                             paste(preserved_in_data, collapse = ", ")))
            }
          }
        }
      }

      # Create NDA files using simplified helper function
      create_nda_files(validation_state, strict = strict, verbose = verbose)

    } else {
      # NEW STRUCTURE - Create ValidationState without NDA validation
      message(sprintf("\nNew structure '%s' (not found in NDA data dictionary)", measure))
      message("Creating ValidationState for new structure with metadata only")

      # Get dataframe
      df <- base::get0(measure)
      
      # Create mock NDA structure
      mock_structure <- list(
        shortName = measure,
        dataElements = data.frame(
          name = character(0),
          type = character(0),
          description = character(0),
          required = character(0),
          stringsAsFactors = FALSE
        )
      )
      
      # Enhance with metadata if available
      if (!is.null(required_field_metadata)) {
        mock_structure <- mergeRequiredMetadata(mock_structure, required_field_metadata, recommended_field_metadata, verbose)
        message("Enhanced new structure with ndar_subject01 required and recommended field metadata")
      }
      
      # Create ValidationState for new structure
      validation_state <- ValidationState$new(measure, api, df, mock_structure)
      validation_state$is_new_structure <- TRUE
      validation_state$bypassed_validation <- TRUE
      
      # Store metadata
      if (!is.null(required_field_metadata)) {
        validation_state$required_metadata <- required_field_metadata
      }
      if (!is.null(recommended_field_metadata)) {
        validation_state$recommended_metadata <- recommended_field_metadata
      }
      
      # Remove DCC fields from dataframe if dcc = FALSE (NEW structures)
      # For new structures, remove ALL DCC fields since there's no base structure to check
      # EXCEPT preserve categorical variables (phenotype, visit, week)
      if (!dcc) {
        # Get categorical variables
        cat_vars <- CategoricalVariables$new()
        categorical_fields <- cat_vars$get_all()
        
        # DCC fields to remove: DCC_FIELDS minus categorical variables
        dcc_fields_to_remove <- setdiff(DCC_FIELDS, categorical_fields)
        dcc_fields_in_data <- intersect(names(df), dcc_fields_to_remove)
        
        if (length(dcc_fields_in_data) > 0) {
          df <- df[, !names(df) %in% dcc_fields_to_remove, drop = FALSE]
          
          # Update both environments
          base::assign(measure, df, envir = wizaRdry_env)
          base::assign(measure, df, envir = .GlobalEnv)
          
          # Update ValidationState dataframe
          validation_state$set_df(df)
          
          if (verbose) {
            message(sprintf("\n[DCC EXCLUDED] Removed %d DCC fields from dataframe (dcc=FALSE): %s",
                           length(dcc_fields_in_data),
                           paste(dcc_fields_in_data, collapse = ", ")))
            
            # Show preserved categorical DCC fields
            preserved_categorical <- intersect(DCC_FIELDS, categorical_fields)
            preserved_in_data <- intersect(names(df), preserved_categorical)
            if (length(preserved_in_data) > 0) {
              message(sprintf("[DCC PRESERVED] Kept %d categorical DCC field(s): %s",
                             length(preserved_in_data),
                             paste(preserved_in_data, collapse = ", ")))
            }
          }
        }
      }

      # Create NDA files using simplified helper function
      create_nda_files(validation_state, strict = strict, verbose = verbose)
    }

    # Update local df variable
    df <- validation_state$get_df()
    
    # Add de-identification summary (verbose mode only)
    if (limited_dataset == FALSE && verbose) {
      message("\nDataset has been de-identified using date-shifting and age-capping.")
    }

    # Add validation summary (non-verbose mode) - LAST THING before completion
    if (!verbose) {
      message("Validation Summary:")
      
      if (validation_state$is_valid) {
        message("- Status: PASSED")
      } else {
        message("- Status: FAILED")
        if (length(validation_state$errors) > 0) {
          message(sprintf("- Errors: %d", length(validation_state$errors)))
        }
      }
      
      message(sprintf("- Structure Type: %s", 
                     if(validation_state$is_new_structure) "NEW" 
                     else if(validation_state$is_modified_structure) "MODIFIED" 
                     else "EXISTING"))
      
      if (!validation_state$is_new_structure && validation_state$is_valid) {
        message(sprintf("- Modified: %s", 
                       if(validation_state$is_modified_structure) "YES" else "NO"))
      }
      
      if (validation_state$is_valid) {
        message(sprintf("- Needs Data Definition: %s (reason: %s)", 
                       if(validation_state$needs_data_definition()) "YES" else "NO",
                       validation_state$get_modification_reason()))
      }
      
      # Show warnings if any
      if (length(validation_state$warnings) > 0) {
        message(sprintf("- Warnings: %d", length(validation_state$warnings)))
        for (warning in validation_state$warnings) {
          message(sprintf("  - %s", warning))
        }
      }
      
      message("")  # Blank line
    }

    # Calculate elapsed time and show completion message
    elapsed_time <- as.numeric(difftime(Sys.time(), start_time, units = "secs"))
    
    if (validation_state$is_valid) {
      message(sprintf("[OK] NDA processing complete in %.1f seconds", elapsed_time))
    } else {
      message("[ERROR] NDA validation failed")
    }

    # Play completion sound AFTER all processing, including data definition export
    if (exists("validation_state") && inherits(validation_state, "ValidationState")) {
      # Use success/fail tone for existing structures; treasure for new (bypassed) ones
      tone <- if (validation_state$is_new_structure) "treasure" else if (validation_state$is_valid) "mario" else "wilhelm"
      if (DEBUG) message("[DEBUG] Playing completion sound: ", tone)
      try(beepr::beep(tone), silent = TRUE)
    } else {
      try(beepr::beep("treasure"), silent = TRUE)
    }
  }, error = function(e) {
    if (DEBUG) message("[DEBUG] Error caught: ", e$message)
    # Check if identifier is valid
    if (length(identifier) == 0 || all(is.na(identifier))) {
      message("An error occurred: ", e$message)  # General error message
    } else {
      message("Error with ", measure, ": ", e$message)  # Specific error message
    }
    NULL  # Return NULL on error
  })

  # Final verification
  if (DEBUG && exists(measure)) {
    final_df <- base::get0(measure)
    if (!is.null(final_df) && is.data.frame(final_df)) {
      message("[DEBUG] FINAL CHECK - Columns in ", measure, ":", paste(head(names(final_df), 10), collapse=", "), "...")
      qualtrics_cols <- intersect(names(final_df), c("StartDate", "EndDate", "Status", "Progress", "ResponseId"))
      if (length(qualtrics_cols) > 0) {
        message("[DEBUG] WARNING: Qualtrics columns still present: ", paste(qualtrics_cols, collapse=", "))
      } else {
        message("[DEBUG] SUCCESS: No Qualtrics columns found in final dataset")
      }
    }
  }

  return(result)  # Return the result of the processing
}

#' Process super required fields from ndar_subject01
#'
#' @description
#' Processes the 5 super required fields from ndar_subject01 metadata:
#' subjectkey, src_subject_id, interview_date, interview_age, sex.
#' These fields are mandatory for ALL NDA submissions.
#' Adds missing required fields and ensures correct data types.
#'
#' @param df Data frame to process
#' @param required_elements Data frame of super required field metadata from ndar_subject01
#' @param verbose Logical - print detailed processing messages
#' @return Modified data frame with super required fields processed
#' @noRd
processRequiredFields <- function(df, required_elements, verbose = FALSE) {
  if (is.null(required_elements) || nrow(required_elements) == 0) {
    return(df)
  }
  
  if (verbose) message("\n--- Processing ALL REQUIRED fields ---")
  
  for (i in 1:nrow(required_elements)) {
    col_name <- required_elements$name[i]
    col_type <- required_elements$type[i]
    
    if (verbose) message(sprintf("Processing required field: %s (%s)", col_name, col_type))
    
    # Determine R data type
    if (grepl("String|GUID", col_type, ignore.case = TRUE)) {
      default_value <- NA_character_
      conversion_func <- as.character
    } else if (grepl("Integer", col_type, ignore.case = TRUE)) {
      default_value <- NA_integer_
      conversion_func <- as.integer
    } else if (grepl("Float", col_type, ignore.case = TRUE)) {
      default_value <- NA_real_
      conversion_func <- as.numeric
    } else if (grepl("Date", col_type, ignore.case = TRUE)) {
      default_value <- NA_character_
      conversion_func <- as.character
    } else {
      default_value <- NA_character_
      conversion_func <- as.character
    }
    
    # Process field
    if (col_name %in% names(df)) {
      # Column exists - ensure correct type and preserve existing values
      existing_data <- df[[col_name]]
      non_na_count <- sum(!is.na(existing_data))
      
      if (non_na_count > 0) {
        if (verbose) message(sprintf("  - Field exists with %d values, ensuring %s type",
                                    non_na_count, col_type))
        df[[col_name]] <- tryCatch({
          conversion_func(existing_data)
        }, error = function(e) {
          if (verbose) message(sprintf("    Warning: Type conversion failed for %s", col_name))
          existing_data
        })
      } else {
        if (verbose) message(sprintf("  - Field exists but empty, setting to %s type", col_type))
        df[[col_name]] <- rep(default_value, nrow(df))
      }
    } else {
      # Column doesn't exist - add it (REQUIRED fields are always added)
      if (verbose) message(sprintf("  - Adding new required field as %s type", col_type))
      df[[col_name]] <- rep(default_value, nrow(df))
    }
  }
  
  if (!verbose) {
    message(sprintf("[OK] Processed %d super required field%s", 
                   nrow(required_elements),
                   if (nrow(required_elements) > 1) "s" else ""))
  }
  
  return(df)
}

#' Process recommended fields from ndar_subject01
#'
#' @description
#' Processes ONLY recommended fields that are common between ndar_subject01 and the dataframe.
#' Ensures correct data types but does NOT add missing fields.
#'
#' @param df Data frame to process
#' @param recommended_elements Data frame of recommended field metadata (already filtered to common only)
#' @param verbose Logical - print detailed processing messages
#' @return Modified data frame with recommended fields processed
#' @noRd
processRecommendedFields <- function(df, recommended_elements, verbose = FALSE) {
  if (is.null(recommended_elements) || nrow(recommended_elements) == 0) {
    if (!verbose) {
      message("[OK] No common recommended fields to process")
    } else {
      message("\n[OK] No common recommended fields to process")
    }
    return(df)
  }
  
  if (verbose) message("\n--- Processing COMMON RECOMMENDED fields ---")
  
  for (i in 1:nrow(recommended_elements)) {
    col_name <- recommended_elements$name[i]
    col_type <- recommended_elements$type[i]
    
    if (verbose) message(sprintf("Processing common recommended field: %s (%s)", col_name, col_type))
    
    # Determine R data type (same logic as required)
    if (grepl("String|GUID", col_type, ignore.case = TRUE)) {
      default_value <- NA_character_
      conversion_func <- as.character
    } else if (grepl("Integer", col_type, ignore.case = TRUE)) {
      default_value <- NA_integer_
      conversion_func <- as.integer
    } else if (grepl("Float", col_type, ignore.case = TRUE)) {
      default_value <- NA_real_
      conversion_func <- as.numeric
    } else if (grepl("Date", col_type, ignore.case = TRUE)) {
      default_value <- NA_character_
      conversion_func <- as.character
    } else {
      default_value <- NA_character_
      conversion_func <- as.character
    }
    
    # Since we filtered for common fields, we know the column exists
    # Just ensure correct type and preserve existing values
    existing_data <- df[[col_name]]
    non_na_count <- sum(!is.na(existing_data))
    
    if (non_na_count > 0) {
      if (verbose) {
        message(sprintf("  - Field exists with %d values, ensuring %s type",
                       non_na_count, col_type))
      }
      df[[col_name]] <- tryCatch({
        conversion_func(existing_data)
      }, error = function(e) {
        if (verbose) message(sprintf("    Warning: Type conversion failed for %s", col_name))
        existing_data
      })
    } else {
      if (verbose) message(sprintf("  - Field exists but empty, setting to %s type", col_type))
      df[[col_name]] <- rep(default_value, nrow(df))
    }
  }
  
  if (!verbose) {
    message(sprintf("[OK] Processed %d common recommended field%s",
                   nrow(recommended_elements),
                   if (nrow(recommended_elements) > 1) "s" else ""))
  }
  
  return(df)
}

#' Add ndar_subject01 elements to dataframe
#'
#' @description
#' Fetches ndar_subject01 from NDA API and adds:
#' - 5 super required fields (subjectkey, src_subject_id, interview_date, interview_age, sex)
#' - Optional DCC fields (7 required + 4 recommended) when dcc = TRUE
#'
#' @param df Data frame to enhance
#' @param measure Measure name
#' @param verbose Logical - print detailed processing messages
#' @param dcc Logical - include DCC fields from ndar_subject01
#' @return List with: df, required_metadata, recommended_metadata
#' @noRd
addNdarSubjectElements <- function(df, measure, verbose = FALSE, dcc = FALSE) {
  # Initialize return structure
  result <- list(
    df = df,
    required_metadata = NULL,
    recommended_metadata = NULL
  )

  tryCatch({
    message("Fetching required and recommended elements from ndar_subject01 API...")

    nda_base_url <- "https://nda.nih.gov/api/datadictionary/v2"
    url <- sprintf("%s/datastructure/ndar_subject01", nda_base_url)

    response <- httr::GET(url, httr::timeout(10))

    if (httr::status_code(response) == 200) {
      raw_content <- rawToChar(response$content)
      if (nchar(raw_content) > 0) {
        subject_structure <- jsonlite::fromJSON(raw_content)

        if ("dataElements" %in% names(subject_structure)) {
          # Force dataElements to base data.frame to avoid tibble evaluation issues
          if (inherits(subject_structure$dataElements, c("tbl_df", "tbl", "data.table"))) {
            subject_structure$dataElements <- as.data.frame(subject_structure$dataElements, stringsAsFactors = FALSE)
          }
          
          # Store complete ndar_subject01 structure and field list for consistent formatting
          # This ensures fields that exist in ndar_subject01 are formatted the same
          # regardless of dcc parameter (fixes formatting inconsistency bug)
          result$ndar_subject01_structure <- subject_structure
          result$ndar_subject01_all_fields <- subject_structure$dataElements$name
          
          if (verbose) {
            message(sprintf("[CACHE] Stored %d ndar_subject01 field names for formatting consistency",
                           length(result$ndar_subject01_all_fields)))
          }
          
          # Get the 5 super required fields that are mandatory for all NDA submissions
          super_required_fields <- SUPER_REQUIRED_FIELDS
          
          # Get REQUIRED elements
          all_required_elements <- subject_structure$dataElements[
            subject_structure$dataElements$required == "Required",
          ]
          
          # Determine which required fields to include based on dcc parameter
          if (dcc) {
            # Include super required + DCC required fields (if they exist in data)
            dcc_required_in_data <- intersect(DCC_REQUIRED_FIELDS, names(df))
            dcc_required_in_ndar <- dcc_required_in_data[dcc_required_in_data %in% all_required_elements$name]
            required_fields_to_include <- c(super_required_fields, dcc_required_in_ndar)
          } else {
            # Only include the 5 super required fields
            required_fields_to_include <- super_required_fields
          }
          
          required_elements <- all_required_elements[
            all_required_elements$name %in% required_fields_to_include,
          ]

          # Get RECOMMENDED elements
          all_recommended_elements <- subject_structure$dataElements[
            subject_structure$dataElements$required == "Recommended",
          ]

          # Filter RECOMMENDED based on dcc parameter
          common_recommended_names <- intersect(all_recommended_elements$name, names(df))
          
          if (dcc) {
            # Include only DCC recommended fields that exist in data
            dcc_recommended_in_data <- intersect(DCC_RECOMMENDED_FIELDS, common_recommended_names)
            recommended_elements <- all_recommended_elements[
              all_recommended_elements$name %in% dcc_recommended_in_data,
            ]
          } else {
            # EXCLUDE all DCC fields from recommended (fixes auto-include bug)
            recommended_names_no_dcc <- setdiff(common_recommended_names, DCC_FIELDS)
            recommended_elements <- all_recommended_elements[
              all_recommended_elements$name %in% recommended_names_no_dcc,
            ]
          }


          # Display what we found
          if (!verbose) {
            message("\n=== STEP 1: Processing Required and Recommended Fields ===\n")
            message("Required Elements:")
          }
          
          if (dcc) {
            dcc_required_count <- sum(required_elements$name %in% DCC_REQUIRED_FIELDS)
            message(sprintf("  Found %d required elements from ndar_subject01 (5 super required + %d DCC required)", 
                           nrow(required_elements), dcc_required_count))
          } else {
            message(sprintf("  Found %d super required elements from ndar_subject01 (subjectkey, src_subject_id, interview_date, interview_age, sex)", 
                           nrow(required_elements)))
          }
          
          if (!verbose) {
            message("\nRecommended Elements:")
          }
          
          if (dcc) {
            message(sprintf("  Found %d DCC recommended elements in dataframe", nrow(recommended_elements)))
          } else {
            message(sprintf("  Found %d recommended elements, %d are common with dataframe (excluding DCC fields)",
                           nrow(all_recommended_elements), nrow(recommended_elements)))
          }
          
          if (nrow(recommended_elements) > 0) {
            message(sprintf("  Common recommended fields: %s",
                           paste(recommended_elements$name, collapse = ", ")))
          }
          
          # PRESERVE ALL METADATA for required and COMMON recommended only
          result$required_metadata <- required_elements
          result$recommended_metadata <- recommended_elements
          
          # Process fields using helper functions
          message("")  # Blank line before processing messages
          df <- processRequiredFields(df, required_elements, verbose)
          df <- processRecommendedFields(df, recommended_elements, verbose)
          
          # Reorder columns: REQUIRED first, then COMMON RECOMMENDED, then others
          required_field_names <- required_elements$name
          recommended_field_names <- recommended_elements$name  # Only common ones
          other_field_names <- setdiff(names(df), c(required_field_names, recommended_field_names))
          
          present_required <- intersect(required_field_names, names(df))
          present_recommended <- intersect(recommended_field_names, names(df))
          
          df <- df[, c(present_required, present_recommended, other_field_names)]
          result$df <- df
          
          message("\n[OK] Successfully processed ndar_subject01 elements")
          if (verbose) {
            message(sprintf("Super required fields (%d): %s",
                            length(present_required), paste(present_required, collapse = ", ")))
            if (length(present_recommended) > 0) {
              message(sprintf("Common recommended fields (%d): %s",
                              length(present_recommended), paste(present_recommended, collapse = ", ")))
            } else {
              message("Common recommended fields (0): none")
            }
          }

        } else {
          message("No dataElements found in ndar_subject01 API response")
        }
      } else {
        message("Empty response from ndar_subject01 API")
      }
    } else {
      message(sprintf("Failed to fetch ndar_subject01 from API (status: %d)", httr::status_code(response)))
    }

  }, error = function(e) {
    message("Error fetching ndar_subject01 from API: ", e$message)
    message("Continuing without required/recommended field updates...")
  })

  return(result)
}

# Function to merge required metadata into new structure definitions
mergeRequiredMetadata <- function(new_structure, required_metadata, recommended_metadata = NULL, verbose = FALSE) {
  if (is.null(required_metadata) || nrow(required_metadata) == 0) {
    return(new_structure)
  }

  tryCatch({
    if (verbose) message("Merging required and recommended ndar_subject01 metadata into new structure...")

    # Combine required and recommended metadata
    all_metadata <- required_metadata
    if (!is.null(recommended_metadata) && nrow(recommended_metadata) > 0) {
      all_metadata <- rbind(required_metadata, recommended_metadata)
      if (verbose) {
        message(sprintf("Combined %d required + %d recommended = %d total elements",
                        nrow(required_metadata), nrow(recommended_metadata), nrow(all_metadata)))
      }
    }

    # For new structures, prepend the combined metadata with their full definitions
    if (is.null(new_structure$dataElements) || nrow(new_structure$dataElements) == 0) {
      # No existing elements, use combined metadata as base
      new_structure$dataElements <- all_metadata
      if (verbose) message(sprintf("Added %d elements as base structure", nrow(all_metadata)))
    } else {
      # Existing elements - merge without overwriting
      existing_names <- new_structure$dataElements$name
      combined_names <- all_metadata$name

      # Find elements not already in the structure
      missing_elements <- all_metadata[!combined_names %in% existing_names, ]

      if (nrow(missing_elements) > 0) {
        # Prepend missing elements
        new_structure$dataElements <- rbind(missing_elements, new_structure$dataElements)
        if (verbose) message(sprintf("Added %d missing elements", nrow(missing_elements)))
      }

      if (verbose) message("Required and recommended elements preserved with complete metadata")
    }

  }, error = function(e) {
    message("Error merging metadata: ", e$message)
  })

  return(new_structure)
}

# Enhanced mergeNdarSubjectIntoExisting function to properly override field definitions

mergeNdarSubjectIntoExisting <- function(existing_structure, required_metadata, recommended_metadata, verbose = FALSE) {
  if (is.null(existing_structure) || is.null(existing_structure$dataElements)) {
    return(existing_structure)
  }

  tryCatch({
    if (verbose) message("Overriding existing structure fields with authoritative ndar_subject01 definitions...")

    # Combine required and recommended metadata from ndar_subject01
    ndar_metadata <- data.frame()
    if (!is.null(required_metadata) && nrow(required_metadata) > 0) {
      ndar_metadata <- rbind(ndar_metadata, required_metadata)
    }
    if (!is.null(recommended_metadata) && nrow(recommended_metadata) > 0) {
      ndar_metadata <- rbind(ndar_metadata, recommended_metadata)
    }

    if (nrow(ndar_metadata) == 0) {
      return(existing_structure)
    }

    # Ensure dataElements is base data.frame to avoid tibble evaluation issues
    if (!is.null(existing_structure$dataElements) && 
        inherits(existing_structure$dataElements, c("tbl_df", "tbl", "data.table"))) {
      existing_structure$dataElements <- as.data.frame(existing_structure$dataElements, stringsAsFactors = FALSE)
    }

    # Get existing and ndar field names
    existing_names <- existing_structure$dataElements$name
    ndar_names <- ndar_metadata$name

    # Find fields that should be REPLACED with ndar_subject01 definitions
    fields_to_replace <- intersect(existing_names, ndar_names)

    # Find fields that are completely new from ndar_subject01
    new_fields <- setdiff(ndar_names, existing_names)

    # Find fields that remain unchanged from original structure
    unchanged_fields <- setdiff(existing_names, ndar_names)

    if (verbose) {
      message(sprintf("Replacing %d fields with ndar_subject01 definitions: %s",
                      length(fields_to_replace), paste(head(fields_to_replace, 8), collapse = ", ")))

      if (length(new_fields) > 0) {
        message(sprintf("Adding %d new ndar_subject01 fields: %s",
                        length(new_fields), paste(head(new_fields, 5), collapse = ", ")))
      }

      message(sprintf("Keeping %d original structure fields unchanged: %s",
                      length(unchanged_fields), paste(head(unchanged_fields, 5), collapse = ", ")))
    }

    # Build the new dataElements in priority order:
    # 1. REQUIRED fields from ndar_subject01 (first priority)
    # 2. RECOMMENDED fields from ndar_subject01 (second priority)
    # 3. Original structure fields that weren't replaced (last)

    new_data_elements <- data.frame()

    # Add REQUIRED fields from ndar_subject01 first
    if (!is.null(required_metadata) && nrow(required_metadata) > 0) {
      required_in_structure <- required_metadata[required_metadata$name %in% c(fields_to_replace, new_fields), ]
      if (nrow(required_in_structure) > 0) {
        new_data_elements <- rbind(new_data_elements, required_in_structure)
        if (verbose) message(sprintf("Added %d required fields from ndar_subject01", nrow(required_in_structure)))
      }
    }

    # Add RECOMMENDED fields from ndar_subject01 second
    if (!is.null(recommended_metadata) && nrow(recommended_metadata) > 0) {
      recommended_in_structure <- recommended_metadata[recommended_metadata$name %in% c(fields_to_replace, new_fields), ]
      if (nrow(recommended_in_structure) > 0) {
        new_data_elements <- rbind(new_data_elements, recommended_in_structure)
        if (verbose) message(sprintf("Added %d recommended fields from ndar_subject01", nrow(recommended_in_structure)))
      }
    }

    # Add unchanged original structure fields last
    unchanged_elements <- existing_structure$dataElements[
      existing_structure$dataElements$name %in% unchanged_fields,
    ]
    if (nrow(unchanged_elements) > 0) {
      new_data_elements <- rbind(new_data_elements, unchanged_elements)
      if (verbose) message(sprintf("Preserved %d original structure fields", nrow(unchanged_elements)))
    }

    # Replace the dataElements with the new prioritized version
    existing_structure$dataElements <- new_data_elements

    if (verbose) {
      message(sprintf("Enhanced structure now has %d total fields with ndar_subject01 definitions taking priority",
                      nrow(existing_structure$dataElements)))

      # Show what the key fields now look like
      key_fields <- c("race", "phenotype", "phenotype_description", "twins_study", "sibling_study")
      present_key_fields <- intersect(key_fields, existing_structure$dataElements$name)
      if (length(present_key_fields) > 0) {
        message("Key ndar_subject01 field definitions now active:")
        for (field in present_key_fields) {
          field_def <- existing_structure$dataElements[existing_structure$dataElements$name == field, ]
          if (nrow(field_def) > 0) {
            message(sprintf("  - %s: %s (%s) - %s",
                          field, field_def$type[1], field_def$required[1],
                          substr(field_def$description[1], 1, 50)))
          }
        }
      }
    }
    
    # Return both structure and list of fields added from ndar_subject01
    # The 'new_fields' variable was already calculated at line 1662
    return(list(
      structure = existing_structure,
      ndar_additions = new_fields
    ))

  }, error = function(e) {
    message("Error merging ndar_subject01 into existing structure: ", e$message)
    # Return structure with empty additions on error
    return(list(
      structure = existing_structure,
      ndar_additions = character(0)
    ))
  })
}

# Add helper function for MongoDB cleanup
disconnectMongo <- function(mongo) {
  if (!is.null(mongo)) {
    tryCatch({
      mongo$disconnect()
      rm(list = deparse(substitute(mongo)), envir = parent.frame())
    }, error = function(e) {
      warning(sprintf("Error disconnecting from MongoDB: %s", e$message))
    })
  }
}


# Cleanup Function
performCleanup <- function() {
  # Placeholder for cleanup operations, like disconnecting from databases
}

# Helper function to preserve MM/DD/YYYY format
# preserveDateFormat <- function(df, limited_dataset = limited_dataset) {
#   if ("interview_date" %in% names(df)) {
#     # Convert to Date first to ensure consistent handling
#     dates <- as.Date(df$interview_date, format = "%m/%d/%Y")
#
#     # Apply format based on limited_dataset flag
#     df$interview_date <- format(dates,
#                                 ifelse(limited_dataset, "%m/%d/%Y", "%m/01/%Y"))
#
#     # Add debug message
#     message("Applying date format with limited_dataset = ", limited_dataset)
#     message("Sample dates after formatting: ",
#             paste(head(df$interview_date), collapse=", "))
#   }
#   return(df)
# }

# Helper function to display time savings.
formatElapsedTime <- function(start_time) {
  time_diff <- Sys.time() - start_time
  units <- attr(time_diff, "units")

  formatted_time <- switch(units,
                           "secs" = sprintf("%.1f seconds", time_diff),
                           "mins" = sprintf("%.1f minutes", time_diff),
                           sprintf("%.1f %s", time_diff, units)
  )

  message("Formatted for NDA in ", formatted_time, ".")
}

#' Alias for 'nda' (DEPRECATED)
#'
#' This function is deprecated. Please use 'nda' instead.
#' This is a legacy alias for the 'nda' function to maintain compatibility with older code.
#'
#' @inheritParams nda
#' @inherit nda return
#' @export
#' @examples
#' \dontrun{
#' # DEPRECATED - use nda() instead
#' prl01 <- ndaRequest("prl01")
#' }
ndaRequest <- function(...) {
  .Deprecated("nda", package = "wizaRdry")
  nda(...)
}

#' Process CAPR data
#' 
#' @param df Data frame containing CAPR data
#' @param instrument_name Name of the instrument
#' @return Processed data frame
#' @importFrom dplyr filter select mutate group_by first contains
#' @importFrom rlang .data

#' @noRd
processCaprData <- function(df, instrument_name) {
  # Debug: Print available columns
  message("Debug: Available columns in processCaprData:")
  message(paste("Columns:", paste(names(df), collapse = ", ")))
  message(paste("Number of rows:", nrow(df)))
  
  # Convert and filter src_subject_id
  if ("src_subject_id" %in% names(df)) {
    df$src_subject_id <- as.numeric(df$src_subject_id)
    # df <- filter(df, between(df$src_subject_id, 10000, 71110)) # between seems() to cause error
    # might be less flexible character to numeric
    # src_subject_id may download as character sometimes
    df <- dplyr::filter(df, .data$src_subject_id > 10000, .data$src_subject_id < 79110)
  } else {
    stop("src_subject_id column is required but not found in the data")
  }
  # include guard clauses for mesaures that require aditional filtering beyond form name
  if (instrument_name == "scid_scoresheet") {
    df <- df %>% dplyr::select(contains(c("src_subject_id", "redcap_event_name", "scid_", "scip_", "mdd_", "pdd_"))) # scid_p18a was misspelled in the dataframe, that is why there is a "scip" variable :)
  }
  df$src_subject_id <- as.character(df$src_subject_id)
  
  # create a visit variable based on redcap_event_name
  ## not over-writing with rename(), so that redcap_event_name can do a "soft retire"
  if ("redcap_event_name" %in% names(df)) {
    df <- df %>% dplyr::mutate(visit = .data$redcap_event_name)
  } else {
    message("Warning: redcap_event_name column not found. Creating placeholder visit column.")
    df <- df %>% dplyr::mutate(visit = NA_character_)
  }
  
  # align redcap_event_name-ing convention with more natural language
  df <- df %>% dplyr::mutate(visit = ifelse(.data$visit == "baseline_arm_1", "bl",
                                            ifelse(.data$visit == "12m_arm_1", "12m",
                                                   ifelse(.data$visit == "24m_arm_1", "24m", NA)
                                            )
  ))
  
  # Check phenotype data - CAPR uses labels by default, so no recoding needed
  if ("phenotype" %in% names(df)) {
    message("Debug: Phenotype column found. Sample values:")
    message(paste("Unique phenotype values:", paste(unique(df$phenotype), collapse = ", ")))
    message(paste("Number of non-NA phenotypes:", sum(!is.na(df$phenotype))))
    
    # No recoding needed since we're getting label data (e.g., "HC", "CHR", "HSC")
    # Just ensure it's a character column
    df$phenotype <- as.character(df$phenotype)
  } else {
    message("Warning: phenotype column not found in data. Creating placeholder phenotype column.")
    df <- df %>% dplyr::mutate(phenotype = NA_character_)
  }
  
  #make sure phenotype doesn't change after baseline visit
  if ("phenotype" %in% names(df) && any(!is.na(df$phenotype))) {
    df <- df %>% 
      dplyr::mutate(visit = factor(.data$visit, levels = c('bl','12m','24m')),
                    phenotype = factor(.data$phenotype)) %>%
      dplyr::group_by(.data$src_subject_id) %>%  
      dplyr::mutate(baseline_pheno = dplyr::first(.data$phenotype)) %>% 
      dplyr::mutate(phenotype = .data$baseline_pheno) %>% 
      dplyr::select(-.data$baseline_pheno) 
    
    # Remove rows where phenotype is NA
    # but first print warning and say how many folks are getting removed
    phenotype_nas <- df[is.na(df$phenotype),]
    if (nrow(phenotype_nas) > 0) {
      message(paste0('removing ', nrow(phenotype_nas),
                   ' subjects because they have NA for phenotype. This generally',
                   ' should not happen. Below are the subject IDs and visit dates ',
                   'for these people. They should be inspected and fixed in redcap'))
      message(paste0(phenotype_nas$src_subject_id, ' ', phenotype_nas$visit))
      df <- df[!is.na(df$phenotype), ]
    }
    
    # Remove rows where phenotype is 'ineligible' or 'withdrawn' (if these labels exist)
    # Note: CAPR might use different labels, so we check what's actually in the data
    ineligible_labels <- c("Ineligible", "Withdrew")
    existing_ineligible <- intersect(unique(df$phenotype), ineligible_labels)
    if (length(existing_ineligible) > 0) {
      message(paste("Removing phenotypes:", paste(existing_ineligible, collapse = ", ")))
      df <- df[!(df$phenotype %in% existing_ineligible), ]
    }
  } else {
    message("Warning: No valid phenotype data found. Skipping phenotype processing.")
  }
  
  # create a site variable based on src_institution_name
  ## not over-writing with rename(), so that redcap_event_name can do a "soft retire"
  if ("src_institution_id" %in% names(df)) {
    df <- df %>% dplyr::mutate(site = .data$src_institution_id)
    # get rid of deprecated variable names is good practice
    df <- subset(df, select = -src_institution_id)
  } else {
    # If src_institution_id doesn't exist, create a placeholder site column
    message("Warning: src_institution_id column not found in data. Creating placeholder site column.")
    df <- df %>% dplyr::mutate(site = NA_character_)
  }
  
  # convert dates - CAPR stores interview dates in superkey, not in individual instruments
  # Note: Date formatting is done in getRedcap before this function, so columns may already be Date objects
  if ("int_start" %in% names(df) && "int_end" %in% names(df)) {
    message("Debug: Interview date columns found:")
    message(paste("int_start sample values:", paste(head(df$int_start, 3), collapse = ", ")))
    message(paste("int_end sample values:", paste(head(df$int_end, 3), collapse = ", ")))
    message(paste("int_start class:", class(df$int_start)[1]))
    message(paste("int_end class:", class(df$int_end)[1]))
    
    # Handle both Date objects and character/numeric dates
    if (inherits(df$int_start, "Date") && inherits(df$int_end, "Date")) {
      # Dates are already properly formatted
      df$int_diff <- as.numeric(df$int_end - df$int_start)
      df$interview_date <- df$int_start
      message("Using pre-formatted Date objects for calculations")
    } else {
      # Convert to dates if they're not already
      df$int_start <- as.Date(df$int_start)
      df$int_end <- as.Date(df$int_end)
      df$int_diff <- as.numeric(df$int_end - df$int_start)
      df$interview_date <- df$int_start
      message("Converted columns to Date objects for calculations")
    }
  } else if ("interview_age" %in% names(df)) {
    message("Debug: CAPR structure detected - interview_age found")
    message(paste("interview_age sample values:", paste(head(df$interview_age, 3), collapse = ", ")))
    
    # Since interview_age is calculated from interview_date - subject_dob,
    # we should try to reconstruct interview_date if possible
    if ("subject_dob" %in% names(df)) {
      message("Attempting to reconstruct interview_date from interview_age and subject_dob")
      
      # Convert subject_dob to Date if it isn't already
      if (!inherits(df$subject_dob, "Date")) {
        df$subject_dob <- as.Date(df$subject_dob)
      }
      
      # Calculate interview_date: subject_dob + interview_age
      # Note: interview_age is in months with >=15 days rounded up
      # Use base R date arithmetic: add months by converting to days
      df$interview_date <- df$subject_dob + (df$interview_age * 30.44)
      df$int_diff <- NA_real_  # Can't calculate duration without start/end times
      
      message("Reconstructed interview_date from interview_age and subject_dob")
    } else {
      message("Warning: subject_dob not found, cannot reconstruct interview_date")
      df$int_diff <- NA_real_
      df$interview_date <- NA
    }
  } else {
    message("Warning: Neither interview date columns nor interview_age found. Skipping interview date processing.")
    message(paste("Available columns:", paste(names(df), collapse = ", ")))
    df$int_diff <- NA_real_
    df$interview_date <- NA
  }
  
  # Clean up CAPR-specific columns that are no longer needed
  columns_to_remove <- c("subject_dob", "group_status")
  existing_columns_to_remove <- intersect(names(df), columns_to_remove)
  
  if (length(existing_columns_to_remove) > 0) {
    message("Cleaning up CAPR-specific columns: ", paste(existing_columns_to_remove, collapse = ", "))
    df <- df[, !names(df) %in% existing_columns_to_remove, drop = FALSE]
  }
  
  # Standardize phenotype and sex values
  if ("phenotype" %in% names(df)) {
    df$phenotype <- tolower(df$phenotype)
    message("Converted phenotype values to lowercase")
  }
  
  if ("sex" %in% names(df)) {
    df$sex <- ifelse(df$sex == "Male", "M", 
                     ifelse(df$sex == "Female", "F", df$sex))
    message("Recoded sex values: Male -> M, Female -> F")
  }
  
  # Reorder columns to put calculated/important fields first
  calculated_fields <- c("phenotype", "site", "visit", "interview_date", "int_diff")
  existing_calculated_fields <- intersect(names(df), calculated_fields)
  other_fields <- setdiff(names(df), existing_calculated_fields)
  
  if (length(existing_calculated_fields) > 0) {
    message("Reordering columns to put calculated fields first")
    df <- df[, c(existing_calculated_fields, other_fields)]
  }
  
  return(df)
}

# Helper function: clean REDCap labels (remove HTML tags)
# Defined at package level for use across functions
# @noRd
rc_clean_label <- function(x) {
  if (is.null(x) || is.na(x)) return("")
  gsub("<br>", " ", x, fixed = TRUE)
}

# Helper function: parse REDCap choices -> numeric codes for ValueRange + mapping text for Notes
# Defined at package level so it's available to both createNdaDataDefinition and exportDataDefinition
# @noRd
rc_parse_choices_codes_and_notes <- function(choices_string) {
  if (is.null(choices_string) || is.na(choices_string) || choices_string == "") {
    return(list(codes = "", notes = ""))
  }
  parts <- strsplit(choices_string, "\\|")[[1]]
  parts <- trimws(parts)
  codes <- character(0)
  note_pairs <- character(0)
  for (p in parts) {
    if (p == "") next
    kv <- strsplit(p, ",")[[1]]
    if (length(kv) >= 2) {
      code <- trimws(kv[1])
      label <- trimws(paste(kv[-1], collapse = ","))
      label <- rc_clean_label(label)
      codes <- c(codes, code)
      note_pairs <- c(note_pairs, paste0(code, "=", label))
    } else {
      # fallback: if no comma, treat as bare code
      code <- trimws(kv[1])
      codes <- c(codes, code)
    }
  }
  list(
    codes = if (length(codes)) paste(codes, collapse = "; ") else "",
    notes = if (length(note_pairs)) paste(note_pairs, collapse = "; ") else ""
  )
}

# Helper function: expand numeric value range string to vector of integers
# Converts NDA value range notation (e.g., "1::14", "1;2;3", "1::7;999") into numeric vector
# Used for comparing NDA ranges with REDCap choices and data values
# @noRd
expand_numeric_value_range <- function(val_range_str) {
  if (is.null(val_range_str) || identical(val_range_str, "") || is.na(val_range_str)) return(NULL)
  parts <- trimws(strsplit(as.character(val_range_str), ";")[[1]])
  parts <- parts[nzchar(parts)]
  if (length(parts) == 0) return(NULL)
  out_vals <- integer(0)
  for (p in parts) {
    if (grepl("::", p, fixed = TRUE)) {
      # Handle range notation (e.g., "1::14")
      bounds <- suppressWarnings(as.numeric(strsplit(p, "::", fixed = TRUE)[[1]]))
      if (length(bounds) == 2 && !any(is.na(bounds))) {
        a <- floor(bounds[1])
        b <- floor(bounds[2])
        if (b >= a && (b - a) <= 10000) { # safety bound to prevent huge expansions
          out_vals <- c(out_vals, seq(a, b))
        }
      }
    } else {
      # Handle individual values (e.g., "999")
      num <- suppressWarnings(as.numeric(p))
      if (!is.na(num)) out_vals <- c(out_vals, floor(num))
    }
  }
  unique(out_vals)
}

#' Create NDA Data Definition File
#'
#' @description
#' Creates Excel data definition file for NDA structure registration or updates.
#' Supports both ValidationState objects (new) and legacy parameters (old).
#'
#' @param submission_template ValidationState object OR legacy list with columns
#' @param nda_structure NDA structure (optional if using ValidationState)
#' @param measure_name Measure name (optional if using ValidationState)
#' @param data_frame Data frame (optional if using ValidationState)
#' @param interactive_mode Interactive mode flag
#' @param selected_fields Character vector of field names to include (optional, from centralized selection)
#' @param skip_prompts Logical - if TRUE, skip all interactive prompts (default: FALSE)
#' @return Invisible NULL
#' @noRd
createNdaDataDefinition <- function(submission_template, nda_structure = NULL, measure_name = NULL, data_frame = NULL, interactive_mode = interactive(), selected_fields = NULL, skip_prompts = FALSE, verbose = FALSE) {

  # NEW PATH: Handle ValidationState objects
  if (inherits(submission_template, "ValidationState")) {
    validation_state <- submission_template
    measure_name <- validation_state$measure_name
    data_frame <- validation_state$get_df()
    nda_structure <- validation_state$nda_structure
    
    # Extract ndar_subject_additions for DCC field tracking
    ndar_subject_additions <- validation_state$ndar_subject_additions
    if (is.null(ndar_subject_additions)) {
      ndar_subject_additions <- character(0)
    }
    
    # Build reason message (used internally, not displayed)
    reason <- validation_state$get_modification_reason()
    
    # In verbose mode, append DCC exclusion info if dcc=FALSE
    if (verbose && !validation_state$dcc) {
      dcc_count <- sum(names(data_frame) %in% DCC_FIELDS)
      if (dcc_count > 0) {
        reason <- sprintf("%s (%d DCC fields excluded)", reason, dcc_count)
      }
    }
    
    # Convert to legacy format for rest of function
    submission_template <- list(columns = names(data_frame))
  } else {
    # LEGACY PATH: submission_template is a list
    message("\n[DATA DEFINITION] Using legacy input format")
    
    # Initialize ndar_subject_additions as empty for legacy path
    ndar_subject_additions <- character(0)
    
    # Try to get the data frame from the global environment if not provided
    if (is.null(data_frame)) {
      data_frame <- tryCatch({
        base::get0(measure_name)
      }, error = function(e) NULL)
    }
  }

  # Load missing data codes from config if available
  missing_data_codes <- NULL
  tryCatch({
    config_env <- ConfigEnv$new("config.yml")
    missing_data_codes <- config_env$get_missing_data_codes()
  }, error = function(e) {
    # Silently fail if config can't be loaded
    missing_data_codes <- NULL
  })

  # Build optional description maps from external sources (REDCap/Qualtrics)
  redcap_label_map <- NULL
  redcap_choices_map <- NULL
  qualtrics_label_map <- NULL

  # Attempt to fetch REDCap metadata and map field_name -> field_label and field_name -> select_choices_or_calculations
  try({
    # Prefer using the provided data_frame if it carries instrument attribute
    if (is.data.frame(data_frame) && !is.null(attr(data_frame, "redcap_instrument"))) {
      rc_meta <- redcap.dict(data_frame)
    } else {
      rc_meta <- redcap.dict(measure_name)
    }
    if (is.data.frame(rc_meta) && all(c("field_name", "field_label") %in% names(rc_meta))) {
      # Create a named vector mapping field names to labels
      redcap_label_map <- rc_meta$field_label
      names(redcap_label_map) <- rc_meta$field_name

      # Create a named vector mapping field names to select_choices_or_calculations for ValueRange
      if ("select_choices_or_calculations" %in% names(rc_meta)) {
        redcap_choices_map <- rc_meta$select_choices_or_calculations
        names(redcap_choices_map) <- rc_meta$field_name
        # Remove NA values
        redcap_choices_map <- redcap_choices_map[!is.na(redcap_choices_map) & redcap_choices_map != ""]
      } else {
        redcap_choices_map <- NULL
      }
    } else {
      redcap_choices_map <- NULL
    }
  }, silent = TRUE)

  # Function to parse RedCap select_choices_or_calculations and convert to NDA ValueRange format
  parse_redcap_choices_to_value_range <- function(choices_string) {
    if (is.null(choices_string) || is.na(choices_string) || choices_string == "") {
      return("")
    }

    # Split by pipe (|) to get individual choices
    choices <- strsplit(choices_string, "\\|")[[1]]
    choices <- trimws(choices)

    # Parse each choice (format: "value, label" or "value, label")
    parsed_choices <- character(0)
    for (choice in choices) {
      if (choice != "") {
        # Split by comma to separate value and label
        parts <- strsplit(choice, ",")[[1]]
        if (length(parts) >= 2) {
          value <- trimws(parts[1])
          label <- trimws(paste(parts[-1], collapse = ","))  # In case label contains commas
          # Format as NDA ValueRange: "value; label"
          parsed_choices <- c(parsed_choices, paste0(value, "; ", label))
        } else if (length(parts) == 1) {
          # If no comma, treat the whole thing as a value
          parsed_choices <- c(parsed_choices, trimws(parts[1]))
        }
      }
    }

    # Join with semicolons for NDA format
    return(paste(parsed_choices, collapse = "; "))
  }

  # Function to infer value ranges from data tuples when choices are not available
  infer_value_range_from_data <- function(data_vector, data_type) {
    if (is.null(data_vector) || length(data_vector) == 0) {
      return("")
    }

    # Remove NAs for analysis
    clean_data <- data_vector[!is.na(data_vector)]
    if (length(clean_data) == 0) {
      return("")
    }

    unique_vals <- unique(clean_data)
    unique_count <- length(unique_vals)

    if (data_type == "Integer") {
      # For integers, check if values look like categorical codes
      if (unique_count <= 20 && all(clean_data >= 0 & clean_data <= 99)) {
        # Check if values are sequential (e.g., 0, 1, 2)
        sorted_vals <- sort(as.numeric(unique_vals))
        
        # Check for sequential pattern
        is_sequential <- all(diff(sorted_vals) == 1)
        
        if (is_sequential && unique_count >= 3) {
          # Use range notation for sequential values (0::2 instead of 0;1;2)
          return(paste0(as.integer(sorted_vals[1]), "::", as.integer(sorted_vals[length(sorted_vals)])))
        } else {
          # Use semicolon-separated list for non-sequential categorical data
          return(paste(sorted_vals, collapse = ";"))
        }
      } else {
        # Likely continuous data - use range notation
        min_val <- min(clean_data, na.rm = TRUE)
        max_val <- max(clean_data, na.rm = TRUE)
        return(paste0(as.integer(min_val), "::", as.integer(max_val)))
      }
    } else if (data_type == "String") {
      # For strings, list unique values if reasonable number
      if (unique_count <= 20) {
        sorted_vals <- sort(as.character(unique_vals))
        return(paste(sorted_vals, collapse = ";"))
      } else {
        return(paste0("Multiple categorical values (", unique_count, " unique)"))
      }
    } else if (data_type == "Float") {
      # For floats, don't specify value ranges (too restrictive)
      return("")
    }

    return("")
  }

  # Helper: clean HTML breaks from labels
  rc_clean_label <- function(x) {
    if (is.null(x) || is.na(x)) return("")
    gsub("<br>", " ", x, fixed = TRUE)
  }

  # Helper: enrich REDCap checkbox children (e.g., parent___1 or parent_other)
  # Using parent field's label and choices to derive description/valueRange/notes
  rc_enrich_checkbox_child <- function(column_name, rc_meta) {
    out <- list(description = NULL, valueRange = NULL, notes = NULL, data_type = NULL, size = NULL)
    if (is.null(rc_meta) || !is.data.frame(rc_meta) || !all(c("field_name", "field_type") %in% names(rc_meta))) return(out)

    parent <- NULL
    suffix <- NULL
    if (grepl("___", column_name, fixed = TRUE)) {
      parent <- sub("___.*$", "", column_name)
      suffix <- sub("^.*___", "", column_name)
    } else if (grepl("_other$", column_name)) {
      parent <- sub("_other$", "", column_name)
      suffix <- "other"
    }

    # Attempt to infer parent via branching_logic if name heuristic fails
    if ((is.null(parent) || !(parent %in% rc_meta$field_name)) && "branching_logic" %in% names(rc_meta)) {
      this_row <- rc_meta[rc_meta$field_name == column_name, , drop = FALSE]
      if (nrow(this_row) == 1) {
        bl <- as.character(this_row$branching_logic %||% "")
        m <- regmatches(bl, regexpr("\\[([A-Za-z0-9_]+)\\(", bl, perl = TRUE))
        if (length(m) == 1 && nzchar(m)) {
          parent_candidate <- sub("^\\[", "", m)
          parent_candidate <- sub("\\($", "", parent_candidate)
          if (parent_candidate %in% rc_meta$field_name) parent <- parent_candidate
        }
      }
    }

    if (is.null(parent) || parent == "") return(out)

    parent_row <- rc_meta[rc_meta$field_name == parent, , drop = FALSE]
    if (nrow(parent_row) != 1) return(out)
    if (!identical(as.character(parent_row$field_type), "checkbox")) return(out)

    parent_label <- as.character(parent_row$field_label %||% "")
    parent_label <- rc_clean_label(parent_label)
    choices <- as.character(parent_row$select_choices_or_calculations %||% "")

    # Build map from coded value -> label
    choice_map <- list()
    if (nzchar(choices)) {
      parts <- strsplit(choices, "\\|")[[1]]
      for (p in trimws(parts)) {
        kv <- strsplit(p, ",")[[1]]
        if (length(kv) >= 2) {
          key <- trimws(kv[1])
          val <- trimws(paste(kv[-1], collapse = ","))
          val <- rc_clean_label(val)
          choice_map[[key]] <- val
        }
      }
    }

    # Numeric child like parent___1
    if (!is.null(suffix) && grepl("^[0-9]+$", suffix)) {
      label <- choice_map[[suffix]] %||% parent_label
      # Compose description as single line: Parent label - choice label
      if (nzchar(parent_label) && nzchar(label) && parent_label != label) {
        out$description <- paste0(parent_label, " - ", label)
      } else {
        out$description <- label
      }
      out$valueRange <- "0;1"
      out$notes <- "0=No;1=Yes"
      # Per current expectation: keep as String with Size 255
      out$data_type <- "String"
      out$size <- 255
      return(out)
    }

    # Other text child
    if (!is.null(suffix) && identical(tolower(suffix), "other")) {
      out$description <- paste0(parent_label, ": If you selected 'other', please describe.")
      out$valueRange <- ""
      out$notes <- ""
      out$data_type <- "String"
      out$size <- 255
      return(out)
    }

    out
  }

  # Helper: parse REDCap choices -> numeric codes only for ValueRange + mapping text for Notes
  rc_parse_choices_codes_and_notes <- function(choices_string) {
    if (is.null(choices_string) || is.na(choices_string) || choices_string == "") {
      return(list(codes = "", notes = ""))
    }
    parts <- strsplit(choices_string, "\\|")[[1]]
    parts <- trimws(parts)
    codes <- character(0)
    note_pairs <- character(0)
    for (p in parts) {
      if (p == "") next
      kv <- strsplit(p, ",")[[1]]
      if (length(kv) >= 2) {
        code <- trimws(kv[1])
        label <- trimws(paste(kv[-1], collapse = ","))
        label <- rc_clean_label(label)
        codes <- c(codes, code)
        note_pairs <- c(note_pairs, paste0(code, "=", label))
      } else {
        # fallback: if no comma, treat as bare code
        code <- trimws(kv[1])
        codes <- c(codes, code)
      }
    }
    list(
      codes = if (length(codes)) paste(codes, collapse = "; ") else "",
      notes = if (length(note_pairs)) paste(note_pairs, collapse = "; ") else ""
    )
  }

  # Helper: use REDCap validation/type to derive ValueRange when appropriate
  rc_derive_value_range_from_validation <- function(meta_row) {
    if (is.null(meta_row) || nrow(meta_row) == 0) return("")
    vtype <- meta_row$text_validation_type_or_show_slider_number
    vmin <- suppressWarnings(as.numeric(meta_row$text_validation_min))
    vmax <- suppressWarnings(as.numeric(meta_row$text_validation_max))
    ftype <- meta_row$field_type

    # Integer-like
    if (!is.null(vtype) && !is.na(vtype) && grepl("integer", vtype, ignore.case = TRUE)) {
      if (!is.na(vmin) && !is.na(vmax)) return(paste0(as.integer(vmin), "::", as.integer(vmax)))
      return("")
    }

    # Slider (numeric)
    if (!is.null(vtype) && !is.na(vtype) && grepl("slider|number", vtype, ignore.case = TRUE)) {
      if (!is.na(vmin) && !is.na(vmax)) return(paste0(vmin, "::", vmax))
      return("")
    }

    # Field type specific
    if (!is.null(ftype) && ftype %in% c("radio", "checkbox", "dropdown")) {
      # Handled via choices, not validation
      return("")
    }

    # Date/time: let NDA handle as type; no numeric range
    if (!is.null(vtype) && grepl("date|datetime", vtype, ignore.case = TRUE)) {
      return("")
    }

    return("")
  }

  # Attempt to fetch Qualtrics column map and map variable -> question text
  try({
    # qualtrics.dict accepts either a data frame or an alias
    if (is.data.frame(data_frame)) {
      q_map <- qualtrics.dict(data_frame)
    } else {
      q_map <- qualtrics.dict(measure_name)
    }
    if (is.data.frame(q_map)) {
      # Heuristically identify variable and question-text columns
      var_cols <- c("qname", "VariableName", "variable", "Variable", "col_name", "column", "Column", "name", "Name")
      txt_cols <- c("description", "QuestionText", "question_text", "Question", "Label", "label", "Description")
      var_col <- var_cols[var_cols %in% names(q_map)]
      txt_col <- txt_cols[txt_cols %in% names(q_map)]
      if (length(var_col) > 0 && length(txt_col) > 0) {
        var_col <- var_col[1]
        txt_col <- txt_col[1]
        # Filter out NAs and create mapping
        valid_rows <- !is.na(q_map[[var_col]]) & !is.na(q_map[[txt_col]])
        if (any(valid_rows)) {
          qualtrics_label_map <- q_map[[txt_col]][valid_rows]
          names(qualtrics_label_map) <- q_map[[var_col]][valid_rows]
        }
      }
    }
  }, silent = TRUE)

  # Handle both character vector and list formats
  if (is.character(submission_template)) {
    submission_template <- list(columns = submission_template)
  }

  # Validate inputs
  if (!is.list(submission_template)) {
    stop("submission_template must be a list or character vector")
  }

  if (!is.list(nda_structure)) {
    stop("nda_structure must be a list")
  }

  if (missing(measure_name) || !is.character(measure_name)) {
    stop("measure_name must be provided as a character string")
  }

  # Extract selected columns from submission template OR use provided selected_fields
  # Handle different possible structures
  selected_columns <- NULL
  
  # Priority 1: Use provided selected_fields parameter (from centralized selection)
  if (!is.null(selected_fields)) {
    selected_columns <- selected_fields
  } else if ("columns" %in% names(submission_template)) {
    selected_columns <- submission_template$columns
  } else if ("selected_fields" %in% names(submission_template)) {
    selected_columns <- submission_template$selected_fields
  } else if (is.character(submission_template)) {
    selected_columns <- submission_template
  } else {
    stop("Could not extract selected columns from submission_template")
  }

  # Define fields that should never be included as new or modified structures
  # They can still be included only if part of essential ndar_subject01 and unchanged
  excluded_from_change <- c("state", "lost_to_followup", "lost_to_follow-up", "study_status")

  # Get essential NDA fields dynamically from ndar_subject01 API
  # This ensures we always have the current required elements with complete metadata
  essential_nda_fields <- tryCatch({
    # Try to get required field metadata from the global environment or nda_structure
    required_metadata <- NULL

    # First, try to get from nda_structure if it contains ndar_subject01 data
    if (!is.null(nda_structure) && "dataElements" %in% names(nda_structure)) {
      required_elements <- nda_structure$dataElements[nda_structure$dataElements$required == "Required", ]
      if (nrow(required_elements) > 0) {
        required_metadata <- required_elements
      }
    }

    # If not found in nda_structure, try to fetch from API directly
    if (is.null(required_metadata)) {
      message("Fetching current ndar_subject01 required elements from NDA API...")
      nda_base_url <- "https://nda.nih.gov/api/datadictionary/v2"
      url <- sprintf("%s/datastructure/ndar_subject01", nda_base_url)

      response <- httr::GET(url, httr::timeout(10))
      if (httr::status_code(response) == 200) {
        raw_content <- rawToChar(response$content)
        if (nchar(raw_content) > 0) {
          subject_structure <- jsonlite::fromJSON(raw_content)
          if ("dataElements" %in% names(subject_structure)) {
            required_metadata <- subject_structure$dataElements[subject_structure$dataElements$required == "Required", ]
          }
        }
      }
    }

    # Extract field names from metadata
    if (!is.null(required_metadata) && nrow(required_metadata) > 0) {
      field_names <- required_metadata$name
      # Exclude internal fields that should never appear
      field_names <- setdiff(field_names, excluded_from_change)
      field_names
    } else {
      # Fallback to essential fields if API fails
      message("Using fallback essential fields list")
      setdiff(c("src_subject_id", "subjectkey", "sex", "interview_age", "interview_date",
        "phenotype", "site", "race", "handedness", "visit", "week", "redcap_event_name"), excluded_from_change)
    }
  }, error = function(e) {
    message("Error fetching ndar_subject01 elements: ", e$message)
    message("Using fallback essential fields list")
    setdiff(c("src_subject_id", "subjectkey", "sex", "interview_age", "interview_date",
      "phenotype", "site", "race", "handedness", "visit", "week", "redcap_event_name"), excluded_from_change)
  })

  # Convert to character vector if needed
  if (is.data.frame(selected_columns)) {
    selected_columns <- names(selected_columns)
  } else if (is.list(selected_columns) && all(sapply(selected_columns, is.character))) {
    selected_columns <- unlist(selected_columns)
  }
  
  # Exclude REDCap completion fields (e.g., "impact_demo_complete")
  # These are REDCap-specific and should not be included in NDA data definitions
  selected_columns <- selected_columns[!grepl("_complete$", selected_columns)]

  # Store original selected_columns before any modifications (for filtering logic later)
  original_selected_columns <- selected_columns

  # Check for required fields from ndar_subject01 that exist in the structure
  # Super-required fields (always included): subjectkey, src_subject_id, interview_date, interview_age, sex
  super_required_fields <- SUPER_REQUIRED_FIELDS
  
  # Get structure field names once for reuse
  structure_field_names <- character(0)
  if (!is.null(nda_structure) && "dataElements" %in% names(nda_structure)) {
    structure_field_names <- if (is.data.frame(nda_structure$dataElements)) {
      nda_structure$dataElements$name
    } else {
      character(0)
    }
    
    # Automatically include super-required fields if they exist in the structure and are not already selected
    missing_super_required <- setdiff(
      intersect(super_required_fields, structure_field_names),
      selected_columns
    )
    if (length(missing_super_required) > 0) {
      selected_columns <- c(selected_columns, missing_super_required)
      if (interactive_mode) {
        message(sprintf("\nAutomatically including super-required fields: %s", 
                       paste(missing_super_required, collapse = ", ")))
      }
    }
    
    # Get required fields from ndar_subject01 that exist in the current structure
    ndar_required_in_structure <- character(0)
    tryCatch({
      nda_base_url <- "https://nda.nih.gov/api/datadictionary/v2"
      url <- sprintf("%s/datastructure/ndar_subject01", nda_base_url)
      response <- httr::GET(url, httr::timeout(10))
      if (httr::status_code(response) == 200) {
        raw_content <- rawToChar(response$content)
        if (nchar(raw_content) > 0) {
          subject_structure <- jsonlite::fromJSON(raw_content)
          if ("dataElements" %in% names(subject_structure)) {
            ndar_required <- subject_structure$dataElements[
              subject_structure$dataElements$required == "Required", 
            ]
            if (nrow(ndar_required) > 0) {
              ndar_required_names <- ndar_required$name
              # Get required fields that exist in structure but are not super-required
              ndar_required_in_structure <- setdiff(
                intersect(ndar_required_names, structure_field_names),
                super_required_fields
              )
            }
          }
        }
      }
    }, error = function(e) {
      # Silently fail if API call fails
    })
  } else {
    ndar_required_in_structure <- character(0)
  }
  
  # Silent auto-add — required fields are always included
  # SKIP if skip_prompts=TRUE (when called from create_nda_files with pre-selected fields)
  if (!skip_prompts && length(ndar_required_in_structure) > 0) {
    missing_required_in_structure <- setdiff(ndar_required_in_structure, selected_columns)

    if (length(missing_required_in_structure) > 0) {
      message(sprintf("\nAdding %d NDA required field(s) to selected columns: %s",
                      length(missing_required_in_structure),
                      paste(missing_required_in_structure, collapse = ", ")))
      selected_columns <- c(selected_columns, missing_required_in_structure)
    }
  }
  
  # Filter selected_columns to only include fields that exist in the current structure
  # This ensures we don't include required fields from ndar_subject01 that aren't in this structure
  # ALWAYS filter: only include fields that exist in structure (except super-required)
  # This applies to both existing and new/modified structures
  if (length(structure_field_names) > 0 && 
      !is.null(nda_structure) && 
      "dataElements" %in% names(nda_structure) &&
      (is.data.frame(nda_structure$dataElements) && nrow(nda_structure$dataElements) > 0)) {
    
    # Check if this is a new/modified structure (has fields not in structure in original selection)
    has_new_or_modified_fields <- any(!original_selected_columns %in% structure_field_names)
    
    # Get list of required fields from ndar_subject01 (for filtering logic)
    ndar_required_all <- character(0)
    tryCatch({
      nda_base_url <- "https://nda.nih.gov/api/datadictionary/v2"
      url <- sprintf("%s/datastructure/ndar_subject01", nda_base_url)
      response <- httr::GET(url, httr::timeout(10))
      if (httr::status_code(response) == 200) {
        raw_content <- rawToChar(response$content)
        if (nchar(raw_content) > 0) {
          subject_structure <- jsonlite::fromJSON(raw_content)
          if ("dataElements" %in% names(subject_structure)) {
            ndar_required <- subject_structure$dataElements[subject_structure$dataElements$required == "Required", ]
            if (nrow(ndar_required) > 0) {
              ndar_required_all <- ndar_required$name
            }
          }
        }
      }
    }, error = function(e) {
      # Silently fail
    })
    
    # Track which fields were added via the prompt (for filtering logic)
    fields_added_by_prompt <- setdiff(selected_columns, original_selected_columns)
    
    # Always keep all original fields (they're in the user's data)
    fields_to_keep <- original_selected_columns
    
    # Add super-required fields (even if not in structure)
    super_required_in_selected <- intersect(selected_columns, super_required_fields)
    fields_to_keep <- unique(c(fields_to_keep, super_required_in_selected))
    
    # For fields added by prompt: only keep them if they exist in structure (or are super-required)
    if (length(fields_added_by_prompt) > 0) {
      prompt_fields_in_structure <- intersect(fields_added_by_prompt, structure_field_names)
      prompt_fields_super_required <- intersect(fields_added_by_prompt, super_required_fields)
      fields_to_keep <- unique(c(fields_to_keep, prompt_fields_in_structure, prompt_fields_super_required))
    }
    
    fields_to_remove <- setdiff(selected_columns, fields_to_keep)
    
    if (length(fields_to_remove) > 0) {
      selected_columns <- fields_to_keep
      if (interactive_mode) {
        message(sprintf("Removed %d field(s) not in current structure: %s", 
                       length(fields_to_remove), paste(fields_to_remove, collapse = ", ")))
      }
    }
  }

  # Extract NDA data elements
  nda_elements <- NULL
  if ("dataElements" %in% names(nda_structure)) {
    nda_elements <- nda_structure$dataElements
  } else if ("fields" %in% names(nda_structure)) {
    nda_elements <- nda_structure$fields
  } else {
    warning("Could not find dataElements or fields in nda_structure")
    nda_elements <- list()
  }

  # Create lookup map for NDA elements
  nda_lookup <- list()
  if (is.data.frame(nda_elements)) {
    # If it's a dataframe, create lookup by name
    for (i in seq_len(nrow(nda_elements))) {
      element_name <- nda_elements$name[i]
      if (!is.na(element_name)) {
        nda_lookup[[element_name]] <- as.list(nda_elements[i, ])
      }
    }
  } else if (is.list(nda_elements)) {
    # If it's already a list, use as is or extract names
    for (element in nda_elements) {
      if (is.list(element) && "name" %in% names(element)) {
        nda_lookup[[element$name]] <- element
      }
    }
  }

  # Helper function to compute metadata from actual data
  compute_field_metadata <- function(field_name, data_vector) {
    if (is.null(data_vector) || length(data_vector) == 0) {
      return(list(
        data_type = "String",
        description = "",
        size = 255,
        valueRange = "",
        required = "Recommended",
        notes = "",
        aliases = ""
      ))
    }

    # Remove NAs for analysis
    clean_data <- data_vector[!is.na(data_vector)]
    total_count <- length(data_vector)
    valid_count <- length(clean_data)
    missing_count <- total_count - valid_count

    if (valid_count == 0) {
      return(list(
        data_type = "String",
        description = "",
        size = 255,
        valueRange = "",
        required = "Recommended",
        notes = "",
        aliases = "",
        missing_info = list(
          missing_count = missing_count,
          missing_percentage = 100.0,
          total_count = total_count
        )
      ))
    }

    # Determine data type based on R class and content
    data_type <- "String"  # default
    size <- 255  # default for strings

    # Check actual R class first
    if (is.integer(data_vector) || (is.numeric(data_vector) && all(clean_data == round(clean_data), na.rm = TRUE))) {
      data_type <- "Integer"
      # For integers, size is based on the maximum number of digits
      max_val <- max(abs(clean_data), na.rm = TRUE)
      size <- max(nchar(as.character(max_val)), 1)
    } else if (is.numeric(data_vector) || is.double(data_vector)) {
      data_type <- "Float"
      # For floats, size includes decimal places
      max_chars <- max(nchar(as.character(clean_data)), na.rm = TRUE)
      size <- min(max_chars, 50)  # Cap at reasonable size
    } else if (is.logical(data_vector)) {
      data_type <- "Integer"  # NDA uses 0/1 for booleans
      size <- 1
    } else if (is.character(data_vector) || is.factor(data_vector)) {
      data_type <- "String"
      # For strings, size is the maximum character length
      char_lengths <- nchar(as.character(clean_data))
      size <- max(char_lengths, na.rm = TRUE)
      if (is.na(size) || size < 1) size <- 255
    } else {
      # Fallback for other types
      data_type <- "String"
      size <- max(nchar(as.character(clean_data)), na.rm = TRUE)
      if (is.na(size) || size < 1) size <- 255
    }

    # Compute value range based on data type
    unique_vals <- unique(clean_data)
    unique_count <- length(unique_vals)
    value_range <- ""

    # Detect user-defined missing value codes from config and data
    user_defined_codes <- character(0)
    if (data_type %in% c("Integer", "Float") && !is.null(missing_data_codes)) {
      # Check for missing data codes defined in config
      for (category in names(missing_data_codes)) {
        if (category != "required" && category != "types" && category != "aliases" && category != "allow_custom") {
          category_values <- missing_data_codes[[category]]
          if (is.character(category_values) || is.numeric(category_values)) {
            for (code in category_values) {
              if (as.numeric(code) %in% data_vector) {
                user_defined_codes <- c(user_defined_codes, as.character(code))
              }
            }
          }
        }
      }
    }

    # Use the enhanced inference function to determine value range
    value_range <- infer_value_range_from_data(data_vector, data_type)

    # Append user-defined missing value codes if they exist
    if (length(user_defined_codes) > 0) {
      if (value_range == "") {
        value_range <- paste(user_defined_codes, collapse = ";")
      } else {
        value_range <- paste0(value_range, ";", paste(user_defined_codes, collapse = ";"))
      }
    }

    # Generate smart description based on field name patterns and data
    description <- generate_field_description(field_name, clean_data, data_type)

    # All computed fields are marked as "Recommended" (requirement 3)
    required <- "Recommended"

    # Keep notes minimal for all field types
    notes_parts <- c()

    # Calculate missing percentage for missing_info (still needed)
    missing_pct <- round((missing_count / total_count) * 100, 1)

    # Document user-defined missing value codes if they exist
    if (length(user_defined_codes) > 0 && !is.null(missing_data_codes)) {
      code_descriptions <- character(0)
      for (code in user_defined_codes) {
        # Find which category this code belongs to
        for (category in names(missing_data_codes)) {
          if (category != "required" && category != "types" && category != "aliases" && category != "allow_custom") {
            category_values <- missing_data_codes[[category]]
            if (as.numeric(code) %in% category_values) {
              code_descriptions <- c(code_descriptions, paste0(code, " = ", category))
              break
            }
          }
        }
      }
      if (length(code_descriptions) > 0) {
        notes_parts <- c(notes_parts, paste0("User-defined codes: ", paste(code_descriptions, collapse = ", ")))
      }
    }

    # Do not append Mean/SD or other statistical summaries to notes

    notes <- paste(notes_parts, collapse = "; ")

    return(list(
      data_type = data_type,
      description = description,
      size = size,
      valueRange = value_range,
      required = required,
      aliases = "",
      notes = notes,
      missing_info = list(
        missing_count = missing_count,
        missing_percentage = missing_pct,
        total_count = total_count
      )
    ))
  }

  # Enhanced helper function to generate smart descriptions
  generate_field_description <- function(field_name, data_vector, data_type) {
    # Pattern matching for common field types
    name_lower <- tolower(field_name)

    # Handle specific patterns first
    if (grepl("^(src_)?subject", name_lower)) {
      return("Subject identifier")
    } else if (grepl("subjectkey|guid", name_lower)) {
      return("Global unique identifier")
    } else if (grepl("interview_age|age", name_lower)) {
      return("Age at interview (in months)")
    } else if (grepl("interview_date|date", name_lower)) {
      return("Interview date")
    } else if (grepl("^sex$|gender", name_lower)) {
      return("Biological sex")
    } else if (grepl("handedness", name_lower)) {
      return("Handedness preference")
    } else if (grepl("arm|group|condition", name_lower)) {
      return("Study arm or condition assignment")
    }

    # Pattern matching for survey/task items
    if (grepl("attention|check", name_lower)) {
      return("Attention check or validation item")
    } else if (grepl("trial|item_\\d+", name_lower)) {
      return("Trial or item response")
    } else if (grepl("response|answer", name_lower)) {
      return("Response data")
    } else if (grepl("score|total", name_lower)) {
      if (data_type %in% c("Integer", "Float")) {
        min_val <- min(data_vector, na.rm = TRUE)
        max_val <- max(data_vector, na.rm = TRUE)
        return(paste0("Score ranging from ", min_val, " to ", max_val))
      } else {
        return("Score or total measurement")
      }
    } else if (grepl("time|duration|rt|latency", name_lower)) {
      if (data_type %in% c("Integer", "Float")) {
        return("Response time or duration (likely in milliseconds)")
      } else {
        return("Time measurement")
      }
    }

    # Task-specific patterns (based on your rgpts fields)
    if (grepl("rgpts|gpts", name_lower)) {
      if (grepl("_a#?\\d+", name_lower)) {
        return("RGPTS Scale A item response")
      } else if (grepl("_b#?\\d+", name_lower)) {
        return("RGPTS Scale B item response")
      } else {
        return("RGPTS questionnaire response")
      }
    }

    # Generic patterns based on data characteristics
    if (data_type == "Integer") {
      unique_count <- length(unique(data_vector[!is.na(data_vector)]))
      min_val <- min(data_vector, na.rm = TRUE)
      max_val <- max(data_vector, na.rm = TRUE)

      if (unique_count <= 10) {
        return(paste0("Numeric scale (", unique_count, " levels: ", min_val, "-", max_val, ")"))
      } else {
        return(paste0("Numeric measurement (range: ", min_val, "-", max_val, ")"))
      }
    } else if (data_type == "Float") {
      # For Float fields, don't generate generic "Numeric measurement" placeholder
      # Let other parts of the code handle meaningful descriptions
      return("")
    } else {
      unique_count <- length(unique(data_vector[!is.na(data_vector)]))
      if (unique_count <= 5) {
        return(paste0("Categorical variable (", unique_count, " categories)"))
      } else if (unique_count <= 20) {
        return(paste0("Categorical response (", unique_count, " possible values)"))
      } else {
        return(paste0("Text field (", unique_count, " unique values)"))
      }
    }
  }

  # Map REDCap field_type/validation to NDA data type
  rc_map_redcap_to_nda_type <- function(meta_row, current_type = NULL) {
    if (is.null(meta_row) || nrow(meta_row) == 0) return(current_type %||% "String")
    ftype <- as.character(meta_row$field_type %||% "")
    vtype <- as.character(meta_row$text_validation_type_or_show_slider_number %||% "")

    # Choice-based
    if (ftype %in% c("radio", "dropdown", "checkbox")) return("Integer")

    # Numeric validations
    if (nzchar(vtype)) {
      if (grepl("^integer$", vtype, ignore.case = TRUE)) return("Integer")
      if (grepl("^(number|float|decimal)$", vtype, ignore.case = TRUE)) return("Float")
      if (grepl("^slider|show_slider_number", vtype, ignore.case = TRUE)) return("Integer")
      if (grepl("date|datetime", vtype, ignore.case = TRUE)) return("Date")
      if (grepl("zipcode|email|phone", vtype, ignore.case = TRUE)) return("String")
    }

    # Calculated fields numeric
    if (ftype %in% c("calc")) return("Float")

    # Text default
    if (ftype %in% c("text", "notes")) return(current_type %||% "String")

    return(current_type %||% "String")
  }

  # For modified data definitions, focus ONLY on changes:
  # 1. New fields (not in NDA)
  # 2. Modified fields (in NDA but changed)
  # Exclude unchanged NDA fields
  # IMPORTANT: ndar_subject01 variables can NEVER be modified

  all_nda_fields <- names(nda_lookup)
  user_added_fields <- selected_columns[!selected_columns %in% all_nda_fields]

  # Identify ndar_subject01 variables that should never be modified
  # These are the 5 super required fields that must be present in all NDA submissions
  ndar_subject01_fields <- SUPER_REQUIRED_FIELDS

  # Initialize lists for different field categories
  new_fields <- character(0)
  modified_fields <- character(0)
  unchanged_fields <- character(0)

  if (!is.null(data_frame)) {
    # Categorize fields based on whether they exist in data and NDA
    for (field in selected_columns) {
      # Skip change categorization for excluded fields: force unchanged
      if (field %in% excluded_from_change) {
        if (field %in% names(data_frame) && field %in% all_nda_fields) {
          unchanged_fields <- c(unchanged_fields, field)
        }
        next
      }
      if (field %in% names(data_frame)) {
        if (field %in% all_nda_fields) {
          # Field exists in both data and NDA - check if it's modified
          if (field %in% ndar_subject01_fields) {
            # ndar_subject01 fields can never be modified
            unchanged_fields <- c(unchanged_fields, field)
          } else {
            # Check if this field has actual modifications
            is_modified <- FALSE

            # Check for missing value code modifications
            if (!is.null(missing_data_codes)) {
              field_data <- data_frame[[field]]
              field_type <- if ("type" %in% names(nda_lookup[[field]])) nda_lookup[[field]]$type else "String"

              if (field_type %in% c("Integer", "Float")) {
                # Check if field has missing value codes not in original NDA range
                original_value_range <- if ("valueRange" %in% names(nda_lookup[[field]])) nda_lookup[[field]]$valueRange else ""

                # Detect missing value codes in the data
                user_defined_codes <- character(0)
                for (category in names(missing_data_codes)) {
                  if (category != "required" && category != "types" && category != "aliases" && category != "allow_custom") {
                    category_values <- missing_data_codes[[category]]
                    if (is.character(category_values) || is.numeric(category_values)) {
                      for (code in category_values) {
                        if (as.numeric(code) %in% field_data) {
                          user_defined_codes <- c(user_defined_codes, as.character(code))
                        }
                      }
                    }
                  }
                }

                # Check if any of these codes are not in the original value range
                if (length(user_defined_codes) > 0) {
                  original_codes <- character(0)
                  if (original_value_range != "") {
                    # Parse original value range to extract individual codes
                    range_parts <- trimws(strsplit(original_value_range, ";")[[1]])
                    for (part in range_parts) {
                      if (!grepl("::", part)) {  # Skip range notation
                        original_codes <- c(original_codes, part)
                      }
                    }
                  }

                  # Find new codes that aren't in the original
                  new_codes <- setdiff(user_defined_codes, original_codes)
                  if (length(new_codes) > 0) {
                    is_modified <- TRUE
                  }
                }
              }
            }

            if (is_modified) {
              modified_fields <- c(modified_fields, field)
            } else {
              unchanged_fields <- c(unchanged_fields, field)
            }
          }
        } else {
          # Field exists in data but not in NDA - it's new
          if (!(field %in% excluded_from_change)) {
            new_fields <- c(new_fields, field)
          } else {
            unchanged_fields <- c(unchanged_fields, field)
          }
        }
      } else {
        # Field doesn't exist in data - skip it
        next
      }
    }
  } else {
    # If no data frame available, treat all selected columns as new
    new_fields <- setdiff(selected_columns, excluded_from_change)
  }

  # Order fields: modified first, then unchanged NDA fields, then new fields at bottom
  # This puts new elements (not in NDA) at the end of the Excel file
  unchanged_selected <- intersect(unchanged_fields, selected_columns)
  
  ordered_columns <- c(modified_fields, unchanged_selected, new_fields)
  
  # Ensure we don't drop any selected columns: append remaining ones and de-duplicate
  remaining_columns <- setdiff(selected_columns, ordered_columns)
  if (length(remaining_columns) > 0) {
    ordered_columns <- c(ordered_columns, remaining_columns)
  }
  ordered_columns <- unique(ordered_columns)

  # Report what's being included in the data definition
  if (verbose) {
    if (length(new_fields) > 0) {
      message(sprintf("Including %d new fields in data definition: %s",
                      length(new_fields), paste(head(new_fields, 5), collapse = ", ")))
    }
    if (length(modified_fields) > 0) {
      message(sprintf("Including %d modified fields in data definition: %s",
                      length(modified_fields), paste(head(modified_fields, 5), collapse = ", ")))
    }
    if (length(unchanged_selected) > 0) {
      message(sprintf("Including %d unchanged NDA fields in data definition: %s",
                      length(unchanged_selected), paste(head(unchanged_selected, 5), collapse = ", ")))
    }
    if (length(unchanged_fields) > 0) {
      message(sprintf("Excluding %d unchanged NDA fields from data definition: %s",
                      length(unchanged_fields), paste(head(unchanged_fields, 5), collapse = ", ")))
    }
  }

  # Initialize data definition structure
  data_definition <- list(
    measure_name = measure_name,
    created_at = Sys.time(),
    total_selected_fields = length(selected_columns),
    fields = list(),
    metadata = list(
      source_template = if ("metadata" %in% names(submission_template)) submission_template$metadata else NULL,
      nda_structure_info = list(
        total_elements = length(nda_lookup),
        structure_name = if ("shortName" %in% names(nda_structure)) nda_structure$shortName else measure_name
      ),
      validation_summary = list(
        matched_fields = 0,
        unmatched_fields = 0,
        computed_fields = 0,
        modified_fields = 0,
        warnings = character(0)
      )
    )
  )

  # Persist NDA element names, verbose flag, ndar_subject additions, structure type, and REDCap choices for downstream export/formatting
  data_definition$metadata$nda_element_names <- names(nda_lookup)
  data_definition$metadata$verbose <- verbose
  data_definition$metadata$ndar_subject_additions <- ndar_subject_additions
  data_definition$metadata$is_new_structure <- validation_state$is_new_structure
  data_definition$metadata$redcap_choices_map <- redcap_choices_map
  data_definition$metadata$ndar_subject01_all_fields <- validation_state$ndar_subject01_all_fields %||% character(0)

  # Process each selected column in the new order
  for (i in seq_along(ordered_columns)) {
    column_name <- ordered_columns[i]
    # Skip internal/excluded fields unconditionally
    if (column_name %in% excluded_from_change) next
    # _complete fields already removed by StandardOutput in ndaRequest.R
    
    # Check if column exists in NDA structure
    if (column_name %in% names(nda_lookup)) {
      # Field found in NDA structure
      nda_field <- nda_lookup[[column_name]]

      # Get missing value information for existing fields
      missing_info <- NULL
      field_data <- NULL
      field_exists_in_data <- FALSE

      if (!is.null(data_frame) && column_name %in% names(data_frame)) {
        field_data <- data_frame[[column_name]]
        field_exists_in_data <- TRUE
        total_count <- length(field_data)
        missing_count <- sum(is.na(field_data))
        missing_percentage <- round((missing_count / total_count) * 100, 1)

        missing_info <- list(
          missing_count = missing_count,
          missing_percentage = missing_percentage,
          total_count = total_count
        )
      }

      # Check if this existing field has missing value codes that aren't in the original NDA value range
      is_modified_structure <- FALSE
      updated_value_range <- NULL
      modification_notes <- NULL
      
      # ENHANCEMENT: Allow REDCap choices to extend NDA value ranges
      # This handles cases like gender_identity where:
      #   - NDA defines 1::12
      #   - REDCap defines 1::14 (includes additional options like "Two-Spirit")
      # We want to use the WIDER range (REDCap) while keeping NDA metadata
      # This ensures consistency regardless of dcc parameter
      if (field_exists_in_data && 
          !is.null(redcap_choices_map) && 
          column_name %in% names(redcap_choices_map)) {
        
        redcap_choices <- redcap_choices_map[[column_name]]
        if (!is.null(redcap_choices) && nzchar(redcap_choices)) {
          # Parse REDCap choices to extract codes
          parsed_redcap <- rc_parse_choices_codes_and_notes(redcap_choices)
          
          if (parsed_redcap$codes != "") {
            # Get NDA value range
            nda_value_range <- if ("valueRange" %in% names(nda_field)) {
              nda_field$valueRange
            } else {
              ""
            }
            
            # Get field type for appropriate handling
            field_type <- if ("type" %in% names(nda_field)) nda_field$type else "String"
            
            # Only extend for Integer and String types (not Float - floats are floats)
            if (field_type %in% c("Integer", "String")) {
              # Expand both ranges to numeric vectors for comparison (Integer only)
              if (field_type == "Integer") {
                nda_vals <- expand_numeric_value_range(nda_value_range)
                redcap_vals <- expand_numeric_value_range(parsed_redcap$codes)
                
                # If REDCap has values not in NDA, extend the range (never shrink)
                if (!is.null(nda_vals) && !is.null(redcap_vals) && length(nda_vals) > 0) {
                  new_vals <- setdiff(redcap_vals, nda_vals)
                  
                  if (length(new_vals) > 0) {
                    # REDCap has additional codes - extend the range
                    all_vals <- sort(unique(c(nda_vals, redcap_vals)))
                    
                    # Check if sequential for range notation (e.g., 1::14)
                    # Otherwise use semicolon notation (e.g., 1::7;999)
                    is_sequential <- length(all_vals) > 1 && all(diff(all_vals) == 1)
                    if (is_sequential) {
                      updated_value_range <- paste0(min(all_vals), "::", max(all_vals))
                    } else {
                      # Build range notation with gaps
                      range_parts <- character(0)
                      i <- 1
                      while (i <= length(all_vals)) {
                        start_val <- all_vals[i]
                        end_val <- start_val
                        
                        # Find consecutive sequence
                        while (i < length(all_vals) && all_vals[i + 1] == all_vals[i] + 1) {
                          i <- i + 1
                          end_val <- all_vals[i]
                        }
                        
                        # Format as range or single value
                        if (end_val - start_val >= 2) {
                          range_parts <- c(range_parts, paste0(start_val, "::", end_val))
                        } else if (end_val == start_val) {
                          range_parts <- c(range_parts, as.character(start_val))
                        } else {
                          # Two consecutive values, list them separately
                          range_parts <- c(range_parts, as.character(start_val), as.character(end_val))
                        }
                        i <- i + 1
                      }
                      updated_value_range <- paste(range_parts, collapse = ";")
                    }
                    
                    # Mark as modified structure (non-super-required fields only)
                    if (!(column_name %in% ndar_subject01_fields)) {
                      is_modified_structure <- TRUE
                      modification_notes <- sprintf(
                        "Extended value range from NDA (%s) to include REDCap codes: added %s",
                        nda_value_range,
                        paste(new_vals, collapse = ", ")
                      )
                    }
                    
                    if (verbose) {
                      message(sprintf(
                        "Field '%s': Extended NDA range %s to %s based on REDCap choices",
                        column_name, nda_value_range, updated_value_range
                      ))
                    }
                  }
                }
              } else if (field_type == "String") {
                # For String types, use REDCap choices if NDA range is empty or less comprehensive
                if (nda_value_range == "" || nchar(parsed_redcap$codes) > nchar(nda_value_range)) {
                  updated_value_range <- parsed_redcap$codes
                  if (!(column_name %in% ndar_subject01_fields)) {
                    is_modified_structure <- TRUE
                    modification_notes <- "Extended value range with REDCap choices"
                  }
                  if (verbose) {
                    message(sprintf(
                      "Field '%s': Using REDCap choices for String field",
                      column_name
                    ))
                  }
                }
              }
            }
          }
        }
      }

      # IMPORTANT: ndar_subject01 variables can NEVER be modified
      if (column_name %in% ndar_subject01_fields) {
        # Force ndar_subject01 fields to remain unchanged
        is_modified_structure <- FALSE
      } else if (field_exists_in_data && !is.null(missing_data_codes)) {
        original_value_range <- if ("valueRange" %in% names(nda_field)) nda_field$valueRange else ""
        field_type <- if ("type" %in% names(nda_field)) nda_field$type else "String"

        # Only check numeric fields for missing value codes
        if (field_type %in% c("Integer", "Float")) {
          # Detect missing value codes in the data
          user_defined_codes <- character(0)
          for (category in names(missing_data_codes)) {
            if (category != "required" && category != "types" && category != "aliases" && category != "allow_custom") {
              category_values <- missing_data_codes[[category]]
              if (is.character(category_values) || is.numeric(category_values)) {
                for (code in category_values) {
                  if (as.numeric(code) %in% field_data) {
                    user_defined_codes <- c(user_defined_codes, as.character(code))
                  }
                }
              }
            }
          }

          # Check if any of these codes are not in the original value range
          if (length(user_defined_codes) > 0) {
            original_codes <- character(0)
            if (original_value_range != "") {
              # Parse original value range to extract individual codes
              range_parts <- trimws(strsplit(original_value_range, ";")[[1]])
              for (part in range_parts) {
                if (!grepl("::", part)) {  # Skip range notation
                  original_codes <- c(original_codes, part)
                }
              }
            }

            # Find new codes that aren't in the original
            new_codes <- setdiff(user_defined_codes, original_codes)
            if (length(new_codes) > 0) {
              is_modified_structure <- TRUE

              # Create updated value range
              if (original_value_range == "") {
                updated_value_range <- paste(new_codes, collapse = ";")
              } else {
                updated_value_range <- paste0(original_value_range, ";", paste(new_codes, collapse = ";"))
              }

              # Create modification notes
              code_descriptions <- character(0)
              for (code in new_codes) {
                for (category in names(missing_data_codes)) {
                  if (category != "required" && category != "types" && category != "aliases" && category != "allow_custom") {
                    category_values <- missing_data_codes[[category]]
                    if (as.numeric(code) %in% category_values) {
                      code_descriptions <- c(code_descriptions, paste0(code, " = ", category))
                      break
                    }
                  }
                }
              }
              modification_notes <- paste0("Modified structure: Added missing value codes: ", paste(code_descriptions, collapse = ", "))
            }
          }
        }
      }

      # Determine source and metadata based on whether structure was modified
      if (is_modified_structure) {
        source <- "nda_modified"
        # Use updated value range
        validation_rules <- list(
          min_value = if ("minimum" %in% names(nda_field)) nda_field$minimum else NULL,
          max_value = if ("maximum" %in% names(nda_field)) nda_field$maximum else NULL,
          allowed_values = updated_value_range,
          pattern = if ("pattern" %in% names(nda_field)) nda_field$pattern else NULL
        )

        # Update NDA metadata with modified value range
        modified_nda_metadata <- nda_field
        modified_nda_metadata$valueRange <- updated_value_range
              # Do not append internal modification notes to exportable metadata

        nda_metadata_to_use <- modified_nda_metadata
        data_definition$metadata$validation_summary$modified_fields <-
          (data_definition$metadata$validation_summary$modified_fields %||% 0) + 1
      } else {
        source <- "nda_validated"
        # Use original validation rules and metadata
        validation_rules <- list(
          min_value = if ("minimum" %in% names(nda_field)) nda_field$minimum else NULL,
          max_value = if ("maximum" %in% names(nda_field)) nda_field$maximum else NULL,
          allowed_values = if ("valueRange" %in% names(nda_field)) nda_field$valueRange else NULL,
          pattern = if ("pattern" %in% names(nda_field)) nda_field$pattern else NULL
        )
        nda_metadata_to_use <- nda_field
        data_definition$metadata$validation_summary$matched_fields <-
          data_definition$metadata$validation_summary$matched_fields + 1
      }

      # Determine final description with fallback hierarchy:
      # 1) NDA description (if present and non-empty)
      # 2) REDCap field_label (if available) - ONLY for non-ndar_subject01 fields
      # 3) Qualtrics question text (if available) - ONLY for non-ndar_subject01 fields
      # 4) Keep as empty (will be handled downstream)
      nda_desc <- if ("description" %in% names(nda_field)) nda_field$description else ""
      fallback_desc <- nda_desc
      if (is.null(fallback_desc) || identical(fallback_desc, "") || is.na(fallback_desc)) {
        # Only apply external fallbacks for fields NOT matched to NDA
        if (!(column_name %in% names(nda_lookup))) {
          if (!is.null(redcap_label_map) && column_name %in% names(redcap_label_map)) {
            fallback_desc <- as.character(redcap_label_map[[column_name]])
          } else if (!is.null(qualtrics_label_map) && column_name %in% names(qualtrics_label_map)) {
            fallback_desc <- as.character(qualtrics_label_map[[column_name]])
          }
          if (!is.null(fallback_desc) && fallback_desc != "") {
            fallback_desc <- rc_clean_label(fallback_desc)
          }
          # If we found a fallback, also enrich nda_metadata to carry it through export
          if (!identical(fallback_desc, "")) {
            nda_metadata_to_use$description <- fallback_desc
          }
        } else {
          fallback_desc <- ""
        }
      }

      # Enrich ValueRange/Notes from REDCap choices ONLY for non-NDA fields
      if (!(column_name %in% names(nda_lookup)) && !is.null(redcap_choices_map) && column_name %in% names(redcap_choices_map)) {
        redcap_choices <- redcap_choices_map[[column_name]]
        if (!is.null(redcap_choices) && !is.na(redcap_choices) && redcap_choices != "") {
          parsed <- rc_parse_choices_codes_and_notes(redcap_choices)
          if (parsed$codes != "") {
            current_value_range <- if ("valueRange" %in% names(nda_metadata_to_use)) nda_metadata_to_use$valueRange else ""
            if (is.null(current_value_range) || current_value_range == "" || is.na(current_value_range)) {
              nda_metadata_to_use$valueRange <- parsed$codes
              validation_rules$allowed_values <- parsed$codes
            }
          }
          # Append mapping into notes (preserve existing)
          if (parsed$notes != "") {
            existing_notes <- if ("notes" %in% names(nda_metadata_to_use)) (nda_metadata_to_use$notes %||% "") else ""
            nda_metadata_to_use$notes <- paste0(existing_notes, ifelse(existing_notes != "", " | ", ""), parsed$notes)
          }
        }
      }

      # If still no ValueRange, attempt to derive from validation/type where possible using REDCap metadata row
      if (!(column_name %in% names(nda_lookup)) && (is.null(nda_metadata_to_use$valueRange) || nda_metadata_to_use$valueRange == "" || is.na(nda_metadata_to_use$valueRange))) {
        if (exists("rc_meta") && is.data.frame(rc_meta) && "field_name" %in% names(rc_meta)) {
          rc_row <- rc_meta[rc_meta$field_name == column_name, , drop = FALSE]
          if (nrow(rc_row) == 1) {
            derived <- rc_derive_value_range_from_validation(rc_row)
            if (derived != "") {
              nda_metadata_to_use$valueRange <- derived
              validation_rules$allowed_values <- derived
            }
          }
        }
      }

      # Create NdaDataStructure from NDA metadata
      # Pass all metadata during construction to enable proper R6 object conversion
      field_struct <- NdaDataStructure$new(
        element_name = column_name,
        data_type = nda_metadata_to_use$type %||% "String",
        size = nda_metadata_to_use$size,
        required = nda_metadata_to_use$required %||% "Recommended",
        element_description = fallback_desc,
        value_range = nda_metadata_to_use$valueRange %||% "",
        notes = nda_metadata_to_use$notes %||% "",
        aliases = nda_metadata_to_use$aliases %||% "",
        selection_order = i,
        source = if(is_modified_structure) "nda_modified" else source,
        missing_info = missing_info,
        validation_rules = validation_rules
      )
      
      # IMPORTANT: Super required fields must ALWAYS be marked as "Required"
      # These are the 5 mandatory fields for all NDA submissions
      if (column_name %in% SUPER_REQUIRED_FIELDS) {
        field_struct$required <- RequirementLevel$new("Required")
      }
      
      # Add modification notes if structure was modified
      if (is_modified_structure && !is.null(modification_notes) && nzchar(modification_notes)) {
        field_struct$source_metadata$add_modification(modification_notes)
      }
      
      # Store as list for backward compatibility with export code
      data_definition$fields[[column_name]] <- field_struct$to_list()

    } else {
      # Field not found in NDA structure - compute metadata from data

      # Get data for this column if data frame is available
      column_data <- NULL
      field_exists_in_data <- FALSE

      if (!is.null(data_frame) && column_name %in% names(data_frame)) {
        column_data <- data_frame[[column_name]]
        field_exists_in_data <- TRUE
      }

      # Compute metadata
      computed_metadata <- compute_field_metadata(column_name, column_data)

      # Apply external fallback descriptions to computed fields when available
      computed_desc <- computed_metadata$description
      rc_label <- NULL
      if (!is.null(redcap_label_map) && column_name %in% names(redcap_label_map)) {
        rc_label <- as.character(redcap_label_map[[column_name]])
        rc_label <- rc_clean_label(rc_label)
      }
      if (!is.null(rc_label) && nzchar(rc_label)) {
        computed_desc <- rc_label
      } else if (is.null(computed_desc) || identical(computed_desc, "") || is.na(computed_desc)) {
        if (!is.null(qualtrics_label_map) && column_name %in% names(qualtrics_label_map)) {
          computed_desc <- as.character(qualtrics_label_map[[column_name]])
        }
      }

      # REDCap checkbox child handling: derive from parent metadata first
      if (exists("rc_meta") && is.data.frame(rc_meta)) {
        cbx <- rc_enrich_checkbox_child(column_name, rc_meta)
        if (!is.null(cbx$description) && cbx$description != "") {
          computed_desc <- cbx$description
        }
        if (!is.null(cbx$valueRange)) {
          computed_metadata$valueRange <- cbx$valueRange
        }
        if (!is.null(cbx$notes) && cbx$notes != "") {
          existing_notes <- computed_metadata$notes %||% ""
          computed_metadata$notes <- paste0(existing_notes, ifelse(existing_notes != "", " | ", ""), cbx$notes)
        }
        if (!is.null(cbx$data_type) && cbx$data_type != "") {
          computed_metadata$data_type <- cbx$data_type
        }
        if (!is.null(cbx$size)) {
          computed_metadata$size <- cbx$size
        }
      }

      # Enhance computed ValueRange with RedCap choices if available
      computed_value_range <- computed_metadata$valueRange
      if ((is.null(computed_value_range) || computed_value_range == "" || is.na(computed_value_range)) &&
          !is.null(redcap_choices_map) && column_name %in% names(redcap_choices_map)) {
        redcap_choices <- redcap_choices_map[[column_name]]
        if (!is.null(redcap_choices) && !is.na(redcap_choices) && redcap_choices != "") {
          parsed <- rc_parse_choices_codes_and_notes(redcap_choices)
          if (parsed$codes != "") {
            computed_metadata$valueRange <- parsed$codes
          }
          # Add mapping to notes
          if (parsed$notes != "") {
            existing_notes <- computed_metadata$notes %||% ""
            computed_metadata$notes <- paste0(existing_notes, ifelse(existing_notes != "", " | ", ""), parsed$notes)
          }
        }
      }

      # If still no ValueRange, attempt derive from REDCap validation/type
      if ((is.null(computed_metadata$valueRange) || computed_metadata$valueRange == "" || is.na(computed_metadata$valueRange)) && exists("rc_meta") && is.data.frame(rc_meta) && "field_name" %in% names(rc_meta)) {
        rc_row <- rc_meta[rc_meta$field_name == column_name, , drop = FALSE]
        if (nrow(rc_row) == 1) {
          derived <- rc_derive_value_range_from_validation(rc_row)
          if (derived != "") {
            computed_metadata$valueRange <- derived
          }
        }
      }

      # Clean description for <br>
      if (!is.null(computed_desc) && computed_desc != "") {
        computed_desc <- rc_clean_label(computed_desc)
      }

      # Apply REDCap-to-NDA type mapping for computed fields if metadata available
      if (exists("rc_meta") && is.data.frame(rc_meta) && "field_name" %in% names(rc_meta)) {
        rc_row <- rc_meta[rc_meta$field_name == column_name, , drop = FALSE]
        if (nrow(rc_row) == 1) {
          computed_metadata$data_type <- rc_map_redcap_to_nda_type(rc_row, computed_metadata$data_type %||% "String")
        }
      }

      # Adjust warning message based on whether field exists in data
      if (field_exists_in_data) {
        # Determine enrichment source
        enrichment <- if (exists("rc_meta") && is.data.frame(rc_meta) && column_name %in% rc_meta$field_name) {
          "from REDCap"
        } else if (!is.null(qualtrics_label_map) && column_name %in% names(qualtrics_label_map)) {
          "from Qualtrics"
        } else {
          "from dataframe"
        }
        warning_msg <- paste("Column", column_name, "not found in NDA data structure - computing metadata", enrichment)
      } else {
        enrichment <- if (exists("rc_meta") && is.data.frame(rc_meta) && column_name %in% rc_meta$field_name) {
          "from REDCap"
        } else if (!is.null(qualtrics_label_map) && column_name %in% names(qualtrics_label_map)) {
          "from Qualtrics"
        } else {
          "from template"
        }
        warning_msg <- paste("Column", column_name, "not found in NDA data structure or data frame - using placeholder metadata", enrichment)
      }

      data_definition$metadata$validation_summary$warnings <-
        c(data_definition$metadata$validation_summary$warnings, warning_msg)

      # Create NdaDataStructure from computed metadata
      # Pass all metadata during construction to enable proper R6 object conversion
      field_struct <- NdaDataStructure$new(
        element_name = column_name,
        data_type = computed_metadata$data_type,
        size = computed_metadata$size,
        required = computed_metadata$required,
        element_description = computed_desc,
        value_range = computed_metadata$valueRange %||% "",
        notes = computed_metadata$notes %||% "",
        aliases = computed_metadata$aliases %||% "",
        selection_order = i,
        source = if (field_exists_in_data) "computed_from_data" else "template_only",
        missing_info = computed_metadata$missing_info,
        validation_rules = list(
          min_value = NULL,
          max_value = NULL,
          allowed_values = computed_metadata$valueRange,
          pattern = NULL
        )
      )
      
      # IMPORTANT: Super required fields must ALWAYS be marked as "Required"
      # These are the 5 mandatory fields for all NDA submissions
      if (column_name %in% SUPER_REQUIRED_FIELDS) {
        field_struct$required <- RequirementLevel$new("Required")
      }
      
      # Store as list for backward compatibility
      data_definition$fields[[column_name]] <- field_struct$to_list()

      data_definition$metadata$validation_summary$unmatched_fields <-
        data_definition$metadata$validation_summary$unmatched_fields + 1
      data_definition$metadata$validation_summary$computed_fields <-
        data_definition$metadata$validation_summary$computed_fields + 1
    }
  }

  # Add summary statistics
  data_definition$summary <- list(
    total_fields = length(selected_columns),
    matched_fields = data_definition$metadata$validation_summary$matched_fields,
    unmatched_fields = data_definition$metadata$validation_summary$unmatched_fields,
    computed_fields = data_definition$metadata$validation_summary$computed_fields,
    modified_fields = data_definition$metadata$validation_summary$modified_fields,
    match_percentage = round(
      (data_definition$metadata$validation_summary$matched_fields / length(selected_columns)) * 100,
      2
    )
  )

  # Add missing value summary across all fields
  total_missing_count <- 0
  total_rows <- 0
  fields_with_missing <- 0

  for (field_name in names(data_definition$fields)) {
    field <- data_definition$fields[[field_name]]
    if (!is.null(field$missing_info)) {
      total_missing_count <- total_missing_count + field$missing_info$missing_count
      total_rows <- max(total_rows, field$missing_info$total_count)
      if (field$missing_info$missing_count > 0) {
        fields_with_missing <- fields_with_missing + 1
      }
    }
  }

  if (total_rows > 0) {
    data_definition$summary$missing_data_summary <- list(
      total_missing_values = total_missing_count,
      total_data_points = total_rows * length(selected_columns),
      overall_missing_percentage = round((total_missing_count / (total_rows * length(selected_columns))) * 100, 1),
      fields_with_missing = fields_with_missing,
      total_fields = length(selected_columns)
    )
  }

  # Auto-export to XLSX and print success message
  tryCatch({
    exportDataDefinition(data_definition)
    # Print success message before Missing Data Summary
    message(sprintf("[OK] Data definition created at: ./tmp/%s_definitions.xlsx", measure_name))
    cat("\n")  # Blank line
  }, error = function(e) {
    warning("Data definition export failed: ", e$message, call. = FALSE)
    message("Note: This is unexpected. Please report this issue.")
    message("Full error: ", toString(e))
    message("Traceback:")
    print(traceback())
  })

  # Print missing value summary
  if (!is.null(data_definition$summary$missing_data_summary)) {
    missing_summary <- data_definition$summary$missing_data_summary
    cat("=== Missing Data Summary ===\n")
    cat("Total missing values:", missing_summary$total_missing_values, "\n")
    cat("Total data points:", missing_summary$total_data_points, "\n")
    cat("Overall missing percentage:", paste0(missing_summary$overall_missing_percentage, "%"), "\n")
    cat("Fields with missing data:", missing_summary$fields_with_missing, "/", missing_summary$total_fields, "\n")
  }

  if (verbose && length(data_definition$metadata$validation_summary$warnings) > 0) {
    cat("\nInfo:\n")
    for (warning in data_definition$metadata$validation_summary$warnings) {
      cat("  -", warning, "\n")
    }
  }

  return(data_definition)
}

#' @noRd
validateDataDefinition <- function(data_definition, strict_validation = FALSE) {

  validation_result <- list(
    is_valid = TRUE,
    errors = character(0),
    warnings = character(0),
    recommendations = character(0)
  )

  # Check required NDA fields
  required_nda_fields <- c("src_subject_id", "subjectkey", "interview_age", "interview_date", "sex")
  selected_field_names <- names(data_definition$fields)

  for (required_field in required_nda_fields) {
    if (!required_field %in% selected_field_names) {
      validation_result$errors <- c(
        validation_result$errors,
        paste("Missing required NDA field:", required_field)
      )
      validation_result$is_valid <- FALSE
    }
  }

  # Check for unmatched fields
  if (data_definition$summary$unmatched_fields > 0) {
    if (strict_validation) {
      validation_result$errors <- c(
        validation_result$errors,
        paste("Unmatched fields found:", data_definition$summary$unmatched_fields)
      )
      validation_result$is_valid <- FALSE
    } else {
      validation_result$warnings <- c(
        validation_result$warnings,
        paste("Unmatched fields found:", data_definition$summary$unmatched_fields)
      )
    }
  }

  # Check match percentage
  if (data_definition$summary$match_percentage < 80) {
    validation_result$recommendations <- c(
      validation_result$recommendations,
      paste0("Low match percentage (", data_definition$summary$match_percentage,
             "%). Consider reviewing field selections.")
    )
  }

  # Print validation results
  cat("\n=== Validation Results ===\n")
  cat("Status:", ifelse(validation_result$is_valid, "VALID", "INVALID"), "\n")

  if (length(validation_result$errors) > 0) {
    cat("\nErrors:\n")
    for (error in validation_result$errors) {
      cat("  \u2717", error, "\n")
    }
  }

  if (length(validation_result$warnings) > 0) {
    cat("\nWarnings:\n")
    for (warning in validation_result$warnings) {
      cat("  \u26a0", warning, "\n")
    }
  }

  if (length(validation_result$recommendations) > 0) {
    cat("\nRecommendations:\n")
    for (rec in validation_result$recommendations) {
      cat("  \u2139", rec, "\n")
    }
  }

  return(validation_result)
}

#' @noRd
exportDataDefinition <- function(data_definition) {
  # Extract verbose flag, ndar_subject_additions, is_new_structure, and redcap_choices_map from data_definition if they exist
  verbose <- if (!is.null(data_definition$metadata$verbose)) data_definition$metadata$verbose else FALSE
  ndar_subject_additions <- if (!is.null(data_definition$metadata$ndar_subject_additions)) {
    data_definition$metadata$ndar_subject_additions
  } else {
    character(0)
  }
  is_new_structure <- if (!is.null(data_definition$metadata$is_new_structure)) {
    data_definition$metadata$is_new_structure
  } else {
    FALSE  # Default to MODIFIED structure behavior (9 columns, no Instructions)
  }
  redcap_choices_map <- if (!is.null(data_definition$metadata$redcap_choices_map)) {
    data_definition$metadata$redcap_choices_map
  } else {
    NULL
  }
  
  # Create directory structure if it doesn't exist
  tmp_path <- file.path(".", "tmp")
  if (!dir.exists(tmp_path)) {
    dir.create(tmp_path, recursive = TRUE)
  }

  # Internal-only fields that must never be exported
  excluded_internal <- c("state", "lost_to_followup", "lost_to_follow-up", "study_status")

  # Create file path - data definitions are always Excel files
  file_path <- file.path(tmp_path, paste0(data_definition$measure_name, "_definitions.xlsx"))

  # Export data definition as Excel file (XLSX)
  # openxlsx is a hard dependency (Imports); proceed directly
  
  # Assemble field data for export
  # openxlsx is a hard dependency (Imports); proceed directly
  
  field_names <- names(data_definition$fields)
           # Ensure field_names is a character vector
           if (is.null(field_names)) {
             field_names <- character(0)
           }
           if (!is.character(field_names)) {
             field_names <- as.character(field_names)
           }
            field_names <- setdiff(field_names, excluded_internal)
             # _complete fields already removed by StandardOutput in ndaRequest.R
             if (length(field_names) == 0) {
               warning("No fields to export")
               return(invisible(NULL))
             }

             # CRITICAL FIX: Reorder fields to ensure super required fields appear FIRST
             # This fixes the bug where interview_date and interview_age appeared at the bottom
             # Order should always be: subjectkey, src_subject_id, interview_date, interview_age, sex
             # Then all other fields in their original order
             
             # FORCE super required fields to the top (ALWAYS)
             super_required_present <- SUPER_REQUIRED_FIELDS[SUPER_REQUIRED_FIELDS %in% field_names]
             other_fields <- setdiff(field_names, SUPER_REQUIRED_FIELDS)
             field_names <- c(super_required_present, other_fields)
             
             if (verbose) {
               message(sprintf("[EXCEL] Field ordering: %d super required first, then %d others",
                              length(super_required_present), length(other_fields)))
             }

             element_names <- field_names

            data_types <- sapply(field_names, function(fname) {
              tryCatch({
                x <- data_definition$fields[[fname]]
                if (!is.null(x$nda_metadata) && "type" %in% names(x$nda_metadata)) {
                  as.character(x$nda_metadata$type %||% "String")
                } else {
                  as.character(x$data_type %||% "String")
                }
              }, error = function(e) "String")
            })

           sizes <- sapply(field_names, function(fname) {
             tryCatch({
               x <- data_definition$fields[[fname]]
               if (!is.null(x$nda_metadata) && "size" %in% names(x$nda_metadata)) {
                 size_val <- x$nda_metadata$size
                 if (is.null(size_val) || is.na(size_val)) "" else as.numeric(size_val)
               } else {
                 ""
               }
             }, error = function(e) "")
           })

           required_vals <- sapply(field_names, function(fname) {
              tryCatch({
                # IMPORTANT: Super required fields must ALWAYS be "Required"
                if (fname %in% SUPER_REQUIRED_FIELDS) {
                  return("Required")
                }
                
                x <- data_definition$fields[[fname]]
                if (!is.null(x$nda_metadata) && "required" %in% names(x$nda_metadata)) {
                  as.character(x$nda_metadata$required %||% "Recommended")
                } else if (!is.null(x$required)) {
                  ifelse(isTRUE(x$required), "Required", "Recommended")
                } else {
                  "Recommended"
                }
              }, error = function(e) "Recommended")
           })

           descriptions <- sapply(field_names, function(fname) {
             tryCatch({
               x <- data_definition$fields[[fname]]
               if (!is.null(x$nda_metadata) && "description" %in% names(x$nda_metadata)) {
                desc_val <- as.character(x$nda_metadata$description %||% "")
                desc_val <- gsub("^Missing field: ", "", desc_val)
                desc_val <- gsub("^Empty field: ", "", desc_val)
                desc_val
               } else {
                 as.character(x$description %||% "")
               }
             }, error = function(e) "")
           })

           value_ranges <- sapply(field_names, function(fname) {
             tryCatch({
               x <- data_definition$fields[[fname]]
               if (!is.null(x$nda_metadata) && "valueRange" %in% names(x$nda_metadata)) {
                 val_range <- x$nda_metadata$valueRange %||% ""
                 if (is.na(val_range) || val_range == "NA") "" else as.character(val_range)
               } else {
                 ""
               }
             }, error = function(e) "")
           })
           value_ranges <- vapply(value_ranges, compress_value_range, character(1))

           notes <- sapply(field_names, function(fname) {
             tryCatch({
               x <- data_definition$fields[[fname]]
               if (!is.null(x$nda_metadata) && "notes" %in% names(x$nda_metadata)) {
                 notes_val <- x$nda_metadata$notes %||% ""
                 notes_val <- if (is.na(notes_val) || notes_val == "NA" || notes_val == "character(0)") "" else as.character(notes_val)
                 # Strip internal modification text if ever present
                 notes_val <- gsub("\\bModified structure: Added missing value codes:.*$", "", notes_val)
                 trimws(notes_val)
               } else if (!is.null(x$notes)) {
                 as.character(x$notes %||% "")
               } else {
                 ""
               }
             }, error = function(e) "")
           })

           aliases <- sapply(field_names, function(fname) {
             tryCatch({
               x <- data_definition$fields[[fname]]
               if (!is.null(x$nda_metadata) && "aliases" %in% names(x$nda_metadata)) {
                 alias_val <- x$nda_metadata$aliases %||% ""
                 normalize_aliases_export(alias_val)
               } else {
                 ""
               }
             }, error = function(e) "")
           })

           # Curator notes to summarize changes (mirror CSV branch)
           # Check if field exists in NDA to avoid marking existing fields as "New field"
           nda_element_names <- data_definition$metadata$nda_element_names
           if (is.null(nda_element_names) || !is.character(nda_element_names) || length(nda_element_names) == 0) {
             nda_element_names <- character(0)
           }
           # Ensure field_names is a character vector
           if (is.null(field_names) || !is.character(field_names)) {
             field_names <- character(0)
           }
           curator_notes <- sapply(field_names, function(fname) {
             if (is.null(fname) || !is.character(fname) || length(fname) != 1) {
               return("")
             }
             x <- data_definition$fields[[fname]]
             if (is.null(x)) return("")
             parts <- character(0)
             # Check if field exists in NDA (ensure both are character vectors)
             exists_in_nda <- if (is.character(fname) && is.character(nda_element_names) && length(fname) == 1) {
               fname %in% nda_element_names
             } else {
               FALSE
             }
             if (!is.null(x$source)) {
               if (identical(x$source, "computed_from_data")) {
                 # If field exists in NDA but was marked as computed, it's likely a modified value range
                 if (exists_in_nda || isTRUE(x$is_modified)) {
                   parts <- c(parts, "Modified value range")
                 } else {
                   parts <- c(parts, "New field (not in NDA)")
                 }
               } else if (identical(x$source, "nda_modified") || isTRUE(x$is_modified)) {
                 parts <- c(parts, "Modified value range")
               } else if (identical(x$source, "template_only")) {
                 parts <- c(parts, "In template only; no data observed")
               } else if (identical(x$source, "nda_validated")) {
                 # Existing NDA fields that match perfectly - no curator notes needed
                 # Return empty string early to skip all curator annotations
                 return("")
               }
             }
             reqFlag <- NULL
              if (!is.null(x$nda_metadata) && is.list(x$nda_metadata) && "required" %in% names(x$nda_metadata)) {
                reqFlag <- x$nda_metadata$required
              } else if (!is.null(x$required)) {
                reqFlag <- ifelse(isTRUE(x$required), "Required", "Recommended")
              }
             if (!is.null(reqFlag) && identical(as.character(reqFlag), "Required")) {
               parts <- c(parts, "Required by NDA")
             }
             # Skip missing percentage in curator notes
             note <- paste(parts, collapse = " | ")
             if (identical(note, "")) {
               note <- paste0("Requesting to add ", fname, " as is")
             }
             note
           })

          # Conditional fields_df creation based on structure type
           if (is_new_structure) {
             # NEW STRUCTURE: 10 columns (with Instructions column)
             fields_df <- data.frame(
               ElementName = element_names,
               DataType = data_types,
               Size = sizes,
               Required = required_vals,
               ElementDescription = descriptions,
               ValueRange = value_ranges,
               Notes = notes,
               Aliases = aliases,
               `Instructions for Creating Template` = rep("", length(element_names)),
               `Notes for Data Curator` = curator_notes,
               stringsAsFactors = FALSE,
               check.names = FALSE
             )
           } else {
             # MODIFIED STRUCTURE: 9 columns (no Instructions column)
             fields_df <- data.frame(
               ElementName = element_names,
               DataType = data_types,
               Size = sizes,
               Required = required_vals,
               ElementDescription = descriptions,
               ValueRange = value_ranges,
               Notes = notes,
               Aliases = aliases,
               `Notes for Data Curator` = curator_notes,
               stringsAsFactors = FALSE,
               check.names = FALSE
             )
           }

           # Sanitize (same as CSV)
           fields_df$ElementName <- as.character(fields_df$ElementName)
           fields_df <- fields_df[!is.na(fields_df$ElementName) & fields_df$ElementName != "", , drop = FALSE]
           for (nm in names(fields_df)) {
             fields_df[[nm]][is.na(fields_df[[nm]])] <- ""
           }
           fields_df$DataType <- as.character(fields_df$DataType)
           fields_df$Required <- as.character(fields_df$Required)
           fields_df$DataType[!fields_df$DataType %in% c("String", "Integer", "Float", "Date", "GUID", "Boolean")] <- "String"
           fields_df$Required[!fields_df$Required %in% c("Required", "Recommended", "Conditional", "No")] <- "No"
           suppressWarnings(sz <- as.numeric(fields_df$Size))
           sz[is.na(sz)] <- NA_real_
           fields_df$Size <- ifelse(is.na(sz), "", as.character(sz))
           non_string_idx <- which(fields_df$DataType != "String")
           if (length(non_string_idx) > 0) {
             fields_df$Size[non_string_idx] <- ""
           }
            fields_df$ValueRange <- vapply(fields_df$ValueRange, compress_value_range, character(1))
            fields_df$Aliases <- vapply(fields_df$Aliases, normalize_aliases_export, character(1))

            # Build workbook with formatting
           wb <- openxlsx::createWorkbook()
           openxlsx::addWorksheet(wb, "Data Definitions")
           openxlsx::writeData(wb, "Data Definitions", fields_df, startRow = 1, startCol = 1, withFilter = TRUE)

           # Header style: yellow fill, bold
           headerStyle <- openxlsx::createStyle(textDecoration = "bold", fgFill = "#FFF2CC", halign = "left", valign = "top", border = "Bottom")
           openxlsx::addStyle(wb, "Data Definitions", headerStyle, rows = 1, cols = seq_len(ncol(fields_df)), gridExpand = TRUE)

            # Populate Instructions column for NEW structures (write to workbook after data is written)
            if (is_new_structure) {
              instructions_col_idx <- which(colnames(fields_df) == "Instructions for Creating Template")
              if (length(instructions_col_idx) == 1 && nrow(fields_df) > 0) {
                # Super required fields get DO NOT DELETE message (highest priority)
                super_required_names <- c("subjectkey", "src_subject_id", "interview_date", "interview_age", "sex")
                idx_super_required <- which(element_names %in% super_required_names)
                
                if (length(idx_super_required) > 0) {
                  instruction_text <- "DO NOT DELETE THIS ROW\nThis is a mandatory element and data must be submitted to this element."
                  for (idx in idx_super_required) {
                    openxlsx::writeData(wb, "Data Definitions", instruction_text, 
                                       startRow = idx + 1, startCol = instructions_col_idx, colNames = FALSE)
                  }
                }
              }
            }

            # Wrap text for description/notes/valueRange/instructions
            wrapStyle <- openxlsx::createStyle(wrapText = TRUE, valign = "top")
            wrap_col_names <- c("ElementDescription", "ValueRange", "Notes", "Notes for Data Curator")
            if (is_new_structure) {
              wrap_col_names <- c(wrap_col_names, "Instructions for Creating Template")
            }
            wrapCols <- which(colnames(fields_df) %in% wrap_col_names)
            if (length(wrapCols) > 0 && nrow(fields_df) > 0) {
              openxlsx::addStyle(wb, "Data Definitions", wrapStyle, rows = 2:(nrow(fields_df) + 1), cols = wrapCols, gridExpand = TRUE)
            }

            # NDA Highlighting Rules
           if (nrow(fields_df) > 0) {
             elementCol <- which(colnames(fields_df) == "ElementName")
             valueCols <- which(colnames(fields_df) %in% c("ValueRange", "Notes"))

            blueFill <- openxlsx::createStyle(fgFill = "#00B0F0")
            yellowFill <- openxlsx::createStyle(fgFill = "#FFFF00")
            redFont <- openxlsx::createStyle(fontColour = "#FF0000")
            
            # Check if openxlsx2 is available for rich text formatting
            has_openxlsx2 <- requireNamespace("openxlsx2", quietly = TRUE)
            
            # Collect rich text edits to apply post-save via openxlsx2
            rich_text_edits <- list()

            # Determine whether each element exists in NDA by lookup retained in metadata,
            # falling back to NDA API element search when not found locally
            nda_element_names <- data_definition$metadata$nda_element_names
            # Ensure nda_element_names is a character vector
            if (is.null(nda_element_names) || !is.character(nda_element_names)) {
              nda_element_names <- character(0)
            }
            
            in_local_nda <- element_names %in% nda_element_names

             # Cached API checker to minimize network calls
             .nda_exists_cache <- new.env(parent = emptyenv())
             nda_element_exists_api <- function(el_name) {
               if (!nzchar(el_name)) return(FALSE)
               if (!is.null(.nda_exists_cache[[el_name]])) return(.nda_exists_cache[[el_name]])
               base_url <- "https://nda.nih.gov/api/datadictionary/v2/datastructure/dataElement/"
               url <- paste0(base_url, utils::URLencode(el_name, reserved = TRUE))
               exists_flag <- FALSE
               try({
                 resp <- httr::GET(url, httr::timeout(8))
                 if (httr::status_code(resp) == 200) {
                   raw <- rawToChar(resp$content)
                   if (nzchar(raw)) {
                     parsed <- tryCatch(jsonlite::fromJSON(raw), error = function(e) NULL)
                     # Endpoint returns an array of structures containing the element; non-empty => exists
                     if (!is.null(parsed) && length(parsed) > 0) exists_flag <- TRUE
                   }
                 }
               }, silent = TRUE)
               .nda_exists_cache[[el_name]] <- exists_flag
               exists_flag
             }

              # CRITICAL FIX: Check against complete ndar_subject01 field list
              # This ensures consistent formatting regardless of dcc parameter
              # Previously, fields only got marked as "from NDA" if dcc=TRUE (merged)
              # Now we check ALL fields against the complete ndar_subject01 list
              
              # Get cached ndar_subject01 field list from data_definition metadata
              ndar_subject01_all_fields <- if (!is.null(data_definition$metadata$ndar_subject01_all_fields)) {
                data_definition$metadata$ndar_subject01_all_fields
              } else {
                character(0)
              }
              
              # Fallback: if not available from metadata, use known fields
              if (length(ndar_subject01_all_fields) == 0) {
                # Include super required + DCC fields + other known ndar_subject01 fields
                ndar_subject01_all_fields <- c(
                  SUPER_REQUIRED_FIELDS,
                  DCC_FIELDS,
                  "country_origin", "gender_identity", "demo_sex_tgender", "visit",
                  "handedness", "comments_misc", "guid", "interview_language"
                  # Add other common fields that exist in ndar_subject01
                )
              }
              
              # Check if field exists in ndar_subject01 (fast lookup, no API calls)
              in_ndar_subject01 <- element_names %in% ndar_subject01_all_fields
              
              # Also check via API for fields not in local structure or ndar_subject01 cache
              api_check_needed <- which(!in_local_nda & !in_ndar_subject01)
              api_exists <- rep(FALSE, length(element_names))
              if (length(api_check_needed) > 0) {
                for (idx in api_check_needed) {
                  api_exists[idx] <- nda_element_exists_api(element_names[idx])
                }
              }

              # Field is "in NDA" if it's in local structure OR ndar_subject01 OR via API
              element_is_in_nda <- in_local_nda | in_ndar_subject01 | api_exists
              
              if (verbose && sum(in_ndar_subject01) > 0) {
                ndar_subject_fields <- element_names[in_ndar_subject01]
                message(sprintf("[EXCEL] Detected %d field(s) from ndar_subject01: %s",
                               length(ndar_subject_fields),
                               paste(head(ndar_subject_fields, 5), collapse=", ")))
              }

             # Fetch NDA element metadata for allowed range comparison (cached)
             .nda_meta_cache <- new.env(parent = emptyenv())
             nda_fetch_element <- function(el_name) {
               if (!nzchar(el_name)) return(NULL)
               if (!is.null(.nda_meta_cache[[el_name]])) return(.nda_meta_cache[[el_name]])
               base_url <- "https://nda.nih.gov/api/datadictionary/v2/dataelement/"
               url <- paste0(base_url, utils::URLencode(el_name, reserved = TRUE))
               out <- NULL
               try({
                 resp <- httr::GET(url, httr::timeout(8))
                 if (httr::status_code(resp) == 200) {
                   raw <- rawToChar(resp$content)
                   if (nzchar(raw)) {
                     out <- tryCatch(jsonlite::fromJSON(raw), error = function(e) NULL)
                   }
                 }
               }, silent = TRUE)
               .nda_meta_cache[[el_name]] <- out
               out
             }

              # Determine which rows are modified (our proposal changes)
              # Note: expand_numeric_value_range() is now defined at package level
             row_modified <- vapply(field_names, function(fname) {
               x <- data_definition$fields[[fname]]
               isTRUE(x$is_modified) || identical(as.character(x$source %||% ""), "nda_modified")
             }, logical(1))

             # Track diagnostics for a Routes sheet
             route_source <- vapply(field_names, function(fname) {
               x <- data_definition$fields[[fname]]
               as.character(x$source %||% "")
             }, character(1))
             applied_blue <- rep(FALSE, length(field_names))
             applied_yellow <- rep(FALSE, length(field_names))
             applied_red <- rep(FALSE, length(field_names))
             nda_range_mismatch <- rep(FALSE, length(field_names))

               # Blue: NEW fields from NDA that weren't in the original structure
                 # These are fields that:
                 #   1. Exist in NDA globally (element_is_in_nda = TRUE) AND not in original structure (!in_local_nda)
                 #   2. OR from ndar_subject01 BUT NOT super required fields
                 # Super required fields (subjectkey, src_subject_id, interview_date, interview_age, sex):
                 #   - Appear NORMAL (no blue highlighting)
                 #   - Keep full metadata (DataType, Required="Required", ValueRange)
                 # Other ndar_subject01 fields get blue highlighting + metadata cleared
                 idx_new_from_nda <- which((element_is_in_nda & !in_local_nda) | 
                                           (in_ndar_subject01 & !(element_names %in% SUPER_REQUIRED_FIELDS)))
               
               if (length(idx_new_from_nda) > 0) {
                 openxlsx::addStyle(wb, "Data Definitions", blueFill, rows = idx_new_from_nda + 1, cols = elementCol, gridExpand = TRUE)
                 applied_blue[idx_new_from_nda] <- TRUE
                 
                 # For new elements (blue): clear all metadata columns and set curator note
                 cols_to_clear <- c("DataType", "Size", "Required", "ElementDescription", "ValueRange", "Notes", "Aliases")
                 clear_cols_idx <- which(colnames(fields_df) %in% cols_to_clear)
                 notes_col <- which(colnames(fields_df) == "Notes for Data Curator")
                 
                 for (idx in idx_new_from_nda) {
                   # Clear metadata columns
                   for (cc in clear_cols_idx) {
                     openxlsx::writeData(wb, "Data Definitions", "", startRow = idx + 1, startCol = cc, colNames = FALSE)
                   }
                   # Set curator note
                   if (length(notes_col) > 0) {
                     field_name <- element_names[idx]
                     openxlsx::writeData(wb, "Data Definitions", sprintf("Requesting to add %s as is", field_name), startRow = idx + 1, startCol = notes_col, colNames = FALSE)
                   }
                 }
                 
                 # Add Instructions for blue-highlighted fields (NEW structures only)
                 if (is_new_structure && exists("instructions_col_idx") && length(instructions_col_idx) == 1) {
                   instruction_text <- "If you found an existing Data Element, only add the exact ElementName and highlight the cell in BLUE. \nYou do not need to populate the other columns if you do not need changes to the Data Element."
                   super_required_names <- c("subjectkey", "src_subject_id", "interview_date", "interview_age", "sex")
                   
                   for (idx in idx_new_from_nda) {
                     element_name <- element_names[idx]
                     # Skip if super required (priority 1 already set)
                     if (!(element_name %in% super_required_names)) {
                       openxlsx::writeData(wb, "Data Definitions", instruction_text, 
                                          startRow = idx + 1, startCol = instructions_col_idx, colNames = FALSE)
                     }
                   }
                 }
               }

              # Yellow: modified NDA elements -> highlight changed definition columns (ValueRange/Notes here)
              idx_modified <- which(row_modified & element_is_in_nda)
              if (length(idx_modified) > 0 && length(valueCols) > 0) {
                # Yellow highlight on ValueRange and Notes cells
                openxlsx::addStyle(wb, "Data Definitions", yellowFill, rows = idx_modified + 1, cols = valueCols, gridExpand = TRUE)
                
                # Only apply red font as fallback when openxlsx2 is NOT available
                # (If openxlsx2 is available, per-token rich text will be applied later)
                if (!has_openxlsx2) {
                  vr_col_only <- which(colnames(fields_df) == "ValueRange")
                  if (length(vr_col_only) == 1) {
                    openxlsx::addStyle(wb, "Data Definitions", redFont, rows = idx_modified + 1, cols = vr_col_only, gridExpand = TRUE)
                  }
                }
                applied_yellow[idx_modified] <- TRUE
                
                # Add Instructions for yellow-highlighted fields (NEW structures only)
                if (is_new_structure && exists("instructions_col_idx") && length(instructions_col_idx) == 1) {
                  instruction_text <- "If you need changes to an existing Data Element, \n1. Add and highlight the ElementName in BLUE\n2. Highlight your changes in YELLOW (see Rules in ReadMe)"
                  super_required_names <- c("subjectkey", "src_subject_id", "interview_date", "interview_age", "sex")
                  
                  for (idx in idx_modified) {
                    element_name <- element_names[idx]
                    # Skip if super required (priority 1 already set)
                    if (!(element_name %in% super_required_names)) {
                      openxlsx::writeData(wb, "Data Definitions", instruction_text,
                                         startRow = idx + 1, startCol = instructions_col_idx, colNames = FALSE)
                    }
                  }
                }
              }

             # Yellow + red text on ValueRange for NDA elements whose exported ValueRange extends beyond NDA-allowed values
             if (length(valueCols) > 0) {
              for (i in which(element_is_in_nda)) {
                # Fast skip when our ValueRange is empty
                ours_str <- as.character(fields_df$ValueRange[i] %||% "")
                if (!nzchar(ours_str)) next
                # Fetch NDA metadata only when needed
                el <- element_names[i]
                nda_meta <- nda_fetch_element(el)
                if (is.null(nda_meta)) next
                nda_allowed <- expand_numeric_value_range(nda_meta$valueRange)
                if (is.null(nda_allowed) || length(nda_allowed) == 0) next
                ours_vals <- expand_numeric_value_range(ours_str)
                if (is.null(ours_vals)) next
                  if (length(setdiff(ours_vals, nda_allowed)) > 0) {
                  # Clear definition columns except ValueRange and Notes
                  clear_cols <- c("DataType", "Size", "Required", "ElementDescription")
                  clear_idx <- which(colnames(fields_df) %in% clear_cols)
                  if (length(clear_idx) > 0) {
                    for (cc in clear_idx) {
                      openxlsx::writeData(wb, "Data Definitions", "", startRow = i + 1, startCol = cc, colNames = FALSE)
                    }
                  }

                  # Apply yellow fill to ValueRange, Notes, and Notes for Data Curator cells to indicate change
                  vr_col <- which(colnames(fields_df) == "ValueRange")
                  notes_col <- which(colnames(fields_df) == "Notes")
                  curator_notes_col <- which(colnames(fields_df) == "Notes for Data Curator")
                  
                  if (length(vr_col) == 1) {
                    openxlsx::addStyle(wb, "Data Definitions", yellowFill, rows = i + 1, cols = vr_col, gridExpand = TRUE)
                  }
                  if (length(notes_col) == 1) {
                    openxlsx::addStyle(wb, "Data Definitions", yellowFill, rows = i + 1, cols = notes_col, gridExpand = TRUE)
                  }
                  if (length(curator_notes_col) == 1) {
                    openxlsx::addStyle(wb, "Data Definitions", yellowFill, rows = i + 1, cols = curator_notes_col, gridExpand = TRUE)
                  }

                  # Build rich-text instructions to apply later with openxlsx2
                  tokens <- trimws(strsplit(ours_str, ";")[[1]])
                  tokens <- tokens[nzchar(tokens)]
                  added_tokens <- character(0)
                  token_is_added <- logical(length(tokens))
                  for (ti in seq_along(tokens)) {
                    tok <- tokens[ti]
                    # Support tokens like "9999=Prefer not to say" by extracting leading code before '='
                    leading_code <- suppressWarnings(as.numeric(sub("=.*$", "", tok)))
                    if (!is.na(leading_code)) {
                      tok_vals <- leading_code
                    } else {
                      tok_vals <- expand_numeric_value_range(tok)
                    }
                  token_is_added[ti] <- length(setdiff(tok_vals %||% numeric(0), nda_allowed)) > 0
                  if (token_is_added[ti]) added_tokens <- c(added_tokens, tok)
                }
                # Validate data before adding to rich_text_edits (prevent openxlsx2 errors)
                if (length(tokens) > 0 && 
                    length(token_is_added) == length(tokens) &&
                    is.character(tokens) && 
                    is.logical(token_is_added)) {
                  rich_text_edits[[length(rich_text_edits) + 1]] <- list(
                    row = i + 1,
                    vr_col = if (length(vr_col) == 1) vr_col else NA_integer_,
                    notes_col = if (length(notes_col) == 1) notes_col else NA_integer_,
                    tokens = tokens,
                    token_is_added = token_is_added,
                    added_tokens = added_tokens
                  )
                }
                  applied_yellow[i] <- TRUE
                  nda_range_mismatch[i] <- TRUE
                  
                  # Write ONLY the proposed range to ValueRange column (not the comparison)
                  # The comparison info is for curator reference and goes in Notes for Data Curator
                  vr_col <- which(colnames(fields_df) == "ValueRange")
                  if (length(vr_col) == 1) {
                    # Write only our proposed range (already in fields_df$ValueRange[i])
                    openxlsx::writeData(wb, "Data Definitions", fields_df$ValueRange[i], 
                                       startRow = i + 1, startCol = vr_col, colNames = FALSE)
                  }
                  
                  # Pull REDCap choices for NDA fields with value range mismatches
                  # This ensures extended fields have proper value labels (1=Female; 2=Male; etc.) in Notes
                  # Enhanced: Show ALL codes with NEW codes in red (when openxlsx2 available)
                  notes_col <- which(colnames(fields_df) == "Notes")
                  notes_text <- ""
                  notes_rich_text_parts <- NULL
                  
                  if (exists("redcap_choices_map") && !is.null(redcap_choices_map) && 
                      element_names[i] %in% names(redcap_choices_map)) {
                    redcap_choices <- redcap_choices_map[[element_names[i]]]
                    if (!is.null(redcap_choices) && nzchar(redcap_choices)) {
                      parsed <- rc_parse_choices_codes_and_notes(redcap_choices)
                      if (parsed$notes != "") {
                        notes_text <- parsed$notes
                        
                        # Parse individual code=label pairs for rich text highlighting
                        if (has_openxlsx2 && !is.null(nda_allowed) && length(nda_allowed) > 0) {
                          note_parts <- trimws(strsplit(parsed$notes, ";", fixed = TRUE)[[1]])
                          note_parts <- note_parts[nzchar(note_parts)]
                          
                          notes_rich_text_parts <- list()
                          for (np_idx in seq_along(note_parts)) {
                            part <- note_parts[np_idx]
                            
                            # Extract code from "13=Two-Spirit" format
                            code_match <- regmatches(part, regexpr("^\\d+", part))
                            if (length(code_match) > 0) {
                              code <- suppressWarnings(as.integer(code_match))
                              # Check if this code is NEW (not in NDA's allowed range)
                              is_new <- !is.na(code) && !(code %in% nda_allowed)
                            } else {
                              is_new <- FALSE
                            }
                            
                            # Add to rich text parts
                            notes_rich_text_parts[[length(notes_rich_text_parts) + 1]] <- list(
                              text = part,
                              color = if (is_new) "FF0000" else "000000"  # Red if new, black if existing
                            )
                            
                            # Add separator (unless last item)
                            if (np_idx < length(note_parts)) {
                              notes_rich_text_parts[[length(notes_rich_text_parts) + 1]] <- list(
                                text = "; ",
                                color = "000000"
                              )
                            }
                          }
                        }
                      }
                    }
                  }
                  
                  # If no REDCap choices, try to infer value labels from actual data
                  if (!nzchar(notes_text) && !is.null(ours_vals) && length(ours_vals) > 0 && 
                      exists("data_frame") && !is.null(data_frame) && is.data.frame(data_frame)) {
                    field_name <- element_names[i]
                    if (field_name %in% names(data_frame)) {
                      # Get unique values from the data column
                      data_vals <- unique(data_frame[[field_name]])
                      data_vals <- data_vals[!is.na(data_vals)]
                      
                      # Only proceed if values are numeric and match our value range
                      if (length(data_vals) > 0 && is.numeric(data_vals)) {
                        data_vals <- sort(as.integer(data_vals))
                        # Filter to only values in our proposed range
                        data_vals <- data_vals[data_vals %in% ours_vals]
                        
                        if (length(data_vals) > 0) {
                          # Build a simple list: "1; 2; 3; 4; ..." 
                          notes_parts_inferred <- as.character(data_vals)
                          notes_text <- paste(notes_parts_inferred, collapse = "; ")
                          
                          # Build rich text parts (highlight new values in red)
                          if (has_openxlsx2 && !is.null(nda_allowed) && length(nda_allowed) > 0) {
                            notes_rich_text_parts <- list()
                            for (dv_idx in seq_along(data_vals)) {
                              val <- data_vals[dv_idx]
                              is_new <- !(val %in% nda_allowed)
                              
                              notes_rich_text_parts[[length(notes_rich_text_parts) + 1]] <- list(
                                text = as.character(val),
                                color = if (is_new) "FF0000" else "000000"
                              )
                              
                              # Add separator (unless last item)
                              if (dv_idx < length(data_vals)) {
                                notes_rich_text_parts[[length(notes_rich_text_parts) + 1]] <- list(
                                  text = "; ",
                                  color = "000000"
                                )
                              }
                            }
                          }
                        }
                      }
                    }
                  }
                  
                  # Write Notes column
                  if (length(notes_col) == 1 && nzchar(notes_text)) {
                    openxlsx::writeData(wb, "Data Definitions", notes_text,
                                       startRow = i + 1, startCol = notes_col, colNames = FALSE)
                    
                    # Store rich text parts for Notes column (if openxlsx2 will be used)
                    if (has_openxlsx2 && !is.null(notes_rich_text_parts) && length(notes_rich_text_parts) > 0) {
                      # Find or create the rich_text_edits entry for this row
                      edit_idx <- which(sapply(rich_text_edits, function(e) e$row == (i + 1)))
                      if (length(edit_idx) == 0) {
                        # Create new entry
                        rich_text_edits[[length(rich_text_edits) + 1]] <- list(
                          row = i + 1,
                          notes_col = notes_col,
                          notes_rich_text_parts = notes_rich_text_parts
                        )
                      } else {
                        # Update existing entry
                        rich_text_edits[[edit_idx[1]]]$notes_rich_text_parts <- notes_rich_text_parts
                      }
                    }
                  }
                  
                  # Update curator notes with specific modification details
                  notes_col_curator <- which(colnames(fields_df) == "Notes for Data Curator")
                  if (length(notes_col_curator) == 1) {
                    # Determine what kind of modification occurred
                    modification_detail <- ""
                    curator_rich_text_parts <- NULL  # For rich text formatting
                    
                    # Check if this is a pure range extension first
                    is_range_extension <- FALSE
                    if (!is.null(ours_vals) && !is.null(nda_allowed)) {
                      our_min <- min(ours_vals)
                      our_max <- max(ours_vals)
                      nda_min <- min(nda_allowed)
                      nda_max <- max(nda_allowed)
                      
                      # It's a range extension if bounds changed and there's only one token that looks like a range
                      if ((our_min < nda_min || our_max > nda_max) && 
                          length(added_tokens) == 1 && 
                          grepl("::", added_tokens[1], fixed = TRUE)) {
                        is_range_extension <- TRUE
                        modification_detail <- sprintf("We extended range from %s::%s to %s::%s", 
                                                      nda_min, nda_max, our_min, our_max)
                      }
                    }
                    
                    # If not a range extension, categorize and format the additions
                    if (!is_range_extension && length(added_tokens) > 0) {
                      # Categorize added tokens: categorical (contain "=") vs numeric ranges
                      categorical_additions <- added_tokens[grepl("=", added_tokens, fixed = TRUE)]
                      numeric_range_additions <- added_tokens[!grepl("=", added_tokens, fixed = TRUE) & 
                                                             grepl("::", added_tokens, fixed = TRUE)]
                      
                      # Extract codes from categorical additions
                      added_codes <- character(0)
                      if (length(categorical_additions) > 0) {
                        added_codes <- sapply(categorical_additions, function(x) {
                          trimws(sub("=.*$", "", x))
                        }, USE.NAMES = FALSE)
                      } else if (length(numeric_range_additions) > 0) {
                        added_codes <- numeric_range_additions
                      }
                      
                      # Format the curator note message
                      if (length(categorical_additions) > 0) {
                        # Categorical values added
                        if (length(added_codes) == 1) {
                          modification_detail <- sprintf("We added another value, %s, to this data element", 
                                                        added_codes[1])
                          # Rich text parts: "We added another value, " + RED(code) + ", to this data element"
                          curator_rich_text_parts <- list(
                            list(text = "Modified value range: We added another value, ", color = "000000"),
                            list(text = added_codes[1], color = "FF0000"),
                            list(text = ", to this data element", color = "000000")
                          )
                        } else if (length(added_codes) == 2) {
                          values_str <- paste(added_codes, collapse = " and ")
                          modification_detail <- sprintf("We added values %s to this data element", values_str)
                          # Rich text: "We added values " + RED(code1) + " and " + RED(code2) + " to this data element"
                          curator_rich_text_parts <- list(
                            list(text = "Modified value range: We added values ", color = "000000"),
                            list(text = added_codes[1], color = "FF0000"),
                            list(text = " and ", color = "000000"),
                            list(text = added_codes[2], color = "FF0000"),
                            list(text = " to this data element", color = "000000")
                          )
                        } else {
                          # 3+ values: "We added values 4, 5, and 6 to this data element"
                          values_str <- paste(paste(utils::head(added_codes, -1), collapse = ", "), 
                                            "and", utils::tail(added_codes, 1))
                          modification_detail <- sprintf("We added values %s to this data element", values_str)
                          # Rich text: alternate between black and red for each code
                          curator_rich_text_parts <- list(
                            list(text = "Modified value range: We added values ", color = "000000")
                          )
                          for (j in seq_along(added_codes)) {
                            curator_rich_text_parts[[length(curator_rich_text_parts) + 1]] <- 
                              list(text = added_codes[j], color = "FF0000")
                            if (j < length(added_codes) - 1) {
                              curator_rich_text_parts[[length(curator_rich_text_parts) + 1]] <- 
                                list(text = ", ", color = "000000")
                            } else if (j == length(added_codes) - 1) {
                              curator_rich_text_parts[[length(curator_rich_text_parts) + 1]] <- 
                                list(text = " and ", color = "000000")
                            }
                          }
                          curator_rich_text_parts[[length(curator_rich_text_parts) + 1]] <- 
                            list(text = " to this data element", color = "000000")
                        }
                      } else if (length(numeric_range_additions) > 0) {
                        # Range segments added (not extensions)
                        if (length(numeric_range_additions) == 1) {
                          modification_detail <- sprintf("We added another range, %s, to this data element", 
                                                        numeric_range_additions[1])
                          # Rich text: "We added another range, " + RED(range) + ", to this data element"
                          curator_rich_text_parts <- list(
                            list(text = "Modified value range: We added another range, ", color = "000000"),
                            list(text = numeric_range_additions[1], color = "FF0000"),
                            list(text = ", to this data element", color = "000000")
                          )
                        } else {
                          ranges_str <- paste(numeric_range_additions, collapse = ", ")
                          modification_detail <- sprintf("We added ranges %s to this data element", ranges_str)
                          # Rich text: alternate between black and red for each range
                          curator_rich_text_parts <- list(
                            list(text = "Modified value range: We added ranges ", color = "000000")
                          )
                          for (j in seq_along(numeric_range_additions)) {
                            curator_rich_text_parts[[length(curator_rich_text_parts) + 1]] <- 
                              list(text = numeric_range_additions[j], color = "FF0000")
                            if (j < length(numeric_range_additions)) {
                              curator_rich_text_parts[[length(curator_rich_text_parts) + 1]] <- 
                                list(text = ", ", color = "000000")
                            }
                          }
                          curator_rich_text_parts[[length(curator_rich_text_parts) + 1]] <- 
                            list(text = " to this data element", color = "000000")
                        }
                      } else {
                        # Fallback for other additions
                        added_str <- paste(added_tokens, collapse = ", ")
                        modification_detail <- sprintf("We added value(s): %s", added_str)
                      }
                    } else if (!is_range_extension && !is.null(ours_vals) && !is.null(nda_allowed)) {
                      # Some other kind of modification
                      modification_detail <- "Value range differs from NDA"
                    } else if (modification_detail == "") {
                      # Fallback
                      modification_detail <- "Value range differs from NDA"
                    }
                    
                    # Write curator note (plain text - will be replaced with rich text later if available)
                    curator_note <- sprintf("Modified value range: %s", modification_detail)
                    openxlsx::writeData(wb, "Data Definitions", curator_note, 
                                       startRow = i + 1, startCol = notes_col_curator, colNames = FALSE)
                    
                    # Add to rich_text_edits for curator note column if we have rich text formatting
                    if (!is.null(curator_rich_text_parts)) {
                      rich_text_edits[[length(rich_text_edits) + 1]] <- list(
                        row = i + 1,
                        curator_col = notes_col_curator,
                        curator_rich_text_parts = curator_rich_text_parts
                      )
                    }
                  }
                }
              }
            }

            # Red text only (no yellow fill): elements not in NDA (new/proposed elements)
            idx_not_in_nda <- which(!element_is_in_nda)
            if (length(idx_not_in_nda) > 0) {
              openxlsx::addStyle(wb, "Data Definitions", redFont, rows = idx_not_in_nda + 1, cols = seq_len(ncol(fields_df)), gridExpand = TRUE)
              applied_red[idx_not_in_nda] <- TRUE
              
              # Add Instructions for red text fields (NEW structures only)
              if (is_new_structure && exists("instructions_col_idx") && length(instructions_col_idx) == 1) {
                instruction_text <- "If you need to add a new data element, populate Columns A-G and change text to RED."
                
                for (idx in idx_not_in_nda) {
                  openxlsx::writeData(wb, "Data Definitions", instruction_text,
                                     startRow = idx + 1, startCol = instructions_col_idx, colNames = FALSE)
                }
              }
            }

             # Add diagnostics worksheet summarizing route and highlighting decisions
             routes_df <- data.frame(
               ElementName = element_names,
               InNDA = element_is_in_nda,
               Source = route_source,
               Modified = row_modified,
               BlueApplied = applied_blue,
               YellowApplied = applied_yellow,
               RedApplied = applied_red,
               NdaRangeMismatch = nda_range_mismatch,
               stringsAsFactors = FALSE,
               check.names = FALSE
             )
             openxlsx::addWorksheet(wb, "Routes")
             openxlsx::writeData(wb, "Routes", routes_df, startRow = 1, startCol = 1, withFilter = TRUE)
             openxlsx::setColWidths(wb, "Routes", cols = seq_len(ncol(routes_df)), widths = "auto")

              # NOTE: With NdaDataStructure, we now ALWAYS show NDA metadata for ALL fields.
              # The old behavior was to clear metadata for "unchanged" NDA fields and only show
              # "Requesting to add X as is" in curator notes. This was bad UX.
              # Now we show full metadata so users can see field definitions without external lookup.
              
              # We no longer clear metadata columns for any NDA fields.
           }
          
          # Auto widths
          openxlsx::setColWidths(wb, "Data Definitions", cols = seq_len(ncol(fields_df)), widths = "auto")

          openxlsx::saveWorkbook(wb, file_path, overwrite = TRUE)

          # Post-process with openxlsx2 to apply per-token rich text if needed
          if (length(rich_text_edits) > 0) {
            # Use has_openxlsx2 from earlier check (line ~1973)
            if (!has_openxlsx2) {
              message("Note: openxlsx2 package not installed - rich text formatting skipped")
              message("      Install with: install.packages('openxlsx2')")
              message("      File created successfully with yellow highlighting")
            } else {
              tryCatch({
                # Validate rich_text_edits structure before processing
                # Note: rich_text_edits can contain two types:
                # 1. ValueRange/Notes edits (have tokens, token_is_added, vr_col, notes_col)
                # 2. Curator note edits (have curator_col, curator_rich_text_parts)
                valid_edits <- Filter(function(edit) {
                  tryCatch({
                    if (!is.list(edit) || is.null(edit$row)) return(FALSE)
                    if (!(is.numeric(edit$row) && length(edit$row) == 1)) return(FALSE)
                    
                    # Check Type 1: ValueRange/Notes edit
                    is_type1 <- FALSE
                    if (!is.null(edit$tokens) && !is.null(edit$token_is_added)) {
                      is_type1 <- (
                        is.character(edit$tokens) &&
                        is.logical(edit$token_is_added) &&
                        length(edit$tokens) == length(edit$token_is_added) &&
                        length(edit$tokens) > 0 &&
                        (is.null(edit$vr_col) || isTRUE(is.na(edit$vr_col)) || 
                         (is.numeric(edit$vr_col) && length(edit$vr_col) == 1)) &&
                        (is.null(edit$notes_col) || isTRUE(is.na(edit$notes_col)) || 
                         (is.numeric(edit$notes_col) && length(edit$notes_col) == 1))
                      )
                    }
                    
                    # Check Type 2: Curator note edit
                    is_type2 <- FALSE
                    if (!is.null(edit$curator_col) && !is.null(edit$curator_rich_text_parts)) {
                      is_type2 <- (
                        is.numeric(edit$curator_col) &&
                        length(edit$curator_col) == 1 &&
                        is.list(edit$curator_rich_text_parts) &&
                        length(edit$curator_rich_text_parts) > 0
                      )
                    }
                    
                    return(is_type1 || is_type2)
                  }, error = function(e) {
                    # If validation fails for this edit, skip it
                    return(FALSE)
                  })
                }, rich_text_edits)
                
                if (length(valid_edits) == 0) {
                  message("Note: No valid rich text edits to apply")
                } else {
                  # Helper to convert numeric column index to Excel column letters
                  num_to_col <- function(n) {
                    letters <- c()
                    while (n > 0) {
                      r <- (n - 1) %% 26
                      letters <- c(LETTERS[r + 1], letters)
                      n <- (n - 1) %/% 26
                    }
                    paste0(letters, collapse = "")
                  }
                  
                  wb2 <- openxlsx2::wb_load(file_path)
                  # Resolve exported functions dynamically to avoid R CMD check NOTE/WARNING
                  create_rich_text_fn <- get("create_rich_text", asNamespace("openxlsx2"))
                  wb_add_rich_text_fn <- get("wb_add_rich_text", asNamespace("openxlsx2"))
                  
                  for (edit in valid_edits) {
                    # Additional safety check: ensure edit$vr_col and edit$notes_col are scalar integers
                    tryCatch({
                      # ValueRange rich text
                      if (!is.na(edit$vr_col) && length(edit$vr_col) == 1 && length(edit$tokens) > 0) {
                        runs <- list()
                        for (ti in seq_along(edit$tokens)) {
                          # Ensure token is scalar character
                          tok <- as.character(edit$tokens[ti])
                          if (length(tok) != 1) next
                          
                          # Ensure token_is_added[ti] is scalar logical
                          is_added <- edit$token_is_added[ti]
                          if (length(is_added) != 1 || is.na(is_added)) is_added <- FALSE
                          
                          col <- if (is_added) "FF0000" else "000000"
                          runs[[length(runs) + 1]] <- create_rich_text_fn(text = tok, color = col)
                          if (ti < length(edit$tokens)) {
                            runs[[length(runs) + 1]] <- create_rich_text_fn(text = "; ", color = "000000")
                          }
                        }
                        if (length(runs) > 0) {
                          dims <- paste0(num_to_col(edit$vr_col), edit$row)
                          wb_add_rich_text_fn(wb2, sheet = "Data Definitions", dims = dims, x = runs)
                        }
                      }
                      # Notes rich text - Support both old format (added_tokens) and new format (notes_rich_text_parts)
                      if (!is.na(edit$notes_col) && length(edit$notes_col) == 1) {
                        runs2 <- list()
                        
                        # NEW FORMAT: Use notes_rich_text_parts if available (full code list with colors)
                        if (!is.null(edit$notes_rich_text_parts) && length(edit$notes_rich_text_parts) > 0) {
                          for (part in edit$notes_rich_text_parts) {
                            if (is.list(part) && !is.null(part$text) && !is.null(part$color)) {
                              runs2[[length(runs2) + 1]] <- create_rich_text_fn(
                                text = as.character(part$text), 
                                color = part$color
                              )
                            }
                          }
                        } 
                        # OLD FORMAT: Fallback to added_tokens only (for backward compatibility)
                        else if (!is.null(edit$added_tokens) && length(edit$added_tokens) > 0) {
                          for (j in seq_along(edit$added_tokens)) {
                            tok2 <- as.character(edit$added_tokens[j])
                            if (length(tok2) != 1) next
                            
                            runs2[[length(runs2) + 1]] <- create_rich_text_fn(text = tok2, color = "FF0000")
                            if (j < length(edit$added_tokens)) {
                              runs2[[length(runs2) + 1]] <- create_rich_text_fn(text = "; ", color = "000000")
                            }
                          }
                        }
                        
                        if (length(runs2) > 0) {
                          dims2 <- paste0(num_to_col(edit$notes_col), edit$row)
                          wb_add_rich_text_fn(wb2, sheet = "Data Definitions", dims = dims2, x = runs2)
                        }
                      }
                      
                      # Curator Notes rich text (Notes for Data Curator column)
                      if (!is.null(edit$curator_col) && !is.na(edit$curator_col) && 
                          length(edit$curator_col) == 1 && 
                          !is.null(edit$curator_rich_text_parts) && 
                          length(edit$curator_rich_text_parts) > 0) {
                        curator_runs <- list()
                        for (part in edit$curator_rich_text_parts) {
                          if (is.list(part) && !is.null(part$text) && !is.null(part$color)) {
                            curator_runs[[length(curator_runs) + 1]] <- 
                              create_rich_text_fn(text = as.character(part$text), 
                                                color = part$color)
                          }
                        }
                        if (length(curator_runs) > 0) {
                          curator_dims <- paste0(num_to_col(edit$curator_col), edit$row)
                          wb_add_rich_text_fn(wb2, sheet = "Data Definitions", 
                                            dims = curator_dims, x = curator_runs)
                        }
                      }
                    }, error = function(e) {
                      # Skip this edit if it fails, continue with others
                      # Silently skip - file already has yellow highlighting
                    })
                  }
                  openxlsx2::wb_save(wb2, file = file_path, overwrite = TRUE)
                }
              }, error = function(e) {
                warning("Rich text formatting failed: ", e$message, call. = FALSE)
              })
            }
          }
}

# Helper function for null coalescing
`%||%` <- function(x, y) if (is.null(x)) y else x

# Compress numeric ValueRange like "1; 2; 3; 4; 5; 9999" -> "1::5;9999"
compress_value_range <- function(val_range) {
  if (is.null(val_range) || identical(val_range, "") || is.na(val_range)) return("")
  parts <- trimws(strsplit(as.character(val_range), ";")[[1]])
  parts <- parts[nzchar(parts)]
  if (length(parts) == 0) return("")
  suppressWarnings(nums <- as.numeric(parts))
  if (any(is.na(nums))) {
    # Not purely numeric; normalize spacing only
    return(paste(parts, collapse = "; "))
  }

  # Keep special outliers separate (e.g., 77, 88, 99, 777, 999, -7, -8, -9, 9999) if they are not contiguous with the main run
  # Strategy: sort ascending to find contiguous runs, then join. This also fixes descending inputs like 5;4;3;2;1;0.
  nums_sorted <- sort(unique(nums))

  runs <- list()
  run_start <- nums_sorted[1]
  run_prev <- nums_sorted[1]
  for (i in seq_along(nums_sorted)) {
    if (i == 1) next
    if (nums_sorted[i] == run_prev + 1) {
      run_prev <- nums_sorted[i]
    } else {
      runs[[length(runs) + 1]] <- c(run_start, run_prev)
      run_start <- nums_sorted[i]
      run_prev <- nums_sorted[i]
    }
  }
  # close last run
  runs[[length(runs) + 1]] <- c(run_start, run_prev)

  # Build tokens: ranges as a::b when length >= 3, otherwise list numbers individually
  tokens <- character(0)
  for (r in runs) {
    a <- r[1]; b <- r[2]
    if (b - a + 1 >= 3) {
      tokens <- c(tokens, paste0(a, "::", b))
    } else if (a == b) {
      tokens <- c(tokens, as.character(a))
    } else {
      tokens <- c(tokens, as.character(a), as.character(b))
    }
  }

  # Preserve any values from the original list that were not numeric (already handled above) or out-of-order duplicates (deduped)
  paste(tokens, collapse = ";")
}

# Normalize aliases to plain comma-separated (no quotes), stripping c() and character(0)
normalize_aliases_export <- function(alias_val) {
  if (is.null(alias_val)) return("")
  # If it's a list, flatten first
  if (is.list(alias_val)) {
    alias_val <- unlist(alias_val, use.names = FALSE)
    if (length(alias_val) == 0) return("")
  }
  # Vector case (after possible flatten)
  if (is.vector(alias_val) && !is.list(alias_val)) {
    alias_vec <- as.character(alias_val)
    alias_vec <- alias_vec[!is.na(alias_vec) & nzchar(alias_vec)]
    if (length(alias_vec) == 0) return("")
    # remove any embedded quotes in vector elements and suppress 'list()'
    alias_vec <- gsub('"', "", alias_vec, fixed = TRUE)
    alias_vec <- alias_vec[alias_vec != "list()"]
    if (length(alias_vec) == 0) return("")
    return(paste(alias_vec, collapse = ","))
  }
  # Character scalar that might contain c() or quotes
  if (is.character(alias_val) && length(alias_val) == 1) {
    s <- as.character(alias_val)
    if (is.na(s) || !nzchar(s)) return("")
    if (identical(trimws(s), "character(0)") || identical(trimws(s), "list()")) return("")
    # remove all embedded quotes first
    s <- gsub('"', "", s, fixed = TRUE)
    # handle character(0) anywhere
    if (grepl("character\\(0\\)", s)) return("")
    # strip c( ... ) wrapper, tolerate spaces using POSIX classes and proper escaping
    if (grepl("^c[[:space:]]*\\(.*\\)[[:space:]]*$", s, perl = TRUE)) {
      s <- sub("^c[[:space:]]*\\((.*)\\)[[:space:]]*$", "\\1", s, perl = TRUE)
    }
    # split by comma
    parts <- trimws(strsplit(s, ",")[[1]])
    parts <- parts[!is.na(parts) & nzchar(parts) & parts != "list()"]
    if (length(parts) == 0) return("")
    return(paste(parts, collapse = ","))
  }
  # Fallback: coerce to character, join, then strip like scalar path
  as_char <- as.character(alias_val)
  as_char <- gsub('"', "", as_char, fixed = TRUE)
  as_char <- as_char[!is.na(as_char) & nzchar(as_char) & !grepl("character\\(0\\)|^list\\(\\)$", as_char)]
  if (length(as_char) == 0) return("")
  s <- paste(as_char, collapse = ",")
  if (grepl("^c[[:space:]]*\\(.*\\)[[:space:]]*$", s, perl = TRUE)) {
    s <- sub("^c[[:space:]]*\\((.*)\\)[[:space:]]*$", "\\1", s, perl = TRUE)
  }
  parts <- trimws(strsplit(s, ",")[[1]])
  parts <- parts[!is.na(parts) & nzchar(parts) & parts != "list()"]
  if (length(parts) == 0) return("")
  paste(parts, collapse = ",")
}

#' StandardOutput R6 Class
#'
#' @description
#' Manages API-specific standard output fields that should be excluded from NDA processing.
#' These are metadata fields automatically generated by data collection platforms.
#'
#' Standard output fields are removed early in the nda() workflow to prevent them from
#' appearing in submission templates or data definition files.
#'
#' @keywords internal
#' @noRd
StandardOutput <- R6::R6Class(
  "StandardOutput",
  
  public = list(
    #' @field redcap Character vector - REDCap standard output fields
    #' - {measure_name}_complete: Form completion status
    #' - record_id: REDCap internal record ID (alias for src_subject_id)
    #' - redcap_event_name: Longitudinal event name
    redcap = c("{measure_name}_complete", "record_id", "redcap_event_name"),
    
    #' @field qualtrics Character vector - Qualtrics standard output fields
    qualtrics = c(
      "StartDate", "EndDate", "Status", "Progress", "Duration (in seconds)",
      "Finished", "RecordedDate", "ResponseId", "DistributionChannel",
      "UserLanguage", "candidateId", "studyId", "measure", "ATTN", "ATTN_1", "SC0",
      "IPAddress", "RecipientLastName", "RecipientFirstName", "RecipientEmail",
      "ExternalReference", "LocationLatitude", "LocationLongitude"
    ),
    
    #' @field mongo Character vector - MongoDB standard output fields
    mongo = c("_id"),
    
    #' @field sql Character vector - SQL standard output fields
    sql = character(0),
    
    #' @field oracle Character vector - Oracle standard output fields
    oracle = character(0),
    
    #' @field csv Character vector - CSV standard output fields
    csv = character(0),
    
    #' @description
    #' Get standard output fields for a specific API
    #'
    #' @param api Character. API type (redcap, qualtrics, mongo, sql, oracle, csv)
    #' @param measure_name Character. Measure name for dynamic substitution (e.g., {measure_name}_complete)
    #' @param categorical_vars CategoricalVariables instance (optional). If provided,
    #'   categorical variables will be excluded from the removal list to prevent accidental deletion
    #'
    #' @return Character vector of field names to remove
    get_fields = function(api, measure_name = NULL, categorical_vars = NULL) {
      if (!api %in% names(self)) {
        warning(sprintf("Unknown API type: %s. Returning empty field list.", api))
        return(character(0))
      }
      
      fields <- self[[api]]
      
      # Replace {measure_name} placeholder if provided
      if (!is.null(measure_name) && length(fields) > 0) {
        fields <- gsub("\\{measure_name\\}", measure_name, fields)
      }
      
      # Preserve categorical variables if provided
      # This prevents accidental removal if a categorical variable has the same name
      # as a standard output field (which would be an error condition)
      if (!is.null(categorical_vars) && inherits(categorical_vars, "CategoricalVariables")) {
        all_categorical <- categorical_vars$get_all()
        
        # Check for overlap (error condition)
        overlap <- intersect(fields, all_categorical)
        if (length(overlap) > 0) {
          stop(sprintf(
            "Field name conflict: The following field(s) appear in both StandardOutput and CategoricalVariables: %s. This is not allowed.",
            paste(overlap, collapse=", ")
          ))
        }
      }
      
      return(fields)
    },
    
    #' @description
    #' Remove standard output fields from a dataframe
    #'
    #' @param df Data frame to clean
    #' @param api Character. API type
    #' @param measure_name Character. Measure name
    #' @param categorical_vars CategoricalVariables instance (optional). Prevents removal of categorical fields
    #' @param verbose Logical. Whether to print verbose messages
    #'
    #' @return Cleaned data frame with standard output fields removed
    remove_from_df = function(df, api, measure_name = NULL, categorical_vars = NULL, verbose = FALSE) {
      if (!is.data.frame(df)) {
        warning("Input is not a data frame. Returning unchanged.")
        return(df)
      }
      
      # Get fields to remove (with categorical variable check)
      fields_to_remove <- self$get_fields(api, measure_name, categorical_vars)
      
      if (length(fields_to_remove) == 0) {
        return(df)
      }
      
      # Find which fields actually exist in df
      existing_to_remove <- intersect(names(df), fields_to_remove)
      
      if (length(existing_to_remove) == 0) {
        if (verbose) {
          message(sprintf("[CLEANUP] No %s standard output fields found in dataframe", toupper(api)))
        }
        return(df)
      }
      
      # Always show removal message (even if verbose=FALSE)
      # This is important for transparency about what's being removed
      if (length(existing_to_remove) > 0) {
        message(sprintf("[CLEANUP] Removing %d %s standard output field(s): %s",
                       length(existing_to_remove),
                       toupper(api),
                       paste(existing_to_remove, collapse=", ")))
      }
      
      # Check if any categorical variables exist in dataframe
      # Show message about preservation (non-verbose mode)
      if (!is.null(categorical_vars)) {
        categorical_in_df <- categorical_vars$get_existing(df)
        if (length(categorical_in_df) > 0) {
          message(sprintf("[CLEANUP] Detected categorical variables: %s - adding to data structure and submission",
                         paste(categorical_in_df, collapse=", ")))
        }
      }
      
      # Remove fields
      df[, !names(df) %in% existing_to_remove, drop = FALSE]
    }
  )
)

#' NDA Data Transformation Functions
#'
#' @description
#' Functions for transforming data to meet NDA requirements.
#' Includes date/age standardization, type conversions, array handling,
#' and missing data code translation.
#'

#' Standardize interview_date for NDA submission
#'
#' @description
#' Converts dates to MM/DD/YYYY format. For de-identification (when limited_dataset=FALSE),
#' shifts dates to MM/01/YYYY (first of month). Handles multiple input date formats.
#'
#' @param df Data frame
#' @param date_cols Character vector of date column names (default: "interview_date")
#' @param verbose Logical - print detailed output
#' @param limited_dataset Logical - if FALSE, performs date-shifting for de-identification
#' @return Data frame with standardized dates
#' @noRd
standardize_dates <- function(df, date_cols = c("interview_date"), verbose = TRUE, limited_dataset = FALSE) {
  date_summary <- list()
  
  for (col in date_cols) {
    if (col %in% names(df)) {
      tryCatch({
        if(verbose) cat(sprintf("\n\nField: %s", col))
        
        # Store original values
        orig_dates <- head(df[[col]], 3)
        
        # Convert to character first to ensure consistency
        dates <- as.character(df[[col]])
        
        # Remove timezone information
        dates <- gsub("\\s+\\d{2}:\\d{2}:\\d{2}.*$", "", dates)
        
        # Try different date formats with better error handling
        date_formats <- c(
          "%Y-%m-%d",    # 2023-12-31
          "%m/%d/%Y",    # 12/31/2023
          "%Y/%m/%d",    # 2023/12/31
          "%d-%m-%Y",    # 31-12-2023
          "%m-%d-%Y",    # 12-31-2023
          "%Y%m%d"       # 20231231 (no separators)
        )
        
        success <- FALSE
        parsed_dates <- NULL
        
        for (format in date_formats) {
          parsed_dates <- tryCatch({
            temp_dates <- as.Date(dates, format = format)
            if(!all(is.na(temp_dates))) {
              temp_dates
            } else {
              NULL
            }
          }, error = function(e) NULL)
          
          if (!is.null(parsed_dates)) {
            if(verbose) cat(sprintf("\n  Detected format: %s", format))
            
            # Choose output format based on limited_dataset flag
            output_format <- ifelse(limited_dataset, "%m/%d/%Y", "%m/01/%Y")
            
            # For dates that couldn't be parsed, keep the original value
            new_dates <- rep(NA, length(dates))
            valid_mask <- !is.na(parsed_dates)
            
            # Format only the successful dates
            new_dates[valid_mask] <- format(parsed_dates[valid_mask], output_format)
            
            # Keep original values for failures (but note them)
            if(any(!valid_mask) && verbose) {
              cat(sprintf("\n  Warning: %d dates could not be parsed", sum(!valid_mask)))
              cat(sprintf("\n  Sample problematic values: %s",
                         paste(head(dates[!valid_mask]), collapse=", ")))
            }
            
            # Update only the successful conversions
            df[[col]] <- new_dates
            
            success <- TRUE
            
            date_summary[[col]] <- list(
              original_format = format,
              sample_before = orig_dates,
              sample_after = head(df[[col]], 3),
              failed_count = sum(!valid_mask)
            )
            break
          }
        }
        
        if(!success && verbose) {
          cat(sprintf("\n  Warning: Could not determine date format"))
          cat(sprintf("\n  Sample values: %s", paste(head(dates), collapse=", ")))
        }
        
      }, error = function(e) {
        if(verbose) {
          cat(sprintf("\n\nError processing dates in %s:", col))
          cat(sprintf("\n  %s", e$message))
        }
      })
    }
  }
  
  if(verbose && length(date_summary) > 0) {
    if(limited_dataset == FALSE) message("\n\nDe-identifying interview_date using date-shifting...")
    cat("Date standardization summary:")
    for(field in names(date_summary)) {
      cat(sprintf("\n- %s", field))
      cat(sprintf("\n  Before: %s", paste(date_summary[[field]]$sample_before, collapse=", ")))
      cat(sprintf("\n  After:  %s", paste(date_summary[[field]]$sample_after, collapse=", ")))
      if(date_summary[[field]]$failed_count > 0) {
        cat(sprintf("\n  Failed: %d dates could not be parsed", date_summary[[field]]$failed_count))
      }
    }
    cat("\n")
  }
  
  return(df)
}

#' Standardize interview_age for NDA submission
#'
#' @description
#' Caps age at 1068 months (89 years) for de-identification when limited_dataset=FALSE.
#' NDA requires age-capping for privacy protection per HIPAA Safe Harbor standard.
#'
#' @param df Data frame
#' @param verbose Logical - print detailed output
#' @param limited_dataset Logical - if FALSE, performs age-capping
#' @return List with two elements: df (modified dataframe) and stats (age statistics)
#' @noRd
standardize_age <- function(df, verbose = TRUE, limited_dataset = FALSE) {
  # Initialize statistics
  age_stats <- list(
    values_capped = 0,
    non_na_count = 0,
    all_na = TRUE,
    max_age = NA
  )
  
  if ("interview_age" %in% names(df)) {
    # Convert to numeric
    df$interview_age <- as.numeric(df$interview_age)
    
    # Check if all values are NA
    age_stats$all_na <- all(is.na(df$interview_age))
    age_stats$non_na_count <- sum(!is.na(df$interview_age))
    
    if (!age_stats$all_na) {
      age_stats$max_age <- max(df$interview_age, na.rm = TRUE)
      
      if (!limited_dataset) {
        # Count values that will be capped
        age_stats$values_capped <- sum(df$interview_age > 1068, na.rm = TRUE)
        
        # Store original for verbose output
        if (verbose) {
          orig_age_stats <- summary(df$interview_age)
        }
        
        # Apply age-capping (cap at 1068 months = 89 years)
        df$interview_age <- pmin(df$interview_age, 1068)
        
        # Verbose output
        if (verbose) {
          cat("\nAge standardization summary:")
          cat("\nBefore:", capture.output(orig_age_stats))
          cat("\nAfter:", capture.output(summary(df$interview_age)))
          if (age_stats$values_capped > 0) {
            cat(sprintf("\nNumber of values capped at 1068 months: %d", age_stats$values_capped))
          } else {
            cat("\nNo values needed capping (all were <= 1068 months)")
          }
          cat("\n")
        }
      }
    }
  }
  
  # Return both dataframe and statistics
  return(list(
    df = df,
    stats = age_stats
  ))
}

#' Standardize handedness values to NDA format
#'
#' @description
#' Converts handedness values to standard codes: L (left), R (right), B (both/ambidextrous)
#'
#' @param value Character vector of handedness values
#' @return Character vector with standardized values
#' @noRd
standardize_handedness <- function(value) {
  # Create mapping for handedness terms
  handedness_map <- c(
    "left" = "L",
    "l" = "L",
    "right" = "R",
    "r" = "R",
    "both" = "B",
    "ambidextrous" = "B"
  )
  
  # Convert to lowercase for consistent matching
  value <- tolower(value)
  
  # Map values using the handedness_map
  mapped_values <- handedness_map[value]
  mapped_values[is.na(mapped_values)] <- value[is.na(mapped_values)]
  
  return(mapped_values)
}

#' Standardize boolean values to 0/1 format
#'
#' @description
#' Converts boolean/text values (TRUE/FALSE, yes/no, etc.) to numeric 0/1 format
#' required by NDA for binary fields.
#'
#' @param value Character or logical vector
#' @return Character vector with "0" or "1" values
#' @noRd
standardize_binary <- function(value) {
  # Create mapping for boolean to numeric terms (including case variations)
  binary_map <- c(
    "true" = "1",
    "false" = "0",
    "t" = "1",
    "f" = "0",
    "TRUE" = "1",
    "FALSE" = "0",
    "True" = "1",
    "False" = "0",
    "yes" = "1",
    "no" = "0",
    "y" = "1",
    "n" = "0"
  )
  
  # Convert value to character without changing case
  value <- as.character(value)
  
  # Keep track of which values were NA originally
  was_na <- is.na(value)
  
  # Map values using the binary_map (exact match)
  mapped_values <- binary_map[value]
  
  # For any unmatched values, try lowercase matching
  still_na <- is.na(mapped_values)
  if(any(still_na) && !all(was_na[still_na])) {
    mapped_values[still_na] <- binary_map[tolower(value[still_na])]
  }
  
  # Keep original values for any remaining unmatched
  mapped_values[is.na(mapped_values) & !was_na] <- value[is.na(mapped_values) & !was_na]
  
  # Restore original NAs
  mapped_values[was_na] <- NA
  
  # Count and report transformations
  n_transformed <- sum(value != mapped_values, na.rm = TRUE)
  if (n_transformed > 0) {
    cat(sprintf("\nTransformed %d boolean values to 0/1 format\n", n_transformed))
  }
  
  return(mapped_values)
}

#' Convert logical columns to character
#'
#' @description
#' Helper function to convert all logical columns to character before validation.
#' Prevents type mismatch errors during NDA validation.
#'
#' @param df Data frame
#' @param verbose Logical - print detailed output
#' @return Data frame with logical columns converted to character
#' @noRd
convert_logical_to_character <- function(df, verbose = FALSE) {
  for(col in names(df)) {
    if(is.logical(df[[col]])) {
      if(verbose) cat(sprintf("\nConverting logical column %s to character", col))
      df[[col]] <- as.character(df[[col]])
    }
  }
  return(df)
}

#' Convert problematic column types to character
#'
#' @description
#' Force complex data types (POSIXct, Date, lists) to character to prevent
#' errors during validation. This is a safety function called early in validation.
#'
#' @param df Data frame
#' @param measure_name Name of measure (for error messages)
#' @param verbose Logical - print detailed output
#' @return Data frame with problematic columns converted
#' @noRd
convert_problematic_column_types <- function(df, measure_name, verbose = FALSE) {
  # Force all complex/problematic data types to character immediately
  for (col in names(df)) {
    tryCatch({
      # Convert POSIXct and other complex classes to character
      if (inherits(df[[col]], "POSIXt") ||
          inherits(df[[col]], "Date") ||
          length(class(df[[col]])) > 1) {
        
        if(verbose) message(sprintf("Column '%s' has a complex class structure. Converting to character.", col))
        df[[col]] <- as.character(df[[col]])
      }
      
      # Test if column is accessible
      dummy <- df[[col]][1]
    }, error = function(e) {
      # If any error, convert to character
      if(verbose) message(sprintf("Column '%s' has an unusable type. Converting to character.", col))
      
      # Try three different approaches to fix problematic columns
      tryCatch({
        df[[col]] <- as.character(df[[col]])
      }, error = function(e2) {
        tryCatch({
          df[[col]] <- as.character(unlist(df[[col]]))
        }, error = function(e3) {
          # Last resort - replace with NAs
          if(verbose) message(sprintf("Could not convert column '%s'. Replacing with NAs.", col))
          df[[col]] <- rep(NA, nrow(df))
        })
      })
    })
  }
  
  return(df)
}

#' Standardize column names
#'
#' @description
#' Standardizes column names by replacing hyphens with underscores and
#' ensuring consistent prefix_number format.
#'
#' @param df Data frame
#' @param structure_name NDA structure name
#' @param verbose Logical - print changes
#' @return Data frame with standardized column names
#' @noRd
standardize_column_names <- function(df, structure_name, verbose = FALSE) {
  if(verbose) cat("\nStandardizing column names...")
  
  # Get structure short name by taking last 2 digits of structure_name
  prefix <- substr(structure_name, nchar(structure_name) - 1, nchar(structure_name))
  
  # Create name mapping function
  standardize_name <- function(name) {
    # Replace hyphens with underscores (but don't lowercase)
    name <- gsub("-", "_", name)
    # Handle prefix if present
    if (grepl(paste0("^", prefix, "[_-]?\\d+$"), name, ignore.case = TRUE)) {
      # Ensure consistent underscore between prefix and number
      name <- gsub(paste0("^(", prefix, ")[_-]?(\\d+)$"), "\\1_\\2", name, ignore.case = TRUE)
    }
    return(name)
  }
  
  # Standardize column names
  old_names <- names(df)
  new_names <- sapply(old_names, standardize_name)
  
  # Report changes
  changed <- old_names != new_names
  if (any(changed) && verbose) {
    cat("\n\nColumn name changes:")
    for (i in which(changed)) {
      cat(sprintf("\n  %s -> %s", old_names[i], new_names[i]))
    }
    
    # Add summary
    cat(sprintf("\n\nSummary: %d names standardized\n", sum(changed)))
  }
  
  # Apply new names
  names(df) <- new_names
  return(df)
}

#' Generic function for field standardization
#'
#' @description
#' Standardizes common field names (e.g., "index" -> "trial" for behavioral tasks)
#'
#' @param df Data frame
#' @param measure_name Name of measure
#' @param verbose Logical - print transformations
#' @return Data frame with standardized field names
#' @noRd
standardize_field_names <- function(df, measure_name, verbose = FALSE) {
  if(verbose) cat("\nStandardizing common field names...")
  
  # Track all transformations for summary
  transformations <- list()
  
  # Handle index -> trial conversion
  if ("index" %in% names(df)) {
    if(verbose) cat("\n\nProcessing 'index' to 'trial' conversion...")
    
    # Store original state for summary
    orig_values <- head(df$index, 3)
    
    # Convert to numeric if not already
    df$index <- as.numeric(df$index)
    
    # Create trial column
    df$trial <- df$index
    
    # Set non-positive values to NA
    df$trial[df$index <= 0] <- NA
    
    # Count transformations
    total_rows <- length(df$index)
    positive_rows <- sum(df$index > 0, na.rm = TRUE)
    
    # Store transformation summary
    transformations[["index_to_trial"]] <- list(
      from = "index",
      to = "trial",
      total = total_rows,
      valid = positive_rows,
      sample_before = orig_values,
      sample_after = head(df$trial, 3)
    )
    
    if(verbose) {
      cat(sprintf("\n  Total rows: %d", total_rows))
      cat(sprintf("\n  Valid rows: %d", positive_rows))
      cat("\n  Sample values:")
      cat(sprintf("\n    Before: %s", paste(orig_values, collapse=", ")))
      cat(sprintf("\n    After:  %s", paste(head(df$trial, 3), collapse=", ")))
    }
    
    # Remove original column
    df$index <- NULL
  }
  
  # Print summary if any transformations occurred
  if(verbose && length(transformations) > 0) {
    cat("\n\nField standardization summary:")
    for(transform_name in names(transformations)) {
      transform <- transformations[[transform_name]]
      cat(sprintf("\n- %s -> %s", transform$from, transform$to))
      cat(sprintf("\n  Processed %d rows (%d valid)",
                 transform$total, transform$valid))
    }
    cat("\n")
  }

  return(df)
}

#' Recode user-defined missing data codes to the NDA-expected special codes
#'
#' @description
#' For each column in \code{df}, looks up the NDA structure's \code{valueRange}
#' for that field, extracts the special (non-range) codes the NDA expects, and
#' replaces any study-specific missing data codes (from config.yml) with those
#' NDA-defined codes. This runs per-field so the target code matches exactly
#' what the NDA structure defines for each measure (e.g., -9 from "1::4;-9").
#'
#' @param df Data frame
#' @param config List - config object from validate_config()
#' @param nda_structure List - NDA structure with a \code{dataElements} data frame
#' @param verbose Logical - print replacement summary
#' @return Data frame with user missing codes replaced by NDA special codes
#' @noRd
recode_missing_data_codes <- function(df, config, nda_structure, verbose = TRUE) {
  if (is.null(config) || is.null(config$missing_data_codes)) return(df)
  if (is.null(nda_structure) || !"dataElements" %in% names(nda_structure)) return(df)

  # Flatten all user-defined codes across all categories into a numeric vector
  all_user_codes <- unique(suppressWarnings(as.numeric(unlist(config$missing_data_codes))))
  all_user_codes <- all_user_codes[!is.na(all_user_codes)]
  if (length(all_user_codes) == 0) return(df)

  elements <- nda_structure$dataElements
  total_replacements <- 0L
  affected_cols <- character(0)

  for (col in names(df)) {
    # Only recode numeric/integer columns that actually contain user missing codes
    if (!is.numeric(df[[col]]) && !is.integer(df[[col]])) next
    user_hits <- !is.na(df[[col]]) & df[[col]] %in% all_user_codes
    if (!any(user_hits)) next

    # Find the NDA field definition for this column
    field_idx <- which(elements$name == col)
    if (length(field_idx) == 0) next
    range_str <- elements$valueRange[field_idx[1]]
    if (is.null(range_str) || is.na(range_str) || range_str == "") next

    # Extract NDA special codes: values listed after ';' that fall outside the
    # primary numeric range (e.g., from "1::4;-9" extract -9)
    nda_special_codes <- .extract_nda_special_codes(range_str)
    if (length(nda_special_codes) == 0) next

    # Use the most-negative special code as the replacement target
    # (most negative is typically the generic "missing / not known" code)
    target_code <- nda_special_codes[which.min(nda_special_codes)]

    n <- sum(user_hits)
    df[[col]][user_hits] <- target_code
    total_replacements <- total_replacements + n
    affected_cols <- c(affected_cols, col)
  }

  if (verbose && total_replacements > 0) {
    message(sprintf(
      "Recoded %d missing value(s) in %d field(s) to NDA-defined special codes: %s",
      total_replacements, length(affected_cols),
      paste(affected_cols, collapse = ", ")
    ))
  }

  return(df)
}

#' Extract NDA special (non-range) codes from a valueRange string
#'
#' @description
#' Given a valueRange like "1::4;-9" or "0::3;-8;-9", returns the numeric
#' codes that are listed explicitly (after ";") and fall outside the primary
#' contiguous range. These are the NDA-defined special/missing codes.
#'
#' @param range_str valueRange string from an NDA data element
#' @return Numeric vector of special codes outside the primary range
#' @noRd
.extract_nda_special_codes <- function(range_str) {
  if (!grepl(";", range_str)) return(numeric(0))
  parts <- trimws(strsplit(range_str, ";")[[1]])
  range_parts   <- parts[grepl("::", parts)]
  special_parts <- suppressWarnings(as.numeric(parts[!grepl("::", parts)]))
  special_parts <- special_parts[!is.na(special_parts)]

  # Exclude specials that fall inside the primary numeric range
  if (length(range_parts) > 0) {
    bounds <- suppressWarnings(as.numeric(strsplit(range_parts[1], "::")[[1]]))
    if (length(bounds) == 2 && !any(is.na(bounds))) {
      special_parts <- special_parts[special_parts < bounds[1] | special_parts > bounds[2]]
    }
  }
  special_parts
}

#' Deprecated functions in wizaRdry
#'
#' These functions are deprecated and may be removed in a future release.
#' Prefer the suggested replacements.
#'
#' @name wizaRdry-deprecated
#' @title Deprecated functions in wizaRdry
#' @section Deprecated functions:
#' \describe{
#'   \item{\code{createCsv(...)}}{Use \code{to.csv(...)} instead.}
#'   \item{\code{createRds(...)}}{Use \code{to.rds(...)} instead.}
#'   \item{\code{createSpss(...)}}{Use \code{to.sav(...)} instead.}
#'   \item{\code{dataFilter(...)}}{Use \code{sift(...)} instead.}
#'   \item{\code{dataMerge(...)}}{Use \code{meld(...)} instead.}
#'   \item{\code{dataRequest(...)}}{Use \code{clean(...)} instead.}
#'   \item{\code{getRedcap(...)}}{Use \code{redcap(...)} instead.}
#'   \item{\code{getSurvey(...)}}{Use \code{qualtrics(...)} instead.}
#'   \item{\code{getTask(...)}}{Use \code{mongo(...)} instead.}
#'   \item{\code{ndaRequest(...)}}{Use \code{nda(...)} instead.}
#' }
#' @seealso help("Deprecated")
NULL

#' NDA Type Safety Classes
#'
#' Complete R6 class system for type-safe NDA data structure management.
#' All Excel columns are represented by typed R6 classes with validation.
#'
#' @name NdaClasses
#' @keywords internal
NULL

# Get config environment for validation settings
get_validation_config <- function() {
  config_file <- Sys.getenv("WIZARDRY_CONFIG", "config.yml")
  if (file.exists(config_file)) {
    tryCatch({
      config <- yaml::read_yaml(config_file)
      return(list(
        strict_validation = config$validation$strict %||% FALSE,
        auto_fix = config$validation$auto_fix %||% TRUE
      ))
    }, error = function(e) {
      return(list(strict_validation = FALSE, auto_fix = TRUE))
    })
  }
  list(strict_validation = FALSE, auto_fix = TRUE)
}

# =============================================================================
# ELEMENT NAME CLASS
# =============================================================================

#' Element Name Management
#'
#' Validates and manages NDA element names
#' 
#' @field name Character - the element name
#' @field is_valid Logical - passes NDA naming rules
#' @field validation_errors Character vector - validation issues
#' 
#' @keywords internal
#' @noRd
ElementName <- R6::R6Class("ElementName",
  public = list(
    value = NULL,  # Alias for name (used by NdaDataStructure)
    name = NULL,
    is_valid = NULL,
    validation_errors = NULL,
    
    initialize = function(name = "") {
      self$name <- name
      self$value <- name  # Keep value in sync with name
      self$validation_errors <- character(0)
      self$validate()
    },
    
    validate = function() {
      config <- get_validation_config()
      self$validation_errors <- character(0)
      
      if (is.null(self$name) || !nzchar(self$name)) {
        self$validation_errors <- c(self$validation_errors, "Name cannot be empty")
        self$is_valid <- FALSE
        
        if (config$strict_validation) {
          stop(sprintf("Invalid element name: %s", paste(self$validation_errors, collapse = "; ")))
        }
        return(invisible(self))
      }
      
      # NDA naming rules
      if (!grepl("^[a-zA-Z]", self$name)) {
        self$validation_errors <- c(self$validation_errors, "Must start with a letter")
      }
      
      if (grepl("[^a-zA-Z0-9_]", self$name)) {
        self$validation_errors <- c(self$validation_errors, 
                                   "Can only contain letters, numbers, and underscores")
      }
      
      if (nchar(self$name) > 50) {
        self$validation_errors <- c(self$validation_errors, 
                                   "Name exceeds 50 character maximum")
      }
      
      self$is_valid <- length(self$validation_errors) == 0
      
      if (!self$is_valid && config$strict_validation) {
        stop(sprintf("Invalid element name '%s': %s", 
                    self$name, 
                    paste(self$validation_errors, collapse = "; ")))
      }
      
      invisible(self)
    },
    
    to_string = function() {
      self$name
    },
    
    matches = function(pattern) {
      grepl(pattern, self$name, ignore.case = TRUE)
    }
  )
)

# =============================================================================
# DATA TYPE CLASS
# =============================================================================

#' Data Type Management
#'
#' Validates and manages NDA data types
#' 
#' @field value Character - the data type value
#' @field valid_types Character vector - allowed NDA types
#' @field is_valid Logical - whether value is a valid NDA type
#' @field validation_errors Character vector - validation issues
#' 
#' @keywords internal
#' @noRd
DataType <- R6::R6Class("DataType",
  public = list(
    value = NULL,
    valid_types = c("String", "Integer", "Float", "Date", "GUID", "Boolean"),
    is_valid = NULL,
    validation_errors = NULL,
    
    initialize = function(type = "String") {
      self$value <- type
      self$validation_errors <- character(0)
      self$validate()
    },
    
    validate = function() {
      config <- get_validation_config()
      self$validation_errors <- character(0)
      
      if (is.null(self$value) || !nzchar(self$value)) {
        self$validation_errors <- c(self$validation_errors, "Type cannot be empty")
        self$is_valid <- FALSE
        
        if (config$strict_validation) {
          stop(sprintf("Invalid data type: %s", paste(self$validation_errors, collapse = "; ")))
        }
        return(invisible(self))
      }
      
      if (!self$value %in% self$valid_types) {
        self$validation_errors <- c(self$validation_errors, 
                                   sprintf("Type '%s' not in valid types: %s", 
                                          self$value, 
                                          paste(self$valid_types, collapse = ", ")))
        self$is_valid <- FALSE
        
        if (config$strict_validation) {
          stop(sprintf("Invalid data type: %s", paste(self$validation_errors, collapse = "; ")))
        }
      } else {
        self$is_valid <- TRUE
      }
      
      invisible(self)
    },
    
    to_string = function() {
      self$value
    },
    
    is_string = function() {
      self$value == "String"
    },
    
    is_numeric = function() {
      self$value %in% c("Integer", "Float")
    },
    
    is_date = function() {
      self$value == "Date"
    },
    
    is_guid = function() {
      self$value == "GUID"
    },
    
    is_boolean = function() {
      self$value == "Boolean"
    },
    
    requires_size = function() {
      self$value == "String"
    }
  )
)

# ==============================================================================
# SHIFT DETECTION & MERGE FUNCTIONS
# ==============================================================================

#' Detect Value Range Shift
#'
#' Detects if user's value range is a shifted version of NDA's range.
#' Example: NDA has 0=Not at all, 4=Very much (0::4)
#'          User has 1=Not at all, 5=Very much (1::5)
#'          → Shifted by +1
#'
#' @param nda_range ValueRange object or string for NDA range
#' @param user_range ValueRange object or string for user's range
#' @return List with:
#'   - is_shifted: TRUE/FALSE
#'   - offset: Integer offset (e.g., +1, -2)
#'   - confidence: "high", "medium", "low"
#'   - evidence: List of evidence supporting detection
#'
#' @keywords internal
#' @noRd
detect_value_range_shift <- function(nda_range, user_range) {
  
  # Convert to ValueRange objects if needed
  if (is.character(nda_range)) {
    nda_range <- ValueRange$new(nda_range)
  }
  if (is.character(user_range)) {
    user_range <- ValueRange$new(user_range)
  }
  
  # Initialize result
  result <- list(
    is_shifted = FALSE,
    offset = 0L,
    confidence = "none",
    evidence = list()
  )
  
  # Can't detect shift if either range is empty
  if (nda_range$is_empty() || user_range$is_empty()) {
    result$evidence$reason <- "One or both ranges are empty"
    return(result)
  }
  
  # Extract numeric codes
  nda_codes <- nda_range$get_numeric_codes()
  user_codes <- user_range$get_numeric_codes()
  
  # Need numeric codes from both ranges
  if (length(nda_codes) == 0 || length(user_codes) == 0) {
    result$evidence$reason <- "Non-numeric codes in one or both ranges"
    return(result)
  }
  
  # Must have same number of values to be a shift
  if (length(nda_codes) != length(user_codes)) {
    result$evidence$reason <- sprintf(
      "Different number of codes (NDA: %d, User: %d)",
      length(nda_codes), length(user_codes)
    )
    return(result)
  }
  
  # Sort codes for comparison
  nda_sorted <- sort(nda_codes)
  user_sorted <- sort(user_codes)
  
  # Calculate offsets between corresponding positions
  offsets <- user_sorted - nda_sorted
  
  # Check if offset is consistent (all the same)
  unique_offsets <- unique(offsets)
  
  if (length(unique_offsets) != 1) {
    result$evidence$reason <- "Inconsistent offsets between codes"
    result$evidence$offsets <- offsets
    return(result)
  }
  
  offset <- unique_offsets[1]
  
  # Zero offset means ranges are identical
  if (offset == 0) {
    result$evidence$reason <- "Ranges are identical (no shift)"
    return(result)
  }
  
  # We detected a consistent shift!
  result$is_shifted <- TRUE
  result$offset <- offset
  result$evidence$nda_codes <- nda_sorted
  result$evidence$user_codes <- user_sorted
  result$evidence$offset_pattern <- offsets
  
  # Determine confidence based on labels (if available)
  nda_labels <- nda_range$get_labels()
  user_labels <- user_range$get_labels()
  
  if (length(nda_labels) > 0 && length(user_labels) > 0) {
    # Compare labels for matching codes
    label_matches <- 0
    label_comparisons <- 0
    
    for (i in seq_along(nda_sorted)) {
      nda_code <- as.character(nda_sorted[i])
      user_code <- as.character(user_sorted[i])
      
      if (nda_code %in% names(nda_labels) && user_code %in% names(user_labels)) {
        label_comparisons <- label_comparisons + 1
        
        nda_label <- tolower(trimws(nda_labels[[nda_code]]))
        user_label <- tolower(trimws(user_labels[[user_code]]))
        
        # Check if labels match or are very similar
        if (nda_label == user_label) {
          label_matches <- label_matches + 1
        } else {
          # Calculate simple similarity (for partial credit)
          max_len <- max(nchar(nda_label), nchar(user_label))
          common_chars <- sum(utf8ToInt(nda_label) %in% utf8ToInt(user_label))
          similarity <- common_chars / max_len
          if (similarity > 0.7) {
            label_matches <- label_matches + 0.5
          }
        }
      }
    }
    
    if (label_comparisons > 0) {
      match_rate <- label_matches / label_comparisons
      result$evidence$label_match_rate <- match_rate
      
      if (match_rate >= 0.8) {
        result$confidence <- "high"
      } else if (match_rate >= 0.5) {
        result$confidence <- "medium"
      } else {
        result$confidence <- "low"
      }
    } else {
      result$confidence <- "medium"
      result$evidence$label_match_rate <- NA
    }
  } else {
    # No labels to compare, medium confidence based on numeric pattern alone
    result$confidence <- "medium"
    result$evidence$label_match_rate <- NA
  }
  
  return(result)
}

#' Merge Value Ranges
#'
#' Intelligently merges value ranges from NDA, REDCap, and data with conflict resolution.
#'
#' Priority order:
#' 1. No NDA range exists → use REDCap or data-inferred range
#' 2. NDA range exists, no user range → use NDA as-is
#' 3. Shift detected → normalize to NDA format, document shift
#' 4. New codes detected (e.g., 999=Undefined) → append to NDA range
#' 5. Conflict with no resolution → use NDA, warn user
#'
#' @param nda_range ValueRange object or string for NDA range (may be NULL/empty)
#' @param redcap_range ValueRange object or string for REDCap range (may be NULL/empty)
#' @param data_range ValueRange object or string for data-inferred range (may be NULL/empty)
#' @param missing_codes Character vector of codes to treat as missing data
#' @return List with:
#'   - merged_range: ValueRange object (final merged result)
#'   - strategy_used: Character describing which merge strategy was applied
#'   - modifications: List of modifications made
#'   - warnings: Character vector of warnings for user
#'
#' @keywords internal
#' @noRd
merge_value_ranges <- function(nda_range = NULL, 
                                redcap_range = NULL, 
                                data_range = NULL,
                                missing_codes = character(0)) {
  
  # Convert to ValueRange objects if needed
  if (!is.null(nda_range) && is.character(nda_range)) {
    nda_range <- ValueRange$new(nda_range)
  }
  if (!is.null(redcap_range) && is.character(redcap_range)) {
    redcap_range <- ValueRange$new(redcap_range)
  }
  if (!is.null(data_range) && is.character(data_range)) {
    data_range <- ValueRange$new(data_range)
  }
  
  # Initialize result
  result <- list(
    merged_range = NULL,
    strategy_used = "unknown",
    modifications = list(),
    warnings = character(0)
  )
  
  # CASE 1: No NDA range exists → use REDCap or data-inferred range
  if (is.null(nda_range) || nda_range$is_empty()) {
    if (!is.null(redcap_range) && !redcap_range$is_empty()) {
      result$merged_range <- redcap_range
      result$strategy_used <- "no_nda_use_redcap"
      result$modifications <- list("Used REDCap range (no NDA range available)")
    } else if (!is.null(data_range) && !data_range$is_empty()) {
      result$merged_range <- data_range
      result$strategy_used <- "no_nda_use_data"
      result$modifications <- list("Used data-inferred range (no NDA range available)")
    } else {
      # No range available from any source
      result$merged_range <- ValueRange$new("")
      result$strategy_used <- "no_range_available"
      result$warnings <- c(result$warnings, "No value range available from any source")
    }
    return(result)
  }
  
  # CASE 2: NDA range exists, no user range → use NDA as-is
  user_range <- if (!is.null(redcap_range) && !redcap_range$is_empty()) {
    redcap_range
  } else if (!is.null(data_range) && !data_range$is_empty()) {
    data_range
  } else {
    NULL
  }
  
  if (is.null(user_range)) {
    result$merged_range <- nda_range
    result$strategy_used <- "nda_only"
    return(result)
  }
  
  # CASE 3: Check for shift
  shift_detection <- detect_value_range_shift(nda_range, user_range)
  
  if (shift_detection$is_shifted) {
    # Normalize to NDA format
    result$merged_range <- nda_range$clone()
    result$strategy_used <- "shift_detected_normalized"
    result$modifications <- list(
      sprintf(
        "Shift detected (offset: %+d, confidence: %s). Normalized to NDA format.",
        shift_detection$offset,
        shift_detection$confidence
      )
    )
    result$warnings <- c(
      result$warnings,
      sprintf(
        "User data appears shifted by %+d from NDA range. Data will be normalized.",
        shift_detection$offset
      )
    )
    return(result)
  }
  
  # CASE 4: Check for new codes (e.g., missing data codes)
  nda_codes <- nda_range$get_numeric_codes()
  user_codes <- user_range$get_numeric_codes()
  
  new_codes <- setdiff(user_codes, nda_codes)
  
  if (length(new_codes) > 0) {
    # Check if new codes are missing data codes
    new_codes_char <- as.character(new_codes)
    missing_in_new <- new_codes_char %in% missing_codes
    
    if (any(missing_in_new)) {
      # Append missing codes to NDA range
      result$merged_range <- nda_range$merge_with(user_range)
      result$strategy_used <- "missing_codes_appended"
      result$modifications <- list(
        sprintf(
          "Appended missing data codes to NDA range: %s",
          paste(new_codes_char[missing_in_new], collapse = ", ")
        )
      )
    } else {
      # New codes exist but aren't recognized as missing codes
      result$merged_range <- nda_range$merge_with(user_range)
      result$strategy_used <- "new_codes_appended"
      result$modifications <- list(
        sprintf(
          "Appended new codes to NDA range: %s",
          paste(new_codes_char, collapse = ", ")
        )
      )
      result$warnings <- c(
        result$warnings,
        sprintf(
          "User data contains codes not in NDA range: %s. Appended to range.",
          paste(new_codes_char, collapse = ", ")
        )
      )
    }
    return(result)
  }
  
  # CASE 5: Ranges are compatible or identical
  if (all(user_codes %in% nda_codes)) {
    result$merged_range <- nda_range$clone()
    result$strategy_used <- "user_subset_of_nda"
    result$modifications <- list("User range is subset of NDA range. Using NDA range.")
    return(result)
  }
  
  # CASE 6: Conflict with no clear resolution → use NDA, warn user
  result$merged_range <- nda_range$clone()
  result$strategy_used <- "conflict_use_nda"
  result$warnings <- c(
    result$warnings,
    sprintf(
      "Value range conflict detected. Using NDA range. NDA: %s | User: %s",
      nda_range$to_nda_format(),
      user_range$to_nda_format()
    )
  )
  result$modifications <- list("Conflict detected. Defaulted to NDA range.")
  
  return(result)
}

# =============================================================================
# SIZE CLASS
# =============================================================================

#' Field Size Management
#'
#' Manages field size with type-specific validation
#' 
#' @field size Integer - field size
#' @field data_type DataType - associated data type
#' @field is_valid Logical - valid for this data type
#' 
#' @keywords internal
#' @noRd
Size <- R6::R6Class("Size",
  public = list(
    size = NULL,
    data_type = NULL,
    is_valid = NULL,
    
    initialize = function(size = NA, data_type = NULL) {
      self$data_type <- data_type
      self$set_size(size)
    },
    
    set_size = function(size) {
      if (is.null(size) || is.na(size) || !nzchar(as.character(size))) {
        self$size <- NA_integer_
        # Check if size is required for this data type
        if (!is.null(self$data_type)) {
          # Handle both DataType object and string
          if (inherits(self$data_type, "DataType")) {
            self$is_valid <- !self$data_type$requires_size()
          } else {
            # String data type
            self$is_valid <- (self$data_type != "String")
          }
        } else {
          self$is_valid <- TRUE
        }
        return(invisible(self))
      }
      
      size_num <- suppressWarnings(as.integer(size))
      
      if (is.na(size_num)) {
        self$size <- NA_integer_
        self$is_valid <- FALSE
      } else {
        self$size <- size_num
        self$is_valid <- self$validate_for_type()
      }
      
      invisible(self)
    },
    
    validate_for_type = function() {
      if (is.null(self$data_type)) return(TRUE)
      
      # Get type string (handle both DataType object and string)
      type_str <- if (inherits(self$data_type, "DataType")) {
        self$data_type$value
      } else {
        as.character(self$data_type)
      }
      
      if (type_str == "String") {
        return(!is.na(self$size) && self$size > 0 && self$size <= 4000)
      }
      
      if (type_str %in% c("Integer", "Float", "Date", "GUID", "Boolean")) {
        return(is.na(self$size))
      }
      
      TRUE
    },
    
    to_string = function() {
      if (is.na(self$size)) return("")
      as.character(self$size)
    },
    
    to_integer = function() {
      self$size
    }
  )
)

# =============================================================================
# REQUIREMENT LEVEL CLASS
# =============================================================================

#' Requirement Level Management
#'
#' Manages NDA requirement levels with validation
#' 
#' @field level Character - "Required", "Recommended", "Conditional", "No"
#' @field is_valid Logical - valid NDA requirement level
#' 
#' @keywords internal
#' @noRd
RequirementLevel <- R6::R6Class("RequirementLevel",
  public = list(
    level = NULL,
    is_valid = NULL,
    
    VALID_LEVELS = c("Required", "Recommended", "Conditional", "No"),
    
    initialize = function(level = "No") {
      self$set_level(level)
    },
    
    set_level = function(level) {
      if (is.null(level) || !nzchar(level)) {
        self$level <- "No"
        self$is_valid <- TRUE
      } else if (level %in% self$VALID_LEVELS) {
        self$level <- level
        self$is_valid <- TRUE
      } else {
        self$level <- level
        self$is_valid <- FALSE
        
        config <- get_validation_config()
        if (config$auto_fix) {
          self$level <- "No"
          self$is_valid <- TRUE
        }
      }
      invisible(self)
    },
    
    is_required = function() {
      self$level == "Required"
    },
    
    is_recommended = function() {
      self$level == "Recommended"
    },
    
    is_conditional = function() {
      self$level == "Conditional"
    },
    
    to_string = function() {
      self$level
    }
  )
)

# =============================================================================
# CONDITION CLASS
# =============================================================================

#' Conditional Logic Management
#'
#' Manages conditional requirements (e.g., "Required if sex=F")
#' 
#' @field condition Character - condition expression
#' @field is_parsed Logical - successfully parsed
#' @field dependent_fields Character vector - fields this depends on
#' 
#' @keywords internal
#' @noRd
Condition <- R6::R6Class("Condition",
  public = list(
    condition = NULL,
    is_parsed = NULL,
    dependent_fields = NULL,
    
    initialize = function(condition = "") {
      self$condition <- condition
      self$dependent_fields <- character(0)
      self$parse()
    },
    
    parse = function() {
      if (is.null(self$condition) || !nzchar(self$condition)) {
        self$is_parsed <- TRUE
        return(invisible(self))
      }
      
      field_pattern <- "\\b([a-zA-Z_][a-zA-Z0-9_]*)\\s*[=<>!]"
      matches <- gregexpr(field_pattern, self$condition, perl = TRUE)
      
      if (matches[[1]][1] != -1) {
        match_text <- regmatches(self$condition, matches)[[1]]
        self$dependent_fields <- unique(gsub("\\s*[=<>!].*$", "", match_text))
        self$is_parsed <- TRUE
      } else {
        self$is_parsed <- FALSE
      }
      
      invisible(self)
    },
    
    to_string = function() {
      self$condition %||% ""
    },
    
    is_empty = function() {
      is.null(self$condition) || !nzchar(self$condition)
    },
    
    get_dependencies = function() {
      self$dependent_fields
    }
  )
)

# =============================================================================
# DESCRIPTION CLASS
# =============================================================================

#' Element Description Management
#'
#' Manages field descriptions with length validation
#' 
#' @field text Character - description text
#' @field source Character - where description came from
#' @field is_valid Logical - meets NDA requirements
#' 
#' @keywords internal
#' @noRd
Description <- R6::R6Class("Description",
  public = list(
    text = NULL,
    source = NULL,
    is_valid = NULL,
    
    MAX_LENGTH = 4000,
    
    initialize = function(text = "", source = "unknown") {
      self$text <- text
      self$source <- source
      self$validate()
    },
    
    validate = function() {
      if (is.null(self$text)) {
        self$is_valid <- FALSE
        return(invisible(self))
      }
      
      cleaned <- gsub("<br\\s*/?>", " ", self$text, ignore.case = TRUE)
      cleaned <- gsub("<[^>]+>", "", cleaned)
      cleaned <- trimws(cleaned)
      
      self$text <- cleaned
      self$is_valid <- nzchar(self$text) && nchar(self$text) <= self$MAX_LENGTH
      
      invisible(self)
    },
    
    to_string = function() {
      self$text %||% ""
    },
    
    set_text = function(text, source = NULL) {
      self$text <- text
      if (!is.null(source)) self$source <- source
      self$validate()
      invisible(self)
    },
    
    append = function(additional) {
      if (nzchar(self$text) && nzchar(additional)) {
        self$text <- paste(self$text, additional, sep = "; ")
      } else if (nzchar(additional)) {
        self$text <- additional
      }
      self$validate()
      invisible(self)
    }
  ),
  active = list(
    value = function() self$text %||% ""
  )
)

# =============================================================================
# VALUE RANGE CLASS
# =============================================================================

#' Value Range Type System
#'
#' Represents a structured value range with parsing, comparison, and merging
#' 
#' @field type Character - "categorical", "range", "mixed", "empty"
#' @field codes Character vector - numeric codes: c("0", "1", "2")
#' @field labels Named character vector - c("0"="No", "1"="Yes")
#' @field ranges List of numeric ranges - list(c(0, 10), c(20, 30))
#' @field raw Character - original string
#' @field source Character - "nda", "redcap", "data", "computed"
#' @field is_shifted Logical - TRUE if detected as shifted version
#' @field shift_offset Integer - offset amount if shifted
#' 
#' @keywords internal
#' @noRd
ValueRange <- R6::R6Class("ValueRange",
  public = list(
    type = NULL,
    codes = NULL,
    labels = NULL,
    ranges = NULL,
    raw = NULL,
    source = NULL,
    is_shifted = FALSE,
    shift_offset = 0,
    
    initialize = function(raw_string = "", source = "unknown") {
      self$raw <- raw_string
      self$source <- source
      if (nzchar(raw_string)) {
        self$parse()
      } else {
        self$type <- "empty"
      }
    },
    
    parse = function() {
      if (is.null(self$raw) || !nzchar(self$raw)) {
        self$type <- "empty"
        return(invisible(self))
      }
      
      parts <- trimws(strsplit(self$raw, ";")[[1]])
      parts <- parts[nzchar(parts)]
      
      if (length(parts) == 0) {
        self$type <- "empty"
        return(invisible(self))
      }
      
      has_labels <- any(grepl("=", parts))
      has_ranges <- any(grepl("::", parts, fixed = TRUE))
      
      if (has_labels) {
        self$type <- if (has_ranges) "mixed" else "categorical"
        self$parse_categorical(parts)
      } else if (has_ranges) {
        self$type <- "range"
        self$parse_ranges(parts)
      } else {
        self$type <- "categorical"
        self$codes <- parts
      }
      
      invisible(self)
    },
    
    parse_categorical = function(parts) {
      self$codes <- character(0)
      self$labels <- character(0)
      
      for (part in parts) {
        if (grepl("=", part)) {
          split_part <- strsplit(part, "=")[[1]]
          if (length(split_part) >= 2) {
            code <- trimws(split_part[1])
            label <- trimws(paste(split_part[-1], collapse = "="))
            self$codes <- c(self$codes, code)
            self$labels <- c(self$labels, label)
            names(self$labels)[length(self$labels)] <- code
          }
        } else if (grepl("::", part, fixed = TRUE)) {
          self$parse_ranges(part)
        } else {
          self$codes <- c(self$codes, trimws(part))
        }
      }
    },
    
    parse_ranges = function(parts) {
      self$ranges <- list()
      
      for (part in parts) {
        if (grepl("::", part, fixed = TRUE)) {
          bounds <- strsplit(part, "::", fixed = TRUE)[[1]]
          if (length(bounds) == 2) {
            start <- suppressWarnings(as.numeric(trimws(bounds[1])))
            end <- suppressWarnings(as.numeric(trimws(bounds[2])))
            if (!is.na(start) && !is.na(end)) {
              self$ranges[[length(self$ranges) + 1]] <- c(start, end)
            }
          }
        }
      }
    },
    
    to_nda_format = function() {
      if (self$type == "empty") return("")
      
      if (self$type == "categorical" && !is.null(self$labels) && length(self$labels) > 0) {
        return(paste(self$codes, collapse = ";"))
      } else if (self$type == "range" && !is.null(self$ranges)) {
        range_strings <- sapply(self$ranges, function(r) {
          paste0(as.integer(r[1]), "::", as.integer(r[2]))
        })
        return(paste(range_strings, collapse = ";"))
      } else if (!is.null(self$codes)) {
        return(paste(self$codes, collapse = ";"))
      }
      
      return(self$raw)
    },
    
    to_notes_format = function() {
      if (is.null(self$labels) || length(self$labels) == 0) {
        return("")
      }
      
      pairs <- sapply(seq_along(self$labels), function(i) {
        code <- names(self$labels)[i]
        label <- self$labels[i]
        paste0(code, "=", label)
      })
      
      paste(pairs, collapse = ";")
    },
    
    get_numeric_codes = function() {
      numeric_codes <- suppressWarnings(as.numeric(self$codes))
      numeric_codes[!is.na(numeric_codes)]
    },
    
    contains_code = function(code) {
      code_str <- as.character(code)
      
      if (code_str %in% self$codes) return(TRUE)
      
      if (!is.null(self$ranges) && length(self$ranges) > 0) {
        code_num <- suppressWarnings(as.numeric(code_str))
        if (!is.na(code_num)) {
          for (range in self$ranges) {
            if (code_num >= range[1] && code_num <= range[2]) {
              return(TRUE)
            }
          }
        }
      }
      
      FALSE
    },
    
    merge_with = function(other, strategy = "append_new") {
      if (strategy == "append_new") {
        new_codes <- setdiff(other$codes, self$codes)
        
        if (length(new_codes) > 0) {
          self$codes <- c(self$codes, new_codes)
          
          if (!is.null(other$labels)) {
            for (code in new_codes) {
              if (code %in% names(other$labels)) {
                self$labels <- c(self$labels, other$labels[code])
                names(self$labels)[length(self$labels)] <- code
              }
            }
          }
        }
      } else if (strategy == "override") {
        self$codes <- other$codes
        self$labels <- other$labels
        self$ranges <- other$ranges
        self$type <- other$type
      }
      
      invisible(self)
    }
  )
)

# =============================================================================
# NOTES CLASS
# =============================================================================

#' Notes Management System
#'
#' Manages field notes including code mappings, descriptions, and metadata
#' 
#' @field text Character - main notes text
#' @field code_mappings Named character vector - code=label pairs
#' @field metadata_notes Character vector - additional metadata notes
#' @field source Character - where notes came from
#' 
#' @keywords internal
#' @noRd
Notes <- R6::R6Class("Notes",
  public = list(
    text = NULL,
    code_mappings = NULL,
    metadata_notes = NULL,
    source = NULL,
    
    initialize = function(text = "", source = "unknown") {
      self$text <- text
      self$source <- source
      self$code_mappings <- character(0)
      self$metadata_notes <- character(0)
      
      if (nzchar(text)) {
        self$parse()
      }
    },
    
    parse = function() {
      if (is.null(self$text) || !nzchar(self$text)) return(invisible(self))
      
      parts <- trimws(strsplit(self$text, "[|;]")[[1]])
      
      for (part in parts) {
        if (grepl("=", part)) {
          split_part <- strsplit(part, "=")[[1]]
          if (length(split_part) >= 2) {
            code <- trimws(split_part[1])
            label <- trimws(paste(split_part[-1], collapse = "="))
            self$code_mappings <- c(self$code_mappings, label)
            names(self$code_mappings)[length(self$code_mappings)] <- code
          }
        } else if (nzchar(part)) {
          self$metadata_notes <- c(self$metadata_notes, part)
        }
      }
      
      invisible(self)
    },
    
    add_mapping = function(code, label) {
      self$code_mappings <- c(self$code_mappings, label)
      names(self$code_mappings)[length(self$code_mappings)] <- code
      self$rebuild_text()
      invisible(self)
    },
    
    add_note = function(note) {
      if (nzchar(note) && !(note %in% self$metadata_notes)) {
        self$metadata_notes <- c(self$metadata_notes, note)
        self$rebuild_text()
      }
      invisible(self)
    },
    
    merge_with = function(other) {
      if (!is.null(other$code_mappings) && length(other$code_mappings) > 0) {
        for (code in names(other$code_mappings)) {
          self$code_mappings[code] <- other$code_mappings[code]
        }
      }
      
      if (!is.null(other$metadata_notes) && length(other$metadata_notes) > 0) {
        new_notes <- setdiff(other$metadata_notes, self$metadata_notes)
        self$metadata_notes <- c(self$metadata_notes, new_notes)
      }
      
      self$rebuild_text()
      invisible(self)
    },
    
    rebuild_text = function() {
      parts <- character(0)
      
      if (length(self$code_mappings) > 0) {
        mappings <- sapply(seq_along(self$code_mappings), function(i) {
          code <- names(self$code_mappings)[i]
          label <- self$code_mappings[i]
          paste0(code, "=", label)
        })
        parts <- c(parts, mappings)
      }
      
      if (length(self$metadata_notes) > 0) {
        parts <- c(parts, self$metadata_notes)
      }
      
      self$text <- paste(parts, collapse = ";")
      invisible(self)
    },
    
    to_string = function() {
      self$text %||% ""
    }
  )
)

# =============================================================================
# ALIASES CLASS
# =============================================================================

#' Field Aliases Management
#'
#' Manages field aliases with proper formatting (no c() wrapper)
#' 
#' @field aliases Character vector - list of aliases
#' @field source Character - where aliases came from
#' 
#' @keywords internal
#' @noRd
Aliases <- R6::R6Class("Aliases",
  public = list(
    aliases = NULL,
    source = NULL,
    
    initialize = function(aliases = NULL, source = "unknown") {
      self$source <- source
      self$set_aliases(aliases)
    },
    
    set_aliases = function(aliases) {
      if (is.null(aliases)) {
        self$aliases <- character(0)
        return(invisible(self))
      }
      
      # Handle list from NDA API (THIS FIXES THE c() WRAPPER BUG)
      if (is.list(aliases)) {
        aliases <- unlist(aliases, use.names = FALSE)
      }
      
      # Handle comma-separated string
      if (is.character(aliases) && length(aliases) == 1 && grepl(",", aliases)) {
        aliases <- trimws(strsplit(aliases, ",")[[1]])
      }
      
      # Clean up
      aliases <- as.character(aliases)
      aliases <- aliases[!is.na(aliases) & nzchar(aliases)]
      aliases <- unique(aliases)
      
      self$aliases <- aliases
      invisible(self)
    },
    
    add = function(alias) {
      if (!is.null(alias) && nzchar(alias) && !(alias %in% self$aliases)) {
        self$aliases <- c(self$aliases, alias)
      }
      invisible(self)
    },
    
    to_string = function() {
      if (length(self$aliases) == 0) return("")
      paste(self$aliases, collapse = ",")
    },
    
    to_vector = function() {
      self$aliases
    },
    
    has = function(alias) {
      alias %in% self$aliases
    },
    
    count = function() {
      length(self$aliases)
    }
  )
)

# =============================================================================
# VALIDATION RULES CLASS
# =============================================================================

#' Validation Rules Management
#'
#' Stores and validates field-level validation rules
#' 
#' @field allowed_values Character vector - permitted values
#' @field min_value Numeric - minimum allowed value
#' @field max_value Numeric - maximum allowed value
#' @field pattern Character - regex pattern for validation
#' @field data_type Character - expected data type
#' @field custom_rules List - custom validation functions
#' 
#' @keywords internal
#' @noRd
ValidationRules <- R6::R6Class("ValidationRules",
  public = list(
    allowed_values = NULL,
    min_value = NULL,
    max_value = NULL,
    pattern = NULL,
    data_type = NULL,
    custom_rules = NULL,
    
    initialize = function(min_value = NULL, max_value = NULL, allowed_values = NULL, pattern = NULL, data_type = NULL) {
      self$min_value <- min_value
      self$max_value <- max_value
      # Safety check: convert list to character vector if needed
      self$allowed_values <- if (is.null(allowed_values)) {
        character(0)
      } else if (is.list(allowed_values)) {
        # Handle case where list is passed instead of vector
        unlist(allowed_values, use.names = FALSE)
      } else {
        allowed_values
      }
      self$pattern <- pattern
      self$data_type <- data_type
      self$custom_rules <- list()
    },
    
    set_from_value_range = function(value_range) {
      if (is.null(value_range)) return(invisible(self))
      
      if (!is.null(value_range$codes)) {
        self$allowed_values <- value_range$codes
      }
      
      if (!is.null(value_range$ranges) && length(value_range$ranges) > 0) {
        all_bounds <- unlist(value_range$ranges)
        self$min_value <- min(all_bounds)
        self$max_value <- max(all_bounds)
      }
      
      invisible(self)
    },
    
    validate = function(value) {
      errors <- character(0)
      
      if (length(self$allowed_values) > 0) {
        if (!(as.character(value) %in% self$allowed_values)) {
          errors <- c(errors, sprintf("Value '%s' not in allowed values", value))
        }
      }
      
      if (!is.null(self$min_value) || !is.null(self$max_value)) {
        num_val <- suppressWarnings(as.numeric(value))
        if (!is.na(num_val)) {
          if (!is.null(self$min_value) && num_val < self$min_value) {
            errors <- c(errors, sprintf("Value %s below minimum %s", num_val, self$min_value))
          }
          if (!is.null(self$max_value) && num_val > self$max_value) {
            errors <- c(errors, sprintf("Value %s above maximum %s", num_val, self$max_value))
          }
        }
      }
      
      list(valid = length(errors) == 0, errors = errors)
    },
    
    to_list = function() {
      list(
        allowed_values = self$allowed_values,
        min_value = self$min_value,
        max_value = self$max_value,
        data_type = self$data_type
      )
    }
  )
)

# =============================================================================
# MISSING INFO CLASS
# =============================================================================

#' Missing Data Information
#'
#' Tracks missing data statistics for a field
#' 
#' @field missing_count Integer - number of missing values
#' @field total_count Integer - total values
#' @field missing_percentage Numeric - percentage missing
#' @field missing_codes Character vector - codes treated as missing
#' 
#' @keywords internal
#' @noRd
MissingInfo <- R6::R6Class("MissingInfo",
  public = list(
    missing_count = NULL,
    total_count = NULL,
    missing_percentage = NULL,
    missing_codes = NULL,
    
    initialize = function(missing_count = 0, total_count = 0, missing_codes = NULL) {
      self$missing_count <- missing_count
      self$total_count <- total_count
      self$missing_codes <- missing_codes
      self$calculate_percentage()
    },
    
    calculate_percentage = function() {
      if (self$total_count > 0) {
        self$missing_percentage <- round((self$missing_count / self$total_count) * 100, 1)
      } else {
        self$missing_percentage <- 0
      }
      invisible(self)
    },
    
    to_list = function() {
      list(
        missing_count = self$missing_count,
        missing_percentage = self$missing_percentage,
        total_count = self$total_count,
        missing_codes = self$missing_codes
      )
    }
  )
)

# =============================================================================
# SOURCE METADATA CLASS
# =============================================================================

#' Source Metadata Tracking
#'
#' Tracks where field metadata came from for provenance
#' 
#' @field primary_source Character - "nda", "redcap", "qualtrics", "data", "computed"
#' @field nda_matched Logical - found in NDA dictionary
#' @field redcap_matched Logical - found in REDCap dictionary
#' @field is_modified Logical - modified from original source
#' @field modifications Character vector - list of modifications made
#' 
#' @keywords internal
#' @noRd
SourceMetadata <- R6::R6Class("SourceMetadata",
  public = list(
    primary_source = NULL,
    nda_matched = NULL,
    redcap_matched = NULL,
    is_modified = NULL,
    modifications = NULL,
    
    initialize = function(primary_source = "unknown") {
      self$primary_source <- primary_source
      self$nda_matched <- FALSE
      self$redcap_matched <- FALSE
      self$is_modified <- FALSE
      self$modifications <- character(0)
    },
    
    mark_nda_match = function() {
      self$nda_matched <- TRUE
      if (self$primary_source == "unknown") {
        self$primary_source <- "nda"
      }
      invisible(self)
    },
    
    mark_redcap_match = function() {
      self$redcap_matched <- TRUE
      if (self$primary_source == "unknown") {
        self$primary_source <- "redcap"
      }
      invisible(self)
    },
    
    add_modification = function(modification) {
      self$is_modified <- TRUE
      self$modifications <- c(self$modifications, modification)
      invisible(self)
    },
    
    to_string = function() {
      parts <- c(self$primary_source)
      if (self$nda_matched) parts <- c(parts, "nda_matched")
      if (self$redcap_matched) parts <- c(parts, "redcap_matched")
      if (self$is_modified) {
        parts <- c(parts, sprintf("modified(%d)", length(self$modifications)))
      }
      paste(parts, collapse = "|")
    }
  )
)

#' Create .sav SPSS file from a data frame
#'
#' This function takes a R data frame and writes it to an SPSS file
#' using the Haven package. The resulting file will be stored in the
#' "tmp" directory with a default name derived from the data frame variable name,
#' but can be customized if desired. The function will prompt for confirmation
#' before creating the file, with an option to remember the user's preference for future calls.
#'
#' @param df Data frame to be exported to SPSS format.
#' @param df_name Optional; custom file name for the saved SPSS file. If not provided,
#'   the name of the data frame variable will be used. The ".sav" extension will
#'   be appended automatically.
#' @param path Character string specifying the directory path where the "tmp" folder
#'   and SPSS file should be created. Defaults to the current working directory.
#' @param skip_prompt Logical. If TRUE (default), skips the confirmation prompt. If FALSE,
#'   will prompt for confirmation unless the user has previously chosen to remember their preference.
#'
#' @return Invisible TRUE if successful. Writes an SPSS file to the designated path and prints a message indicating
#'   the file's location.
#'
#' @examples
#' \dontrun{
#' # Create a sample data frame
#' sample_df <- data.frame(
#'   id = 1:3,
#'   score = c(85, 92, 78),
#'   group = c("A", "B", "A")
#' )
#'
#' # Basic usage with prompt
#' to.sav(sample_df)
#'
#' # Custom filename
#' to.sav(sample_df, "participants_data")
#'
#' # Skip the confirmation prompt
#' to.sav(sample_df, skip_prompt = TRUE)
#'
#' # Save in a different directory
#' to.sav(sample_df, path = "path/to/project")
#' }
#'
#' @import haven
#' @export
to.sav <- function(df, df_name = NULL, path = ".", skip_prompt = TRUE) { # skip_prompt set to TRUE so user can flag FALSE if needed
  # Get the name of the data frame for display
  df_display_name <- if (!is.null(df_name)) {
    df_name
  } else {
    deparse(substitute(df))
  }

  # Check for user preferences file
  user_prefs_file <- file.path(path, ".wizaRdry_prefs")
  user_prefs <- list(shown_tree = FALSE, auto_create = FALSE, auto_clean = FALSE, auto_nda = FALSE,
                     auto_nda_template = FALSE, auto_csv = FALSE, auto_rds = FALSE, auto_sav = FALSE)

  if (file.exists(user_prefs_file)) {
    tryCatch({
      user_prefs <- readRDS(user_prefs_file)
      # Add the auto_sav field if it doesn't exist
      if (is.null(user_prefs$auto_sav)) {
        user_prefs$auto_sav <- FALSE
      }
    }, error = function(e) {
      # If file exists but can't be read, create a new one
      user_prefs <- list(shown_tree = FALSE, auto_create = FALSE, auto_clean = FALSE, auto_nda = FALSE,
                         auto_nda_template = FALSE, auto_csv = FALSE, auto_rds = FALSE, auto_sav = FALSE)
    })
  }

  # If skip_prompt is TRUE or user has previously set auto_sav to TRUE, bypass the prompt
  if (!skip_prompt | !user_prefs$auto_sav) {
    response <- readline(prompt = sprintf("Would you like to create the SPSS data file for %s now? y/n ",
                                          df_display_name))

    while (!tolower(response) %in% c("y", "n")) {
      response <- readline(prompt = "Please enter either y or n: ")
    }

    # If response is 'y', save their preference as yes for later and update auto_sav to TRUE
    if (tolower(response) == "y") {
      user_prefs$auto_sav <- TRUE
      saveRDS(user_prefs, user_prefs_file)
    }

    if (tolower(response) == "n") {
      message(".sav creation cancelled.")
      invokeRestart("abort")  # This exits without the "Error:" prefix
    }
  }


  # Use df_name if provided, otherwise derive from df variable name
  filename <- if (!is.null(df_name)) {
    df_name
  } else {
    deparse(substitute(df))
  }

  # Create tmp directory if it doesn't exist
  tmp_path <- file.path(path, "tmp")
  if (!dir.exists(tmp_path)) {
    dir.create(tmp_path)
  }

  # Construct the file path
  file_path <- file.path(tmp_path, paste0(filename, '.sav'))

  # Write the data frame to an SPSS file
  write_sav(df, file_path)

  # Notify user of file creation
  message(paste0("Extract created at ", file_path, "\n"))

  return(invisible(TRUE))
}

#' Alias for 'to.sav' (DEPRECATED)
#'
#' This function is deprecated. Please use 'to.sav' instead.
#' This is a legacy alias for the 'to.sav' function to maintain compatibility with older code.
#'
#' @param ... Additional arguments passed through to \code{to.sav()}.
#' @inherit to.sav return
#' @export
#' @examples
#' \dontrun{
#' # DEPRECATED - use to.sav() instead
#' createSpss(prl01)
#' }
createSpss <- function(...) {
  .Deprecated("to.sav", package = "wizaRdry")
  to.sav(...)
}

# Modified ConfigEnv class with more flexible missing value categories and SQL support
ConfigEnv <- R6::R6Class("ConfigEnv",
                         public = list(
                           # Store configuration
                           config = NULL,
                           config_file = NULL,

                           # Define validation specs for each API
                           api_specs = list(
                             mongo = list(
                               required = c("database")
                             ),
                             qualtrics = list(
                               required = c("survey_ids")
                             ),
                             redcap = list(
                               required = c("superkey", "primary_key")
                             ),
                             sql = list(
                               required = c()  # All SQL config settings are optional - superkey/primary_key only needed for auto-joins
                             ),
                             missing_data_codes = list(
                               required = c(),  # No required fields - all are optional
                               types = c("skipped", "refused", "unknown", "missing"),  # Add "missing" as valid category type
                               aliases = list(                                         # Add support for aliases
                                 "missing" = c("undefined", "na", "null"),
                                 "unknown" = c("undefined", "na", "null"),
                                 "skipped" = c("not_applicable", "na", "skip"),
                                 "refused" = c("declined", "no_answer")
                               ),
                               allow_custom = TRUE  # Flag to allow custom categories beyond the predefined ones
                             )
                           ),

                            initialize = function(config_file = "config.yml") {
                             # Check if config file exists
                             if (!file.exists(config_file)) {
                               stop(config_file, " not found. Please create this file with the required API configurations.")
                             }
                             # Store the config file path
                             self$config_file <- config_file
                             # Load configuration
                             self$config <- config::get(file = config_file)
                             # Process variable substitutions
                             self$process_substitutions()
                           },

                           # Method to handle variable substitutions like ${study_alias}
                           process_substitutions = function() {
                             # Process mongo database name if it references study_alias
                             if (!is.null(self$config$mongo) &&
                                 !is.null(self$config$mongo$database) &&
                                 self$config$mongo$database == "${study_alias}") {
                               if (!is.null(self$config$study_alias)) {
                                 self$config$mongo$database <- self$config$study_alias
                               } else {
                                 warning("Cannot substitute ${study_alias} in mongo.database: study_alias is not defined in config")
                               }
                             }
                               # Process sql database name if it references study_alias
                               if (!is.null(self$config$sql) &&
                                   !is.null(self$config$sql$database) &&
                                   self$config$sql$database == "${study_alias}") {
                                 if (!is.null(self$config$study_alias)) {
                                   self$config$sql$database <- self$config$study_alias
                                 } else {
                                   warning("Cannot substitute ${study_alias} in sql.database: study_alias is not defined in config")
                                 }
                               }
                             # Add more substitution rules as needed
                           },

                           # Get a specific configuration value
                           get_value = function(path) {
                             # Split the path by dots
                             parts <- strsplit(path, "\\.")[[1]]
                             # Start with the root config
                             result <- self$config
                             # Navigate through the path
                             for (part in parts) {
                               if (is.null(result) || !part %in% names(result)) {
                                 return(NULL)
                               }
                               result <- result[[part]]
                             }
                             return(result)
                           },

                           # Check if a configuration value exists
                           has_value = function(path) {
                             !is.null(self$get_value(path))
                           },

                           # Check which APIs are configured
                           get_configured_apis = function() {
                             configured_apis <- character(0)
                             for (api_type in names(self$api_specs)) {
                               if (self$has_value(api_type)) {
                                 configured_apis <- c(configured_apis, api_type)
                               }
                             }
                             return(configured_apis)
                           },

                           # Normalize a missing value category name (handle aliases)
                           normalize_missing_category = function(category) {
                             # If the category is already a valid type, return it
                             if (category %in% self$api_specs$missing_data_codes$types) {
                               return(category)
                             }

                             # Check if it's an alias for a known category
                             for (valid_type in names(self$api_specs$missing_data_codes$aliases)) {
                               aliases <- self$api_specs$missing_data_codes$aliases[[valid_type]]
                               if (category %in% aliases) {
                                 return(valid_type)
                               }
                             }

                             # If allow_custom is TRUE, return the original category
                             if (self$api_specs$missing_data_codes$allow_custom) {
                               return(category)
                             }

                             # Otherwise return NULL (invalid category)
                             return(NULL)
                           },

                           # Validate specific API configuration
                           validate_config = function(api_type = NULL) {
                             # If no API type specified, validate core config
                             if (is.null(api_type)) {
                               return(self$validate_core_config())
                             }
                             # Check if the API type is supported
                             if (!api_type %in% names(self$api_specs)) {
                               stop("Unknown API type: '", api_type, "'. Valid options are: ",
                                    paste(names(self$api_specs), collapse=", "))
                             }
                             # Check if this API is actually configured
                             if (!self$has_value(api_type)) {
                               # If the API section doesn't exist, skip validation but return TRUE
                               # message("The '", api_type, "' section is not defined in config.yml, skipping validation.")
                               return(TRUE)
                             }
                             all_errors <- c()
                             # Get API specs
                             specs <- self$api_specs[[api_type]]
                             # Check required fields (only if any are specified)
                             if (length(specs$required) > 0) {
                               for (field in specs$required) {
                                 field_path <- paste0(api_type, ".", field)
                                 if (!self$has_value(field_path)) {
                                   all_errors <- c(all_errors, paste("Missing '", field, "' setting in the ", api_type, " section"))
                                 }
                               }
                             }
                             # API-specific additional validations
                             if (api_type == "mongo") {
                               # Check if database is empty after substitution
                               if (self$has_value("mongo.database") && nchar(self$get_value("mongo.database")) == 0) {
                                 all_errors <- c(all_errors, "The 'database' setting cannot be empty")
                               }
                             } else if (api_type == "qualtrics") {
                               # Check if survey_ids exists and is a list
                               if (self$has_value("qualtrics.survey_ids")) {
                                 survey_ids <- self$get_value("qualtrics.survey_ids")
                                 if (!is.list(survey_ids)) {
                                   all_errors <- c(all_errors, "The 'survey_ids' setting must be a nested structure")
                                 } else {
                                   # Check if there are any institutions defined
                                   if (length(names(survey_ids)) == 0) {
                                     all_errors <- c(all_errors, "No institutions defined in 'survey_ids'")
                                   }
                                 }
                               }
                             } else if (api_type == "redcap") {
                               # Any redcap-specific validations
                               if (self$has_value("redcap.superkey") && nchar(self$get_value("redcap.superkey")) == 0) {
                                 all_errors <- c(all_errors, "The 'superkey' setting cannot be empty")
                               }
                               if (self$has_value("redcap.primary_key") && nchar(self$get_value("redcap.primary_key")) == 0) {
                                 all_errors <- c(all_errors, "The 'primary_key' setting cannot be empty")
                               }
                            } else if (api_type == "sql") {
                               # SQL-specific validations
                              # Check for primary_key setting if specified
                               if (self$has_value("sql.primary_key") && nchar(self$get_value("sql.primary_key")) == 0) {
                                 all_errors <- c(all_errors, "The 'primary_key' setting cannot be empty")
                               }

                              # Check for superkey table if specified
                               if (self$has_value("sql.superkey") && nchar(self$get_value("sql.superkey")) == 0) {
                                 all_errors <- c(all_errors, "The 'superkey' setting cannot be empty")
                               }

                              # Check for pii_fields if specified
                               if (self$has_value("sql.pii_fields")) {
                                 pii_fields <- self$get_value("sql.pii_fields")
                                 if (!is.vector(pii_fields) || !is.character(pii_fields)) {
                                   all_errors <- c(all_errors, "The 'pii_fields' setting must be a character vector")
                                 }
                               }

                              # Ensure at least one of database or schema exists when needed
                              if (!self$has_value("sql.database") && !self$has_value("sql.schema")) {
                                message("Note: Neither 'database' nor 'schema' specified under sql; some features may attempt auto-detection or require fully qualified names.")
                              }
                             } else if (api_type == "missing_data_codes") {
                               # Get the current missing_data_codes configuration
                               missing_value_config <- self$get_value("missing_data_codes")

                               # If any categories are defined, validate them
                               if (!is.null(missing_value_config)) {
                                 config_categories <- names(missing_value_config)

                                 # More lenient validation when allow_custom is TRUE
                                 if (!specs$allow_custom) {
                                   # Check that all provided categories are valid or aliases of valid categories
                                   valid_categories <- c()
                                   unexpected_categories <- c()

                                   for (category in config_categories) {
                                     normalized_category <- self$normalize_missing_category(category)
                                     if (!is.null(normalized_category)) {
                                       valid_categories <- c(valid_categories, category)
                                     } else {
                                       unexpected_categories <- c(unexpected_categories, category)
                                     }
                                   }

                                   if (length(unexpected_categories) > 0) {
                                     known_aliases <- unlist(specs$aliases)
                                     all_errors <- c(all_errors, paste0("Unexpected categories in missing_data_codes: ",
                                                                        paste(unexpected_categories, collapse=", "),
                                                                        ". Allowed categories are: ",
                                                                        paste(c(specs$types, known_aliases), collapse=", ")))
                                   }
                                 }

                                 # For each provided category, check its values
                                 for (category in config_categories) {
                                   category_values <- self$get_value(paste0("missing_data_codes.", category))
                                   # Check if the category is a list or vector
                                   if (!is.vector(category_values) && !is.list(category_values)) {
                                     all_errors <- c(all_errors, paste0("The '", category, "' category must be a list or vector of values"))
                                   }
                                   # Check if the list has at least one value
                                   if (length(category_values) == 0) {
                                     all_errors <- c(all_errors, paste0("The '", category, "' category must contain at least one value"))
                                   }
                                 }

                                 # Missing categories are ok - we'll use defaults
                                 missing_categories <- setdiff(specs$types, config_categories)
                                 if (length(missing_categories) > 0) {
                                   message("Note: The following missing value categories are not defined and will use R's default NA: ",
                                           paste(missing_categories, collapse=", "))
                                 }
                               } else {
                                 # If missing_data_codes section is completely empty, just inform the user
                                 message("Note: No missing value categories defined. Default R NA values will be used for all missing value types.")
                               }
                             }

                             # If we found any errors, report them all at once
                             if (length(all_errors) > 0) {
                               stop(api_type, " configuration errors in ", self$config_file, ":\n- ",
                                    paste(all_errors, collapse="\n- "), call. = FALSE)
                             }
                             # else {
                             #   message("The ", api_type, " configuration in ", self$config_file, " is valid.")
                             # }
                             return(TRUE)
                           },

                           # Validate core configuration
                           validate_core_config = function() {
                             all_errors <- c()
                             # Check required global fields
                             required_fields <- c("study_alias", "identifier")
                             for (field in required_fields) {
                               if (!self$has_value(field)) {
                                 all_errors <- c(all_errors, paste("Missing required '", field, "' setting in the root configuration"))
                               }
                             }
                             # If we found any errors, report them all at once
                             if (length(all_errors) > 0) {
                               stop("Core configuration errors in ", self$config_file, ":\n- ",
                                    paste(all_errors, collapse="\n- "), call. = FALSE)
                             } else {
                               # message("The core configuration in ", self$config_file, " is valid.")
                             }
                             return(TRUE)
                           },

                           # Get missing value codes for a specific category
                           get_missing_data_codes = function(category = NULL) {
                             # If no category specified, return all missing values
                             if (is.null(category)) {
                               return(self$get_value("missing_data_codes"))
                             }

                             # If we have an exact match for the category, use it
                             if (self$has_value(paste0("missing_data_codes.", category))) {
                               values <- self$get_value(paste0("missing_data_codes.", category))
                               if (!is.null(values)) {
                                 return(values)
                               }
                             }

                             # Try to normalize the category name (handle aliases)
                             normalized_category <- self$normalize_missing_category(category)
                             if (!is.null(normalized_category) && normalized_category != category) {
                               # Check if the normalized category exists in config
                               if (self$has_value(paste0("missing_data_codes.", normalized_category))) {
                                 values <- self$get_value(paste0("missing_data_codes.", normalized_category))
                                 if (!is.null(values)) {
                                   return(values)
                                 }
                               }
                             }

                             # If no values are defined for any variant, return NULL to indicate default NA should be used
                             return(NULL)
                           }
                         )
)

# Create a function to validate configuration and return the config
validate_config <- function(api_type = NULL, config_file = "config.yml") {
  config_env <- ConfigEnv$new(config_file)
  # If no specific API type is provided, just validate core config
  if (is.null(api_type)) {
    validation_result <- config_env$validate_core_config()
  } else {
    # If a specific API type is requested, validate just that one
    validation_result <- config_env$validate_config(api_type)
  }
  # If validation passes, return the config
  if (validation_result) {
    return(config_env$config)
  } else {
    return(NULL)  # Or handle failure appropriately
  }
}

# Helper function to get missing value codes
get_missing_data_codes <- function(category = NULL, config_file = "config.yml") {
  config_env <- ConfigEnv$new(config_file)
  return(config_env$get_missing_data_codes(category))
}

#' Add Prefix to Column Names Based on DataFrame Name
#'
#' This function adds a prefix to column names derived from the dataframe name (excluding "_clean").
#' It preserves standard identifier and metadata columns by excluding them from the prefixing operation.
#' This is useful for maintaining clear variable origins when merging multiple datasets.
#'
#' @param df A data frame whose columns need to be prefixed
#' @param excluded_cols Optional character vector of column names that should not be prefixed.
#'        Defaults to standard NDA and identifier columns.
#'
#' @return A data frame with prefixed column names
#'
#' @details
#' The function:
#' 1. Extracts a prefix from the data frame name (removing "_clean" suffix if present)
#' 2. Only adds the prefix to columns that don't already have it
#' 3. Preserves standard identifier/metadata columns untouched
#'
#' @examples
#' \donttest{
#' # Create sample data
#' social_prl_clean <- data.frame(
#'   src_subject_id = c("S001", "S002"),
#'   trial = c(1, 2),
#'   rt = c(450, 520),
#'   accuracy = c(1, 0)
#' )
#'
#' # Add prefix to column names
#' social_prl_clean <- addColumnPrefix(social_prl_clean)
#'
#' # Result will have columns:
#' # "src_subject_id", "social_prl_trial", "social_prl_rt", "social_prl_accuracy"
#' }
#'
#' @noRd
addColumnPrefix <- function(df) {
  # Determine the prefix from the dataframe name, excluding "_clean"
  prefix <- sub("_clean$", "", deparse(substitute(df)))
  
  # Specify columns that should not have the prefix added
  excluded_cols <- c("src_subject_id", "visit", "phenotype", "interview_age", "interview_date", "site", "sex", "subjectkey", "arm")
  
  # Identify columns that are not in the excluded list and don't already start with the prefix
  cols_to_prefix <- names(df)[!names(df) %in% excluded_cols & !startsWith(names(df), prefix)]
  
  # Loop through each column name that needs the prefix added
  new_names <- setNames(names(df), names(df))
  for (col in cols_to_prefix) {
    new_names[col] <- paste0(prefix, "_", col)
  }
  
  # Rename the columns in the dataframe
  names(df) <- new_names
  
  return(df)
}

# Example of usage:
# Assuming `social_prl_clean` is your dataframe
# social_prl_clean <- addColumnPrefix(social_prl_clean)

#' Generate clean data frames from cleaning scripts created in the ./clean directory
#'
#' This function processes requests for clean data sequentially for specified measures.
#' It makes a request to the appropriate API for the named measure or measures
#' and runs the associated data cleaning routines. It then runs a series of
#' unit tests to verify that the data quality standards are met.
#'
#' @param ... Strings, specifying the measures to process, which can be a Mongo collection, REDCap instrument, or Qualtrics survey.
#' @param csv Optional; Boolean, if TRUE creates a .csv extract in ./tmp.
#' @param rdata Optional; Boolean, if TRUE creates an .rdata extract in ./tmp.
#' @param spss Optional; Boolean, if TRUE creates a .sav extract in ./tmp.
#' @param skip_prompt Logical. If TRUE (default), skips confirmation prompts. If FALSE,
#'   prompts for confirmation unless the user has previously chosen to remember their preference.
#' @return Prints the time taken for the data request process.
#' @export
#' @examples
#' \dontrun{
#'   clean("prl", csv=TRUE)
#'   clean("rgpts", "kamin", rdata=TRUE)
#'
#'   # Skip confirmation prompts
#'   clean("prl", csv=TRUE, skip_prompt=TRUE)
#' }
#'
#' @author Joshua Kenney <joshua.kenney@yale.edu>
clean <- function(..., csv = FALSE, rdata = FALSE, spss = FALSE, skip_prompt = TRUE) { #skip_prompt to true so preferences can be set on first run

  # Define base path
  path <- "." # Or whatever directory you're working from

  # Required Libraries Setup

  # Prepare lists for REDCap, Qualtrics, and MongoDB
  csv_list <- tools::file_path_sans_ext(list.files("./clean/csv"))
  redcap_list <- tools::file_path_sans_ext(list.files("./clean/redcap"))
  qualtrics_list <- tools::file_path_sans_ext(list.files("./clean/qualtrics"))
  mongo_list <- tools::file_path_sans_ext(list.files("./clean/mongo"))
  oracle_list <- tools::file_path_sans_ext(list.files("./clean/oracle"))
  sql_list <- tools::file_path_sans_ext(list.files("./clean/sql"))

  # Get identifier from config
  config <- validate_config()
  identifier <- config$identifier
  if (is.null(identifier) || identifier == "") {
    stop("No identifier specified in the config file.")
  }

  # Split identifier if it's a comma-separated string
  if (is.character(identifier)) {
    identifier <- strsplit(identifier, ",")[[1]]
  }

  start_time <- Sys.time()

  # Source necessary R scripts from the 'api' directory

  # Create a mapping to store the source type for newly created scripts
  new_script_sources <- list()

  # Validate Measures Function
  validateMeasures <- function(data_list) {
    # Check if input is a dataframe
    if (is.data.frame(data_list)) {
      # Get the name of the dataframe as a string
      data_list <- deparse(substitute(data_list))
    }

    # Ensure data_list is a character vector (in case it's a single string)
    if (!is.character(data_list)) {
      data_list <- as.character(data_list)
    }


    # Check for user preferences file
    user_prefs_file <- file.path(path, ".wizaRdry_prefs")
    user_prefs <- list(shown_tree = FALSE, auto_create = FALSE, auto_clean = FALSE)

    if (file.exists(user_prefs_file)) {
      tryCatch({
        user_prefs <- readRDS(user_prefs_file)
        # Add the auto_clean field if it doesn't exist
        if (is.null(user_prefs$auto_clean)) {
          user_prefs$auto_clean <- FALSE
        }
      }, error = function(e) {
        # If file exists but can't be read, create a new one
        user_prefs <- list(shown_tree = FALSE, auto_create = FALSE, auto_clean = FALSE)
      })
    }

    # Validate measures against predefined lists
    invalid_scripts <- Filter(function(measure) !measure %in% c(csv_list, redcap_list, qualtrics_list, mongo_list, oracle_list, sql_list), data_list)

    # If we have invalid scripts to create and need to prompt
    if (length(invalid_scripts) > 0) {
      # If skip_prompt is TRUE or user has previously set auto_clean to TRUE, bypass the prompt
      if (!skip_prompt | !user_prefs$auto_clean) {
        script_word <- ifelse(length(invalid_scripts) > 1, "scripts", "script")
        response <- readline(prompt = sprintf("Would you like to create cleaning %s for %s now? y/n ",
                                              script_word,
                                              paste(invalid_scripts, collapse = ", ")))

        while (!tolower(response) %in% c("y", "n")) {
          response <- readline(prompt = "Please enter either y or n: ")
        }

        # If user selects y, set auto_clean to TRUE and update preferences
        if (tolower(response) == "y") {
          user_prefs$auto_clean <- TRUE
          saveRDS(user_prefs, user_prefs_file)
        }

        if (tolower(response) == "n") {
          message("Cleaning script creation cancelled.")
          invokeRestart("abort")  # This exits without the "Error:" prefix
        }
      }

      # If response is "y", allow user to select api:
      api_selection <- function() {
        options <- c("mongo", "qualtrics", "redcap")

        cat("Select script type (choose only one):\n")

        for (i in 1:length(options)) {
          cat(i, ":", options[i], "\n")
        }

        # Get user choice
        choice <- as.numeric(readline("Enter number to select option: "))

        while(is.na(choice) || choice < 1 || choice > length(options)) {
          cat("Please enter a valid number between 1 and", length(options), "\n")
          choice <- as.numeric(readline("Enter number to select option: "))
        }

        # Return the selected API
        return(options[choice])
      }

      # Use the function - select API once for all scripts
      selected_api <- api_selection()

      # Process each invalid script
      for (script_name in invalid_scripts) {
        message(sprintf("\nProcessing script: %s", script_name))

        # Store the selected API for this script in our mapping
        new_script_sources[[script_name]] <<- selected_api

        clean_templates <- list(
          mongo = list(
            path = sprintf(file.path(path, "clean", "mongo", "%s.R"), script_name),  # Added .R extension
            content = paste(
              "#",
              sprintf("# clean/mongo/%s.R", script_name),
              "#",
              '# config:  database name is defined in config.yml',
              '# secrets: connectionString is defined in secrets.R',
              '# encrypt: the *.pem file must be placed in the root of this repository',
              "#",
              "# return a list of the collection(s) from MongoDB",
              "# mongo.index()",
              "#",
              "# get collection from MongoDB",
              "# IMPORTANT: both variable name and script filename must match",
              sprintf("%s <- mongo(\"%s\")", script_name, script_name),
              "",
              "# cleaning script code...",
              "",
              "# final df must be named like the R script and appended with _clean",
              sprintf("%s_clean <- %s", script_name, script_name),
              sep = "\n"
            )
          ),
          qualtrics = list(
            path = sprintf(file.path(path, "clean", "qualtrics", "%s.R"), script_name),  # Added .R extension
            content = paste(
              "#",
              sprintf("# clean/qualtrics/%s.R", script_name),
              "#",
              "# get survey from Qualtrics database",
              "# config:  surveys are defined in config.yml as key-value pairs",
              "# secrets: baseUrls and apiKeys are defined in secrets.R",
              "#",
              "# return a list of the survey(s) from Qualtrics",
              "# qualtrics.index()",
              "#",
              "# get collection from Qualtrics",
              "# IMPORTANT: both variable name and script filename must match",
              sprintf("%s <- qualtrics(\"%s\")", script_name, script_name),
              "",
              "# cleaning script code...",
              "",
              "# IMPORTANT: final df must be appended with _clean",
              sprintf("%s_clean <- %s", script_name, script_name),
              sep = "\n"
            )
          ),
          redcap = list(
            path = sprintf(file.path(path, "clean", "redcap", "%s.R"), script_name),  # Added .R extension
            content = paste(
              "#",
              sprintf("# clean/redcap/%s.R", script_name),
              "#",
              "# config:  superkey instrument is defined in config.yml",
              "# secrets: uri and token are defined in secrets.R",
              "#",
              "# return a list of the instrument_name(s) from REDCap",
              "# redcap.index()",
              "#",
              "# get the instrument_name from REDCap",
              "# IMPORTANT: both variable name and script filename must match",
              sprintf("%s <- redcap(\"%s\")", script_name, script_name),
              "",
              "# cleaning script code...",
              "",
              "# IMPORTANT: final df must be appended with _clean",
              sprintf("%s_clean <- %s", script_name, script_name),
              sep = "\n"
            )
          ),
          oracle = list(
            path = sprintf(file.path(path, "clean", "oracle", "%s.R"), script_name),  # Added .R extension
            content = paste(
              "#",
              sprintf("# clean/oracle/%s.R", script_name),
              "#",
              "# config:  coming soon...",
              "# secrets: coming soon...",
              "#",
              "# return a list of tables from ORACLE",
              "# oracle.index()",
              "#",
              "# get the table from ORACLE",
              "# IMPORTANT: both variable name and script filename must match",
              sprintf("%s <- oracle(\"%s\")", script_name, script_name),
              "",
              "# cleaning script code...",
              "",
              "# IMPORTANT: final df must be appended with _clean",
              sprintf("%s_clean <- %s", script_name, script_name),
              sep = "\n"
            )
          ),
          sql = list(
            path = sprintf(file.path(path, "clean", "sql", "%s.R"), script_name),  # Added .R extension
            content = paste(
              "#",
              sprintf("# clean/sql/%s.R", script_name),
              "#",
              "# config:  coming soon...",
              "# secrets: coming soon...",
              "#",
              "# return a list of tables from SQL",
              "# sql.index()",
              "#",
              "# get the table from SQL",
              "# IMPORTANT: both variable name and script filename must match",
              sprintf("%s <- sql(\"%s\")", script_name, script_name),
              "",
              "# cleaning script code...",
              "",
              "# IMPORTANT: final df must be appended with _clean",
              sprintf("%s_clean <- %s", script_name, script_name),
              sep = "\n"
            )
          ),
          csv = list(
            path = sprintf(file.path(path, "clean", "csv", "%s.R"), script_name),  # Added .R extension
            content = paste(
              "#",
              sprintf("# clean/csv/%s.R", script_name),
              "#",
              "# get the data from CSV file",
              "# IMPORTANT: both variable name and script filename must match",
              sprintf("%s <- read.csv(\"%s\".csv)", script_name, script_name),
              "",
              "# cleaning script code...",
              "",
              "# IMPORTANT: final df must be appended with _clean",
              sprintf("%s_clean <- %s", script_name, script_name),
              sep = "\n"
            )
          )
        )

        template <- clean_templates[[selected_api]]

        # Create directory if it doesn't exist
        dir_path <- dirname(template$path)
        if (!dir.exists(dir_path)) {
          dir.create(dir_path, recursive = TRUE)
        }

        if (!file.exists(template$path)) {
          writeLines(template$content, template$path)
          message(sprintf("Created file: %s", template$path))
          # open script for editing
          rstudioapi::navigateToFile(sprintf("%s",template$path))
        } else {
          message(sprintf("File already exists: %s (skipped)", template$path))
        }
      }
    }

    # After creating new scripts in validateMeasures, update the lists
    csv_list <<- tools::file_path_sans_ext(list.files("./clean/csv"))
    redcap_list <<- tools::file_path_sans_ext(list.files("./clean/redcap"))
    qualtrics_list <<- tools::file_path_sans_ext(list.files("./clean/qualtrics"))
    mongo_list <<- tools::file_path_sans_ext(list.files("./clean/mongo"))
    oracle_list <<- tools::file_path_sans_ext(list.files("./clean/oracle"))
    sql_list <<- tools::file_path_sans_ext(list.files("./clean/sql"))

    # Return the data_list invisibly instead of stopping execution
    return(invisible(data_list))
  }

  # Compile data list and validate measures
  data_list <- list(...)

  # This is so the function doesn't break if user enters a variable storing a character vector
  # or a list of strings
  # in other words it let's you do this:
  # vars_i_want <- c('demo','sps','sips_p')
  # clean(vars_i_want)
  if (length(data_list) == 1) {
    data_list = data_list[[1]]
  }

  # Validate measures and potentially create new scripts
  validateMeasures(data_list)

  # Process each measure using processData function
  for (measure in data_list) {
    # Check if this is a newly created script with a known source
    if (measure %in% names(new_script_sources)) {
      sourceCategory <- new_script_sources[[measure]]
    } else {
      # Otherwise determine the source from updated lists
      sourceCategory <- ifelse(measure %in% redcap_list, "redcap",
                               ifelse(measure %in% qualtrics_list, "qualtrics",
                                      ifelse(measure %in% mongo_list, "mongo",
                                             ifelse(measure %in% oracle_list, "oracle",
                                                    ifelse(measure %in% sql_list, "sql", "csv")))))
    }

    processData(measure, sourceCategory, csv, rdata, spss, identifier)
  }

  # Clean up and record processing time
  print(Sys.time() - start_time)  # Print time taken for processing

  # Flush environment

  return(invisible(NULL))
}

processData <- function(measure, source, csv, rdata, spss, identifier) {
  # Check if input is a dataframe
  if (is.data.frame(measure)) {
    # Get the name of the dataframe as a string
    measure <- deparse(substitute(measure))
  }

  # Ensure data_list is a character vector (in case it's a single string)
  if (!is.character(measure)) {
    measure <- as.character(measure)
  }
  # Construct the path to the measure's cleaning script
  file_path <- sprintf("./clean/%s/%s.R", source, measure)
  message("\nProcessing ", measure, " from ./clean/", source, "/", measure, ".R")

  # Setup cleanup on exit
  on.exit({
    if (exists("mongo_conn") && !is.null(mongo_conn)) {
      tryCatch({
        mongo_conn$disconnect()
      }, error = function(e) {
        warning(sprintf("Error disconnecting from MongoDB: %s", e$message))
      })
    }
    # Clear the mongo connection from memory
    if (exists("mongo_conn")) {
      rm(mongo_conn)
    }
    gc()  # Force garbage collection
  })

  result <- tryCatch({
    base::source(file_path)  # Execute the cleaning script for the measure
    # Ensure testSuite is sourced and then called
    # Call testSuite with identifier
    testSuite(measure, source, file_path, identifier)

    df_name <- paste0(measure, "_clean")  # Construct the name of the cleaned data frame

    # Assuming createExtract is a function to create data extracts
    createExtract(base::get(df_name), df_name, csv, rdata, spss)  # Create data extracts
  }, error = function(e) {
    # Check if identifier is valid (you can modify this logic based on your criteria)
    if (length(identifier) == 0 || all(is.na(identifier))) {
      message("An error occurred: ", e$message)  # General error message
    } else {
      message("Error with ./clean/", source, "/", measure, ".R: ", e$message)  # Specific error message
    }
    NULL  # Return NULL on error
  })

  return(result)  # Return the result of the processing
}

# Add helper function for MongoDB cleanup
disconnectMongo <- function(mongo) {
  if (!is.null(mongo)) {
    tryCatch({
      mongo$disconnect()
      rm(list = deparse(substitute(mongo)), envir = parent.frame())
    }, error = function(e) {
      warning(sprintf("Error disconnecting from MongoDB: %s", e$message))
    })
  }
}

#' Alias for 'clean' (DEPRECATED)
#'
#' This function is deprecated. Please use 'clean' instead.
#' This is a legacy alias for the 'clean' function to maintain compatibility with older code.
#'
#' @inheritParams clean
#' @inherit clean return
#' @export
#' @examples
#' \dontrun{
#' # DEPRECATED - use clean() instead
#' prl <- dataRequest("prl")
#' }
dataRequest <- function(...) {
  .Deprecated("clean", package = "wizaRdry")
  clean(...)
}

#' NdaDataStructure R6 Class
#'
#' @description
#' Represents a single field (data element) in an NDA data structure.
#' This is a typed struct (similar to Go structs) that enforces schema consistency
#' and provides validation for NDA field definitions.
#'
#' @details
#' This class replaces ad-hoc list construction for NDA field definitions.
#' It provides:
#' - Type safety and validation for field definitions
#' - Consistent structure across all code paths
#' - Factory methods for creating fields from different sources
#' - Helper methods for common operations
#' - Direct mapping to Excel export columns
#'
#' The field structure matches the NDA data dictionary schema:
#' ElementName, DataType, Size, Required, ElementDescription, ValueRange, Notes, Aliases
#'
#' Uses typed R6 classes from NdaClasses.R for all fields (ElementName, DataType, 
#' Size, RequirementLevel, Description, ValueRange, Notes, Aliases, etc.)
#'
#' @keywords internal
NdaDataStructure <- R6::R6Class("NdaDataStructure",
  public = list(
    #' @field element_name ElementName object - field name (ElementName in Excel)
    element_name = NULL,
    
    #' @field data_type DataType object - data type (String, Integer, Float, Date, GUID, Boolean)
    data_type = NULL,
    
    #' @field size Size object - size for String types
    size = NULL,
    
    #' @field required RequirementLevel object - requirement level (Required, Recommended, Conditional, No)
    required = NULL,
    
    #' @field element_description Description object - field description
    element_description = NULL,
    
    #' @field value_range ValueRange object - allowed values or range
    value_range = NULL,
    
    #' @field notes Notes object - field notes
    notes = NULL,
    
    #' @field aliases Aliases object - field aliases
    aliases = NULL,
    
    #' @field selection_order Integer - order in which field was selected
    selection_order = NULL,
    
    #' @field selected_for_submission Logical - whether field is selected for NDA submission
    selected_for_submission = TRUE,
    
    #' @field source_metadata SourceMetadata object - field source tracking
    source_metadata = NULL,
    
    #' @field missing_info MissingInfo object - missing data information
    missing_info = NULL,
    
    #' @field validation_rules ValidationRules object - validation rules
    validation_rules = NULL,
    
    #' @description
    #' Create a new NdaDataStructure instance
    #' @param element_name Field name (required) - accepts string or ElementName object
    #' @param data_type Data type (default: "String") - accepts string or DataType object
    #' @param size Size for String types - accepts numeric or Size object
    #' @param required Requirement level (default: "No") - accepts string or RequirementLevel object
    #' @param element_description Field description - accepts string or Description object
    #' @param value_range Allowed values or range - accepts string or ValueRange object
    #' @param notes Field notes - accepts string or Notes object
    #' @param aliases Field aliases - accepts string, list, or Aliases object
    #' @param selection_order Selection order
    #' @param source Field source (legacy - use source_metadata instead)
    #' @param source_metadata SourceMetadata object
    #' @param missing_info MissingInfo object or list
    #' @param validation_rules ValidationRules object or list
    #' @param ... Additional fields
    #' @return A new NdaDataStructure object
    initialize = function(element_name, 
                         data_type = "String",
                         size = NULL,
                         required = "No",
                         element_description = "",
                         value_range = "",
                         notes = "",
                         aliases = "",
                         selection_order = NULL,
                         source = NULL,
                         source_metadata = NULL,
                         missing_info = NULL,
                         validation_rules = NULL,
                         ...) {
      # Validate required fields
      if (missing(element_name) || is.null(element_name) || element_name == "") {
        stop("element_name is required")
      }
      
      # Convert to typed objects if needed
      self$element_name <- if (inherits(element_name, "ElementName")) {
        element_name
      } else {
        ElementName$new(element_name)
      }
      
      self$data_type <- if (inherits(data_type, "DataType")) {
        data_type
      } else {
        DataType$new(data_type)
      }
      
      self$size <- if (is.null(size)) {
        Size$new(NULL, self$data_type$value)
      } else if (inherits(size, "Size")) {
        size
      } else {
        Size$new(size, self$data_type$value)
      }
      
      self$required <- if (inherits(required, "RequirementLevel")) {
        required
      } else {
        RequirementLevel$new(required)
      }
      
      self$element_description <- if (inherits(element_description, "Description")) {
        element_description
      } else {
        Description$new(element_description %||% "")
      }
      
      self$value_range <- if (inherits(value_range, "ValueRange")) {
        value_range
      } else {
        ValueRange$new(value_range %||% "")
      }
      
      self$notes <- if (inherits(notes, "Notes")) {
        notes
      } else {
        Notes$new(notes %||% "")
      }
      
      self$aliases <- if (inherits(aliases, "Aliases")) {
        aliases
      } else {
        Aliases$new(aliases %||% "")
      }
      
      # Metadata fields
      self$selection_order <- selection_order
      
      # Source metadata
      self$source_metadata <- if (!is.null(source_metadata)) {
        source_metadata
      } else if (!is.null(source)) {
        # Legacy compatibility - convert string source to SourceMetadata
        SourceMetadata$new(primary_source = source)
      } else {
        SourceMetadata$new()
      }
      
      # Missing info
      self$missing_info <- if (inherits(missing_info, "MissingInfo")) {
        missing_info
      } else if (is.list(missing_info)) {
        MissingInfo$new(
          missing_count = missing_info$missing_count %||% 0L,
          total_count = missing_info$total_count %||% 0L
        )
      } else {
        NULL
      }
      
      # Validation rules
      self$validation_rules <- if (inherits(validation_rules, "ValidationRules")) {
        validation_rules
      } else if (is.list(validation_rules)) {
        ValidationRules$new(
          min_value = validation_rules$min_value,
          max_value = validation_rules$max_value,
          allowed_values = validation_rules$allowed_values,
          pattern = validation_rules$pattern
        )
      } else {
        NULL
      }
      
      # Handle additional args
      extra_args <- list(...)
      for (name in names(extra_args)) {
        if (name %in% names(self)) {
          self[[name]] <- extra_args[[name]]
        }
      }
    },
    
    #' @description
    #' Convert to Excel row (returns named list for data.frame row)
    #' @return Named list with Excel column names and values
    to_excel_row = function() {
      list(
        ElementName = self$element_name$value,
        DataType = self$data_type$value,
        Size = self$size$to_string(),
        Required = self$required$value,
        ElementDescription = self$element_description$value,
        ValueRange = self$value_range$to_nda_format(),
        Notes = self$notes$to_string(),
        Aliases = self$aliases$to_string()
      )
    },
    
    #' @description
    #' Convert to legacy list format for backward compatibility
    #' @return List with field definition
    to_list = function() {
      list(
        name = self$element_name$value,
        selection_order = self$selection_order,
        selected_for_submission = self$selected_for_submission,
        source = self$source_metadata$primary_source,
        data_type = self$data_type$value,
        description = self$element_description$value,
        required = self$required$value == "Required",
        validation_rules = if (!is.null(self$validation_rules)) {
          self$validation_rules$to_list()
        } else {
          NULL
        },
        nda_metadata = list(
          name = self$element_name$value,
          type = self$data_type$value,
          size = if (self$size$is_valid) self$size$value else NULL,
          required = self$required$value,
          description = self$element_description$value,
          valueRange = self$value_range$to_nda_format(),
          notes = self$notes$to_string(),
          aliases = self$aliases$to_string()
        ),
        missing_info = if (!is.null(self$missing_info)) {
          list(
            missing_count = self$missing_info$missing_count,
            missing_percentage = self$missing_info$missing_percentage,
            total_count = self$missing_info$total_count
          )
        } else {
          NULL
        },
        is_modified = self$source_metadata$is_modified,
        modification_notes = paste(self$source_metadata$modifications, collapse = "; ")
      )
    },
    
    #' @description
    #' Check if field is a super required field
    #' @return Logical
    is_super_required = function() {
      self$element_name$value %in% SUPER_REQUIRED_FIELDS
    },
    
    #' @description
    #' Check if field came from ndar_subject01
    #' @return Logical
    is_from_ndar_subject = function() {
      ndar_fields <- c(SUPER_REQUIRED_FIELDS, 
                      c("ethnic_group", "site", "study", "subsiteid"))
      self$element_name$value %in% ndar_fields
    },
    
    #' @description
    #' Check if field is a DCC required field
    #' @return Logical
    is_dcc_required = function() {
      self$element_name$value %in% DCC_REQUIRED_FIELDS
    },
    
    #' @description
    #' Check if field is a DCC recommended field
    #' @return Logical
    is_dcc_recommended = function() {
      self$element_name$value %in% DCC_RECOMMENDED_FIELDS
    },
    
    #' @description
    #' Create a modified copy of this field
    #' @param value_range New value range (string or ValueRange object)
    #' @param notes New notes (string or Notes object)
    #' @param modification_note Description of modification
    #' @param ... Other fields to modify
    #' @return New NdaDataStructure object
    modify = function(value_range = NULL, notes = NULL, modification_note = NULL, ...) {
      clone <- self$clone(deep = TRUE)
      
      # Mark as modified
      clone$source_metadata$add_modification(modification_note %||% "Field modified")
      
      # Update value range
      if (!is.null(value_range)) {
        clone$value_range <- if (inherits(value_range, "ValueRange")) {
          value_range
        } else {
          ValueRange$new(value_range)
        }
      }
      
      # Update notes
      if (!is.null(notes)) {
        clone$notes <- if (inherits(notes, "Notes")) {
          notes
        } else {
          Notes$new(notes)
        }
      }
      
      # Update other fields
      extra_args <- list(...)
      for (name in names(extra_args)) {
        if (name %in% names(clone)) {
          value <- extra_args[[name]]
          
          # Try to convert to appropriate type based on field name
          if (name == "element_name" && !inherits(value, "ElementName")) {
            clone[[name]] <- ElementName$new(value)
          } else if (name == "data_type" && !inherits(value, "DataType")) {
            clone[[name]] <- DataType$new(value)
          } else if (name == "size" && !inherits(value, "Size")) {
            clone[[name]] <- Size$new(value, clone$data_type$value)
          } else if (name == "required" && !inherits(value, "RequirementLevel")) {
            clone[[name]] <- RequirementLevel$new(value)
          } else if (name == "element_description" && !inherits(value, "Description")) {
            clone[[name]] <- Description$new(value)
          } else if (name == "aliases" && !inherits(value, "Aliases")) {
            clone[[name]] <- Aliases$new(value)
          } else {
            clone[[name]] <- value
          }
        }
      }
      
      clone
    },
    
    #' @description
    #' Print method for NdaDataStructure
    #' @return Self (invisibly)
    print = function() {
      cat(sprintf("NdaDataStructure: %s\n", self$element_name$value))
      cat(sprintf("  Type: %s", self$data_type$value))
      if (self$size$is_valid) {
        cat(sprintf(" (size: %s)", self$size$value))
      }
      cat("\n")
      cat(sprintf("  Required: %s\n", self$required$value))
      
      range_str <- self$value_range$to_nda_format()
      if (!is.null(range_str) && range_str != "") {
        range_preview <- if (nchar(range_str) > 50) {
          paste0(substring(range_str, 1, 50), "...")
        } else {
          range_str
        }
        cat(sprintf("  ValueRange: %s\n", range_preview))
      }
      
      cat(sprintf("  Source: %s\n", self$source_metadata$to_string()))
      
      if (self$source_metadata$is_modified) {
        cat("  [MODIFIED]\n")
        if (length(self$source_metadata$modifications) > 0) {
          cat(sprintf("    %s\n", paste(self$source_metadata$modifications, collapse = "\n    ")))
        }
      }
      
      if (!is.null(self$missing_info)) {
        cat(sprintf("  Missing: %d/%d (%.1f%%)\n", 
                   self$missing_info$missing_count,
                   self$missing_info$total_count,
                   self$missing_info$missing_percentage))
      }
      
      invisible(self)
    },
    
    #' @description
    #' Merge value ranges from multiple sources
    #' @param nda_range ValueRange from NDA (may be NULL)
    #' @param redcap_range ValueRange from REDCap (may be NULL)
    #' @param data_range ValueRange from data (may be NULL)
    #' @param missing_codes Character vector of missing data codes
    #' @return Self (invisibly) with merged value_range
    merge_value_ranges = function(nda_range = NULL, 
                                  redcap_range = NULL, 
                                  data_range = NULL,
                                  missing_codes = character(0)) {
      
      merge_result <- merge_value_ranges(
        nda_range = nda_range %||% self$value_range,
        redcap_range = redcap_range,
        data_range = data_range,
        missing_codes = missing_codes
      )
      
      # Update value range
      self$value_range <- merge_result$merged_range
      
      # Record modifications
      if (length(merge_result$modifications) > 0) {
        for (mod in merge_result$modifications) {
          self$source_metadata$add_modification(mod)
        }
      }
      
      # Store warnings for later retrieval
      if (length(merge_result$warnings) > 0) {
        private$merge_warnings <- c(private$merge_warnings, merge_result$warnings)
      }
      
      invisible(self)
    },
    
    #' @description
    #' Get any warnings from merge operations
    #' @return Character vector of warnings
    get_merge_warnings = function() {
      private$merge_warnings
    },
    
    #' @description
    #' Clear merge warnings
    #' @return Self (invisibly)
    clear_merge_warnings = function() {
      private$merge_warnings <- character(0)
      invisible(self)
    }
  ),
  
  private = list(
    merge_warnings = character(0)
  )
)

#' Create NdaDataStructure from NDA API response
#'
#' @description
#' Factory function to create an NdaDataStructure from an NDA data element
#' (as returned from the NDA API or nda_lookup)
#'
#' @param nda_element List - NDA data element from API
#' @param selection_order Integer - selection order
#' @param missing_info MissingInfo object or list - missing data information
#' @return NdaDataStructure object
#' @keywords internal
#' @noRd
nda_structure_from_nda <- function(nda_element, selection_order = NULL, missing_info = NULL) {
  if (is.null(nda_element) || !is.list(nda_element)) {
    stop("nda_element must be a list")
  }
  
  # Extract and validate name
  name <- nda_element$name %||% stop("nda_element must have 'name' field")
  
  # Create typed objects for each field
  element_name <- ElementName$new(name)
  data_type <- DataType$new(nda_element$type %||% "String")
  size <- Size$new(nda_element$size, data_type$value)
  required <- RequirementLevel$new(nda_element$required %||% "No")
  
  # Handle condition if present
  if (!is.null(nda_element$condition) && nda_element$condition != "") {
    required <- RequirementLevel$new("Conditional", condition = nda_element$condition)
  }
  
  element_description <- Description$new(nda_element$description %||% "")
  value_range <- ValueRange$new(nda_element$valueRange %||% "")
  notes <- Notes$new(nda_element$notes %||% "")
  
  # CRITICAL FIX: Handle aliases from NDA API
  # NDA API returns: list(c("alias1", "alias2", "alias3"))
  # We need to flatten and convert to comma-separated string
  aliases_value <- nda_element$aliases
  if (is.list(aliases_value)) {
    aliases_value <- unlist(aliases_value, use.names = FALSE)
  }
  aliases <- Aliases$new(aliases_value %||% "")
  
  # Create source metadata
  source_metadata <- SourceMetadata$new(primary_source = "nda")
  source_metadata$mark_nda_match()
  
  # Convert missing_info if it's a list
  if (is.list(missing_info) && !inherits(missing_info, "MissingInfo")) {
    missing_info <- MissingInfo$new(
      missing_count = missing_info$missing_count %||% 0L,
      total_count = missing_info$total_count %||% 0L
    )
  }
  
  # Create structure
  structure <- NdaDataStructure$new(
    element_name = element_name,
    data_type = data_type,
    size = size,
    required = required,
    element_description = element_description,
    value_range = value_range,
    notes = notes,
    aliases = aliases,
    selection_order = selection_order,
    source_metadata = source_metadata,
    missing_info = missing_info
  )
  
  structure
}

#' Create NdaDataStructure from data frame column
#'
#' @description
#' Factory function to create an NdaDataStructure by computing metadata
#' from actual data in a column
#'
#' @param column_name Character - column name
#' @param column_data Vector - column data
#' @param selection_order Integer - selection order
#' @param description Character - optional description override
#' @return NdaDataStructure object
#' @keywords internal
#' @noRd
nda_structure_from_data <- function(column_name, column_data, selection_order = NULL, description = NULL) {
  # Note: compute_field_metadata is defined inside createNdaDataDefinition function
  # This factory function is currently unused but kept for potential future use
  # For now, use simple defaults since this isn't called at runtime
  computed <- list(
    data_type = "String",
    size = 255,
    required = "Recommended",
    description = "",
    valueRange = ""
  )
  
  # Create typed objects
  element_name <- ElementName$new(column_name)
  data_type <- DataType$new(computed$data_type)
  size <- Size$new(computed$size, computed$data_type)
  required <- RequirementLevel$new(computed$required %||% "No")
  element_description <- Description$new(description %||% computed$description %||% "")
  value_range <- ValueRange$new(computed$valueRange %||% "")
  notes <- Notes$new("")
  aliases <- Aliases$new("")
  
  # Create source metadata
  source_metadata <- SourceMetadata$new(primary_source = "computed_from_data")
  
  # Create missing info if we have data
  missing_info <- NULL
  if (!is.null(column_data)) {
    total_count <- length(column_data)
    missing_count <- sum(is.na(column_data))
    
    missing_info <- MissingInfo$new(
      missing_count = missing_count,
      total_count = total_count
    )
  }
  
  # Create structure
  structure <- NdaDataStructure$new(
    element_name = element_name,
    data_type = data_type,
    size = size,
    required = required,
    element_description = element_description,
    value_range = value_range,
    notes = notes,
    aliases = aliases,
    selection_order = selection_order,
    source_metadata = source_metadata,
    missing_info = missing_info
  )
  
  structure
}

#' Create NdaDataStructure from REDCap field
#'
#' @description
#' Factory function to create an NdaDataStructure from REDCap data dictionary
#'
#' @param redcap_field List - REDCap field from data dictionary
#' @param column_data Vector - actual data from column (optional, for missing info)
#' @param selection_order Integer - selection order
#' @return NdaDataStructure object
#' @keywords internal
#' @noRd
nda_structure_from_redcap <- function(redcap_field, column_data = NULL, selection_order = NULL) {
  if (is.null(redcap_field) || !is.list(redcap_field)) {
    stop("redcap_field must be a list")
  }
  
  # Extract field name
  field_name <- redcap_field$field_name %||% redcap_field$name %||% stop("redcap_field must have 'field_name'")
  
  # Map REDCap field type to NDA data type
  redcap_type <- redcap_field$field_type %||% "text"
  nda_type <- switch(redcap_type,
    "text" = "String",
    "notes" = "String",
    "calc" = "Float",
    "dropdown" = "Integer",
    "radio" = "Integer",
    "checkbox" = "Integer",
    "yesno" = "Integer",
    "truefalse" = "Integer",
    "slider" = "Integer",
    "String"  # default
  )
  
  # Create typed objects
  element_name <- ElementName$new(field_name)
  data_type <- DataType$new(nda_type)
  
  # Infer size from validation or default
  size_val <- NULL
  if (nda_type == "String") {
    size_val <- redcap_field$text_validation_max %||% 255
  }
  size <- Size$new(size_val, nda_type)
  
  # Map required
  required_val <- if (isTRUE(redcap_field$required_field) || redcap_field$required_field == "y") {
    "Required"
  } else {
    "No"
  }
  required <- RequirementLevel$new(required_val)
  
  # Description
  desc <- redcap_field$field_label %||% redcap_field$field_note %||% ""
  element_description <- Description$new(desc)
  
  # Value range from choices
  choices <- redcap_field$select_choices_or_calculations %||% ""
  value_range <- ValueRange$new(choices)
  
  # Notes (can include choice labels)
  notes <- Notes$new(choices)
  
  # No aliases from REDCap
  aliases <- Aliases$new("")
  
  # Source metadata
  source_metadata <- SourceMetadata$new(primary_source = "redcap")
  source_metadata$mark_redcap_match()
  
  # Missing info if we have data
  missing_info <- NULL
  if (!is.null(column_data)) {
    total_count <- length(column_data)
    missing_count <- sum(is.na(column_data))
    missing_info <- MissingInfo$new(
      missing_count = missing_count,
      total_count = total_count
    )
  }
  
  # Create structure
  structure <- NdaDataStructure$new(
    element_name = element_name,
    data_type = data_type,
    size = size,
    required = required,
    element_description = element_description,
    value_range = value_range,
    notes = notes,
    aliases = aliases,
    selection_order = selection_order,
    source_metadata = source_metadata,
    missing_info = missing_info
  )
  
  structure
}

#' Parse composite Qualtrics survey into component data frames by variable prefix
#'
#' This function fetches a Qualtrics data frame containing multiple surveys and
#' separates it into individual data frames for each survey detected in the data.
#' It identifies the appropriate identifier column (e.g., participantId, workerId)
#' and splits the data based on column name prefixes.
#'
#' @param qualtrics_alias Character string specifying the Qualtrics survey alias to retrieve.
#' @param prefix Character string; default NULL, if specified returns only the dataframe with this prefix
#' @param institution Character string; default NULL, specify location
#' @param label Logical; default TRUE, returns coded values as labels instead of raw values.
#' @param interview_date Logical or Date String, returns all data before date
#' @param complete Logical; default FALSE, if TRUE only returns rows where Progress == 100
#' @param lower default TRUE convert prefixes to lower case
#'
#' @return Creates multiple dataframes in the global environment, one for each survey
#'   detected in the data. Each dataframe is named after its survey prefix.
#'
#' @details
#' The function performs the following steps:
#' \itemize{
#'   \item Retrieves the raw Qualtrics data using the getSurvey() function
#'   \item Identifies which identifier column to use (participantId, workerId, PROLIFIC_PID, or src_subject_id)
#'   \item Determines survey prefixes by analyzing column names
#'   \item Creates separate dataframes for each survey prefix found
#'   \item Assigns each dataframe to the global environment with names matching the survey prefixes
#' }
#'
#' @examples
#' \dontrun{
#' # Parse a a Qualtrics survey into its component dataframes
#' qualtrics.rune("combined_surveys", label = FALSE)
#'
#' # After running, access individual survey dataframes directly:
#' head(pss)  # Access the PSS survey dataframe
#' head(cesd) # Access the CESD survey dataframe
#' 
#' # Parse a single Qualtrics survey from composite survey
#' rgpts <- qualtrics.rune("combined_surveys", prefix = "rgpts")
#' }
#'
#' @importFrom dplyr filter select
#' @export
qualtrics.rune <- function(qualtrics_alias, prefix = NULL, institution = NULL, label = FALSE, interview_date = NULL, complete = FALSE, lower = TRUE){

  df <- qualtrics(qualtrics_alias, institution = institution, label = label, interview_date = interview_date, complete = complete)

  # Define potential identifiers
  identifiers <- c("participantId", "workerId", "PROLIFIC_PID", "src_subject_id")

  # Filter to keep only existing keys in the dataframe
  existing_keys <- identifiers[identifiers %in% names(df)]
  # Check if any identifiers exist in the dataframe
  if (length(existing_keys) == 0) {
    stop("No valid identifiers found in the dataframe.")
  }

  # Print existing identifiers for debugging
  #print("Existing identifiers:")
  #print(existing_keys)

  # Find the first identifier with non-NA values
  identifier <- NA
  for (key in existing_keys) {
    non_na_count <- sum(!is.na(df[[key]]))

    # Debug print to check how many non-NA values exist in each column
    #print(paste("Checking identifier:", key, "with", non_na_count, "non-NA values"))

    if (non_na_count > 0) {  # As long as there's at least 1 non-NA value
      identifier <- key
      break
    }
  }

  # If no column has any non-NA values, issue a warning or stop
  if (is.na(identifier)) {
    stop(paste("No identifier found without NA values or multiple identifiers exist:", existing_keys))
  }

  # Print the detected identifier for debugging
  message(paste("Detected identifier:", identifier))

  # Define common columns to include if they exist
  common_columns <- c(
    "record_id", "subjectkey", "site", "phenotype", "visit", "week",
    "state", "status", "lost_to_followup", "lost_to_follow-up",
    "interview_age", "interview_date"
  )

  # Filter to only include common columns that exist in the original dataframe
  common_columns_exist <- common_columns[common_columns %in% names(df)]

  # Exclude non-survey and specific columns from survey prefix detection
  non_survey_columns <- c(existing_keys, common_columns)
  survey_columns <- names(df)[!names(df) %in% non_survey_columns & grepl("_", names(df))]

  # Extract unique survey prefixes from survey-specific column names
  extract_first_part <- function(string) {
    parts <- strsplit(string, "_")[[1]]
    return(parts[1])
  }
  survey_prefixes <- unique(sapply(survey_columns, extract_first_part))

  # Exclude prefixes that might still be problematic
  excluded_prefixes <- c("PROLIFIC", "interview")
  survey_prefixes <- survey_prefixes[!survey_prefixes %in% excluded_prefixes]

  # Create a list of dataframes, one for each survey
  output = list()
  for (survey_prefix in survey_prefixes) {
    survey_specific_columns <- grep(paste0("^", survey_prefix, "_"), names(df), value = TRUE)
    if (length(survey_specific_columns) > 0) {
      # Create subset dataframe with identifier, common columns, and survey-specific columns
      all_columns <- c(identifier, common_columns_exist, survey_specific_columns)
      subset_df <- df[, all_columns, drop = FALSE]

      # Apply lowercase transformation if requested
      if (lower) {
        # Always preserve the identifier column name and common columns
        cols_to_transform <- !names(subset_df) %in% c(identifier, common_columns_exist)
        names(subset_df)[cols_to_transform] <- tolower(names(subset_df)[cols_to_transform])
      }

      # Add to output list with lowercase prefix as key
      output[[tolower(survey_prefix)]] <- subset_df
    }
  }

  # If prefix parameter is specified, return only that dataframe
  if (!is.null(prefix)) {
    # Normalize prefix to match the output list keys (lowercase)
    prefix_key <- if (lower) tolower(prefix) else prefix
    if (prefix_key %in% names(output)) {
      return(output[[prefix_key]])
    } else {
      stop(paste("Prefix '", prefix, "' not found. Available prefixes:", paste(names(output), collapse = ", ")))
    }
  }

  # Use parent.frame() instead of globalenv() for CRAN compliance
  list2env(output, parent.frame())
  # Add an explicit return of invisible(NULL) to suppress output
  invisible(NULL)
}

#' Parse composite MongoDB collection into component data frames by variable prefix
#'
#' This function fetches a MongoDB collection containing multiple collections and
#' separates it into individual data frames for each collection detected in the data.
#' It identifies the appropriate identifier column (e.g., participantId, workerId)
#' and splits the data based on column name prefixes.
#'
#' @param collection Character string specifying the Mongo collection
#' @param prefix Character string; default NULL, if specified returns only the dataframe with this prefix
#' @param db_name Character string specifying the Mongo database
#' @param lower default TRUE convert prefixes to lower case
#'
#' @return If prefix is specified, returns a single dataframe with that prefix. Otherwise,
#'   creates multiple dataframes in the global environment, one for each survey
#'   detected in the data. Each dataframe is named after its survey prefix.
#'
#' @details
#' The function performs the following steps:
#' \itemize{
#'   \item Retrieves the raw Qualtrics data using the getSurvey() function
#'   \item Identifies which identifier column to use (participantId, workerId, PROLIFIC_PID, or src_subject_id)
#'   \item Determines survey prefixes by analyzing column names
#'   \item Creates separate dataframes for each survey prefix found
#'   \item Assigns each dataframe to the global environment with names matching the survey prefixes
#' }
#'
#' @examples
#' \dontrun{
#' # Parse a MongoDB collection into its component dataframes
#' mongo.rune("combined_surveys")
#'
#' # After running, access individual survey dataframes directly:
#' head(pss)  # Access the PSS survey dataframe
#' head(cesd) # Access the CESD survey dataframe
#' 
#' # Parse a single survey from composite collection
#' rgpts <- mongo.rune("combined_surveys", prefix = "rgpts")
#' }
#'
#' @importFrom dplyr filter select
#' @export
mongo.rune <- function(collection, prefix = NULL, db_name = NULL, lower = TRUE ){

  df <- mongo(collection, db_name)

  # Define potential identifiers
  identifiers <- c("participantId", "workerId", "PROLIFIC_PID", "src_subject_id")

  # Filter to keep only existing keys in the dataframe
  existing_keys <- identifiers[identifiers %in% names(df)]

  # Check if any identifiers exist in the dataframe
  if (length(existing_keys) == 0) {
    stop("No valid identifiers found in the dataframe.")
  }

  # Print existing identifiers for debugging
  #print("Existing identifiers:")
  #print(existing_keys)

  # Find the first identifier with non-NA values
  identifier <- NA
  for (key in existing_keys) {
    non_na_count <- sum(!is.na(df[[key]]))

    # Debug print to check how many non-NA values exist in each column
    #print(paste("Checking identifier:", key, "with", non_na_count, "non-NA values"))

    if (non_na_count > 0) {  # As long as there's at least 1 non-NA value
      identifier <- key
      break
    }
  }

  # If no column has any non-NA values, issue a warning or stop
  if (is.na(identifier)) {
    stop(paste("No identifier found without NA values or multiple identifiers exist:", existing_keys))
  }

  # Print the detected identifier for debugging
  message(paste("Detected identifier:", identifier))

  # Define common columns to include if they exist
  common_columns <- c(
    "record_id", "subjectkey", "site", "phenotype", "visit", "week",
    "state", "status", "lost_to_followup", "lost_to_follow-up",
    "interview_age", "interview_date"
  )

  # Filter to only include common columns that exist in the original dataframe
  common_columns_exist <- common_columns[common_columns %in% names(df)]

  # Exclude non-survey and specific columns from survey prefix detection
  non_survey_columns <- c(existing_keys, common_columns)
  survey_columns <- names(df)[!names(df) %in% non_survey_columns & grepl("_", names(df))]

  # Extract unique survey prefixes from survey-specific column names
  extract_first_part <- function(string) {
    parts <- strsplit(string, "_")[[1]]
    return(parts[1])
  }
  survey_prefixes <- unique(sapply(survey_columns, extract_first_part))

  # Exclude prefixes that might still be problematic
  excluded_prefixes <- c("PROLIFIC", "interview")
  survey_prefixes <- survey_prefixes[!survey_prefixes %in% excluded_prefixes]

  # Create a list of dataframes, one for each survey
  output = list()
  for (survey_prefix in survey_prefixes) {
    survey_specific_columns <- grep(paste0("^", survey_prefix, "_"), names(df), value = TRUE)
    if (length(survey_specific_columns) > 0) {
      # Create subset dataframe with identifier, common columns, and survey-specific columns
      all_columns <- c(identifier, common_columns_exist, survey_specific_columns)
      subset_df <- df[, all_columns, drop = FALSE]

      # Apply lowercase transformation if requested
      if (lower) {
        # Always preserve the identifier column name and common columns
        cols_to_transform <- !names(subset_df) %in% c(identifier, common_columns_exist)
        names(subset_df)[cols_to_transform] <- tolower(names(subset_df)[cols_to_transform])
      }

      # Add to output list with lowercase prefix as key
      output[[tolower(survey_prefix)]] <- subset_df
    }
  }

  # If prefix parameter is specified, return only that dataframe
  if (!is.null(prefix)) {
    # Normalize prefix to match the output list keys (lowercase)
    prefix_key <- if (lower) tolower(prefix) else prefix
    if (prefix_key %in% names(output)) {
      return(output[[prefix_key]])
    } else {
      stop(paste("Prefix '", prefix, "' not found. Available prefixes:", paste(names(output), collapse = ", ")))
    }
  }

  # Use parent.frame() instead of globalenv() for CRAN compliance
  list2env(output, parent.frame())
  # Add an explicit return of invisible(NULL) to suppress output
  invisible(NULL)
}

#' Parse composite REDCap instrument into component data frames by variable prefix
#'
#' This function fetches a REDCap instrument and separates it into individual data frames
#' for each survey/collection detected in the data based on column name prefixes.
#' It identifies the appropriate identifier column and splits the data accordingly.
#'
#' @param instrument_name Name of the REDCap instrument
#' @param prefix Character string; default NULL, if specified returns only the dataframe with this prefix
#' @param raw_or_label Whether to return raw or labeled values
#' @param redcap_event_name Optional event name filter. Can be a single string
#'        or a vector of event names (e.g., \code{c("event1", "event2")})
#' @param batch_size Number of records to retrieve per batch
#' @param records Optional vector of specific record IDs
#' @param fields Optional vector of specific fields
#' @param pii Logical; if FALSE (default), remove fields marked as PII. TRUE keeps PII.
#' @param interview_date Optional; date filtering parameter
#' @param date_format Default ymd define date format for interview_date
#' @param lower default TRUE convert prefixes to lower case
#'
#' @return If prefix is specified, returns a single dataframe with that prefix. Otherwise,
#'   creates multiple dataframes in the parent environment, one for each survey
#'   detected in the data. Each dataframe is named after its survey prefix.
#'
#' @examples
#' \dontrun{
#' # Parse a REDCap instrument into its component dataframes
#' redcap.rune("baseline_assessment")
#'
#' # After running, access individual survey dataframes directly:
#' head(pss)  # Access the PSS survey dataframe
#' head(cesd) # Access the CESD survey dataframe
#' 
#' # Parse a single survey from composite instrument
#' rgpts <- redcap.rune("baseline_assessment", prefix = "rgpts")
#' }
#'
#' @export
redcap.rune <- function(instrument_name, prefix = NULL, raw_or_label = "raw",
                        redcap_event_name = NULL, batch_size = 1000,
                        records = NULL, fields = NULL, pii = FALSE,
                        interview_date = NULL, date_format = "ymd", lower = TRUE) {
  # Fetch the data using redcap()
  df <- redcap(instrument_name, raw_or_label, redcap_event_name, batch_size,
               records, fields, pii, interview_date, date_format)

  # Define potential identifiers
  identifiers <- c("record_id", "participantId", "workerId", "PROLIFIC_PID", "src_subject_id")

  # Filter to keep only existing keys in the dataframe
  existing_keys <- identifiers[identifiers %in% names(df)]

  # Check if any identifiers exist in the dataframe
  if (length(existing_keys) == 0) {
    stop("No valid identifiers found in the dataframe.")
  }

  # Find the first identifier with non-NA values
  identifier <- NA
  for (key in existing_keys) {
    non_na_count <- sum(!is.na(df[[key]]))
    if (non_na_count > 0) {  # As long as there's at least 1 non-NA value
      identifier <- key
      break
    }
  }

  # If no column has any non-NA values, issue a warning or stop
  if (is.na(identifier)) {
    stop(paste("No identifier found without NA values or multiple identifiers exist:", existing_keys))
  }

  # Print the detected identifier for debugging
  message(paste("Detected identifier:", identifier))

  # Define common columns to include if they exist
  common_columns <- c(
    "record_id", "subjectkey", "site", "phenotype", "visit", "week",
    "state", "status", "lost_to_followup", "lost_to_follow-up",
    "interview_age", "interview_date", "redcap_event_name"
  )

  # Filter to only include common columns that exist in the original dataframe
  common_columns_exist <- common_columns[common_columns %in% names(df)]

  # Exclude non-survey and specific columns from survey prefix detection
  non_survey_columns <- c(existing_keys, common_columns)
  survey_columns <- names(df)[!names(df) %in% non_survey_columns & grepl("_", names(df))]

  # Extract unique survey prefixes from survey-specific column names
  extract_first_part <- function(string) {
    parts <- strsplit(string, "_")[[1]]
    return(parts[1])
  }
  survey_prefixes <- unique(sapply(survey_columns, extract_first_part))

  # Exclude prefixes that might still be problematic
  excluded_prefixes <- c("PROLIFIC", "interview", "redcap")
  survey_prefixes <- survey_prefixes[!survey_prefixes %in% excluded_prefixes]

  # Create a list of dataframes, one for each survey
  output = list()
  for (survey_prefix in survey_prefixes) {
    survey_specific_columns <- grep(paste0("^", survey_prefix, "_"), names(df), value = TRUE)
    if (length(survey_specific_columns) > 0) {
      # Create subset dataframe with identifier, common columns, and survey-specific columns
      all_columns <- c(identifier, common_columns_exist, survey_specific_columns)
      # Make sure all columns exist in df
      all_columns <- all_columns[all_columns %in% names(df)]
      subset_df <- df[, all_columns, drop = FALSE]

      # Apply lowercase transformation if requested
      if (lower) {
        # Always preserve the identifier column name and common columns
        cols_to_transform <- !names(subset_df) %in% c(identifier, common_columns_exist)
        names(subset_df)[cols_to_transform] <- tolower(names(subset_df)[cols_to_transform])
      }

      # Add to output list with lowercase prefix as key if requested
      if (lower) {
        output[[tolower(survey_prefix)]] <- subset_df
      } else {
        output[[survey_prefix]] <- subset_df
      }
    }
  }

  # If prefix parameter is specified, return only that dataframe
  if (!is.null(prefix)) {
    # Normalize prefix to match the output list keys
    prefix_key <- if (lower) tolower(prefix) else prefix
    if (prefix_key %in% names(output)) {
      return(output[[prefix_key]])
    } else {
      stop(paste("Prefix '", prefix, "' not found. Available prefixes:", paste(names(output), collapse = ", ")))
    }
  }

  # Use parent.frame() instead of globalenv() for CRAN compliance
  list2env(output, parent.frame())

  # Add an explicit return of invisible(NULL) to suppress output
  invisible(NULL)
}

#' Parse composite data frame into component data frames by variable prefix
#'
#' This function takes a data frame containing multiple measures and
#' separates it into individual data frames for each measure detected in the data.
#' It identifies the appropriate identifier column (e.g., participantId, workerId)
#' and splits the data based on column name prefixes.
#'
#' @param df a dataframe containing multiple, prefixed measures
#' @param prefix Character string; default NULL, if specified returns only the dataframe with this prefix
#' @param lower default TRUE convert prefixes to lower case
#'
#' @return If prefix is specified, returns a single dataframe with that prefix. Otherwise,
#'   creates multiple dataframes in the global environment, one for each survey
#'   detected in the data. Each dataframe is named after its survey prefix.
#'
#' @details
#' The function performs the following steps:
#' \itemize{
#'   \item Identifies which identifier column to use (participantId, workerId, PROLIFIC_PID, or src_subject_id)
#'   \item Determines survey prefixes by analyzing column names
#'   \item Creates separate dataframes for each survey prefix found
#'   \item Assigns each dataframe to the global environment with names matching the survey prefixes
#' }
#'
#' @examples
#' # Parse a data frame containing multiple surveys
#' combined_df <- data.frame(
#'   record_id = c("REC001", "REC002", "REC003", "REC004"),
#'   src_subject_id = c("SUB001", "SUB002", "SUB003", "SUB004"),
#'   subjectkey = c("KEY001", "KEY002", "KEY003", "KEY004"),
#'   site = c("Yale", "NU", "Yale", "NU"),
#'   phenotype = c("A", "B", "A", "C"),
#'   visit = c(1, 2, 2, 1),
#'   state = c("complete", "completed baseline", "in progress", NA),
#'   status = c(NA, NA, NA, "complete"),
#'   lost_to_followup = c(FALSE, FALSE, TRUE, NA),
#'   interview_date = c("2023-01-15", "2023/02/20", NA, "2023-03-10"),
#'   foo_1 = c(1, 3, 5, 7),
#'   foo_2 = c("a", "b", "c", "d"),
#'   bar_1 = c(2, 4, 6, 8),
#'   bar_2 = c("w", "x", "y", "z")
#' )
#' rune(combined_df)
#'
#' # After running, access individual survey dataframes directly:
#' head(foo)  # Access the foo dataframe
#' head(bar)  # Access the bar dataframe
#' 
#' # Parse a single survey from composite dataframe
#' foo_df <- rune(combined_df, prefix = "foo")
#'
#' @importFrom dplyr filter select
#' @export
rune <- function(df, prefix = NULL, lower = TRUE){

  # Define potential identifiers
  identifiers <- c("participantId", "workerId", "PROLIFIC_PID", "src_subject_id")

  # Filter to keep only existing keys in the dataframe
  existing_keys <- identifiers[identifiers %in% names(df)]

  # Check if any identifiers exist in the dataframe
  if (length(existing_keys) == 0) {
    stop("No valid identifiers found in the dataframe.")
  }

  # Print existing identifiers for debugging
  #print("Existing identifiers:")
  #print(existing_keys)

  # Find the first identifier with non-NA values
  identifier <- NA
  for (key in existing_keys) {
    non_na_count <- sum(!is.na(df[[key]]))

    # Debug print to check how many non-NA values exist in each column
    #print(paste("Checking identifier:", key, "with", non_na_count, "non-NA values"))

    if (non_na_count > 0) {  # As long as there's at least 1 non-NA value
      identifier <- key
      break
    }
  }

  # If no column has any non-NA values, issue a warning or stop
  if (is.na(identifier)) {
    stop(paste("No identifier found without NA values or multiple identifiers exist:", existing_keys))
  }

  # Print the detected identifier for debugging
  message(paste("Detected identifier:", identifier))

  # Define common columns to include if they exist
  common_columns <- c(
    "record_id", "subjectkey", "site", "phenotype", "visit", "week",
    "state", "status", "lost_to_followup", "lost_to_follow-up",
    "interview_age", "interview_date"
  )

  # Filter to only include common columns that exist in the original dataframe
  common_columns_exist <- common_columns[common_columns %in% names(df)]

  # Exclude non-survey and specific columns from survey prefix detection
  non_survey_columns <- c(existing_keys, common_columns)
  survey_columns <- names(df)[!names(df) %in% non_survey_columns & grepl("_", names(df))]

  # Extract unique survey prefixes from survey-specific column names
  extract_first_part <- function(string) {
    parts <- strsplit(string, "_")[[1]]
    return(parts[1])
  }
  survey_prefixes <- unique(sapply(survey_columns, extract_first_part))

  # Exclude prefixes that might still be problematic
  excluded_prefixes <- c("PROLIFIC", "interview")
  survey_prefixes <- survey_prefixes[!survey_prefixes %in% excluded_prefixes]

  # Create a list of dataframes, one for each survey
  output = list()
  for (survey_prefix in survey_prefixes) {
    survey_specific_columns <- grep(paste0("^", survey_prefix, "_"), names(df), value = TRUE)
    if (length(survey_specific_columns) > 0) {
      # Create subset dataframe with identifier, common columns, and survey-specific columns
      all_columns <- c(identifier, common_columns_exist, survey_specific_columns)
      subset_df <- df[, all_columns, drop = FALSE]

      # Apply lowercase transformation if requested
      if (lower) {
        # Always preserve the identifier column name and common columns
        cols_to_transform <- !names(subset_df) %in% c(identifier, common_columns_exist)
        names(subset_df)[cols_to_transform] <- tolower(names(subset_df)[cols_to_transform])
      }

      # Add to output list with lowercase prefix as key
      output[[tolower(survey_prefix)]] <- subset_df
    }
  }

  # If prefix parameter is specified, return only that dataframe
  if (!is.null(prefix)) {
    # Normalize prefix to match the output list keys (lowercase)
    prefix_key <- if (lower) tolower(prefix) else prefix
    if (prefix_key %in% names(output)) {
      return(output[[prefix_key]])
    } else {
      stop(paste("Prefix '", prefix, "' not found. Available prefixes:", paste(names(output), collapse = ", ")))
    }
  }

  # Use parent.frame() instead of globalenv() for CRAN compliance
  list2env(output, parent.frame())
}

# ============================================================================
# UTILITY FUNCTIONS
# ============================================================================

#' Safe Readline with Error Handling
#' @param prompt Character - prompt to display
#' @param default Character - default value if error/empty
#' @return Character - user input or default
#' @noRd
safe_readline <- function(prompt, default = "") {
  result <- tryCatch({
    readline(prompt = prompt)
  }, error = function(e) {
    return(default)
  }, interrupt = function(e) {
    message("\nInput cancelled (Esc pressed)")
    return(default)
  })
  if (is.null(result) || trimws(result) == "") default else result
}

#' Detect Fields Not in NDA Structure
#' @param df Data frame
#' @param elements NDA structure dataElements
#' @param verbose Logical
#' @return Character vector of unexpected field names
#' @noRd
detect_unexpected_fields <- function(df, elements, verbose = TRUE) {
  df_cols <- names(df)
  valid_fields <- elements$name
  unexpected <- setdiff(df_cols, valid_fields)
  
  if (length(unexpected) > 0 && verbose) {
    message(sprintf("\n[FIELD MAPPING] Detected %d unexpected field(s) not in NDA structure",
                   length(unexpected)))
  }
  
  return(unexpected)
}

#' Check if Field is Part of REDCap Matrix
#' @description
#' REDCap matrices have a matrix_group_name in the data dictionary.
#' We should skip auto-renaming these as they follow a specific structure.
#' Falls back to heuristic if dictionary unavailable.
#' 
#' @param field_name Character - field to check
#' @param api Character - API type (only applies to "redcap")
#' @param measure_name Character - instrument name for REDCap dictionary lookup
#' @return Logical - TRUE if field is part of a matrix
#' @noRd
is_redcap_matrix_field <- function(field_name, api, measure_name) {
  # Only applies to REDCap
  if (tolower(api) != "redcap") {
    return(FALSE)
  }
  
  # Try to fetch REDCap dictionary
  redcap_dict <- tryCatch({
    redcap.dict(measure_name)
  }, error = function(e) {
    return(NULL)
  })
  
  if (is.null(redcap_dict)) {
    # Fallback heuristic: fieldname_N or fieldname_N_N pattern
    # More than 1 numeric part suggests matrix structure
    parts <- strsplit(field_name, "_")[[1]]
    num_parts <- sum(grepl("^\\d+$", parts))
    return(num_parts > 1)
  }
  
  # Check if field has matrix_group_name defined
  if (field_name %in% redcap_dict$field_name) {
    matrix_group <- redcap_dict$matrix_group_name[redcap_dict$field_name == field_name]
    return(!is.na(matrix_group) && nchar(matrix_group) > 0)
  }
  
  return(FALSE)
}

# ============================================================================
# SIMILARITY & COMPATIBILITY
# ============================================================================

#' Calculate Similarity Scores for a Field
#' @param field_name Character - field to match
#' @param valid_fields Character vector - NDA structure fields
#' @param verbose Logical
#' @return Named numeric vector of similarity scores (sorted descending)
#' @noRd
calculate_field_similarities <- function(field_name, valid_fields, verbose = TRUE) {
  # Calculate Levenshtein similarity for all valid fields
  similarities <- sapply(valid_fields, function(name) {
    calculate_levenshtein(field_name, name)
  })
  
  # Remove NA values
  similarities <- similarities[!is.na(similarities)]
  
  # Sort by similarity (descending)
  similarities <- sort(similarities, decreasing = TRUE)
  
  return(similarities)
}

#' Check Value Range Compatibility for Rename
#' @param source_field Character - source field name
#' @param target_field Character - target field name
#' @param elements NDA structure dataElements
#' @param interactive_mode Logical
#' @param verbose Logical
#' @return Logical - TRUE if rename should proceed, FALSE otherwise
#' @noRd
check_rename_compatibility <- function(source_field, target_field, elements,
                                      interactive_mode = TRUE, verbose = TRUE) {
  # Get value ranges
  source_range <- get_field_value_range(source_field, elements)
  target_range <- get_field_value_range(target_field, elements)
  
  # Check compatibility
  range_check <- check_value_range_compatibility(
    source_range, target_range, 
    source_field, target_field, 
    verbose
  )
  
  if (!range_check$compatible) {
    if (verbose) {
      message(sprintf("\n*** VALUE RANGE CONFLICT DETECTED:"))
      message(sprintf("   Source field '%s' range: %s", source_field, source_range))
      message(sprintf("   Target field '%s' range: %s", target_field, target_range))
      message(sprintf("   %s", range_check$message))
      message(sprintf("   Recommendation: %s", range_check$recommendation))
    }
    
    if (interactive_mode) {
      proceed_input <- safe_readline(
        prompt = "Proceed with rename despite conflict? (y/n): ",
        default = "n"
      )
      while (!tolower(proceed_input) %in% c("y", "n")) {
        proceed_input <- safe_readline(
          prompt = "Please enter 'y' or 'n': ",
          default = "n"
        )
      }
      return(tolower(proceed_input) == "y")
    } else {
      if (verbose) message("Skipping rename due to conflict (non-interactive mode)")
      return(FALSE)
    }
  }
  
  return(TRUE)
}

# ============================================================================
# USER PROMPTS
# ============================================================================

#' Prompt User for High Confidence Rename (>90% similarity)
#' @param field_name Character - source field
#' @param best_match Character - suggested target field
#' @param similarity Numeric - similarity score (0-1)
#' @param elements NDA structure dataElements
#' @param interactive_mode Logical
#' @param verbose Logical
#' @return List with: action ("rename", "keep", "drop"), target (field name if rename)
#' @noRd
prompt_high_confidence_rename <- function(field_name, best_match, similarity,
                                         elements, interactive_mode = TRUE,
                                         verbose = TRUE) {
  result <- list(action = "keep", target = NULL)
  
  if (!interactive_mode) {
    # Non-interactive: auto-rename high confidence matches if compatible
    if (check_rename_compatibility(field_name, best_match, elements, FALSE, verbose)) {
      result$action <- "rename"
      result$target <- best_match
    }
    return(result)
  }
  
  # Interactive: prompt user
  rename_input <- safe_readline(
    prompt = sprintf("Rename '%s' to '%s' (%.1f%% match)? (y/n/m for manual): ",
                    field_name, best_match, similarity * 100),
    default = "y"
  )
  
  # Validate input
  while (!tolower(rename_input) %in% c("y", "n", "m")) {
    rename_input <- safe_readline(
      prompt = "Please enter 'y', 'n', or 'm': ",
      default = "y"
    )
  }
  
  if (tolower(rename_input) == "m") {
    # Manual entry
    manual_input <- safe_readline(
      prompt = sprintf("Enter the NDA field name for '%s': ", field_name),
      default = ""
    )
    while (manual_input == "" || is.na(manual_input)) {
      manual_input <- safe_readline(
        prompt = sprintf("Please enter a valid field name for '%s': ", field_name),
        default = ""
      )
    }
    result$target <- manual_input
    result$action <- "rename"
    if (verbose) message(sprintf("Using manual field name: '%s'", manual_input))
    
  } else if (tolower(rename_input) == "y") {
    # Check compatibility
    if (check_rename_compatibility(field_name, best_match, elements, interactive_mode, verbose)) {
      result$action <- "rename"
      result$target <- best_match
    } else {
      # Compatibility check failed, ask to drop/keep
      result <- prompt_field_action(field_name, FALSE, interactive_mode, verbose)
    }
  } else {
    # User said no, ask what to do with field
    result <- prompt_field_action(field_name, FALSE, interactive_mode, verbose)
  }
  
  return(result)
}

#' Prompt User for Low Confidence Rename (<90% similarity) with Pagination
#' @param field_name Character - source field
#' @param similarities Named numeric vector - sorted similarity scores
#' @param elements NDA structure dataElements
#' @param interactive_mode Logical
#' @param auto_drop_unknown Logical
#' @param verbose Logical
#' @return List with: action ("rename", "keep", "drop"), target (field name if rename)
#' @noRd
prompt_low_confidence_rename <- function(field_name, similarities, elements,
                                        interactive_mode = TRUE,
                                        auto_drop_unknown = FALSE,
                                        verbose = TRUE) {
  result <- list(action = if(auto_drop_unknown) "drop" else "keep", target = NULL)
  
  if (!interactive_mode) {
    return(result)
  }
  
  # Show paginated options
  page_size <- 5
  total <- length(similarities)
  page <- 1
  
  repeat {
    start_idx <- (page - 1) * page_size + 1
    end_idx <- min(page * page_size, total)
    
    message(sprintf("\nField: %s", field_name))
    message(sprintf("Top matches %d-%d of %d:", start_idx, end_idx, total))
    
    for (i in seq(start_idx, end_idx)) {
      local_num <- i - start_idx + 1
      match_name <- names(similarities)[i]
      match_score <- similarities[i]
      message(sprintf("%d. %s (%.1f%% match)", local_num, match_name, match_score * 100))
    }
    
    prompt_msg <- "Select 1-5, 'n' next, 'p' prev, 'm' manual, Enter skip, Esc quit: "
    user_input <- safe_readline(prompt = prompt_msg, default = "")
    
    # Empty = skip
    if (is.null(user_input) || trimws(user_input) == "") {
      break
    }
    
    # Manual entry
    if (tolower(user_input) == "m") {
      manual_input <- safe_readline(
        prompt = sprintf("Enter the NDA field name for '%s': ", field_name),
        default = ""
      )
      while (manual_input == "" || is.na(manual_input)) {
        manual_input <- safe_readline(
          prompt = "Please enter a valid field name: ",
          default = ""
        )
      }
      result$target <- manual_input
      result$action <- "rename"
      if (verbose) message(sprintf("Using manual field name: '%s'", manual_input))
      break
    }
    
    # Pagination
    if (tolower(user_input) == "n") {
      if (page * page_size < total) {
        page <- page + 1
      } else {
        message("Already at last page.")
      }
      next
    }
    
    if (tolower(user_input) == "p") {
      if (page > 1) {
        page <- page - 1
      } else {
        message("Already at first page.")
      }
      next
    }
    
    # Number selection (1-5)
    if (grepl("^[1-5]$", user_input)) {
      match_idx <- as.integer(user_input)
      if (match_idx >= 1 && match_idx <= (end_idx - start_idx + 1)) {
        global_index <- start_idx + match_idx - 1
        selected_match <- names(similarities)[global_index]
        
        if (verbose) {
          message(sprintf("Selected: %s (%.1f%% match)", 
                     selected_match, similarities[global_index] * 100))
        }
        
        # Check compatibility
        if (check_rename_compatibility(field_name, selected_match, elements, interactive_mode, verbose)) {
          result$action <- "rename"
          result$target <- selected_match
        } else {
          # Compatibility failed, ask what to do
          result <- prompt_field_action(field_name, auto_drop_unknown, interactive_mode, verbose)
        }
        break
      } else {
        message(sprintf("Please enter a number between 1 and %d", (end_idx - start_idx + 1)))
        next
      }
    }
    
    message("Invalid input. Choose 1-5, 'n', 'p', 'm', or Enter.")
  }
  
  # If user skipped pagination, ask what to do with field
  if (result$action != "rename") {
    result <- prompt_field_action(field_name, auto_drop_unknown, interactive_mode, verbose)
  }
  
  return(result)
}

#' Prompt User for Field Action (Keep or Drop)
#' @param field_name Character - field name
#' @param auto_drop_unknown Logical - default to drop
#' @param interactive_mode Logical
#' @param verbose Logical
#' @return List with: action ("keep" or "drop"), target = NULL
#' @noRd
prompt_field_action <- function(field_name, auto_drop_unknown = FALSE,
                               interactive_mode = TRUE, verbose = TRUE) {
  result <- list(action = if(auto_drop_unknown) "drop" else "keep", target = NULL)
  
  if (!interactive_mode) {
    if (verbose && result$action == "drop") {
      message(sprintf("Dropping field '%s' (auto_drop_unknown=TRUE)", field_name))
    }
    return(result)
  }
  
  drop_input <- safe_readline(
    prompt = sprintf("Drop field '%s'? (y/n): ", field_name),
    default = if(auto_drop_unknown) "y" else "n"
  )
  
  while (!tolower(drop_input) %in% c("y", "n")) {
    drop_input <- safe_readline(
      prompt = "Please enter 'y' or 'n': ",
      default = if(auto_drop_unknown) "y" else "n"
    )
  }
  
  if (tolower(drop_input) == "y") {
    result$action <- "drop"
    if (verbose) message(sprintf("Will drop field '%s'", field_name))
  } else {
    result$action <- "keep"
    if (verbose) message(sprintf("Keeping field '%s'", field_name))
  }
  
  return(result)
}

# ============================================================================
# FIELD OPERATIONS
# ============================================================================

#' Apply Field Rename to Data Frame
#' @param df Data frame
#' @param source_field Character - old name
#' @param target_field Character - new name
#' @param verbose Logical
#' @return Data frame with renamed field
#' @noRd
apply_field_rename <- function(df, source_field, target_field, verbose = TRUE) {
  # Safely handle special characters in field names
  field_data <- df[[source_field]]
  df[[target_field]] <- field_data
  
  if (verbose) {
    message(sprintf("[RENAME] '%s' -> '%s'", source_field, target_field))
  }
  
  return(df)
}

# ============================================================================
# MAIN ORCHESTRATOR
# ============================================================================

#' Process Unexpected Fields with Interactive Mapping
#' @description
#' Main orchestrator for handling fields in data that don't exist in NDA structure.
#' Prompts user to rename, keep, or drop each unexpected field.
#' 
#' @param df Data frame
#' @param elements NDA structure dataElements
#' @param structure_name Character - NDA structure name
#' @param measure_name Character - measure name
#' @param api Character - API type (redcap, qualtrics, etc.)
#' @param verbose Logical
#' @param auto_drop_unknown Logical
#' @param interactive_mode Logical
#' @return List with:
#'   - df: Updated data frame
#'   - renames: Named character vector (old_name = new_name)
#'   - renamed_fields: Character vector of original field names that were renamed
#'   - columns_to_drop: Character vector of fields to drop
#' @noRd
process_unexpected_fields <- function(df, elements, structure_name, measure_name, api,
                                     verbose = TRUE, auto_drop_unknown = FALSE,
                                     interactive_mode = TRUE) {
  result <- list(
    df = df,
    renames = character(),
    renamed_fields = character(),
    columns_to_drop = character()
  )
  
  # Detect unexpected fields
  unexpected_fields <- detect_unexpected_fields(df, elements, verbose)
  
  if (length(unexpected_fields) == 0) {
    if (verbose) message("[FIELD MAPPING] No unexpected fields detected")
    return(result)
  }
  
  if (verbose) {
    message(sprintf("\n=== Interactive Field Mapping ==="))
    message("Processing unexpected fields...")
  }
  
  # Process each unexpected field
  for (field in unexpected_fields) {
    if (verbose) message(sprintf("\n[%d/%d] Processing field: '%s'", 
                                which(unexpected_fields == field),
                                length(unexpected_fields), field))
    
    # Check for special characters
    has_special_chars <- grepl("[#\\$%&\\*\\+/:<=>\\?@\\[\\\\\\]\\^\\{\\|\\}~]", field)
    if (has_special_chars && verbose) {
      message(sprintf("  Note: Field contains special characters"))
    }
    
    # Check if REDCap matrix field (skip renaming)
    if (is_redcap_matrix_field(field, api, measure_name)) {
      if (verbose) {
        matrix_group <- get_matrix_group_name(field, api, measure_name)
        if (nchar(matrix_group) > 0) {
          message(sprintf("  Skipping - REDCap matrix field (group: %s)", matrix_group))
        } else {
          message(sprintf("  Skipping - REDCap matrix field detected"))
        }
      }
      next
    }

    # Auto-drop DCC and internal wizaRdry fields — never belong in NDA structures
    if (field %in% NDAR_SKIP_FIELDS) {
      result$columns_to_drop <- c(result$columns_to_drop, field)
      next
    }

    # Calculate similarity scores
    similarities <- calculate_field_similarities(field, elements$name, verbose = FALSE)
    
    if (length(similarities) == 0) {
      # No matches found
      if (verbose) message("  No similarity matches found")
      action <- prompt_field_action(field, auto_drop_unknown, interactive_mode, verbose)
      
      if (action$action == "drop") {
        result$columns_to_drop <- c(result$columns_to_drop, field)
      }
      next
    }
    
    # Get best match
    best_match <- names(similarities)[1]
    best_score <- similarities[1]
    
    # High confidence match (>90%)
    if (best_score > 0.9) {
      if (verbose) {
        message(sprintf("  High confidence match: %s (%.1f%%)", best_match, best_score * 100))
      }
      
      action <- prompt_high_confidence_rename(
        field, best_match, best_score, elements,
        interactive_mode, verbose
      )
      
    } else {
      # Low confidence match (<90%)
      if (verbose) {
        message(sprintf("  Best match below 90%% threshold (%.1f%%)", best_score * 100))
      }
      
      action <- prompt_low_confidence_rename(
        field, similarities, elements,
        interactive_mode, auto_drop_unknown, verbose
      )
    }
    
    # Execute action
    if (action$action == "rename") {
      result$df <- apply_field_rename(result$df, field, action$target, verbose)
      result$renamed_fields <- c(result$renamed_fields, field)
      # Store as named vector: old_name = new_name
      result$renames <- c(result$renames, stats::setNames(action$target, field))
    } else if (action$action == "drop") {
      result$columns_to_drop <- c(result$columns_to_drop, field)
    }
    # action == "keep" means do nothing (field stays as-is, becomes new field in structure)
  }
  
  # Summary
  if (verbose && (length(result$renamed_fields) > 0 || length(result$columns_to_drop) > 0)) {
    message(sprintf("\n[FIELD MAPPING] Summary:"))
    if (length(result$renamed_fields) > 0) {
      message(sprintf("  - Renamed: %d field(s)", length(result$renamed_fields)))
      for (i in seq_along(result$renames)) {
        old_name <- names(result$renames)[i]
        new_name <- result$renames[i]
        message(sprintf("    - %s -> %s", old_name, new_name))
      }
    }
    if (length(result$columns_to_drop) > 0) {
      message(sprintf("  - Dropped: %d field(s)", length(result$columns_to_drop)))
      message(sprintf("    - %s", paste(result$columns_to_drop, collapse = ", ")))
    }
  }
  
  return(result)
}

# ============================================================================
# SCRIPT UPDATE (AUTO-INJECTION)
# ============================================================================

#' Update Cleaning Script with Field Operations
#' @description
#' Injects or updates auto-generated section in cleaning script with
#' rename and drop operations from interactive field mapping.
#' 
#' @param script_path Character - path to cleaning script (e.g., "./nda/redcap/cde_phq901.R")
#' @param measure_name Character - dataframe name
#' @param renames Named character vector - old_name = new_name
#' @param drops Character vector - fields to drop
#' @param verbose Logical
#' @return Logical - TRUE if successful, FALSE otherwise
#' @noRd
update_cleaning_script <- function(script_path, measure_name, renames, drops, assignments = c(), verbose = TRUE) {
  # Check if script exists
  if (!file.exists(script_path)) {
    if (verbose) message(sprintf("[SCRIPT UPDATE] Script not found: %s", script_path))
    return(FALSE)
  }

  # Read script
  script_lines <- readLines(script_path, warn = FALSE)

  # Find existing auto-generated section first so we can parse and preserve mappings
  auto_gen_start <- grep("^# ========== AUTO-GENERATED: Field Operations", script_lines)
  auto_gen_end <- grep("^# =====================================================================", script_lines)

  # Parse existing renames and drops so they survive the rewrite
  existing_renames <- c()
  existing_drops <- character(0)

  if (length(auto_gen_start) > 0 && length(auto_gen_end) > 0) {
    section_lines <- script_lines[auto_gen_start[1]:auto_gen_end[length(auto_gen_end)]]

    for (line in section_lines) {
      # Rename lines carry a trailing comment "# Renamed: old -> new"
      m <- regmatches(line, regexec("# Renamed: (\\w+) -> (\\w+)", line))[[1]]
      if (length(m) == 3) {
        existing_renames[m[2]] <- m[3]   # names = old field, value = new field
        next
      }

      # Drop lines contain "!names(measure_name) %in% c(...)"
      drop_m <- regmatches(line, regexec(
        paste0("!names\\(", measure_name, "\\)\\s*%in%\\s*c\\(([^)]+)\\)"), line))[[1]]
      if (length(drop_m) == 2) {
        field_strs <- strsplit(drop_m[2], ",")[[1]]
        fields <- gsub("['\"]", "", trimws(field_strs))
        existing_drops <- c(existing_drops, fields[nchar(fields) > 0])
      }
    }
  }

  # Merge: incoming values take precedence; existing fill in everything else
  merged_renames <- existing_renames
  for (nm in names(renames)) merged_renames[nm] <- renames[nm]

  merged_drops <- unique(c(existing_drops, drops))

  # Generate new auto-generated section using the fully merged mappings
  new_section <- generate_auto_generated_section(measure_name, merged_renames, merged_drops, assignments)
  
  if (length(auto_gen_start) > 0 && length(auto_gen_end) > 0) {
    # Replace existing section
    before_section <- if (auto_gen_start[1] > 1) script_lines[1:(auto_gen_start[1] - 1)] else character()
    after_section <- if (auto_gen_end[length(auto_gen_end)] < length(script_lines)) {
      script_lines[(auto_gen_end[length(auto_gen_end)] + 1):length(script_lines)]
    } else {
      character()
    }
    
    new_script <- c(before_section, "", new_section, "", after_section)
  } else {
    # Find insertion point (before final line, which is usually blank or comment)
    # Insert before the last non-empty line
    insert_point <- length(script_lines)
    while (insert_point > 1 && trimws(script_lines[insert_point]) == "") {
      insert_point <- insert_point - 1
    }
    
    before_section <- script_lines[1:insert_point]
    after_section <- if (insert_point < length(script_lines)) {
      script_lines[(insert_point + 1):length(script_lines)]
    } else {
      character()
    }
    
    new_script <- c(before_section, "", new_section, "", after_section)
  }
  
  # Write back to file
  tryCatch({
    writeLines(new_script, script_path)
    if (verbose) {
      message(sprintf("[SCRIPT UPDATE] Updated cleaning script: %s", script_path))
      message("[SCRIPT UPDATE] Changes will persist on next run")
    }
    return(TRUE)
  }, error = function(e) {
    if (verbose) message(sprintf("[SCRIPT UPDATE] Failed to write script: %s", e$message))
    return(FALSE)
  })
}

#' Generate Auto-Generated Section Code
#' @param measure_name Character - dataframe name
#' @param renames Named character vector - old_name = new_name
#' @param drops Character vector - fields to drop
#' @return Character vector - lines of code
#' @noRd
generate_auto_generated_section <- function(measure_name, renames, drops, assignments = c()) {
  lines <- character()

  # Header
  lines <- c(lines, "# ========== AUTO-GENERATED: Field Operations (DO NOT EDIT) ==========")

  # Renames section
  if (length(renames) > 0) {
    lines <- c(lines, "# Renames (from interactive field mapping)")
    for (i in seq_along(renames)) {
      old_name <- names(renames)[i]
      new_name <- renames[i]
      lines <- c(lines, sprintf("%s$%s <- %s$%s  # Renamed: %s -> %s",
                               measure_name, new_name, measure_name, old_name, old_name, new_name))
      lines <- c(lines, sprintf("%s$%s <- NULL  # Remove original field", measure_name, old_name))
    }
    lines <- c(lines, "")
  }

  # Assignments section (required field values added automatically)
  if (length(assignments) > 0) {
    lines <- c(lines, "# Assignments (required field values \u2014 added automatically by wizaRdry)")
    for (i in seq_along(assignments)) {
      field <- names(assignments)[i]
      value <- assignments[i]
      lines <- c(lines, sprintf("%s$%s <- %s", measure_name, field, value))
    }
    lines <- c(lines, "")
  }

  # Drops section
  if (length(drops) > 0) {
    lines <- c(lines, "# Drops (from interactive field mapping + standard cleanup)")
    drop_vector <- paste0("c('", paste(drops, collapse = "', '"), "')")
    lines <- c(lines, sprintf("%s <- %s[, !names(%s) %%in%% %s, drop = FALSE]",
                             measure_name, measure_name, measure_name, drop_vector))
  } else {
    lines <- c(lines, "# No fields marked for drop")
  }

  # Footer
  lines <- c(lines, "# =====================================================================")

  return(lines)
}

#' Check for Presence of Key Variables in a Data Frame
#' 
#' This function checks whether specified variables exist as column names in a given data frame.
#' It reports which variables are present and which are missing, and returns the list of 
#' present variables for further processing.
#' 
#' @param df_name Character string. The name of the data frame to check (must exist in the global environment)
#' @param variables Character vector. Names of the variables (columns) to check for in the data frame
#' @param type Character string. A descriptor of the data frame type, used in output messages (default: "data")
#' 
#' @return Character vector containing the names of the variables that are present in the data frame
#' 
#' @details 
#' The function retrieves the data frame from the global environment using \code{base::get()}.
#' It then compares the specified variables with the column names of the data frame.
#' The function prints messages about which variables are present and which are missing.
#' Unlike earlier versions that would stop execution if variables were missing, this version 
#' continues execution and simply reports the missing variables.
#' 
#' @examples
#' # Create a sample data frame
#' test_df <- data.frame(id = 1:5, name = letters[1:5], value = rnorm(5))
#' 
#' # Check for the presence of specific keys
#' present_vars <- checkKeys("test_df", c("id", "name", "missing_var"), type = "test")
#' 
#' # Use the returned present variables for further operations
#' if("id" %in% present_vars) {
#'   # Proceed with operations requiring the 'id' column
#' }
#' 
#' @noRd
checkKeys <- function(df_name, variables, type = "data") {
  # Attempt to retrieve the dataframe from the global environment
  df <- base::get(df_name)
  
  # Find any missing variables
  missing_vars <- setdiff(variables, names(df))
  # Find present variables
  present_vars <- intersect(variables, names(df))
  
  # Print present variables
  if (length(present_vars) > 0) {
    message(paste("Candidate keys present in ", type, " data of ", df_name, ": ", paste(present_vars, collapse=", "), "."))
  }
  
  # Handle missing variables with tryCatch
  tryCatch({
    if (length(missing_vars) > 0) {
      # Instead of stopping, just print the message about missing variables
      message(paste("Missing variables in", type, "data of", df_name, ":", paste(missing_vars, collapse=", "), ". Merging on present variables:", paste(present_vars, collapse=", "), "."))
    } else {
      # Only print this message if there are no missing variables
      message(paste("All required variables are present in ", type, " data of ", df_name, "."))
    }
  }, error = function(e) {
    message("An unexpected error occurred: ", e$message)
  })
  # Return the list of present variables for further checks
  return(present_vars)
}

#' Create NDA Submission Template
#'
#' This function creates a CSV template file for National Data Archive (NDA) submissions.
#' It extracts the data from a specified data frame and formats it according to NDA requirements,
#' with the structure name split into base name and suffix in the first line.
#' The function will prompt for confirmation before creating the file,
#' with an option to remember the user's preference for future calls.
#'
#' @param df Data frame to be used as template or character string naming a data frame
#'        in the global environment.
#' @param path Character string specifying the directory path where the "tmp" folder
#'        and template file should be created. Defaults to the current working directory.
#' @param skip_prompt Logical. If TRUE (default), skips the confirmation prompt. If FALSE,
#'        will prompt for confirmation unless the user has previously chosen to remember their preference.
#' @param selected_fields Character vector of field names to include in template. If NULL (default),
#'        uses all fields from data frame. Used by create_nda_files() for centralized field selection.
#' @param skip_prompts Logical. If TRUE, skip ALL interactive prompts (used when called from
#'        create_nda_files() with pre-selected fields). Default: FALSE.
#' @param verbose Logical. If TRUE, show detailed progress messages. Default: FALSE.
#'
#' @return Invisible TRUE if successful. Creates a CSV file at the specified path
#'         and prints a message with the file location.
#'
#' @details
#' The function will:
#' 1. Create a 'tmp' directory if it doesn't exist
#' 2. Parse the structure name into base and suffix components (e.g., "eefrt01" -> "eefrt" and "01")
#' 3. Write the structure name components as the first line
#' 4. Write column headers as the second line
#' 5. Write the data rows below
#'
#' @examples
#' \dontrun{
#'   # First create some sample data
#'   eefrt01 <- data.frame(
#'     src_subject_id = c("SUB001", "SUB002"),
#'     interview_age = c(240, 360),
#'     interview_date = c("01/01/2023", "02/15/2023"),
#'     response_time = c(450, 520)
#'   )
#'
#'   # Create the NDA template using the data frame directly
#'   to.nda(eefrt01)
#'
#'   # Or using the name as a string
#'   to.nda("eefrt01")
#'
#'   # Skip the confirmation prompt
#'   to.nda(eefrt01, skip_prompt = TRUE)
#' }
#'
#' @export
to.nda <- function(df, path = ".", skip_prompt = TRUE, selected_fields = NULL, skip_prompts = FALSE, verbose = FALSE) { #set skip_prompt to TRUE so users need to specify to see prompt after prefs are set
  # Check for user preferences file
  user_prefs_file <- file.path(path, ".wizaRdry_prefs")
  user_prefs <- list(shown_tree = FALSE, auto_create = FALSE, auto_clean = FALSE, auto_nda = FALSE, auto_nda_template = FALSE)

  if (file.exists(user_prefs_file)) {
    tryCatch({
      user_prefs <- readRDS(user_prefs_file)
      # Add the auto_nda_template field if it doesn't exist
      if (is.null(user_prefs$auto_nda_template)) {
        user_prefs$auto_nda_template <- FALSE
      }
    }, error = function(e) {
      # If file exists but can't be read, create a new one
      user_prefs <- list(shown_tree = FALSE, auto_create = FALSE, auto_clean = FALSE, auto_nda = FALSE, auto_nda_template = FALSE)
    })
  }

  # Determine whether df is a data frame or a string
  if (is.character(df)) {
    df_name <- df
    if (!exists(df_name, envir = .GlobalEnv)) {
      stop(paste("Data frame", df_name, "not found in global environment"))
    }
  } else {
    df_name <- deparse(substitute(df))
  }

  # If skip_prompt is TRUE or user has previously set auto_nda_template to TRUE, bypass the prompt
  if (!skip_prompt | !user_prefs$auto_nda_template) {
    response <- readline(prompt = sprintf("Would you like to create the NDA submission template for %s now? y/n ",
                                          df_name))

    while (!tolower(response) %in% c("y", "n")) {
      response <- readline(prompt = "Please enter either y or n: ")
    }

    # If user selects y, save their preference automatically and set auto_nda_template to TRUE.
    if (tolower(response) == "y") {
      user_prefs$auto_nda_template <- TRUE
      saveRDS(user_prefs, user_prefs_file)
    }

    if (tolower(response) == "n") {
      message("NDA submission template creation cancelled.")
      invokeRestart("abort")  # This exits without the "Error:" prefix
    }
  }

  # Create directory structure if it doesn't exist
  tmp_path <- file.path(path, "tmp")
  if (!dir.exists(tmp_path)) {
    dir.create(tmp_path)
  }

  # Define structure_name
  structure_name <- df_name

  # Create the file path
  file_path <- file.path(tmp_path, paste0(structure_name, '_submission.csv'))

  # Get the data frame
  if (is.character(df)) {
    template <- base::get(df, envir = .GlobalEnv)
  } else {
    template <- df
  }

  # Drop old pre-transformed fields (ending with .1, .2, etc.) that correspond to renamed fields
  # These are old versions of fields that were renamed during validation
  template_cols <- names(template)
  old_transformed_fields <- template_cols[grepl("\\.\\d+$", template_cols)]
  
  if (length(old_transformed_fields) > 0) {
    fields_to_drop <- character(0)
    for (old_field in old_transformed_fields) {
      # Extract base name by removing .1, .2, etc. suffix
      base_name <- sub("\\.\\d+$", "", old_field)
      # Check if the base name (without suffix) exists in the template
      if (base_name %in% template_cols) {
        # This is an old version of a renamed field - drop it
        fields_to_drop <- c(fields_to_drop, old_field)
      }
    }
    
    if (length(fields_to_drop) > 0) {
      template <- template[, !names(template) %in% fields_to_drop, drop = FALSE]
      message(sprintf("Dropped %d old pre-transformed field(s): %s", 
                     length(fields_to_drop), paste(fields_to_drop, collapse = ", ")))
      template_cols <- names(template)  # Update after dropping
    }
  }

  # Define fields to exclude from submission templates
  excluded_from_template <- c("state", "lost_to_followup", "lost_to_follow-up")
  
  # Super-required fields (always included): subjectkey, src_subject_id, interview_date, interview_age, sex
  super_required_fields <- SUPER_REQUIRED_FIELDS
  
  # Get structure field names to check which required fields exist in this structure
  structure_field_names <- character(0)
  ndar_required_in_structure <- character(0)
  
  # Try to fetch the structure to see which fields it contains
  tryCatch({
    nda_base_url <- "https://nda.nih.gov/api/datadictionary/v2"
    structure_url <- sprintf("%s/datastructure/%s", nda_base_url, structure_name)
    structure_response <- httr::GET(structure_url, httr::timeout(10))
    if (httr::status_code(structure_response) == 200) {
      raw_content <- rawToChar(structure_response$content)
      if (nchar(raw_content) > 0) {
        structure_data <- jsonlite::fromJSON(raw_content)
        if ("dataElements" %in% names(structure_data)) {
          structure_field_names <- structure_data$dataElements$name
        }
      }
    }
  }, error = function(e) {
    # Silently fail if structure fetch fails
  })
  
  # Get required fields from ndar_subject01 that exist in the current structure
  if (length(structure_field_names) > 0) {
    tryCatch({
      nda_base_url <- "https://nda.nih.gov/api/datadictionary/v2"
      url <- sprintf("%s/datastructure/ndar_subject01", nda_base_url)
      
      response <- httr::GET(url, httr::timeout(10))
      if (httr::status_code(response) == 200) {
        raw_content <- rawToChar(response$content)
        if (nchar(raw_content) > 0) {
          subject_structure <- jsonlite::fromJSON(raw_content)
          if ("dataElements" %in% names(subject_structure)) {
            ndar_required <- subject_structure$dataElements[subject_structure$dataElements$required == "Required", ]
            if (nrow(ndar_required) > 0) {
              ndar_required_names <- ndar_required$name
              # Get required fields that exist in structure but are not super-required
              ndar_required_in_structure <- setdiff(
                intersect(ndar_required_names, structure_field_names),
                super_required_fields
              )
            }
          }
        }
      }
    }, error = function(e) {
      # Silently fail if API call fails
    })
  }
  
  # Automatically include super-required fields if they exist in structure but not in template
  missing_super_required <- setdiff(
    intersect(super_required_fields, structure_field_names),
    template_cols
  )
  
  if (length(missing_super_required) > 0) {
    # Add missing super-required fields with NA values
    for (field in missing_super_required) {
      template[[field]] <- NA
    }
    message(sprintf("\nAutomatically including super-required fields: %s", 
                   paste(missing_super_required, collapse = ", ")))
    template_cols <- names(template)  # Update after adding fields
  }
  
  # Note: DCC required fields and timepoint fields removed - users should include them in their NDA script if needed
  
  # Use selected_fields if provided (from centralized field selection)
  # Otherwise, prompt user in interactive mode
  if (!is.null(selected_fields) && skip_prompts) {
    # Use pre-selected fields from create_nda_files()
    # STEP 1: Add any missing fields from selected_fields to template
    fields_to_add <- setdiff(selected_fields, template_cols)
    if (length(fields_to_add) > 0) {
      for (field in fields_to_add) {
        # Add field with NA values
        # Note: structure_field_names check removed because we trust selected_fields
        # which includes super required fields that may not be in structure yet
        template[[field]] <- NA
      }
    }
    
    # STEP 2: Filter template to ONLY include selected_fields (in the order specified)
    # This removes any fields that were excluded (e.g., DCC fields when dcc=FALSE)
    template <- template[, intersect(selected_fields, names(template)), drop = FALSE]
    template_cols <- names(template)
  } else if (length(ndar_required_in_structure) > 0 && !skip_prompts) {
    # Silent auto-add — required fields are always included
    missing_required_in_structure <- setdiff(ndar_required_in_structure, template_cols)

    if (length(missing_required_in_structure) > 0) {
      message(sprintf("\nAdding %d NDA required field(s) to template: %s",
                      length(missing_required_in_structure),
                      paste(missing_required_in_structure, collapse = ", ")))
      for (field in missing_required_in_structure) template[[field]] <- NA
      template_cols <- names(template)
    }
  }
  
  # Get essential NDA fields dynamically from ndar_subject01 API
  # This ensures we always have the current required elements
  essential_nda_fields <- tryCatch({
    if (verbose) {
      message("Fetching current ndar_subject01 required elements from NDA API...")
    }
    nda_base_url <- "https://nda.nih.gov/api/datadictionary/v2"
    url <- sprintf("%s/datastructure/ndar_subject01", nda_base_url)
    
    response <- httr::GET(url, httr::timeout(10))
    if (httr::status_code(response) == 200) {
      raw_content <- rawToChar(response$content)
      if (nchar(raw_content) > 0) {
        subject_structure <- jsonlite::fromJSON(raw_content)
        if ("dataElements" %in% names(subject_structure)) {
          required_metadata <- subject_structure$dataElements[subject_structure$dataElements$required == "Required", ]
          if (nrow(required_metadata) > 0) {
            field_names <- required_metadata$name
            if (verbose) {
              message(sprintf("Found %d required ndar_subject01 elements for template", length(field_names)))
            }
            field_names
          } else {
            stop("No required elements found")
          }
        } else {
          stop("No dataElements in API response")
        }
      } else {
        stop("Empty API response")
      }
    } else {
      stop(sprintf("API request failed with status %d", httr::status_code(response)))
    }
  }, error = function(e) {
    message("Error fetching ndar_subject01 elements: ", e$message)
    message("Using fallback essential fields list")
    c("src_subject_id", "subjectkey", "sex", "interview_age", "interview_date", 
      "phenotype", "site", "race", "handedness", "visit", "week", "redcap_event_name")
  })
  
  # Filter template columns:
  # Just remove explicitly excluded fields
  # The dataframe has already been cleaned up by the nda script and DCC removal logic,
  # so we trust that all remaining fields should be in the template
  excluded_cols <- template_cols[template_cols %in% excluded_from_template]
  filtered_cols <- template_cols[!template_cols %in% excluded_from_template]
  
  template <- template[, filtered_cols, drop = FALSE]
  
  # Report what was filtered
  if (length(excluded_cols) > 0) {
    message(sprintf("Excluded %d fields from submission template: %s", 
                    length(excluded_cols), paste(excluded_cols, collapse = ", ")))
  }

  # Open a connection to overwrite the file
  con <- file(file_path, "w")

  # Split structure name into base name and suffix
  # Check if the name ends with 2 digits
  if (grepl("\\d{2}$", structure_name)) {
    structure_short_name <- substr(structure_name, 1, nchar(structure_name) - 2)
    structure_suffix <- substr(structure_name, nchar(structure_name) - 1, nchar(structure_name))
  } else {
    # If no suffix, use whole name and "01" as default
    structure_short_name <- structure_name
    structure_suffix <- "01"
  }

  # Remove columns with NA names BEFORE writing anything (prevents parser warnings)
  valid_cols <- !is.na(names(template))
  if (!all(valid_cols)) {
    invalid_count <- sum(!valid_cols)
    if (invalid_count > 0) {
      message(sprintf("Note: Removing %d column(s) with invalid (NA) names", invalid_count))
    }
    template <- template[, valid_cols, drop = FALSE]
  }

  # Write the line with separated components
  writeLines(paste0(structure_short_name, ",", structure_suffix), con)

  # Write column headers manually (now with NA columns already removed)
  writeLines(paste(names(template), collapse = ","), con)

  # Close the connection to save changes
  close(con)

  # Append the data without column headers
  write.table(template, file_path, row.names = FALSE, col.names = FALSE, append = TRUE,
              quote = TRUE, sep = ",", na = "")

  # Message removed - caller (ndaFileCreation.R) handles user messaging
  return(invisible(TRUE))
}

#' Alias for 'to.nda'
#'
#' This is a legacy alias for the 'to.nda' function to maintain compatibility with older code.
#'
#' @inheritParams to.nda
#' @inherit to.nda return
#' @noRd
#' @examples
#' \dontrun{
#' createNda(prl01)
#' }
createNdaSubmissionTemplate <- to.nda

#' ValidationState R6 Class
#'
#' @description
#' Manages NDA validation state and tracks modifications to data structures.
#' Central object for tracking validation results, modifications, and determining
#' whether data definition files need to be created.
#'
#' @details
#' This class replaces the fragile attribute-passing pattern used previously.
#' It provides a structured way to track:
#' - Validation status (valid/invalid, new/existing structure)
#' - Value range violations
#' - New fields added
#' - Required field status
#' - Metadata from ndar_subject01
#'
#' The key method `needs_data_definition()` determines whether a data definition
#' file should be created based on whether the structure is new or modified.
#'
#' @keywords internal
ValidationState <- R6::R6Class("ValidationState",
  public = list(
    #' @field measure_name Character - name of the measure/structure
    measure_name = NULL,
    
    #' @field api Character - API type (redcap, qualtrics, mongo, etc.)
    api = NULL,
    
    #' @field data_env DataEnvironment - manages dataframe across environments
    data_env = NULL,
    
    #' @field nda_structure List - NDA structure definition from API
    nda_structure = NULL,
    
    #' @field is_valid Logical - whether validation passed
    is_valid = TRUE,
    
    #' @field is_new_structure Logical - whether structure is new (not in NDA)
    is_new_structure = FALSE,
    
    #' @field is_modified_structure Logical - whether existing structure has modifications
    is_modified_structure = FALSE,
    
    #' @field bypassed_validation Logical - whether validation was bypassed (new structures)
    bypassed_validation = FALSE,
    
    #' @field value_range_violations List - fields with value range violations
    value_range_violations = list(),
    
    #' @field new_fields Character vector - fields in data not in NDA structure
    new_fields = character(),
    
    #' @field ndar_subject_additions Character vector - DCC fields added from ndar_subject01
    ndar_subject_additions = character(),
    
    #' @field ndar_subject01_all_fields Character vector - ALL field names from ndar_subject01 (~150 fields)
    #' Used for consistent formatting in Excel exports regardless of dcc parameter
    ndar_subject01_all_fields = character(),
    
    #' @field renamed_fields Character vector - fields that were renamed
    renamed_fields = character(),
    
    #' @field dropped_fields Character vector - fields that were dropped
    dropped_fields = character(),
    
    #' @field missing_required Character vector - required fields with missing data
    missing_required = character(),
    
    #' @field required_metadata Data frame - ndar_subject01 required field metadata
    required_metadata = NULL,
    
    #' @field recommended_metadata Data frame - ndar_subject01 recommended field metadata
    recommended_metadata = NULL,
    
    #' @field warnings Character vector - warning messages
    warnings = character(),
    
    #' @field errors Character vector - error messages
    errors = character(),
    
    #' @field dcc Logical - whether DCC fields should be validated
    dcc = FALSE,
    
    #' @description
    #' Create a new ValidationState instance
    #' @param measure_name Name of the measure/structure
    #' @param api API type (redcap, qualtrics, mongo, csv, oracle, sql)
    #' @param df Initial dataframe
    #' @param nda_structure NDA structure definition (NULL for new structures)
    #' @param dcc Logical - whether DCC fields should be validated
    #' @return A new ValidationState object
    initialize = function(measure_name, api, df, nda_structure = NULL, dcc = FALSE) {
      if (missing(measure_name) || is.null(measure_name) || measure_name == "") {
        stop("measure_name is required and cannot be empty")
      }
      if (missing(api) || is.null(api) || api == "") {
        stop("api is required and cannot be empty")
      }
      if (missing(df) || is.null(df) || !is.data.frame(df)) {
        stop("df must be a valid data.frame")
      }
      
      self$measure_name <- measure_name
      self$api <- api
      self$data_env <- DataEnvironment$new(measure_name, df)
      self$nda_structure <- nda_structure
      self$dcc <- dcc
      
      # Determine if this is a new structure
      self$is_new_structure <- is.null(nda_structure) || 
                               is.null(nda_structure$dataElements) ||
                               nrow(nda_structure$dataElements) == 0
    },
    
    #' @description
    #' Get current dataframe
    #' @return Data frame
    get_df = function() {
      self$data_env$get_df()
    },
    
    #' @description
    #' Update dataframe in all environments
    #' @param df New dataframe
    #' @return Self (invisibly) for method chaining
    set_df = function(df) {
      self$data_env$set_df(df)
      invisible(self)
    },
    
    #' @description
    #' Add a value range violation
    #' @param field Field name
    #' @param expected Expected value range (NULL if no range defined)
    #' @param actual Vector of violating values
    #' @return Self (invisibly) for method chaining
    add_value_range_violation = function(field, expected, actual) {
      self$value_range_violations[[field]] <- list(
        expected = expected,
        actual = actual,
        count = length(actual)
      )
      self$is_modified_structure <- TRUE
      
      # Log the violation for debugging
      message(sprintf("[VIOLATION] Field '%s': expected=%s, found %d violating value(s)",
                     field, 
                     if(is.null(expected)) "NULL (no range defined)" else expected,
                     length(actual)))
      
      invisible(self)
    },
    
    #' @description
    #' Add violations of a specific type (e.g., DCC violations)
    #' @param type Character - type of violations ("dcc_required", "dcc_recommended", etc.)
    #' @param violations List - violations to add
    #' @return Self (invisibly) for method chaining
    add_violations = function(type, violations) {
      # Store violations by type (not currently used in reporting but available for future use)
      # For now, just track that we have violations
      if (length(violations) > 0) {
        # Could store them in a typed list if needed:
        # self$typed_violations[[type]] <- violations
      }
      invisible(self)
    },
    
    #' @description
    #' Set validation status
    #' @param valid Logical - TRUE if validation passed, FALSE otherwise
    #' @return Self (invisibly) for method chaining
    set_valid = function(valid) {
      self$is_valid <- valid
      invisible(self)
    },
    
    #' @description
    #' Check if structure has modifications requiring data definition
    #' @return Logical
    has_modifications = function() {
      length(self$new_fields) > 0 || 
      length(self$value_range_violations) > 0 ||
      length(self$ndar_subject_additions) > 0
    },
    
    #' @description
    #' Determine if data definition file is needed
    #' @return Logical - TRUE if data definition should be created
    needs_data_definition = function() {
      self$is_new_structure || self$has_modifications()
    },
    
    #' @description
    #' Get human-readable modification reason
    #' @return Character string describing why structure is modified
    get_modification_reason = function() {
      if (self$is_new_structure) {
        return("new structure")
      }
      
      reasons <- character()
      
      if (length(self$new_fields) > 0) {
        reasons <- c(reasons, sprintf("%d new field%s", 
                                     length(self$new_fields),
                                     if(length(self$new_fields) > 1) "s" else ""))
      }
      
      if (length(self$ndar_subject_additions) > 0) {
        reasons <- c(reasons, sprintf("%d DCC field%s added", 
                                     length(self$ndar_subject_additions),
                                     if(length(self$ndar_subject_additions) > 1) "s" else ""))
      }
      
      if (length(self$value_range_violations) > 0) {
        reasons <- c(reasons, sprintf("%d value range difference%s", 
                                     length(self$value_range_violations),
                                     if(length(self$value_range_violations) > 1) "s" else ""))
      }
      
      if (length(reasons) == 0) {
        return("unmodified")
      }
      
      paste(reasons, collapse = " and ")
    },
    
    #' @description
    #' Convert to list for backward compatibility with old validation_results
    #' @return List with validation results
    to_list = function() {
      list(
        valid = self$is_valid,
        df = self$get_df(),
        message = self$get_modification_reason(),
        bypassed_validation = self$bypassed_validation,
        value_range_violations = self$value_range_violations,
        missing_required = self$missing_required,
        warnings = self$warnings,
        errors = self$errors
      )
    },
    
    #' @description
    #' Print method for ValidationState
    #' @return Self (invisibly)
    print = function() {
      cat("ValidationState:\n")
      cat(sprintf("  Measure: %s\n", self$measure_name))
      cat(sprintf("  API: %s\n", self$api))
      cat(sprintf("  Status: %s\n", if(self$is_valid) "VALID" else "INVALID"))
      cat(sprintf("  Structure Type: %s\n", 
                  if(self$is_new_structure) "NEW" 
                  else if(self$is_modified_structure) "MODIFIED" 
                  else "EXISTING"))
      
      if (!self$is_new_structure) {
        cat(sprintf("  Modified: %s\n", 
                    if(self$is_modified_structure) "YES" else "NO"))
      }
      
      if (length(self$value_range_violations) > 0) {
        cat(sprintf("  Value Range Violations: %d field(s)\n", 
                    length(self$value_range_violations)))
      }
      
      if (length(self$new_fields) > 0) {
        cat(sprintf("  New Fields: %d\n", length(self$new_fields)))
      }
      
      if (length(self$missing_required) > 0) {
        cat(sprintf("  Missing Required: %d field(s)\n", 
                    length(self$missing_required)))
      }
      
      cat(sprintf("  Needs Data Definition: %s\n", 
                  if(self$needs_data_definition()) "YES" else "NO"))
      
      cat(sprintf("  Modification Reason: %s\n", self$get_modification_reason()))
      
      invisible(self)
    }
  )
)

#' Create NDA Submission Template
#'
#' This function creates a CSV template file for National Data Archive (NDA) submissions.
#' It extracts the data from a specified data frame and formats it according to NDA requirements,
#' with the structure name split into base name and suffix in the first line.
#' The function will prompt for confirmation before creating the file,
#' with an option to remember the user's preference for future calls.
#'
#' @param df Data frame to be used as template or character string naming a data frame
#'        in the global environment.
#' @param path Character string specifying the directory path where the "tmp" folder
#'        and template file should be created. Defaults to the current working directory.
#' @param skip_prompt Logical. If TRUE (default), skips the confirmation prompt. If FALSE,
#'        will prompt for confirmation unless the user has previously chosen to remember their preference.
#'
#' @return Invisible TRUE if successful. Creates a CSV file at the specified path
#'         and prints a message with the file location.
#'
#' @details
#' The function will:
#' 1. Create a 'tmp' directory if it doesn't exist
#' 2. Parse the structure name into base and suffix components (e.g., "eefrt01" -> "eefrt" and "01")
#' 3. Write the structure name components as the first line
#' 4. Write column headers as the second line
#' 5. Write the data rows below
#'
#' @examples
#' \dontrun{
#'   # First create some sample data
#'   eefrt01 <- data.frame(
#'     src_subject_id = c("SUB001", "SUB002"),
#'     interview_age = c(240, 360),
#'     interview_date = c("01/01/2023", "02/15/2023"),
#'     response_time = c(450, 520)
#'   )
#'
#'   # Create the NDA template using the data frame directly
#'   to.nda(eefrt01)
#'
#'   # Or using the name as a string
#'   to.nda("eefrt01")
#'
#'   # Skip the confirmation prompt
#'   to.nda(eefrt01, skip_prompt = TRUE)
#' }
#'
#' @export
to.nda <- function(df, path = ".", skip_prompt = TRUE) { #set skip_prompt to TRUE so users need to specify to see prompt after prefs are set
  # Check for user preferences file
  user_prefs_file <- file.path(path, ".wizaRdry_prefs")
  user_prefs <- list(shown_tree = FALSE, auto_create = FALSE, auto_clean = FALSE, auto_nda = FALSE, auto_nda_template = FALSE)

  if (file.exists(user_prefs_file)) {
    tryCatch({
      user_prefs <- readRDS(user_prefs_file)
      # Add the auto_nda_template field if it doesn't exist
      if (is.null(user_prefs$auto_nda_template)) {
        user_prefs$auto_nda_template <- FALSE
      }
    }, error = function(e) {
      # If file exists but can't be read, create a new one
      user_prefs <- list(shown_tree = FALSE, auto_create = FALSE, auto_clean = FALSE, auto_nda = FALSE, auto_nda_template = FALSE)
    })
  }

  # Determine whether df is a data frame or a string
  if (is.character(df)) {
    df_name <- df
    if (!exists(df_name, envir = .GlobalEnv)) {
      stop(paste("Data frame", df_name, "not found in global environment"))
    }
  } else {
    df_name <- deparse(substitute(df))
  }

  # If skip_prompt is TRUE or user has previously set auto_nda_template to TRUE, bypass the prompt
  if (!skip_prompt | !user_prefs$auto_nda_template) {
    response <- readline(prompt = sprintf("Would you like to create the NDA submission template for %s now? y/n ",
                                          df_name))

    while (!tolower(response) %in% c("y", "n")) {
      response <- readline(prompt = "Please enter either y or n: ")
    }

    # If user selects y, save their preference automatically and set auto_nda_template to TRUE.
    if (tolower(response) == "y") {
      user_prefs$auto_nda_template <- TRUE
      saveRDS(user_prefs, user_prefs_file)
    }

    if (tolower(response) == "n") {
      message("NDA submission template creation cancelled.")
      invokeRestart("abort")  # This exits without the "Error:" prefix
    }
  }

  # Create directory structure if it doesn't exist
  tmp_path <- file.path(path, "tmp")
  if (!dir.exists(tmp_path)) {
    dir.create(tmp_path)
  }

  # Define structure_name
  structure_name <- df_name

  # Create the file path
  file_path <- file.path(tmp_path, paste0(structure_name, '_submission.csv'))

  # Get the data frame
  if (is.character(df)) {
    template <- base::get(df, envir = .GlobalEnv)
  } else {
    template <- df
  }

  # Open a connection to overwrite the file
  con <- file(file_path, "w")

  # Split structure name into base name and suffix
  # Check if the name ends with 2 digits
  if (grepl("\\d{2}$", structure_name)) {
    structure_short_name <- substr(structure_name, 1, nchar(structure_name) - 2)
    structure_suffix <- substr(structure_name, nchar(structure_name) - 1, nchar(structure_name))
  } else {
    # If no suffix, use whole name and "01" as default
    structure_short_name <- structure_name
    structure_suffix <- "01"
  }

  # Write the line with separated components
  writeLines(paste0(structure_short_name, ",", structure_suffix), con)

  # Write column headers manually
  writeLines(paste(names(template), collapse = ","), con)

  # Close the connection to save changes
  close(con)

  # Append the data without column headers
  write.table(template, file_path, row.names = FALSE, col.names = FALSE, append = TRUE,
              quote = TRUE, sep = ",", na = "")

  message(sprintf("\nSubmission Template created at: %s \n", file_path))

  return(invisible(TRUE))
}

#' Alias for 'to.nda'
#'
#' This is a legacy alias for the 'to.nda' function to maintain compatibility with older code.
#'
#' @inheritParams to.nda
#' @inherit to.nda return
#' @noRd
#' @examples
#' \dontrun{
#' createNda(prl01)
#' }
createNdaSubmissionTemplate <- to.nda

# globals.R
# Global variables used by wizaRdry package (CRAN compliant)

utils::globalVariables(c(
  # Existing variables
  "connectionString", "apiKeys", "baseUrls", "config",
  "uri", "token", "dob", "surveyIds", "pb", "value",
  "nda_base_url",
  # Additional variables
  "duplicates", "nda_base_url", "src_institution_id",
  "subject_dob", "mongo_conn", "checkKeys", "createExtract",
  "View", "write.csv", "write.table", "ymd", "mdy",
  "SUPER_REQUIRED_FIELDS",  # NDA super required fields constant
  "NdaDataStructure",  # R6 class for NDA field definitions
  "DCC_REQUIRED_FIELDS",  # DCC required fields constant
  "DCC_RECOMMENDED_FIELDS",  # DCC recommended fields constant
  "DCC_FIELDS",  # All DCC fields constant
  "NDAR_SKIP_FIELDS"  # DCC + internal fields to auto-drop
))

#' NDA File Creation Helper Functions
#'
#' @description
#' Functions for creating NDA submission templates and data definition files
#' based on ValidationState results.
#'

#' Get educational description for structure type
#'
#' @description
#' Returns a user-friendly explanation of what each structure type means
#' and what actions the researcher needs to take.
#'
#' @param structure_type Character - one of "NEW", "MODIFIED", or "EXISTING"
#' @return Character string with educational description
#' @noRd
get_structure_description <- function(structure_type) {
  descriptions <- list(
    NEW = paste0(
      "This data structure does not yet exist in NDA. The draft data definitions file\n",
      "must be submitted to DCC and approved by NDA. When approved, you can create your\n",
      "submission file."
    ),
    
    EXISTING = paste0(
      "Your data structure already exists in NDA and requires no modifications. You can\n",
      "submit your submission file at any time."
    ),
    
    MODIFIED = paste0(
      "While your data structure already exists in NDA, you have made modifications to\n",
      "value ranges and/or added new elements. Please submit the draft definitions file\n",
      "to DCC. When approved by NDA, you can create your submission file."
    )
  )
  
  return(descriptions[[structure_type]])
}

#' Check if files should be created based on validation state
#'
#' @description
#' Validation gate that determines whether file creation should proceed.
#' In strict mode: validation failure → no files created
#' In lenient mode: validation failure → files created with warnings
#'
#' @param validation_state ValidationState object
#' @param strict Logical - if TRUE, enforce strict validation (no files on failure)
#' @param verbose Logical - print detailed messages
#' @return Logical - TRUE if files should be created, FALSE if validation failed
#' @noRd
should_create_nda_files <- function(validation_state, strict = TRUE, verbose = TRUE) {
  # Only check ValidationState objects (not legacy lists)
  if (!inherits(validation_state, "ValidationState")) {
    return(TRUE)  # Legacy format always proceeds
  }
  
  # LENIENT MODE with validation failures: Create files anyway (with warnings)
  if (!validation_state$is_valid && !strict) {
    if (verbose) {
      message("[WARN] Creating files despite validation failure (lenient mode)")
    }
    return(TRUE)  # Proceed to file creation
  }
  
  # STRICT MODE with validation failures: Skip file creation
  if (!validation_state$is_valid) {
    if (verbose) {
      message("[VALIDATION FAILED] Skipping file creation due to validation errors")
      if (length(validation_state$errors) > 0) {
        message("Validation errors:")
        for (err in validation_state$errors) {
          message(sprintf("  - %s", err))
        }
      }
    } else {
      message("[SKIPPED - Validation failed]")
    }
    return(FALSE)  # Do not create files
  }
  
  return(TRUE)  # Validation passed, create files
}

#' Create NDA files based on ValidationState
#'
#' @description
#' Determines which NDA files to create (submission template and/or data definition)
#' based on whether the structure is new, existing and unmodified, or existing and modified.
#'
#' Decision logic:
#' - NEW structure: Create data definition only (can't submit to non-existent structure)
#' - EXISTING unmodified: Create submission template only
#' - EXISTING modified: Create both submission template and data definition
#'
#' @param validation_state ValidationState object OR legacy list (for backward compatibility)
#' @param measure Measure name (used for legacy list format)
#' @param strict Logical - if TRUE, enforce strict validation (no files on failure)
#' @param verbose Logical - print detailed messages
#' @return Invisible NULL
#' @noRd
create_nda_files <- function(validation_state, measure = NULL, strict = TRUE, verbose = TRUE) {
  
  # Validation gate: Check if files should be created
  if (!should_create_nda_files(validation_state, strict, verbose)) {
    return(invisible(NULL))
  }
  
  # Handle both ValidationState and legacy list formats
  if (inherits(validation_state, "ValidationState")) {
    # NEW path: Use ValidationState
    is_new <- validation_state$is_new_structure
    needs_definition <- validation_state$needs_data_definition()
    reason <- validation_state$get_modification_reason()
    measure_name <- validation_state$measure_name
    nda_structure <- validation_state$nda_structure
    
    # Determine modification type
    modification_type <- if (is_new) {
      "NEW"
    } else if (needs_definition) {
      "MODIFIED"
    } else {
      "EXISTING"
    }
    
    # Print STEP 4 header with structure type
    if (verbose) {
      message(sprintf("\n=== Creating NDA Files for '%s' ===", measure_name))
      message(sprintf("Structure type: %s", if(is_new) "NEW" else "EXISTING"))
      message(sprintf("Modification status: %s", reason))
    } else {
      # Context-aware header based on structure type
      step4_header <- if (is_new) {
        "=== STEP 4: New Data Structure ==="
      } else if (needs_definition) {
        "=== STEP 4: Modified Data Structure ==="
      } else {
        "=== STEP 4: Existing Data Structure ==="
      }
      message(sprintf("\n%s", step4_header))
    }
    
    # Print educational description (non-verbose only)
    if (!verbose) {
      description <- get_structure_description(modification_type)
      message(description)
      message("")  # Blank line
      
      # Add lenient mode warning if validation failed
      if (!strict && !validation_state$is_valid) {
        message("[WARN] Running in lenient mode - creating DRAFT files despite validation errors")
        message("")  # Blank line
      }
    }
    
    # Show modification details for MODIFIED structures (non-verbose only)
    if (!verbose && needs_definition && !is_new) {
      message("")  # Blank line before changes
      message("  Changes detected:")
      
      # Show new fields
      if (length(validation_state$new_fields) > 0) {
        for (field in validation_state$new_fields) {
          message(sprintf("    - Added new element: %s", field))
        }
      }
      
      # Show DCC additions
      if (length(validation_state$ndar_subject_additions) > 0) {
        for (field in validation_state$ndar_subject_additions) {
          message(sprintf("    - Added DCC field: %s", field))
        }
      }
      
      # Show value range changes
      if (length(validation_state$value_range_violations) > 0) {
        for (field in names(validation_state$value_range_violations)) {
          violation <- validation_state$value_range_violations[[field]]
          if (!is.null(violation$expected)) {
            # Extract just the unique violating values for cleaner display
            violating_vals <- unique(violation$actual)
            message(sprintf("    - Updated value range: %s (extended from %s to include: %s)",
                           field,
                           violation$expected,
                           paste(violating_vals, collapse = ", ")))
          } else {
            message(sprintf("    - Added value range for new element: %s", field))
          }
        }
      }
      message("")
    }
    
    # Use centralized field selection (user prompted ONCE)
    field_selection <- select_nda_fields(
      validation_state = validation_state,
      nda_structure = nda_structure,
      verbose = verbose
    )
    
    if (is_new) {
      # NEW structure: Handle submission file based on strict mode
      
      # STEP 4A: Submission Template
      if (!verbose) {
        message("\n=== Submission Template ===")
      }
      
      if (!strict) {
        # LENIENT MODE: Create DRAFT submission file
        tryCatch({
          # Create file
          createNdaSubmissionTemplate(
            measure_name,
            skip_prompt = TRUE,
            selected_fields = field_selection$selected_fields,
            skip_prompts = TRUE,
            verbose = verbose
          )
          
          # Rename to _draft suffix
          original_path <- sprintf("./tmp/%s_submission.csv", measure_name)
          draft_path <- sprintf("./tmp/%s_submission_draft.csv", measure_name)
          if (file.exists(original_path)) {
            file.rename(original_path, draft_path)
          }
          
          if (!verbose) {
            field_count <- length(field_selection$selected_fields)
            message(sprintf("[DRAFT] ./tmp/%s_submission_draft.csv (%d fields)", measure_name, field_count))
          } else {
            message("[DRAFT] Draft submission file created for testing")
          }
        }, error = function(e) {
          warning(sprintf("Error creating draft submission file: %s", e$message))
        })
      } else {
        # STRICT MODE: Skip submission file
        if (!verbose) {
          message("[SKIP] Structure must be registered with NDA before creating submission file")
        } else {
          message("[SKIP] Skipping submission file (structure doesn't exist in NDA yet)")
        }
      }
      
      # STEP 4B: Data Definition (always created)
      if (!verbose) {
        message("")
        message("=== Data Definition ===")
      } else {
        message("\n[NEW STRUCTURE]")
        message("[OK] Creating data definition (for structure registration)")
      }
      
      tryCatch({
        createNdaDataDefinition(
          validation_state,
          selected_fields = field_selection$selected_fields,
          skip_prompts = TRUE,
          verbose = verbose
        )
      }, error = function(e) {
        warning(sprintf("Error creating data definition: %s", e$message))
      })

      # STEP 4C: Subject count (always written for new structures)
      tryCatch({
        df <- validation_state$get_df()
        id_col <- if ("src_subject_id" %in% names(df)) "src_subject_id" else "subjectkey"
        n_subjects <- length(unique(df[[id_col]]))
        count_path <- sprintf("./tmp/%s_subject_count.txt", measure_name)
        writeLines(as.character(n_subjects), count_path)
        if (!verbose) {
          message(sprintf("[OK] ./tmp/%s_subject_count.txt (%d unique subjects)", measure_name, n_subjects))
        } else {
          message(sprintf("[OK] Subject count written: %d unique subjects -> %s", n_subjects, count_path))
        }
      }, error = function(e) {
        warning(sprintf("Error writing subject count: %s", e$message))
      })

      # STEP 4D: Category and description template (always created for new structures)
      tryCatch({
        cat_path <- sprintf("./tmp/%s_category-and-description.xlsx", measure_name)
        wb <- openxlsx::createWorkbook()

        # Sheet 1: Filled template row
        openxlsx::addWorksheet(wb, "Category and Description Templa")
        template_df <- data.frame(
          `Structure Title`  = "",
          `Shortname`        = measure_name,
          `Category (see Categories Tab)` = "",
          `Description(One or two sentences about the purpose of the structure/measure/assessment/questionnaire)` = "",
          `Reference`        = "",
          check.names = FALSE,
          stringsAsFactors = FALSE
        )
        openxlsx::writeData(wb, "Category and Description Templa", template_df, startRow = 1)
        headerStyle <- openxlsx::createStyle(textDecoration = "bold")
        openxlsx::addStyle(wb, "Category and Description Templa", headerStyle,
                           rows = 1, cols = 1:5, gridExpand = TRUE)

        # Sheet 2: Categories reference list
        openxlsx::addWorksheet(wb, "Categories")
        categories_df <- data.frame(
          `Category List` = c(
            "ADHD","Abuse","Acoustics","Activity","Addiction","Adverse Events",
            "Aggression","Anger","Anxiety","Arousal","Attention","Behavior",
            "Body Image","Cognitive","Conflict","Coping","DTI","DTI, MRI, fMRI",
            "Delusions","Demographics","Depression","Diagnostic","EEG","EGG","EMG",
            "ERP","Eating Disorder","Emotions","Enrollment","Entry Form","Evaluated",
            "Exit Form","Experimental","Exposure","Eye Tracking","Fatigue","Fear",
            "Food","Gen Test","Hallucinations","IQ","Life Events","Loneliness","MEG",
            "MRI","Mania","Med History","Mood","Network","Neurophysiology","OCD",
            "Omics","PTSD","Pain","Panic","Parenting","Perception","Personality",
            "Phobia","Phys Characteristics","Phys Exam","Pleasure","Psychophysiology",
            "Psychosis","Quality of Life","Questionnaire","Relationships",
            "Resolve Identifiers","Satisfaction","Seizure","Self-Harm","Sensory",
            "Sexuality","Side Effects","Sleep","Social Adjustment",
            "Social Responsiveness","Social Support","Socioeconomic","Spectroscopy",
            "Speech-Language","Stigma","Stress","Substance Use","Suicide","Summary",
            "Task Based","Taste","Temperament","Trauma","Treatment","Victimization",
            "Violence","Vision","Well-Being","X-Ray"
          ),
          check.names = FALSE,
          stringsAsFactors = FALSE
        )
        openxlsx::writeData(wb, "Categories", categories_df, startRow = 1)
        headerStyle2 <- openxlsx::createStyle(textDecoration = "bold")
        openxlsx::addStyle(wb, "Categories", headerStyle2, rows = 1, cols = 1)

        openxlsx::saveWorkbook(wb, cat_path, overwrite = TRUE)

        if (!verbose) {
          message(sprintf("[OK] ./tmp/%s_category-and-description.xlsx", measure_name))
        } else {
          message(sprintf("[OK] Category/description template created: %s", cat_path))
        }
      }, error = function(e) {
        warning(sprintf("Error creating category/description template: %s", e$message))
      })

    } else {
      # EXISTING structure: Create submission file (and data definition if modified)
      
      # STEP 4A: Submission Template
      if (!verbose) {
        message("\n=== Submission Template ===")
      } else {
        message("\n[EXISTING STRUCTURE]")
      }
      
      # Determine prefix and filename based on structure type and mode
      if (!strict && needs_definition) {
        # DRAFT mode: MODIFIED in lenient mode
        submission_prefix <- "[DRAFT]"
        submission_filename <- sprintf("%s_submission_draft.csv", measure_name)
        should_rename_to_draft <- TRUE
      } else if (!validation_state$is_valid) {
        # WARN mode: EXISTING with validation errors
        submission_prefix <- "[WARN]"
        submission_filename <- sprintf("%s_submission.csv", measure_name)
        should_rename_to_draft <- FALSE
      } else {
        # OK mode: Validation passed
        submission_prefix <- "[OK]"
        submission_filename <- sprintf("%s_submission.csv", measure_name)
        should_rename_to_draft <- FALSE
      }
      
      tryCatch({
        createNdaSubmissionTemplate(
          measure_name,
          skip_prompt = TRUE,  # Skip initial confirmation prompt
          selected_fields = field_selection$selected_fields,
          skip_prompts = TRUE,   # Skip field selection prompts
          verbose = verbose
        )
        
        # Rename to draft if needed
        if (should_rename_to_draft) {
          original_path <- sprintf("./tmp/%s_submission.csv", measure_name)
          draft_path <- sprintf("./tmp/%s_submission_draft.csv", measure_name)
          if (file.exists(original_path)) {
            file.rename(original_path, draft_path)
          }
        }
        
        if (!verbose) {
          field_count <- length(field_selection$selected_fields)
          message(sprintf("%s ./tmp/%s (%d fields)", submission_prefix, submission_filename, field_count))
        } else {
          message("[OK] Submission file created successfully")
        }
      }, error = function(e) {
        warning(sprintf("Error creating submission file: %s", e$message))
      })
      
      # STEP 4B: Data Definition (only if modified)
      if (needs_definition) {
        if (!verbose) {
          message("")  # Blank line
          message("=== Data Definition ===")
        } else {
          message("")  # Blank line
          message(sprintf("[OK] Creating data definition (reason: %s)", reason))
        }
        
        tryCatch({
          createNdaDataDefinition(
            validation_state,
            selected_fields = field_selection$selected_fields,
            skip_prompts = TRUE,
          verbose = verbose
        )
      }, error = function(e) {
        warning(sprintf("Error creating data definition: %s", e$message))
      })
      } else {
        if (!verbose) {
          message("")  # Blank line
          message("=== Data Definition ===")
          message("[SKIP] Structure unmodified (no data definition needed)")
        } else {
          message("")  # Blank line
          message("[SKIP] Skipping data definition (structure unmodified)")
        }
      }
    }
    
    # Add final file creation summary (non-verbose only)
    if (!verbose) {
      message("")  # Blank line
      files_created <- character()
      
      if (!is_new) {
        files_created <- c(files_created, "submission template")
      }
      
      if (needs_definition) {
        files_created <- c(files_created, "data definition")
      }
      

    }
    
  } else {
    # LEGACY path: validation_state is actually a plain list
    if (verbose) {
      message("\n[LEGACY MODE] Using old validation_results format")
    }
    
    # Extract information from legacy format
    is_new <- isTRUE(validation_state$bypassed_validation)
    nda_structure <- attr(validation_state, "nda_structure")
    
    if (is.null(measure)) {
      stop("measure parameter required when using legacy validation_results format")
    }
    
    # Check for modifications in legacy format
    has_violations <- !is.null(validation_state$value_range_violations) && 
                     length(validation_state$value_range_violations) > 0
    
    # Check for new fields
    has_new_fields <- FALSE
    if (!is.null(nda_structure) && "dataElements" %in% names(nda_structure)) {
      df <- validation_state$df
      if (!is.null(df) && is.data.frame(df)) {
        structure_fields <- nda_structure$dataElements$name
        df_fields <- names(df)
        new_fields <- setdiff(df_fields, structure_fields)
        # _complete fields already removed by StandardOutput in ndaRequest.R
        super_required <- SUPER_REQUIRED_FIELDS
        new_fields <- setdiff(new_fields, super_required)
        has_new_fields <- length(new_fields) > 0
      }
    }
    
    needs_definition <- is_new || has_violations || has_new_fields
    
    # Apply same logic as ValidationState path
    if (is_new) {
      if (verbose) message("Creating data definition for new structure (legacy mode)")
      tryCatch({
        submission_template <- list(columns = names(validation_state$df))
        createNdaDataDefinition(submission_template, nda_structure, measure, verbose = verbose)
      }, error = function(e) {
        warning(sprintf("Error creating data definition: %s", e$message))
      })
    } else {
      if (verbose) message("Creating submission template for existing structure (legacy mode)")
      tryCatch({
        createNdaSubmissionTemplate(measure)
      }, error = function(e) {
        warning(sprintf("Error creating submission template: %s", e$message))
      })
      
      if (needs_definition) {
        if (verbose) message("Creating data definition for modified structure (legacy mode)")
        tryCatch({
          submission_template <- list(columns = names(validation_state$df))
          createNdaDataDefinition(submission_template, nda_structure, measure, verbose = verbose)
        }, error = function(e) {
          warning(sprintf("Error creating data definition: %s", e$message))
        })
      } else {
        if (verbose) message("Skipping data definition - unmodified structure (legacy mode)")
      }
    }
  }
  
  invisible(NULL)
}

#' Retrieve Survey Data from Qualtrics
#'
#' @param qualtrics_alias The alias for the Qualtrics survey to be retrieved.
#' @param ... Optional column names to filter for. Only rows with non-missing values
#'        in ALL specified columns will be returned. This is useful for filtering
#'        data to only include complete cases for specific variables of interest.
#' @param institution Optional. The institution name (e.g., "temple" or "nu"). If NULL, all institutions will be searched.
#' @param label Logical indicating whether to return coded values or their associated labels (default is FALSE).
#' @param interview_date Optional; can be either:
#'        - A date string in various formats (ISO, US, etc.) to filter data up to that date
#'        - A boolean TRUE to return only rows with non-NA interview_date values
#' @param complete Logical; default FALSE, if TRUE only returns rows where Progress == 100
#'
#' @return A cleaned and harmonized data frame containing the survey data with superkeys first.
#' @importFrom dplyr %>% select mutate
#' @export
#' @examples
#' \dontrun{
#' # Get survey by alias (will search all institutions)
#' survey_data <- qualtrics("rgpts")
#' }
qualtrics <- function(qualtrics_alias, ..., institution = NULL, label = FALSE, interview_date = NULL, complete = FALSE) {
  # Load necessary source files

  # Validate config
  cfg <- validate_config("qualtrics")

  # Get secrets using get_secret() to keep it secret, keep it safe
  baseUrls <- get_secret("baseUrls")
  apiKeys <- get_secret("apiKeys")

  # Get survey ID
  survey_id <- NULL

  if (!is.null(institution)) {
    # Check if institution exists
    if (!(institution %in% names(cfg$qualtrics$survey_ids))) {
      stop(paste("Institution", institution, "not found in ./config.yml configuration."))
    }

    # Check if survey exists in specified institution
    if (!(qualtrics_alias %in% names(cfg$qualtrics$survey_ids[[institution]]))) {
      stop(paste("Survey", qualtrics_alias, "not found in institution", institution))
    }

    survey_id <- cfg$qualtrics$survey_ids[[institution]][[qualtrics_alias]]
  } else {
    # Search all institutions
    found <- FALSE
    for (inst in names(cfg$qualtrics$survey_ids)) {
      if (qualtrics_alias %in% names(cfg$qualtrics$survey_ids[[inst]])) {
        survey_id <- cfg$qualtrics$survey_ids[[inst]][[qualtrics_alias]]
        institution <- inst
        found <- TRUE
        break
      }
    }

    if (!found) {
      stop(sprintf("Qualtrics survey '%s' not found in any institution.", qualtrics_alias))
    }
  }

  message(sprintf("Retrieving '%s' survey from %s Qualtrics...", qualtrics_alias, institution))

  # Connect to Qualtrics with the specific institution's credentials
  connectQualtrics(institution = institution)

  # Show loading animation (if implemented)
  if (exists("show_loading_animation")) {
    show_loading_animation()
  }

  # Fetch the data
  df <- qualtRics::fetch_survey(
    surveyID = survey_id,
    verbose = FALSE,
    label = label,
    convert = label,
    add_column_map = TRUE
  )

  if (!is.data.frame(df)) {
    stop(paste("fetch_survey did not return a data frame for", qualtrics_alias))
  }

  # Filter for complete responses if requested
  if (complete) {
    if ("Progress" %in% names(df)) {
      original_rows <- nrow(df)
      df <- df[df$Progress == 100, ]
      kept_rows <- nrow(df)
      message(sprintf("Filtered to complete responses (Progress == 100): kept %d of %d rows.", kept_rows, original_rows))
    } else {
      warning("'complete = TRUE' was specified but 'Progress' column not found in data. Returning all rows.")
    }
  }

  # Get identifier from config
  identifier <- cfg$identifier

  # Create a copy of the original dataframe to preserve original values
  original_df <- df

  # Advanced date parsing function that handles multiple formats
  parseAnyDate <- function(date_string) {
    if (is.na(date_string) || is.null(date_string)) {
      return(NA)
    }

    # Try multiple date formats sequentially
    date <- NULL

    # Try ISO format (YYYY-MM-DD)
    if (grepl("^\\d{4}-\\d{1,2}-\\d{1,2}$", date_string)) {
      date <- tryCatch(ymd(date_string), error = function(e) NULL)
    }
    # Try US format (MM/DD/YYYY)
    else if (grepl("^\\d{1,2}/\\d{1,2}/\\d{4}$", date_string)) {
      date <- tryCatch(mdy(date_string), error = function(e) NULL)
    }
    # Try European format (DD.MM.YYYY)
    else if (grepl("^\\d{1,2}\\.\\d{1,2}\\.\\d{4}$", date_string)) {
      date <- tryCatch(dmy(date_string), error = function(e) NULL)
    }
    # Try Canadian format (YYYY/MM/DD)
    else if (grepl("^\\d{4}/\\d{1,2}/\\d{1,2}$", date_string)) {
      date <- tryCatch(ymd(date_string), error = function(e) NULL)
    }
    # Try other format (DD-MM-YYYY)
    else if (grepl("^\\d{1,2}-\\d{1,2}-\\d{4}$", date_string)) {
      date <- tryCatch(dmy(date_string), error = function(e) NULL)
    }
    # Try abbreviated month name (15-Jan-2023 or Jan 15, 2023)
    else if (grepl("[A-Za-z]", date_string)) {
      date <- tryCatch(parse_date_time(date_string, c("dmy", "mdy")), error = function(e) NULL)
    }

    # If all attempts fail, return NA
    if (is.null(date) || all(is.na(date))) {
      warning("Failed to parse date: ", date_string, ". Treating as NA.")
      return(NA)
    }

    return(as.Date(date))
  }

  # Handle interview_date filtering
  if ("interview_date" %in% names(df)) {
    # Create a temporary date column for filtering but don't modify the original
    df$temp_date <- sapply(df$interview_date, parseAnyDate)

    # Handle the interview_date parameter
    if (!is.null(interview_date)) {
      if (is.logical(interview_date) && interview_date == TRUE) {
        # Keep only rows with non-NA interview_date values
        rows_to_keep <- !is.na(df$temp_date)
        df <- df[rows_to_keep, ]
        original_df <- original_df[rows_to_keep, ]
      } else if (is.character(interview_date) || inherits(interview_date, "Date")) {
        # Filter by specific date
        input_date <- tryCatch({
          if (inherits(interview_date, "Date")) {
            interview_date
          } else {
            parseAnyDate(interview_date)
          }
        }, error = function(e) {
          stop("Failed to parse interview_date parameter: ", interview_date)
        })

        if (is.na(input_date)) {
          stop("Failed to parse interview_date parameter: ", interview_date)
        }

        # If we're here with a valid date, keep rows with dates before or equal to input_date
        rows_to_keep <- !is.na(df$temp_date) & df$temp_date <= input_date
        df <- df[rows_to_keep, ]
        original_df <- original_df[rows_to_keep, ]
      } else {
        stop("interview_date must be either a date string or TRUE")
      }
    }

    # Remove temporary column after filtering
    df$temp_date <- NULL
  }

  # Harmonize the data
  clean_df <- qualtricsHarmonization(df, identifier, qualtrics_alias)

  # List of allowed superkey columns to prioritize
  allowed_superkey_cols <- c(
    "record_id",
    "src_subject_id",
    "subjectkey",
    "site",
    "subsiteid",
    "sex",
    "race",
    "ethnic_group",
    "phenotype",
    "phenotype_description",
    "state",
    "status",
    "lost_to_followup",
    "lost_to_follow-up",
    "twins_study",
    "sibling_study",
    "family_study",
    "sample_taken",
    "interview_date",
    "interview_age",
    "visit",
    "week"
  )

  # Reorder columns to have superkeys first
  if (is.data.frame(clean_df) && ncol(clean_df) > 0) {
    # Identify which superkey columns are actually in the data
    present_superkeys <- intersect(allowed_superkey_cols, names(clean_df))

    # Get all other columns (non-superkeys)
    other_cols <- setdiff(names(clean_df), present_superkeys)

    # If there are matching superkeys, reorder the columns
    if (length(present_superkeys) > 0) {
      # Create new column order with superkeys first, then other columns
      new_order <- c(present_superkeys, other_cols)

      # Reorder the dataframe
      clean_df <- clean_df[, new_order, drop = FALSE]
    }
  }

  # Check if any column requests were passed via ...
  dots_args <- list(...)
  if (length(dots_args) > 0) {
    # Convert the dots arguments to a character vector
    requested_cols <- as.character(unlist(dots_args))

    # Find which of the requested columns actually exist in the data
    existing_cols <- intersect(requested_cols, names(clean_df))

    if (length(existing_cols) > 0) {
      # Display the names of the columns that were found
      message(sprintf("Found %d of %d requested columns: %s",
                      length(existing_cols),
                      length(requested_cols),
                      paste(existing_cols, collapse = ", ")))

      # Create a filter to keep only rows where ALL requested columns have data
      rows_to_keep <- rep(TRUE, nrow(clean_df))

      for (col in existing_cols) {
        # Check if column values are not NA
        not_na <- !is.na(df[[col]])

        # For non-NA values, check if they're not empty strings (if character type)
        not_empty <- rep(TRUE, nrow(clean_df))
        if (is.character(clean_df[[col]])) {
          not_empty <- clean_df[[col]] != ""
        }

        # Combine the conditions - both not NA and not empty (if applicable)
        has_data <- not_na & not_empty

        # Update the rows_to_keep vector
        rows_to_keep <- rows_to_keep & has_data
      }

      # Apply the filter to keep only rows with data in all requested columns
      original_rows <- nrow(clean_df)
      clean_df <- clean_df[rows_to_keep, ]
      kept_rows <- nrow(clean_df)

      message(sprintf("Kept %d of %d rows where all requested columns have values.",
                      kept_rows, original_rows))
    } else {
      if (length(requested_cols) > 0) {
        warning("None of the requested columns were found in the dataset.")
      }
    }
  }

  return(clean_df)
}

#################
## Helper Functions
#################

#' Connect to Qualtrics API
#'
#' This helper function sets up the connection to the Qualtrics API using credentials stored in a file or environment variables.
#' It is called internally by the 'qualtrics' function.
#'
#' @importFrom config get
#' @import qualtRics
#' @noRd
connectQualtrics <- function(institution = NULL) {
  # Validate secrets
  validate_secrets("qualtrics")

  # Get secrets using get_secret() to keep it secret, keep it safe
  baseUrls <- get_secret("baseUrls")
  apiKeys <- get_secret("apiKeys")

  if (!exists("apiKeys") || !exists("baseUrls")) {
    stop("apiKeys and/or baseUrls arrays not found in secrets.R")
  }

  if (length(apiKeys) != length(baseUrls)) {
    stop("apiKeys and baseUrls arrays must have the same length.")
  }

  # If institution is specified, try to find the matching credentials
  if (!is.null(institution)) {
    # Get institution names from config
    cfg <- validate_config("qualtrics")
    institution_names <- names(cfg$qualtrics$survey_ids)
    
    # Find the index of the institution
    inst_index <- match(institution, institution_names)
    
    if (!is.na(inst_index) && inst_index <= length(apiKeys)) {
      # Use the specific institution's credentials
      message(sprintf("Connecting to Qualtrics using %s credentials...", institution))
      
      suppressMessages({
        tryCatch({
          result <- qualtRics::qualtrics_api_credentials(
            api_key = apiKeys[inst_index],
            base_url = baseUrls[inst_index],
            install = TRUE,
            overwrite = TRUE
          )

          # If credentials were set successfully, also read them into current session
          if (file.exists("~/.Renviron")) {
            readRenviron("~/.Renviron")
          }

          return(TRUE)
        }, error = function(e) {
          stop(sprintf("Failed to connect with %s credentials: %s", institution, e$message))
        })
      })
    } else {
      warning(sprintf("Institution '%s' not found in credentials. Trying all credentials...", institution))
    }
  }

  # Fallback: try all credentials if institution not found or not specified
  suppressMessages({
    for (i in seq_along(apiKeys)) {
      tryCatch({
        # Set credentials and load environment manually to avoid restart message
        result <- qualtRics::qualtrics_api_credentials(
          api_key = apiKeys[i],
          base_url = baseUrls[i],
          install = TRUE,
          overwrite = TRUE
        )

        # If credentials were set successfully, also read them into current session
        if (file.exists("~/.Renviron")) {
          readRenviron("~/.Renviron")
        }

        return(TRUE)
      }, error = function(e) {
        if (i == length(apiKeys)) {
          stop("Failed to connect with any credentials provided in ./secrets.R")
        }
      })
    }
  })
}

#' Harmonize Data
#'
#' Performs data cleaning and harmonization on the fetched Qualtrics survey data.
#'
#' @param df Data frame containing Qualtrics survey data.
#' @param identifier The unique identifier for survey respondents.
#' @param qualtrics_alias The alias for the Qualtrics survey.
#' @return Harmonized data frame.
#' @importFrom dplyr mutate
#' @noRd
qualtricsHarmonization <- function(df, identifier, qualtrics_alias) {
  if (!is.data.frame(df)) {
    stop("Input to qualtricsHarmonization is not a data frame.")
  }

  # Validate config
  cfg <- validate_config("qualtrics")

  # Check for visit variable, if not add baseline
  `%!in%` <- Negate(`%in%`)
  if ("visit" %!in% colnames(df) && cfg$study_alias == 'capr') {
    df$visit <- "bl"
  }

  # If visit variable exists, standardize values
  if ("visit" %in% colnames(df) && cfg$study_alias == 'capr') {
    df$visit <- ifelse(is.na(df$visit), "bl",
                       ifelse(df$visit == "0", "bl",
                              ifelse(df$visit == "12", "12m",
                                     ifelse(df$visit == "24", "24m", df$visit))))
  }

  # Additional processing can be uncommented and modified as needed
  # df$src_subject_id <- as.numeric(df$src_subject_id)
  # df$interview_date <- as.Date(df$interview_date, "%m/%d/%Y")
  # df$measure <- qualtrics_alias

  # convert dates (from string ("m/d/Y") to iso date format)
  if ("interview_date" %in% colnames(df)) {
    df$interview_date <- parse_dates_to_iso(df$interview_date, "interview_date")
  }

  suppressWarnings(return(df))
}

#' Display table of available Qualtrics surveys
#'
#' Retrieves a list of all available surveys from the Qualtrics API. Shows all
#' surveys pulled down from Qualtrics, with alias and institution information
#' merged from config.yml where available.
#'
#' @param institution Optional; the institution identifier to use. If NULL, uses all
#'   institutions specified in the configuration file (or all available credentials if no config).
#' @param all Logical; deprecated parameter kept for backward compatibility. All surveys
#'   are now shown by default. Default is FALSE.
#'
#' @return A data frame containing the IDs and names of all available surveys
#'   from the Qualtrics API. Surveys with aliases configured in config.yml will
#'   show the alias and institution; unmapped surveys will show NA for these fields.
#'
#' @export
qualtrics.index <- function(institution = NULL, all = FALSE) {
  # Temporarily suppress warnings
  old_warn <- options("warn")

  tryCatch({
    # Load necessary source files for helper functions

    # Load required secrets and configuration
    validate_secrets("qualtrics")

    # Get secrets using get_secret() to keep it secret, keep it safe
    baseUrls <- get_secret("baseUrls")
    apiKeys <- get_secret("apiKeys")

    cfg <- validate_config("qualtrics")

    # Build mapping of all configured surveys (if config exists)
    all_mapped_surveys <- data.frame(
      id = character(0),
      alias = character(0),
      institution = character(0),
      stringsAsFactors = FALSE
    )

    # Get institution names from config (if available)
    institution_names <- character(0)
    if (!is.null(cfg$qualtrics$survey_ids) && length(cfg$qualtrics$survey_ids) > 0) {
      institution_names <- names(cfg$qualtrics$survey_ids)
      
      # Build the complete mapping from configuration
      for (inst in institution_names) {
        inst_config <- cfg$qualtrics$survey_ids[[inst]]
        if (length(inst_config) > 0) {
          inst_df <- data.frame(
            id = unlist(inst_config),
            alias = names(inst_config),
            institution = rep(inst, length(inst_config)),
            stringsAsFactors = FALSE
          )
          all_mapped_surveys <- rbind(all_mapped_surveys, inst_df)
        }
      }
    }

    # Determine which institutions to query
    institutions_to_query <- character(0)
    if (is.null(institution)) {
      # If no institution specified, query all institutions from config
      if (length(institution_names) > 0) {
        institutions_to_query <- institution_names
      } else {
        # No config, but we still want to fetch surveys
        # We'll connect without specifying institution and use generic names
        institutions_to_query <- paste0("credential_", seq_along(apiKeys))
      }
    } else {
      # Institution specified
      if (institution %in% institution_names) {
        institutions_to_query <- institution
      } else {
        warning(paste0("Institution '", institution, "' not found in configuration. Attempting to connect anyway..."))
        institutions_to_query <- institution
      }
    }

    # Initialize surveys data frame to store ALL surveys from API
    all_surveys <- data.frame(
      id = character(0),
      name = character(0),
      lastModified = character(0),
      institution = character(0),
      stringsAsFactors = FALSE
    )

    # Fetch ALL surveys from each institution's Qualtrics API
    for (i in seq_along(institutions_to_query)) {
      inst <- institutions_to_query[i]
      tryCatch({
        # Connect to this institution (use NULL if it's a generic credential name)
        inst_to_connect <- if (inst %in% institution_names || (!is.null(institution) && institution == inst)) inst else NULL
        
        # For generic credentials, try connecting with index
        if (is.null(inst_to_connect) && grepl("^credential_", inst)) {
          cred_index <- as.integer(sub("^credential_", "", inst))
          if (!is.na(cred_index) && cred_index <= length(apiKeys)) {
            # Manually set credentials for this index
            suppressMessages({
              qualtRics::qualtrics_api_credentials(
                api_key = apiKeys[cred_index],
                base_url = baseUrls[cred_index],
                install = TRUE,
                overwrite = TRUE
              )
            })
            if (file.exists("~/.Renviron")) {
              readRenviron("~/.Renviron")
            }
          } else {
            connectQualtrics(institution = NULL)
          }
        } else {
          connectQualtrics(institution = inst_to_connect)
        }
        
        # Get ALL surveys from this institution's API
        inst_surveys <- qualtRics::all_surveys()
        
        if (nrow(inst_surveys) > 0) {
          # Add institution identifier
          inst_surveys$institution <- inst
          all_surveys <- rbind(all_surveys, inst_surveys)
        }
        
      }, error = function(e) {
        warning(paste0("Failed to fetch surveys from institution '", inst, "': ", e$message))
      })
    }

    if (nrow(all_surveys) == 0) {
      message("No surveys found in Qualtrics API.")
      options(old_warn)
      return(invisible(NULL))
    }

    # Merge with the configured surveys data to get aliases (where available)
    # Use left join to keep all surveys from API, adding alias info where it exists
    surveys <- merge(all_surveys, all_mapped_surveys, by = "id", all.x = TRUE, suffixes = c("_api", "_config"))
    
    # Combine institution columns (prefer config if available, otherwise use API)
    if ("institution_config" %in% names(surveys)) {
      surveys$institution <- ifelse(!is.na(surveys$institution_config), 
                                     surveys$institution_config, 
                                     surveys$institution_api)
      surveys$institution_api <- NULL
      surveys$institution_config <- NULL
    } else {
      names(surveys)[names(surveys) == "institution_api"] <- "institution"
    }

    # Filter by institution if specified (after fetching all)
    if (!is.null(institution) && is.character(institution)) {
      surveys <- surveys[surveys$institution == institution, ]
      if (nrow(surveys) == 0) {
        message(paste0("No surveys found for institution '", institution, "'."))
        options(old_warn)
        return(invisible(NULL))
      }
    }

    # Format the output
    if (nrow(surveys) > 0) {
      # Sort by name for easier reading
      surveys <- surveys[order(surveys$name), ]

      # Create simple output table
      result <- data.frame(
        ID = surveys$id,
        Alias = surveys$alias,
        Institution = surveys$institution,
        Name = surveys$name,
        Last_Modified = surveys$lastModified,
        stringsAsFactors = FALSE
      )

      # Calculate counts for message
      if (!is.null(institution) && is.character(institution)) {
        # Single institution
        count_msg <- paste0("Found ", nrow(result), " surveys from ", institution)
      } else {
        # Multiple institutions - count by institution
        inst_counts <- table(surveys$institution)
        count_parts <- paste0(names(inst_counts), ":", inst_counts, collapse = ", ")
        count_msg <- paste0("Found ", nrow(result), " surveys (", count_parts, ")")
      }

      # Print the results
      message(count_msg)
      print(result, row.names = TRUE)

      # Restore previous warning setting
      options(old_warn)

      return(invisible(surveys))
    } else {
      message("No surveys found.")

      # Restore previous warning setting
      options(old_warn)

      return(invisible(NULL))
    }
  }, error = function(e) {
    # Restore previous warning setting before stopping
    options(old_warn)

    stop(paste("Error connecting to Qualtrics:", e$message))
  })
}

#' Fetch Qualtrics survey metadata to be stored in data frame
#'
#' This function extracts column mappings from the metadata of a Qualtrics survey data frame.
#' It can accept either a data frame containing Qualtrics data, a variable name as string,
#' or a survey alias string.
#'
#' @param survey_alias Can either be an existing dataframe, variable name as string, or survey alias string
#' @param exclude_embedded Only select QIDs
#' @return A list containing the mappings of column names to survey questions.
#' @export
qualtrics.dict <- function(survey_alias, exclude_embedded = TRUE) {
  # First handle the case of a non-existent variable being passed without quotes
  var_name <- NULL

  # Only try to get the name if survey_alias is missing
  if (missing(survey_alias)) {
    stop("Survey alias is required")
  }

  # Capture the actual call
  call_expr <- substitute(survey_alias)

  # Check if it's a symbol (variable name) that doesn't exist
  if (is.symbol(call_expr) && !exists(as.character(call_expr))) {
    var_name <- as.character(call_expr)
    message(sprintf("Object '%s' not found, using as survey alias instead.", var_name))
    survey_alias <- var_name
  }

  # Now proceed with normal function logic

  # Check if input is a data frame
  if (is.data.frame(survey_alias)) {
    # Input is already a data frame, use it directly
    colmap <- qualtRics::extract_colmap(respdata = survey_alias)

    # Filter to include only QID fields
    if (exclude_embedded && !is.null(colmap)) {
      if ("ImportId" %in% names(colmap)) {
        colmap <- colmap[!is.na(colmap$ImportId) & grepl("^QID", colmap$ImportId), ]
      } else if ("qid" %in% names(colmap)) {
        colmap <- colmap[!is.na(colmap$qid) & grepl("^QID", colmap$qid), ]
      }
    }

    return(colmap)
  }

  # Input is a string
  if (is.character(survey_alias)) {
    # First, check if it's a variable name in the global environment
    if (exists(survey_alias)) {
      var_data <- base::get(survey_alias)

      # Check if the variable is a data frame
      if (is.data.frame(var_data)) {
        message(sprintf("Using existing data frame '%s' from environment.", survey_alias))
        colmap <- qualtRics::extract_colmap(respdata = var_data)

        # Filter to include only QID fields
        if (exclude_embedded && !is.null(colmap)) {
          if ("ImportId" %in% names(colmap)) {
            colmap <- colmap[!is.na(colmap$ImportId) & grepl("^QID", colmap$ImportId), ]
          } else if ("qid" %in% names(colmap)) {
            colmap <- colmap[!is.na(colmap$qid) & grepl("^QID", colmap$qid), ]
          }
        }

        return(colmap)
      }
    }

    # Not a variable or not a data frame, treat as survey alias
    message(sprintf("Fetching dictionary for alias '%s' from Qualtrics.", survey_alias))

    # Temporarily suppress warnings and disable progress bars
    old_warn <- options("warn")
    old_opt <- options(qualtRics.progress = FALSE)
    on.exit({options(old_warn); options(old_opt)}, add = TRUE)

    # Get survey data with suppressed output
    survey_data <- suppressMessages(wizaRdry::qualtrics(survey_alias))
    colmap <- qualtRics::extract_colmap(respdata = survey_data)

    # Filter to include only QID fields
    if (exclude_embedded && !is.null(colmap)) {
      if ("ImportId" %in% names(colmap)) {
        colmap <- colmap[!is.na(colmap$ImportId) & grepl("^QID", colmap$ImportId), ]
      } else if ("qid" %in% names(colmap)) {
        colmap <- colmap[!is.na(colmap$qid) & grepl("^QID", colmap$qid), ]
      }
    }

    return(colmap)
  }

  # Invalid input type
  stop("Input must be either a data frame or a string (survey alias or variable name).")
}

#' Convert dates to ISO format robustly
#'
#' This function attempts to intelligently parse dates in various formats
#' and convert them to ISO format (YYYY-MM-DD).
#'
#' @param date_vector A vector of date strings to be parsed
#' @param column_name The name of the column being parsed (for error messages)
#' @return A Date vector in ISO format (YYYY-MM-DD)
#' @importFrom lubridate parse_date_time
#' @noRd
parse_dates_to_iso <- function(date_vector, column_name = "date") {
  if (is.null(date_vector) || length(date_vector) == 0) {
    return(date_vector)
  }

  # Skip if already in Date format
  if (inherits(date_vector, "Date")) {
    return(date_vector)
  }

  # If already a POSIXct or POSIXlt, convert to Date
  if (inherits(date_vector, "POSIXt")) {
    return(as.Date(date_vector))
  }

  # Convert to character if not already
  date_vector <- as.character(date_vector)

  # Remove any NA values for analysis
  non_na_dates <- date_vector[!is.na(date_vector) & date_vector != ""]

  if (length(non_na_dates) == 0) {
    # All NA or empty, just return a vector of NAs
    return(as.Date(date_vector))
  }

  # Define a set of possible date formats to try
  possible_formats <- c(
    # American formats
    "mdy", "mdY", "m/d/y", "m/d/Y", "m-d-y", "m-d-Y",
    # European/ISO formats
    "ymd", "Ymd", "y/m/d", "Y/m/d", "y-m-d", "Y-m-d",
    # Other common formats
    "dmy", "dmY", "d/m/y", "d/m/Y", "d-m-y", "d-m-Y",
    # Month name formats
    "mdy_b", "mdY_b", "b_d_y", "b_d_Y",
    "dmy_b", "dmY_b", "d_b_y", "d_b_Y",
    "ymd_b", "Ymd_b", "y_b_d", "Y_b_d"
  )

  # Try to detect the date format
  tryCatch({
    # Sample the first few non-NA dates to guess format
    sample_size <- min(100, length(non_na_dates))
    sample_dates <- non_na_dates[1:sample_size]

    # Try parsing with each format and keep track of success rate
    format_success <- numeric(length(possible_formats))

    for (i in seq_along(possible_formats)) {
      parsed_dates <- suppressWarnings(
        lubridate::parse_date_time(sample_dates, possible_formats[i], quiet = TRUE)
      )
      format_success[i] <- sum(!is.na(parsed_dates)) / length(sample_dates)
    }

    # Find the format with the highest success rate
    best_format_idx <- which.max(format_success)
    best_format <- possible_formats[best_format_idx]

    # If the best format doesn't parse at least 50% of dates, try combo of top formats
    if (format_success[best_format_idx] < 0.5) {
      # Get top 3 formats
      top_formats <- possible_formats[order(format_success, decreasing = TRUE)[1:3]]

      # Try parsing with these formats
      parsed_dates <- suppressWarnings(
        lubridate::parse_date_time(date_vector, top_formats, quiet = TRUE)
      )
    } else {
      # Parse all dates with the best format
      parsed_dates <- suppressWarnings(
        lubridate::parse_date_time(date_vector, best_format, quiet = TRUE)
      )
    }

    # Convert to Date class
    result <- as.Date(parsed_dates)

    # Basic validation: check for impossibly old dates (before 1900) or future dates
    result[result < as.Date("1900-01-01") | result > Sys.Date() + 30] <- NA

    # Log stats about parsing
    success_rate <- sum(!is.na(result)) / length(date_vector) * 100
    message(sprintf("Parsed %s: %.1f%% successful using %s format",
                    column_name, success_rate,
                    ifelse(format_success[best_format_idx] < 0.5,
                           paste(top_formats, collapse=", "), best_format)))

    return(result)
  }, error = function(e) {
    # Fallback: try base R's as.Date with common formats
    warning(sprintf("Advanced date parsing failed for %s: %s. Falling back to basic parsing.",
                    column_name, e$message))

    fallback_formats <- c("%Y-%m-%d", "%m/%d/%Y", "%d/%m/%Y", "%Y/%m/%d")
    for (fmt in fallback_formats) {
      parsed <- suppressWarnings(as.Date(date_vector, format = fmt))
      if (sum(!is.na(parsed)) / length(parsed) > 0.5) {
        message(sprintf("Basic parsing of %s succeeded with format: %s", column_name, fmt))
        return(parsed)
      }
    }

    # If all else fails, return NA
    warning(sprintf("All date parsing methods failed for %s", column_name))
    return(as.Date(rep(NA, length(date_vector))))
  })
}

#' Alias for 'qualtrics' (DEPRECATED)
#'
#' This function is deprecated. Please use 'qualtrics' instead.
#' This is a legacy alias for the 'qualtrics' function to maintain compatibility with older code.
#'
#' @inheritParams qualtrics
#' @inherit qualtrics return
#' @export
#' @examples
#' \dontrun{
#' # DEPRECATED - use qualtrics() instead
#' survey_data <- getSurvey("your_survey_alias")
#' }
getSurvey <- function(...) {
  .Deprecated("qualtrics", package = "wizaRdry")
  qualtrics(...)
}

# First, install R6 if you don't have it

#' @noRd
#' @importFrom R6 R6Class
SecretsEnv <- R6::R6Class("SecretsEnv",
                          public = list(
                            #' @field config_specs List of specifications for each API
                            config_specs = list(
                              redcap = list(
                                required = c("uri", "token"),
                                types = c(uri = "character", token = "character")
                              ),
                              mongo = list(
                                required = c("connectionString"),
                                types = c(connectionString = "character")
                              ),
                              qualtrics = list(
                                required = c("apiKeys", "baseUrls"),
                                types = c(apiKeys = "vector", baseUrls = "vector")
                              ),
                              sql = list(
                                required = c("uid", "pwd"),
                                optional_connection = c("host", "DSN", "DBQ"),  # At least one required (DSN/DBQ follow Oracle naming convention)
                                types = c(host = "character", DSN = "character", DBQ = "character", uid = "character", pwd = "character")
                              )
                            ),

                            #' @field secrets_file Path to the secrets file
                            secrets_file = NULL,

                            #' @field config Configuration object loaded from config.yml
                            config = NULL,

                            #' @description Initialize a new SecretsEnv object
                            #' @param secrets_file Path to the secrets file
                            #' @param config_file Path to the configuration file
                            #' @return A new SecretsEnv object
                            initialize = function(secrets_file = "secrets.R", config_file = "config.yml") {
                              # Check if secrets file exists
                              if (!file.exists(secrets_file)) {
                                stop(secrets_file, " file not found. Please create this file and define the required API variables.")
                              }

                              # Store the secrets file path
                              self$secrets_file <- secrets_file

                              # Load secrets if not already loaded
                              load_secrets(secrets_file)

                              # Try to load config to determine which APIs are configured
                              tryCatch({
                                self$config <- config::get(file = config_file)
                              }, error = function(e) {
                                warning("Could not load config.yml: ", e$message,
                                        ". Will validate all API types.")
                              })

                              # Only validate APIs that are configured
                              apis_to_validate <- self$get_configured_apis()

                              for (api_type in apis_to_validate) {
                                self$validate_config(api_type)
                              }
                            },

                            #' @description Determine which APIs are configured
                            #' @return Character vector of configured API names
                            get_configured_apis = function() {
                              if (is.null(self$config)) {
                                # If config couldn't be loaded, return an empty list
                                return(character(0))
                              }

                              # Check which API sections exist in the config
                              configured_apis <- character(0)

                              if (!is.null(self$config$mongo)) {
                                configured_apis <- c(configured_apis, "mongo")
                              }

                              if (!is.null(self$config$qualtrics)) {
                                configured_apis <- c(configured_apis, "qualtrics")
                              }

                              if (!is.null(self$config$redcap)) {
                                configured_apis <- c(configured_apis, "redcap")
                              }

                              if (!is.null(self$config$sql)) {
                                configured_apis <- c(configured_apis, "sql")
                              }

                              return(configured_apis)
                            },

                            #' @description Validate the configuration for a specific API
                            #' @param api_type The API type to validate
                            #' @return TRUE if validation passes
                            validate_config = function(api_type) {
                              if (!api_type %in% names(self$config_specs)) {
                                stop("Unknown API type: '", api_type, "'. Valid types are: ",
                                     base::paste(names(self$config_specs), collapse=", "))
                              }

                              specs <- self$config_specs[[api_type]]
                              all_errors <- c()

                              # Special handling for sql API: check for connection target (host, DSN, or DBQ)
                              if (api_type == "sql" && !is.null(specs$optional_connection)) {
                                # Check if at least one connection target is provided
                                connection_targets <- specs$optional_connection
                                has_connection_target <- any(base::sapply(connection_targets, function(var) {
                                  tryCatch({
                                    get_secret(var)
                                    return(TRUE)
                                  }, error = function(e) {
                                    return(FALSE)
                                  })
                                }))
                                
                                if (!has_connection_target) {
                                  all_errors <- c(all_errors, paste("Missing connection target. Please provide one of:",
                                                                    base::paste(connection_targets, collapse=", ")))
                                }
                              }

                              # Check that required variables exist in secrets environment
                              missing_vars <- specs$required[!base::sapply(specs$required, function(var) {
                                tryCatch({
                                  get_secret(var)
                                  return(TRUE)
                                }, error = function(e) {
                                  return(FALSE)
                                })
                              })]

                              if (length(missing_vars) > 0) {
                                all_errors <- c(all_errors, paste("Missing variables:",
                                                                  base::paste(missing_vars, collapse=", ")))
                              }

                              # Only check existing variables for type and emptiness
                              existing_vars <- specs$required[base::sapply(specs$required, function(var) {
                                tryCatch({
                                  get_secret(var)
                                  return(TRUE)
                                }, error = function(e) {
                                  return(FALSE)
                                })
                              })]
                              
                              # Also include optional connection variables if they exist (for sql API)
                              if (api_type == "sql" && !is.null(specs$optional_connection)) {
                                optional_existing <- specs$optional_connection[base::sapply(specs$optional_connection, function(var) {
                                  tryCatch({
                                    get_secret(var)
                                    return(TRUE)
                                  }, error = function(e) {
                                    return(FALSE)
                                  })
                                })]
                                existing_vars <- c(existing_vars, optional_existing)
                              }

                              # Check variable types and emptiness for variables that exist
                              for (type_name in unique(specs$types)) {
                                # Get variables expected to be of this type that exist
                                vars_of_type_names <- names(specs$types[specs$types == type_name])
                                vars_of_type <- vars_of_type_names[vars_of_type_names %in% existing_vars]

                                if (length(vars_of_type) > 0) {
                                  # Check which ones fail the type check
                                  failing_vars <- vars_of_type[base::sapply(vars_of_type, function(var) {
                                    var_value <- get_secret(var)

                                    # Check the appropriate type
                                    !switch(type_name,
                                            "character" = is.character(var_value),
                                            "vector" = is.vector(var_value),
                                            FALSE  # Default case for unknown types
                                    )
                                  })]

                                  # Report if any variables fail this type check
                                  if (length(failing_vars) > 0) {
                                    if (type_name == "character") {
                                      all_errors <- c(all_errors, paste("Type error:",
                                                                        base::paste(failing_vars, collapse=", "),
                                                                        "must be defined as character strings using quotes."))
                                    } else if (type_name == "vector") {
                                      all_errors <- c(all_errors, paste("Type error:",
                                                                        base::paste(failing_vars, collapse=", "),
                                                                        "must be defined as vectors using c() function."))
                                    }
                                  }

                                  # Only check emptiness for variables of the correct type
                                  correct_type_vars <- vars_of_type[!vars_of_type %in% failing_vars]

                                  # For character variables, check for empty strings
                                  if (type_name == "character" && length(correct_type_vars) > 0) {
                                    empty_vars <- correct_type_vars[base::sapply(correct_type_vars, function(var) {
                                      var_value <- get_secret(var)
                                      nchar(var_value) == 0 || all(trimws(var_value) == "")
                                    })]

                                    if (length(empty_vars) > 0) {
                                      all_errors <- c(all_errors, paste("Empty value error:",
                                                                        base::paste(empty_vars, collapse=", "),
                                                                        "cannot be empty strings."))
                                    }
                                  }

                                  # For vector variables, check if they're empty
                                  if (type_name == "vector" && length(correct_type_vars) > 0) {
                                    empty_vars <- correct_type_vars[base::sapply(correct_type_vars, function(var) {
                                      var_value <- get_secret(var)
                                      length(var_value) == 0
                                    })]

                                    if (length(empty_vars) > 0) {
                                      all_errors <- c(all_errors, paste("Empty vector error:",
                                                                        base::paste(empty_vars, collapse=", "),
                                                                        "cannot be empty vectors."))
                                    }
                                  }

                                  # Special checks for specific API types
                                  if (api_type == "redcap" && "uri" %in% correct_type_vars) {
                                    uri_value <- get_secret("uri")

                                    if (grepl("api/$", uri_value)) {
                                      # Format is correct, do nothing
                                    }
                                    # Check if it ends with "api" (missing trailing slash)
                                    else if (grepl("api$", uri_value)) {
                                      # Add trailing slash
                                      fixed_uri <- paste0(uri_value, "/")

                                      # Update the variable in the secrets environment
                                      assign_secret("uri", fixed_uri)

                                      # Update the secrets.R file
                                      if (file.exists(self$secrets_file)) {
                                        # Read and update file content
                                        file_content <- readLines(self$secrets_file)
                                        uri_pattern <- "^\\s*uri\\s*<-\\s*[\"\'](.*)[\"\']\\s*$"
                                        uri_line_index <- grep(uri_pattern, file_content)

                                        if (length(uri_line_index) > 0) {
                                          file_content[uri_line_index] <- gsub(uri_pattern,
                                                                               paste0("uri <- \"", fixed_uri, "\""),
                                                                               file_content[uri_line_index])
                                          writeLines(file_content, self$secrets_file)
                                          message("Note: Added trailing slash to uri in ", self$secrets_file,
                                                  " (", uri_value, " -> ", fixed_uri, ")")
                                        } else {
                                          message("Note: Added trailing slash to uri in memory, but couldn't update ",
                                                  self$secrets_file, " automatically.")
                                        }
                                      }
                                    }
                                    # Not ending with "api" at all
                                    else {
                                      # Don't add a slash, add an error instead
                                      all_errors <- c(all_errors, paste("URI format error:",
                                                                        "Please verify REDCap API endpoint in the API Playground"))
                                    }
                                  }

                                  # Check REDCap token length
                                  if (api_type == "redcap" && "token" %in% correct_type_vars) {
                                    token_value <- get_secret("token")
                                    if (nchar(token_value) < 32) {
                                      all_errors <- c(all_errors, paste("Token length error:",
                                                                        "token must be at least 32 characters long for REDCap API."))
                                    }
                                  }

                                  # SQL validation - ensure connection parameters are valid
                                  if (api_type == "sql") {
                                    # Can add specific SQL validation logic here if needed
                                    # For example, checking connection string format
                                  }
                                }
                              }

                              # If we found any errors, report them all at once
                              if (length(all_errors) > 0) {
                                stop(api_type, " API configuration errors in secrets.R:\n- ",
                                     paste(all_errors, collapse="\n- "), call. = FALSE)
                              } else {
                                # message("All ", api_type, " API credentials in secrets.R are valid.")
                              }

                              return(TRUE)
                            }
                          )
)

#' @noRd
validate_secrets <- function(api_type = NULL, secrets_file = "secrets.R") {
  secrets <- SecretsEnv$new(secrets_file)

  # If api_type is provided, validate just that API
  if (!is.null(api_type)) {
    # Check if the API is configured before validating
    configured_apis <- secrets$get_configured_apis()

    if (api_type %in% configured_apis) {
      return(secrets$validate_config(api_type))
    } else {
      # If this API isn't configured, return TRUE without validating
      message("API type '", api_type, "' is not configured in config.yml, skipping validation.")
      return(TRUE)
    }
  }

  # Otherwise, validation was already done during initialization
  return(TRUE)
}

#
# function: redcap(instrument_name)
# input: instrument_name from table below
#

# Get full file paths of all R files in the api directory
# base::source all files using lapply()

# Initialize functions needed for the progress bar
#' @noRd
initializeLoadingAnimation <- function(steps) {
  # Get console width
  width <- tryCatch({
    if (requireNamespace("cli", quietly = TRUE)) {
      cli::console_width() - 10  # Leave some margin
    } else {
      getOption("width", 80) - 10  # Fallback to R's width setting
    }
  }, error = function(e) 80)  # Default if all else fails

  list(
    steps = steps,
    current = 0,
    width = width,
    start_time = Sys.time()
  )
}

#' @noRd
updateLoadingAnimation <- function(pb, current) {
  pb$current <- current
  percentage <- round(current / pb$steps * 100)
  filled <- round(pb$width * current / pb$steps)
  bar <- paste0(
    strrep("=", filled),
    strrep(" ", pb$width - filled)
  )
  cat(sprintf("\r|%s| %3d%%", bar, percentage))  # Removed extra spaces before bar
  utils::flush.console()
}

#' @noRd
completeLoadingAnimation <- function(pb) {
  updateLoadingAnimation(pb, pb$steps)
  cat("\n")
}

#' Format a time duration in a human-readable way
#'
#' @name formatDuration
#' @param duration The duration to format in seconds or minutes
#' @return A formatted string representing the duration
#' @noRd
formatDuration <- function(duration) {
  secs <- as.numeric(duration, units = "secs")
  if (secs < 60) {
    return(sprintf("%.1f seconds", secs))
  } else {
    mins <- floor(secs / 60)
    remaining_secs <- round(secs %% 60, 1)
    if (remaining_secs > 0) {
      return(sprintf("%d minutes and %.1f seconds", mins, remaining_secs))
    } else {
      return(sprintf("%d minutes", mins))
    }
  }
}

#' Fetch data from REDCap to be stored in a data frame
#'
#' Retrieves data from a REDCap instrument and ensures subject identifiers
#' are propagated across all events
#'
#' @param instrument_name Name of the REDCap instrument
#' @param ... Optional column names to filter for. Only rows with non-missing values
#'        in ALL specified columns will be returned. This is useful for filtering
#'        data to only include complete cases for specific variables of interest.
#' @param raw_or_label Whether to return raw or labeled values
#' @param redcap_event_name Optional event name filter. Can be a single string
#'        or a vector of event names (e.g., \code{c("event1", "event2")})
#' @param batch_size Number of records to retrieve per batch
#' @param records Optional vector of specific record IDs
#' @param fields Optional vector of specific fields
#' @param pii Logical; if FALSE (default), remove fields marked as PII. TRUE keeps PII.
#' @param interview_date Optional; can be either:
#'        - A date string in various formats (ISO, US, etc.) to filter data up to that date
#'        - A boolean TRUE to return only rows with non-NA interview_date values
#' @param date_format Default ymd define date format for interview_date
#' @param complete Option boolean TRUE will return only forms marked as complete in REDCap
#'
#' @return A data frame containing the requested REDCap data
#' @export
#' @examples
#' \dontrun{
#' # Get data from a specific instrument
#' data <- redcap("demographics")
#' }
redcap <- function(instrument_name = NULL, ..., raw_or_label = "raw",
                   redcap_event_name = NULL, batch_size = 1000,
                   records = NULL, fields = NULL, pii = FALSE,
                   interview_date = NULL, date_format = "ymd", complete = NULL) {
  start_time <- Sys.time()

  # Validate secrets and config - MUST COME FIRST
  validate_secrets("redcap")
  config <- validate_config("redcap")

  # Validate that the primary key is properly configured
  if (is.null(config$redcap$primary_key) || !is.character(config$redcap$primary_key)) {
    message("Warning: Primary key not properly configured. Defaulting to 'record_id'")
    config$redcap$primary_key <- "record_id"
  }

  # Log what primary key we're using
  message(sprintf("Using '%s' as the primary key", config$redcap$primary_key))

  # Define the allowed superkey columns explicitly
  allowed_superkey_cols <- c(
    config$redcap$primary_key,
    "src_subject_id",
    "subjectkey",
    "site",
    "subsiteid",
    "src_institution_id",
    "sex",
    "race",
    "ethnic_group",
    "phenotype",
    "phenotype_description",
    "state",
    "status",
    "lost_to_followup",
    "lost_to_follow-up",
    "twins_study",
    "sibling_study",
    "family_study",
    "sample_taken",
    "visit",
    "week",
    "interview_age",
    "group_status",
    "int_start",
    "int_end"
  )

  # Validate and normalize redcap_event_name parameter
  if (!is.null(redcap_event_name)) {
    # Convert to character vector if it's not already
    if (!is.character(redcap_event_name)) {
      stop("redcap_event_name must be a character vector or NULL")
    }
    # Remove any NA values and trim whitespace
    redcap_event_name <- trimws(redcap_event_name[!is.na(redcap_event_name)])
    # If after filtering we have an empty vector, set to NULL
    if (length(redcap_event_name) == 0) {
      redcap_event_name <- NULL
    }
  }

  # Validate date_format parameter
  if (!date_format %in% c("mdy", "dmy", "ymd")) {
    stop("date_format must be one of 'mdy', 'dmy', or 'ymd'")
  }

  # Get secrets using get_secret() to keep it secret, keep it safe
  uri <- get_secret("uri")
  token <- get_secret("token")

  # Test to see if API connection is working
  message("Validating REDCap API connection and user permissions...")
  api_test <- tryCatch({
    # Use a call to redcap_instruments to see if API connnection is working
    forms_test <- REDCapR::redcap_instruments(redcap_uri = uri, token = token, verbose = FALSE)

    # Add conditions to see if API connection is working, if not, throw an error
    if (!is.null(forms_test$success) && !forms_test$success) {
      stop("API connection failed - check your user permissions")
    }

    if (is.null(forms_test$data) || nrow(forms_test$data) == 0) {
      stop("API connected but no instruments accessible - check your export permissions")
    }

    # If permissions are correct and everything is working, notify the user
    message("\nREDCap API connection and permissions validated successfully\n")
    TRUE

    # If permissions are not correct, throw an error with message
  }, error = function(e) {
    stop(sprintf(paste(
      "REDCap API validation failed: \n",
      "Please verify:\n",
      "1. Your REDCap user has 'API Export' and 'API Import/Update' rights\n",
      "2. The API module is enabled for your REDCap project\n",
      "3. Your API token and URI are correct in your secrets configuration"
    )), call. = FALSE)
  })

  # Input validation
  if (is.null(instrument_name)) {
    forms_data <- REDCapR::redcap_instruments(redcap_uri = uri, token = token, verbose = FALSE)$data
    forms_filtered <- forms_data[!grepl("nda", forms_data$instrument_name), ]
    random_instrument <- sample(forms_filtered$instrument_name, 1)
    forms_table <- paste(capture.output(message(redcap.index())), collapse = "\n")
    example_text <- sprintf("\n\nExample:\n%s <- getRedcap(\"%s\")", random_instrument, random_instrument)
    stop(sprintf("No REDCap Instrument Name provided!\n%s%s",
                 forms_table, example_text),
         call. = FALSE)
  }

  # Check if the config$redcap$superkey exists
  if (is.null(config$redcap$superkey)) {
    stop("No superkey form defined in ./config.yml. Please check your REDCap configuration.")
  }

  # Check if the instrument exists before trying to retrieve data
  tryCatch({
    forms_data <- REDCapR::redcap_instruments(redcap_uri = uri, token = token, verbose = FALSE)$data
    # Ensure instrument_name is properly trimmed
    instrument_name <- trimws(instrument_name)
    if (instrument_name %in% forms_data$instrument_name) {
      # Instrument exists, continue
    } else {
      # Format available instruments in a readable way
      available_forms <- paste(sort(forms_data$instrument_name), collapse = "\n- ")
      stop(sprintf("\nInstrument '%s' not found in REDCap for %s.\n\nAvailable instruments:\n- %s",
                   instrument_name, toupper(config$study_alias), available_forms))
    }
  }, error = function(e) {
    if (grepl("not found in REDCap", e$message)) {
      # This is our custom error, pass it through
      stop(e$message, call. = FALSE)
    } else {
      # This is an unexpected error
      stop(sprintf("Error connecting to REDCap: %s", e$message), call. = FALSE)
    }
  })

  # Determine if the requested instrument is the configured superkey form
  is_superkey_request <- FALSE
  if (!is.null(config$redcap$superkey)) {
    is_superkey_request <- identical(trimws(instrument_name), trimws(config$redcap$superkey))
  }

  # First get metadata to identify PII fields
  metadata <- NULL
  pii_fields <- c()
  # Get the full metadata to identify date fields
  metadata <- tryCatch({
    REDCapR::redcap_metadata_read(
      redcap_uri = uri,
      token = token,
      verbose = FALSE
    )$data
  }, error = function(e) {
    message("Warning: Could not retrieve metadata.")
    return(NULL)
  })

  # Find date fields in the metadata
  date_fields <- c()
  if (!is.null(metadata) && "field_name" %in% names(metadata) && "text_validation_type_or_show_slider_number" %in% names(metadata)) {
    # Identify all date fields
    date_validation_types <- c("date_mdy", "date_dmy", "date_ymd", "datetime_mdy", "datetime_dmy", "datetime_ymd", "datetime_seconds_mdy", "datetime_seconds_dmy", "datetime_seconds_ymd")
    date_fields <- metadata$field_name[metadata$text_validation_type_or_show_slider_number %in% date_validation_types]
    date_fields <- c(date_fields, "interview_date") # Add interview_date explicitly
    date_fields <- date_fields[!is.na(date_fields)]
    if (length(date_fields) > 0) {
      message(sprintf("Found %d date fields that will be properly formatted: %s",
                      length(date_fields),
                      paste(date_fields, collapse = ", ")))
    }
  }

  if (!pii && !is.null(metadata) && "field_name" %in% names(metadata) && "identifier" %in% names(metadata)) {
    pii_fields <- metadata$field_name[metadata$identifier == "y"]
    # Filter out NA values and print only the non-NA field names
    pii_fields <- pii_fields[!is.na(pii_fields)]
    if (length(pii_fields) > 0) {
      message(sprintf("Found %d PII fields that will be excluded: %s",
                      length(pii_fields),
                      paste(pii_fields, collapse = ", ")))
    }
  }

  # Now decide which fields to request based on the PII exclusion
  selected_fields <- NULL
  if (!pii && length(pii_fields) > 0) {
    # If fields parameter is provided, exclude PII fields from it
    if (!is.null(fields)) {
      selected_fields <- setdiff(fields, pii_fields)
      if (length(selected_fields) == 0) {
        message("Warning: All requested fields are marked as PII. Results may be empty.")
      }
    } else {
      # If no fields specified, we'll exclude PII fields after retrieving data
      # This is because we can't know all fields in advance
      selected_fields <- NULL
    }
  } else {
    # Use the fields parameter as-is if not excluding PII
    selected_fields <- fields
  }

  # Progress bar
  pb <- initializeLoadingAnimation(20)
  event_name_display <- if (!is.null(redcap_event_name)) {
    if (length(redcap_event_name) == 1) {
      sprintf(" %s", redcap_event_name)
    } else {
      sprintf(" [%s]", paste(redcap_event_name, collapse = ", "))
    }
  } else {
    ""
  }
  message(sprintf("\nImporting records from REDCap form: %s%s%s",
                  instrument_name,
                  event_name_display,
                 ifelse(!pii && length(pii_fields) > 0, " (excluding PII)", "")))
  for (i in 1:20) {
    updateLoadingAnimation(pb, i)
    Sys.sleep(0.1)
  }
  completeLoadingAnimation(pb)
  message("")

  # 1. First, get the superkey data (always using "label")
  # IMPORTANT: Do NOT filter superkey by redcap_event_name here.
  # We want superkey values consolidated across ALL events for each subject,
  # then propagated to the instrument data (which may be event-filtered).
  superkey_response <- REDCapR::redcap_read(
    redcap_uri = uri,
    token = token,
    forms = config$redcap$superkey,
    batch_size = batch_size,
    records = records,
    raw_or_label = "label",  # Always use label for superkey
    raw_or_label_headers = "raw",
    verbose = FALSE
  )

  # 2. Then, get the instrument data with user's raw_or_label preference
  instrument_response <- REDCapR::redcap_read(
    redcap_uri = uri,
    token = token,
    forms = instrument_name,
    batch_size = batch_size,
    records = records,
    events = redcap_event_name,  # Filter by event names if specified
    fields = selected_fields,
    raw_or_label = raw_or_label,  # Use user's preference
    raw_or_label_headers = "raw",
    verbose = FALSE
  )

  # Filter superkey columns to only allowed columns
  super_key_cols <- intersect(names(superkey_response$data), allowed_superkey_cols)

  # Add redcap_event_name to super_key_cols if it exists in either dataset
  if ("redcap_event_name" %in% names(superkey_response$data) ||
      "redcap_event_name" %in% names(instrument_response$data)) {
    super_key_cols <- c(super_key_cols, "redcap_event_name")
  }

  # If excluding PII, remove PII fields from superkey columns
  if (!pii && length(pii_fields) > 0) {
    super_key_cols <- setdiff(super_key_cols, pii_fields)
  }

  # If the requested instrument IS the superkey, return it directly without joins/propagation
  if (is_superkey_request) {
    message("Requested instrument matches configured superkey; returning without joins or propagation.")
    df <- instrument_response$data
  } else if ("redcap_event_name" %in% names(superkey_response$data)) {
    # First, ensure the primary key exists in the superkey data
    if (!(config$redcap$primary_key %in% names(superkey_response$data))) {
      stop(sprintf("Primary key '%s' not found in superkey data", config$redcap$primary_key))
    }

    # For each subject, collect all non-NA values across events
    subjects <- unique(superkey_response$data[[config$redcap$primary_key]])

    if (length(subjects) == 0) {
      message("No subjects found in superkey data. Creating empty consolidated superkey.")
      # Create an empty data frame with the primary key as the first column
      consolidated_superkey <- data.frame(matrix(ncol = 0, nrow = 0))
    } else {
      # Create a data frame with the primary key column
      primary_key_col <- config$redcap$primary_key
      consolidated_superkey <- data.frame(matrix(ncol = 1, nrow = length(subjects)))
      colnames(consolidated_superkey) <- primary_key_col
      consolidated_superkey[[primary_key_col]] <- subjects

      # Add the rest of the columns
      for (col in super_key_cols) {
        if (col != primary_key_col && col != "redcap_event_name") {
          # Only try to process columns that exist in the source data
          if (col %in% names(superkey_response$data)) {
            consolidated_superkey[[col]] <- NA
            for (subject_id in subjects) {
              # Get all values for this subject across all events
              subject_rows <- superkey_response$data[superkey_response$data[[primary_key_col]] == subject_id, ]
              non_na_values <- subject_rows[[col]][!is.na(subject_rows[[col]])]
              if (length(non_na_values) > 0) {
                consolidated_superkey[consolidated_superkey[[primary_key_col]] == subject_id, col] <- non_na_values[1]
              }
            }
          }
        }
      }
    }
  } else if (!is_superkey_request) {
    # If no redcap_event_name, just use the superkey data as is, but only the allowed columns
    allowed_cols <- intersect(names(superkey_response$data), super_key_cols)

    if (length(allowed_cols) == 0) {
      message("No matching columns found in superkey data. Creating empty consolidated superkey.")
      consolidated_superkey <- data.frame()
    } else {
      consolidated_superkey <- superkey_response$data[, allowed_cols, drop = FALSE]
    }
  }

  # FIXED: Merge the consolidated superkey with the instrument data
  # If superkey requested, we already set df above; otherwise proceed with merge logic
  if (!exists("df", inherits = FALSE) && nrow(consolidated_superkey) == 0) {
    message("Warning: Consolidated superkey is empty. Using instrument data as is.")
    df <- instrument_response$data
  } else if (!exists("df", inherits = FALSE) && nrow(instrument_response$data) == 0) {
    message("Warning: Instrument data is empty. Using consolidated superkey as is.")
    df <- consolidated_superkey
  } else if (!exists("df", inherits = FALSE) && "redcap_event_name" %in% names(instrument_response$data)) {
    # Keep the original event names from the instrument data
    df <- instrument_response$data

    # Check if the primary key exists in the instrument data
    if (!(config$redcap$primary_key %in% names(df))) {
      message(sprintf("Warning: Primary key '%s' not found in instrument data. Cannot merge superkey information.", config$redcap$primary_key))
    } else {
      # Merge in the consolidated superkey data (excluding event_name)
      for (subject_id in unique(df[[config$redcap$primary_key]])) {
        # Check if this subject exists in the superkey data
        if (config$redcap$primary_key %in% names(consolidated_superkey)) {
          superkey_data <- consolidated_superkey[consolidated_superkey[[config$redcap$primary_key]] == subject_id,
                                                 !(names(consolidated_superkey) %in% c("redcap_event_name")),
                                                 drop = FALSE]

          if (nrow(superkey_data) > 0) {
            # Apply the superkey data to all rows for this subject
            for (col in names(superkey_data)) {
              if (col != config$redcap$primary_key) {
                if (col %in% names(df)) {
                  df[df[[config$redcap$primary_key]] == subject_id, col] <- superkey_data[[col]][1]
                } else {
                  # Add the column if it doesn't exist in the instrument data
                  df[[col]] <- NA
                  df[df[[config$redcap$primary_key]] == subject_id, col] <- superkey_data[[col]][1]
                }
              }
            }
          }
        }
      }
    }
  } else if (!exists("df", inherits = FALSE)) {
    # Standard merge if there's no event name in the instrument data
    message("Performing standard merge of superkey and instrument data")
    # Make sure at least the primary key is in both data frames
    if (!(config$redcap$primary_key %in% names(consolidated_superkey)) ||
        !(config$redcap$primary_key %in% names(instrument_response$data))) {
      message(sprintf("Warning: Primary key '%s' missing from one of the data frames. Concatenating data instead of merging.", config$redcap$primary_key))
      df <- rbind(consolidated_superkey, instrument_response$data)
    } else {
      df <- merge(
        consolidated_superkey,
        instrument_response$data,
        by = config$redcap$primary_key,
        all.y = TRUE
      )
    }
  }

  # Continue with the existing propagation logic
  if (config$redcap$primary_key %in% names(df) && "redcap_event_name" %in% names(df)) {
    # Keep only fields that exist in our dataframe
    super_key_cols <- super_key_cols[super_key_cols %in% names(df)]
    if (length(super_key_cols) > 0) {
      message("\nPropagating superkey across all events for each subject...")
      # For each subject
      for (subject_id in unique(df[[config$redcap$primary_key]])) {
        # Get all rows for this subject
        subject_rows <- which(df[[config$redcap$primary_key]] == subject_id)
        # For each super key field, find a non-NA value across all events
        for (key_field in super_key_cols) {
          # Skip redcap_event_name to preserve the original values
          if (key_field == "redcap_event_name") {
            next
          }
          key_values <- df[subject_rows, key_field]
          non_na_values <- key_values[!is.na(key_values)]
          if (length(non_na_values) > 0) {
            # Propagate the first non-NA value to all events for this subject
            df[subject_rows, key_field] <- non_na_values[1]
          }
        }
      }
    }
  }

  # Fix the date fields - Convert numeric dates to proper date format
  # Process date fields - this includes interview_date
  if (length(date_fields) > 0) {
    message("\nConverting date fields to proper date format...")
    for (date_field in date_fields) {
      if (date_field %in% names(df)) {
        # Only process if the column exists and has some non-NA values
        if (any(!is.na(df[[date_field]]))) {
          # Convert to character first to handle both numeric and character inputs
          dates_char <- as.character(df[[date_field]])
          # Check if dates are numeric (SPSS/SAS format) - typically represented as days since 1960-01-01
          if (all(grepl("^\\d+(\\.\\d+)?$", dates_char[!is.na(dates_char)]))) {
            # Convert from numeric to date
            # For SPSS/SAS dates (days since 1960-01-01)
            numeric_dates <- as.numeric(dates_char)
            date_values <- as.Date(numeric_dates, origin = "1960-01-01")
            # Now format the date according to the requested format
            if (date_format == "mdy") {
              df[[date_field]] <- format(date_values, "%m/%d/%Y")
            } else if (date_format == "dmy") {
              df[[date_field]] <- format(date_values, "%d/%m/%Y")
            } else if (date_format == "ymd") {
              df[[date_field]] <- format(date_values, "%Y-%m-%d")
            }
            message(sprintf("  Converted %s from numeric to %s format", date_field, date_format))
          }
          # Check if they are already in a date-like format
          else if (any(grepl("-|/", dates_char[!is.na(dates_char)]))) {
            # Try to parse existing dates and reformat them
            parsed_dates <- tryCatch({
              # Try different parsing approaches
              if (all(grepl("/", dates_char[!is.na(dates_char)]))) {
                # Likely MM/DD/YYYY or DD/MM/YYYY
                parts <- strsplit(dates_char[!is.na(dates_char)][1], "/")[[1]]
                if (length(parts) == 3) {
                  if (as.numeric(parts[1]) > 12) { # DD/MM/YYYY
                    as.Date(dates_char, format = "%d/%m/%Y")
                  } else { # Default to MM/DD/YYYY
                    as.Date(dates_char, format = "%m/%d/%Y")
                  }
                }
              } else if (all(grepl("-", dates_char[!is.na(dates_char)]))) {
                # Likely YYYY-MM-DD
                as.Date(dates_char)
              } else {
                # Try generic parsing with lubridate
                if (date_format == "mdy") {
                  lubridate::mdy(dates_char)
                } else if (date_format == "dmy") {
                  lubridate::dmy(dates_char)
                } else {
                  lubridate::ymd(dates_char)
                }
              }
            }, error = function(e) {
              # If parsing fails, return NA
              message(sprintf("  Warning: Could not parse dates in %s, leaving as is", date_field))
              return(NULL)
            })
            if (!is.null(parsed_dates)) {
              # Format according to preference
              if (date_format == "mdy") {
                df[[date_field]] <- format(parsed_dates, "%m/%d/%Y")
              } else if (date_format == "dmy") {
                df[[date_field]] <- format(parsed_dates, "%d/%m/%Y")
              } else if (date_format == "ymd") {
                df[[date_field]] <- format(parsed_dates, "%Y-%m-%d")
              }
              message(sprintf("  Reformatted %s to %s format", date_field, date_format))
            }
          }
        }
      }
    }
  }

  # For interview_age columns
  age_cols <- grep("_interview_age$", base::names(df))
  if (length(age_cols) > 0) {
    base::names(df)[age_cols] <- "interview_age"
  }

  # For interview_date columns - more robust handling
  date_patterns <- c("_interview_date$", "interview_date")
  date_cols <- NULL
  for (pattern in date_patterns) {
    found_cols <- grep(pattern, base::names(df), ignore.case = TRUE)
    if (length(found_cols) > 0) {
      date_cols <- found_cols
      break  # Stop at first pattern that finds matches
    }
  }
  if (!is.null(date_cols) && length(date_cols) > 0) {
    base::names(df)[date_cols] <- "interview_date"
  }

  # Apply redcap_event_name filter if specified
  if (!is.null(redcap_event_name)) {
    if (!"redcap_event_name" %in% names(df)) {
      stop("Cannot filter by redcap_event_name: column not found in data")
    }
    # Use %in% to support both single strings and vectors
    df <- df[df$redcap_event_name %in% redcap_event_name, ]
  }

  # Study-specific processing (legacy - errors should be fixed in redcap...)
  if (config$study_alias == "capr") {
    df <- processCaprData(df, instrument_name)
  }

  # Make a final pass to remove any PII fields that might have been included
  if (!pii && length(pii_fields) > 0) {
    pii_cols_present <- intersect(names(df), pii_fields)
    pii_cols_present <- pii_cols_present[!is.na(pii_cols_present)]
    if (length(pii_cols_present) > 0) {
      message(sprintf("\nRemoving %d PII fields from final dataset: %s",
                      length(pii_cols_present),
                      paste(pii_cols_present, collapse = ", ")))
      df <- df[, !names(df) %in% pii_fields, drop = FALSE]
    }
  }

  # Create a copy of the original dataframe to preserve original values
  original_df <- df

  # Advanced date parsing function that handles multiple formats
  parseAnyDate <- function(date_string) {
    if (is.na(date_string) || is.null(date_string)) {
      return(NA)
    }
    # Try multiple date formats sequentially
    date <- NULL
    # Try ISO format (YYYY-MM-DD)
    if (grepl("^\\d{4}-\\d{1,2}-\\d{1,2}$", date_string)) {
      date <- tryCatch(lubridate::ymd(date_string), error = function(e) NULL)
    }
    # Try US format (MM/DD/YYYY)
    else if (grepl("^\\d{1,2}/\\d{1,2}/\\d{4}$", date_string)) {
      date <- tryCatch(lubridate::mdy(date_string), error = function(e) NULL)
    }
    # Try European format (DD.MM.YYYY)
    else if (grepl("^\\d{1,2}\\.\\d{1,2}\\.\\d{4}$", date_string)) {
      date <- tryCatch(lubridate::dmy(date_string), error = function(e) NULL)
    }
    # Try Canadian format (YYYY/MM/DD)
    else if (grepl("^\\d{4}/\\d{1,2}/\\d{1,2}$", date_string)) {
      date <- tryCatch(lubridate::ymd(date_string), error = function(e) NULL)
    }
    # Try other format (DD-MM-YYYY)
    else if (grepl("^\\d{1,2}-\\d{1,2}-\\d{4}$", date_string)) {
      date <- tryCatch(lubridate::dmy(date_string), error = function(e) NULL)
    }
    # Try abbreviated month name (15-Jan-2023 or Jan 15, 2023)
    else if (grepl("[A-Za-z]", date_string)) {
      date <- tryCatch(lubridate::parse_date_time(date_string, c("dmy", "mdy")), error = function(e) NULL)
    }
    # If all attempts fail, return NA
    if (is.null(date) || all(is.na(date))) {
      warning("Failed to parse date: ", date_string, ". Treating as NA.")
      return(NA)
    }
    return(as.Date(date))
  }

  # Handle interview_date filtering
  if ("interview_date" %in% names(df)) {
    # Create a temporary date column for filtering but don't modify the original
    df$temp_date <- sapply(df$interview_date, parseAnyDate)
    # Handle the interview_date parameter
    if (!is.null(interview_date)) {
      if (is.logical(interview_date) && interview_date == TRUE) {
        # Keep only rows with non-NA interview_date values
        rows_to_keep <- !is.na(df$temp_date)
        df <- df[rows_to_keep, ]
        original_df <- original_df[rows_to_keep, ]
      } else if (is.character(interview_date) || inherits(interview_date, "Date")) {
        # Filter by specific date
        input_date <- tryCatch({
          if (inherits(interview_date, "Date")) {
            interview_date
          } else {
            parseAnyDate(interview_date)
          }
        }, error = function(e) {
          stop("Failed to parse interview_date parameter: ", interview_date)
        })
        if (is.na(input_date)) {
          stop("Failed to parse interview_date parameter: ", interview_date)
        }
        # If we're here with a valid date, keep rows with dates before or equal to input_date
        rows_to_keep <- !is.na(df$temp_date) & df$temp_date <= input_date
        df <- df[rows_to_keep, ]
        original_df <- original_df[rows_to_keep, ]
      } else {
        stop("interview_date must be either a date string or TRUE")
      }
    }
    # Remove temporary column after filtering
    df$temp_date <- NULL
  }

  # Reorder columns to put superkey columns first
  if (ncol(df) > 0) {
    # Get names of all superkey columns that are in the dataset
    superkey_cols_in_data <- intersect(
      c(config$redcap$primary_key, "redcap_event_name", super_key_cols),
      names(df)
    )
    # Get names of all instrument columns
    instrument_cols <- setdiff(names(df), superkey_cols_in_data)
    # Create the new column order
    new_col_order <- c(superkey_cols_in_data, instrument_cols)
    # Reorder the dataframe
    df <- df[, new_col_order, drop = FALSE]
  }

  # Attach the instrument name as an attribute without an extra parameter
  attr(df, "redcap_instrument") <- instrument_name

  # allow sort by complete
  if (!is.null(instrument_name) && !is.null(complete)) {
    # Construct the complete variable name based on instrument_name
    complete_var <- paste0(instrument_name, "_complete")
    # Check if the complete variable exists in the dataframe
    if (complete_var %in% names(df)) {
      # Check for NA values in the complete variable
      na_counts <- sum(is.na(df[[complete_var]]))
      if (na_counts > 0) {
        message(sprintf("Note: Found %d rows with NA values in %s", na_counts, complete_var))
      }
      # If complete is TRUE, keep only complete records (exclude NA)
      # If complete is FALSE, keep only incomplete records (exclude NA)
      if (is.logical(complete)) {
        if (complete) {
          message("Filtering for complete records only...")
          # Handle both raw (1) and label ("Complete") formats
          if (raw_or_label == "raw") {
            df <- df[!is.na(df[[complete_var]]) & df[[complete_var]] == 2, ]
          } else {
            df <- df[!is.na(df[[complete_var]]) & df[[complete_var]] == "Complete", ]
          }
        } else {
          message("Filtering for incomplete records only...")
          # Handle both raw (0) and label ("Incomplete") formats
          if (raw_or_label == "raw") {
            df <- df[!is.na(df[[complete_var]]) & df[[complete_var]] == 0, ]
          } else {
            df <- df[!is.na(df[[complete_var]]) & df[[complete_var]] == "Incomplete", ]
          }
        }
      } else {
        warning("The 'complete' parameter should be TRUE or FALSE. Ignoring filter.")
      }
    } else {
      warning(paste0("Complete variable '", complete_var, "' not found in the dataset. Ignoring filter."))
    }
  }

  # Check if any column requests were passed via ...
  dots_args <- list(...)
  if (length(dots_args) > 0) {
    # Convert the dots arguments to a character vector
    requested_cols <- as.character(unlist(dots_args))
    # Find which of the requested columns actually exist in the data
    existing_cols <- intersect(requested_cols, names(df))
    if (length(existing_cols) > 0) {
      # Display the names of the columns that were found
      message(sprintf("Found %d of %d requested columns: %s",
                      length(existing_cols),
                      length(requested_cols),
                      paste(existing_cols, collapse = ", ")))
      # Create a filter to keep only rows where ALL requested columns have data
      rows_to_keep <- rep(TRUE, nrow(df))
      for (col in existing_cols) {
        # Check if column values are not NA
        not_na <- !is.na(df[[col]])
        # For non-NA values, check if they're not empty strings (if character type)
        not_empty <- rep(TRUE, nrow(df))
        if (is.character(df[[col]])) {
          not_empty <- df[[col]] != ""
        }
        # Combine the conditions - both not NA and not empty (if applicable)
        has_data <- not_na & not_empty
        # Update the rows_to_keep vector
        rows_to_keep <- rows_to_keep & has_data
      }
      # Apply the filter to keep only rows with data in all requested columns
      original_rows <- nrow(df)
      df <- df[rows_to_keep, ]
      kept_rows <- nrow(df)
      message(sprintf("Kept %d of %d rows where all requested columns have values.",
                      kept_rows, original_rows))
    } else {
      if (length(requested_cols) > 0) {
        warning("None of the requested columns were found in the dataset.")
      }
    }
  }

  # Show duration
  end_time <- Sys.time()
  duration <- difftime(end_time, start_time, units = "secs")
  message(sprintf("\nData frame '%s' retrieved in %s.", instrument_name, formatDuration(duration)))
  return(df)
}


#' Display table of available REDCap instruments and their labels
#'
#' Retrieves a list of all available REDCap forms as a formatted table
#'
#' @return A formatted table (kable) of available REDCap instruments/forms
#' @importFrom REDCapR redcap_instruments
#' @importFrom knitr kable
#' @export
redcap.index <- function() {
  # Load required packages

  # Validate secrets
  tryCatch({
    validate_secrets("redcap")

    # Get secrets using get_secret() to keep it secret, keep it safe
    uri <- get_secret("uri")
    token <- get_secret("token")

  }, error = function(e) {
    message("Error loading or validating REDCap secrets: ", e$message)
    return(NULL)
  })

  # Attempt to fetch instruments from REDCap
  tryCatch({
    forms_result <- REDCapR::redcap_instruments(redcap_uri = uri, token = token, verbose = FALSE)

    # Check if the operation was successful
    if (forms_result$success) {
      return(knitr::kable(forms_result$data, format = "simple"))
    } else {
      message("REDCap API returned an error: ", forms_result$status_message)
      return(NULL)
    }
  }, error = function(e) {
    message("Error connecting to REDCap: ", e$message)
    return(NULL)
  }, warning = function(w) {
    message("Warning during REDCap connection: ", w$message)
  })
}

#' Fetch REDCap data dictionary to be stored in data frame
#'
#' This function extracts metadata/dictionary information from REDCap. It can accept
#' either an instrument name to fetch new data, an existing data frame with instrument
#' attributes, or a variable name as string.
#'
#' @param instrument_name Can either be an instrument name to fetch new data, a data frame
#'   returned by redcap(), or a variable name as string
#' @return A data frame containing the data dictionary/metadata for the specified instrument
#' @importFrom REDCapR redcap_metadata_read
#' @export
redcap.dict <- function(instrument_name) {
  # First handle the case of a non-existent variable being passed without quotes
  var_name <- NULL

  # Only try to get the name if instrument_name is missing
  if (missing(instrument_name)) {
    stop("Instrument name is required")
  }

  # Capture the actual call
  call_expr <- substitute(instrument_name)

  # Check if it's a symbol (variable name) that doesn't exist
  if (is.symbol(call_expr) && !exists(as.character(call_expr))) {
    var_name <- as.character(call_expr)
    message(sprintf("Object '%s' not found, using as instrument name instead.", var_name))
    instrument_name <- var_name
  }

  # Now proceed with normal function logic

  # Check if input is a data frame with redcap_instrument attribute
  if (is.data.frame(instrument_name) && !is.null(attr(instrument_name, "redcap_instrument"))) {
    inst <- attr(instrument_name, "redcap_instrument")

    # Fetch metadata using the instrument name
    # Validate secrets
    validate_secrets("redcap")

    # Get secrets using get_secret() to keep it secret, keep it safe
    uri <- get_secret("uri")
    token <- get_secret("token")

    metadata <- REDCapR::redcap_metadata_read(
      redcap_uri = uri,
      token = token,
      forms = inst,
      verbose = FALSE
    )$data

    return(metadata)
  }

  # Check if input is a regular data frame
  if (is.data.frame(instrument_name)) {
    message("Using provided REDCap metadata data frame.")
    return(instrument_name)
  }

  # Input is a string
  if (is.character(instrument_name)) {
    # Check if it's a variable name in the global environment
    if (exists(instrument_name)) {
      var_data <- base::get(instrument_name)

      # Check if the variable is a data frame with instrument attribute
      if (is.data.frame(var_data) && !is.null(attr(var_data, "redcap_instrument"))) {
        inst <- attr(var_data, "redcap_instrument")
        message(sprintf("Retrieving metadata for instrument '%s' from variable '%s'.",
                        inst, instrument_name))

        # Fetch metadata using the instrument name
        # Validate secrets
        validate_secrets("redcap")

        # Get secrets using get_secret() to keep it secret, keep it safe
        uri <- get_secret("uri")
        token <- get_secret("token")

        metadata <- REDCapR::redcap_metadata_read(
          redcap_uri = uri,
          token = token,
          forms = inst,
          verbose = FALSE
        )$data

        return(metadata)
      }

      # Check if the variable is just a data frame
      if (is.data.frame(var_data)) {
        message(sprintf("Using existing metadata data frame '%s' from environment.", instrument_name))
        return(var_data)
      }
    }

    # Not a variable or not a data frame, treat as instrument name
    message(sprintf("Fetching metadata for instrument '%s' from REDCap.", instrument_name))
    # Validate secrets
    validate_secrets("redcap")

    # Get secrets using get_secret() to keep it secret, keep it safe
    uri <- get_secret("uri")
    token <- get_secret("token")

    # Fetch metadata from REDCap
    metadata <- REDCapR::redcap_metadata_read(
      redcap_uri = uri,
      token = token,
      verbose = FALSE
    )$data

    dictionary <- metadata[metadata$form_name == instrument_name, ]

    message(sprintf("Retrieved metadata for instrument '%s' from REDCap.", instrument_name))
    return(dictionary)
  }

  # Invalid input type
  stop("Input must be either a data frame, a string variable name, or an instrument name string.")
}

#' Alias for 'redcap' (DEPRECATED)
#'
#' This function is deprecated. Please use 'redcap' instead.
#' This is a legacy alias for the 'redcap' function to maintain compatibility with older code.
#'
#' @inheritParams redcap
#' @inherit redcap return
#' @export
#' @examples
#' \dontrun{
#' # DEPRECATED - use redcap() instead
#' survey_data <- getRedcap("demographics")
#' }
getRedcap <- function(...) {
  .Deprecated("redcap", package = "wizaRdry")
  redcap(...)
}

#' Initialize the wizaRdry directory structure inside an R project
#'
#' Creates the standard directory structure required for the wizaRdry package to function properly.
#' This includes folders for data cleaning scripts, NDA submission templates, and temporary outputs.
#' It can detect and repair incomplete directory structures, and optionally create an R project.
#'
#' @param study_alias Character string specifying the short name for the study e.g. impact, capr, sing
#' @param path Character string specifying the directory path where the structure should be created.
#'        Defaults to the current working directory.
#' @param overwrite Logical. If TRUE, will overwrite existing files. If FALSE (default),
#'        will not replace existing files.
#' @param repair Logical. If TRUE, will attempt to repair an incomplete directory structure.
#'        If FALSE, will abort with an error message when encountering an incomplete structure.
#' @param show_tree Logical. If TRUE (default on first run), will display a visual file tree.
#'        Set to FALSE to suppress the tree view.
#' @param create_project Logical. If TRUE, will create an R project file if one doesn't exist.
#'        If FALSE (default), will not create an R project.
#' @param examples Logical. If TRUE (default when not repairing), will create example script templates.
#'        If FALSE (default when repairing), will skip creating example scripts.
#' @param skip_prompt Logical. If TRUE (default), will skip the initial confirmation prompt if y/n preferences are not set yet. FALSE if specified.
#'
#' @return Invisible TRUE if successful.
#'
#' @details
#' The function creates the following directory structure:
#' - clean/
#'   - csv/
#'   - mongo/
#'   - qualtrics/
#'   - redcap/
#'   - oracle/
#'   - sql/
#' - nda/
#'   - csv/
#'   - mongo/
#'   - qualtrics/
#'   - redcap/
#'   - oracle/
#'   - sql/
#' - tmp/
#'
#' It also creates template config.yml and secrets.R files, and optionally an R project file.
#'
#' @examples
#' \dontrun{
#' # Initialize in current directory
#' scry()
#'
#' # Repair structure in current directory
#' scry(repair = TRUE)
#'
#' # Initialize in a specific directory with an R project
#' scry("path/to/project", create_project = TRUE, repair = TRUE)
#'
#' # Skip the tree display
#' scry(repair = TRUE, show_tree = FALSE)
#'
#' # Explicitly create example scripts when repairing
#' scry(repair = TRUE, examples = TRUE)
#'
#' # Skip the confirmation prompt
#' scry(skip_prompt = TRUE)
#' }
#'
#' @export
scry <- function(study_alias = NULL, path = ".", overwrite = FALSE, repair = FALSE, show_tree = NULL,
                 create_project = FALSE, examples = FALSE, skip_prompt = TRUE) {
  # Check for user preferences file
  user_prefs_file <- file.path(path, ".wizaRdry_prefs")
  user_prefs <- list(shown_tree = FALSE, auto_create = FALSE)

  if (file.exists(user_prefs_file)) {
    tryCatch(
      {
        user_prefs <- readRDS(user_prefs_file)
      },
      error = function(e) {
        # If file exists but can't be read, create a new one
        user_prefs <- list(shown_tree = FALSE, auto_create = FALSE)
      }
    )
  }

  # Check if this looks like a wizaRdry project structure
  has_clean <- dir.exists(file.path(path, "clean"))
  has_nda <- dir.exists(file.path(path, "nda"))
  has_config <- file.exists(file.path(path, "config.yml"))
  has_secrets <- file.exists(file.path(path, "secrets.R"))
  has_main <- file.exists(file.path(path, "main.R"))

  structure_exists <- has_clean || has_nda || has_config || has_secrets || has_main


  # If skip_prompt is TRUE or user has previously set auto_create to TRUE, bypass the prompt
  if (!structure_exists && (!skip_prompt || !user_prefs$auto_create)) {
    response <- readline(prompt = "No wizaRdry project found. Would you like to create the wizaRdry project structure? y/n ")

    while (!tolower(response) %in% c("y", "n")) {
      response <- readline(prompt = "Please enter either y or n: ")
    }

    # If user says yes, this will save their preference as y and update auto_create to TRUE
    if (tolower(response) == "y") {
      user_prefs$auto_create <- TRUE
      saveRDS(user_prefs, user_prefs_file)
    }

    if (tolower(response) == "n") {
      # Instead of stopping with an error, return invisibly
      return(invisible(NULL))
    }
  }

  # If study_alias is NULL (not provided), try to detect from .Rproj file
  if (is.null(study_alias)) {
    # Look for .Rproj files in the specified path
    rproj_files <- list.files(path, pattern = "\\.Rproj$", full.names = FALSE)

    if (length(rproj_files) > 0) {
      # Extract project name from the first .Rproj file found
      project_name <- sub("\\.Rproj$", "", rproj_files[1])
      study_alias <- project_name
      message(paste0("Using project name as study alias: ", study_alias))
    } else {
      # Default to FALSE if no .Rproj file found
      study_alias <- FALSE
    }
  }

  # Define directory structure
  expected_dirs <- c(
    file.path(path, "clean"),
    file.path(path, "clean", "csv"),
    file.path(path, "clean", "mongo"),
    file.path(path, "clean", "qualtrics"),
    file.path(path, "clean", "redcap"),
    file.path(path, "clean", "oracle"),
    file.path(path, "clean", "sql"),
    file.path(path, "nda"),
    file.path(path, "nda", "csv"),
    file.path(path, "nda", "mongo"),
    file.path(path, "nda", "qualtrics"),
    file.path(path, "nda", "redcap"),
    file.path(path, "nda", "oracle"),
    file.path(path, "nda", "sql"),
    file.path(path, "tmp")
  )

  # Define expected files
  expected_files <- c(
    file.path(path, "config.yml"),
    file.path(path, "secrets.R"),
    file.path(path, "main.R")
  )

  # If structure partially exists but is incomplete
  structure_complete <- all(sapply(expected_dirs, dir.exists)) &&
    all(sapply(expected_files, file.exists))

  # Determine if we should show the tree
  if (is.null(show_tree)) {
    show_tree <- !user_prefs$shown_tree
  }

  # Handle incomplete structures
  if (structure_exists && !structure_complete) {
    if (!repair) {
      missing_dirs <- expected_dirs[!sapply(expected_dirs, dir.exists)]
      missing_files <- expected_files[!sapply(expected_files, file.exists)]

      stop(
        "Incomplete wizaRdry directory structure detected.\n",
        "Missing components:\n",
        if (length(missing_dirs) > 0) paste0("  Directories: ", paste(missing_dirs, collapse = ", "), "\n") else "",
        if (length(missing_files) > 0) paste0("  Files: ", paste(missing_files, collapse = ", "), "\n") else "",
        "Use scry(repair = TRUE) to repair the structure."
      )
    } else {
      message("Repairing incomplete wizaRdry directory structure...")
    }
  }

  # Create directories
  created <- character(0)
  for (dir in expected_dirs) {
    if (!dir.exists(dir)) {
      dir.create(dir, recursive = TRUE)
      created <- c(created, dir)
    } else if (overwrite) {
      created <- c(created, paste0(dir, " (already exists)"))
    } else {
      created <- c(created, paste0(dir, " (skipped, already exists)"))
    }
  }

  # Create .gitkeep files to ensure empty directories are tracked by git
  for (dir in expected_dirs) {
    if (dir != path) { # Don't add .gitkeep to the root directory
      gitkeep_file <- file.path(dir, ".gitkeep")
      if (!file.exists(gitkeep_file)) {
        file.create(gitkeep_file)
      }
    }
  }

  # Add tmp directory contents to .gitignore
  gitignore_file <- file.path(path, ".gitignore")
  tmp_pattern <- "tmp/*"
  secrets_pattern <- "secrets.R"
  pem_pattern <- "*.pem"
  prefs_pattern <- ".wizaRdry_prefs"

  if (file.exists(gitignore_file)) {
    gitignore_content <- readLines(gitignore_file)
    need_to_update <- FALSE

    # Check for secrets.R
    if (!any(grepl(secrets_pattern, gitignore_content, fixed = TRUE))) {
      gitignore_content <- c(gitignore_content, secrets_pattern)
      need_to_update <- TRUE
      message("Added secrets.R to .gitignore")
    }

    # Check for tmp/* pattern
    if (!any(grepl(tmp_pattern, gitignore_content, fixed = TRUE))) {
      gitignore_content <- c(gitignore_content, tmp_pattern)
      need_to_update <- TRUE
      message("Added tmp/* to .gitignore")
    }

    # Check for *.pem pattern
    if (!any(grepl(pem_pattern, gitignore_content, fixed = TRUE))) {
      gitignore_content <- c(gitignore_content, pem_pattern)
      need_to_update <- TRUE
      message("Added *.pem to .gitignore")
    }

    # Check for .wizaRdry_prefs pattern
    if (!any(grepl(prefs_pattern, gitignore_content, fixed = TRUE))) {
      gitignore_content <- c(gitignore_content, prefs_pattern)
      need_to_update <- TRUE
      message("Added .wizaRdry_prefs to .gitignore")
    }

    if (need_to_update) {
      writeLines(gitignore_content, gitignore_file)
    }
  } else {
    # Create new .gitignore with all patterns
    writeLines(c(secrets_pattern, tmp_pattern, pem_pattern, prefs_pattern), gitignore_file)
    message("Created .gitignore with secrets.R, *.pem, tmp/*, and .wizaRdry_prefs patterns")
  }

  # Ensure .gitkeep exists in tmp directory but is tracked
  # This allows the tmp directory to be committed while its contents are ignored
  tmp_gitkeep <- file.path(path, "tmp", ".gitkeep")
  if (!file.exists(tmp_gitkeep)) {
    file.create(tmp_gitkeep)
    message("Created .gitkeep in tmp directory")
  }

  # Create template config.yml file
  config_file <- file.path(path, "config.yml")
  config_template <- paste(
    "default:",
    paste0("  study_alias: ", ifelse(isFALSE(study_alias), "test", tolower(study_alias))),
    "  identifier: src_subject_id",
    "  mongo:",
    "    #database: ${study_alias}",
    "  qualtrics:",
    "    #survey_ids:",
    "      Institution1:",
    "        foo: \"SV_\"",
    "      Institution2:",
    "        bar: \"SV_\"",
    "  redcap:",
    "    #primary_key: record_id",
    "    #superkey: ndar_subject01",
    "  # Configuration for both SQL and ORACLE",
    "  sql:",
    "    # Primary key and superkey table configuration",
    "    #primary_key: 'PARTICIPANTIDENTIFIER'",
    "    #superkey: 'COMPASS_DATA.STUDYPARTICIPANTS'",
    "    # Schema of interest (for displaying tables)",
    "    #schema: 'COMPASS_DATA'",
    "    # Fields marked as PII to excude (personally identifiable information)",
    "    #pii_fields:",
    "      - 'DATEOFBIRTH'",
    "      - 'DOB'",
    "      - 'EMAIL'",
    "      - 'FIRSTNAME'",
    "      - 'LASTNAME'",
    "      - 'SSN'",
    "      - 'PHONE'",
    "      - 'ADDRESS'",
    "  missing_data_codes:",
    "    #skipped:",
    "      #- -888   # Skip pattern/branching logic",
    "    #refused:",
    "      #- -9999  # Explicitly declined to answer",
    "      #- -1",
    "    #missing:",
    "      #- 777    # Missing for unknown reasons",
    "    #undefined:",
    "      #- -555   # Otherwise undefined value",
    sep = "\n"
  )

  if (!file.exists(config_file) || overwrite) {
    writeLines(config_template, config_file)
    created <- c(created, config_file)
  } else {
    created <- c(created, paste0(config_file, " (skipped, already exists)"))
  }

  # Create template secrets.R file
  secrets_file <- file.path(path, "secrets.R")
  secrets_template <- paste(
    "# THE FIRST  RULE OF SECRETS.R IS YOU DO NOT PUSH SECRETS.R",
    "# THE SECOND RULE OF SECRETS.R IS YOU DO NOT PUSH SECRETS.R",
    "# THE THIRD  RULE OF SECRETS.R IS YOU DO NOT MODIFY .GITIGNORE",
    "",
    "# Qualtrics",
    "apiKeys <- c(\"\")",
    "baseUrls <- c(\"\")",
    "",
    "# REDCap",
    "uri <- \"\"",
    "token <- \"\"",
    "",
    "# Mongo",
    "connectionString <- \"\"",
    "",
    "# SQL or ORACLE Server",
    "host <- \"\"",
    "uid <- \"\"",
    "pwd <- \"\"",
    sep = "\n"
  )

  if (!file.exists(secrets_file) || overwrite) {
    writeLines(secrets_template, secrets_file)
    created <- c(created, secrets_file)
  } else {
    created <- c(created, paste0(secrets_file, " (skipped, already exists)"))
  }

  # Create template main.R file
  main_file <- file.path(path, "main.R")

  if (!file.exists(main_file) || overwrite) {
    writeLines(c(
      "# Main analysis script for this wizaRdry project",
      "",
      "# Load wizaRdry library:",
      "if(!require(wizaRdry)) {install.packages('wizaRdry')}; library(wizaRdry)",
      "",
      "# Check available REDCap forms:",
      "# redcap.index()",
      "",
      "# Example NDA request:",
      "# Create remediation script in nda/",
      "# nda(\"cde_dsm5crossad01\")",
      "",
      "# Your analysis code here",
      ""
    ), main_file)
    created <- c(created, main_file)
  } else {
    created <- c(created, paste0(main_file, " (skipped, already exists)"))
  }

  # Define template files with their content
  clean_templates <- list(
    collection = list(
      path = file.path(path, "clean", "mongo", "collection.R"),
      content = paste(
        "#",
        "# clean/mongo/collection.R",
        "#",
        "# config:  database name is defined in config.yml",
        "# secrets: connectionString is defined in secrets.R",
        "# encrypt: the *.pem file must be placed in the root of this repository",
        "#",
        "# return a list of the collection(s) from MongoDB",
        "# mongo.index()",
        "#",
        "# get collection from MongoDB",
        "# IMPORTANT: both variable name and script filename must match",
        "collection <- mongo(\"collection\")",
        "",
        "# cleaning script code...",
        "",
        "# final df must be named like the R script and appended with _clean",
        "collection_clean <- collection",
        sep = "\n"
      )
    ),
    survey = list(
      path = file.path(path, "clean", "qualtrics", "survey.R"),
      content = paste(
        "#",
        "# clean/qualtrics/survey.R",
        "#",
        "# get survey from Qualtrics database",
        "# config:  surveys are defined in config.yml as key-value pairs",
        "# secrets: baseUrls and apiKeys are defined in secrets.R",
        "#",
        "# return a list of the survey(s) from Qualtrics",
        "# qualtrics.index()",
        "#",
        "# get survey from Qualtrics",
        "# IMPORTANT: both variable name and script filename must match",
        "survey <- qualtrics(\"survey\")",
        "",
        "# cleaning script code...",
        "",
        "# IMPORTANT: final df must be appended with _clean",
        "survey_clean <- survey",
        sep = "\n"
      )
    )
  )

  nda_templates <- list(
    cde_dsm5crossad01 = list(
      path = file.path(path, "nda", "redcap", "cde_dsm5crossad01.R"),
      content = paste(
        "#",
        "# nda/redcap/cde_dsm5crossad01.R",
        "#",
        "# config:  superkey instrument is defined in config.yml",
        "# secrets: uri and token are defined in secrets.R",
        "#",
        "# return a list of the instrument_name(s) from REDCap",
        "# redcap.index()",
        "#",
        "# get the instrument_name dsm_5 from REDCap",
        "# IMPORTANT: both variable name and script filename must match the NDA data structure alias",
        "cde_dsm5crossad01 <- redcap(\"dsm_5\")",
        "",
        "# nda remediation code...",
        "",
        "# IMPORTANT: final df name must still match the NDA data structure alias",
        "",
        sep = "\n"
      )
    )
  )

  # Create example template files only if examples=TRUE
  if (examples) {
    # Create clean templates
    for (template_name in names(clean_templates)) {
      template <- clean_templates[[template_name]]
      if (!file.exists(template$path)) {
        writeLines(template$content, template$path)
        created <- c(created, template$path)
      } else {
        created <- c(created, paste0(template$path, " (skipped, already exists)"))
      }
    }

    # Create NDA templates
    for (template_name in names(nda_templates)) {
      template <- nda_templates[[template_name]]
      if (!file.exists(template$path)) {
        writeLines(template$content, template$path)
        created <- c(created, template$path)
      } else {
        created <- c(created, paste0(template$path, " (skipped, already exists)"))
      }
    }
  } else {
    # Skip creating example templates
    message("Skipping creation of example scripts (examples=FALSE)")
  }

  # Create R project file if requested
  if (create_project) {
    # Get directory name for default project name
    dir_name <- basename(normalizePath(path))

    # Check for existing .Rproj files
    rproj_files <- list.files(path, pattern = "\\.Rproj$", full.names = FALSE)

    if (length(rproj_files) == 0) {
      # No project file exists, create one
      project_file <- file.path(path, paste0(dir_name, ".Rproj"))

      project_template <- paste(
        "Version: 1.0",
        "",
        "RestoreWorkspace: No",
        "SaveWorkspace: No",
        "AlwaysSaveHistory: Default",
        "",
        "EnableCodeIndexing: Yes",
        "UseSpacesForTab: Yes",
        "NumSpacesForTab: 2",
        "Encoding: UTF-8",
        "",
        "RnwWeave: Sweave",
        "LaTeX: pdfLaTeX",
        "",
        "AutoAppendNewline: Yes",
        "StripTrailingWhitespace: Yes",
        sep = "\n"
      )

      writeLines(project_template, project_file)
      created <- c(created, project_file)

      message(paste0("Created R project: ", project_file))
    } else {
      message(paste0("R project already exists: ", rproj_files[1]))
    }
  }

  # Determine appropriate message based on action
  if (structure_exists && !structure_complete && repair) {
    header_message <- "The wizaRdry crystal ball has repaired your project structure:"
  } else if (!structure_exists) {
    header_message <- "The wizaRdry crystal ball has created your project structure:"
  } else {
    header_message <- "The wizaRdry crystal ball has verified your project structure:"
  }

  # Print header message only
  message(header_message)

  # Display file tree if requested
  if (show_tree) {
    # Display tree without a header
    display_tree(path)

    # Save preference that tree has been shown
    user_prefs$shown_tree <- TRUE
    saveRDS(user_prefs, user_prefs_file)

    message("\nNote: This tree view will only be shown once.")
    message("Use scry(show_tree = TRUE) to display it again in the future.")
  }

  message("\nYour next enchantments:")
  message("1. Update config.yml with your study-specific configuration")
  message("2. Add your API credentials to secrets.R (DO NOT COMMIT this file)")
  message("3. Create data cleaning scripts with clean()")
  message("4. Create NDA remediation scripts with nda()")

  if (create_project && length(rproj_files) == 0) {
    message("5. Open the newly created .Rproj file to work in this project environment")
  }

  return(invisible(TRUE))
}

#' Display a file tree structure similar to the Unix tree command
#'
#' @param path The path to display as a tree
#' @return NULL (called for side effects)
#' @keywords internal
display_tree <- function(path) {
  # Function to list all files and directories in a directory
  list_all <- function(dir_path) {
    # Get all files and directories in the directory
    all_items <- list.files(dir_path,
      all.files = FALSE, include.dirs = TRUE,
      recursive = FALSE, full.names = TRUE
    )

    # Separate dirs and files
    is_dir <- file.info(all_items)$isdir
    dirs <- all_items[is_dir]
    files <- all_items[!is_dir]

    # Check specifically for main.R file and add it if it exists
    main_r_file <- file.path(dir_path, "main.R")
    if (file.exists(main_r_file) && !(main_r_file %in% files)) {
      files <- c(files, main_r_file)
    }

    # Sort each group
    dirs <- sort(dirs)
    files <- sort(files)

    # Combine with directories first
    return(c(dirs, files))
  }

  # Function to process a directory
  process_dir <- function(dir_path, prefix) {
    sub_items <- list_all(dir_path)

    # Update counts
    dir_count <<- dir_count + sum(file.info(sub_items)$isdir)
    file_count <<- file_count + sum(!file.info(sub_items)$isdir)

    # Process each item
    for (j in seq_along(sub_items)) {
      sub_item <- sub_items[j]
      is_sub_last <- (j == length(sub_items))
      is_sub_dir <- file.info(sub_item)$isdir

      # Determine connector
      sub_connector <- if (is_sub_last) "\u2514\u2500\u2500 " else "\u251c\u2500\u2500 "
      sub_prefix <- if (is_sub_last) "    " else "\u2502   "

      # Print item
      sub_name <- basename(sub_item)
      if (is_sub_dir) {
        cat(paste0(prefix, sub_connector, sub_name, "\n"))
        # Recurse
        process_dir(sub_item, paste0(prefix, sub_prefix))
      } else {
        cat(paste0(prefix, sub_connector, sub_name, "\n"))
      }
    }
  }

  # Initialize stats counters
  dir_count <- 0
  file_count <- 0

  # Start with root
  cat("\n")
  cat(".\n")

  # Get top-level items
  items <- list_all(path)

  # Count directories and files at top level
  dir_count <- sum(file.info(items)$isdir)
  file_count <- sum(!file.info(items)$isdir)

  # Process each top-level item
  for (i in seq_along(items)) {
    item <- items[i]
    is_last <- (i == length(items))
    is_dir <- file.info(item)$isdir

    # Determine connector
    connector <- if (is_last) "\u2514\u2500\u2500 " else "\u251c\u2500\u2500 "

    # Print item
    item_name <- basename(item)
    if (is_dir) {
      cat(paste0(connector, item_name, "\n"))

      # Process items in this directory
      process_dir(item, if (is_last) "    " else "\u2502   ")
    } else {
      cat(paste0(connector, item_name, "\n"))
    }
  }

  # Print stats
  cat("\n")
  cat(paste0(dir_count, " directories, ", file_count, " files\n"))

  return(invisible(NULL))
}

#' NDA Validation Helper Functions
#'
#' @description
#' Core validation logic for NDA data structure compliance.
#' Includes value range checking, GUID validation, and violation tracking.
#'

#' Determine if validation messages should be shown
#'
#' @description
#' Helper function to centralize message suppression logic.
#' Messages are shown in strict mode OR when verbose is enabled.
#' In lenient non-verbose mode, individual messages are suppressed
#' to avoid duplication with summary warnings.
#'
#' @param strict Logical - if TRUE, enforce strict validation
#' @param verbose Logical - if TRUE, show detailed messages
#' @return Logical - TRUE if messages should be shown
#' @noRd
should_show_validation_message <- function(strict, verbose) {
  strict || verbose
}

#' Check for value range violations and update ValidationState
#'
#' @description
#' THE KEY FUNCTION for fixing value range violation tracking.
#' Scans all fields for violations and updates the ValidationState object.
#' This replaces the scattered violation checking in the old validate_structure().
#'
#' @param state ValidationState object to update
#' @param elements NDA structure dataElements
#' @param verbose Logical - print detailed output
#' @return List of violations detected (for logging/debugging)
#' @noRd
check_value_range_violations <- function(state, elements, verbose = FALSE) {
  if (verbose) cat("\n\nChecking value ranges...")
  
  df <- state$get_df()
  violations_detected <- list()
  df_cols <- names(df)
  
  for (col in intersect(df_cols, elements$name)) {
    element <- elements[elements$name == col, ]
    
    if (nrow(element) == 0) next
    
    value_range <- element$valueRange
    data_type <- element$type
    
    # Case 1: Field has no valueRange - check if this is an unbounded type
    if (is.null(value_range) || is.na(value_range) || value_range == "") {
      # String, GUID, Date, Integer, Float without valueRange are VALID (unbounded)
      # These types don't require a defined range to be compliant
      unbounded_types <- c("String", "GUID", "Date", "Integer", "Float")
      
      if (data_type %in% unbounded_types) {
        if (verbose) {
          cat(sprintf("\n\nField: %s", col))
          cat(sprintf("\n  Type: %s (no valueRange defined)", data_type))
          cat(sprintf("\n  Status: VALID - Unbounded %s field", data_type))
        }
        next  # NOT a violation - skip to next field
      }
      
      # For unknown/other types without valueRange, flag for investigation
      # This is a rare edge case
      if (has_data_values(df[[col]])) {
        unique_vals <- get_unique_values(df[[col]])
        
        state$add_value_range_violation(
          field = col,
          expected = NULL,
          actual = unique_vals
        )
        
        violations_detected[[col]] <- list(
          type = "unknown_type_no_range",
          field = col,
          data_type = data_type,
          expected = NULL,
          actual = unique_vals,
          count = length(unique_vals)
        )
        
        if (verbose) {
          cat(sprintf("\n\nField: %s (no valueRange defined)", col))
          cat(sprintf("\n  Type: %s (unknown/unhandled type)", data_type))
          cat(sprintf("\n  Data contains %d unique values", length(unique_vals)))
          cat(sprintf("\n  Sample values: %s", 
                     paste(head(unique_vals, 5), collapse = ", ")))
          if (length(unique_vals) > 5) {
            cat(sprintf(" (and %d more)", length(unique_vals) - 5))
          }
        }
      }
      next
    }
    
    # Case 2: Field has valueRange - check for violations
    if (verbose) {
      cat(sprintf("\n\nField: %s", col))
      cat(sprintf("\n  Type: %s", element$type))
      cat(sprintf("\n  Expected range: %s", value_range))
    }
    
    violating_values <- get_violations(df[[col]], value_range)
    
    if (length(violating_values) > 0) {
      state$add_value_range_violation(
        field = col,
        expected = value_range,
        actual = violating_values
      )
      
      state$is_valid <- FALSE
      
      violations_detected[[col]] <- list(
        type = "range_violation",
        field = col,
        expected_range = value_range,
        violating_values = violating_values,
        count = length(violating_values)
      )
      
      if (verbose) {
        cat("\n  ERROR: Value range violations found:")
        cat(sprintf("\n    Invalid values: %s",
                   paste(head(violating_values, 5), collapse = ", ")))
        if (length(violating_values) > 5) {
          cat(sprintf(" (and %d more...)", length(violating_values) - 5))
        }
      }
    } else if (verbose) {
      cat("\n  All values within expected range")
    }
  }
  
  if (verbose && length(violations_detected) > 0) {
    cat(sprintf("\n\nSummary: %d field(s) with value range issues", 
               length(violations_detected)))
  }
  
  return(violations_detected)
}

#' Check if field has actual data values (not just NAs)
#' @param field_values Vector of field values
#' @return Logical
#' @noRd
has_data_values <- function(field_values) {
  !all(is.na(field_values)) && length(unique(field_values[!is.na(field_values)])) > 0
}

#' Get unique non-NA values as sorted character vector
#' @param field_values Vector of field values
#' @return Character vector of unique values
#' @noRd
get_unique_values <- function(field_values) {
  unique_vals <- unique(field_values[!is.na(field_values)])
  as.character(sort(unique_vals))
}

#' Detect fields in dataframe not in NDA structure
#' @param df Data frame
#' @param elements NDA structure dataElements
#' @return Character vector of new field names
#' @noRd
detect_new_fields <- function(df, elements, dcc = FALSE) {
  df_cols <- names(df)
  structure_cols <- elements$name
  
  # Find new fields
  new_fields <- setdiff(df_cols, structure_cols)
  
  # Exclude special API-specific fields
  exclude_patterns <- c("_complete$", "^record_id$", "^redcap_event_name$")
  for (pattern in exclude_patterns) {
    new_fields <- new_fields[!grepl(pattern, new_fields)]
  }
  
  # Exclude super-required fields (added by addNdarSubjectElements)
  super_required <- SUPER_REQUIRED_FIELDS
  new_fields <- setdiff(new_fields, super_required)
  
  # Exclude DCC fields when dcc=FALSE
  # These fields may exist in the user's data but aren't being submitted to NDA
  # when dcc=FALSE, so they shouldn't trigger "MODIFIED" structure status
  if (!dcc) {
    new_fields <- setdiff(new_fields, DCC_FIELDS)
  }
  
  return(new_fields)
}

#' Get violations for a specific field value against its range
#'
#' @description
#' Core function that checks if values violate the specified range.
#' Handles numeric ranges (::), categorical values (;), and mixed ranges.
#' Extracted from original ndaValidator.R lines 2040-2127.
#'
#' @param value Vector of values to check
#' @param range_str Value range string from NDA structure (e.g., "1::10" or "A;B;C")
#' @return Character vector of violating values
#' @noRd
get_violations <- function(value, range_str) {
  if (is.null(range_str) || is.na(range_str) || range_str == "") return(character(0))
  
  # First check if there are non-numeric values (when expected numeric)
  if (grepl("::", range_str)) {
    # Numeric range expected - check for non-numeric values
    if (!is.numeric(value)) {
      # Try to convert to numeric to find which values can't be converted
      value_numeric <- suppressWarnings(as.numeric(as.character(value)))
      non_numeric_mask <- !is.na(value) & is.na(value_numeric)
      
      if (any(non_numeric_mask)) {
        non_numeric_values <- unique(value[non_numeric_mask])
        if (length(non_numeric_values) > 0) {
          return(sort(as.character(non_numeric_values)))
        }
      }
    }
  }
  
  # Special case for mixed ranges like "1::26;77"
  if (grepl("::", range_str) && grepl(";", range_str)) {
    # Split by semicolon first
    parts <- strsplit(range_str, ";")[[1]]
    # Get the numeric range from the first part
    range_part <- parts[grepl("::", parts)][1]
    range <- as.numeric(strsplit(range_part, "::")[[1]])
    # Get individual values from other parts
    individual_values <- as.numeric(parts[!grepl("::", parts)])
    # Combine valid values: numbers in range plus individual values
    valid_values <- c(seq(from = range[1], to = range[2]), individual_values)
    
    # Check for violations - handle non-numeric values appropriately
    if (is.numeric(value)) {
      invalid_mask <- !value %in% valid_values
      invalid_mask[is.na(invalid_mask)] <- FALSE
      return(sort(unique(value[invalid_mask])))
    } else {
      # For non-numeric values, we need to convert to numeric first
      value_numeric <- suppressWarnings(as.numeric(as.character(value)))
      # Identify values that can be converted but are outside range
      convertible_mask <- !is.na(value_numeric)
      invalid_mask <- convertible_mask & !value_numeric %in% valid_values
      invalid_mask[is.na(invalid_mask)] <- FALSE
      
      result <- sort(unique(value[invalid_mask]))
      # If no numeric violations, return empty
      if (length(result) == 0) return(character(0))
      return(result)
    }
  }
  
  # Rest of the function for simple ranges
  if (grepl("::", range_str)) {
    # Numeric range check
    range <- as.numeric(strsplit(range_str, "::")[[1]])
    
    # Handle non-numeric values appropriately
    if (is.numeric(value)) {
      invalid_mask <- value < range[1] | value > range[2]
      invalid_mask[is.na(invalid_mask)] <- FALSE
      return(sort(unique(value[invalid_mask])))
    } else {
      # For non-numeric values, we need to convert to numeric first
      value_numeric <- suppressWarnings(as.numeric(as.character(value)))
      # Identify values that can be converted but are outside range
      convertible_mask <- !is.na(value_numeric)
      invalid_mask <- convertible_mask & (value_numeric < range[1] | value_numeric > range[2])
      invalid_mask[is.na(invalid_mask)] <- FALSE
      
      result <- sort(unique(value[invalid_mask]))
      # If no numeric violations, return empty
      if (length(result) == 0) return(character(0))
      return(result)
    }
  } else if (grepl(";", range_str)) {
    # Categorical values
    valid_values <- trimws(strsplit(range_str, ";")[[1]])
    
    # Convert to character for comparison
    value_char <- as.character(value)
    invalid_mask <- !value_char %in% valid_values
    invalid_mask[is.na(invalid_mask)] <- FALSE
    return(sort(unique(value[invalid_mask])))
  }
  
  return(character(0))
}

#' Check required and recommended fields for missing data
#'
#' @description
#' Validates that fields contain sufficient non-NA data according to NDA requirements.
#' In strict mode, ANY NA in required fields or ALL NA in recommended fields = failure.
#' In lenient mode, these trigger warnings only.
#'
#' IMPORTANT: Only checks the 5 super required fields from ndar_subject01.
#' Structure-level required fields (like phq9_1, phq9_2, etc.) are NOT validated.
#'
#' @param state ValidationState object
#' @param elements Data elements from NDA structure
#' @param super_required_fields Character vector - the 5 mandatory NDA fields to check
#' @param strict Logical - if TRUE, enforce strict validation
#' @param verbose Logical - print detailed output
#' @return List with two elements: required_violations and recommended_violations
#' @keywords internal
#' @noRd
check_field_data_completeness <- function(state, elements, super_required_fields = SUPER_REQUIRED_FIELDS, strict = TRUE, verbose = FALSE) {
  df <- state$get_df()
  
  # Force elements to be base data.frame to avoid tibble/data.table evaluation issues
  # When dplyr is loaded, jsonlite::fromJSON returns tibbles which have different
  # scoping behavior during subsetting operations (e.g., tibble[condition, ])
  if (!is.data.frame(elements) || inherits(elements, c("tbl_df", "tbl", "data.table"))) {
    elements <- as.data.frame(elements, stringsAsFactors = FALSE)
  }
  
  required_violations <- list()
  recommended_violations <- list()
  
  # Get required elements - ONLY check super required fields
  # Structure-level required fields (like phq9_*) are NOT validated
  # This prevents false positives for fields that are required in the structure
  # but not part of the core 5 super required fields from ndar_subject01
  required_elements <- elements[
    elements$required == "Required" & 
    elements$name %in% super_required_fields,
  ]
  
  # Get recommended elements (check all of them)
  recommended_elements <- elements[elements$required == "Recommended", ]
  
  # Check REQUIRED fields for ANY NA
  if (nrow(required_elements) > 0) {
    for (i in 1:nrow(required_elements)) {
      field_name <- required_elements$name[i]
      
      if (field_name %in% names(df)) {
        field_values <- df[[field_name]]
        na_count <- sum(is.na(field_values))
        total_count <- length(field_values)
        
        # In strict mode: ANY NA is a violation
        # In lenient mode: ANY NA is also a violation (but handled as warning)
        if (na_count > 0) {
          required_violations[[field_name]] <- list(
            field = field_name,
            total_rows = total_count,
            na_count = na_count,
            percent_missing = round((na_count / total_count) * 100, 1),
            issue = if (na_count == total_count) {
              "All values are NA"
            } else {
              sprintf("%d of %d values are NA (%.1f%%)", na_count, total_count, (na_count / total_count) * 100)
            }
          )
          
          # Show individual messages only in strict mode OR verbose mode
          # In lenient non-verbose mode, suppress to avoid duplication with summary
          if (should_show_validation_message(strict, verbose)) {
            message(sprintf("[REQUIRED FIELD ISSUE] %s: %s", field_name, required_violations[[field_name]]$issue))
          }
        }
      } else {
        # Field doesn't exist at all
        required_violations[[field_name]] <- list(
          field = field_name,
          total_rows = nrow(df),
          na_count = nrow(df),
          percent_missing = 100,
          issue = "Field missing from dataframe"
        )
        
        # Show individual messages only in strict mode OR verbose mode
        if (should_show_validation_message(strict, verbose)) {
          message(sprintf("[REQUIRED FIELD ISSUE] %s: Field missing from dataframe", field_name))
        }
      }
    }
  }
  
  # Check RECOMMENDED fields for ALL NA (only matters in strict mode)
  if (strict && nrow(recommended_elements) > 0) {
    for (i in 1:nrow(recommended_elements)) {
      field_name <- recommended_elements$name[i]
      
      if (field_name %in% names(df)) {
        field_values <- df[[field_name]]
        
        # Only flag if ALL values are NA
        if (all(is.na(field_values))) {
          recommended_violations[[field_name]] <- list(
            field = field_name,
            total_rows = length(field_values),
            na_count = length(field_values),
            percent_missing = 100,
            issue = "All values are NA"
          )
          
          # Show individual messages only in strict mode OR verbose mode
          if (should_show_validation_message(strict, verbose)) {
            message(sprintf("[RECOMMENDED FIELD ISSUE] %s: All values are NA", field_name))
          }
        }
      }
    }
  }
  
  return(list(
    required = required_violations,
    recommended = recommended_violations
  ))
}

#' Check DCC Fields for Missing Data
#'
#' @description
#' Validates that DCC (Data Coordinating Center) fields contain sufficient
#' non-NA data when dcc = TRUE. Checks both required and recommended DCC fields.
#'
#' @param state ValidationState object
#' @param elements Data elements from NDA structure
#' @param dcc_required_fields Character vector - DCC required fields to check
#' @param dcc_recommended_fields Character vector - DCC recommended fields to check
#' @param strict Logical - if TRUE, enforce strict validation
#' @param verbose Logical - print detailed output
#' @return List with two elements: required_violations and recommended_violations
#' @keywords internal
#' @noRd
check_dcc_fields <- function(state, elements, 
                             dcc_required_fields = DCC_REQUIRED_FIELDS,
                             dcc_recommended_fields = DCC_RECOMMENDED_FIELDS,
                             strict = TRUE, verbose = FALSE) {
  df <- state$get_df()
  
  # Force elements to be base data.frame
  if (!is.data.frame(elements) || inherits(elements, c("tbl_df", "tbl", "data.table"))) {
    elements <- as.data.frame(elements, stringsAsFactors = FALSE)
  }
  
  required_violations <- list()
  recommended_violations <- list()
  
  # Get DCC required elements that are in the structure
  dcc_required_in_structure <- elements[
    elements$required == "Required" & 
    elements$name %in% dcc_required_fields,
  ]
  
  # Get DCC recommended elements that are in the structure
  dcc_recommended_in_structure <- elements[
    elements$required == "Recommended" &
    elements$name %in% dcc_recommended_fields,
  ]
  
  # Check DCC required fields - ANY NA = violation
  if (nrow(dcc_required_in_structure) > 0) {
    for (i in seq_len(nrow(dcc_required_in_structure))) {
      field_name <- dcc_required_in_structure$name[i]
      
      # Safety: Convert to character if it's a list
      if (is.list(field_name)) {
        field_name <- as.character(unlist(field_name))
      }
      field_name <- as.character(field_name)
      
      if (field_name %in% names(df)) {
        field_data <- df[[field_name]]
        na_count <- sum(is.na(field_data))
        
        if (na_count > 0) {
          issue <- sprintf("DCC required field has %d/%d missing values",
                          na_count, length(field_data))
          required_violations[[field_name]] <- list(
            field = field_name,
            issue = issue,
            na_count = na_count,
            total_count = length(field_data)
          )
          
          if (should_show_validation_message(strict, verbose)) {
            message(sprintf("  [%s] %s: %s",
                           if (strict) "ERROR" else "WARN",
                           field_name,
                           issue))
          }
        }
      }
    }
  }
  
  # Check DCC recommended fields - ALL NA = violation (strict mode only)
  if (strict && nrow(dcc_recommended_in_structure) > 0) {
    for (i in seq_len(nrow(dcc_recommended_in_structure))) {
      field_name <- dcc_recommended_in_structure$name[i]
      
      # Safety: Convert to character if it's a list
      if (is.list(field_name)) {
        field_name <- as.character(unlist(field_name))
      }
      field_name <- as.character(field_name)
      
      if (field_name %in% names(df)) {
        field_data <- df[[field_name]]
        
        if (all(is.na(field_data))) {
          issue <- "DCC recommended field is entirely missing (all NA)"
          recommended_violations[[field_name]] <- list(
            field = field_name,
            issue = issue,
            na_count = length(field_data),
            total_count = length(field_data)
          )
          
          if (should_show_validation_message(strict, verbose)) {
            message(sprintf("  [%s] %s: %s",
                           if (strict) "ERROR" else "WARN",
                           field_name,
                           issue))
          }
        }
      }
    }
  }
  
  result <- base::list(
    required_violations = required_violations,
    recommended_violations = recommended_violations
  )
  return(result)
}

#' Print validation summary
#' @param state ValidationState object
#' @noRd
print_validation_summary <- function(state) {
  message("\n\nValidation Summary:")
  message(sprintf("- Status: %s", if(state$is_valid) "PASSED" else "FAILED"))
  message(sprintf("- Structure Type: %s", 
                 if(state$is_new_structure) "NEW" 
                 else if(state$is_modified_structure) "MODIFIED" 
                 else "EXISTING"))
  
  if (!state$is_new_structure) {
    message(sprintf("- Modified: %s", 
                   if(state$is_modified_structure) "YES" else "NO"))
  }
  
  if (length(state$value_range_violations) > 0) {
    message(sprintf("- Value Range Violations: %d field(s) (%s)",
                   length(state$value_range_violations),
                   paste(names(state$value_range_violations), collapse = ", ")))
  }
  
  if (length(state$new_fields) > 0) {
    message(sprintf("- New Fields: %d (%s)",
                   length(state$new_fields),
                   paste(state$new_fields, collapse = ", ")))
  }
  
  if (length(state$missing_required) > 0) {
    message(sprintf("- Missing Required: %d field(s) (%s)",
                   length(state$missing_required),
                   paste(state$missing_required, collapse = ", ")))
  }
  
  message(sprintf("- Needs Data Definition: %s (reason: %s)",
                 if(state$needs_data_definition()) "YES" else "NO",
                 state$get_modification_reason()))
}

#' Create Data Extracts in Various Formats
#'
#' This function creates data extracts from a given dataframe in one or more of 
#' the following formats: CSV, RDS, and SPSS.
#' It sources the necessary functions for creating each file type from their 
#' respective scripts and executes them based on the arguments passed. The user 
#' specifies the desired output formats using boolean flags for each file type.
#'
#' @param df The dataframe to be exported.
#' @param df_name The base name for the output file(s) without extension.
#' @param csv Logical; if TRUE, a CSV file is created using the `createCsv` function.
#' @param rdata Logical; if TRUE, an RDS file is created using the `createRds` function.
#' @param spss Logical; if TRUE, an SPSS file is created using the `createSpss` function.
#' @examples
#' createExtract(mtcars, "mtcars_export", csv = TRUE, rds = TRUE, spss = FALSE)
#' @noRd
#' @author Joshua Kenney <joshua.kenney@yale.edu>

createExtract <- function(df, df_name, csv=NULL, rdata=NULL, spss=NULL) {
  
  
  if (!is.null(csv) && csv) {
    createCsv(df, df_name)
  }
  if (!is.null(rdata) && rdata) {
    createRds(df, df_name)
  }
  if (!is.null(spss) && spss) {
    createSpss(df, df_name)
  }
}

#' Check for Duplicate Script Names
#'
#' This function checks for duplicate script names across the data source folders
#' to prevent naming conflicts. Checks clean/ and nda/ directories separately.
#'
#' @param path Character string specifying the directory path to check.
#'        Defaults to the current working directory.
#' @param quiet Logical. If TRUE, will not print messages about non-duplicate scripts.
#'        Defaults to TRUE.
#'
#' @return Invisibly returns a list of duplicate script names and their locations.
#'         If no duplicates are found, returns NULL.
#' 
#' @noRd
checkDuplicateScripts <- function(path = ".", quiet = TRUE) {
  duplicates_found <- FALSE
  result <- list()
  
  # Check clean/ subdirectories
  clean_dirs <- c(
    redcap = file.path(path, "clean", "redcap"),
    qualtrics = file.path(path, "clean", "qualtrics"),
    mongo = file.path(path, "clean", "mongo")
  )
  
  # Check if clean directories exist
  clean_dir_exists <- sapply(clean_dirs, dir.exists)
  
  if (any(clean_dir_exists)) {
    # Get all script names from each folder
    scripts <- list()
    for (dir_name in names(clean_dirs)[clean_dir_exists]) {
      scripts[[dir_name]] <- tools::file_path_sans_ext(
        list.files(clean_dirs[dir_name], pattern = "\\.R$")
      )
    }
    
    # Create a map to count occurrences of each script name
    script_counts <- table(unlist(scripts))
    
    # Find duplicates (count > 1)
    duplicates <- names(script_counts[script_counts > 1])
    
    # Check if there are any duplicates
    if (length(duplicates) > 0) {
      duplicates_found <- TRUE
      message("WARNING: Found duplicate script names in clean/ folders:")
      
      # For each duplicate, show which folders they appear in
      clean_duplicate_info <- list()
      for (dup in duplicates) {
        locations <- names(scripts)[sapply(scripts, function(x) dup %in% x)]
        locations_full <- sprintf("clean/%s", locations)
        
        message(sprintf("  - Script '%s.R' appears in: %s", 
                        dup, paste(locations_full, collapse = ", ")))
        
        clean_duplicate_info[[dup]] <- locations_full
      }
      
      result$clean <- clean_duplicate_info
    } else if (!quiet) {
      message("No duplicate script names found in clean/ folders.")
    }
  }
  
  # Check nda/ subdirectories
  nda_dirs <- c(
    redcap = file.path(path, "nda", "redcap"),
    qualtrics = file.path(path, "nda", "qualtrics"),
    mongo = file.path(path, "nda", "mongo")
  )
  
  # Check if nda directories exist
  nda_dir_exists <- sapply(nda_dirs, dir.exists)
  
  if (any(nda_dir_exists)) {
    # Get all script names from each folder
    scripts <- list()
    for (dir_name in names(nda_dirs)[nda_dir_exists]) {
      scripts[[dir_name]] <- tools::file_path_sans_ext(
        list.files(nda_dirs[dir_name], pattern = "\\.R$")
      )
    }
    
    # Create a map to count occurrences of each script name
    script_counts <- table(unlist(scripts))
    
    # Find duplicates (count > 1)
    duplicates <- names(script_counts[script_counts > 1])
    
    # Check if there are any duplicates
    if (length(duplicates) > 0) {
      duplicates_found <- TRUE
      message("WARNING: Found duplicate script names in nda/ folders:")
      
      # For each duplicate, show which folders they appear in
      nda_duplicate_info <- list()
      for (dup in duplicates) {
        locations <- names(scripts)[sapply(scripts, function(x) dup %in% x)]
        locations_full <- sprintf("nda/%s", locations)
        
        message(sprintf("  - Script '%s.R' appears in: %s", 
                        dup, paste(locations_full, collapse = ", ")))
        
        nda_duplicate_info[[dup]] <- locations_full
      }
      
      result$nda <- nda_duplicate_info
    } else if (!quiet) {
      message("No duplicate script names found in nda/ folders.")
    }
  }
  
  if (duplicates_found) {
    message("\nDuplicate scripts can cause conflicts when processing data.")
    message("Please rename one of the scripts to avoid conflicts.")
    return(invisible(result))
  } else {
    if (!quiet) {
      message("No duplicate script names found.")
    }
    return(invisible(NULL))
  }
}

#' Check for Presence of NDA Required Variables in a Data Frame
#'
#' This function checks if a cleaned data frame contains all the variables required by the National Data Archive (NDA).
#' The set of required variables can be adjusted based on the specific requirements of the study and the presence
#' of certain variables like 'visit' or 'week'. The function is useful for ensuring data compliance before submission.
#'
#' @param measure_alias A string representing the alias name of the dataset to be checked.
#' @param measure_type A string indicating the type of measure, used to adjust the list of required variables if necessary.
#' @param nda_required_variables A vector of strings representing the initial set of NDA required variables to check for.
#'        This parameter is overwritten inside the function but can be used to extend the functionality in the future.
#' @return This function does not return a value but uses the `testthat` package to assert the presence of all NDA required variables and provides detailed feedback if any are missing.
#' @examples
#' \donttest{
#' ndaRequiredVariablesExist("your_dataset_alias", "qualtrics", c("src_subject_id", "phenotype"))
#' }
#' @importFrom testthat test_that expect_true
#' @importFrom dplyr setdiff
#' @note While currently the function overwrites the 'nda_required_variables' parameter internally, future versions may allow for dynamic adjustment based on 'measure_type'.
#'       It assumes that the dataset has been cleaned and is named according to a standard naming convention with a '_clean' suffix.
#' @noRd
ndaRequiredVariablesExist <- function(measure_alias, measure_type, nda_required_variables) {


  # append _clean to the measure in question
  output_df_name <- paste0(measure_alias, "_clean")

  # store clean dataframe in df_clean
  df_clean <- base::get(output_df_name)

  # Initial list of NDA required variables
  nda_required_variables <- c("src_subject_id", "phenotype", "site",
                              "subjectkey", "sex", "interview_date", "interview_age")

  # Adjust NDA required variables based on presence of 'visit' or 'week'
  adjusted_nda_required <- nda_required_variables

  # alter required variables for redcap measures (no interview_date or interview_age)
  if (measure_type=="redcap") {
    adjusted_nda_required <- nda_required_variables <- c("src_subject_id", "phenotype", "site",
                                                         "subjectkey", "sex")
  }

  # If 'visit' and 'week' are not both required, adjust the list accordingly:
  if (!("visit" %in% colnames(df_clean)) && ("week" %in% colnames(df_clean))) {
    adjusted_nda_required <- dplyr::setdiff(adjusted_nda_required, "visit")  # Remove 'visit' if it's not there but 'week' is
  } else if (("visit" %in% colnames(df_clean)) && !("week" %in% colnames(df_clean))) {
    adjusted_nda_required <- dplyr::setdiff(adjusted_nda_required, "week")  # Remove 'week' if it's not there but 'visit' is
  } # If neither or both are present, no changes needed to adjusted_nda_required


  # Now check if the output dataframe contains all adjusted NDA required variables
  missing_vars <- dplyr::setdiff(adjusted_nda_required, colnames(df_clean))

  if (length(missing_vars) == 0) {
    base::cat("nda required variables exist: ")
    return(invisible(TRUE))
  }
  
  error_msg <- paste("SCRIPT ERROR: All NDA required variables are not present in '", measure_alias, 
                    "'. Please make sure the following variable(s) are present in the clean df: ", 
                    paste(missing_vars, collapse = ", "), ".")
  
  tryCatch({
    test_that("Check for missing NDA required variables", {
      testthat::expect_true(FALSE, info = error_msg)
    })
  }, error = function(e) {
    message(error_msg)
  })

}

#' Check if a Clean Data Frame Exists
#'
#' This function checks if a cleaned data frame, specified by the measure alias, exists in the global environment.
#' The function is designed to verify the existence of data frames intended to have been cleaned and prepared
#' under a specific naming convention (suffix '_clean').
#'
#' @param measure_alias A string representing the alias name of the dataset to be checked.
#' @param measure_type A string indicating the type of measure, currently not utilized in the function but reserved for future use.
#' @return This function does not return a value but outputs a message indicating whether the specified clean data frame exists.
#' @examples
#' \donttest{
#' cleanDataFrameExists("your_dataset_alias", "qualtrics")
#' }
#' @importFrom testthat test_that expect_true
#' @note This function assumes that the dataset, if cleaned and prepared correctly, has been named according to a standard
#'       naming convention with a '_clean' suffix. The measure_type parameter is included for potential future functionality
#'       but is not currently used.
#' @noRd
cleanDataFrameExists <- function(measure_alias, measure_type) {


  # append _clean to the measure in question
  output_df_name <- paste0(measure_alias, "_clean")

  if (exists(output_df_name)){
    base::cat("Clean data frame exists: ")
    return(invisible(TRUE))
  }
  
  error_msg <- paste("SCRIPT ERROR: The script did not create '", output_df_name, "' dataframe.")
  
  tryCatch({
    test_that("Clean df exists", {
      testthat::expect_true(FALSE, info = error_msg)
    })
  }, error = function(e) {
    message(error_msg)
  })

}

#' Check Column Naming Conventions Against NDA Requirements
#'
#' This function checks that all non-NDA columns in a dataset follow a specified naming convention,
#' prefixed with the measure alias. It is designed to support different types of datasets like Qualtrics and REDCap.
#'
#' @param measure_alias A string representing the prefix expected for non-NDA columns.
#' @param measure_type A string specifying the type of dataset, supports 'qualtrics' or 'redcap'.
#' @param nda_required_variables A vector of strings listing the NDA required variables which are excluded from the naming convention check.
#' @return The function does not return a value but outputs an error message if any non-NDA columns do not follow the naming convention.
#' @examples
#' \donttest{
#' checkColumnPrefix("dataset_name", "qualtrics", c("nda_var1", "nda_var2"))
#' }
#' @importFrom testthat test_that expect_true
#' @importFrom dplyr setdiff
#' @note This function assumes that the dataset follows a specific naming convention where non-NDA columns should be prefixed with the measure alias followed by an underscore.
#'       The actual dataset name is expected to be the measure alias suffixed with '_clean'.
#' @noRd
checkColumnPrefix <- function(measure_alias, measure_type, nda_required_variables, identifier) {

  # Construct dataframe name based on measure_alias
  df_name <- paste0(measure_alias, "_clean")

  # Fetch the dataframe
  df <- base::get(df_name)

  # Determine columns to exclude based on measure_type and identifier
  # NDA required variables are only excluded when identifier is "src_subject_id"
  cols_to_exclude <- character(0)
  
  if (identifier == "src_subject_id") {
    # Include NDA required variables when using src_subject_id
    cols_to_exclude <- c(cols_to_exclude, nda_required_variables)
  }
  
  # Always add identifier to excluded columns
  cols_to_exclude <- c(cols_to_exclude, identifier, "interview_date")
  
  # Add measure-specific columns to exclude
  if (measure_type == "qualtrics") {
    cols_to_exclude <- c(cols_to_exclude, "ResponseId")
  } else if (measure_type == "redcap") {
    cols_to_exclude <- c(cols_to_exclude, "int_start", "int_end")
  }
  
  # Get non-NDA columns (all columns minus excluded ones)
  non_nda_cols <- dplyr::setdiff(colnames(df), cols_to_exclude)

  # Check naming convention for non-NDA columns
  actual_non_conform <- non_nda_cols[!grepl(paste0("^", measure_alias, "_"), non_nda_cols)]
  
  if (length(actual_non_conform) == 0) {
    base::cat("Columns have correct naming convention ")
    return(invisible(TRUE))
  }
  
  # If there are non-conforming columns, show error
  error_msg <- paste0("SCRIPT ERROR: The following non-NDA columns in '", df_name,
                      "' do not follow the correct naming convention starting with '", measure_alias, "_':\n",
                      paste(actual_non_conform, collapse = ", "))
  
  tryCatch({
    test_that("Check column naming convention", {
      expect_true(FALSE, info = error_msg)
    })
  }, error = function(e) {
    message(error_msg)
  })
}

#' Check Interview Age Validity
#'
#' This function checks whether the 'interview_age' for all records in a specified dataset
#' falls within the acceptable age range (12 to 70 years, converted into months).
#' If any record falls outside this range, the function will list the subject IDs that do not meet this condition.
#'
#' @param measure_alias A string representing the name of the dataset to check.
#'        The dataset should have a column 'interview_age' and 'src_subject_id'.
#' @return The function itself does not return a value but will output a message listing the subject IDs
#'         with 'interview_age' outside the 12 to 70 years range if such cases exist.
#' @examples
#' \donttest{
#' checkInterviewAge("your_dataset_alias")
#' }
#' @importFrom testthat test_that expect_true
#' @note This function requires the 'testthat' package for unit testing.
#'       It is designed to operate on datasets that follow a specific naming convention,
#'       appending '_clean' to the measure_alias to construct the dataframe name.
#'       The function throws an error if the 'interview_age' falls outside the specified range.
#' @noRd
checkInterviewAge <- function(measure_alias) {

  months_in_a_year <- 12
  # define age range
  min_age <- 12 * months_in_a_year  # 144 months
  max_age <- 70 * months_in_a_year  # 840 months

  # Construct the expected dataframe name
  output_df_name <- paste0(measure_alias, "_clean")

  # Retrieve the dataframe based on constructed name
  df_clean <- base::get(output_df_name)  # specify the environment if needed

  # Check for NA values
  has_na <- base::any(base::is.na(df_clean$interview_age))
  
  if (has_na) {
    # Filter out NA values for checking
    non_na_ages <- df_clean$interview_age[!base::is.na(df_clean$interview_age)]
    non_na_subjects <- df_clean$src_subject_id[!base::is.na(df_clean$interview_age)]
    rows_not_meeting_condition <- non_na_subjects[non_na_ages < min_age | non_na_ages > max_age]
    is_valid <- all(non_na_ages >= min_age & non_na_ages <= max_age)
  } else {
    rows_not_meeting_condition <- df_clean$src_subject_id[df_clean$interview_age < min_age | df_clean$interview_age > max_age]
    is_valid <- all(df_clean$interview_age >= min_age & df_clean$interview_age <= max_age)
  }
  
  if (is_valid) {
    base::cat("Check interview_age is between ", min_age, " and ", max_age, ": ", sep = "")
    if (has_na) {
      base::cat("NA values found while checking interview age!")
    }
    return(invisible(TRUE))
  }
  
  error_msg <- paste("DATA ERROR: All values in 'interview_age' should be greater than ", min_age, 
                    " and less than ", max_age, ". src_subject_id not meeting condition: ", 
                    paste(rows_not_meeting_condition, collapse = ", "))
  
  tryCatch({
    test_that(paste0("Check interview_age is between ", min_age, " and ", max_age), {
      testthat::expect_true(FALSE, info = error_msg)
    })
  }, error = function(e) {
    message(error_msg)
  })
}
# in the case where there are NA values in interview_age column, we should not be using testthat to check normally
# for the condition bc it will flag NA values

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/secrets.R
\name{assign_secret}
\alias{assign_secret}
\title{Assign a Secret Value}
\usage{
assign_secret(name, value)
}
\arguments{
\item{name}{Name of the secret to set}

\item{value}{Value to assign}
}
\value{
Invisibly returns TRUE if successful
}
\description{
Assign a Secret Value
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getSql.R
\name{sql}
\alias{sql}
\title{Fetch data from SQL database to be stored in a data frame}
\usage{
sql(
  table_name = NULL,
  ...,
  fields = NULL,
  where_clause = NULL,
  join_primary_keys = TRUE,
  custom_query = NULL,
  max_rows = NULL,
  date_format = NULL,
  batch_size = 1000,
  pii = FALSE,
  interview_date = NULL,
  all = FALSE
)
}
\arguments{
\item{table_name}{Name of the SQL table or view to query}

\item{...}{Optional column names to filter for. Only rows with non-missing values
in ALL specified columns will be returned.}

\item{fields}{Optional vector of specific fields to select}

\item{where_clause}{Optional WHERE clause to filter results (without the "WHERE" keyword)}

\item{join_primary_keys}{Boolean, whether to join with the primary keys table (default: TRUE)}

\item{custom_query}{Optional custom SQL query to execute instead of building one}

\item{max_rows}{Optional limit on the number of rows to return}

\item{date_format}{Optional format for date fields (default uses ISO format)}

\item{batch_size}{Number of records to retrieve per batch for large datasets}

\item{pii}{Logical; if FALSE (default), remove fields marked as PII. TRUE keeps PII.}

\item{interview_date}{Optional; can be either:
- A date string in various formats (ISO, US, etc.) to filter data up to that date
- A boolean TRUE to return only rows with non-NA interview_date values}

\item{all}{Logical; if TRUE, use LEFT OUTER JOIN instead of INNER JOIN (default: FALSE),
similar to the 'all' parameter in base R's merge() function}
}
\value{
A data frame containing the requested SQL data
}
\description{
Retrieves data from a SQL table and optionally joins it with a primary keys table
as specified in the configuration.
}
\examples{
\dontrun{
# Get data from a specific table
data <- sql("participants")

# Get data with a where clause
survey_data <- sql("vw_surveyquestionresults",
                  where_clause = "resultidentifier = 'NRS'")

# Get all records, including those without matching primary key
all_data <- sql("candidate", all = TRUE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getQualtrics.R
\name{qualtrics}
\alias{qualtrics}
\title{Retrieve Survey Data from Qualtrics}
\usage{
qualtrics(
  qualtrics_alias,
  ...,
  institution = NULL,
  label = FALSE,
  interview_date = NULL,
  complete = FALSE
)
}
\arguments{
\item{qualtrics_alias}{The alias for the Qualtrics survey to be retrieved.}

\item{...}{Optional column names to filter for. Only rows with non-missing values
in ALL specified columns will be returned. This is useful for filtering
data to only include complete cases for specific variables of interest.}

\item{institution}{Optional. The institution name (e.g., "temple" or "nu"). If NULL, all institutions will be searched.}

\item{label}{Logical indicating whether to return coded values or their associated labels (default is FALSE).}

\item{interview_date}{Optional; can be either:
- A date string in various formats (ISO, US, etc.) to filter data up to that date
- A boolean TRUE to return only rows with non-NA interview_date values}

\item{complete}{Logical; default FALSE, if TRUE only returns rows where Progress == 100}
}
\value{
A cleaned and harmonized data frame containing the survey data with superkeys first.
}
\description{
Retrieve Survey Data from Qualtrics
}
\examples{
\dontrun{
# Get survey by alias (will search all institutions)
survey_data <- qualtrics("rgpts")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/NdaClasses.R
\name{NdaClasses}
\alias{NdaClasses}
\title{NDA Type Safety Classes}
\description{
Complete R6 class system for type-safe NDA data structure management.
All Excel columns are represented by typed R6 classes with validation.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ValidationState.R
\name{ValidationState}
\alias{ValidationState}
\title{ValidationState R6 Class}
\description{
Manages NDA validation state and tracks modifications to data structures.
Central object for tracking validation results, modifications, and determining
whether data definition files need to be created.
}
\details{
This class replaces the fragile attribute-passing pattern used previously.
It provides a structured way to track:
\itemize{
\item Validation status (valid/invalid, new/existing structure)
\item Value range violations
\item New fields added
\item Required field status
\item Metadata from ndar_subject01
}

The key method \code{needs_data_definition()} determines whether a data definition
file should be created based on whether the structure is new or modified.
}
\keyword{internal}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{measure_name}}{Character - name of the measure/structure}

\item{\code{api}}{Character - API type (redcap, qualtrics, mongo, etc.)}

\item{\code{data_env}}{DataEnvironment - manages dataframe across environments}

\item{\code{nda_structure}}{List - NDA structure definition from API}

\item{\code{is_valid}}{Logical - whether validation passed}

\item{\code{is_new_structure}}{Logical - whether structure is new (not in NDA)}

\item{\code{is_modified_structure}}{Logical - whether existing structure has modifications}

\item{\code{bypassed_validation}}{Logical - whether validation was bypassed (new structures)}

\item{\code{value_range_violations}}{List - fields with value range violations}

\item{\code{new_fields}}{Character vector - fields in data not in NDA structure}

\item{\code{ndar_subject_additions}}{Character vector - DCC fields added from ndar_subject01}

\item{\code{ndar_subject01_all_fields}}{Character vector - ALL field names from ndar_subject01 (~150 fields)
Used for consistent formatting in Excel exports regardless of dcc parameter}

\item{\code{renamed_fields}}{Character vector - fields that were renamed}

\item{\code{dropped_fields}}{Character vector - fields that were dropped}

\item{\code{missing_required}}{Character vector - required fields with missing data}

\item{\code{required_metadata}}{Data frame - ndar_subject01 required field metadata}

\item{\code{recommended_metadata}}{Data frame - ndar_subject01 recommended field metadata}

\item{\code{warnings}}{Character vector - warning messages}

\item{\code{errors}}{Character vector - error messages}

\item{\code{dcc}}{Logical - whether DCC fields should be validated}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-ValidationState-new}{\code{ValidationState$new()}}
\item \href{#method-ValidationState-get_df}{\code{ValidationState$get_df()}}
\item \href{#method-ValidationState-set_df}{\code{ValidationState$set_df()}}
\item \href{#method-ValidationState-add_value_range_violation}{\code{ValidationState$add_value_range_violation()}}
\item \href{#method-ValidationState-add_violations}{\code{ValidationState$add_violations()}}
\item \href{#method-ValidationState-set_valid}{\code{ValidationState$set_valid()}}
\item \href{#method-ValidationState-has_modifications}{\code{ValidationState$has_modifications()}}
\item \href{#method-ValidationState-needs_data_definition}{\code{ValidationState$needs_data_definition()}}
\item \href{#method-ValidationState-get_modification_reason}{\code{ValidationState$get_modification_reason()}}
\item \href{#method-ValidationState-to_list}{\code{ValidationState$to_list()}}
\item \href{#method-ValidationState-print}{\code{ValidationState$print()}}
\item \href{#method-ValidationState-clone}{\code{ValidationState$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ValidationState-new"></a>}}
\if{latex}{\out{\hypertarget{method-ValidationState-new}{}}}
\subsection{Method \code{new()}}{
Create a new ValidationState instance
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ValidationState$new(measure_name, api, df, nda_structure = NULL, dcc = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{measure_name}}{Name of the measure/structure}

\item{\code{api}}{API type (redcap, qualtrics, mongo, csv, oracle, sql)}

\item{\code{df}}{Initial dataframe}

\item{\code{nda_structure}}{NDA structure definition (NULL for new structures)}

\item{\code{dcc}}{Logical - whether DCC fields should be validated}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A new ValidationState object
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ValidationState-get_df"></a>}}
\if{latex}{\out{\hypertarget{method-ValidationState-get_df}{}}}
\subsection{Method \code{get_df()}}{
Get current dataframe
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ValidationState$get_df()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Data frame
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ValidationState-set_df"></a>}}
\if{latex}{\out{\hypertarget{method-ValidationState-set_df}{}}}
\subsection{Method \code{set_df()}}{
Update dataframe in all environments
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ValidationState$set_df(df)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{df}}{New dataframe}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Self (invisibly) for method chaining
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ValidationState-add_value_range_violation"></a>}}
\if{latex}{\out{\hypertarget{method-ValidationState-add_value_range_violation}{}}}
\subsection{Method \code{add_value_range_violation()}}{
Add a value range violation
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ValidationState$add_value_range_violation(field, expected, actual)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{field}}{Field name}

\item{\code{expected}}{Expected value range (NULL if no range defined)}

\item{\code{actual}}{Vector of violating values}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Self (invisibly) for method chaining
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ValidationState-add_violations"></a>}}
\if{latex}{\out{\hypertarget{method-ValidationState-add_violations}{}}}
\subsection{Method \code{add_violations()}}{
Add violations of a specific type (e.g., DCC violations)
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ValidationState$add_violations(type, violations)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{type}}{Character - type of violations ("dcc_required", "dcc_recommended", etc.)}

\item{\code{violations}}{List - violations to add}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Self (invisibly) for method chaining
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ValidationState-set_valid"></a>}}
\if{latex}{\out{\hypertarget{method-ValidationState-set_valid}{}}}
\subsection{Method \code{set_valid()}}{
Set validation status
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ValidationState$set_valid(valid)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{valid}}{Logical - TRUE if validation passed, FALSE otherwise}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Self (invisibly) for method chaining
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ValidationState-has_modifications"></a>}}
\if{latex}{\out{\hypertarget{method-ValidationState-has_modifications}{}}}
\subsection{Method \code{has_modifications()}}{
Check if structure has modifications requiring data definition
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ValidationState$has_modifications()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Logical
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ValidationState-needs_data_definition"></a>}}
\if{latex}{\out{\hypertarget{method-ValidationState-needs_data_definition}{}}}
\subsection{Method \code{needs_data_definition()}}{
Determine if data definition file is needed
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ValidationState$needs_data_definition()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Logical - TRUE if data definition should be created
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ValidationState-get_modification_reason"></a>}}
\if{latex}{\out{\hypertarget{method-ValidationState-get_modification_reason}{}}}
\subsection{Method \code{get_modification_reason()}}{
Get human-readable modification reason
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ValidationState$get_modification_reason()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Character string describing why structure is modified
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ValidationState-to_list"></a>}}
\if{latex}{\out{\hypertarget{method-ValidationState-to_list}{}}}
\subsection{Method \code{to_list()}}{
Convert to list for backward compatibility with old validation_results
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ValidationState$to_list()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
List with validation results
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ValidationState-print"></a>}}
\if{latex}{\out{\hypertarget{method-ValidationState-print}{}}}
\subsection{Method \code{print()}}{
Print method for ValidationState
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ValidationState$print()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Self (invisibly)
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-ValidationState-clone"></a>}}
\if{latex}{\out{\hypertarget{method-ValidationState-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{ValidationState$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deprecated.R
\name{wizaRdry-deprecated}
\alias{wizaRdry-deprecated}
\title{Deprecated functions in wizaRdry}
\description{
Deprecated functions in wizaRdry
}
\details{
These functions are deprecated and may be removed in a future release.
Prefer the suggested replacements.
}
\section{Deprecated functions}{

\describe{
\item{\code{createCsv(...)}}{Use \code{to.csv(...)} instead.}
\item{\code{createRds(...)}}{Use \code{to.rds(...)} instead.}
\item{\code{createSpss(...)}}{Use \code{to.sav(...)} instead.}
\item{\code{dataFilter(...)}}{Use \code{sift(...)} instead.}
\item{\code{dataMerge(...)}}{Use \code{meld(...)} instead.}
\item{\code{dataRequest(...)}}{Use \code{clean(...)} instead.}
\item{\code{getRedcap(...)}}{Use \code{redcap(...)} instead.}
\item{\code{getSurvey(...)}}{Use \code{qualtrics(...)} instead.}
\item{\code{getTask(...)}}{Use \code{mongo(...)} instead.}
\item{\code{ndaRequest(...)}}{Use \code{nda(...)} instead.}
}
}

\seealso{
help("Deprecated")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getRedcap.R
\name{getRedcap}
\alias{getRedcap}
\title{Alias for 'redcap' (DEPRECATED)}
\usage{
getRedcap(...)
}
\arguments{
\item{...}{Optional column names to filter for. Only rows with non-missing values
in ALL specified columns will be returned. This is useful for filtering
data to only include complete cases for specific variables of interest.}
}
\value{
A data frame containing the requested REDCap data
}
\description{
This function is deprecated. Please use 'redcap' instead.
This is a legacy alias for the 'redcap' function to maintain compatibility with older code.
}
\examples{
\dontrun{
# DEPRECATED - use redcap() instead
survey_data <- getRedcap("demographics")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/createNda.R, R/createNdaSubmissionTemplate.R
\name{to.nda}
\alias{to.nda}
\title{Create NDA Submission Template}
\usage{
to.nda(
  df,
  path = ".",
  skip_prompt = TRUE,
  selected_fields = NULL,
  skip_prompts = FALSE,
  verbose = FALSE
)

to.nda(
  df,
  path = ".",
  skip_prompt = TRUE,
  selected_fields = NULL,
  skip_prompts = FALSE,
  verbose = FALSE
)
}
\arguments{
\item{df}{Data frame to be used as template or character string naming a data frame
in the global environment.}

\item{path}{Character string specifying the directory path where the "tmp" folder
and template file should be created. Defaults to the current working directory.}

\item{skip_prompt}{Logical. If TRUE (default), skips the confirmation prompt. If FALSE,
will prompt for confirmation unless the user has previously chosen to remember their preference.}

\item{selected_fields}{Character vector of field names to include in template. If NULL (default),
uses all fields from data frame. Used by create_nda_files() for centralized field selection.}

\item{skip_prompts}{Logical. If TRUE, skip ALL interactive prompts (used when called from
create_nda_files() with pre-selected fields). Default: FALSE.}

\item{verbose}{Logical. If TRUE, show detailed progress messages. Default: FALSE.}
}
\value{
Invisible TRUE if successful. Creates a CSV file at the specified path
and prints a message with the file location.

Invisible TRUE if successful. Creates a CSV file at the specified path
and prints a message with the file location.
}
\description{
This function creates a CSV template file for National Data Archive (NDA) submissions.
It extracts the data from a specified data frame and formats it according to NDA requirements,
with the structure name split into base name and suffix in the first line.
The function will prompt for confirmation before creating the file,
with an option to remember the user's preference for future calls.

This function creates a CSV template file for National Data Archive (NDA) submissions.
It extracts the data from a specified data frame and formats it according to NDA requirements,
with the structure name split into base name and suffix in the first line.
The function will prompt for confirmation before creating the file,
with an option to remember the user's preference for future calls.
}
\details{
The function will:
\enumerate{
\item Create a 'tmp' directory if it doesn't exist
\item Parse the structure name into base and suffix components (e.g., "eefrt01" -> "eefrt" and "01")
\item Write the structure name components as the first line
\item Write column headers as the second line
\item Write the data rows below
}

The function will:
\enumerate{
\item Create a 'tmp' directory if it doesn't exist
\item Parse the structure name into base and suffix components (e.g., "eefrt01" -> "eefrt" and "01")
\item Write the structure name components as the first line
\item Write column headers as the second line
\item Write the data rows below
}
}
\examples{
\dontrun{
  # First create some sample data
  eefrt01 <- data.frame(
    src_subject_id = c("SUB001", "SUB002"),
    interview_age = c(240, 360),
    interview_date = c("01/01/2023", "02/15/2023"),
    response_time = c(450, 520)
  )

  # Create the NDA template using the data frame directly
  to.nda(eefrt01)

  # Or using the name as a string
  to.nda("eefrt01")

  # Skip the confirmation prompt
  to.nda(eefrt01, skip_prompt = TRUE)
}

\dontrun{
  # First create some sample data
  eefrt01 <- data.frame(
    src_subject_id = c("SUB001", "SUB002"),
    interview_age = c(240, 360),
    interview_date = c("01/01/2023", "02/15/2023"),
    response_time = c(450, 520)
  )

  # Create the NDA template using the data frame directly
  to.nda(eefrt01)

  # Or using the name as a string
  to.nda("eefrt01")

  # Skip the confirmation prompt
  to.nda(eefrt01, skip_prompt = TRUE)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/createSpss.R
\name{to.sav}
\alias{to.sav}
\title{Create .sav SPSS file from a data frame}
\usage{
to.sav(df, df_name = NULL, path = ".", skip_prompt = TRUE)
}
\arguments{
\item{df}{Data frame to be exported to SPSS format.}

\item{df_name}{Optional; custom file name for the saved SPSS file. If not provided,
the name of the data frame variable will be used. The ".sav" extension will
be appended automatically.}

\item{path}{Character string specifying the directory path where the "tmp" folder
and SPSS file should be created. Defaults to the current working directory.}

\item{skip_prompt}{Logical. If TRUE (default), skips the confirmation prompt. If FALSE,
will prompt for confirmation unless the user has previously chosen to remember their preference.}
}
\value{
Invisible TRUE if successful. Writes an SPSS file to the designated path and prints a message indicating
the file's location.
}
\description{
This function takes a R data frame and writes it to an SPSS file
using the Haven package. The resulting file will be stored in the
"tmp" directory with a default name derived from the data frame variable name,
but can be customized if desired. The function will prompt for confirmation
before creating the file, with an option to remember the user's preference for future calls.
}
\examples{
\dontrun{
# Create a sample data frame
sample_df <- data.frame(
  id = 1:3,
  score = c(85, 92, 78),
  group = c("A", "B", "A")
)

# Basic usage with prompt
to.sav(sample_df)

# Custom filename
to.sav(sample_df, "participants_data")

# Skip the confirmation prompt
to.sav(sample_df, skip_prompt = TRUE)

# Save in a different directory
to.sav(sample_df, path = "path/to/project")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getSql.R
\name{sql.desc}
\alias{sql.desc}
\title{Get SQL table columns/metadata}
\usage{
sql.desc(table_name)
}
\arguments{
\item{table_name}{Name of the table to get metadata for}
}
\value{
A data frame with column information
}
\description{
Get SQL table columns/metadata
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getRedcap.R
\name{redcap.dict}
\alias{redcap.dict}
\title{Fetch REDCap data dictionary to be stored in data frame}
\usage{
redcap.dict(instrument_name)
}
\arguments{
\item{instrument_name}{Can either be an instrument name to fetch new data, a data frame
returned by redcap(), or a variable name as string}
}
\value{
A data frame containing the data dictionary/metadata for the specified instrument
}
\description{
This function extracts metadata/dictionary information from REDCap. It can accept
either an instrument name to fetch new data, an existing data frame with instrument
attributes, or a variable name as string.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getMongo.R
\name{mongo.index}
\alias{mongo.index}
\title{Display table of available MongoDB collections}
\usage{
mongo.index(database = NULL)
}
\arguments{
\item{database}{Optional; the name of the database to connect to. If NULL, uses the database
specified in the configuration file.}
}
\value{
A character vector containing the names of all available collections
in the configured MongoDB database.
}
\description{
Retrieves a list of all available collections in the configured MongoDB database.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getOracle.R
\name{oracle.test}
\alias{oracle.test}
\title{Test Oracle database connection}
\usage{
oracle.test()
}
\value{
A logical value indicating whether the connection was successful
}
\description{
Tests the connection to the Oracle database using the configured DSN and credentials.
This is a simple connectivity test that doesn't perform any data operations.
}
\examples{
\dontrun{
# Test the Oracle connection
if (oracle.test()) {
  message("Oracle connection successful!")
} else {
  message("Oracle connection failed!")
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getOracle.R
\name{oracle.query}
\alias{oracle.query}
\title{Perform a direct Oracle query with minimal processing}
\usage{
oracle.query(query, pii = FALSE, schema = NULL)
}
\arguments{
\item{query}{The SQL query to execute}

\item{pii}{Logical; if FALSE (default), remove fields marked as PII. TRUE keeps PII.}

\item{schema}{Optional schema name to qualify table names in the query}
}
\value{
A data frame with the query results
}
\description{
Perform a direct Oracle query with minimal processing
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/createSpss.R
\name{createSpss}
\alias{createSpss}
\title{Alias for 'to.sav' (DEPRECATED)}
\usage{
createSpss(...)
}
\arguments{
\item{...}{Additional arguments passed through to \code{to.sav()}.}
}
\value{
Invisible TRUE if successful. Writes an SPSS file to the designated path and prints a message indicating
the file's location.
}
\description{
This function is deprecated. Please use 'to.sav' instead.
This is a legacy alias for the 'to.sav' function to maintain compatibility with older code.
}
\examples{
\dontrun{
# DEPRECATED - use to.sav() instead
createSpss(prl01)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataMerge.R
\name{meld}
\alias{meld}
\title{Merge two or more data frames magically according to their candidate key}
\usage{
meld(
  ...,
  by = NULL,
  all = TRUE,
  no.dups = FALSE,
  csv = FALSE,
  rdata = FALSE,
  spss = FALSE
)
}
\arguments{
\item{...}{Clean data frames to be merged.}

\item{by}{A vector of strings specifying the column names to be used as merge keys. If NULL,
the function automatically determines common keys from the provided data frames.}

\item{all}{Logical; if TRUE, performs an OUTER JOIN. If FALSE, performs an INNER JOIN.}

\item{no.dups}{Logical; if TRUE, duplicates are removed post-merge.}

\item{csv}{Logical; if TRUE, the merged data frame is exported as a CSV file.}

\item{rdata}{Logical; if TRUE, the merged data frame is saved as an Rda file.}

\item{spss}{Logical; if TRUE, the merged data frame is exported as an SPSS file.}
}
\value{
A merged data frame based on the specified or common candidate keys.
}
\description{
This function simplifies the process of merging multiple cleaned data frames by automatically determining
common merge keys or utilizing user-specified keys. Supports both inner and outer join methods,
and offers options for exporting the merged data.
}
\examples{
\dontrun{
# Create sample dataframes for demonstration
df1 <- data.frame(
  src_subject_id = c("S001", "S002", "S003"),
  visit = c(1, 2, 1),
  measure1 = c(10, 15, 12),
  stringsAsFactors = FALSE
)

df2 <- data.frame(
  src_subject_id = c("S001", "S002", "S004"),
  visit = c(1, 2, 2),
  measure2 = c(85, 92, 78),
  stringsAsFactors = FALSE
)

# Perform an OUTER JOIN using default keys:
merged1 <- meld(df1, df2, all = TRUE)

# Perform an INNER JOIN using specified keys:
merged2 <- meld(df1, df2, by = "src_subject_id", all = FALSE)
}
}
\author{
Joshua Kenney \href{mailto:joshua.kenney@yale.edu}{joshua.kenney@yale.edu}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataFilter.R
\name{dataFilter}
\alias{dataFilter}
\title{Alias for 'sift' (DEPRECATED)}
\usage{
dataFilter(...)
}
\arguments{
\item{...}{Additional arguments passed through to \code{sift()}.}
}
\value{
A filtered dataframe based on the provided parameters,
and containing only the columns specified in 'cols'. If no columns
are specified, returns the entire dataframe with applied row filters.
}
\description{
This function is deprecated. Please use 'sift' instead.
This is a legacy alias for the 'sift' function to maintain compatibility with older code.
}
\examples{
\dontrun{
# DEPRECATED - use sift() instead
filtered <- dataFilter(df, sex="F")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataParse.R
\name{qualtrics.rune}
\alias{qualtrics.rune}
\title{Parse composite Qualtrics survey into component data frames by variable prefix}
\usage{
qualtrics.rune(
  qualtrics_alias,
  prefix = NULL,
  institution = NULL,
  label = FALSE,
  interview_date = NULL,
  complete = FALSE,
  lower = TRUE
)
}
\arguments{
\item{qualtrics_alias}{Character string specifying the Qualtrics survey alias to retrieve.}

\item{prefix}{Character string; default NULL, if specified returns only the dataframe with this prefix}

\item{institution}{Character string; default NULL, specify location}

\item{label}{Logical; default TRUE, returns coded values as labels instead of raw values.}

\item{interview_date}{Logical or Date String, returns all data before date}

\item{complete}{Logical; default FALSE, if TRUE only returns rows where Progress == 100}

\item{lower}{default TRUE convert prefixes to lower case}
}
\value{
Creates multiple dataframes in the global environment, one for each survey
detected in the data. Each dataframe is named after its survey prefix.
}
\description{
This function fetches a Qualtrics data frame containing multiple surveys and
separates it into individual data frames for each survey detected in the data.
It identifies the appropriate identifier column (e.g., participantId, workerId)
and splits the data based on column name prefixes.
}
\details{
The function performs the following steps:
\itemize{
\item Retrieves the raw Qualtrics data using the getSurvey() function
\item Identifies which identifier column to use (participantId, workerId, PROLIFIC_PID, or src_subject_id)
\item Determines survey prefixes by analyzing column names
\item Creates separate dataframes for each survey prefix found
\item Assigns each dataframe to the global environment with names matching the survey prefixes
}
}
\examples{
\dontrun{
# Parse a a Qualtrics survey into its component dataframes
qualtrics.rune("combined_surveys", label = FALSE)

# After running, access individual survey dataframes directly:
head(pss)  # Access the PSS survey dataframe
head(cesd) # Access the CESD survey dataframe

# Parse a single Qualtrics survey from composite survey
rgpts <- qualtrics.rune("combined_surveys", prefix = "rgpts")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getQualtrics.R
\name{getSurvey}
\alias{getSurvey}
\title{Alias for 'qualtrics' (DEPRECATED)}
\usage{
getSurvey(...)
}
\arguments{
\item{...}{Optional column names to filter for. Only rows with non-missing values
in ALL specified columns will be returned. This is useful for filtering
data to only include complete cases for specific variables of interest.}
}
\value{
A cleaned and harmonized data frame containing the survey data with superkeys first.
}
\description{
This function is deprecated. Please use 'qualtrics' instead.
This is a legacy alias for the 'qualtrics' function to maintain compatibility with older code.
}
\examples{
\dontrun{
# DEPRECATED - use qualtrics() instead
survey_data <- getSurvey("your_survey_alias")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataMerge.R
\name{dataMerge}
\alias{dataMerge}
\title{Alias for 'meld' (DEPRECATED)}
\usage{
dataMerge(...)
}
\arguments{
\item{...}{Clean data frames to be merged.}
}
\value{
A merged data frame based on the specified or common candidate keys.
}
\description{
This function is deprecated. Please use 'meld' instead.
This is a legacy alias for the 'meld' function to maintain compatibility with older code.
}
\examples{
\dontrun{
# DEPRECATED - use meld() instead
merged <- dataMerge(df1_clean, df2_clean)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataRequest.R
\name{dataRequest}
\alias{dataRequest}
\title{Alias for 'clean' (DEPRECATED)}
\usage{
dataRequest(...)
}
\arguments{
\item{...}{Strings, specifying the measures to process, which can be a Mongo collection, REDCap instrument, or Qualtrics survey.}
}
\value{
Prints the time taken for the data request process.
}
\description{
This function is deprecated. Please use 'clean' instead.
This is a legacy alias for the 'clean' function to maintain compatibility with older code.
}
\examples{
\dontrun{
# DEPRECATED - use clean() instead
prl <- dataRequest("prl")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataParse.R
\name{redcap.rune}
\alias{redcap.rune}
\title{Parse composite REDCap instrument into component data frames by variable prefix}
\usage{
redcap.rune(
  instrument_name,
  prefix = NULL,
  raw_or_label = "raw",
  redcap_event_name = NULL,
  batch_size = 1000,
  records = NULL,
  fields = NULL,
  pii = FALSE,
  interview_date = NULL,
  date_format = "ymd",
  lower = TRUE
)
}
\arguments{
\item{instrument_name}{Name of the REDCap instrument}

\item{prefix}{Character string; default NULL, if specified returns only the dataframe with this prefix}

\item{raw_or_label}{Whether to return raw or labeled values}

\item{redcap_event_name}{Optional event name filter. Can be a single string
or a vector of event names (e.g., \code{c("event1", "event2")})}

\item{batch_size}{Number of records to retrieve per batch}

\item{records}{Optional vector of specific record IDs}

\item{fields}{Optional vector of specific fields}

\item{pii}{Logical; if FALSE (default), remove fields marked as PII. TRUE keeps PII.}

\item{interview_date}{Optional; date filtering parameter}

\item{date_format}{Default ymd define date format for interview_date}

\item{lower}{default TRUE convert prefixes to lower case}
}
\value{
If prefix is specified, returns a single dataframe with that prefix. Otherwise,
creates multiple dataframes in the parent environment, one for each survey
detected in the data. Each dataframe is named after its survey prefix.
}
\description{
This function fetches a REDCap instrument and separates it into individual data frames
for each survey/collection detected in the data based on column name prefixes.
It identifies the appropriate identifier column and splits the data accordingly.
}
\examples{
\dontrun{
# Parse a REDCap instrument into its component dataframes
redcap.rune("baseline_assessment")

# After running, access individual survey dataframes directly:
head(pss)  # Access the PSS survey dataframe
head(cesd) # Access the CESD survey dataframe

# Parse a single survey from composite instrument
rgpts <- redcap.rune("baseline_assessment", prefix = "rgpts")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/createCsv.R
\name{to.csv}
\alias{to.csv}
\title{Create .csv file from a data frame}
\usage{
to.csv(df, df_name = NULL, path = ".", skip_prompt = TRUE)
}
\arguments{
\item{df}{Data frame to be exported to CSV format.}

\item{df_name}{Optional; a custom file name for the saved CSV file.
If not provided, the name of the data frame variable is used.
The function adds the ".csv" extension automatically.}

\item{path}{Character string specifying the directory path where the "tmp" folder
and CSV file should be created. Defaults to the current working directory.}

\item{skip_prompt}{Logical. If TRUE (default), skips the confirmation prompt. If FALSE,
will prompt for confirmation unless the user has previously chosen to remember their preference.}
}
\value{
Invisible TRUE if successful. The function writes a CSV file to the specified path
and prints a message indicating the file's location.
}
\description{
This function exports a given R data frame to a CSV file format.
The resulting file is saved in the "tmp" directory. If a filename
is not specified, the function uses the name of the data frame variable.
The ".csv" extension is appended automatically to the filename.
The function will prompt for confirmation before creating the file,
with an option to remember the user's preference for future calls.
}
\examples{
\dontrun{
# Create a sample data frame
sample_df <- data.frame(
  id = 1:3,
  name = c("Alice", "Bob", "Charlie")
)

# Basic usage with prompt
to.csv(sample_df)

# Custom filename
to.csv(sample_df, "participants_data")

# Skip the confirmation prompt
to.csv(sample_df, skip_prompt = TRUE)

# Save in a different directory
to.csv(sample_df, path = "path/to/project")
}

}
\author{
Joshua Kenney \href{mailto:joshua.kenney@yale.edu}{joshua.kenney@yale.edu}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getQualtrics.R
\name{qualtrics.dict}
\alias{qualtrics.dict}
\title{Fetch Qualtrics survey metadata to be stored in data frame}
\usage{
qualtrics.dict(survey_alias, exclude_embedded = TRUE)
}
\arguments{
\item{survey_alias}{Can either be an existing dataframe, variable name as string, or survey alias string}

\item{exclude_embedded}{Only select QIDs}
}
\value{
A list containing the mappings of column names to survey questions.
}
\description{
This function extracts column mappings from the metadata of a Qualtrics survey data frame.
It can accept either a data frame containing Qualtrics data, a variable name as string,
or a survey alias string.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getQualtrics.R
\name{qualtrics.index}
\alias{qualtrics.index}
\title{Display table of available Qualtrics surveys}
\usage{
qualtrics.index(institution = NULL, all = FALSE)
}
\arguments{
\item{institution}{Optional; the institution identifier to use. If NULL, uses all
institutions specified in the configuration file (or all available credentials if no config).}

\item{all}{Logical; deprecated parameter kept for backward compatibility. All surveys
are now shown by default. Default is FALSE.}
}
\value{
A data frame containing the IDs and names of all available surveys
from the Qualtrics API. Surveys with aliases configured in config.yml will
show the alias and institution; unmapped surveys will show NA for these fields.
}
\description{
Retrieves a list of all available surveys from the Qualtrics API. Shows all
surveys pulled down from Qualtrics, with alias and institution information
merged from config.yml where available.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getMongo.R
\name{mongo}
\alias{mongo}
\title{Fetch data from MongoDB to be stored in a data frame - UPDATED VERSION}
\usage{
mongo(
  collection,
  ...,
  database = NULL,
  identifier = NULL,
  chunk_size = NULL,
  verbose = FALSE,
  interview_date = NULL
)
}
\arguments{
\item{collection}{The name of the MongoDB collection}

\item{...}{Optional column names to filter for. Only rows with non-missing values
in ALL specified columns will be returned. This is useful for filtering
data to only include complete cases for specific variables of interest.}

\item{database}{The database name (optional)}

\item{identifier}{Field to use as identifier (optional)}

\item{chunk_size}{Number of records per chunk (optional)}

\item{verbose}{Logical; if TRUE, displays detailed progress messages. Default is FALSE.}

\item{interview_date}{Optional; can be either:
- A date string in various formats (ISO, US, etc.) to filter data up to that date
- A boolean TRUE to return only rows with non-NA interview_date values}
}
\value{
A data frame containing the MongoDB data with superkeys first
}
\description{
Fetch data from MongoDB to be stored in a data frame - UPDATED VERSION
}
\examples{
\dontrun{
# Get data from MongoDB collection
data <- mongo("collection")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataParse.R
\name{mongo.rune}
\alias{mongo.rune}
\title{Parse composite MongoDB collection into component data frames by variable prefix}
\usage{
mongo.rune(collection, prefix = NULL, db_name = NULL, lower = TRUE)
}
\arguments{
\item{collection}{Character string specifying the Mongo collection}

\item{prefix}{Character string; default NULL, if specified returns only the dataframe with this prefix}

\item{db_name}{Character string specifying the Mongo database}

\item{lower}{default TRUE convert prefixes to lower case}
}
\value{
If prefix is specified, returns a single dataframe with that prefix. Otherwise,
creates multiple dataframes in the global environment, one for each survey
detected in the data. Each dataframe is named after its survey prefix.
}
\description{
This function fetches a MongoDB collection containing multiple collections and
separates it into individual data frames for each collection detected in the data.
It identifies the appropriate identifier column (e.g., participantId, workerId)
and splits the data based on column name prefixes.
}
\details{
The function performs the following steps:
\itemize{
\item Retrieves the raw Qualtrics data using the getSurvey() function
\item Identifies which identifier column to use (participantId, workerId, PROLIFIC_PID, or src_subject_id)
\item Determines survey prefixes by analyzing column names
\item Creates separate dataframes for each survey prefix found
\item Assigns each dataframe to the global environment with names matching the survey prefixes
}
}
\examples{
\dontrun{
# Parse a MongoDB collection into its component dataframes
mongo.rune("combined_surveys")

# After running, access individual survey dataframes directly:
head(pss)  # Access the PSS survey dataframe
head(cesd) # Access the CESD survey dataframe

# Parse a single survey from composite collection
rgpts <- mongo.rune("combined_surveys", prefix = "rgpts")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataParse.R
\name{rune}
\alias{rune}
\title{Parse composite data frame into component data frames by variable prefix}
\usage{
rune(df, prefix = NULL, lower = TRUE)
}
\arguments{
\item{df}{a dataframe containing multiple, prefixed measures}

\item{prefix}{Character string; default NULL, if specified returns only the dataframe with this prefix}

\item{lower}{default TRUE convert prefixes to lower case}
}
\value{
If prefix is specified, returns a single dataframe with that prefix. Otherwise,
creates multiple dataframes in the global environment, one for each survey
detected in the data. Each dataframe is named after its survey prefix.
}
\description{
This function takes a data frame containing multiple measures and
separates it into individual data frames for each measure detected in the data.
It identifies the appropriate identifier column (e.g., participantId, workerId)
and splits the data based on column name prefixes.
}
\details{
The function performs the following steps:
\itemize{
\item Identifies which identifier column to use (participantId, workerId, PROLIFIC_PID, or src_subject_id)
\item Determines survey prefixes by analyzing column names
\item Creates separate dataframes for each survey prefix found
\item Assigns each dataframe to the global environment with names matching the survey prefixes
}
}
\examples{
# Parse a data frame containing multiple surveys
combined_df <- data.frame(
  record_id = c("REC001", "REC002", "REC003", "REC004"),
  src_subject_id = c("SUB001", "SUB002", "SUB003", "SUB004"),
  subjectkey = c("KEY001", "KEY002", "KEY003", "KEY004"),
  site = c("Yale", "NU", "Yale", "NU"),
  phenotype = c("A", "B", "A", "C"),
  visit = c(1, 2, 2, 1),
  state = c("complete", "completed baseline", "in progress", NA),
  status = c(NA, NA, NA, "complete"),
  lost_to_followup = c(FALSE, FALSE, TRUE, NA),
  interview_date = c("2023-01-15", "2023/02/20", NA, "2023-03-10"),
  foo_1 = c(1, 3, 5, 7),
  foo_2 = c("a", "b", "c", "d"),
  bar_1 = c(2, 4, 6, 8),
  bar_2 = c("w", "x", "y", "z")
)
rune(combined_df)

# After running, access individual survey dataframes directly:
head(foo)  # Access the foo dataframe
head(bar)  # Access the bar dataframe

# Parse a single survey from composite dataframe
foo_df <- rune(combined_df, prefix = "foo")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/createRda.R
\name{createRds}
\alias{createRds}
\title{Alias for 'to.rds' (DEPRECATED)}
\usage{
createRds(...)
}
\arguments{
\item{...}{Additional arguments passed through to \code{to.rds()}.}
}
\value{
Invisible TRUE if successful. The function writes an RDS file to the specified path
and prints a message indicating the file's location.
}
\description{
This function is deprecated. Please use 'to.rds' instead.
This is a legacy alias for the 'to.rds' function to maintain compatibility with older code.
}
\examples{
\dontrun{
# DEPRECATED - use to.rds() instead
createRds(prl01)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scry.R
\name{display_tree}
\alias{display_tree}
\title{Display a file tree structure similar to the Unix tree command}
\usage{
display_tree(path)
}
\arguments{
\item{path}{The path to display as a tree}
}
\value{
NULL (called for side effects)
}
\description{
Display a file tree structure similar to the Unix tree command
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataWizardry-package.R
\docType{package}
\name{wizaRdry-package}
\alias{wizaRdry}
\alias{wizaRdry-package}
\title{wizaRdry: A Magical Framework for Collaborative & Reproducible Data Analysis}
\description{
A comprehensive data analysis framework for NIH-funded research that streamlines workflows for both data cleaning and preparing NIH Data Archive ('NDA') submission templates. Provides unified access to multiple data sources ('REDCap', 'MongoDB', 'Qualtrics', 'SQL', 'ORACLE') through interfaces to their APIs, with specialized functions for data cleaning, filtering, merging, and parsing. Features automatic validation, field harmonization, and memory-aware processing to enhance reproducibility in multi-site collaborative research as described in Mittal et al. (2021) \doi{10.20900/jpbs.20210011}.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/belieflab/wizaRdry}
  \item Report bugs at \url{https://github.com/belieflab/wizaRdry/issues}
}

}
\author{
\strong{Maintainer}: Joshua G. Kenney \email{joshua.kenney@yale.edu}

Authors:
\itemize{
  \item Trevor F. Williams \email{trevormsu@gmail.com}
  \item Minerva K. Pappu \email{minerva.pappu@yale.edu}
  \item Michael J. Spilka \email{michael.j.spilka@gmail.com}
}

Other contributors:
\itemize{
  \item Danielle N. Pratt \email{danielle.pratt33@gmail.com} [contributor]
  \item Victor J. Pokorny \email{victor.pokorny@northwestern.edu} [contributor]
  \item Santiago Castiello de Obeso \email{santiago.castiellodeobeso@yale.edu} [contributor]
  \item Praveen Suthaharan \email{praveen.suthaharan@yale.edu} [contributor]
  \item Christian R. Horgan \email{christian.horgan@yale.edu} [contributor]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getSql.R
\name{sql.index}
\alias{sql.index}
\title{Get a list of tables from the SQL database}
\usage{
sql.index(schema = NULL)
}
\arguments{
\item{schema}{Optional schema name to filter tables}
}
\value{
A data frame with table information
}
\description{
Get a list of tables from the SQL database
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ndaRequest.R
\name{nda}
\alias{nda}
\title{Generate validated NDA submission templates created in the ./nda directory}
\usage{
nda(
  ...,
  csv = FALSE,
  rdata = FALSE,
  spss = FALSE,
  limited_dataset = FALSE,
  skip_prompt = TRUE,
  verbose = FALSE,
  strict = TRUE,
  dcc = FALSE
)
}
\arguments{
\item{...}{Strings, specifying the measures to process, which can be a Mongo collection, REDCap instrument, or Qualtrics survey.}

\item{csv}{Optional; Boolean, if TRUE creates a .csv extract in ./tmp.}

\item{rdata}{Optional; Boolean, if TRUE creates an .rdata extract in ./tmp.}

\item{spss}{Optional; Boolean, if TRUE creates a .sav extract in ./tmp.}

\item{limited_dataset}{Optional; Boolean, if TRUE does not perform date-shifting of interview_date or age-capping of interview_age}

\item{skip_prompt}{Logical. If TRUE (default), skips confirmation prompts unless preferences aren't set yet. If FALSE,
prompts for confirmation unless the user has previously chosen to remember their preference.}

\item{verbose}{Logical. If TRUE, shows detailed processing information. If FALSE (default), shows only essential user-facing messages.}

\item{strict}{Logical. If TRUE (default), enforce strict NDA validation: required fields with ANY missing data or
recommended fields with ALL missing data will cause validation failure. If FALSE (lenient mode), missing data
triggers warnings but allows processing to continue.}

\item{dcc}{Logical. If TRUE, include 11 DCC (Data Coordinating Center) fields from ndar_subject01
(7 required + 4 recommended). Default FALSE.}
}
\value{
Prints the time taken for the data request process.
}
\description{
This function processes requests for clean data sequentially for specified measures.
It makes a request to the NIH NDA API for the named data structures
and runs the associated data remediation routines. It then runs a series of
unit tests to verify that the data quality standards are met.
}
\examples{
\dontrun{
  nda("prl", csv=TRUE)
  nda("rgpts", "kamin", rdata=TRUE)

  # Skip confirmation prompts
  nda("prl", csv=TRUE, skip_prompt=TRUE)
  
  # Show detailed processing information
  nda("prl", verbose=TRUE)
  
  # Use lenient validation mode (allow missing data with warnings)
  nda("prl", strict=FALSE)
  
  # Include DCC fields from ndar_subject01
  nda("prl", dcc=TRUE)
}

}
\author{
Joshua Kenney \href{mailto:joshua.kenney@yale.edu}{joshua.kenney@yale.edu}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getRedcap.R
\name{redcap}
\alias{redcap}
\title{Fetch data from REDCap to be stored in a data frame}
\usage{
redcap(
  instrument_name = NULL,
  ...,
  raw_or_label = "raw",
  redcap_event_name = NULL,
  batch_size = 1000,
  records = NULL,
  fields = NULL,
  pii = FALSE,
  interview_date = NULL,
  date_format = "ymd",
  complete = NULL
)
}
\arguments{
\item{instrument_name}{Name of the REDCap instrument}

\item{...}{Optional column names to filter for. Only rows with non-missing values
in ALL specified columns will be returned. This is useful for filtering
data to only include complete cases for specific variables of interest.}

\item{raw_or_label}{Whether to return raw or labeled values}

\item{redcap_event_name}{Optional event name filter. Can be a single string
or a vector of event names (e.g., \code{c("event1", "event2")})}

\item{batch_size}{Number of records to retrieve per batch}

\item{records}{Optional vector of specific record IDs}

\item{fields}{Optional vector of specific fields}

\item{pii}{Logical; if FALSE (default), remove fields marked as PII. TRUE keeps PII.}

\item{interview_date}{Optional; can be either:
- A date string in various formats (ISO, US, etc.) to filter data up to that date
- A boolean TRUE to return only rows with non-NA interview_date values}

\item{date_format}{Default ymd define date format for interview_date}

\item{complete}{Option boolean TRUE will return only forms marked as complete in REDCap}
}
\value{
A data frame containing the requested REDCap data
}
\description{
Retrieves data from a REDCap instrument and ensures subject identifiers
are propagated across all events
}
\examples{
\dontrun{
# Get data from a specific instrument
data <- redcap("demographics")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/NdaDataStructure.R
\name{NdaDataStructure}
\alias{NdaDataStructure}
\title{NdaDataStructure R6 Class}
\description{
Represents a single field (data element) in an NDA data structure.
This is a typed struct (similar to Go structs) that enforces schema consistency
and provides validation for NDA field definitions.
}
\details{
This class replaces ad-hoc list construction for NDA field definitions.
It provides:
\itemize{
\item Type safety and validation for field definitions
\item Consistent structure across all code paths
\item Factory methods for creating fields from different sources
\item Helper methods for common operations
\item Direct mapping to Excel export columns
}

The field structure matches the NDA data dictionary schema:
ElementName, DataType, Size, Required, ElementDescription, ValueRange, Notes, Aliases

Uses typed R6 classes from NdaClasses.R for all fields (ElementName, DataType,
Size, RequirementLevel, Description, ValueRange, Notes, Aliases, etc.)
}
\keyword{internal}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{element_name}}{ElementName object - field name (ElementName in Excel)}

\item{\code{data_type}}{DataType object - data type (String, Integer, Float, Date, GUID, Boolean)}

\item{\code{size}}{Size object - size for String types}

\item{\code{required}}{RequirementLevel object - requirement level (Required, Recommended, Conditional, No)}

\item{\code{element_description}}{Description object - field description}

\item{\code{value_range}}{ValueRange object - allowed values or range}

\item{\code{notes}}{Notes object - field notes}

\item{\code{aliases}}{Aliases object - field aliases}

\item{\code{selection_order}}{Integer - order in which field was selected}

\item{\code{selected_for_submission}}{Logical - whether field is selected for NDA submission}

\item{\code{source_metadata}}{SourceMetadata object - field source tracking}

\item{\code{missing_info}}{MissingInfo object - missing data information}

\item{\code{validation_rules}}{ValidationRules object - validation rules}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-NdaDataStructure-new}{\code{NdaDataStructure$new()}}
\item \href{#method-NdaDataStructure-to_excel_row}{\code{NdaDataStructure$to_excel_row()}}
\item \href{#method-NdaDataStructure-to_list}{\code{NdaDataStructure$to_list()}}
\item \href{#method-NdaDataStructure-is_super_required}{\code{NdaDataStructure$is_super_required()}}
\item \href{#method-NdaDataStructure-is_from_ndar_subject}{\code{NdaDataStructure$is_from_ndar_subject()}}
\item \href{#method-NdaDataStructure-is_dcc_required}{\code{NdaDataStructure$is_dcc_required()}}
\item \href{#method-NdaDataStructure-is_dcc_recommended}{\code{NdaDataStructure$is_dcc_recommended()}}
\item \href{#method-NdaDataStructure-modify}{\code{NdaDataStructure$modify()}}
\item \href{#method-NdaDataStructure-print}{\code{NdaDataStructure$print()}}
\item \href{#method-NdaDataStructure-merge_value_ranges}{\code{NdaDataStructure$merge_value_ranges()}}
\item \href{#method-NdaDataStructure-get_merge_warnings}{\code{NdaDataStructure$get_merge_warnings()}}
\item \href{#method-NdaDataStructure-clear_merge_warnings}{\code{NdaDataStructure$clear_merge_warnings()}}
\item \href{#method-NdaDataStructure-clone}{\code{NdaDataStructure$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-NdaDataStructure-new"></a>}}
\if{latex}{\out{\hypertarget{method-NdaDataStructure-new}{}}}
\subsection{Method \code{new()}}{
Create a new NdaDataStructure instance
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{NdaDataStructure$new(
  element_name,
  data_type = "String",
  size = NULL,
  required = "No",
  element_description = "",
  value_range = "",
  notes = "",
  aliases = "",
  selection_order = NULL,
  source = NULL,
  source_metadata = NULL,
  missing_info = NULL,
  validation_rules = NULL,
  ...
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{element_name}}{Field name (required) - accepts string or ElementName object}

\item{\code{data_type}}{Data type (default: "String") - accepts string or DataType object}

\item{\code{size}}{Size for String types - accepts numeric or Size object}

\item{\code{required}}{Requirement level (default: "No") - accepts string or RequirementLevel object}

\item{\code{element_description}}{Field description - accepts string or Description object}

\item{\code{value_range}}{Allowed values or range - accepts string or ValueRange object}

\item{\code{notes}}{Field notes - accepts string or Notes object}

\item{\code{aliases}}{Field aliases - accepts string, list, or Aliases object}

\item{\code{selection_order}}{Selection order}

\item{\code{source}}{Field source (legacy - use source_metadata instead)}

\item{\code{source_metadata}}{SourceMetadata object}

\item{\code{missing_info}}{MissingInfo object or list}

\item{\code{validation_rules}}{ValidationRules object or list}

\item{\code{...}}{Additional fields}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A new NdaDataStructure object
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-NdaDataStructure-to_excel_row"></a>}}
\if{latex}{\out{\hypertarget{method-NdaDataStructure-to_excel_row}{}}}
\subsection{Method \code{to_excel_row()}}{
Convert to Excel row (returns named list for data.frame row)
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{NdaDataStructure$to_excel_row()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Named list with Excel column names and values
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-NdaDataStructure-to_list"></a>}}
\if{latex}{\out{\hypertarget{method-NdaDataStructure-to_list}{}}}
\subsection{Method \code{to_list()}}{
Convert to legacy list format for backward compatibility
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{NdaDataStructure$to_list()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
List with field definition
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-NdaDataStructure-is_super_required"></a>}}
\if{latex}{\out{\hypertarget{method-NdaDataStructure-is_super_required}{}}}
\subsection{Method \code{is_super_required()}}{
Check if field is a super required field
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{NdaDataStructure$is_super_required()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Logical
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-NdaDataStructure-is_from_ndar_subject"></a>}}
\if{latex}{\out{\hypertarget{method-NdaDataStructure-is_from_ndar_subject}{}}}
\subsection{Method \code{is_from_ndar_subject()}}{
Check if field came from ndar_subject01
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{NdaDataStructure$is_from_ndar_subject()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Logical
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-NdaDataStructure-is_dcc_required"></a>}}
\if{latex}{\out{\hypertarget{method-NdaDataStructure-is_dcc_required}{}}}
\subsection{Method \code{is_dcc_required()}}{
Check if field is a DCC required field
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{NdaDataStructure$is_dcc_required()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Logical
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-NdaDataStructure-is_dcc_recommended"></a>}}
\if{latex}{\out{\hypertarget{method-NdaDataStructure-is_dcc_recommended}{}}}
\subsection{Method \code{is_dcc_recommended()}}{
Check if field is a DCC recommended field
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{NdaDataStructure$is_dcc_recommended()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Logical
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-NdaDataStructure-modify"></a>}}
\if{latex}{\out{\hypertarget{method-NdaDataStructure-modify}{}}}
\subsection{Method \code{modify()}}{
Create a modified copy of this field
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{NdaDataStructure$modify(
  value_range = NULL,
  notes = NULL,
  modification_note = NULL,
  ...
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{value_range}}{New value range (string or ValueRange object)}

\item{\code{notes}}{New notes (string or Notes object)}

\item{\code{modification_note}}{Description of modification}

\item{\code{...}}{Other fields to modify}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
New NdaDataStructure object
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-NdaDataStructure-print"></a>}}
\if{latex}{\out{\hypertarget{method-NdaDataStructure-print}{}}}
\subsection{Method \code{print()}}{
Print method for NdaDataStructure
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{NdaDataStructure$print()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Self (invisibly)
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-NdaDataStructure-merge_value_ranges"></a>}}
\if{latex}{\out{\hypertarget{method-NdaDataStructure-merge_value_ranges}{}}}
\subsection{Method \code{merge_value_ranges()}}{
Merge value ranges from multiple sources
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{NdaDataStructure$merge_value_ranges(
  nda_range = NULL,
  redcap_range = NULL,
  data_range = NULL,
  missing_codes = character(0)
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{nda_range}}{ValueRange from NDA (may be NULL)}

\item{\code{redcap_range}}{ValueRange from REDCap (may be NULL)}

\item{\code{data_range}}{ValueRange from data (may be NULL)}

\item{\code{missing_codes}}{Character vector of missing data codes}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Self (invisibly) with merged value_range
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-NdaDataStructure-get_merge_warnings"></a>}}
\if{latex}{\out{\hypertarget{method-NdaDataStructure-get_merge_warnings}{}}}
\subsection{Method \code{get_merge_warnings()}}{
Get any warnings from merge operations
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{NdaDataStructure$get_merge_warnings()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Character vector of warnings
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-NdaDataStructure-clear_merge_warnings"></a>}}
\if{latex}{\out{\hypertarget{method-NdaDataStructure-clear_merge_warnings}{}}}
\subsection{Method \code{clear_merge_warnings()}}{
Clear merge warnings
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{NdaDataStructure$clear_merge_warnings()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Self (invisibly)
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-NdaDataStructure-clone"></a>}}
\if{latex}{\out{\hypertarget{method-NdaDataStructure-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{NdaDataStructure$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DataEnvironment.R
\name{DataEnvironment}
\alias{DataEnvironment}
\title{DataEnvironment R6 Class}
\description{
Manages dataframe storage in package environment (.pkg_env$.wizaRdry_env) with optional
convenience assignment to calling environment. CRAN-compliant environment management.
}
\details{
This class provides a clean interface for getting and setting dataframes using the package
environment (.pkg_env) as the authoritative source, with optional assignment to the calling
environment for user convenience. This eliminates global environment pollution and follows
R package best practices.
}
\keyword{internal}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{measure_name}}{Character string - name of the measure/dataframe}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-DataEnvironment-new}{\code{DataEnvironment$new()}}
\item \href{#method-DataEnvironment-get_df}{\code{DataEnvironment$get_df()}}
\item \href{#method-DataEnvironment-set_df}{\code{DataEnvironment$set_df()}}
\item \href{#method-DataEnvironment-get_colnames}{\code{DataEnvironment$get_colnames()}}
\item \href{#method-DataEnvironment-nrow}{\code{DataEnvironment$nrow()}}
\item \href{#method-DataEnvironment-ncol}{\code{DataEnvironment$ncol()}}
\item \href{#method-DataEnvironment-print}{\code{DataEnvironment$print()}}
\item \href{#method-DataEnvironment-clone}{\code{DataEnvironment$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DataEnvironment-new"></a>}}
\if{latex}{\out{\hypertarget{method-DataEnvironment-new}{}}}
\subsection{Method \code{new()}}{
Create a new DataEnvironment instance
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{DataEnvironment$new(measure_name, df)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{measure_name}}{Name of the measure/dataframe}

\item{\code{df}}{Initial dataframe to store}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A new DataEnvironment object
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DataEnvironment-get_df"></a>}}
\if{latex}{\out{\hypertarget{method-DataEnvironment-get_df}{}}}
\subsection{Method \code{get_df()}}{
Get dataframe from package environment or calling environment
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{DataEnvironment$get_df()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
The dataframe stored in package environment
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DataEnvironment-set_df"></a>}}
\if{latex}{\out{\hypertarget{method-DataEnvironment-set_df}{}}}
\subsection{Method \code{set_df()}}{
Set dataframe in package environment with optional calling environment assignment
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{DataEnvironment$set_df(df)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{df}}{Data frame to set}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
Self (invisibly) for method chaining
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DataEnvironment-get_colnames"></a>}}
\if{latex}{\out{\hypertarget{method-DataEnvironment-get_colnames}{}}}
\subsection{Method \code{get_colnames()}}{
Get column names from the dataframe
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{DataEnvironment$get_colnames()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Character vector of column names
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DataEnvironment-nrow"></a>}}
\if{latex}{\out{\hypertarget{method-DataEnvironment-nrow}{}}}
\subsection{Method \code{nrow()}}{
Get number of rows in the dataframe
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{DataEnvironment$nrow()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Integer number of rows
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DataEnvironment-ncol"></a>}}
\if{latex}{\out{\hypertarget{method-DataEnvironment-ncol}{}}}
\subsection{Method \code{ncol()}}{
Get number of columns in the dataframe
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{DataEnvironment$ncol()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Integer number of columns
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DataEnvironment-print"></a>}}
\if{latex}{\out{\hypertarget{method-DataEnvironment-print}{}}}
\subsection{Method \code{print()}}{
Print method for DataEnvironment
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{DataEnvironment$print()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
Self (invisibly)
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DataEnvironment-clone"></a>}}
\if{latex}{\out{\hypertarget{method-DataEnvironment-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{DataEnvironment$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scry.R
\name{scry}
\alias{scry}
\title{Initialize the wizaRdry directory structure inside an R project}
\usage{
scry(
  study_alias = NULL,
  path = ".",
  overwrite = FALSE,
  repair = FALSE,
  show_tree = NULL,
  create_project = FALSE,
  examples = FALSE,
  skip_prompt = TRUE
)
}
\arguments{
\item{study_alias}{Character string specifying the short name for the study e.g. impact, capr, sing}

\item{path}{Character string specifying the directory path where the structure should be created.
Defaults to the current working directory.}

\item{overwrite}{Logical. If TRUE, will overwrite existing files. If FALSE (default),
will not replace existing files.}

\item{repair}{Logical. If TRUE, will attempt to repair an incomplete directory structure.
If FALSE, will abort with an error message when encountering an incomplete structure.}

\item{show_tree}{Logical. If TRUE (default on first run), will display a visual file tree.
Set to FALSE to suppress the tree view.}

\item{create_project}{Logical. If TRUE, will create an R project file if one doesn't exist.
If FALSE (default), will not create an R project.}

\item{examples}{Logical. If TRUE (default when not repairing), will create example script templates.
If FALSE (default when repairing), will skip creating example scripts.}

\item{skip_prompt}{Logical. If TRUE (default), will skip the initial confirmation prompt if y/n preferences are not set yet. FALSE if specified.}
}
\value{
Invisible TRUE if successful.
}
\description{
Creates the standard directory structure required for the wizaRdry package to function properly.
This includes folders for data cleaning scripts, NDA submission templates, and temporary outputs.
It can detect and repair incomplete directory structures, and optionally create an R project.
}
\details{
The function creates the following directory structure:
\itemize{
\item clean/
\itemize{
\item csv/
\item mongo/
\item qualtrics/
\item redcap/
\item oracle/
\item sql/
}
\item nda/
\itemize{
\item csv/
\item mongo/
\item qualtrics/
\item redcap/
\item oracle/
\item sql/
}
\item tmp/
}

It also creates template config.yml and secrets.R files, and optionally an R project file.
}
\examples{
\dontrun{
# Initialize in current directory
scry()

# Repair structure in current directory
scry(repair = TRUE)

# Initialize in a specific directory with an R project
scry("path/to/project", create_project = TRUE, repair = TRUE)

# Skip the tree display
scry(repair = TRUE, show_tree = FALSE)

# Explicitly create example scripts when repairing
scry(repair = TRUE, examples = TRUE)

# Skip the confirmation prompt
scry(skip_prompt = TRUE)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getOracle.R
\name{oracle.desc}
\alias{oracle.desc}
\title{Get Oracle table columns/metadata}
\usage{
oracle.desc(table_name, schema = NULL)
}
\arguments{
\item{table_name}{Name of the table to get metadata for}

\item{schema}{Optional schema name}
}
\value{
A data frame with column information
}
\description{
Get Oracle table columns/metadata
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getRedcap.R
\name{redcap.index}
\alias{redcap.index}
\title{Display table of available REDCap instruments and their labels}
\usage{
redcap.index()
}
\value{
A formatted table (kable) of available REDCap instruments/forms
}
\description{
Retrieves a list of all available REDCap forms as a formatted table
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataRequest.R
\name{clean}
\alias{clean}
\title{Generate clean data frames from cleaning scripts created in the ./clean directory}
\usage{
clean(..., csv = FALSE, rdata = FALSE, spss = FALSE, skip_prompt = TRUE)
}
\arguments{
\item{...}{Strings, specifying the measures to process, which can be a Mongo collection, REDCap instrument, or Qualtrics survey.}

\item{csv}{Optional; Boolean, if TRUE creates a .csv extract in ./tmp.}

\item{rdata}{Optional; Boolean, if TRUE creates an .rdata extract in ./tmp.}

\item{spss}{Optional; Boolean, if TRUE creates a .sav extract in ./tmp.}

\item{skip_prompt}{Logical. If TRUE (default), skips confirmation prompts. If FALSE,
prompts for confirmation unless the user has previously chosen to remember their preference.}
}
\value{
Prints the time taken for the data request process.
}
\description{
This function processes requests for clean data sequentially for specified measures.
It makes a request to the appropriate API for the named measure or measures
and runs the associated data cleaning routines. It then runs a series of
unit tests to verify that the data quality standards are met.
}
\examples{
\dontrun{
  clean("prl", csv=TRUE)
  clean("rgpts", "kamin", rdata=TRUE)

  # Skip confirmation prompts
  clean("prl", csv=TRUE, skip_prompt=TRUE)
}

}
\author{
Joshua Kenney \href{mailto:joshua.kenney@yale.edu}{joshua.kenney@yale.edu}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dataFilter.R
\name{sift}
\alias{sift}
\title{Filter data frame by superkey parameters, rows, and columns}
\usage{
sift(
  df,
  rows = NULL,
  cols = NULL,
  record_id = NULL,
  src_subject_id = NULL,
  subjectkey = NULL,
  site = NULL,
  subsiteid = NULL,
  sex = NULL,
  race = NULL,
  ethnic_group = NULL,
  phenotype = NULL,
  phenotype_description = NULL,
  status = NULL,
  lost_to_followup = NULL,
  twins_study = NULL,
  sibling_study = NULL,
  family_study = NULL,
  sample_taken = NULL,
  visit = NULL,
  week = NULL,
  arm = NULL,
  interview_date = NULL
)
}
\arguments{
\item{df}{Dataframe to be filtered and trimmed based on the provided parameters.}

\item{rows}{Optional; either a single row name or a vector of row names to be retained in the final output.
If NULL or empty, all rows in the dataframe are retained.}

\item{cols}{Optional; either a single column name or a vector of column names to be retained in the final output.
If NULL or empty, all columns in the dataframe are retained.#' Data Filter}

\item{record_id}{Optional; either a single record_id or a vector of record_ids to filter the dataframe by}

\item{src_subject_id}{Optional; either a single subject ID or a vector of subject IDs to filter the dataframe by}

\item{subjectkey}{Optional; either a single subjectkey or a vector of subjectkeys to filter the dataframe by}

\item{site}{Optional; either a single site value or a vector of site values to filter the dataframe by (e.g., Yale, NU)}

\item{subsiteid}{Optional; either a single subsiteid or a vector of subsiteids to filter the dataframe by}

\item{sex}{Optional; either a single sex value or a vector of sex values at birth to filter the dataframe by (e.g., 'M', 'F')}

\item{race}{Optional; either a single race value or a vector of race values to filter the dataframe by}

\item{ethnic_group}{Optional; either a single ethnic_group value or a vector of ethnic_group values to filter the dataframe by}

\item{phenotype}{Optional; either a single phenotype value or a vector of phenotype values to filter the dataframe by}

\item{phenotype_description}{Optional; either a single phenotype_description or a vector of phenotype_descriptions to filter the dataframe by}

\item{status}{Optional; either a single status string or a vector of status conditions to filter the dataframe by. Used
if either 'state' or 'status' column exists in the dataframe. Can include values like 'complete',
'completed baseline', 'completed 12m', 'completed 24m', etc.}

\item{lost_to_followup}{Optional; either a single value or a vector of values to filter the dataframe by (checks both 'lost_to_followup' and 'lost_to_follow-up' columns)}

\item{twins_study}{Optional; either a single twins_study value or a vector of twins_study values to filter the dataframe by}

\item{sibling_study}{Optional; either a single sibling_study value or a vector of sibling_study values to filter the dataframe by}

\item{family_study}{Optional; either a single family_study value or a vector of family_study values to filter the dataframe by}

\item{sample_taken}{Optional; either a single sample_taken value or a vector of sample_taken values to filter the dataframe by}

\item{visit}{Optional; either a single visit value or a vector of visit values to filter the dataframe by.
Only used if 'visit' column exists in the dataframe.}

\item{week}{Optional; either a single week value or a vector of week values to filter the dataframe by.
Only used if 'week' column exists in the dataframe.}

\item{arm}{Optional; either a single arm value or a vector of arm values to filter the dataframe by (e.g., drug, placebo)}

\item{interview_date}{Optional; can be either:
- A date string in various formats (ISO, US, etc.) to filter data up to that date
- A boolean TRUE to return only rows with non-NA interview_date values}
}
\value{
A filtered dataframe based on the provided parameters,
and containing only the columns specified in 'cols'. If no columns
are specified, returns the entire dataframe with applied row filters.
}
\description{
Filter data frame by superkey parameters, rows, and columns
}
\examples{
# Create a sample dataframe
sample_df <- data.frame(
  record_id = c("REC001", "REC002", "REC003", "REC004"),
  src_subject_id = c("SUB001", "SUB002", "SUB003", "SUB004"),
  subjectkey = c("KEY001", "KEY002", "KEY003", "KEY004"),
  site = c("Yale", "NU", "Yale", "NU"),
  phenotype = c("A", "B", "A", "C"),
  visit = c(1, 2, 2, 1),
  state = c("complete", "completed baseline", "in progress", NA),
  status = c(NA, NA, NA, "complete"),
  lost_to_followup = c(FALSE, FALSE, TRUE, NA),
  interview_date = c("2023-01-15", "2023/02/20", NA, "2023-03-10")
)

# Set row names for demonstration
rownames(sample_df) <- c("foo", "bar", "baz", "qux")

# Filter by specific date
filtered1 <- sift(sample_df, 
                 cols = c("src_subject_id", "phenotype"), 
                 visit = 2, 
                 interview_date = "01/31/2023")
                 
# Filter to include only rows with non-NA interview dates
filtered2 <- sift(sample_df, 
                 interview_date = TRUE)
                 
# Filter by status (works with either state or status column)
filtered3 <- sift(sample_df,
                 status = c("complete", "completed baseline"))
                 
# Filter with specific row names
filtered4 <- sift(sample_df,
                 rows = c("foo", "qux"))
                 
# Filter with vector of visit values
filtered6 <- sift(sample_df,
                 visit = c(1, 2))
                 
# Filter by lost_to_followup
filtered10 <- sift(sample_df,
                 lost_to_followup = FALSE)
                 
# Filter by src_subject_id
filtered11 <- sift(sample_df,
                 src_subject_id = c("SUB001", "SUB004"))
                 
# Multiple filters combined
filtered12 <- sift(sample_df,
                 site = "Yale",
                 visit = 1,
                 cols = c("record_id", "src_subject_id", "site"))
}
